Package: blockr.core
Title: Graphical Web-Framework for Data Manipulation and Visualization
Version: 0.1.3
Authors@R: 
    c(person(given = "Nicolas",
             family = "Bennett",
             role = c("aut", "cre"),
             email = "nicolas@cynkra.com"),
      person(given = "David",
             family = "Granjon",
             role = c("aut"),
             email = "david@cynkra.com"),
      person(given = "Christoph",
             family = "Sax",
             role = c("aut"),
             email = "christoph@cynkra.com"),
      person(family = "Bristol Myers Squibb",
             role = "fnd")
    )
Description: A framework for data manipulation and visualization using a
    web-based point and click user interface where analysis pipelines are decomposed into re-usable and parameterizable blocks.
URL: https://bristolmyerssquibb.github.io/blockr.core/
BugReports: https://github.com/BristolMyersSquibb/blockr.core/issues
License: GPL (>= 3)
Encoding: UTF-8
Imports: shiny (>= 1.5.0), DT, bslib, bsicons, datasets, grDevices,
        graphics, methods, utils, jsonlite, vctrs, generics, rlang,
        htmltools, evaluate, shinyFiles, digest, cli, glue, yaml
Suggests: testthat (>= 3.0.0), memuse, withr, shinytest2, chromote,
        roxy.shinylive, roxygen2, knitr, rmarkdown, quarto, scoutbaR,
        thematic, ids, ellmer
Config/testthat/edition: 3
VignetteBuilder: quarto
Collate: 'block-args.R' 'block-class.R' 'block-eval.R' 'block-meta.R'
        'block-registry.R' 'block-roclet.R' 'block-server.R'
        'block-ui.R' 'blocks-class.R' 'board-class.R' 'board-loader.R'
        'board-lock.R' 'board-option.R' 'board-options.R'
        'board-plugins.R' 'board-server.R' 'board-ui.R' 'utils-dt.R'
        'data-block.R' 'data-dataset.R' 'data-static.R' 'file-block.R'
        'file-browser.R' 'file-upload.R' 'link-class.R' 'links-class.R'
        'parser-block.R' 'parser-csv.R' 'plot-block.R' 'plot-scatter.R'
        'plugin-block.R' 'plugin-blocks.R' 'plugin-code.R'
        'plugin-control.R' 'plugin-links.R' 'plugin-notification.R'
        'plugin-serdes.R' 'plugin-stack.R' 'plugin-stacks.R'
        'reactives.R' 'stack-class.R' 'stack-ui.R' 'stacks-class.R'
        'str-value.R' 'text-block.R' 'text-glue.R' 'transform-block.R'
        'transform-fixed.R' 'transform-head.R' 'transform-merge.R'
        'transform-rbind.R' 'transform-subset.R' 'utils-assertions.R'
        'utils-cnd.R' 'utils-code.R' 'utils-expr.R' 'utils-graph.R'
        'utils-logging.R' 'utils-misc.R' 'utils-pkg.R' 'utils-ply.R'
        'utils-serdes.R' 'utils-serve.R' 'utils-shiny.R'
        'utils-tests.R' 'zzz-onload.R'
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-12 20:07:25 UTC; coder
Author: Nicolas Bennett [aut, cre],
  David Granjon [aut],
  Christoph Sax [aut],
  Bristol Myers Squibb [fnd]
Maintainer: Nicolas Bennett <nicolas@cynkra.com>
Repository: CRAN
Date/Publication: 2026-07-12 21:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("$",links)
S3method("$<-",links)
S3method("[",board_options)
S3method("[",link)
S3method("[",links)
S3method("[",plugins)
S3method("[",stack)
S3method("[<-",blocks)
S3method("[<-",board_options)
S3method("[<-",links)
S3method("[<-",plugins)
S3method("[<-",stack)
S3method("[<-",stacks)
S3method("[[",board_options)
S3method("[[",link)
S3method("[[",links)
S3method("[[",plugins)
S3method("[[",stack)
S3method("[[<-",blocks)
S3method("[[<-",board_options)
S3method("[[<-",links)
S3method("[[<-",plugins)
S3method("[[<-",stack)
S3method("[[<-",stacks)
S3method("names<-",blocks)
S3method("names<-",links)
S3method("names<-",plugins)
S3method("names<-",stacks)
S3method(add_block_to_stack,board)
S3method(anyDuplicated,plugins)
S3method(anyDuplicated,stack)
S3method(as.character,link)
S3method(as.character,stack)
S3method(as.data.frame,link)
S3method(as.data.frame,links)
S3method(as.list,blocks)
S3method(as.list,board_options)
S3method(as.list,link)
S3method(as.list,links)
S3method(as.list,plugin)
S3method(as.list,plugins)
S3method(as.list,stack)
S3method(as.list,stacks)
S3method(as.matrix,board)
S3method(as.matrix,links)
S3method(as_block,block)
S3method(as_blocks,block)
S3method(as_blocks,blocks)
S3method(as_blocks,board)
S3method(as_blocks,list)
S3method(as_board_option,board_option)
S3method(as_board_options,board)
S3method(as_board_options,board_option)
S3method(as_board_options,board_options)
S3method(as_board_options,list)
S3method(as_link,character)
S3method(as_link,data.frame)
S3method(as_link,link)
S3method(as_link,list)
S3method(as_links,"NULL")
S3method(as_links,board)
S3method(as_links,data.frame)
S3method(as_links,link)
S3method(as_links,links)
S3method(as_links,list)
S3method(as_plugin,list)
S3method(as_plugin,plugin)
S3method(as_plugin,plugins)
S3method(as_plugins,list)
S3method(as_plugins,plugin)
S3method(as_plugins,plugins)
S3method(as_stack,character)
S3method(as_stack,list)
S3method(as_stack,stack)
S3method(as_stacks,board)
S3method(as_stacks,character)
S3method(as_stacks,list)
S3method(as_stacks,stack)
S3method(as_stacks,stacks)
S3method(block_eval,block)
S3method(block_eval,plot_block)
S3method(block_eval_trigger,block)
S3method(block_eval_trigger,plot_block)
S3method(block_inputs,block)
S3method(block_inputs,board)
S3method(block_output,block)
S3method(block_output,data_block)
S3method(block_output,parser_block)
S3method(block_output,plot_block)
S3method(block_output,text_block)
S3method(block_output,transform_block)
S3method(block_render_trigger,block)
S3method(block_render_trigger,data_block)
S3method(block_render_trigger,parser_block)
S3method(block_render_trigger,transform_block)
S3method(block_server,block)
S3method(block_summary,block)
S3method(block_ui,block)
S3method(block_ui,board)
S3method(block_ui,data_block)
S3method(block_ui,parser_block)
S3method(block_ui,plot_block)
S3method(block_ui,text_block)
S3method(block_ui,transform_block)
S3method(blockr_app_options,board)
S3method(blockr_app_plugins,board)
S3method(blockr_app_server,board)
S3method(blockr_app_ui,board)
S3method(blockr_deser,block)
S3method(blockr_deser,blockr_ctor)
S3method(blockr_deser,blocks)
S3method(blockr_deser,board)
S3method(blockr_deser,board_option)
S3method(blockr_deser,board_options)
S3method(blockr_deser,link)
S3method(blockr_deser,links)
S3method(blockr_deser,list)
S3method(blockr_deser,stack)
S3method(blockr_deser,stacks)
S3method(blockr_ser,block)
S3method(blockr_ser,blockr_ctor)
S3method(blockr_ser,blocks)
S3method(blockr_ser,board)
S3method(blockr_ser,board_option)
S3method(blockr_ser,board_options)
S3method(blockr_ser,fixed_block)
S3method(blockr_ser,link)
S3method(blockr_ser,links)
S3method(blockr_ser,llm_model_option)
S3method(blockr_ser,stack)
S3method(blockr_ser,stacks)
S3method(blockr_test_exports,board)
S3method(board_options,block)
S3method(board_options,block_registry_entry)
S3method(board_options,board)
S3method(board_options,data_block)
S3method(board_options,parser_block)
S3method(board_options,transform_block)
S3method(board_plugins,board)
S3method(board_server,board)
S3method(board_ui,"NULL")
S3method(board_ui,board)
S3method(board_ui,board_options)
S3method(board_ui,plugin)
S3method(board_ui,plugins)
S3method(c,block)
S3method(c,blocks)
S3method(c,board_option)
S3method(c,board_options)
S3method(c,link)
S3method(c,links)
S3method(c,plugin)
S3method(c,plugins)
S3method(c,stack)
S3method(c,stacks)
S3method(duplicated,stack)
S3method(expr_server,block)
S3method(expr_ui,block)
S3method(format,block)
S3method(format,blocks)
S3method(format,board)
S3method(format,board_option)
S3method(format,board_options)
S3method(format,link)
S3method(format,links)
S3method(format,llm_model_option)
S3method(format,plugin)
S3method(format,plugins)
S3method(format,stack)
S3method(format,stacks)
S3method(insert_block_ui,board)
S3method(insert_stack_ui,board)
S3method(intersect,stack)
S3method(is.element,stack)
S3method(is_acyclic,board)
S3method(is_acyclic,links)
S3method(is_acyclic,matrix)
S3method(modify_board_links,board)
S3method(modify_board_stacks,board)
S3method(names,board_options)
S3method(names,links)
S3method(names,plugins)
S3method(print,block)
S3method(print,blocks)
S3method(print,board)
S3method(print,board_option)
S3method(print,board_options)
S3method(print,link)
S3method(print,plugin)
S3method(print,plugins)
S3method(print,stack)
S3method(print,stacks)
S3method(remove_block_from_stack,board)
S3method(remove_block_ui,board)
S3method(remove_stack_ui,board)
S3method(replay,error)
S3method(replay,message)
S3method(replay,warning)
S3method(restore_board,board)
S3method(rm_blocks,board)
S3method(serialize_board,board)
S3method(serve,block)
S3method(serve,board)
S3method(serve,character)
S3method(setdiff,stack)
S3method(setequal,stack)
S3method(show_condition,default)
S3method(show_condition,error)
S3method(show_condition,list)
S3method(show_condition,message)
S3method(show_condition,warning)
S3method(sort,board)
S3method(stack_ui,board)
S3method(stack_ui,stack)
S3method(toolbar_ui,board)
S3method(union,stack)
S3method(validate_board,board)
S3method(validate_board,default)
S3method(validate_board_option,board_name_option)
S3method(validate_board_option,dark_mode_option)
S3method(validate_board_option,default)
S3method(validate_board_option,filter_rows_option)
S3method(validate_board_option,llm_model_option)
S3method(validate_board_option,n_rows_option)
S3method(validate_board_option,page_size_option)
S3method(validate_board_option,show_conditions_option)
S3method(validate_board_option,thematic_option)
S3method(validate_stack,default)
S3method(validate_stack,stack)
S3method(vec_restore,blocks)
S3method(vec_restore,links)
S3method(vec_restore,plugins)
S3method(vec_restore,stack)
S3method(vec_restore,stacks)
export("block_name<-")
export("board_blocks<-")
export("board_links<-")
export("board_options<-")
export("board_stacks<-")
export("stack_blocks<-")
export("stack_name<-")
export(.)
export(..)
export(abort_not_null)
export(add_block_to_stack)
export(adjective_animal)
export(as_block)
export(as_blocks)
export(as_board_option)
export(as_board_options)
export(as_link)
export(as_links)
export(as_log_level)
export(as_plugin)
export(as_plugins)
export(as_stack)
export(as_stacks)
export(available_blocks)
export(available_stack_blocks)
export(bbquote)
export(block_arity)
export(block_eval)
export(block_eval_trigger)
export(block_inputs)
export(block_metadata)
export(block_name)
export(block_output)
export(block_render_trigger)
export(block_server)
export(block_summary)
export(block_ui)
export(blockr_abort)
export(blockr_app_options)
export(blockr_app_plugins)
export(blockr_app_server)
export(blockr_app_ui)
export(blockr_deser)
export(blockr_inform)
export(blockr_option)
export(blockr_ser)
export(blockr_test_exports)
export(blockr_warn)
export(blocks)
export(board_block_ids)
export(board_blocks)
export(board_ctor)
export(board_link_ids)
export(board_links)
export(board_option_category)
export(board_option_ctor)
export(board_option_default)
export(board_option_id)
export(board_option_ids)
export(board_option_server)
export(board_option_transform)
export(board_option_trigger)
export(board_option_ui)
export(board_option_value)
export(board_option_values)
export(board_options)
export(board_plugins)
export(board_server)
export(board_stack_ids)
export(board_stacks)
export(board_ui)
export(cat_logger)
export(chr_mply)
export(chr_ply)
export(chr_xtr)
export(clear_board)
export(cnd_logger)
export(coal)
export(combine_board_options)
export(create_block)
export(ctor_fun)
export(ctor_name)
export(ctor_pkg)
export(ctrl_block)
export(ctrl_block_server)
export(ctrl_block_ui)
export(custom_options)
export(custom_plugins)
export(dbl_mply)
export(dbl_ply)
export(dbl_xtr)
export(debug_log_level)
export(default_block_name)
export(default_board_options)
export(default_category)
export(default_icon)
export(default_stack_name)
export(destroy_module)
export(edit_block)
export(edit_block_server)
export(edit_block_ui)
export(edit_stack)
export(edit_stack_server)
export(edit_stack_ui)
export(error_log_level)
export(eval_env)
export(export_code)
export(export_safely)
export(expr_server)
export(expr_ui)
export(fatal_log_level)
export(filebrowser_volumes)
export(forward_ctor)
export(generate_code)
export(generate_code_server)
export(generate_code_ui)
export(generate_plugin_args)
export(get_board_option_or_default)
export(get_board_option_or_null)
export(get_board_option_value)
export(get_board_option_values)
export(get_log_level)
export(get_s3_method)
export(get_serve_obj)
export(get_session)
export(has_length)
export(id_to_sentence_case)
export(include_mermaid)
export(info_log_level)
export(insert_block_ui)
export(insert_stack_ui)
export(int_mply)
export(int_ply)
export(int_xtr)
export(is_acyclic)
export(is_block)
export(is_blockr_ctor)
export(is_blocks)
export(is_board)
export(is_board_option)
export(is_board_options)
export(is_bool)
export(is_count)
export(is_intish)
export(is_link)
export(is_links)
export(is_number)
export(is_plugin)
export(is_plugins)
export(is_scalar)
export(is_stack)
export(is_stacks)
export(is_string)
export(lgl_mply)
export(lgl_ply)
export(lgl_xtr)
export(links)
export(list_blocks)
export(log_debug)
export(log_error)
export(log_fatal)
export(log_info)
export(log_trace)
export(log_warn)
export(lst_xtr)
export(manage_blocks)
export(manage_blocks_server)
export(manage_blocks_ui)
export(manage_links)
export(manage_links_server)
export(manage_links_ui)
export(manage_stacks)
export(manage_stacks_server)
export(manage_stacks_ui)
export(map)
export(modify_board_links)
export(modify_board_stacks)
export(new_block)
export(new_board)
export(new_board_name_option)
export(new_board_option)
export(new_board_options)
export(new_csv_block)
export(new_dark_mode_option)
export(new_data_block)
export(new_dataset_block)
export(new_file_block)
export(new_filebrowser_block)
export(new_filter_rows_option)
export(new_fixed_block)
export(new_glue_block)
export(new_head_block)
export(new_link)
export(new_llm_model_option)
export(new_merge_block)
export(new_mock_session)
export(new_n_rows_option)
export(new_page_size_option)
export(new_parser_block)
export(new_plot_block)
export(new_plugin)
export(new_rbind_block)
export(new_scatter_block)
export(new_show_conditions_option)
export(new_stack)
export(new_static_block)
export(new_subset_block)
export(new_text_block)
export(new_thematic_option)
export(new_transform_block)
export(new_upload_block)
export(not_null)
export(notify)
export(notify_user)
export(notify_user_server)
export(notify_user_ui)
export(paste_enum)
export(pkg_avail)
export(pkg_file)
export(pkg_name)
export(pkg_version)
export(plugin_id)
export(plugin_server)
export(plugin_ui)
export(plugin_validator)
export(plugins)
export(preserve_board)
export(preserve_board_server)
export(preserve_board_ui)
export(rand_names)
export(register_block)
export(register_blocks)
export(registry_id_from_block)
export(registry_metadata)
export(remove_block_from_stack)
export(remove_block_ui)
export(remove_stack_ui)
export(resolve_ctor)
export(restore_board)
export(reval)
export(reval_if)
export(rm_blocks)
export(sample_letters)
export(serialize_board)
export(serve)
export(set_blockr_options)
export(set_board_option_value)
export(set_names)
export(sink_msg)
export(stack_blocks)
export(stack_name)
export(stack_ui)
export(stacks)
export(suggested_categories)
export(to_sentence_case)
export(toolbar_ui)
export(topo_sort)
export(trace_log_level)
export(unlst)
export(unregister_blocks)
export(validate_board)
export(validate_board_option)
export(validate_board_options)
export(validate_data_inputs)
export(validate_links)
export(validate_plugins)
export(validate_stack)
export(warn_log_level)
export(with_mock_context)
export(with_mock_session)
export(write_log)
import(shiny)
import(vctrs)
importFrom(cli,qty)
importFrom(evaluate,replay)
importFrom(generics,intersect)
importFrom(generics,is.element)
importFrom(generics,setdiff)
importFrom(generics,setequal)
importFrom(generics,union)
importFrom(grDevices,col2rgb)
importFrom(graphics,plot)
importFrom(methods,is)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# blockr.core

<!-- badges: start -->

[![cran](https://www.r-pkg.org/badges/version/blockr.core)](https://CRAN.R-project.org/package=blockr.core)
[![lifecycle](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![status](https://github.com/BristolMyersSquibb/blockr.core/actions/workflows/ci.yaml/badge.svg)](https://github.com/BristolMyersSquibb/blockr.core/actions/workflows/ci.yaml)
[![coverage](https://codecov.io/gh/BristolMyersSquibb/blockr.core/graph/badge.svg?token=VoOPRU65KA)](https://app.codecov.io/gh/BristolMyersSquibb/blockr.core)
<!-- badges: end -->

## Overview

blockr is a framework for **data analysis**, using a **web-based** point
and click user interface. Use blockr to create web apps that perform
**visual programming**, leveraging the power of **R**.

## Installation

You can install the development version of blockr.core from
[GitHub](https://github.com/) with:

``` r
pak::pak("BristolMyersSquibb/blockr.core")
```

## What is blockr.core?

`blockr.core` is the entry point to blockr, providing the core
functionality needed to get started. It can be run as a stand-alone app,
but is designed with extensibility in mind. A rich **API** is provided
to connect your own blocks and override the default behaviors (e.g., you
can add your own GUI layer or serialization).

## What is a block?

A **block** is a fundamental unit of a data analysis workflow. A single
block performs a single **task**, such as loading or filtering or
visualization. Blocks can be joined together using a **DAG** (direct
acyclic graph) to create powerful data **workflows**.

## Who is it for?

blockr targets two main user groups:

- **non developers** are empowered to create insightful data workflows
  using pre-built blocks that can be **easily** connected, all without
  writing a single line of code.
- **developers** are provided with a set of tools to create new blocks
  and extend blockr, thereby enhancing the entire framework and
  fostering **collaboration** within teams and organisations.

## Key features

1.  **User-Friendly Interface**: Build data pipelines with intuitive
    interface, like with
    [blockr.dock](https://github.com/BristolMyersSquibb/blockr.dock).
2.  **Flexibility**: Easily add, remove, or rearrange blocks in your
    pipeline.
3.  **Extensibility**: Developers can create custom blocks to extend
    functionality
    ([blockr.dplyr](https://github.com/BristolMyersSquibb/blockr.dplyr),
    [blockr.ai](https://github.com/BristolMyersSquibb/blockr.ai),
    [blockr.io](https://github.com/BristolMyersSquibb/blockr.io),
    [blockr.sdtm](https://github.com/BristolMyersSquibb/blockr.sdtm), …)
4.  **Reproducibility**: Pipelines created with `blockr.core` are easily
    shareable and reproducible, with exportable code.
5.  **Interactivity**: Real-time feedback as you build and modify your
    pipeline.

## Getting started

To get started, we invite you to read this
[vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/get-started.html).

## Extending blockr

We started to build blockr extensions to enrich blockr.core’s
capabilities:

- [blockr.dplyr](https://github.com/BristolMyersSquibb/blockr.dplyr): an
  entry point to the `dplyr` ecosystem.
- [blockr.ai](https://github.com/BristolMyersSquibb/blockr.ai):
  LLM-powered blocks.
- [blockr.io](https://github.com/BristolMyersSquibb/blockr.io): handle
  different data input format (xpt, csv, …) and output for export.
- [blockr.sdtm](https://github.com/BristolMyersSquibb/blockr.sdtm):
  pharmaverse datasets blocks.

If you wish to learn to develop your own blocks, have a look at this
[vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/create-block.html).

library(blockr.core)

serve(
  new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("ChickWeight"),
      c = new_merge_block("Time")
    ),
    links = c(
      ac = new_link("a", "c", "x"),
      bc = new_link("b", "c", "y")
    ),
    stacks = list(ac = c("a", "c"))
  ),
  "my_board"
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----setup--------------------------------------------------------------------
library(blockr.core)


## ----empty-board, eval = FALSE------------------------------------------------
# serve(
#   new_board()
# )


## ----empty-board-shinylive_url, echo = FALSE, results = 'asis'----------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "webr::install(\"blockr.dplyr\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(blockr.dplyr)",
    "library(bsicons)",
    knitr::knit_code$get("empty-board")
  ),
  collapse = "\n"
)

empty_board_url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", empty_board_url))


## ----empty-board-demo, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()----
knitr::include_url(empty_board_url, height = "800px")


## ----echo=FALSE---------------------------------------------------------------
include_mermaid("board-dag")


## ----eval = FALSE-------------------------------------------------------------
# serve(new_dataset_block("iris"))


## ----eval = FALSE-------------------------------------------------------------
# serve(
#   new_merge_block(by = "name"),
#   data = list(x = datasets::BOD, y = datasets::ChickWeight)
# )


## ----eval = FALSE-------------------------------------------------------------
# serve(new_board())


## ----simple-board, eval = FALSE-----------------------------------------------
# serve(
#   new_board(
#     blocks = c(
#       a = new_dataset_block("BOD"),
#       b = new_dataset_block("ChickWeight"),
#       c = new_merge_block("Time")
#     ),
#     links = c(
#       ac = new_link("a", "c", "x"),
#       bc = new_link("b", "c", "y")
#     ),
#     stacks = list(ac = c("a", "c"))
#   )
# )


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "library(blockr.core)",
    "library(bsicons)",
    knitr::knit_code$get("simple-board")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", url))


## ----shinylive_iframe, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()----
knitr::include_url(url, height = "800px")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----packages, message=FALSE, warning=FALSE-----------------------------------
library(shiny)
library(bslib)
library(scoutbaR)
library(blockr.core)


## ----custom-plugin-setup, eval = FALSE, echo=FALSE----------------------------
# chr_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
#   vapply(x, fun, character(length), ..., USE.NAMES = use_names)
# }
# 
# #' @keywords internal
# lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
#   vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
# }
# 
# dropNulls <- function(x) {
#   x[!lgl_ply(x, is.null)]
# }
# 
# blk_icon <- function(category) {
#   switch(
#     category,
#     "data" = "table",
#     "file" = "file-import",
#     "parse" = "cogs",
#     "plot" = "chart-line",
#     "transform" = "wand-magic-sparkles",
#     "table" = "table"
#   )
# }
# 
# blk_choices <- function() {
#   blk_cats <- sort(
#     unique(chr_ply(available_blocks(), \(b) attr(b, "category")))
#   )
# 
#   lapply(blk_cats, \(cat) {
#     scout_section(
#       label = cat,
#       .list = dropNulls(
#         unname(
#           lapply(available_blocks(), \(choice) {
#             if (attr(choice, "category") == cat) {
#               scout_action(
#                 id = attr(choice, "classes")[1],
#                 label = attr(choice, "name"),
#                 description = attr(choice, "description"),
#                 icon = blk_icon(cat)
#               )
#             }
#           })
#         )
#       )
#     )
#   })
# }


## ----custom-plugin-server, eval=FALSE-----------------------------------------
# new_manage_blocks_server <- function(id, board, update, ...) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       # Trigger add block
#       observeEvent(
#         input$add_block,
#         {
#           update_scoutbar(
#             session,
#             "scoutbar",
#             revealScoutbar = TRUE
#           )
#         }
#       )
# 
#       observeEvent(input$scoutbar, {
#         new_blk <- as_blocks(create_block(input$scoutbar))
#         update(
#           list(blocks = list(add = new_blk))
#         )
#       })
# 
#       NULL
#     }
#   )
# }


## ----custom-plugin-app, eval=FALSE--------------------------------------------
#| code-fold: true
# main_ui <- function(id, board) {
#   ns <- NS(id)
#   board_ui(
#     ns("board"),
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     )
#   )
# }
# 
# main_server <- function(id, board) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       ns <- session$ns
# 
#       # Board module
#       board_server(
#         "board",
#         board,
#         plugins = plugins(
#           manage_blocks(server = new_manage_blocks_server, ui = new_manage_blocks_ui)
#         ),
#         callbacks = list()
#       )
#     }
#   )
# }
# 
# board <- new_board()
# 
# ui <- page_fluid(
#   main_ui("app", board)
# )
# 
# server <- function(input, output, session) {
#   main_server("app", board)
# }
# 
# shinyApp(ui, server)


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    knitr::knit_code$get("packages"),
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-plugin-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive_iframe, echo = FALSE, eval = TRUE------------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)


## ----custom-board-ui, eval=FALSE----------------------------------------------
# board_ui.custom_board <- function(id, x, plugins = list(), ...) {
#   plugins <- as_plugins(plugins)
#   div(
#     id = paste0(id, "_board"),
#     board_ui(id, plugins[["manage_blocks"]], x),
#     div(
#       id =  paste0(id, "_blocks"),
#       block_ui(id, x)
#     )
#   )
# }


## ----custom-block-ui, eval=FALSE----------------------------------------------
# get_block_registry <- function(x) {
#   stopifnot(is_block(x))
#   available_blocks()[[strsplit(attr(x, "ctor"), "new_")[[1]][2]]]
# }
# 
# block_ui.custom_board <- function(id, x, blocks = NULL, ...) {
#   block_card <- function(x, id, ns) {
#     id <- paste0("block_", id)
# 
#     blk_info <- get_block_registry(x)
# 
#     div(
#       class = "m-2",
#       id = ns(id),
#       shinyNextUI::card(
#         variant = "bordered",
#         shinyNextUI::card_header(
#           className = "d-flex justify-content-between",
#           icon(blk_icon(attr(blk_info, "category"))),
#           sprintf(
#             "Block: %s (id: %s)",
#             attr(blk_info, "name"),
#             gsub("block_", "", id)
#           ),
#           shinyNextUI::tooltip(
#             icon("info-circle"),
#             content = tagList(
#               p(
#               icon("lightbulb"),
#               "How to use this block?",
#               ),
#               p(attr(blk_info, "description"), ".")
#             )
#           )
#         ),
#         shinyNextUI::divider(),
#         shinyNextUI::card_body(
#           expr_ui(ns(id), x),
#           block_ui(ns(id), x)
#         ),
#         shinyNextUI::divider(),
#         shinyNextUI::card_footer(
#           sprintf(
#             "Type: %s; Package: %s",
#             attr(blk_info, "category"),
#             attr(blk_info, "package")
#           )
#         )
#       )
#     )
#   }
# 
#   stopifnot(is.character(id) && length(id) == 1L)
# 
#   if (is.null(blocks)) {
#     blocks <- board_blocks(x)
#   } else if (is.character(blocks)) {
#     blocks <- board_blocks(x)[blocks]
#   }
# 
#   stopifnot(is_blocks(blocks))
# 
#   tagList(
#     Map(
#       block_card,
#       blocks,
#       names(blocks),
#       MoreArgs = list(ns = NS(id)),
#       USE.NAMES = FALSE
#     )
#   )
# }


## ----custom-plugin-ui-nextui, eval=FALSE--------------------------------------
# add_rm_block_ui <- function(id, board) {
#   tagList(
#     scoutbar(
#       NS(id, "scoutbar"),
#       placeholder = "Search for a block",
#       actions = blk_choices(),
#       theme = "dark",
#       showRecentSearch = TRUE
#     ),
#     shinyNextUI::actionButton(
#       NS(id, "add_block"),
#       "New block",
#       icon = icon("circle-plus"),
#     )
#   )
# }


## ----custom-ui-app, eval=FALSE------------------------------------------------
#| code-fold: true
# board <- new_board(class = "custom_board")
# 
# ui <- nextui_page(
#   board_ui(
#     "board",
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   board_server(
#     "board",
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     ),
#     callbacks = list()
#   )
# }
# 
# shinyApp(ui, server)


## ----shinylive2_url, echo = FALSE, results = 'asis'---------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(shiny)",
    "library(scoutbaR)",
    "library(blockr.core)",
    "library(shinyNextUI)",
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui-nextui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-block-ui"),
    knitr::knit_code$get("custom-board-ui"),
    knitr::knit_code$get("custom-ui-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive2_iframe, echo = FALSE, eval = TRUE-----------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----setup--------------------------------------------------------------------
library(blockr.core)


## ----echo=FALSE---------------------------------------------------------------
include_mermaid("block-anatomy")


## ----eval = FALSE-------------------------------------------------------------
# ui <- function(id) {
#   tagList(
#     # Wrap widgets in `tagList()`
#     textInput(
#       NS(id, "my_input") # Use `shiny::NS()` to construct namespaces
#     )
#   )
# }


## ----echo=FALSE---------------------------------------------------------------
include_mermaid("block-inputs")


## ----eval = FALSE-------------------------------------------------------------
# server <- function(id, data) {
#   moduleServer(id, function(input, output, session) {
#     # Reactive logic goes here
# 
#     # Return a list with "expr" and "state"
#     list(
#       expr = reactive(quote(identity(data))),
#       state = list(
#         input_one = reactive(input_one()),
#         input_two = reactive(input_two())
#       )
#     )
#   })
# }


## ----eval = FALSE-------------------------------------------------------------
# example_constructor <- function(ui_state = character(), ...) {
#   ui <- function(id) {
#     tagList(textInput(NS(id, "ui_state")))
#   }
# 
#   server <- function(id, data) {
#     moduleServer(id, function(input, output, session) {
#       # Reactive logic goes here
# 
#       # Return a list with "expr" and "state"
#       list(
#         expr = reactive(quote(identity(data))),
#         state = list(
#           # name must match what is defined in the constructor signature
#           ui_state = <STATE_VALUE>
#         )
#       )
#     })
#   }
# 
#   # Return call to `new_block()`
#   new_block(
#     server = server,
#     ui = ui,
#     class = "my_block",
#     ...
#   )
# }


## -----------------------------------------------------------------------------
new_head_block <- function(n = 6L, ...) {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          n_rows <- reactiveVal(n)

          observeEvent(input$n, n_rows(input$n))

          observeEvent(
            nrow(data()),
            updateNumericInput(
              inputId = "n",
              value = n_rows(),
              min = 1L,
              max = nrow(data())
            )
          )

          list(
            expr = reactive(
              bquote(utils::head(data, n = .(n)), list(n = n_rows()))
            ),
            state = list(
              n = n_rows
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        numericInput(
          inputId = NS(id, "n"),
          label = "Number of rows",
          value = n,
          min = 1L
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data) || is.matrix(data))
    },
    class = "head_block",
    ...
  )
}


## ----eval = FALSE-------------------------------------------------------------
# serve(new_head_block(n = 10L), list(data = mtcars))


## ----eval = FALSE-------------------------------------------------------------
# serve(
#   new_merge_block(by = "Time"),
#   data = list(x = datasets::BOD, y = datasets::ChickWeight)
# )


## ----esquisse-server, eval=FALSE----------------------------------------------
# esquisse_block_server <- function(id, data) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       results <- esquisse::esquisse_server(
#         id = "esquisse",
#         data_rv = data
#       )
#       list(
#         expr = reactive({
#           bquote(
#             list(
#               dat = as.data.frame(.(dat)),
#               filters = .(filters)
#             ),
#             list(
#               filters = results$code_plot,
#               dat = results$data
#             )
#           )
#         }),
#         state = list()
#       )
#     }
#   )
# }


## ----esquisse-complex-block, eval=FALSE---------------------------------------
# new_complex_block <- function(server, ui, class, ctor = sys.parent(), ...) {
#   new_block(server, ui, c(class, "complex_block"), ctor, ...)
# }


## ----esquisse-block-output, eval=FALSE----------------------------------------
# #' @export
# block_output.complex_block <- function(x, result, session) {
#   session$output$filters <- renderPrint(result$filters)
#   dt_result(result$dat, x, session)
# }


## ----esquisse-block-ui, eval=FALSE--------------------------------------------
# #' @export
# block_ui.complex_block <- function(id, x, ...) {
#   tagList(
#     h1("Transformed data from {esquisse}"),
#     DT::dataTableOutput(NS(id, "result")),
#     verbatimTextOutput(NS(id, "filters"))
#   )
# }


## ----esquisse-ui, eval=FALSE--------------------------------------------------
# esquisse_block_ui <- function(id) {
#   tagList(
#     esquisse::esquisse_ui(
#       id = NS(id, "esquisse"),
#       header = FALSE # dont display gadget title
#     )
#   )
# }


## ----esquisse-ctor, eval=FALSE------------------------------------------------
# new_esquisse_block <- function(...) {
#   new_complex_block(
#     server = esquisse_block_server,
#     ui = esquisse_block_ui,
#     class = "esquisse_block",
#     dat_valid = NULL,
#     allow_empty_state = TRUE,
#     ...
#   )
# }


## ----esquisse-app, eval=FALSE-------------------------------------------------
# serve(
#   new_board(
#     blocks = list(
#       a = new_dataset_block(iris),
#       b = new_esquisse_block()
#     ),
#     links = list(
#       new_link("a", "b", "data")
#     )
#   )
# )


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(esquisse)",
    "library(plotly)",
    knitr::knit_code$get("esquisse-server"),
    knitr::knit_code$get("esquisse-ui"),
    knitr::knit_code$get("esquisse-complex-block"),
    knitr::knit_code$get("esquisse-ctor"),
    knitr::knit_code$get("esquisse-block-ui"),
    knitr::knit_code$get("esquisse-block-output"),
    knitr::knit_code$get("esquisse-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive_iframe, echo = FALSE, eval = TRUE------------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)

test_that("head block constructor", {

  blk <- new_head_block()

  expect_s3_class(blk, "head_block")

  testServer(
    block_expr_server(blk),
    {
      expect_equal(nrw(), 6L)
      expect_equal(til(), FALSE)

      expect_equal(session$returned$state$n(), 6L)
      expect_equal(session$returned$state$direction(), "head")

      session$setInputs(n = 10L, tail = TRUE)

      session$flushReact()

      expect_equal(session$returned$state$n(), 10L)
      expect_equal(session$returned$state$direction(), "tail")
    },
    args = list(data = mtcars)
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        head(datasets::mtcars)
      )
    },
    args = list(
      x = blk,
      data = list(data = function() datasets::mtcars)
    )
  )
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----setup--------------------------------------------------------------------
library(blockr.core)


## flowchart LR
##   subgraph block[Block]
##     subgraph ctor[constructor]
##       block_ui[UI]
##       subgraph block_server[server]
##         blk_state[State]
##         blk_expr[Expression]
##       end
##     end
##   end

## ----eval = FALSE-------------------------------------------------------------
# ui <- function(id) {
#   tagList(
#     # Wrap widgets in `tagList()`
#     textInput(
#       NS(id, "my_input") # Use `shiny::NS()` to construct namespaces
#     )
#   )
# }


## flowchart TB
##   data_blk[data block 1]
##   data_blk_2[data block 2]
##   data_blk_3[data block 3]
##   data_blk_4[data block 4]
##   select_blk[select block]
##   join_blk[join block]
##   rbind_blk[rbind block]
##   data_blk --> |data| select_blk
##   data_blk_2 -->|data1| join_blk
##   data_blk_3 --> |data2| join_blk
## 
##   data_blk --> |1| rbind_blk
##   select_blk --> |2| rbind_blk
##   data_blk_4 --> |3| rbind_blk
##   join_blk --> |4| rbind_blk

## ----eval = FALSE-------------------------------------------------------------
# server <- function(id, data) {
#   moduleServer(id, function(input, output, session) {
#     # Reactive logic goes here
# 
#     # Return a list with "expr" and "state"
#     list(
#       expr = reactive(quote(identity(data))),
#       state = list(
#         input_one = reactive(input_one()),
#         input_two = reactive(input_two())
#       )
#     )
#   })
# }


## ----eval = FALSE-------------------------------------------------------------
# example_constructor <- function(ui_state = character(), ...) {
#   ui <- function(id) {
#     tagList(textInput(NS(id, "ui_state")))
#   }
# 
#   server <- function(id, data) {
#     moduleServer(id, function(input, output, session) {
#       # Reactive logic goes here
# 
#       # Return a list with "expr" and "state"
#       list(
#         expr = reactive(quote(identity(data))),
#         state = list(
#           # name must match what is defined in the constructor signature
#           ui_state = <STATE_VALUE>
#         )
#       )
#     })
#   }
# 
#   # Return call to `new_block()`
#   new_block(
#     server = server,
#     ui = ui,
#     class = "my_block",
#     ...
#   )
# }


## -----------------------------------------------------------------------------
new_head_block <- function(n = 6L, ...) {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          n_rows <- reactiveVal(n)

          observeEvent(input$n, n_rows(input$n))

          observeEvent(
            nrow(data()),
            updateNumericInput(
              inputId = "n",
              value = n_rows(),
              min = 1L,
              max = nrow(data())
            )
          )

          list(
            expr = reactive(
              bquote(utils::head(data, n = .(n)), list(n = n_rows()))
            ),
            state = list(
              n = n_rows
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        numericInput(
          inputId = NS(id, "n"),
          label = "Number of rows",
          value = n,
          min = 1L
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data) || is.matrix(data))
    },
    class = "head_block",
    ...
  )
}


## ----eval = FALSE-------------------------------------------------------------
# serve(new_head_block(n = 10L), list(data = mtcars))


## ----eval = FALSE-------------------------------------------------------------
# serve(
#   new_merge_block(by = "Time"),
#   data = list(x = datasets::BOD, y = datasets::ChickWeight)
# )


## ----esquisse-server, eval=FALSE----------------------------------------------
# esquisse_block_server <- function(id, data) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       results <- esquisse::esquisse_server(
#         id = "esquisse",
#         data_rv = data
#       )
#       list(
#         expr = reactive({
#           bquote(
#             list(
#               dat = as.data.frame(.(dat)),
#               filters = .(filters)
#             ),
#             list(
#               filters = results$code_plot,
#               dat = results$data
#             )
#           )
#         }),
#         state = list()
#       )
#     }
#   )
# }


## ----esquisse-complex-block, eval=FALSE---------------------------------------
# new_complex_block <- function(server, ui, class, ctor = sys.parent(), ...) {
#   new_block(server, ui, c(class, "complex_block"), ctor, ...)
# }


## ----esquisse-block-output, eval=FALSE----------------------------------------
# #' @export
# block_output.complex_block <- function(x, result, session) {
#   session$output$filters <- renderPrint(result$filters)
#   # result must come at the end of the output list if you have multiple outputs
#   session$output$result <- dt_result(result$dat, session)
# }


## ----esquisse-block-ui, eval=FALSE--------------------------------------------
# #' @export
# block_ui.complex_block <- function(id, x, ...) {
#   tagList(
#     h1("Transformed data from {esquisse}"),
#     DT::dataTableOutput(NS(id, "result")),
#     verbatimTextOutput(NS(id, "filters"))
#   )
# }


## ----esquisse-ui, eval=FALSE--------------------------------------------------
# esquisse_block_ui <- function(id) {
#   tagList(
#     esquisse::esquisse_ui(
#       id = NS(id, "esquisse"),
#       header = FALSE # dont display gadget title
#     )
#   )
# }


## ----esquisse-ctor, eval=FALSE------------------------------------------------
# new_esquisse_block <- function(...) {
#   new_complex_block(
#     server = esquisse_block_server,
#     ui = esquisse_block_ui,
#     class = "esquisse_block",
#     dat_valid = NULL,
#     allow_empty_state = TRUE,
#     ...
#   )
# }


## ----esquisse-app, eval=FALSE-------------------------------------------------
# serve(
#   new_board(
#     blocks = list(
#       a = new_dataset_block(iris),
#       b = new_esquisse_block()
#     ),
#     links = list(
#       new_link("a", "b", "data")
#     )
#   )
# )


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(esquisse)",
    "library(plotly)",
    knitr::knit_code$get("esquisse-server"),
    knitr::knit_code$get("esquisse-ui"),
    knitr::knit_code$get("esquisse-complex-block"),
    knitr::knit_code$get("esquisse-ctor"),
    knitr::knit_code$get("esquisse-block-ui"),
    knitr::knit_code$get("esquisse-block-output"),
    knitr::knit_code$get("esquisse-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive_iframe, echo = FALSE, eval = TRUE------------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----setup--------------------------------------------------------------------
library(blockr.core)


## ----empty-board, eval = FALSE------------------------------------------------
# serve(
#   new_board()
# )


## ----empty-board-shinylive_url, echo = FALSE, results = 'asis'----------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "webr::install(\"blockr.dplyr\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(blockr.dplyr)",
    "library(bsicons)",
    knitr::knit_code$get("empty-board")
  ),
  collapse = "\n"
)

empty_board_url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", empty_board_url))


## ----empty-board-demo, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()----
knitr::include_url(empty_board_url, height = "800px")


## flowchart TD
##   subgraph board[Board]
## 
##     subgraph stack1[Stack 1]
##       direction TB
##       import[Data 1]
##       transform[Transform]
##       visualize[Visualize]
##       import2[Data 2]
##       merge[Merge data]
##       transform2[Subset columns]
##       import --> transform --> |filter| merge --> |left join| visualize
##       import2 --> transform2 --> |subset| merge
##     end
## 
##     subgraph stack2[Stack 2]
##       direction TB
##       import3[Data 3]
##       llm[Transform LLM]
##       visualise_llm[Visualize LLM]
##       import3 --> llm --> visualise_llm
##     end
##   end

## ----eval = FALSE-------------------------------------------------------------
# serve(new_dataset_block("iris"))


## ----eval = FALSE-------------------------------------------------------------
# serve(
#   new_merge_block(by = "name"),
#   data = list(x = datasets::BOD, y = datasets::ChickWeight)
# )


## ----eval = FALSE-------------------------------------------------------------
# serve(new_board())


## ----simple-board, eval = FALSE-----------------------------------------------
# serve(
#   new_board(
#     blocks = c(
#       a = new_dataset_block("BOD"),
#       b = new_dataset_block("ChickWeight"),
#       c = new_merge_block("Time")
#     ),
#     links = c(
#       ac = new_link("a", "c", "x"),
#       bc = new_link("b", "c", "y")
#     ),
#     stacks = list(ac = c("a", "c"))
#   )
# )


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "library(blockr.core)",
    "library(bsicons)",
    knitr::knit_code$get("simple-board")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", url))


## ----shinylive_iframe, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()----
knitr::include_url(url, height = "800px")

#' Block UI
#'
#' The UI associated with a block is created via the generics `expr_ui()` and
#' `block_ui()`. The former is mainly responsible for user inputs that are
#' specific to every block type (i.e. a `subset_block` requires different user
#' inputs compared to a `head_block`, see [new_transform_block()]) and
#' essentially calls the UI function passed as `ui` to [new_block()]. UI that
#' represents block outputs typically is shared among similar block types (i.e.
#' blocks with shared inheritance structure, such as `subset_block` and
#' `head_block`, which both inherit from `transform_block`). This type of UI us
#' created by `block_ui()` and block inheritance is used to deduplicate shared
#' functionality (i.e. by implementing a method for the `transform_block` class
#' only instead of separate methods for `subset_block` and `head_block`.
#' Working in tandem with `block_ui()`, the generic `block_output()` generates
#' the output to be displayed by the UI portion defined via `block_ui()`.
#'
#' The result of `block_output()`, which is evaluated in the [block_server()]
#' context is assigned to `output$result`. Consequently, when referencing
#' the block result in `block_ui()`, this naming convention has to be followed
#' by referring to this as something like `shiny::NS(id, "result")`.
#'
#' @param id Namespace ID
#' @param x Object for which to generate a UI container
#' @param ... Generic consistency
#'
#' @return Both `expr_ui()` and `block_ui()` are expected to return shiny UI
#' (e.g. objects wrapped in a [shiny::tagList()]). For rendering the UI,
#' `block_output()` is required to return the result of a shiny render function.
#' For example, a transform block might show the resulting `data.frame` as an
#' HTML table using the DT package. The corresponding `block_ui()` function
#' would then contain UI created by [DT::dataTableOutput()] and rendering in
#' `block_output()` would then be handled by [DT::renderDT()].
#'
#' @export
block_ui <- function(id, x, ...) {
  UseMethod("block_ui", x)
}

#' @export
block_ui.block <- function(id, x, ...) {
  NULL
}

#' @rdname block_ui
#' @export
expr_ui <- function(id, x, ...) {
  UseMethod("expr_ui", x)
}

#' @export
expr_ui.block <- function(id, x, ...) {

  if (...length()) {
    blockr_abort(
      paste(
        "Unknown arguments", paste_enum(...names()), "in call to `expr_ui()`."
      ),
      class = "superfluous_expr_ui_args"
    )
  }

  do.call(block_expr_ui(x), list(id = NS(id, "expr")))
}

#' @param result Block result
#' @param session Shiny session object
#' @rdname block_ui
#' @export
block_output <- function(x, result, session) {
  UseMethod("block_output")
}

#' @export
block_output.block <- function(x, result, session) {
  NULL
}

test_that("dummy block ui test", {

  new_identity_block <- function() {
    new_block(
      function(id, data) {
        moduleServer(
          "expression",
          function(input, output, session) {
            list(
              expr = reactive(quote(identity(data))),
              state = list()
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "identity_block",
      block_metadata = list()
    )
  }

  expect_null(
    block_ui("identity_block", new_identity_block())
  )

  expect_s3_class(
    expr_ui("identity_block", new_identity_block()),
    "shiny.tag.list"
  )

  expect_error(
    expr_ui("identity_block", new_identity_block(), abc = 1),
    class = "superfluous_expr_ui_args"
  )
})

band_members <- data.frame(
  name = c("Mick", "John", "Paul"),
  band = c("Stones", "Beatles", "Beatles")
)

band_instruments <- data.frame(
  name = c("John", "Paul", "Keith"),
  plays = c("guitar", "bass", "guitar")
)

#' Board UI
#'
#' As counterpart to [board_server()], `board_ui()` is responsible for rendering
#' UI for a board module. This top-level entry point for customizing board
#' appearance and functionality can be overridden by sub-classing the boar
#' object and providing an implementation for this sub-class. Such an
#' implementation is expected to handle UI for plugins and all board
#' components, including blocks, links and stacks, but may rely on
#' functionality that generates UI for these components, such as [block_ui()]
#' or [stack_ui()], as well as already available UI provided by plugins
#' themselves. Additionally, `toolbar_ui()` is responsible for creating a
#' toolbar UI component from several plugin UI components.
#'
#' Dynamic UI updates are handled by functions `insert_block_ui()` and
#' `remove_block_ui()` for adding and removing block-level UI elements to and
#' from `board` UI, whenever blocks are added or removed. These update functions
#' are provided as S3 generics with implementations for `board` and can be
#' extended if so desired.
#'
#' @param id Namespace ID
#' @param x Board
#' @param ... Generic consistency
#'
#' @return A `board_ui()` implementation is expected to return [shiny::tag] or
#' [shiny::tagList()] objects, as does `toolbar_ui()`, while updater functions
#' (`insert_block_ui()` and `remove_block_ui()`) are called for their side
#' effects (which includes UI updates such as [shiny::insertUI()],
#' [shiny::removeUI()]) and return the board object passed as `x` invisibly.
#'
#' @export
board_ui <- function(id, x, ...) {
  UseMethod("board_ui", x)
}

#' @param plugins UI for board plugins
#' @param options Board options (`NULL` defaults to the union of board, block
#' and registry sourced options)
#' @rdname board_ui
#' @export
board_ui.board <- function(id, x, plugins = board_plugins(x), options = NULL,
                           ...) {
  tagList(
    toolbar_ui(id, x, plugins, options),
    board_ui(id, plugins[["notify_user"]], x),
    div(
      id = paste0(id, "_board"),
      stack_ui(id, x, edit_ui = get_plugin("edit_stack", plugins)),
      block_ui(
        id,
        x,
        edit_ui = get_plugin("edit_block", plugins),
        ctrl_ui = get_plugin("ctrl_block", plugins)
      )
    )
  )
}

#' @rdname board_ui
#' @export
board_ui.NULL <- function(id, x, ...) NULL

#' @section Board-level block UI:
#' While the contents of block-level UI are created by dispatching `block_ui()`
#' on blocks another dispatch on `board` (see [new_board()]) occurs as well.
#' This can be used to control how blocks are integrated into the board UI. For
#' the default board, this uses [bslib::card()] to represent blocks. For boards
#' that extend the default `board` class, control is available for how blocks
#' are displayed by providing a board-specific `block_ui()` method.
#'
#' @param blocks (Additional) blocks (or IDs) for which to generate the UI
#' @param edit_ui,ctrl_ui Block plugin UI
#'
#' @rdname block_ui
#' @export
block_ui.board <- function(id, x, blocks = NULL, edit_ui = NULL, ctrl_ui = NULL,
                           ...) {

  block_card <- function(x, block_id, board_ns, card_elems, ctrl) {

    blk_id <- board_ns(paste0("block_", block_id))

    if (!is.null(ctrl)) {
      ctrl_ui <- ctrl(NS(blk_id, "ctrl_block"), x)
    } else {
      ctrl_ui <- NULL
    }

    bslib::card(
      id = paste0(block_id, "_block"),
      card_elems(
        x,
        NS(blk_id, "edit_block"),
        bslib::card_body(
          ctrl_ui,
          expr_ui(blk_id, x),
          block_ui(blk_id, x)
        )
      )
    )
  }

  stopifnot(is_string(id))

  if (is.null(blocks)) {
    blocks <- board_blocks(x)
  } else if (is.character(blocks)) {
    blocks <- board_blocks(x)[blocks]
  }

  stopifnot(is_blocks(blocks))

  if (is.null(edit_ui)) {
    edit_ui <- function(x, id, ...) {
      tagList(
        bslib::card_header(block_name(x)),
        ...
      )
    }
  } else {
    edit_ui <- get_plugin_ui(edit_ui)
  }

  if (!is.null(ctrl_ui)) {
    ctrl_ui <- get_plugin_ui(ctrl_ui)
  }

  args <- list(
    board_ns = NS(id),
    card_elems = edit_ui,
    ctrl = ctrl_ui
  )

  do.call(
    div,
    c(
      id = paste0(id, "_blocks"),
      map(block_card, blocks, names(blocks), MoreArgs = args)
    )
  )
}

#' @param blocks (Additional) blocks (or IDs) for which to generate the UI
#' @param session Shiny session
#' @rdname board_ui
#' @export
insert_block_ui <- function(id, x, blocks = NULL, ...,
                            session = get_session()) {

  UseMethod("insert_block_ui", x)
}

#' @rdname board_ui
#' @export
insert_block_ui.board <- function(id, x, blocks = NULL, ...,
                                  session = get_session()) {

  stopifnot(is_string(id))

  insertUI(
    paste0("#", id, "_blocks"),
    "beforeEnd",
    block_ui(id, x, blocks, ...),
    immediate = TRUE
  )

  invisible(x)
}

#' @rdname board_ui
#' @export
remove_block_ui <- function(id, x, blocks = NULL, ...,
                            session = get_session()) {

  UseMethod("remove_block_ui", x)
}

#' @rdname board_ui
#' @export
remove_block_ui.board <- function(id, x, blocks = NULL, ...,
                                  session = get_session()) {

  if (is.null(blocks)) {

    stopifnot(is_string(id))

    removeUI(
      paste0("#", id, "_blocks > div"),
      multiple = TRUE,
      immediate = TRUE
    )

  } else {

    stopifnot(is.character(blocks))

    for (block in blocks) {
      removeUI(
        paste0("#", block, "_block"),
        immediate = TRUE
      )
    }
  }

  invisible(x)
}

#' @rdname board_ui
#' @export
toolbar_ui <- function(id, x, plugins = list(), ...) {
  UseMethod("toolbar_ui", x)
}

#' @rdname board_ui
#' @export
toolbar_ui.board <- function(id, x, plugins = list(), options = NULL, ...) {

  plugins <- as_plugins(plugins)

  toolbar_plugins <- c("preserve_board", "manage_blocks", "manage_links",
                       "manage_stacks", "generate_code")

  toolbar_plugins <- plugins[intersect(toolbar_plugins, names(plugins))]

  if (is.null(options)) {
    options <- combine_board_options(
      board_options(x),
      lapply(board_blocks(x), board_options),
      lapply(available_blocks(), board_options)
    )
  }

  stopifnot(is_board_options(options))

  toolbar_ui <- do.call(
    tagList,
    list(
      board_ui(id, toolbar_plugins, x),
      bslib::popover(
        bsicons::bs_icon("gear", size = "1.5em"),
        board_ui(id, options),
        title = "Board options"
      )
    )
  )

  do.call(
    div,
    c(
      class = paste(
        "d-flex justify-content-evenly align-items-center",
        "bg-light-subtle sticky-top border rounded-4",
        "m-2 gap-5 p-2"
      ),
      toolbar_ui
    )
  )
}

test_that("dummy board ui test", {

  expect_s3_class(board_ui("board", new_board()), "shiny.tag.list")

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(from = "a", to = "b")
  )

  expect_s3_class(board_ui("board", board), "shiny.tag.list")
})

#' Testing utilities
#'
#' Several utilities for unit testing, mainly with [shiny::testServer()] that
#' have proven themselves useful for testing this package are exported for
#' re-use in other packages.
#'
#' @param board A board object
#' @param mode Edit plugins, such as `manage_blocks` get an additional argument
#' `update` over read plugins such as `preserve_board`.
#'
#' @return For testing plugins, `generate_plugin_args()` returns objects that
#' mimic how plugins are called in the board server (with reactive components
#' re-created on a `NULL` reactive domain so they outlive the
#' [shiny::testServer()] session used to set them up), `sink_msg()` is called
#' mainly for the side-effect of muting shiny messages (and returns them
#' invisibly), `with_mock_session()` returns `NULL` (invisibly) and
#' `with_mock_context()` returns the result of a call to
#' [shiny::withReactiveDomain()]. Finally, `get_s3_method()` returns a
#' class-specific implementation of the specified generic (and throws an error
#' if none is found).
#'
#' @rdname testing
#' @export
generate_plugin_args <- function(board, ..., mode = c("edit", "read")) {

  withr::local_envvar(BLOCKR_LOG_LEVEL = "")
  withr::local_options(blockr.log_level = "warn")

  session <- edit_plugin_args <- read_plugin_args <- NULL

  mode <- match.arg(mode)

  res_plugin_args <- list()

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()
      res_plugin_args <<- snapshot_reactives(
        switch(mode, edit = edit_plugin_args, read = read_plugin_args)
      )
    },
    args = list(x = board, ...)
  )

  res_plugin_args
}

# Re-create the reactive parts of the plugin args on a NULL reactive domain, so
# they survive the testServer() session that built them (which shiny tears down
# on close, taking session-scoped reactives with it; see #196). reactiveVal and
# reactiveExpr both become read-only reactives, since callers only read them.
snapshot_reactives <- function(x) {

  if (is.reactivevalues(x)) {
    vals <- lapply(isolate(reactiveValuesToList(x)), snapshot_reactives)
    return(withReactiveDomain(NULL, do.call(reactiveValues, vals)))
  }

  if (inherits(x, "reactive")) {

    # Defer evaluation errors (e.g. a block with unset inputs) to read time, as
    # the live reactive would have, rather than failing the whole snapshot.
    captured <- tryCatch(
      list(value = isolate(x())),
      error = function(e) list(cond = e)
    )

    if (is.null(captured$cond)) {
      val <- captured$value
      return(withReactiveDomain(NULL, reactive(val)))
    }

    cnd <- captured$cond
    return(withReactiveDomain(NULL, reactive(stop(cnd))))
  }

  if (is.list(x) && is.null(attr(x, "class"))) {
    return(lapply(x, snapshot_reactives))
  }

  x
}

#' @param ... Forwarded to [utils::capture.output()]
#' @rdname testing
#' @export
sink_msg <- function(...) {
  invisible(utils::capture.output(..., type = "message"))
}

#' @rdname testing
#' @export
new_mock_session <- function() MockShinySession$new()

#' @inheritParams shiny::testServer
#' @rdname testing
#' @export
with_mock_session <- function(expr, session = new_mock_session()) {

  empty_module <- function() {
    moduleServer(rand_names(), function(input, output, session) { })
  }

  on.exit(if (!session$isClosed()) session$close())

  quosure <- rlang::enquo(expr)

  with_mock_context(session, empty_module())

  parent_clone <- rlang::env_clone(parent.env(session$env))
  clone <- rlang::env_clone(session$env, parent_clone)
  mask <- rlang::new_data_mask(clone, parent_clone)

  with_mock_context(
    session,
    rlang::eval_tidy(quosure, mask, rlang::caller_env())
  )

  invisible()
}

#' @rdname testing
#' @export
with_mock_context <- function(session, expr) {
  isolate(
    withReactiveDomain(
      session,
      {
        withr::with_options(
          list(shiny.allowoutputreads = TRUE),
          {
            shinyOptions(cache = session$appcache)
            expr
          }
        )
      }
    )
  )
}

#' @param generic Generic function name (passed as string)
#' @param object S3 Object
#' @rdname testing
#' @export
get_s3_method <- function(generic, object) {

  for (cls in class(object)) {
    res <- utils::getS3method(generic, cls, optional = TRUE)
    if (is.function(res)) {
      return(res)
    }
  }

  blockr_abort(
    "No function found for generic `{generic}()` and classes {class(object)}.",
    class = "generic_method_not_found"
  )
}

#' @param x Reactive object to use in [shiny::exportTestValues()]
#' @rdname testing
#' @export
export_safely <- function(x) {
  # https://github.com/rstudio/shiny/issues/3768
  r_quo <- rlang::enquo(x)
  rlang::inject(
    reactive(
      tryCatch(!!r_quo, error = function(e) e)
    )
  )
}

test_that("csv block constructor", {

  file <- withr::local_tempfile()
  utils::write.csv(datasets::BOD, file, row.names = FALSE)

  blk <- new_csv_block()

  expect_s3_class(blk, "csv_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_equal(
        session$returned$result(),
        datasets::BOD,
        ignore_attr = TRUE
      )
    },
    args = list(x = blk, data = list(file = function() file))
  )
})

test_that("scetter plot block constructor", {

  blk <- new_scatter_block()

  expect_s3_class(blk, "scatter_block")

  testServer(
    block_expr_server(blk),
    {
      expect_identical(x_col(), character())
      expect_identical(y_col(), character())

      expect_identical(cols(), colnames(iris))

      expect_identical(session$returned$state$x(), character())
      expect_identical(session$returned$state$y(), character())

      session$setInputs(xcol = "Sepal.Length", ycol = "Sepal.Width")

      expect_identical(x_col(), "Sepal.Length")
      expect_identical(y_col(), "Sepal.Width")

      expect_identical(session$returned$state$x(), "Sepal.Length")
      expect_identical(session$returned$state$y(), "Sepal.Width")
    },
    args = list(data = function() iris)
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      expr <- session$makeScope("expr")
      expr$setInputs(xcol = "Sepal.Length", ycol = "Sepal.Width")
      session$flushReact()

      expect_s3_class(
        session$returned$result(),
        "evaluate_evaluation"
      )
    },
    args = list(x = blk, data = list(data = function() iris))
  )
})

test_that("upload block constructor", {

  file <- withr::local_tempfile(lines = "abc")

  blk <- new_upload_block()

  expect_s3_class(blk, "upload_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$makeScope("expr")$setInputs(upload = list(datapath = file))
      session$flushReact()
      expect_identical(
        session$returned$result(),
        file
      )
    },
    args = list(x = blk)
  )
})

test_that("dataset block constructor", {

  blk <- new_dataset_block("mtcars")

  expect_s3_class(blk, "dataset_block")

  testServer(
    block_expr_server(blk),
    {
      expect_equal(dat(), "mtcars")
      session$setInputs(dataset = "iris")
      expect_equal(dat(), "iris")
    }
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        datasets::mtcars
      )
    },
    args = list(x = blk)
  )
})

test_that("glue block constructor", {

  blk <- new_glue_block()

  expect_s3_class(blk, "glue_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_null(session$returned$result())

      expr <- session$makeScope("expr")
      expr$setInputs(text = "The dataset has {nrow(data)} rows.")

      session$flushReact()

      expect_identical(
        session$returned$result(),
        paste0("The dataset has ", nrow(mtcars), " rows.")
      )
    },
    args = list(
      x = blk,
      data = list(
        ...args = reactives(
          data = function() mtcars
        )
      )
    )
  )
})

test_that("subset block constructor", {

  blk <- new_subset_block()

  expect_s3_class(blk, "subset_block")

  testServer(
    block_expr_server(blk),
    {
      expect_equal(sub(), "")
      expect_equal(sel(), "")

      expect_equal(session$returned$state$subset(), "")
      expect_equal(session$returned$state$select(), "")
    },
    args = list(data = mtcars)
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      expr <- session$makeScope("expr")

      expr$setInputs(
        subset = "",
        select = "c(Sepal.Width, Sepal.Length)",
        eval = 1
      )

      session$flushReact()

      res <- session$returned$result()

      expect_named(res, c("Sepal.Width", "Sepal.Length"))
      expect_identical(nrow(res), nrow(iris))

      expr$setInputs(
        subset = "Sepal.Width > 3",
        select = "",
        eval = 2
      )

      session$flushReact()

      res <- session$returned$result()

      expect_named(res, names(iris))
      expect_identical(nrow(res), nrow(iris[iris$Sepal.Width > 3, ]))

      expr$setInputs(
        subset = "Sepal.Width > 3",
        select = "c(Sepal.Width, Sepal.Length)",
        eval = 3
      )

      session$flushReact()

      res <- session$returned$result()

      expect_named(res, c("Sepal.Width", "Sepal.Length"))
      expect_identical(nrow(res), nrow(iris[iris$Sepal.Width > 3, ]))

      expr$setInputs(
        subset = "",
        select = "",
        eval = 4
      )

      session$flushReact()

      res <- session$returned$result()

      expect_named(res, names(iris))
      expect_identical(nrow(res), nrow(iris))
    },
    args = list(x = blk, data = list(data = function() iris))
  )
})

test_that("block server", {

  blk <- new_dataset_block("iris")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      expect_equal(session$returned$result(), iris)

      session$makeScope("expr")$setInputs(dataset = "mtcars")
      expect_equal(session$returned$result(), mtcars)
    },
    args = list(x = blk, data = list())
  )

  new_identity_block <- function() {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(
              expr = reactive(quote(identity(data))),
              state = list()
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "identity_block",
      block_metadata = list()
    )
  }

  idt <- new_identity_block()

  testServer(
    get_s3_method("block_server", idt),
    {
      session$flushReact()
      expect_equal(session$returned$result(), iris)
    },
    args = list(x = idt, data = list(data = function() iris))
  )
})

test_that("block conditions", {

  new_conds_block <- function() {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(
              expr = reactive(
                quote(
                  {
                    message("hello")
                    warning("oh no")
                    message("world")
                    identity(data)
                  }
                )
              ),
              state = list(),
              cond = reactiveValues(
                message = "message",
                warning = c("warning 1", "warning 2")
              )
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "conds_block",
      block_metadata = list()
    )
  }

  cnd <- new_conds_block()

  withr::with_options(
    list(blockr.show_conditions = c("warning", "error")),
    testServer(
      get_s3_method("block_server", cnd),
      {
        session$flushReact()

        expect_s3_class(cond, "reactivevalues")

        expect_true("block" %in% names(cond))
        expect_named(cond$block, c("warning", "error"), ignore.order = TRUE)

        expect_length(cond$block$warning, 2L)
        expect_length(cond$block$error, 0L)

        for (x in unlst(cond$block)) {
          expect_s3_class(x, "blk_cnd")
        }

        expect_true("eval" %in% names(cond))
        expect_named(cond$eval, c("warning", "error"), ignore.order = TRUE)

        expect_length(cond$eval$warning, 1L)
        expect_length(cond$eval$error, 0L)

        for (x in unlst(cond$eval)) {
          expect_s3_class(x, "blk_cnd")
        }
      },
      args = list(x = cnd, data = list(data = function() iris))
    )
  )

  withr::with_options(
    list(blockr.show_conditions = c("message", "warning")),
    testServer(
      get_s3_method("block_server", cnd),
      {
        session$flushReact()

        expect_s3_class(cond, "reactivevalues")

        expect_true("block" %in% names(cond))
        expect_named(cond$block, c("warning", "message"), ignore.order = TRUE)

        expect_length(cond$block$warning, 2L)
        expect_length(cond$block$message, 1L)

        for (x in unlst(cond$block)) {
          expect_s3_class(x, "blk_cnd")
        }

        expect_true("eval" %in% names(cond))
        expect_named(cond$eval, c("warning", "message"), ignore.order = TRUE)

        expect_length(cond$eval$warning, 1L)
        expect_length(cond$eval$message, 2L)

        for (x in unlst(cond$eval)) {
          expect_s3_class(x, "blk_cnd")
        }
      },
      args = list(x = cnd, data = list(data = function() iris))
    )
  )
})

test_that("block expr validation", {

  blk <- new_dataset_block()

  expect_error(
    check_expr_val(NULL, blk),
    class = "expr_server_return_type_invalid"
  )

  expect_error(
    check_expr_val(list(), blk),
    class = "expr_server_return_required_component_missing"
  )

  expect_error(
    check_expr_val(list(expr = "a", state = "b"), blk),
    class = "expr_server_return_expr_invalid"
  )

  expect_error(
    check_expr_val(list(expr = reactiveVal(), state = "b"), blk),
    class = "expr_server_return_state_type_invalid"
  )

  expect_error(
    check_expr_val(list(expr = reactiveVal(), state = list()), blk),
    class = "expr_server_return_state_missing_component"
  )

  expect_error(
    check_expr_val(
      list(
        expr = reactiveVal(),
        state = list(dataset = "a", package = "b", ui = NULL)
      ),
      blk
    ),
    class = "expr_server_return_state_invalid_component"
  )

  expect_error(
    check_expr_val(
      list(
        expr = reactiveVal(),
        state = list(dataset = reactiveVal("a"), package = reactiveVal("b")),
        cond = NULL
      ),
      blk
    ),
    class = "expr_server_return_cond_invalid"
  )

  expect_error(
    with_mock_session(
      check_expr_val(
        list(
          expr = reactiveVal(),
          state = list(dataset = reactiveVal("a"), package = reactiveVal("b")),
          cond = reactiveValues(foo = "abc")
        ),
        blk
      )
    ),
    class = "expr_server_return_cond_invalid"
  )

  with_mock_session(
    {
      check_expr_val(
        list(
          expr = reactiveVal(),
          state = list(
            dataset = reactiveVal("a"),
            package = reactiveVal("b")
          ),
          cond = reactiveValues(message = 123)
        ),
        blk
      )
      sink_msg(expect_warning(session$flushReact()))
    }
  )
})

test_that("block server exposes a reactive conditions frame", {

  blk <- new_dataset_block("iris")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      expect_true(is.reactive(conditions))
      expect_named(
        conditions(),
        c("block", "phase", "severity", "message", "id")
      )
      expect_identical(nrow(conditions()), 0L)

      cond$eval <- list(error = list(new_blk_cnd("boom")))

      session$flushReact()

      res <- conditions()

      expect_identical(nrow(res), 1L)
      expect_identical(res$phase, "eval")
      expect_identical(res$severity, "error")
      expect_identical(res$message, "boom")

      cond$eval <- list(error = list())

      session$flushReact()

      expect_identical(nrow(conditions()), 0L)
    },
    args = list(x = blk, data = list())
  )
})

test_that("an unconnected block is waiting and explains why", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_head_block()
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$a(), "ready")
      expect_identical(rv$eval$b(), "waiting")

      expect_identical(rv$blocks$a$server$result(), datasets::iris)
      expect_null(rv$blocks$b$server$result())

      # `waiting` surfaces a status-phase explanation (a warning, not an error)
      cnds <- rv$blocks$b$server$conditions()
      expect_false("error" %in% cnds$severity)
      expect_true(any(cnds$phase == "status" & cnds$severity == "warning"))
    },
    args = list(x = board)
  )
})

test_that("a waiting block evaluates once its input is connected", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_head_block()
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()
      expect_identical(rv$eval$b(), "waiting")

      board_update(
        list(links = list(add = links(ab = new_link("a", "b", "data"))))
      )
      session$flushReact()

      expect_identical(rv$eval$b(), "ready")
      expect_identical(
        rv$blocks$b$server$result(),
        utils::head(datasets::iris)
      )
      expect_identical(nrow(rv$blocks$b$server$conditions()), 0L)
    },
    args = list(x = board)
  )
})

test_that("a waiting upstream holds its downstream waiting (cascade)", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_head_block(),
      c = new_head_block()
    ),
    links = links(bc = new_link("b", "c", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$b(), "waiting")
      expect_identical(rv$eval$c(), "waiting")
      expect_false("error" %in% rv$conditions()$severity)

      board_update(
        list(links = list(add = links(ab = new_link("a", "b", "data"))))
      )
      session$flushReact()

      expect_identical(rv$eval$b(), "ready")
      expect_identical(rv$eval$c(), "ready")
      expect_identical(
        rv$blocks$c$server$result(),
        utils::head(datasets::iris)
      )
    },
    args = list(x = board)
  )
})

test_that("a block returns to waiting when its input is disconnected", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_head_block()
    ),
    links = links(ab = new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$b(), "ready")
      expect_identical(rv$blocks$b$server$result(), utils::head(datasets::iris))

      board_update(list(links = list(rm = "ab")))
      session$flushReact()

      expect_identical(rv$eval$b(), "waiting")
      expect_null(rv$blocks$b$server$result())
    },
    args = list(x = board)
  )
})

test_that("output clears when a ready block's input is disconnected", {

  blk <- new_head_block()
  inputs_ready <- reactiveVal(TRUE)

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      expect_false(is.null(output$result))

      inputs_ready(FALSE)
      session$flushReact()

      expect_null(session$returned$result())

      # a cleared output holds no value, so fetching it errors
      expect_error(output$result)
    },
    args = list(
      x = blk,
      data = list(data = reactive(utils::head(datasets::iris))),
      inputs_ready = inputs_ready
    )
  )
})

test_that("a block whose expression errors is failed, not ready", {

  boom_block <- function() {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(expr = reactive(quote(stop("boom"))), state = list())
          }
        )
      },
      function(id) tagList(),
      class = "boom_block",
      block_metadata = list()
    )
  }

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = boom_block()),
    links = links(ab = new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$b(), "failed")
      expect_null(rv$blocks$b$server$result())

      cnds <- rv$blocks$b$server$conditions()
      expect_true(any(cnds$phase == "eval" & cnds$severity == "error"))
    },
    args = list(x = board)
  )
})

cnd_row <- function(block, severity, id, message = "m", phase = "eval") {
  data.frame(
    block = block,
    phase = phase,
    severity = severity,
    message = message,
    id = id
  )
}

record_notifications <- function() {
  rec <- new.env(parent = emptyenv())
  rec$events <- character()
  rec
}

test_that("notify_user shows and clears conditions per block", {

  rec <- record_notifications()

  local_mocked_bindings(
    showNotification = function(ui, ..., id = NULL, session = NULL) {
      rec$events <- c(rec$events, paste0("show:", id))
      id
    },
    removeNotification = function(id, session = NULL) {
      rec$events <- c(rec$events, paste0("remove:", id))
      invisible()
    }
  )

  cond_a <- reactiveVal(empty_conditions_frame())

  board <- reactiveValues(
    blocks = list(a = list(server = list(conditions = cond_a)))
  )

  testServer(
    notify_user_server,
    {
      session$flushReact()
      expect_identical(rec$events, character())

      cond_a(cnd_row("a", "error", "e1", "boom"))
      session$flushReact()
      expect_identical(rec$events, "show:a-error-e1")

      cond_a(empty_conditions_frame())
      session$flushReact()
      expect_identical(rec$events, c("show:a-error-e1", "remove:a-error-e1"))

      expect_null(session$returned)
    },
    args = list(board = board)
  )
})

test_that("notify_user shows a separate toast per block for a shared message", {

  rec <- record_notifications()

  local_mocked_bindings(
    showNotification = function(ui, ..., id = NULL, session = NULL) {
      rec$events <- c(rec$events, paste0("show:", id))
      id
    },
    removeNotification = function(id, session = NULL) {
      rec$events <- c(rec$events, paste0("remove:", id))
      invisible()
    }
  )

  cond_a <- reactiveVal(empty_conditions_frame())
  cond_b <- reactiveVal(empty_conditions_frame())

  board <- reactiveValues(
    blocks = list(
      a = list(server = list(conditions = cond_a)),
      b = list(server = list(conditions = cond_b))
    )
  )

  testServer(
    notify_user_server,
    {
      session$flushReact()

      cond_a(cnd_row("a", "error", "e1", "boom"))
      session$flushReact()

      cond_b(cnd_row("b", "error", "e1", "boom"))
      session$flushReact()

      expect_identical(rec$events, c("show:a-error-e1", "show:b-error-e1"))

      cond_a(empty_conditions_frame())
      session$flushReact()

      expect_identical(
        rec$events,
        c("show:a-error-e1", "show:b-error-e1", "remove:a-error-e1")
      )
    },
    args = list(board = board)
  )
})

test_that("notify_user clears notifications of a removed block", {

  rec <- record_notifications()

  local_mocked_bindings(
    showNotification = function(ui, ..., id = NULL, session = NULL) {
      rec$events <- c(rec$events, paste0("show:", id))
      id
    },
    removeNotification = function(id, session = NULL) {
      rec$events <- c(rec$events, paste0("remove:", id))
      invisible()
    }
  )

  cond_a <- reactiveVal(cnd_row("a", "error", "ea", "boom"))
  cond_b <- reactiveVal(cnd_row("b", "warning", "wb", "careful"))

  board <- reactiveValues(
    blocks = list(
      a = list(server = list(conditions = cond_a)),
      b = list(server = list(conditions = cond_b))
    )
  )

  testServer(
    notify_user_server,
    {
      session$flushReact()
      expect_setequal(rec$events, c("show:a-error-ea", "show:b-warning-wb"))

      rec$events <- character()

      board$blocks <- board$blocks["a"]
      session$flushReact()

      expect_identical(rec$events, "remove:b-warning-wb")
    },
    args = list(board = board)
  )
})

test_that("notify_user reads only the changed block's conditions", {

  local_mocked_bindings(
    showNotification = function(ui, ..., id = NULL, session = NULL) id,
    removeNotification = function(id, session = NULL) invisible()
  )

  reads <- new.env(parent = emptyenv())
  reads$a <- 0L
  reads$b <- 0L

  val_a <- reactiveVal(empty_conditions_frame())
  val_b <- reactiveVal(cnd_row("b", "error", "eb", "B"))

  cond_a <- function() {
    reads$a <- reads$a + 1L
    val_a()
  }

  cond_b <- function() {
    reads$b <- reads$b + 1L
    val_b()
  }

  board <- reactiveValues(
    blocks = list(
      a = list(server = list(conditions = cond_a)),
      b = list(server = list(conditions = cond_b))
    )
  )

  testServer(
    notify_user_server,
    {
      session$flushReact()

      reads_a <- reads$a
      reads_b <- reads$b

      for (i in seq_len(5L)) {
        val_a(cnd_row("a", "error", paste0("ea", i), "A"))
        session$flushReact()
      }

      expect_gt(reads$a, reads_a)
      expect_identical(reads$b, reads_b)
    },
    args = list(board = board)
  )
})

test_that("notif_frame gates by show_conditions", {

  frame <- rbind(
    cnd_row("a", "error", "e1", "boom"),
    cnd_row("b", "warning", "w1", "careful"),
    cnd_row("c", "message", "m1", "fyi")
  )

  with_mock_session(
    {
      gated <- withr::with_options(
        list(blockr.show_conditions = c("warning", "error")),
        notif_frame(frame, session)
      )

      expect_setequal(gated$key, c("a-error-e1", "b-warning-w1"))

      full <- withr::with_options(
        list(blockr.show_conditions = c("message", "warning", "error")),
        notif_frame(frame, session)
      )

      expect_setequal(
        full$key,
        c("a-error-e1", "b-warning-w1", "c-message-m1")
      )
    }
  )
})

test_that("notif_frame collapses a block's duplicate keys to one row", {

  dup <- rbind(
    cnd_row("a", "error", "e1", "boom"),
    cnd_row("a", "error", "e1", "boom", phase = "render")
  )

  with_mock_session(
    {
      res <- withr::with_options(
        list(blockr.show_conditions = c("warning", "error")),
        notif_frame(dup, session)
      )

      expect_identical(nrow(res), 1L)
      expect_identical(res$key, "a-error-e1")
    }
  )
})

test_that("rbind block constructor", {

  blk <- new_rbind_block()

  expect_s3_class(blk, "rbind_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        rbind(iris[1:3, ], iris[4:6, ])
      )
    },
    args = list(
      x = blk,
      data = list(
        ...args = reactives(
          function() iris[1:3, ],
          function() iris[4:6, ]
        )
      )
    )
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        rbind(a = iris[1:3, ], b = iris[4:6, ])
      )
    },
    args = list(
      x = blk,
      data = list(
        ...args = reactives(
          a = function() iris[1:3, ],
          b = function() iris[4:6, ]
        )
      )
    )
  )

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        rbind(iris[1:3, ], a = iris[4:6, ])
      )
    },
    args = list(
      x = blk,
      data = list(
        ...args = reactives(
          function() iris[1:3, ],
          a = function() iris[4:6, ]
        )
      )
    )
  )
})

test_that("named variadic inputs produce named rbind arguments", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(
      ac = new_link("a", "c", "left"),
      bc = new_link("b", "c", "right")
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(
        rv$blocks$c$server$result(),
        rbind(left = datasets::BOD, right = datasets::BOD)
      )

      code <- export_wrapped_code(
        lapply(lst_xtr(rv$blocks, "server", "expr"), reval),
        rv$board
      )

      expect_match(code, "rbind(left = a, right = b)", fixed = TRUE)
    },
    args = list(x = board)
  )
})

test_that("unnamed variadic inputs produce a positional rbind", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(
      ac = new_link("a", "c"),
      bc = new_link("b", "c")
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(
        rv$blocks$c$server$result(),
        rbind(datasets::BOD, datasets::BOD)
      )

      code <- export_wrapped_code(
        lapply(lst_xtr(rv$blocks, "server", "expr"), reval),
        rv$board
      )

      expect_match(code, "rbind(a, b)", fixed = TRUE)
    },
    args = list(x = board)
  )
})

test_that("positional ...args follow container order, not any sort", {

  blk <- new_rbind_block()

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      expect_identical(
        session$returned$result(),
        rbind(iris[4:6, ], iris[1:3, ])
      )
    },
    args = list(
      x = blk,
      data = list(
        ...args = reactives(
          function() iris[4:6, ],
          function() iris[1:3, ]
        )
      )
    )
  )
})

test_that("editing a variadic link preserves its argument order", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(ac = new_link("a", "c"), bc = new_link("b", "c"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_match(
        export_wrapped_code(
          lapply(lst_xtr(rv$blocks, "server", "expr"), reval),
          rv$board
        ),
        "rbind(a, b)",
        fixed = TRUE
      )

      board_update(
        list(
          links = list(
            add = links(ac = new_link("a", "c", "left")),
            rm = "ac"
          )
        )
      )
      session$flushReact()

      expect_match(
        export_wrapped_code(
          lapply(lst_xtr(rv$blocks, "server", "expr"), reval),
          rv$board
        ),
        "rbind(left = a, b)",
        fixed = TRUE
      )
    },
    args = list(x = board)
  )
})

test_that("a variadic block below its input minimum is waiting", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_rbind_block()
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$b(), "waiting")
      expect_null(rv$blocks$b$server$result())
      expect_false("error" %in% rv$blocks$b$server$conditions()$severity)

      board_update(
        list(links = list(add = links(ab = new_link("a", "b"))))
      )
      session$flushReact()

      expect_identical(rv$eval$b(), "ready")
      expect_identical(rv$blocks$b$server$result(), datasets::iris)
    },
    args = list(x = board)
  )
})

test_that("snapshot_reactives detaches from the source reactives", {

  src <- reactiveValues(
    val = 1L,
    block = list(
      server = list(
        expr = reactive("E"),
        cond = reactiveValues(eval = "c")
      )
    )
  )

  snap <- snapshot_reactives(src)

  expect_true(is.reactivevalues(snap))

  isolate({

    expect_identical(snap$val, 1L)
    expect_identical(snap$block$server$expr(), "E")

    expect_true(is.reactivevalues(snap$block$server$cond))
    expect_identical(
      reactiveValuesToList(snap$block$server$cond),
      list(eval = "c")
    )
  })

  isolate(src$val <- 2L)
  isolate(src$block$server$cond$eval <- "d")

  expect_identical(isolate(snap$val), 1L)
  expect_identical(isolate(snap$block$server$cond$eval), "c")
})

test_that("snapshot_reactives defers evaluation errors to read time", {

  snap <- snapshot_reactives(list(thrower = reactive(stop("boom"))))

  expect_error(isolate(snap$thrower()), "boom")
})

test_that("generate_plugin_args returns session-independent args", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("ChickWeight"),
      c = new_merge_block(by = "Time")
    ),
    links = links(
      from = c("a", "b"),
      to = c("c", "c"),
      input = c("x", "y")
    )
  )

  args <- generate_plugin_args(board)

  expect_true(is.reactivevalues(args$board))

  isolate({

    exprs <- lst_xtr_reval(args$board$blocks, "server", "expr")
    expect_named(exprs, c("a", "b", "c"), ignore.order = TRUE)

    conditions <- args$board$blocks[["a"]]$server$conditions
    expect_true(is.function(conditions))
    expect_s3_class(conditions(), "data.frame")
  })
})

test_that("static block constructor", {

  blk <- new_static_block(datasets::mtcars)

  expect_s3_class(blk, "static_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        datasets::mtcars
      )
    },
    args = list(x = blk)
  )
})

#' Transform block constructors
#'
#' Many data transformations are be provided by blocks constructed via
#' `new_transform_block()`, including examples where a single `data.frame` is
#' transformed into another (e.g. `subset_block`), and two or more `data.frame`s
#' are combined (e.g. `merge_block` or `rbind_block`).
#'
#' @param ... Forwarded to `new_transform_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_transform_block()` inherit from
#' `transform_block`.
#'
#' @seealso Real-world transform blocks (e.g. `select`, `mutate`, `filter` and
#' `join`) built on `new_transform_block()` are provided by the
#' [blockr.dplyr](https://bristolmyerssquibb.github.io/blockr.dplyr/) package.
#'
#' @export
new_transform_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "transform_block"), ctor, ...)
}

#' @export
block_output.transform_block <- function(x, result, session) {
  dt_result(result, x, session)
}

#' @export
block_ui.transform_block <- function(id, x, ...) {
  tagList(
    DT::dataTableOutput(NS(id, "result"))
  )
}

#' @export
block_render_trigger.transform_block <- dt_render_trigger

#' @export
board_options.transform_block <- function(x, ...) {
  combine_board_options(
    new_n_rows_option(...),
    new_page_size_option(...),
    new_filter_rows_option(...),
    NextMethod()
  )
}

#' Quoting utilities
#'
#' Block expressions in blockr are evaluated in a 2-step manner: first in the
#' context of arguments supplied by the user via UI elements and in a second
#' step in the context of input data. The function [base::bquote()] does not
#' allow for terms wrapped in `.()` or `..()` to be missing and this makes it
#' incompatible with this 2-step approach. A drop-in replacement, provided as
#' `bbquote()` addresses this shortcoming.
#'
#' A block like `new_head_block()` is expected to return an expression of the
#' form `utils::head(data, n = 10L)`, which will then be evaluated in an
#' environment where he have a name `data` bound to some dataset. In order to
#' perform some manipulations of such block expressions it is required to
#' somehow mark the terms that correspond to input data and for that we can
#' use the syntax introduced by [base::bquote()]. What we would prefer to have
#' as block expression therefore is not the above, but something like
#' `utils::head(.(data), n = 10L)`, as this affords us more possibilities for
#' performing substitutions (and therefore generates cleaner code).
#'
#' In order to interpolate certain arguments in a first step, we unfortunately
#' cannot use [base::bquote()], but we can use `bbquote()` instead to generate
#' the desired expression.
#'
#' ```r
#' bquote(utils::head(.(data), n = .(n)), list(n = 10L))
#' #> Error in eval(e[[2L]], where) : object 'data' not found
#' bbquote(utils::head(.(data), n = .(n)), list(n = 10L))
#' #> utils::head(.(data), n = 10L)
#' ```
#'
#' This also works with `..()` and splicing.
#'
#' @inheritParams base::bquote
#'
#' @examples
#' bbquote(utils::head(.(data), n = .(n)), list(n = 10L))
#' bbquote(c(.(a), ..(bc)), list(a = "a"))
#' bbquote(c(.(a), ..(bc)), list(a = "a", bc = c("b", "c")), splice = TRUE)
#'
#' @return A language object in the same way as returned by [base::bquote()].
#' Functions `.()` and `..()` throw errors when invoked an only exist to mask
#' check notes "no visible global function definition" for their use.
#'
#' @export
bbquote <- function(expr, where = parent.frame(), splice = FALSE) {

  maybe_dots <- function(x) {
    is.call(x) && length(x) == 2 && is.name(x[[2]])
  }

  is_dot <- function(x) {
    maybe_dots(x) && identical(x[[1]], quote(.))
  }

  is_dots <- function(x) {
    maybe_dots(x) && identical(x[[1]], quote(..))
  }

  find_dots <- function(x) {
    if (is_dot(x)) {
      list(list(var = as.character(x[[2]]), splice = FALSE))
    } else if (is_dots(x)) {
      list(list(var = as.character(x[[2]]), splice = TRUE))
    } else if (is.call(x) || is.pairlist(x)) {
      unlst(lapply(as.list(x), find_dots))
    } else {
      NULL
    }
  }

  process_splices <- function(e) {

    if (!is.call(e)) {
      return(e)
    }

    e_list <- as.list(e)
    names_e <- coal(names(e_list), rep("", length(e_list)))

    for (i in seq_along(e_list)) {
      if (nchar(names_e[i]) && is_dots(e_list[[i]])) {
        names_e[i] <- ""
      } else {
        e_list[[i]] <- process_splices(e_list[[i]])
      }
    }

    names(e_list) <- names_e

    as.call(e_list)
  }

  expr_sub <- substitute(expr)

  if (missing(expr)) {
    return(bquote())
  }

  if (is.name(expr_sub)) {

    expr_val <- tryCatch(
      eval(expr_sub, parent.frame()),
      error = function(e) NULL
    )

    if (!is.null(expr_val) && (is.call(expr_val) || is.expression(expr_val))) {
      expr_sub <- expr_val
    }
  }

  dot_vars_list <- find_dots(expr_sub)

  if (length(dot_vars_list) > 0) {
    dot_vars <- chr_xtr(dot_vars_list, "var")
    spl_vars <- unique(dot_vars[lgl_xtr(dot_vars_list, "splice")])
    dot_vars <- unique(dot_vars)
  } else {
    dot_vars <- character(0)
    spl_vars <- character(0)
  }

  if (is.list(where)) {
    available <- names(where)
  } else {
    available <- ls(envir = where, all.names = TRUE)
  }

  subst_list <- list()

  for (var in dot_vars) {

    if (var %in% available) {

      if (is.list(where)) {
        subst_list[[var]] <- where[[var]]
      } else {
        subst_list[[var]] <- get(var, envir = where)
      }

    } else {

      if (var %in% spl_vars) {
        subst_list[[var]] <- set_names(list(as_dots_call(var)), var)
      } else {
        subst_list[[var]] <- as_dot_call(var)
      }
    }
  }

  process_splices(
    eval(call("bquote", expr_sub, subst_list, splice = splice))
  )
}

as_dot_call <- function(x) {
  call(".", as.name(x))
}

as_dots_call <- function(x) {
  call("..", as.name(x))
}

#' @param x Object
#' @rdname bbquote
#' @export
. <- function(x) {
  blockr_abort(
    "Function `.()` is not intended to be called.",
    class = "dot_should_not_be_called"
  )
}

#' @rdname bbquote
#' @export
.. <- function(x) {
  blockr_abort(
    "Function `..()` is not intended to be called.",
    class = "dots_should_not_be_called"
  )
}

test_that("bbquote", {

  expect_identical(
    bbquote(func(.(a), .(a)), list(a = "a")),
    quote(func("a", "a"))
  )

  expect_identical(
    bbquote(func(..(a), ..(a)), list(a = 1:2), splice = TRUE),
    quote(func(1L, 2L, 1L, 2L))
  )

  expr1 <- quote(func(..(a), ..(b)))

  expect_identical(
    bbquote(expr1, list()),
    expr1
  )

  expect_identical(
    expr2 <- bbquote(expr1, list(a = list(x = 1, y = 2)), splice = TRUE),
    quote(func(x = 1, y = 2, ..(b)))
  )

  expect_identical(
    expr3 <- bbquote(expr2, list(b = list(z = 3)), splice = TRUE),
    quote(func(x = 1, y = 2, z = 3))
  )

  expr4 <- bbquote(
    plot(..(args), main = .(title)),
    list(args = list(x = 1:10, y = 1:10)),
    splice = TRUE
  )

  expect_identical(
    deparse(expr4),
    "plot(x = 1:10, y = 1:10, main = .(title))"
  )

  expr5 <- bbquote(expr4, list(title = "My Plot"))

  expect_identical(
    deparse(expr5),
    "plot(x = 1:10, y = 1:10, main = \"My Plot\")"
  )

  expect_identical(
    bbquote(func(a = 1L, ..(a)), list()),
    quote(func(a = 1L, ..(a)))
  )

  expect_identical(
    bbquote(func(a = 1L, ..(a)), list(a = c("a", "b")), splice = TRUE),
    quote(func(a = 1L, "a", "b"))
  )

  expect_identical(
    bbquote(func(a = 1L, .(a)), list(a = "a")),
    quote(func(a = 1L, "a"))
  )

  expect_identical(
    eval(
      bbquote(
        {
          b <- 1L
          .(a) + b
        },
        list(a = 2L)
      )
    ),
    3L
  )

  expect_identical(
    eval(
      bbquote(
        {
          a <- 1L
          .(a) + a
        },
        list(a = 2L)
      )
    ),
    3L
  )

  expect_identical(
    eval(bbquote(1L + 2L, list())),
    3L
  )

  expect_identical(
    eval(bbquote(1L + 2L, list(a = 3L))),
    3L
  )

  expect_identical(bquote(), bbquote())

  expect_identical(
    bbquote(func(.(a), .(a)), list2env(list(a = "a"), parent = emptyenv())),
    quote(func("a", "a"))
  )

  expect_error(.("a"), class = "dot_should_not_be_called")
  expect_error(..("a"), class = "dots_should_not_be_called")
})

#' Functional programming utilities
#'
#' A set of wrappers for [base::vapply()] with some convenient defaults.
#' Intended mainly for internal use, but available for dependent packages
#' wherever this is deemed convenient.
#'
#' @param x Object to iterate over
#' @param fun Function to apply to each component
#' @param ... Forwarded to `fun`
#' @param length Expected result length
#' @param use_names Name the result using `names(x)`
#'
#' @keywords internal
#'
#' @return The result of a call to [base::lapply()], [base::vapply()] or
#' [base::Map()].
#'
#' @export
chr_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, character(length), ..., USE.NAMES = use_names)
}

#' @rdname chr_ply
#' @export
lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
}

#' @rdname chr_ply
#' @export
int_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, integer(length), ..., USE.NAMES = use_names)
}

#' @rdname chr_ply
#' @export
dbl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, double(length), ..., USE.NAMES = use_names)
}

#' @rdname chr_ply
#' @export
chr_mply <- function(..., length = 1L) {
  chr_ply(map(...), identity, length = length)
}

#' @rdname chr_ply
#' @export
lgl_mply <- function(..., length = 1L) {
  lgl_ply(map(...), identity, length = length)
}

#' @rdname chr_ply
#' @export
int_mply <- function(..., length = 1L) {
  int_ply(map(...), identity, length = length)
}

#' @rdname chr_ply
#' @export
dbl_mply <- function(..., length = 1L) {
  dbl_ply(map(...), identity, length = length)
}

#' @param i Index to extract
#' @rdname chr_ply
#' @export
chr_xtr <- function(x, i, ...) chr_ply(x, `[[`, i, ...)

#' @rdname chr_ply
#' @export
lgl_xtr <- function(x, i, ...) lgl_ply(x, `[[`, i, ...)

#' @rdname chr_ply
#' @export
int_xtr <- function(x, i, ...) int_ply(x, `[[`, i, ...)

#' @rdname chr_ply
#' @export
dbl_xtr <- function(x, i, ...) dbl_ply(x, `[[`, i, ...)

#' @rdname chr_ply
#' @export
lst_xtr <- function(x, ...) {

  for (i in c(...)) {
    x <- lapply(x, `[[`, i)
  }

  x
}

#' @rdname chr_ply
#' @export
map <- function(fun, ..., use_names = FALSE) {
  Map(fun, ..., USE.NAMES = use_names)
}

test_that("ply", {

  dat <- replicate(5, sample(letters, 10), simplify = FALSE)

  expect_length(
    chr_ply(dat, paste0, collapse = ""),
    5
  )

  expect_length(
    lgl_ply(dat, `==`, "a", length = 10L),
    5 * 10
  )

  expect_length(
    int_ply(dat, match, "a", length = 10L),
    5 * 10
  )

  dat <- replicate(5, sample(seq(0, 1, by = 0.1), 10), simplify = FALSE)

  expect_length(dbl_ply(dat, sum), 5)

  expect_length(chr_mply(paste, letters, LETTERS), length(letters))

  a <- sample(c(TRUE, FALSE), 10, replace = TRUE)
  b <- sample(c(TRUE, FALSE), 10, replace = TRUE)

  expect_length(lgl_mply(`|`, a, b), 10)

  expect_length(int_mply(`+`, seq.int(10), seq.int(10)), 10)
  expect_length(dbl_mply(`+`, seq(0, 1, by = 0.1), seq(0, 1, by = 0.1)), 11)

  expect_length(
    chr_xtr(replicate(5, sample(letters, 10), simplify = FALSE), 3L),
    5
  )

  expect_length(
    lgl_xtr(replicate(5, rep(TRUE, 5), simplify = FALSE), 3L),
    5
  )

  expect_length(
    int_xtr(replicate(5, rep(1L, 5), simplify = FALSE), 3L),
    5
  )

  expect_length(
    dbl_xtr(replicate(5, rep(1.5, 5), simplify = FALSE), 3L),
    5
  )

  expect_length(
    lst_xtr(replicate(5, rep(1.5, 5), simplify = FALSE), 3L),
    5
  )

  expect_named(
    map(paste, set_names(nm = letters), set_names(LETTERS, letters)),
    NULL
  )
})

#' Code generation plugin module
#'
#' All code necessary for reproducing a data analysis as set up in blockr can
#' be made available to the user. Several ways of providing such a script or
#' code snippet are conceivable and currently implemented, we have a modal
#' with copy-to-clipboard functionality. This is readily extensible, for example
#' by offering a download button, by providing this functionality as a
#' `generate_code` module.
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `generate_code` is returned by
#' `generate_code()`, while the UI component (e.g. `generate_code_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `generate_code_server()`) is expected to return `NULL`.
#'
#' @export
generate_code <- function(server = generate_code_server,
                          ui = generate_code_ui) {

  new_plugin(server, ui, class = "generate_code")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname generate_code
#' @export
generate_code_server <- function(id, board, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      code <- reactive(
        if (isTRUE(code_export_ready(board))) {
          export_wrapped_code(
            lst_xtr_reval(board$blocks, "server", "expr"),
            board$board
          )
        }
      )

      observeEvent(
        input$code_mod,
        showModal(
          modalDialog(
            title = "Generated code",
            code_modal_body(code(), session$ns("code_out")),
            easyClose = TRUE,
            footer = NULL,
            size = "l"
          )
        )
      )

      NULL
    }
  )
}

#' @rdname generate_code
#' @export
generate_code_ui <- function(id, board) {
  tagList(
    actionButton(
      NS(id, "code_mod"),
      "Show code",
      icon = icon("code")
    )
  )
}

code_export_ready <- function(board) {

  status <- reactiveValuesToList(board$eval)

  if (!setequal(names(status), board_block_ids(board$board))) {
    return(FALSE)
  }

  all(chr_ply(status, reval_if) %in% c("ready", "dormant"))
}

code_modal_body <- function(script, out_id) {

  if (is.null(script)) {
    return(
      div(
        class = "text-muted",
        paste(
          "The board is not ready. Finish configuring all blocks",
          "before exporting code."
        )
      )
    )
  }

  div(
    id = out_id,
    class = "text-decoration-none position-relative",
    pre(paste0(script, collapse = "\n"))
  )
}

test_that("generate code", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("ChickWeight"),
      c = new_merge_block(by = "Time")
    ),
    links = links(
      from = c("a", "b"),
      to = c("c", "c"),
      input = c("x", "y")
    )
  )

  testServer(
    generate_code_server,
    {
      res <- code()

      expect_type(res, "character")
      expect_length(res, 1L)

      session$setInputs(code_mod = 1)
    },
    args = generate_plugin_args(board)
  )
})

test_that("code export is gated on all blocks being ready", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      c = new_merge_block(by = "Time")
    )
  )

  testServer(
    generate_code_server,
    {
      expect_false(code_export_ready(board))
      expect_null(code())

      session$setInputs(code_mod = 1)
    },
    args = generate_plugin_args(board)
  )
})

test_that("code_export_ready settles on ready or dormant", {

  reactiveConsole(TRUE)
  on.exit(reactiveConsole(FALSE))

  board <- new_board(
    blocks = c(a = new_dataset_block("BOD"), b = new_dataset_block("BOD"))
  )

  make_rv <- function(...) {
    list(eval = do.call(reactiveValues, list(...)), board = board)
  }

  expect_true(
    code_export_ready(make_rv(a = reactive("ready"), b = reactive("ready")))
  )
  expect_true(
    code_export_ready(make_rv(a = reactive("ready"), b = reactive("dormant")))
  )

  expect_false(
    code_export_ready(make_rv(a = reactive("ready"), b = reactive("waiting")))
  )
  expect_false(
    code_export_ready(make_rv(a = reactive("ready"), b = reactive("unset")))
  )
  expect_false(
    code_export_ready(make_rv(a = reactive("ready"), b = reactive("failed")))
  )

  expect_false(
    code_export_ready(make_rv(a = reactive("ready")))
  )
})

test_that("code modal body shows script or a not-ready note", {

  note <- code_modal_body(NULL, "out")

  expect_s3_class(note, "shiny.tag")
  expect_match(as.character(note), "not ready")

  body <- code_modal_body("y <- 1", "out")

  expect_match(as.character(body), "<pre>")
  expect_match(as.character(body), "y &lt;- 1")
})

test_that("dummy add/rm block ui test", {
  expect_s3_class(generate_code_ui("gen", new_board()), "shiny.tag.list")
})

new_blk_cnd <- function(x) {
  structure(
    x,
    id = digest::digest(x, "xxh3_64", serialize = FALSE),
    class = "blk_cnd"
  )
}

as_blk_cnd <- function(x) {
  UseMethod("as_blk_cnd")
}

#' @noRd
#' @export
as_blk_cnd.blk_cnd <- function(x) {
  x
}

#' @noRd
#' @export
as_blk_cnd.condition <- function(x) {
  new_blk_cnd(fmt_cnd_msg(x))
}

#' @noRd
#' @export
as_blk_cnd.character <- function(x) {
  new_blk_cnd(x)
}

blk_cnds <- function(x, block = NA_character_) {

  phases <- names(x)

  rows <- lapply(
    c("error", "warning", "message"),
    function(sev) {

      msgs <- lst_xtr(x[phases], sev)
      lens <- lengths(msgs)

      if (!any(lens)) {
        return(NULL)
      }

      flat <- unlst(msgs)

      data.frame(
        phase = rep(phases, lens),
        severity = sev,
        message = chr_ply(flat, cnd_message),
        id = chr_ply(flat, cnd_id),
        row.names = NULL
      )
    }
  )

  out <- do.call(rbind, rows)

  if (is.null(out)) {
    out <- data.frame(
      phase = character(),
      severity = character(),
      message = character(),
      id = character()
    )
  }

  res <- cbind(block = rep(block, nrow(out)), out)
  row.names(res) <- NULL

  res
}

is_blk_cnd <- function(x) {
  inherits(x, "blk_cnd")
}

cnd_id <- function(x) {
  attr(x, "id")
}

cnd_message <- function(x) {
  as.character(x)
}

empty_conditions_frame <- function() {
  blk_cnds(list())
}

is_list_of_blk_cnds <- function(x) {
  is.list(x) && all(lgl_ply(x, is_blk_cnd))
}

empty_block_condition <- function() {
  list(
    error = list(),
    warning = list(),
    message = list()
  )
}

fmt_cnd_msg <- function(x) {
  sub("\n$", "", conditionMessage(x))
}

show_condition <- function(x) {
  UseMethod("show_condition")
}

#' @noRd
#' @export
show_condition.list <- function(x) {
  invisible(lapply(x, show_condition))
}

#' @noRd
#' @export
show_condition.default <- function(x) {
  invisible(NULL)
}

#' @noRd
#' @export
show_condition.message <- function(x) {
  message(fmt_cnd_msg(x))
}

#' @noRd
#' @export
show_condition.warning <- function(x) {
  warning(fmt_cnd_msg(x))
}

#' @noRd
#' @export
show_condition.error <- function(x) {
  stop(fmt_cnd_msg(x))
}

#' @importFrom evaluate replay
#' @export
replay.message <- function(x) {
  log_info(fmt_cnd_msg(x))
}

#' @export
replay.warning <- function(x) {
  log_warn(fmt_cnd_msg(x))
}

#' @export
replay.error <- function(x) {
  log_error(fmt_cnd_msg(x))
}

msg_handler <- function(cond, conds) {
  stopifnot(is.environment(cond), is.character(conds))
  function(m) {
    if ("message" %in% conds) {
      cond$message <- c(cond$message, list(as_blk_cnd(m)))
    }
    log_info(fmt_cnd_msg(m))
    tryInvokeRestart("muffleMessage")
  }
}

warn_handler <- function(cond, conds) {
  stopifnot(is.environment(cond), is.character(conds))
  function(w) {
    if ("warning" %in% conds) {
      cond$warning <- c(cond$warning, list(as_blk_cnd(w)))
    }
    log_warn(fmt_cnd_msg(w))
    tryInvokeRestart("muffleWarning")
  }
}

err_handler <- function(cond, conds, err_val = NULL) {
  stopifnot(is.environment(cond), is.character(conds))
  function(e) {
    if ("error" %in% conds) {
      cond$error <- list(as_blk_cnd(e))
    }
    log_error(fmt_cnd_msg(e))
    err_val
  }
}

capture_conditions <- function(expr, rv, slot, error_val = NULL,
                               session = get_session()) {

  stopifnot(
    is.reactivevalues(rv),
    is_string(slot), slot %in% names(rv)
  )

  cnds <- coal(
    isolate(
      get_board_option_or_null("show_conditions", session)
    ),
    match.arg(
      blockr_option("show_conditions", c("warning", "error")),
      c("message", "warning", "error"),
      several.ok = TRUE
    )
  )

  cond <- list2env(empty_block_condition()[cnds])

  res <- try_catch_continue(
    expr,
    message = msg_handler(cond, cnds),
    warning = warn_handler(cond, cnds),
    error = err_handler(cond, cnds, error_val)
  )

  cond <- as.list(cond)
  curr <- set_names(
    coal(rv[[slot]], empty_block_condition())[cnds],
    cnds
  )

  if (any(lengths(cond))) {

    chk <- lgl_mply(
      Negate(setequal),
      lapply(cond, chr_ply, attr, "id"),
      lapply(curr, chr_ply, attr, "id")
    )

    if (any(chk)) {
      rv[[slot]] <- cond
    }

  } else if (any(lengths(curr) > 0L)) {
    rv[[slot]] <- empty_block_condition()[cnds]
  }

  res
}

try_catch_continue <- function(expr, message, warning, error) {
  tryCatch(
    withCallingHandlers(
      expr,
      message = message,
      warning = warning
    ),
    error = error
  )
}

glue_plur <- function(..., envir = parent.frame()) {
  cli::pluralize(..., .envir = envir)
}

#' Blockr conditions
#'
#' Wrappers for [rlang::abort()], [rlang::warn()] and [rlang::inform()]. In
#' addition to `class`, conditions inherit from "blockr_error".
#'
#' @param ... Forwarded to [cli::pluralize()]
#' @param class Condition class
#' @param envir Forwarded to [cli::pluralize()]
#'
#' @return Called for side-effect of signaling conditions.
#'
#' @export
blockr_abort <- function(..., class = character(), envir = parent.frame()) {
  rlang::abort(glue_plur(..., envir = envir), c(class, "blockr_error"))
}

#' @param frequency,frequency_id Forwarded to [rlang::warn()]
#' @rdname blockr_abort
#' @export
blockr_warn <- function(..., class = character(), envir = parent.frame(),
                        frequency = "always", frequency_id = NULL) {

  rlang::warn(
    glue_plur(..., envir = envir),
    c(class, "blockr_warning"),
    .frequency = frequency,
    .frequency_id = frequency_id
  )
}

#' @rdname blockr_abort
#' @export
blockr_inform <- function(..., class = character(), envir = parent.frame(),
                          frequency = "always", frequency_id = NULL) {

  rlang::inform(
    glue_plur(..., envir = envir),
    c(class, "blockr_message"),
    .frequency = frequency,
    .frequency_id = frequency_id
  )
}

test_that("conditions", {

  with_mock_session(
    {
      vals <- reactiveValues(test = NULL)

      res <- withr::with_options(
        list(blockr.show_conditions = c("warning", "error")),
        capture_conditions(
          {
            message("hello")
            warning("test1")
            warning("test2")
            stop("bye")
          },
          rv = vals,
          slot = "test",
          error_val = NULL
        )
      )

      expect_null(res)

      cnds <- vals$test

      expect_named(cnds, c("warning", "error"), ignore.order = TRUE)

      expect_length(cnds$warning, 2L)
      expect_length(cnds$error, 1L)

      res <- withr::with_options(
        list(blockr.show_conditions = c("message", "warning", "error")),
        capture_conditions(
          {
            message("hello")
            warning("test1")
            warning("test2")
            stop("bye")
          },
          rv = vals,
          slot = "test",
          error_val = NULL
        )
      )

      expect_null(res)

      cnds <- vals$test

      expect_named(cnds, c("message", "warning", "error"), ignore.order = TRUE)

      expect_length(cnds$message, 1L)
      expect_length(cnds$warning, 2L)
      expect_length(cnds$error, 1L)
    }
  )

  test_val <- "world"

  expect_error(
    blockr_abort("hello {test_val}", class = "test_error"),
    class = "test_error"
  )

  expect_warning(
    blockr_warn("hello {test_val}", class = "test_warning"),
    class = "test_warning"
  )

  expect_message(
    blockr_inform("hello {test_val}", class = "test_message"),
    class = "test_message"
  )

  cnd <- new_blk_cnd("xyz")

  expect_s3_class(cnd, "blk_cnd")
  expect_identical(cnd, as_blk_cnd(cnd))
})

test_that("blk_cnd data api", {

  cnd <- new_blk_cnd("xyz")

  expect_s3_class(cnd, "blk_cnd")
  expect_true(is_blk_cnd(cnd))
  expect_false(is_blk_cnd("xyz"))

  expect_identical(cnd_message(cnd), "xyz")
  expect_identical(cnd_id(cnd), attr(cnd, "id"))
  expect_null(cnd_id("xyz"))

  expect_identical(cnd, as_blk_cnd(cnd))
})

test_that("blk_cnds flattens a cond", {

  empty <- blk_cnds(list())

  expect_s3_class(empty, "data.frame")
  expect_identical(nrow(empty), 0L)
  expect_named(empty, c("block", "phase", "severity", "message", "id"))
  expect_identical(empty, empty_conditions_frame())

  expect_identical(
    nrow(blk_cnds(list(eval = NULL, render = NULL), block = "a")),
    0L
  )

  cond <- list(
    state = list(error = list(new_blk_cnd("state bad"))),
    eval = list(
      warning = list(new_blk_cnd("w1"), new_blk_cnd("w2")),
      error = list(new_blk_cnd("boom"))
    ),
    block = list(message = list(new_blk_cnd("hi")))
  )

  res <- blk_cnds(cond, block = "blk1")

  expect_named(res, c("block", "phase", "severity", "message", "id"))
  expect_identical(nrow(res), 5L)
  expect_setequal(unique(res$block), "blk1")
  expect_setequal(res$severity, c("error", "warning", "message"))

  eval_warn <- res[res$phase == "eval" & res$severity == "warning", ]
  expect_setequal(eval_warn$message, c("w1", "w2"))

  expect_identical(
    res$id,
    chr_ply(res$message, function(m) cnd_id(new_blk_cnd(m)))
  )

  expect_identical(
    res$id[res$message == "boom"],
    cnd_id(new_blk_cnd("boom"))
  )
})

# CI hardening for the chromote-driven shinytest2 e2e tests.

# Give Chrome longer to open its remote-debugging port. chromote's default is
# 10s (`getOption("chromote.timeout", 10)` in launch_chrome_impl), too short on
# loaded CI runners -- Windows especially -- where launch intermittently aborts
# with "Chrome debugging port not open after 10 seconds". Scoped to the suite
# so the option is restored when testing finishes.
withr::local_options(
  chromote.timeout = 30,
  .local_envir = testthat::teardown_env()
)

# Chrome leaves scratch dirs (com.google.Chrome.* / org.chromium.Chromium.*,
# scoped_dir variants included) in the session temp parent (`$TMPDIR`), which
# R CMD check then flags as "detritus in the temp directory" -- a NOTE the CI
# gate fails on. `app$stop()` only ends the shinytest2 session; the shared
# browser stays alive until R exits, so its scratch is never reclaimed. At the
# end of the suite, close the browser (a clean shutdown reclaims its scratch),
# then sweep anything that still leaked -- but only what this run created, so a
# shared `$TMPDIR` on a developer machine keeps its other dirs.
local({

  pat <- "^(com\\.google\\.Chrome|org\\.chromium\\.Chromium)"
  tmp_parent <- dirname(tempdir())
  before <- list.files(tmp_parent, pattern = pat)

  withr::defer(
    {
      if (isTRUE(chromote::has_default_chromote_object())) {
        try(chromote::default_chromote_object()$close(), silent = TRUE)
      }

      unlink(
        file.path(
          tmp_parent,
          setdiff(list.files(tmp_parent, pattern = pat), before)
        ),
        recursive = TRUE,
        force = TRUE
      )
    },
    testthat::teardown_env()
  )
})

#' Remove entries from a `reactiveValues` object
#'
#' shiny offers no public way to delete a key from a [shiny::reactiveValues()]
#' object -- assigning `NULL` stores a `NULL` value but leaves the key in
#' `names()`. `trim_rv()` removes the named entries outright and invalidates
#' the affected reactive dependencies, so a key can be truly dropped (and
#' later re-added) -- for instance when a variadic block argument is unlinked.
#'
#' @param x A `reactiveValues` object.
#' @param rm Character vector of keys to remove; all must be present in `x`.
#'
#' @return `x`, invisibly.
#'
#' @export
trim_rv <- function(x, rm) {

  stopifnot(is.reactivevalues(x))

  internals <- .subset2(x, "impl")

  stopifnot(all(rm %in% internals$.nameOrder))

  # No public removal exists, so mirror the internals ReactiveValues$set()
  # touches: drop the value and its name, then invalidate the same
  # dependencies an update would (the key's own, plus names / as-list).
  for (key in rm) {

    internals$.values$remove(key)

    if (internals$.dependents$containsKey(key)) {
      internals$.dependents$get(key)$invalidate()
    }
  }

  internals$.nameOrder <- setdiff(internals$.nameOrder, rm)

  if (isTRUE(internals$.hasRetrieved$names)) {
    internals$.namesDeps$invalidate()
  }

  if (isTRUE(internals$.hasRetrieved$asList)) {
    internals$.valuesDeps$invalidate()
  }

  if (isTRUE(internals$.hasRetrieved$asListAll)) {
    internals$.allValuesDeps$invalidate()
  }

  invisible(x)
}

make_read_only <- function(x) {

  stopifnot(is.reactivevalues(x))

  res <- unclass(x)
  res[["readonly"]] <- TRUE
  class(res) <- class(x)

  res
}

destroy_outputs <- function(ns_prefix, session = get_session()) {

  for (id in starts_with(list_outputs(session), ns_prefix)) {
    destroy_output(id, session)
  }

  invisible()
}

list_outputs <- function(session = get_session()) {
  coal(
    union(
      names(session$.__enclos_env__$private$.outputs),
      names(session$.__enclos_env__$private$.outputOptions)
    ),
    character()
  )
}

destroy_output <- function(id, session = get_session()) {

  log_trace("destroying output {id}")

  session$defineOutput(id, NULL, NULL)
  session$.__enclos_env__$private$.outputs[[id]] <- NULL
  session$.__enclos_env__$private$.outputOptions[[id]] <- NULL

  invisible()
}

destroy_inputs <- function(ns_prefix, session = get_session()) {

  for (id in starts_with(names(session$input), ns_prefix)) {
    destroy_input(id, session)
  }

  invisible()
}

destroy_input <- function(id, session = get_session()) {

  log_trace("destroying input {id}")

  session$manageInputs(
    set_names(list(NULL), id),
    now = TRUE
  )

  input <- .subset2(session$input, "impl")

  input$.values$remove(id)
  input$.nameOrder <- setdiff(input$.nameOrder, id)

  invisible()
}

invalidate_inputs <- function(session = get_session()) {

  log_trace("invalidating inputs of domain {session$ns(NULL)}")

  input <- .subset2(session$input, "impl")

  input$.namesDeps$invalidate()
  input$.valuesDeps$invalidate()

  invisible()
}

destroy_observers <- function(ns_prefix, session = get_session()) {

  if (isTRUE(blockr_option("observe_hook_disabled", FALSE))) {
    return(invisible(FALSE))
  }

  obs <- get0("observers", envir = session$userData)

  if (is.null(obs)) {
    log_debug("cannot destroy uncaptured observers")
    return(invisible(FALSE))
  }

  for (i in starts_with(names(obs), ns_prefix)) {

    for (x in obs[[i]]) {

      log_trace(
        "destroying observer {x$.reactId} of domain {x$.domain$ns(NULL)}"
      )

      x$destroy()
    }

    obs[[i]] <- NULL
  }

  assign("observers", obs, envir = session$userData)

  invisible(TRUE)
}

#' Destroy a Shiny module
#'
#' Cleans up inputs, outputs and/or observers associated with a Shiny module.
#'
#' @note This function relies on non-documented, internal Shiny APIs that are
#' not officially supported or recommended to use. These internals may break
#' if Shiny's internal source code changes. Use with caution and only
#' whenever necessary.
#'
#' @param id Module namespace id.
#' @param what Character vector indicating which components to destroy.
#'   Defaults to all of `"inputs"`, `"outputs"` and `"observers"`.
#' @param session Shiny session object.
#'
#' @return The namespaced id (invisibly).
#'
#' @keywords internal
#' @export
destroy_module <- function(id, what = c("inputs", "outputs", "observers"),
                           session = get_session()) {

  what <- match.arg(what, several.ok = TRUE)

  log_debug("destroying module {id}, component{?s} {what}")

  ns <- session$ns(id)

  if ("inputs" %in% what) {
    destroy_inputs(id, session)
  }

  if ("outputs" %in% what) {
    destroy_outputs(ns, session)
  }

  if ("observers" %in% what) {
    destroy_observers(ns, session)
  }

  invisible(ns)
}

trace_observe <- function() {

  if (isTRUE(blockr_option("observe_hook_disabled", FALSE))) {
    return(invisible(FALSE))
  }

  if (is_observe_traced()) {
    return(invisible(FALSE))
  }

  log_debug("hooking observer capturing")

  suppressMessages(
    trace(
      shiny::observe,
      exit = quote(
        {
          if (!is.null(domain)) { # nocov start

            obs <- get0(
              "observers",
              envir = domain$userData,
              inherits = FALSE
            )

            if (is.null(obs)) {
              obs <- list()
            }

            dom <- domain$ns(NULL)

            if (!length(dom)) {
              dom <- ""
            }

            cur <- returnValue()

            blockr.core::log_trace(
              "capturing observer {cur$.reactId} of domain {dom}"
            )

            obs[[dom]] <- c(obs[[dom]], cur)

            assign("observers", obs, envir = domain$userData)
          } # nocov end
        }
      ),
      print = FALSE
    )
  )

  invisible(untrace_observe)
}

untrace_observe <- function() {

  if (isTRUE(blockr_option("observe_hook_disabled", FALSE))) {
    return(invisible(FALSE))
  }

  if (!is_observe_traced()) {
    return(invisible(FALSE))
  }

  log_debug("removing observer capture hook")

  suppressMessages(untrace(shiny::observe))

  invisible(TRUE)
}

is_observe_traced <- function() {
  inherits(shiny::observe, "functionWithTrace")
}

#' Shiny utilities
#'
#' Utility functions for shiny:
#' - `get_session`: See [shiny::getDefaultReactiveDomain()].
#' - `generate_plugin_args`: Meant for unit testing plugins.
#' - `notify`: Glue-capable wrapper for [shiny::showNotification()].
#'
#' @return Either `NULL` or a shiny session object for `get_session()`, a list
#' of arguments for plugin server functions in the case of
#' `generate_plugin_args()` and `notify()` is called for the side-effect of
#' displaying a browser notification (and returns `NULL` invisibly).
#'
#' @export
get_session <- function() {
  getDefaultReactiveDomain()
}

#' @param close_button Passed as `closeButton` to [shiny::showNotification()]
#' @param glue,log Whether to [glue::glue()]-interpolate `...` and whether to
#' emit a log message. Set both to `FALSE` to surface pre-formatted,
#' already-logged text (e.g. a captured condition message, which may contain
#' braces that would otherwise fail interpolation).
#'
#' @inheritParams write_log
#' @inheritParams shiny::showNotification
#'
#' @rdname get_session
#' @export
notify <- function(..., envir = parent.frame(), action = NULL, duration = 5,
                   close_button = TRUE, id = NULL,
                   type = c("message", "warning", "error"),
                   glue = TRUE, log = TRUE, session = get_session()) {

  type <- match.arg(type)

  msg <- if (glue) glue_plur(..., envir = envir) else paste0(...)
  msg <- HTML(cli::ansi_html(msg))

  showNotification(
    msg,
    action = action,
    duration = duration,
    closeButton = close_button,
    id = id,
    type = type,
    session = session
  )

  if (log) {
    switch(
      type,
      message = log_info(msg, envir = envir, use_glue = FALSE),
      warning = log_warn(msg, envir = envir, use_glue = FALSE),
      error = log_error(msg, envir = envir, use_glue = FALSE)
    )
  }

  invisible(NULL)
}

notify_remove <- function(id, session = get_session()) {
  removeNotification(id, session = session)
}

test_that("shiny utils", {

  with_mock_session(
    {
      expect_null(invalidate_inputs())
      expect_false(destroy_observers("test"))
    }
  )

  withr::with_options(
    list(blockr.observe_hook_disabled = TRUE),
    with_mock_session(
      expect_false(destroy_observers("test"))
    )
  )

  with_mock_session(
    {
      expect_null(invalidate_inputs())
      expect_false(destroy_observers("test"))
    }
  )

  withr::with_options(
    list(blockr.observe_hook_disabled = TRUE),
    {
      expect_false(trace_observe())
      expect_false(untrace_observe())
    }
  )

  withr::defer(
    {
      if (is_observe_traced()) {
        trace_observe()
      }
    }
  )

  withr::local_options(
    blockr.observe_hook_disabled = FALSE
  )

  untrace_observe()

  expect_false(untrace_observe())
  expect_type(trace_observe(), "closure")
  expect_false(trace_observe())
  expect_true(untrace_observe())
})

test_that("trim_rv removes keys, invalidates readers, and allows re-add", {

  with_mock_session(
    {
      rv <- reactiveValues(a = 1, b = 2, c = 3)

      seen <- NULL
      observe(seen <<- list(names = names(rv), b = rv[["b"]]))
      session$flushReact()

      expect_setequal(seen$names, c("a", "b", "c"))
      expect_identical(seen$b, 2)

      # Unlike `rv[["b"]] <- NULL` (which leaves a NULL-valued key behind),
      # trim_rv() drops the key outright and re-runs the reader.
      trim_rv(rv, "b")
      session$flushReact()

      expect_setequal(seen$names, c("a", "c"))
      expect_null(seen$b)

      # The slot is genuinely free, so the same key can be re-added.
      rv[["b"]] <- 99
      session$flushReact()

      expect_setequal(seen$names, c("a", "c", "b"))
      expect_identical(seen$b, 99)

      expect_error(trim_rv(rv, "absent"))
    }
  )
})

test_that("trim_rv runs outside a reactive context", {

  rv <- reactiveValues(a = 1, b = 2, c = 3)

  expect_silent(trim_rv(rv, "b"))
  expect_setequal(isolate(names(rv)), c("a", "c"))

  expect_error(trim_rv(rv, "absent"))
})

test_that("trim_rv invalidates reactiveValuesToList readers", {

  with_mock_session(
    {
      rv <- reactiveValues(a = 1, b = 2, c = 3)

      as_seen <- NULL
      all_seen <- NULL
      observe(as_seen <<- reactiveValuesToList(rv))
      observe(all_seen <<- reactiveValuesToList(rv, all.names = TRUE))
      session$flushReact()

      expect_true(all(c("a", "b", "c") %in% names(as_seen)))
      expect_true(all(c("a", "b", "c") %in% names(all_seen)))

      trim_rv(rv, "b")
      session$flushReact()

      expect_false("b" %in% names(as_seen))
      expect_false("b" %in% names(all_seen))
    }
  )
})

test_that("notify(glue = FALSE) surfaces literal text with braces", {

  shown <- NULL

  local_mocked_bindings(
    showNotification = function(ui, ...) {
      shown <<- as.character(ui)
      invisible()
    }
  )

  with_mock_session(
    notify(
      "object {x} not found",
      type = "error",
      glue = FALSE,
      log = FALSE,
      session = session
    )
  )

  expect_match(shown, "object {x} not found", fixed = TRUE)
})

test_that("notify() interpolates by default and errors on bad braces", {

  local_mocked_bindings(showNotification = function(ui, ...) invisible())

  with_mock_session(
    expect_error(
      notify("object {x} not found", type = "error", session = session)
    )
  )
})

test_that("notify(log = FALSE) skips the log entry", {

  logged <- new.env(parent = emptyenv())
  logged$n <- 0L

  local_mocked_bindings(
    showNotification = function(ui, ...) invisible(),
    log_error = function(...) {
      logged$n <- logged$n + 1L
      invisible()
    }
  )

  with_mock_session(
    {
      notify("boom", type = "error", glue = FALSE, log = FALSE,
             session = session)
      expect_identical(logged$n, 0L)

      notify("boom", type = "error", glue = FALSE, session = session)
      expect_identical(logged$n, 1L)
    }
  )
})

test_that("notify_remove removes a notification by id", {

  removed <- NULL

  local_mocked_bindings(
    removeNotification = function(id, ...) {
      removed <<- id
      invisible()
    }
  )

  with_mock_session(notify_remove("toast-1", session))

  expect_identical(removed, "toast-1")
})

#' @section Merge block:
#' Joining together two `data.frame`s, based on a set of index columns, using
#' [base::merge()] is available as `merge_block`. Depending on values passed
#' as `all_x`/`all_y` the result will correspond to an "inner", "outer", "left"
#' or "right" join. See [base::merge()] for details. This block class serves
#' as an example for a transform block that takes exactly two data inputs `x`
#' and `y` to produce a single `data.frame` as output.
#'
#' @param by Column(s) tp join on
#' @param all_x,all_y Join type, see [base::merge()]
#'
#' @block merge block
#' @blockDescr Joining or datasets
#' @blockCategory transform
#' @blockIcon union
#' @blockGuidance Joins the two inputs with `base::merge()`. `by` names the
#'   shared key column(s); omit it to join on all common columns. all_x and
#'   all_y set the join type: both FALSE for an inner join, both TRUE for a full
#'   outer join, all_x alone for a left join, all_y alone for a right join.
#' @blockKeywords merge, join, inner join, outer join, left join, right join
#' @blockArg by Shared key column(s) to join on.
#'   [type] arg_array(arg_string())
#' @blockArg all_x Keep unmatched rows from the first input.
#'   [type] arg_boolean()
#' @blockArg all_y Keep unmatched rows from the second input.
#'   [type] arg_boolean()
#' @blockExamples list(
#'   list(all_x = TRUE, all_y = FALSE),
#'   list(all_x = TRUE, all_y = TRUE)
#' )
#'
#' @rdname new_transform_block
#' @export
new_merge_block <- function(by = character(), all_x = FALSE, all_y = FALSE,
                            ...) {

  by_choices <- function(x, y) {
    intersect(colnames(x), colnames(y))
  }

  new_transform_block(
    function(id, x, y) {
      moduleServer(
        id,
        function(input, output, session) {

          sels <- reactiveVal(by)

          allx <- reactiveVal(all_x)
          ally <- reactiveVal(all_y)

          observeEvent(input$by, sels(input$by))

          observeEvent(
            input$type,
            {
              allx("all.x" %in% input$type)
              ally("all.y" %in% input$type)
            },
            ignoreNULL = FALSE,
            ignoreInit = TRUE
          )

          cols <- reactive(by_choices(x(), y()))

          observe(
            updateSelectInput(
              session,
              inputId = "by",
              choices = cols(),
              selected = sels()
            )
          )

          list(
            expr = reactive(
              bbquote(
                merge(.(x), .(y), by = .(cols), all.x = .(allx),
                      all.y = .(ally)),
                list(cols = sels(), allx = allx(), ally = ally())
              )
            ),
            state = list(
              by = sels,
              all_x = allx,
              all_y = ally
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        selectInput(
          inputId = NS(id, "by"),
          label = "By columns",
          choices = list(),
          multiple = TRUE
        ),
        checkboxGroupInput(
          inputId = NS(id, "type"),
          label = "Join type",
          choices = c("all.x", "all.y"),
          selected = c("all.x", "all.y")[c(all_x, all_y)]
        )
      )
    },
    dat_valid = function(x, y) {
      stopifnot(is.data.frame(x), is.data.frame(y))
    },
    expr_type = "bquoted",
    class = "merge_block",
    ...
  )
}

test_that("merge block constructor", {

  blk <- new_merge_block()

  expect_s3_class(blk, "merge_block")

  testServer(
    block_expr_server(blk),
    {
      expect_identical(sels(), character())
      expect_identical(allx(), FALSE)
      expect_identical(ally(), FALSE)

      expect_identical(session$returned$state$by(), character())
      expect_identical(session$returned$state$all_x(), FALSE)
      expect_identical(session$returned$state$all_y(), FALSE)

      session$flushReact()
      session$setInputs(by = "name", type = "all.y")

      expect_identical(sels(), "name")
      expect_identical(allx(), FALSE)
      expect_identical(ally(), TRUE)

      expect_identical(session$returned$state$by(), "name")
      expect_identical(session$returned$state$all_x(), FALSE)
      expect_identical(session$returned$state$all_y(), TRUE)
    },
    args = list(
      x = function() band_members,
      y = function() band_instruments
    )
  )
})

test_that("edit blocks", {

  testServer(
    edit_block_server,
    {
      expect_null(update())

      session$setInputs(block_name_in = "iris")

      session$flushReact()

      res <- update()$blocks

      expect_length(res$add, 0L)
      expect_length(res$rm, 0L)
      expect_length(res$mod, 1L)

      expect_named(res$mod, "a")
      expect_identical(res$mod[["a"]], list(block_name = "iris"))

      expect_null(session$returned)
    },
    args = list(
      block_id = "a",
      board = reactiveValues(
        board = new_board(blocks = list(a = new_dataset_block("iris")))
      ),
      update = reactiveVal()
    )
  )

  testServer(
    edit_block_server,
    {
      expect_null(update())

      session$setInputs(
        add_block_after = 1,
        registry_select = "subset_block",
        block_id = "c",
        input_select = "data",
        link_id = "bc",
        confirm_insert = 1
      )

      session$flushReact()

      blk <- update()$blocks

      expect_length(blk$add, 1L)
      expect_length(blk$mod, 0L)
      expect_length(blk$rm, 0L)

      lnk <- update()$links

      expect_length(lnk$add, 1L)
      expect_length(lnk$mod, 0L)
      expect_length(lnk$rm, 0L)

      expect_null(session$returned)
    },
    args = list(
      block_id = "b",
      board = reactiveValues(
        board = new_board(
          blocks = c(
            a = new_dataset_block("iris"),
            b = new_merge_block()
          ),
          links = list(from = "a", to = "b", input = "x"),
          stacks = list(ab = c("a", "b"))
        )
      ),
      update = reactiveVal()
    )
  )

  testServer(
    edit_block_server,
    {
      expect_null(update())

      session$setInputs(
        add_block_before = 1,
        registry_select = "dataset_block",
        block_id = "c",
        input_select = "y",
        link_id = "cb",
        confirm_insert = 1
      )

      session$flushReact()

      blk <- update()$blocks

      expect_length(blk$add, 1L)
      expect_length(blk$mod, 0L)
      expect_length(blk$rm, 0L)

      lnk <- update()$links

      expect_length(lnk$add, 1L)
      expect_length(lnk$mod, 0L)
      expect_length(lnk$rm, 0L)

      expect_null(session$returned)
    },
    args = list(
      block_id = "b",
      board = reactiveValues(
        board = new_board(
          blocks = c(
            a = new_dataset_block("iris"),
            b = new_merge_block()
          ),
          links = list(from = "a", to = "b", input = "x"),
          stacks = list(ab = c("a", "b"))
        )
      ),
      update = reactiveVal()
    )
  )

  testServer(
    edit_block_server,
    {
      expect_null(update())

      session$setInputs(
        add_block_before = 1,
        registry_select = "dataset_block",
        block_id = "c",
        input_select = "y",
        link_id = "cb",
        add_to_stack = TRUE,
        confirm_insert = 1
      )

      session$flushReact()

      blk <- update()$blocks

      expect_length(blk$add, 1L)
      expect_length(blk$mod, 0L)
      expect_length(blk$rm, 0L)

      lnk <- update()$links

      expect_length(lnk$add, 1L)
      expect_length(lnk$mod, 0L)
      expect_length(lnk$rm, 0L)

      stk <- update()$stacks

      expect_length(stk$add, 1L)
      expect_length(stk$mod, 0L)
      expect_length(stk$rm, 1L)

      expect_null(session$returned)
    },
    args = list(
      block_id = "b",
      board = reactiveValues(
        board = new_board(
          blocks = c(
            a = new_dataset_block("iris"),
            b = new_merge_block()
          ),
          links = list(from = "a", to = "b", input = "x"),
          stacks = list(ab = c("a", "b"))
        )
      ),
      update = reactiveVal()
    )
  )
})

test_that("dummy edit block ui test", {
  expect_s3_class(
    edit_block_ui(new_dataset_block(), "edit_block"),
    "shiny.tag.list"
  )
})

#' @import vctrs shiny
#' @keywords internal
#' @importFrom cli qty
#' @importFrom grDevices col2rgb
#' @importFrom graphics plot
#' @importFrom methods is
NULL

# `datasets` has no function exports (only lazy data), so @importFrom
# can't anchor it; reference `iris` here to silence the unused-Imports NOTE.
ignore_unused_imports <- function() datasets::iris

#' @param env An environment that is resolved to a package name
#' @rdname set_names
#' @export
pkg_name <- function(env = parent.frame()) {
  utils::packageName(env)
}

#' @param pkg A string.valued package name or an environment passed to
#' [pkg_name()]
#' @rdname set_names
#' @export
pkg_version <- function(pkg = parent.frame()) {

  if (is.environment(pkg)) {
    pkg <- pkg_name(pkg)
  }

  utils::packageVersion(pkg)
}

#' @rdname set_names
#' @export
pkg_file <- function(..., pkg = parent.frame()) {
  system.file(..., package = pkg_name(pkg))
}

#' @rdname set_names
#' @export
pkg_avail <- function(...) {
  pkgs <- c(...)
  log_trace("checking availability of package{?s} {pkgs}")
  all(lgl_ply(pkgs, requireNamespace, quietly = TRUE))
}

#' Include a pre-rendered mermaid diagram
#'
#' Renders mermaid `.mmd` source to `.svg` when the source is newer than the
#' existing SVG and `mmdc` is available, then includes the SVG via
#' [knitr::include_graphics()].
#'
#' @param name Diagram name (without extension), resolved relative to a
#'   `mermaid/` directory alongside the calling document.
#' @param chromium_args Character vector of extra Chromium flags passed to
#'   `mmdc` via a temporary puppeteer config file.
#'
#' @keywords internal
#'
#' @export
include_mermaid <- function(name,
                            chromium_args = c("--no-sandbox")) {

  mmd <- file.path("mermaid", paste0(name, ".mmd"))
  svg <- file.path("mermaid", paste0(name, ".svg"))

  needs_render <- !file.exists(svg) ||
    (file.exists(mmd) && file.mtime(mmd) > file.mtime(svg))

  if (needs_render && nzchar(Sys.which("mmdc"))) {
    cfg <- tempfile(fileext = ".json")
    on.exit(unlink(cfg))
    writeLines(
      sprintf('{"args":[%s]}', paste0('"', chromium_args, '"', collapse = ",")),
      cfg
    )
    system2(
      "mmdc",
      c("-i", mmd, "-o", svg, "--puppeteerConfigFile", cfg, "--quiet")
    )
  }

  knitr::include_graphics(svg)
}

#' Random IDs
#'
#' Randomly generated unique IDs are used throughout the package, created by
#' `rand_names()`. If random strings are required that may not clash with a set
#' of existing values, this can be guaranteed by passing them as `old_names`.
#' A [blockr_option()] `rand_id` can be set to swap out the function responsible
#' for ID generation.
#'
#' @param old_names Disallowed IDs
#' @param n Number of IDs to generate
#' @param max_tries Max number of attempts to create IDs that do not intersect
#' with `old_names`
#' @param id_fun A function with a single argument `n` that generates random
#' IDs. A value of `NULL` defaults to [ids::adjective_animal()] if available and
#' `sample_letters` otherwise.
#'
#' @examples
#' rand_names()
#' rand_names(n = 5L)
#' rand_names(id_fun = sample_letters)
#'
#' @return A character vector of length `n` where each entry contains `length`
#' characters (all among `chars` and start/end with `prefix`/`suffix`), is
#' guaranteed to be unique and not present among values passed as `old_names`.
#'
#' @export
rand_names <- function(old_names = character(0L), n = 1L, max_tries = 100L,
                       id_fun = blockr_option("rand_id", NULL)) {

  stopifnot(is.character(old_names), is_count(n), is_count(max_tries))

  if (is.null(id_fun)) {
    if (pkg_avail("ids")) {
      id_fun <- adjective_animal
    } else {
      id_fun <- sample_letters
    }
  } else {
    stopifnot(is.function(id_fun), identical(names(formals(id_fun)), "n"))
  }

  new_names <- character(0L)
  counter <- 0L

  while (length(new_names) < n && counter < max_tries) {

    counter <- counter + 1L

    candidates <- unique(id_fun(n - length(new_names)))

    collisions <- candidates %in% c(old_names, new_names)

    new_names <- c(new_names, candidates[!collisions])
  }

  if (length(new_names) < n) {
    blockr_abort(
      "Failed to create {n} unique ID{?s} within {max_tries} attempt{?s}.",
      class = "id_creation_unsuccessful"
    )
  }

  new_names
}

#' @rdname rand_names
#' @export
adjective_animal <- function(n) {
  ids::adjective_animal(n, max_len = 8L)
}

#' @rdname rand_names
#' @export
sample_letters <- function(n) {
  paste(sample(letters, 8, replace = TRUE), collapse = "")
}

inherits <- function(x, ..., agg = NULL) {

  res <- lgl_ply(c(...), function(y) base::inherits(x, y))

  if (is.null(agg)) {
    return(res)
  }

  agg(res)
}

#' Miscellaneous utilities
#'
#' Several internal utility functions are exported for convenience in case
#' dependent packages can make use of this functionality.
#'
#' @param object,nm See [stats::setNames()]
#'
#' @return Function `set_names()` returns a names object, `paste_enum()` a
#' string, `coal()` the first non-null object and `unlst()` performs the same
#' action as [base::unlist()] but with differing defaults.
#'
#' @keywords internal
#'
#' @export
set_names <- function(object = nm, nm) {
  names(object) <- nm
  object
}

#' @param x Character vector to [base::paste()]
#' @param sep,conj Separation strings for all but last and last positions
#' @param quotes Quotes to wrap each entry in `x` with
#'
#' @rdname set_names
#' @export
paste_enum <- function(x, sep = ", ", conj = " and ", quotes = "`") {

  if (length(x) == 0L) {
    return(x)
  }

  if (length(x) == 1L) {
    return(paste0(quotes, x, quotes))
  }

  paste0(
    paste0(quotes, x[seq_len(length(x) - 1L)], quotes, collapse = sep),
    conj,
    quotes, x[length(x)], quotes
  )
}

#' @param ... Set of objects to iterate over
#' @param fail_all Error if no non-null objects are present
#' @param test_fun Function to test each element with
#'
#' @rdname set_names
#' @export
coal <- function(..., fail_all = TRUE, test_fun = is.null) {

  for (i in seq_len(...length())) {
    x <- ...elt(i)
    if (test_fun(x)) next else return(x)
  }

  if (isTRUE(fail_all)) {
    blockr_abort(
      "No value encountered that does not fail `test_fun`.",
      class = "coal_null_return_disallowed"
    )
  }

  x
}

#' @rdname set_names
#' @export
reval <- function(x) x()

#' @rdname set_names
#' @export
reval_if <- function(x) if (is.function(x)) x() else x

lst_xtr_reval <- function(x, ...) {
  lapply(lst_xtr(x, ...), reval)
}

#' @param recursive,use_names See [base::unlist()]
#' @rdname set_names
#' @export
unlst <- function(x, recursive = FALSE, use_names = FALSE) {
  unlist(x, recursive = recursive, use.names = use_names)
}

na_to_empty <- function(x) replace(x, is.na(x), "")

#' Blockr Options
#'
#' Retrieves options via [base::getOption()] or [base::Sys.getenv()], in that
#' order, and prefixes the option name passed as `name` with `blockr.` or
#' `blockr_` respectively. Additionally, the name is converted to lower case
#' for `getOption()` and upper case for environment variables. In case no value
#' is available for a given `name`, `default` is returned.
#'
#' @param name Option name
#' @param default Default value
#'
#' @examples
#' blockr_option("test-example", "default")
#'
#' options(`blockr.test-example` = "non-default")
#' blockr_option("test-example", "default")
#'
#' Sys.setenv(`BLOCKR_TEST-EXAMPLE` = "another value")
#' tryCatch(
#'   blockr_option("test-example", "default"),
#'   error = function(e) conditionMessage(e)
#' )
#' options(`blockr.test-example` = NULL)
#' blockr_option("test-example", "default")
#'
#' Sys.unsetenv("BLOCKR_TEST-EXAMPLE")
#' blockr_option("test-example", "default")
#'
#' @return The value set as option `name` or `default` if not set. In case of
#' the option being available only as environment variable, the value will be
#' a string and if available as [base::options()] entry it may be of any R type.
#'
#' @export
blockr_option <- function(name, default) {

  opt <- tolower(paste0("blockr.", name))
  env <- toupper(paste0("blockr_", name))

  res_opt <- getOption(opt, default = NULL)
  res_env <- Sys.getenv(env, unset = "")

  if (is.null(res_opt) && identical(res_env, "")) {
    return(default)
  }

  if (identical(res_opt, res_env)) {
    return(res_opt)
  }

  if (is.null(res_opt) && !identical(res_env, "")) {
    return(res_env)
  }

  if (identical(res_env, "") && !is.null(res_opt)) {
    return(res_opt)
  }

  blockr_abort(
    "Conflicting options set for {name}: check environment variable {env} ",
    "and option {opt}.",
    class = "conflicing_blockr_option"
  )
}

#' @param ... Option key value pairs as named arguments
#' @rdname blockr_option
#' @export
set_blockr_options <- function(...) {

  if (...length() == 0L) {
    return(invisible())
  }

  vals <- list(...)
  name <- names(vals)

  stopifnot(
    length(vals) == length(name),
    all(nzchar(name)),
    anyDuplicated(name) == 0L
  )

  env <- toupper(paste0("blockr_", name))
  evl <- nzchar(Sys.getenv(env))

  if (any(evl)) {
    Sys.unsetenv(env[evl])
  }

  names(vals) <- tolower(paste0("blockr.", name))

  do.call(options, vals)
}

dot_sym <- function(i) {
  paste0(".arg", i)
}

arg_refs <- function(nms) {

  unnamed <- !nzchar(nms)

  replace(nms, unnamed, dot_sym(seq_len(sum(unnamed))))
}

dot_arg_refs <- function(x) {

  nms <- names(x)

  if (is.null(nms)) {
    nms <- character(length(x))
  }

  set_names(arg_refs(nms), nms)
}

dot_arg_values <- function(x) {
  set_names(as.list(x), unname(dot_arg_refs(x)))
}

exprs_to_lang <- function(exprs) {

  if (rlang::is_syntactic_literal(exprs)) {
    return(exprs)
  }

  if (is.expression(exprs)) {
    exprs <- as.list(exprs)
  }

  if (is.list(exprs)) {
    exprs <- as.call(c(quote(`{`), exprs))
  }

  stopifnot(typeof(exprs) == "language")

  exprs
}

starts_with <- function(x, prefix) {
  x[startsWith(x, prefix)]
}

#' @param ctor Function (either a string, a function or number used to index
#' the call stack
#' @param ctor_pkg The package where `ctor` is defined (either a string or
#' `NULL` which will use the function environment)
#'
#' @rdname rand_names
#' @export
resolve_ctor <- function(ctor, ctor_pkg = NULL) {

  try <- NULL

  if (is.numeric(ctor)) {

    if (ctor < 0L) {
      func <- rlang::caller_fn(-ctor)
      call <- as.character(rlang::caller_call(-ctor)[[1L]])
    } else {
      func <- sys.function(ctor)
      call <- deparse(sys.call(ctor)[[1L]])
    }

    if (any(grepl("::", call, fixed = TRUE))) {

      if (length(call) == 1L) {
        call <- strsplit(call, "::", fixed = TRUE)[[1L]]
      } else {
        stopifnot(length(call) == 3L)
        call <- call[-1L]
      }

      stopifnot(length(call) == 2L)

      ctor <- call[2L]

      if (is.null(ctor_pkg)) {
        ctor_pkg <- call[1L]
      } else {
        stopifnot(identical(ctor_pkg, call[1L]))
      }

    } else {

      if (is.null(ctor_pkg)) {
        ctor_pkg <- pkg_name(environment(func))
      }

      if (not_null(ctor_pkg)) {
        ctor <- call
      }
    }

    if (is_string(ctor_pkg) && is_string(ctor)) {
      try <- get0(ctor, asNamespace(ctor_pkg), mode = "function",
                  inherits = FALSE)
    }

    if (is.null(try)) {
      ctor <- func
      ctor_pkg <- NULL
    }
  }

  if (is.null(ctor_pkg)) {
    return(new_blockr_ctor(ctor))
  }

  new_blockr_ctor(try, ctor, ctor_pkg)
}

#' @rdname rand_names
#' @export
forward_ctor <- function(x) {

  x <- coal(x, 0L)

  if (is.numeric(x)) {

    stopifnot(is_scalar(x), is_intish(x))

    if (x > 0L) {
      x <- -x
    }

    x <- x - 1L
  }

  x
}

new_blockr_ctor <- function(fun, nme = NULL, pkg = NULL) {

  if (is.null(nme) && is.null(pkg)) {

    if (is.function(fun)) {
      return(structure(fun, class = "blockr_ctor"))
    }

    stopifnot(is_string(fun))

    fun <- strsplit(fun, "::", fixed = TRUE)[[1L]]

    stopifnot(length(fun) == 2L)

    nme <- fun[2L]
    pkg <- fun[1L]
    fun <- NULL
  }

  stopifnot(is_string(nme), is_string(pkg))

  structure(
    coal(fun, get_ctor_from_pkg(nme, pkg)),
    fun = nme,
    pkg = pkg,
    class = "blockr_ctor"
  )
}

get_ctor_from_pkg <- function(name, package) {
  get0(name, asNamespace(package), mode = "function", inherits = FALSE)
}

#' @rdname rand_names
#' @export
is_blockr_ctor <- function(x) {
  inherits(x, "blockr_ctor")
}

#' @rdname rand_names
#' @export
ctor_name <- function(x) {
  stopifnot(is_blockr_ctor(x))
  attr(x, "fun")
}

#' @rdname rand_names
#' @export
ctor_pkg <- function(x) {
  stopifnot(is_blockr_ctor(x))
  attr(x, "pkg")
}

#' @rdname rand_names
#' @export
ctor_fun <- function(x) {
  stopifnot(is_blockr_ctor(x))
  attributes(x) <- NULL
  x
}

#' @param x Character vector to transform
#' @param replace,with Mapped to [base::gsub()]
#' @rdname rand_names
#' @export
to_sentence_case <- function(x, replace = character(), with = character()) {

  to_case_collapse <- function(x) {

    if (length(x) == 0L) {
      return("")
    }

    first <- x[1L]
    first <- paste0(toupper(substr(first, 1, 1)), tolower(substring(first, 2)))

    if (length(x) == 1L) {
      return(first)
    }

    paste(first, paste0(tolower(x[-1L]), collapse = " "))
  }

  if (is.null(x)) {
    x <- ""
  }

  len_rep <- length(replace)
  len_wit <- length(with)

  stopifnot(
    is.character(x), is.character(replace), is.character(with),
    length(x) > 0, len_wit == length(replace) || len_wit == 1L
  )

  x <- replace(x, is.na(x), "")

  if (len_wit == 1L) {
    with <- rep(with, len_rep)
  }

  for (i in seq_len(len_rep)) {
    x <- gsub(replace[i], with[i], x)
  }

  chr_ply(strsplit(trimws(x), " ", fixed = TRUE), to_case_collapse)
}

#' @rdname rand_names
#' @export
id_to_sentence_case <- function(x) {
  to_sentence_case(x, c("([A-Z])", "_", "-", "\\."), c(" \\1", " ", " ", " "))
}

format_head <- function(x, n = 3L) {

  if (length(x) > 1) {

    res <- paste(chr_ply(utils::head(x, n = n), format), collapse = ", ")

    if (length(x) > n) {
      res <- paste0(res, ", ...")
    }

    return(res)
  }

  format(x)
}

get_attr <- function(i, x) attr(x, i, exact = TRUE)

test_that("merge app", {

  skip_on_cran()

  app_path <- system.file("examples", "block", "merge", "app.R",
                          package = "blockr.core")

  app <- try(
    shinytest2::AppDriver$new(
      app_path,
      name = "merge",
      seed = 42,
      load_timeout = 30 * 1000
    )
  )

  testthat::skip_if(
    inherits(app, "try-error"),
    "Cannot start shinytest2 merge app."
  )

  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`block-expr-by` = "Time")
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`block-expr-type` = "all.x")
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`block-expr-type` = c("all.x", "all.y"))
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`block-expr-type` = "all.y")
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`block-expr-type` = character(0))
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$stop()
})

test_that("rbind app", {

  skip_on_cran()

  app_path <- system.file("examples", "block", "rbind", "app.R",
                          package = "blockr.core")

  app <- try(
    shinytest2::AppDriver$new(
      app_path,
      name = "rbind",
      seed = 42,
      load_timeout = 30 * 1000
    )
  )

  testthat::skip_if(
    inherits(app, "try-error"),
    "Cannot start shinytest2 rbind app."
  )

  on.exit(app$stop(), add = TRUE)

  res <- app$get_values(export = TRUE)$export[["result"]]

  expect_s3_class(res, "data.frame")
  expect_identical(dim(res), c(2L, 1L))
  expect_identical(res[["a"]], c(1, 2))
})

test_that("board app", {

  skip_on_cran()

  app_path <- system.file("examples", "board", "empty", "app.R",
                          package = "blockr.core")

  app <- try(
    shinytest2::AppDriver$new(
      app_path,
      name = "board",
      seed = 42,
      load_timeout = 30 * 1000
    )
  )

  testthat::skip_if(
    inherits(app, "try-error"),
    "Cannot start shinytest2 empty board app."
  )

  app$click("my_board-manage_blocks-add_block")
  app$wait_for_value(input = "my_board-manage_blocks-block_id")

  app$set_inputs(`my_board-manage_blocks-block_id` = "a")
  app$set_inputs(`my_board-manage_blocks-registry_select` = "dataset_block")

  app$wait_for_idle()
  app$click("my_board-manage_blocks-confirm_add")

  app$wait_for_value(input = "my_board-block_a-expr-dataset")
  app$set_inputs(`my_board-block_a-expr-dataset` = "BOD")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$click("my_board-manage_blocks-add_block")
  app$wait_for_value(input = "my_board-manage_blocks-block_id")

  app$set_inputs(`my_board-manage_blocks-block_id` = "b")
  app$set_inputs(`my_board-manage_blocks-registry_select` = "dataset_block")

  app$wait_for_idle()
  app$click("my_board-manage_blocks-confirm_add")

  app$wait_for_value(input = "my_board-block_b-expr-dataset")
  app$set_inputs(`my_board-block_b-expr-dataset` = "ChickWeight")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$click("my_board-manage_blocks-add_block")
  app$wait_for_value(input = "my_board-manage_blocks-block_id")

  app$set_inputs(`my_board-manage_blocks-block_id` = "c")
  app$set_inputs(`my_board-manage_blocks-registry_select` = "merge_block")

  app$wait_for_idle()
  app$click("my_board-manage_blocks-confirm_add")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$click("my_board-manage_links-links_mod")

  app$wait_for_idle()

  app$set_inputs(`my_board-manage_links-new_link_id` = "ac")
  app$click("my_board-manage_links-add_link")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-ac_from` = "a")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-ac_to` = "c")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-ac_input` = "x")

  app$set_inputs(`my_board-manage_links-new_link_id` = "bc")
  app$click("my_board-manage_links-add_link")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-bc_from` = "b")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-bc_to` = "c")
  app$wait_for_idle()
  app$set_inputs(`my_board-manage_links-bc_input` = "y")

  app$click("my_board-manage_links-modify_links")

  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`my_board-block_c-expr-by` = "Time")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$stop()
})

test_that("board stacks", {

  skip_on_cran()

  app_path <- system.file("examples", "board", "stack", "app.R",
                          package = "blockr.core")

  app <- try(
    shinytest2::AppDriver$new(
      app_path,
      name = "stack",
      seed = 42,
      load_timeout = 30 * 1000
    )
  )

  testthat::skip_if(
    inherits(app, "try-error"),
    "Cannot start shinytest2 board stacks app."
  )

  app$click("my_board-manage_stacks-stacks_mod")
  app$wait_for_value(input = "my_board-manage_stacks-ac_blocks")
  app$set_inputs(`my_board-manage_stacks-ac_blocks` = c("a", "c", "b"))
  app$click("my_board-manage_stacks-modify_stacks")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$set_inputs(`my_board-stack_ac-stack_name_in` = "abc")
  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$click("my_board-stack_ac-rm_stack")
  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$click("my_board-manage_stacks-stacks_mod")
  app$set_inputs(`my_board-manage_stacks-new_stack_id` = "abc")
  app$click("my_board-manage_stacks-add_stack")
  app$wait_for_value(input = "my_board-manage_stacks-abc_name")
  app$set_inputs(`my_board-manage_stacks-abc_name` = "ABC")
  app$set_inputs(`my_board-manage_stacks-abc_blocks` = c("a", "b", "c"))
  app$click("my_board-manage_stacks-modify_stacks")

  app$wait_for_idle()
  app$expect_values(export = TRUE, screenshot_args = FALSE)

  app$stop()
})

test_that("exprs_to_lang", {

  eval_exprs <- function(exprs) {
    eval(exprs_to_lang(exprs), list())
  }

  deparse_exprs <- function(exprs) {
    deparse(call("local", exprs_to_lang(exprs)))
  }

  expected_eval <- 3
  expected_dparse <- c(
    "local({",
    "    x <- 1",
    "    y <- 2",
    "    x + y", "})"
  )

  a <- quote(
    {
      x <- 1
      y <- 2
      x + y
    }
  )

  expect_identical(eval_exprs(a), expected_eval)
  expect_identical(deparse_exprs(a), expected_dparse)

  b <- parse(
    text = "
      x <- 1
      y <- 2
      x + y
    "
  )

  expect_identical(eval_exprs(b), expected_eval)
  expect_identical(deparse_exprs(b), expected_dparse)

  c <- call("{", quote(x <- 1), quote(y <- 2), quote(x + y))

  expect_identical(eval_exprs(c), expected_eval)
  expect_identical(deparse_exprs(c), expected_dparse)

  d <- list(quote(x <- 1), quote(y <- 2), quote(x + y))

  expect_identical(eval_exprs(d), expected_eval)
  expect_identical(deparse_exprs(d), expected_dparse)

  e <- expression(x <- 1, y <- 2, x + y)

  expect_identical(eval_exprs(e), expected_eval)
  expect_identical(deparse_exprs(e), expected_dparse)

  f <- quote(1 + 2)

  expect_identical(eval_exprs(f), expected_eval)
  expect_identical(deparse_exprs(f), "local(1 + 2)")

  g <- quote(sum(1, 2))

  expect_identical(eval_exprs(g), expected_eval)
  expect_identical(deparse_exprs(g), "local(sum(1, 2))")

  f <- 3

  expect_identical(eval_exprs(f), expected_eval)
  expect_identical(deparse_exprs(f), "local(3)")
})

test_that("rand_names()", {

  id_1 <- rand_names(n = 1L)

  expect_type(id_1, "character")
  expect_length(id_1, 1L)
  expect_true(grepl("_", id_1, fixed = TRUE))

  expect_identical(
    rand_names(n = 1L, id_fun = function(n) rep("a", n)),
    "a"
  )

  expect_error(
    rand_names(n = 2L, max_tries = 5, id_fun = function(n) rep("a", n)),
    class = "id_creation_unsuccessful"
  )

  expect_error(
    rand_names("a", n = 1L, max_tries = 5, id_fun = function(n) rep("a", n)),
    class = "id_creation_unsuccessful"
  )

  id_2 <- with_mocked_bindings(
    rand_names(),
    pkg_avail = function(x) {
      if (identical(x, "ids")) {
        FALSE
      } else {
        requireNamespace(x, quietly = TRUE)
      }
    }
  )

  expect_identical(nchar(id_2), 8L)
  expect_type(id_2, "character")
  expect_length(id_2, 1L)
  expect_false(grepl("_", id_2, fixed = TRUE))
})

test_that("sentence case", {

  expect_identical(
    to_sentence_case("tiny_aidi", "_", " "),
    "Tiny aidi"
  )

  expect_identical(
    to_sentence_case("key_sick_puma", "_", " "),
    "Key sick puma"
  )

  expect_identical(
    to_sentence_case("HAZY_PINK_RHEA", "_", " "),
    "Hazy pink rhea"
  )

  expect_identical(
    to_sentence_case(c("used_sad_olm", "bad_numb_rhea"), "_", " "),
    c("Used sad olm", "Bad numb rhea")
  )

  expect_identical(
    to_sentence_case("firmHare", "([A-Z])", " \\1"),
    "Firm hare"
  )

  expect_identical(
    to_sentence_case("thinCuteDove", "([A-Z])", " \\1"),
    "Thin cute dove"
  )

  expect_identical(
    to_sentence_case("FitFunFly", "([A-Z])", " \\1"),
    "Fit fun fly"
  )

  expect_identical(
    to_sentence_case("313d964aca2"),
    "313d964aca2"
  )

  expect_identical(
    to_sentence_case("a13d964aca2"),
    "A13d964aca2"
  )

  expect_identical(
    to_sentence_case("d12c577e-f94a-4d8a-b6cf-eab63b854fcf"),
    "D12c577e-f94a-4d8a-b6cf-eab63b854fcf"
  )

  expect_identical(
    to_sentence_case("d12c577e-f94a-4d8a-b6cf-eab63b854fcf", "-", " "),
    "D12c577e f94a 4d8a b6cf eab63b854fcf"
  )

  expect_identical(
    to_sentence_case(""),
    ""
  )

  expect_identical(
    to_sentence_case(NA_character_),
    ""
  )

  expect_identical(
    to_sentence_case(NULL),
    ""
  )

  expect_identical(
    id_to_sentence_case(
      c("thinCuteDove", "gold-busy-rail", "key_sick_puma", "left.tart.naga")
    ),
    c("Thin cute dove", "Gold busy rail", "Key sick puma", "Left tart naga")
  )
})

test_that("misc", {

  expect_identical(paste_enum(character()), character())

  expect_error(
    coal(NULL),
    class = "coal_null_return_disallowed"
  )

  res_1 <- withr::with_options(
    list(blockr.test_opt = "b"),
    blockr_option("test_opt", "a")
  )

  expect_identical("a", blockr_option("test_opt", "a"))

  expect_identical(
    "b",
    withr::with_options(
      list(blockr.test_opt = "b"),
      blockr_option("test_opt", "a")
    )
  )

  expect_identical(
    "b",
    withr::with_envvar(
      list(BLOCKR_TEST_OPT = "b"),
      blockr_option("test_opt", "a")
    )
  )

  expect_identical(
    "b",
    withr::with_options(
      list(blockr.test_opt = "b"),
      withr::with_envvar(
        list(BLOCKR_TEST_OPT = "b"),
        blockr_option("test_opt", "a")
      )
    )
  )

  expect_error(
    withr::with_options(
      list(blockr.test_opt = "b"),
      withr::with_envvar(
        list(BLOCKR_TEST_OPT = "c"),
        blockr_option("test_opt", "a")
      )
    ),
    class = "conflicing_blockr_option"
  )

  withr::defer(
    {
      Sys.unsetenv("BLOCKR_TEST_OPT")
      options(`blockr.test_opt` = NULL)
    }
  )

  expect_null(set_blockr_options())

  withr::with_options(
    list(blockr.test_opt = NULL),
    expect_identical(
      set_blockr_options(test_opt = "a"),
      list(blockr.test_opt = NULL)
    )
  )

  withr::with_options(
    list(blockr.test_opt = "a"),
    expect_identical(
      set_blockr_options(test_opt = "b"),
      list(blockr.test_opt = "a")
    )
  )

  withr::with_envvar(
    list(BLOCKR_TEST_OPT = "b"),
    withr::with_options(
      list(blockr.test_opt = "a"),
      {
        expect_identical(
          set_blockr_options(test_opt = "c"),
          list(blockr.test_opt = "a")
        )
        expect_identical(blockr_option("test_opt"), "c")
        expect_identical(Sys.getenv("BLOCKR_TEST_OPT"), "")
      }
    )
  )

  expect_s3_class(
    new_blockr_ctor("blockr.core::new_dataset_block"),
    "blockr_ctor"
  )

  expect_identical(forward_ctor(1L), -2L)

  # test for namespaced ctor in `resolve_ctor()`
  expect_s3_class(
    blockr.core::new_dataset_block(),
    "block"
  )
})

#' Similar to [`blocks`](new_block()) and [`links`][new_link()], multiple
#' stacks can be handled using `stacks` container objects. These hand out
#' (and guarantee) unique IDs (implemented as [base::names()]) and ensure
#' further consistency criteria, such as disallowing membership in multiple
#' stacks at the same time for a given block.
#'
#' Stack container objects (`stacks` objects) can be created with `stacks()`
#' or `as_stacks()` and inheritance can be tested via `is_stacks()`. Further
#' basic operations such as concatenation, subsetting and sub-assignments is
#' available by means of base R generics.
#'
#' @rdname new_stack
#' @export
stacks <- function(...) {

  args <- list(...)

  args <- lapply(args, as_stack)

  if (is.null(names(args))) {
    names(args) <- rand_names(n = length(args))
  }

  miss <- is.na(names(args)) | names(args) == ""

  if (any(miss)) {
    names(args)[miss] <- rand_names(names(args)[!miss], sum(miss))
  }

  validate_stacks(
    new_vctr(args, class = "stacks")
  )
}

#' @rdname new_stack
#' @export
is_stacks <- function(x) {
  inherits(x, "stacks")
}

validate_stacks <- function(x) {

  if (!is_stacks(x)) {
    blockr_abort(
      "Expecting stacks to inherit from \"stacks\".",
      class = "stacks_class_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a `stacks` object to behave list-like.",
      class = "stacks_type_invalid"
    )
  }

  if (!all(lgl_ply(x, is_stack))) {
    blockr_abort(
      "Expecting a `stacks` object to contain a set of stacks.",
      class = "stacks_contains_invalid"
    )
  }

  ids <- names(x)

  if (length(ids) != length(x) || any(is.na(ids) | !nchar(ids))) {
    blockr_abort(
      "Stack IDs are required to be nonempty strings.",
      class = "stacks_names_invalid"
    )
  }

  if (anyDuplicated(ids) != 0) {
    blockr_abort(
      "Stack IDs are required to be unique.",
      class = "stacks_names_invalid"
    )
  }

  if (!stack_blocks_unique(x)) {
    blockr_abort(
      "Blocks cannot be in mutliple stacks at the same time.",
      class = "stacks_blocks_invalid"
    )
  }

  x
}

stack_blocks_unique <- function(x) {

  seq <- seq_along(x)

  for (i in seq) {
    for (j in seq[seq > i]) {
      if (length(intersect(x[[i]], x[[j]]))) {
        return(FALSE)
      }
    }
  }

  TRUE
}

#' @export
format.stacks <- function(x, ...) {

  out <- lapply(x, format)
  out <- Map(c, names(x), lapply(out, function(x) paste0(" ", x)))
  out <- lapply(out, c, "")
  out <- unlst(out)

  c(
    paste0("<", class(x)[1L], "[", length(x), "]>"),
    if (length(x)) "",
    out[-length(out)]
  )
}

#' @export
print.stacks <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.stacks <- function(x, ...) {
  str_value_collection(x, chr_ply(x, str_value))
}

#' @export
str.stacks <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @rdname new_stack
#' @export
as_stacks <- function(x, ...) {
  UseMethod("as_stacks")
}

#' @export
as_stacks.stacks <- function(x, ...) {
  validate_stacks(x)
}

#' @export
as_stacks.list <- function(x, ...) {
  do.call(stacks, x)
}

#' @export
as_stacks.character <- function(x, ...) {
  as_stacks(list(x))
}

#' @export
as_stacks.stack <- function(x, ...) {
  as_stacks(list(x))
}

#' @export
as_stacks.board <- function(x, ...) {
  board_stacks(x)
}

#' @export
as.list.stacks <- function(x, ...) {
  vec_data(x)
}

#' @export
`names<-.stacks` <- function(x, value) {

  if (is.null(value)) {
    value <- rep("", length(x))
  } else if (anyDuplicated(value) != 0L) {
    blockr_abort(
      "IDs are required to be unique.",
      class = "stacks_names_unique_invalid"
    )
  }

  attr(x, "names") <- value

  x
}

#' @export
vec_restore.stacks <- function(x, to, ...) {
  validate_stacks(NextMethod())
}

harmonize_list_of_stacks <- function(x) {
  if (is_stack(x)) {
    list(x)
  } else if (is_stacks(x)) {
    as.list(x)
  } else if (is.list(x) && all(lgl_ply(x, is_stack))) {
    x
  } else {
    list(as_stack(x))
  }
}

#' @export
c.stacks <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_stacks),
    recursive = FALSE
  )

  as_stacks(res)
}

#' @export
`[<-.stacks` <- function(x, i, ..., value) {

  i <- vec_as_location(i, length(x), names(x))

  if (is.null(value)) {
    return(stacks_slice(x, -i))
  }

  value <- as_stacks(value)

  trg_ids <- names(x)[i]

  if (is.null(names(value))) {
    names(value) <- trg_ids
  }

  new_ids <- names(value)

  if (!setequal(new_ids, trg_ids)) {
    blockr_abort(
      "Replacing IDs {trg_ids} with {new_ids} is not allowed.",
      class = "stacks_assignment_name_invalid"
    )
  }

  stacks_assign(x, i, value[trg_ids])
}

#' @export
`[[<-.stacks` <- function(x, i, ..., value) {

  i <- vec_as_location2(i, length(x), names(x))
  val <- set_names(list(as_stack(value)), names(x)[i])

  stacks_assign(x, i, as_stacks(val))
}

stacks_slice <- function(...) {
  validate_stacks(vec_slice(...))
}

stacks_assign <- function(...) {
  validate_stacks(vec_assign(...))
}

test_that("edit blocks", {

  testServer(
    edit_stack_server,
    {
      expect_null(update())

      session$setInputs(stack_name_in = "data")

      session$flushReact()

      res <- update()$stacks

      expect_length(res$add, 0L)
      expect_length(res$rm, 0L)
      expect_length(res$mod, 1L)

      expect_named(res$mod, "a")
      expect_identical(res$mod[["a"]], list(name = "data"))

      expect_null(session$returned)
    },
    args = list(
      stack_id = "a",
      board = reactiveValues(
        board = new_board(
          blocks = blocks(a = new_dataset_block("iris")),
          stacks = stacks(a = "a")
        )
      ),
      update = reactiveVal()
    )
  )

  testServer(
    edit_stack_server,
    {
      expect_null(update())

      session$setInputs(rm_stack = 1)

      session$flushReact()

      res <- update()$stacks

      expect_length(res$add, 0L)
      expect_length(res$mod, 0L)

      expect_identical(res$rm, "a")

      expect_null(session$returned)
    },
    args = list(
      stack_id = "a",
      board = reactiveValues(
        board = new_board(
          blocks = blocks(a = new_dataset_block("iris")),
          stacks = stacks(a = "a")
        )
      ),
      update = reactiveVal()
    )
  )
})

test_that("dummy edit block ui test", {
  expect_s3_class(
    edit_stack_ui("edit_stack", new_stack()),
    "shiny.tag.list"
  )
})

test_that("stacks class", {

  x <- new_stack(letters[1:5], "stack a")
  y <- new_stack(letters[6:8], "stack b")

  stk <- stacks(x, y)

  expect_s3_class(stk, "stacks")
  expect_true(is_stacks(stk))
  expect_length(stk, 2)
  expect_named(stk)

  expect_error(
    validate_stacks(1),
    class = "stacks_class_invalid"
  )

  expect_error(
    validate_stacks(
      structure(1, class = c("stacks", "vctrs_vctr", "list"))
    ),
    class = "stacks_type_invalid"
  )

  expect_error(
    validate_stacks(
      structure(list(1), class = c("stacks", "vctrs_vctr", "list"))
    ),
    class = "stacks_contains_invalid"
  )

  expect_error(
    validate_stacks(
      structure(list(x), class = c("stacks", "vctrs_vctr", "list"))
    ),
    class = "stacks_names_invalid"
  )

  expect_error(
    validate_stacks(
      structure(list(a = x, a = y), class = c("stacks", "vctrs_vctr", "list"))
    ),
    class = "stacks_names_invalid"
  )

  expect_error(
    validate_stacks(
      structure(list(a = x, b = x), class = c("stacks", "vctrs_vctr", "list"))
    ),
    class = "stacks_blocks_invalid"
  )

  names(stk) <- c("a", "b")

  expect_named(stk, c("a", "b"))

  expect_error(
    names(stk) <- c("a", "a"),
    class = "stacks_names_unique_invalid"
  )

  expect_error(
    names(stk)[1] <- "b",
    class = "stacks_names_unique_invalid"
  )

  expect_snapshot(print(stk))

  expect_identical(as_stacks(stk), stk)
  expect_identical(as_stacks(as.list(stk)), stk)

  expect_s3_class(as_stacks(x), "stacks")
  expect_true(is_stacks(as_stacks(x)))
  expect_length(as_stacks(x), 1)

  stks <- c(a = x, b = list(y), c = letters[9:10])

  expect_s3_class(as_stacks(stks), "stacks")
  expect_true(is_stacks(as_stacks(stks)))
  expect_length(as_stacks(stks), 3)
  expect_named(stks, c("a", "b", "c"))

  stkss <- c(stks, d = letters[11:15])

  expect_s3_class(as_stacks(stkss), "stacks")
  expect_true(is_stacks(as_stacks(stkss)))
  expect_length(as_stacks(stkss), 4)
  expect_named(stkss, c("a", "b", "c", "d"))

  orig <- stacks(a = x, b = y)
  modi <- orig

  modi["b"] <- stacks(b = y)

  expect_identical(orig, modi)

  modi["b"] <- list(b = y)

  expect_identical(orig, modi)

  expect_error(
    modi["b"] <- y,
    class = "stacks_assignment_name_invalid"
  )

  modi[["b"]] <- y

  expect_identical(orig, modi)
})

test_that("a stacks container has a compact str_value()", {

  stk <- stacks(s1 = new_stack(c("a", "b"), name = "my stack"))

  expect_identical(
    str_value(stk),
    "<stacks[1]>\n  s1: <stack> \"my stack\": a, b"
  )

  expect_identical(str_value(stacks()), "<stacks[0]>")

  expect_identical(capture.output(str(stk))[1L], " <stacks[1]>")
})

dt_result <- function(result, block, session) {

  info_string <- function(total) {
    paste0(
      "\"Showing \" + start + \" to \" + end + \" of ",
      if (is.null(total)) "0" else if (is.na(total)) "??" else total,
      " entries\""
    )
  }

  need_pagination <- function(dat_row, show_row, page) {
    if (is.na(dat_row)) rows > page else min(dat_row, rows) > page
  }

  default_opts <- board_options(block)

  rows <- get_board_option_or_default("n_rows", default_opts, session)
  page <- get_board_option_or_default("page_size", default_opts, session)

  dom <- "rti"

  if (not_null(result) && need_pagination(nrow(result), rows, page)) {
    dom <- paste0(dom, "p")
  }

  if (get_board_option_or_default("filter_rows", default_opts, session)) {
    dom <- paste0("f", dom)
  }

  opts <- list(
    processing = FALSE,
    infoCallback = DT::JS(
      "function(settings, start, end, max, total, pre) {",
      "let res = ", info_string(nrow(result)), ";",
      "return res;",
      "}"
    ),
    dom = dom,
    pageLength = page,
    ordering = FALSE
  )

  DT::renderDT(
    {
      dat <- as.data.frame(utils::head(result, rows))
      dt <- DT::datatable(
        dat,
        selection = "none",
        options = opts
      )

      # Skip formatStyle for empty dataframes (0 columns)
      # formatStyle with columns=character(0) breaks DT in DAG boards
      if (ncol(dat) > 0) {
        DT::formatStyle(
          dt,
          columns = names(dat),
          whiteSpace = "pre-wrap"
        )
      } else {
        dt
      }
    },
    server = TRUE
  )
}

dt_render_trigger <- function(x, session = get_session()) {

  invisible(
    get_board_option_values(
      "n_rows",
      "page_size",
      "filter_rows",
      if_not_found = "null",
      session = session
    )
  )
}

#' Stacks
#'
#' Multiple (related) blocks can be grouped together into stacks. Such a
#' grouping has no functional implications, rather it is an organizational tool
#' to help users manage more complex pipelines. Stack objects constitute a set
#' of attributes, the most important of which is `blocks` (a character vector
#' of block IDs). Each `stack` may have an arbitrary `name` and the class can
#' be extended by adding further attributes, maybe something like `color`,
#' coupled with sub-classing.
#'
#' Individual stacks can be created using `new_stack()` or `as_stack()` and
#' inheritance can be tested with `is_stack()`. Attributes can be retrieved
#' (and modified) with `stack_blocks()`/`stack_blocks<-()` and
#' `stack_name()`/`stack_name<-()`, while validation is available as
#' (generic) `validate_stack()`.
#'
#' @param blocks Set of blocks
#' @param name Stack name
#' @param ... Extensibility
#' @param ctor,pkg Constructor information (used for serialization)
#' @param class (Optional) stack sub-class
#'
#' @examples
#' stk <- new_stack(letters[1:5], "Alphabet 1")
#'
#' stack_blocks(stk)
#' stack_name(stk)
#' stack_name(stk) <- "Alphabet start"
#'
#' stks <- c(start = stk, cont = new_stack(letters[6:10], "Alphabet cont."))
#' names(stks)
#'
#' tryCatch(
#'   stack_blocks(stks[[2]]) <- letters[4:8],
#'   error = function(e) conditionMessage(e)
#' )
#'
#' @return Construction and coercion via `new_stack()`/`as_stack()` and
#' `stacks()`/`as_stacks()` results in `stack` and `stacks` objects,
#' respectively, while inheritance testing via `is_stack()` and `is_stacks()`
#' returns scalar logicals. Attribute getters `stack_name()` and
#' `stack_blocks()` return scalar and vector-valued character vectors while
#' setters `stack_name()<-` and `stack_blocks()<-` return modified stack
#' objects.
#'
#' @export
new_stack <- function(blocks = character(), name = default_stack_name, ...,
                      ctor = "new_stack", pkg = pkg_name(),
                      class = character()) {

  if (is_blocks(blocks)) {
    blocks <- names(blocks)
  }

  if (!length(blocks)) {
    blocks <- character()
  }

  if (is.null(name)) {
    name <- default_stack_name
  }

  if (is.function(name)) {
    name <- name()
  }

  stopifnot(is_string(name))

  validate_stack(
    structure(
      blocks,
      name = name,
      ...,
      ctor = new_blockr_ctor(NULL, ctor, pkg),
      class = c(class, "stack")
    )
  )
}

#' @rdname new_stack
#' @export
default_stack_name <- function() {
  "Stack"
}

#' @param x Stack object
#' @rdname new_stack
#' @export
is_stack <- function(x) {
  inherits(x, "stack")
}

#' @rdname new_stack
#' @export
stack_blocks <- function(x) {
  as.character(x)
}

#' @param value Replacement value
#' @rdname new_stack
#' @export
`stack_blocks<-` <- function(x, value) {
  vec_restore(new_stack(value), x)
}

#' @rdname new_stack
#' @export
stack_name <- function(x, name) {
  attr(x, "name")
}

#' @rdname new_stack
#' @export
`stack_name<-` <- function(x, value) {
  attr(x, "name") <- value
  x
}

#' @section Partial-arg updates:
#' `update_stack()` is an S3 generic that produces a modified stack by
#' merging a named-list `delta` of constructor argument values onto an
#' existing stack. The default `.stack` method reconstructs the object
#' via its stored constructor, treating the underlying character vector
#' (block IDs) as `blocks` and every other attribute as a named
#' constructor argument. Sub-class owners (e.g. `dock_stack` adding
#' `color`) only need to register a method when their constructor
#' deviates from this convention. The reserved delta key `blocks`
#' replaces the member block IDs; every other key updates the named
#' attribute.
#'
#' @param delta A named list of constructor argument values to apply on
#' top of `x`'s current fields.
#'
#' @rdname new_stack
#' @export
update_stack <- function(x, delta) {
  UseMethod("update_stack")
}

#' @export
update_stack.stack <- function(x, delta) {

  ctor <- attr(x, "ctor")

  reserved <- c("class", "ctor")
  attrs <- attributes(x)
  attrs <- attrs[setdiff(names(attrs), reserved)]

  args <- c(list(blocks = stack_blocks(x)), attrs)
  args[names(delta)] <- delta

  do.call(
    ctor_fun(ctor),
    c(
      args,
      list(ctor = coal(ctor_name(ctor), ctor_fun(ctor)), pkg = ctor_pkg(ctor))
    )
  )
}

#' @rdname new_stack
#' @export
validate_stack <- function(x) {
  UseMethod("validate_stack")
}

#' @export
validate_stack.stack <- function(x) {

  if (!is_stack(x)) {
    blockr_abort(
      "Expecting a stack to inherit from `stack`.",
      class = "stack_class_invalid"
    )
  }

  if (!is_string(attr(x, "name"))) {
    blockr_abort(
      "Expecting the stack name to be a string.",
      class = "stack_name_invalid"
    )
  }

  if (!is.character(x)) {
    blockr_abort(
      "Expecting stack blocks to be character-like.",
      class = "stack_type_invalid"
    )
  }

  if (anyNA(x) || !all(nzchar(x))) {
    blockr_abort(
      "Expecting the stack blocks to be strings.",
      class = "stack_blocks_invalid"
    )
  }

  nme <- names(x)

  if (!is.null(nme) || any(nzchar(nme))) {
    blockr_warn(
      "Names are ignored in stack objects.",
      class = "named_stack_obejct_warn"
    )
  }

  if (anyDuplicated(x) != 0L) {
    blockr_abort(
      "Stack blocks have to be unique.",
      class = "stack_block_duplicates"
    )
  }

  x
}

#' @export
validate_stack.default <- function(x) {
  blockr_abort(
    "Expecting a stack to inherit from `stack`.",
    class = "stack_class_invalid"
  )
}

#' @export
format.stack <- function(x, ...) {

  out <- paste0("[", length(x), "]")

  for (cl in rev(setdiff(class(x), c("list", "vctrs_vctr")))) {
    out <- paste0("<", cl, out, ">")
  }

  if (length(x)) {
    blks <- paste0("Blocks: ", paste_enum(as.character(x), quotes = "\""))
  } else {
    blks <- "No blocks"
  }

  c(
    out,
    paste0("Name: \"", attr(x, "name"), "\""),
    strwrap(blks, exdent = 2)
  )
}

#' @export
print.stack <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.stack <- function(x, ...) {

  out <- paste0("<", class(x)[1L], "> \"", stack_name(x), "\"")

  blks <- stack_blocks(x)

  if (length(blks)) {
    out <- paste0(out, ": ", paste0(blks, collapse = ", "))
  }

  out
}

#' @export
str.stack <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
duplicated.stack <- function(x, incomparables = FALSE, ...) {
  duplicated(as.character(x), incomparables = incomparables, ...)
}

#' @export
anyDuplicated.stack <- function(x, incomparables = FALSE, ...) {
  anyDuplicated(as.character(x), incomparables = incomparables, ...)
}

#' @export
as.character.stack <- function(x, ...) c(unclass(x))

#' @export
as.list.stack <- function(x, ...) {
  list(
    blocks = as.character(x),
    name = stack_name(x)
  )
}

#' @rdname new_stack
#' @export
as_stack <- function(x) UseMethod("as_stack")

#' @export
as_stack.stack <- function(x) validate_stack(x)

#' @export
as_stack.character <- function(x) new_stack(x)

#' @export
as_stack.list <- function(x) {
  do.call(new_stack, x[c("blocks", "name")])
}

#' @export
vec_restore.stack <- function(x, to, ...) {
  validate_stack(NextMethod())
}

#' @importFrom generics intersect
#' @export
intersect.stack <- function(x, y, ...) {
  vec_restore(intersect(as.character(x), as.character(y), ...), x)
}

#' @importFrom generics union
#' @export
union.stack <- function(x, y, ...) {
  vec_restore(union(as.character(x), as.character(y), ...), x)
}

#' @importFrom generics setdiff
#' @export
setdiff.stack <- function(x, y, ...) {
  vec_restore(setdiff(as.character(x), as.character(y), ...), x)
}

#' @importFrom generics setequal
#' @export
setequal.stack <- function(x, y, ...) {
  setequal(as.character(x), as.character(y), ...)
}

#' @importFrom generics is.element
#' @export
is.element.stack <- function(el, set, ...) {
  is.element(as.character(el), as.character(set), ...)
}

#' @export
`[.stack` <- function(x, i, ...) {
  vec_slice(as.character(x), vec_as_location(i, length(x)))
}

#' @export
`[[.stack` <- function(x, i, ...) {
  vec_slice(as.character(x), vec_as_location2(i, length(x)))
}

#' @export
`[<-.stack` <- function(x, i, ..., value) {
  blockr_abort(
    "Subassignment of stack objects is not suported. Use set operations ",
    "instead.",
    class = "stack_subassignment_invalid"
  )
}

#' @export
`[[<-.stack` <- function(x, i, ..., value) {
  blockr_abort(
    "Subassignment of stack objects is not suported. Use set operations",
    "instead.",
    class = "stack_subassignment_invalid"
  )
}

#' @export
c.stack <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_stacks),
    recursive = FALSE
  )

  as_stacks(res)
}

test_that("add/rm stacks", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    )
  )

  testServer(
    manage_stacks_server,
    {
      expect_null(session$returned)
      expect_null(update())

      expect_identical(upd$add, stacks())
      expect_identical(upd$rm, character())
      expect_null(upd$edit)

      expect_null(update())

      session$setInputs(add_stack = 1)

      session$flushReact()

      expect_s3_class(upd$add, "stacks")
      expect_length(upd$add, 1L)

      stk <- names(upd$add)
      new <- as_stacks(
        set_names(list(new_stack(name = stack_name(upd$add[[1L]]))), stk)
      )

      expect_identical(upd$add, new)

      upd$edit <- list(row = names(upd$add), col = "blocks", val = "a")
      stack_blocks(new[[1]]) <- "a"

      session$flushReact()
      expect_identical(upd$add, new)

      upd$edit <- list(row = names(upd$add), col = "name", val = "my stack")
      stack_name(new[[1]]) <- "my stack"

      session$flushReact()
      expect_identical(upd$add, new)

      session$setInputs(modify_stacks = 1)

      expect_identical(
        update()$stacks,
        list(add = new, rm = NULL, mod = NULL)
      )

      expect_identical(upd$add, stacks())
      expect_identical(upd$rm, character())
      expect_identical(upd$mod, stacks())

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )

  testServer(
    manage_stacks_server,
    {
      expect_null(upd$edit)
      expect_length(upd$add, 0L)
      expect_length(upd$mod, 0L)

      session$flushReact()

      expect_length(upd$obs, 1L)
      expect_named(upd$obs, "ac")
      expect_type(upd$obs, "list")

      expect_length(upd$obs[["ac"]], 2L)
      expect_named(upd$obs[["ac"]], c("name", "blocks"))
      expect_type(upd$obs[["ac"]], "list")

      for (i in c("name", "blocks")) {
        expect_s3_class(upd$obs[["ac"]][[i]], "Observer")
      }

      session$setInputs(ac_name = "some stack")
      expect_identical(
        upd$edit,
        list(row = "ac", col = "name", val = "some stack")
      )
      expect_length(upd$add, 0L)
      expect_length(upd$mod, 1L)

      session$setInputs(ac_blocks = c("a", "b"))
      expect_identical(
        upd$edit,
        list(row = "ac", col = "blocks", val = c("a", "b"))
      )
      expect_length(upd$add, 0L)
      expect_length(upd$mod, 1L)
    },
    args = list(
      board = list(
        board = new_board(
          blocks = c(
            a = new_dataset_block("iris"),
            b = new_dataset_block("mtcars"),
            c = new_subset_block(),
            d = new_subset_block()
          ),
          stacks = stacks(ac = c("a", "c"))
        )
      ),
      update = reactiveVal()
    )
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    stacks = stacks(ab = c("a", "b"))
  )

  testServer(
    manage_stacks_server,
    {
      expect_null(session$returned)
      expect_null(update())

      expect_s3_class(upd$curr, "stacks")
      expect_length(upd$curr, 1L)
      expect_named(upd$curr, "ab")

      expect_identical(upd$rm, character())

      session$setInputs(stacks_dt_rows_selected = 1, rm_stack = 1)

      expect_identical(upd$rm, "ab")

      session$setInputs(modify_stacks = 1)

      expect_identical(
        update()$stacks,
        list(add = NULL, rm = "ab", mod = NULL)
      )

      expect_identical(upd$add, stacks())
      expect_identical(upd$rm, character())
      expect_identical(upd$mod, stacks())
      expect_identical(upd$curr, stacks())

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("add/rm stacks return validation", {

  brd <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    stacks = stacks(ab = c("a", "b"))
  )

  pay <- list(stacks = list(add = stacks(a = "a"), rm = "ab"))

  expect_identical(
    validate_board_update(pay, brd),
    pay
  )

  expect_error(
    validate_board_update(list(stacks = "a"), new_board()),
    class = "board_update_component_type_invalid"
  )

  expect_error(
    validate_board_update(list(stacks = list(abc = NULL)), new_board()),
    class = "board_update_component_components_invalid"
  )

  expect_error(
    validate_board_update(list(stacks = list(add = "a")), new_board()),
    class = "board_update_add_component_invalid"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(add = stacks(a = new_stack()))),
      new_board(
        blocks = c(a = new_dataset_block()),
        stacks = stacks(a = "a")
      )
    ),
    class = "board_update_stacks_add_invalid"
  )

  expect_error(
    validate_board_update(list(stacks = list(rm = 1)), new_board()),
    class = "board_update_rm_component_invalid"
  )

  expect_error(
    validate_board_update(list(stacks = list(rm = "a")), new_board()),
    class = "board_update_stacks_rm_invalid"
  )
})

test_that("merge_staged_stacks overlays staged edits on refreshed stacks", {

  applied <- stacks(
    k1 = new_stack(name = "One"),
    k2 = new_stack(name = "Two")
  )

  expect_identical(
    merge_staged_stacks(applied, stacks(), character(), stacks()),
    applied
  )

  added <- stacks(z = new_stack(name = "Zed"))
  expect_identical(
    merge_staged_stacks(applied, added, character(), stacks()),
    c(applied, added)
  )

  modded <- stacks(k1 = new_stack(name = "Renamed"))
  in_place <- applied
  in_place[["k1"]] <- modded[["k1"]]
  expect_identical(
    merge_staged_stacks(applied, stacks(), character(), modded),
    in_place
  )

  expect_named(
    merge_staged_stacks(applied, added, "k2", stacks()),
    c("k1", "z")
  )
})

test_that("manage stacks keeps staged edits when the board re-emits", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    )
  )

  testServer(
    manage_stacks_server,
    {
      session$flushReact()

      session$setInputs(add_stack = 1)
      session$flushReact()

      row <- names(upd$add)
      expect_length(upd$curr, 1L)

      upd$edit <- list(row = row, col = "blocks", val = "a")
      session$flushReact()

      expect_named(upd$add, row)
      expect_true(row %in% names(upd$curr))

      board$board <- new_board(
        blocks = c(
          a = new_dataset_block("iris"),
          b = new_subset_block()
        )
      )
      session$flushReact()

      expect_true(row %in% names(upd$curr))
      expect_named(upd$add, row)
      expect_identical(upd$curr[[row]], upd$add[[row]])

      session$setInputs(modify_stacks = 1)

      res <- update()$stacks

      expect_named(res$add, row)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("manage stacks redraws only when the id set changes", {

  redraws <- 0L

  local_mocked_bindings(
    replaceData = function(...) redraws <<- redraws + 1L,
    .package = "DT"
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    stacks = stacks(ab = c("a", "b"))
  )

  testServer(
    manage_stacks_server,
    {
      session$flushReact()

      expect_identical(names(upd$curr), "ab")
      expect_setequal(names(upd$obs), "ab")

      base <- redraws

      upd$curr <- stacks(ab = new_stack(blocks = "a", name = "ab"))
      session$flushReact()

      expect_identical(names(upd$curr), "ab")
      expect_identical(redraws, base)
      expect_setequal(names(upd$obs), "ab")

      upd$curr <- stacks(
        ab = new_stack(blocks = "a", name = "ab"),
        cd = new_stack(blocks = "b", name = "cd")
      )
      session$flushReact()

      expect_gt(redraws, base)
      expect_setequal(names(upd$obs), c("ab", "cd"))
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("manage stacks ignores no-op board re-emits", {

  stack_syncs <- 0L

  real_merge_stacks <- merge_staged_stacks

  local_mocked_bindings(
    merge_staged_stacks = function(...) {
      stack_syncs <<- stack_syncs + 1L
      real_merge_stacks(...)
    }
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    stacks = stacks(s1 = new_stack(blocks = "a", name = "One"))
  )

  testServer(
    manage_stacks_server,
    {
      session$flushReact()

      expect_gt(stack_syncs, 0L)

      stack_base <- stack_syncs

      board$board <- new_board(
        blocks = c(
          a = new_dataset_block("iris"),
          b = new_subset_block()
        ),
        stacks = stacks(s1 = new_stack(blocks = "a", name = "One"))
      )
      session$flushReact()

      expect_identical(stack_syncs, stack_base)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("dummy ad/rm stack ui test", {
  expect_s3_class(manage_stacks_ui("stack", new_board()), "shiny.tag.list")
  expect_s3_class(stacks_modal(NS("stacks")), "shiny.tag")
})

test_that("stack class", {

  x <- new_stack(letters[1:5])

  expect_s3_class(x, "stack")
  expect_true(is_stack(x))
  expect_length(x, 5)

  expect_setequal(as_stack(x), x)
  expect_setequal(as_stack(letters[1:5]), x)
  expect_setequal(as_stack(list(blocks = letters[1:5])), x)

  expect_setequal(as_stack(as.character(x)), x)
  expect_setequal(as_stack(as.list(x)), x)

  expect_identical(duplicated(x), rep(FALSE, length(x)))
  expect_identical(anyDuplicated(x), 0L)

  expect_error(
    validate_stack(1),
    class = "stack_class_invalid"
  )

  expect_error(
    validate_stack(structure(1, name = 2, class = "stack")),
    class = "stack_name_invalid"
  )

  expect_error(
    validate_stack(structure(1, name = "a", class = "stack")),
    class = "stack_type_invalid"
  )

  expect_error(
    validate_stack(structure("", name = "a", class = "stack")),
    class = "stack_blocks_invalid"
  )

  expect_error(
    validate_stack(structure(NA_character_, name = "a", class = "stack")),
    class = "stack_blocks_invalid"
  )

  expect_error(
    validate_stack(structure(c("a", "a"), name = "a", class = "stack")),
    class = "stack_block_duplicates"
  )

  stack_name(x) <- "my stack"

  expect_snapshot(print(x))

  y <- new_stack("a")

  expect_length(setdiff(x, y), 4L)
  expect_s3_class(setdiff(x, y), "stack")
  expect_identical(setdiff(x, y), setdiff(x, "a"))

  expect_length(setdiff(letters[1:5], y), 4L)
  expect_length(union(x, y), length(x))
  expect_length(intersect(x, y), length(y))

  expect_true(setequal(x, x))
  expect_false(setequal(y, x))

  expect_true(is.element(y, x))
  expect_false(is.element("y", x))

  expect_type(x[1], "character")
  expect_type(x[[1]], "character")

  expect_error(x["a"])
  expect_error(x[[1:2]])

  expect_error(
    x[1] <- "c",
    class = "stack_subassignment_invalid"
  )

  expect_error(
    x[[1]] <- "c",
    class = "stack_subassignment_invalid"
  )

  stks <- c(x, letters[6:8])

  expect_s3_class(stks, "stacks")
  expect_true(is_stacks(stks))
  expect_length(stks, 2)
})

test_that("stacks have a compact str_value()", {

  stk <- new_stack(c("a", "b"), name = "my stack")

  expect_identical(str_value(stk), "<stack> \"my stack\": a, b")
  expect_identical(str_value(new_stack(name = "empty")), "<stack> \"empty\"")

  out <- capture.output(res <- withVisible(str(stk)))

  expect_identical(out, " <stack> \"my stack\": a, b")
  expect_false(res$visible)
  expect_identical(res$value, stk)
})

test_that("str_value() is extensible via NextMethod()", {

  registerS3method(
    "str_value", "test_substack",
    function(x, ...) paste0(NextMethod(), " [extra]"),
    envir = asNamespace("blockr.core")
  )

  sub <- new_stack(c("a", "b"), name = "S")
  class(sub) <- c("test_substack", class(sub))

  expect_identical(str_value(sub), "<test_substack> \"S\": a, b [extra]")

  expect_identical(
    capture.output(str(sub)),
    " <test_substack> \"S\": a, b [extra]"
  )
})

#' Plugin module for managing board links
#'
#' Logic and user experience for adding new, removing and modifying existing
#' links to/from the board can be customized or enhanced by providing an
#' alternate version of this plugin. The default implementation provides a
#' table-based UI, presented in a modal.
#'
#' Updates are mediated via the [shiny::reactiveVal()] object passed as
#' `update`, where link updates are communicated as list entry `stacks` with
#' components `add`, `rm` or `mod`, where
#' * `add` is either `NULL` or a `links` object (link IDs may not already
#'   exists),
#' * `rm` is either `NULL` or a character vector of (existing) link IDs,
#' * `mod` is either `NULL` or a `links` object (where link IDs must already
#'   exist).
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `manage_links` is returned by
#' `manage_links()`, while the UI component (e.g. `manage_links_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `manage_links_server()`) is expected to return `NULL`.
#'
#' @export
manage_links <- function(server = manage_links_server, ui = manage_links_ui) {
  new_plugin(server, ui, class = "manage_links")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param update Reactive value object to initiate board updates
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname manage_links
#' @export
manage_links_server <- function(id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      observeEvent(
        input$links_mod,
        showModal(links_modal(session$ns))
      )

      upd <- reactiveValues(
        add = links(),
        rm = character(),
        curr = isolate(board_links(board$board)),
        obs = list(),
        edit = NULL
      )

      applied_links <- deduped_board_reactive(board, board_links)

      observeEvent(
        applied_links(),
        {
          upd$curr <- merge_staged_links(applied_links(), upd$add, upd$rm)
        }
      )

      output$links_dt <- DT::renderDataTable(
        link_dt(isolate(upd$curr), session$ns, isolate(board$board)),
        server = TRUE
      )

      links_proxy <- DT::dataTableProxy("links_dt", session)

      create_link_obs_observer(input, board, upd, session, links_proxy)

      edit_link_observer(upd, board)

      add_link_observer(input, board, upd, session)

      rm_link_observer(input, board, upd, session)

      cancel_link_observer(input, board, upd, session)

      modify_link_observer(input, board, upd, session, links_proxy, update)

      NULL
    }
  )
}

#' @param board The initial `board` object
#' @rdname manage_links
#' @export
manage_links_ui <- function(id, board) {
  tagList(
    actionButton(
      NS(id, "links_mod"),
      "Edit links",
      icon = icon("link")
    )
  )
}

link_dt <- function(dat, ns, board) {
  res <- DT::datatable(
    dt_board_link(dat, ns, board),
    options = list(
      pageLength = 5,
      preDrawCallback = DT::JS(
        "function() { Shiny.unbindAll(this.api().table().node()); }"
      ),
      drawCallback = DT::JS(
        "function() { Shiny.bindAll(this.api().table().node()); }"
      ),
      dom = "tp",
      ordering = FALSE,
      columnDefs = list(
        list(targets = 0, width = "125px")
      )
    ),
    rownames = FALSE,
    escape = FALSE
  )

  DT::formatStyle(res, 1L, `vertical-align` = "middle")
}

dt_board_link <- function(lnk, ns, board) {

  blks <- board_blocks(board)
  arity <- int_ply(blks, block_arity, use_names = TRUE)

  from_ids <- rep(list(names(blks)), length(lnk))

  to_avail <- arity
  cnt_to <- c(table(filter_empty(lnk$to)))

  to_avail[names(cnt_to)] <- int_mply(`-`, to_avail[names(cnt_to)], cnt_to)
  to_avail[is.na(to_avail)] <- 1L

  to_avail <- lapply(
    lapply(lnk$to, filter_empty),
    union,
    names(to_avail)[to_avail > 0L]
  )

  cur_inp <- lapply(split(lnk$input, lnk$to), filter_empty)

  in_avail <- c(
    lapply(blks[!is.na(arity)], block_inputs),
    cur_inp[intersect(names(cur_inp), names(arity)[is.na(arity)])]
  )

  rm_inp <- lapply(
    seq_along(lnk),
    function(i) split(lnk$input[-i], lnk$to[-i])[[lnk$to[i]]]
  )

  data.frame(
    ID = lnk$id,
    From = chr_mply(
      dt_selectize,
      lapply(paste0(lnk$id, "_from"), ns),
      lnk$from,
      Map(setdiff, from_ids, lnk$to)
    ),
    To = chr_mply(
      dt_selectize,
      lapply(paste0(lnk$id, "_to"), ns),
      lnk$to,
      Map(setdiff, to_avail, lnk$from)
    ),
    Input = chr_mply(
      dt_selectize,
      lapply(paste0(lnk$id, "_input"), ns),
      lnk$input,
      Map(setdiff, in_avail[lnk$to], rm_inp),
      is.na(arity[lnk$to])
    )
  )
}

dt_selectize <- function(id, val, choices, create = FALSE) {

  if (isTRUE(create)) {
    opts <- list(create = TRUE)
  } else {
    opts <- NULL
  }

  res <- selectizeInput(id, label = NULL, choices = c("", choices),
                        selected = val, options = opts)

  res <- htmltools::tagQuery(
    res
  )$addAttrs(
    style = "width: 175px; margin-bottom: 0;"
  )$allTags()

  as.character(res)
}

create_dt_link_obs <- function(ids, upd, ...) {

  create_obs <- function(col, row, upd, input, blks, sess) {

    inp <- paste0(row, "_", col)

    log_debug("creating link DT observer ", inp)

    observeEvent(
      input[[inp]],
      {
        new <- input[[inp]]
        cur <- upd$curr[[row]][[col]]

        if (new == cur || (new == "" && col != "input")) {
          return()
        }

        if (col == "from") {

          to_avail <- int_ply(blks, block_arity, use_names = TRUE)

          cnt <- c(table(filter_empty(upd$curr$to)))

          to_avail[names(cnt)] <- int_mply(`-`, to_avail[names(cnt)], cnt)
          to_avail[is.na(to_avail)] <- 1L

          to_avail <- c(
            upd$curr[[row]][["to"]],
            names(to_avail)[to_avail > 0L]
          )

          updateSelectizeInput(
            sess,
            inputId = paste0(row, "_to"),
            choices = c("", setdiff(to_avail, new)),
            selected = upd$curr[[row]][["to"]]
          )

        } else if (col == "to") {

          ids <- names(blks)

          updateSelectizeInput(
            sess,
            inputId = paste0(row, "_from"),
            choices = c("", setdiff(ids, new)),
            selected = upd$curr[[row]][["from"]]
          )

          if (identical(new, "")) {

            updateSelectizeInput(
              sess,
              inputId = paste0(row, "_input"),
              choices = list()
            )

          } else {

            blk <- blks[[which(ids == new)]]
            ary <- block_arity(blk)
            hit <- upd$curr$to == new

            if (is.na(ary)) {
              inp <- upd$curr[[row]][["input"]]
              opt <- list(create = TRUE)
            } else {
              inp <- setdiff(block_inputs(blk), upd$curr$input[hit])
              opt <-  NULL
            }

            updateSelectizeInput(
              sess,
              inputId = paste0(row, "_input"),
              choices = c("", inp),
              selected = upd$curr[[row]][["input"]],
              options = opt
            )
          }

        } else if (col != "input") {

          blockr_abort(
            "Unexpected input: column {col}.",
            class = "unexpected_link_col_input"
          )
        }

        upd$edit <- list(row = row, col = col, val = new)
      },
      ignoreInit = TRUE
    )
  }

  create_obs_for_id <- function(id, ...) {
    lapply(
      set_names(nm = c("from", "to", "input")),
      create_obs,
      id,
      ...
    )
  }

  upd$obs[ids] <- lapply(ids, create_obs_for_id, upd, ...)

  upd
}

destroy_dt_link_obs <- function(ids, update) {

  for (row in ids) {
    for (col in c("from", "to", "input")) {
      log_debug("destroying link DT observer ", row, " ", col)
      update$obs[[row]][[col]]$destroy()
    }
    update$obs[[row]] <- NULL
  }

  update
}

links_modal <- function(ns) {
  modalDialog(
    title = "Board links",
    DT::dataTableOutput(ns("links_dt")),
    footer = tagList(
      htmltools::tagQuery(
        textInput(ns("new_link_id"), NULL, placeholder = "Next ID")
      )$addAttrs(
        style = "width: 180px; margin: 0 8px;"
      )$allTags(),
      tags$style(
        type = "text/css",
        paste0("#", ns("new_link_id"), " {padding: 0.75em 2em; margin: 4px;}")
      ),
      tags$style(
        type = "text/css",
        paste0("#", ns("new_link_id"), "-label {text-align: left;}")
      ),
      actionButton(ns("add_link"), "Add", icon = icon("plus")),
      actionButton(ns("rm_link"), "Remove", icon = icon("minus")),
      actionButton(ns("cancel_links"), "Cancel", class = "btn-danger"),
      actionButton(ns("modify_links"), "OK", class = "btn-success")
    ),
    size = "l"
  )
}

create_link_obs_observer <- function(input, rv, upd, session, proxy) {

  observeEvent(
    names(upd$curr),
    {
      ids <- names(upd$curr)

      if (setequal(ids, names(upd$obs))) {
        return()
      }

      DT::replaceData(
        proxy,
        dt_board_link(upd$curr, session$ns, rv$board),
        rownames = FALSE
      )

      create_dt_link_obs(setdiff(ids, names(upd$obs)), upd, input,
                         board_blocks(rv$board), session)
      destroy_dt_link_obs(setdiff(names(upd$obs), ids), upd)
    }
  )
}

edit_link_observer <- function(upd, rv) {

  observeEvent(
    upd$edit,
    {
      row <- upd$edit$row
      col <- upd$edit$col

      if (!row %in% upd$rm && row %in% board_link_ids(rv$board)) {
        upd$rm <- c(upd$rm, row)
      }

      new <- do.call(
        `$<-`,
        list(upd$curr[row], col, coal(upd$edit$val, ""))
      )

      upd$curr[row] <- new

      if (row %in% names(upd$add)) {
        upd$add[row] <- new
      } else {
        upd$add <- c(upd$add, new)
      }
    }
  )
}

merge_staged_links <- function(applied, add, rm) {

  keep <- setdiff(names(applied), setdiff(rm, names(add)))
  out <- applied[keep]

  edited <- intersect(keep, names(add))

  if (length(edited)) {
    out[edited] <- add[edited]
  }

  c(out, add[setdiff(names(add), names(applied))])
}

add_link_observer <- function(input, rv, upd, sess) {

  observeEvent(
    input$add_link,
    {
      total <- sum(block_arity(rv$board))

      if (is.na(total) || length(upd$curr) < total) {

        new <- new_link(from = "", to = "", input = "")

        if (length(input$new_link_id) && nchar(input$new_link_id)) {

          updateTextInput(
            session = sess,
            inputId = "new_link_id",
            label = NULL,
            value = "",
            placeholder = "Next ID"
          )

          if (input$new_link_id %in% names(upd$curr)) {
            notify(
              "Please choose a unique link ID.",
              type = "warning",
              session = sess
            )
            return()
          }

          new <- set_names(list(new), input$new_link_id)
        }

        upd$curr <- c(upd$curr, new)
        upd$add <- c(upd$add, upd$curr[length(upd$curr)])

      } else {

        notify(
          "No new links can be added. Remove a row first.",
          type = "warning",
          session = sess
        )
      }
    }
  )
}

rm_link_observer <- function(input, rv, upd, sess) {

  observeEvent(
    input$rm_link,
    {
      sel <- input$links_dt_rows_selected

      if (length(sel)) {

        ids <- names(upd$curr[sel])

        upd$rm <- c(upd$rm, ids[ids %in% board_link_ids(rv$board)])

        upd$add <- upd$add[setdiff(names(upd$add), ids)]
        upd$curr <- upd$curr[setdiff(names(upd$curr), ids)]

      } else {

        notify("No row selected", type = "warning", session = sess)
      }
    }
  )
}

cancel_link_observer <- function(input, rv, upd, session) {

  observeEvent(
    input$cancel_links,
    {
      removeModal(session)

      upd$add <- links()
      upd$rm <- character()
      upd$curr <- board_links(rv$board)
    }
  )
}

modify_link_observer <- function(input, rv, upd, session, proxy, res) {

  observeEvent(
    input$modify_links,
    {
      if (!length(upd$add) && !length(upd$rm)) {

        notify(
          "No changes specified.",
          type = "warning",
          session = session
        )

        return()
      }

      new <- try_catch_continue(
        modify_board_links(rv$board, upd$add, upd$rm),
        message = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "message",
                 glue = FALSE, session = session)
        },
        warning = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "warning",
                 glue = FALSE, session = session)
        },
        error = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "error",
                 glue = FALSE, session = session)
          NULL
        }
      )

      if (is.null(new)) {
        return()
      }

      res(
        list(
          links = list(
            add = if (length(upd$add)) upd$add,
            rm = if (length(upd$rm)) upd$rm
          )
        )
      )

      upd$add <- links()
      upd$rm <- character()
      upd$curr <- board_links(new)

      DT::replaceData(
        proxy,
        dt_board_link(upd$curr, session$ns, rv$board),
        rownames = FALSE
      )

      removeModal(session)
    }
  )
}

test_that("add/rm links", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    )
  )

  testServer(
    manage_links_server,
    {
      expect_null(session$returned)
      expect_null(update())

      expect_identical(upd$add, links())
      expect_identical(upd$rm, character())
      expect_null(upd$edit)

      expect_null(update())

      session$setInputs(add_link = 1)

      expect_s3_class(upd$add, "links")
      expect_length(upd$add, 1L)

      lnk <- names(upd$add)
      expect_identical(upd$add, as_links(set_names(list(new_link()), lnk)))

      session$setInputs(add_link = 2)

      expect_identical(upd$add, as_links(set_names(list(new_link()), lnk)))

      upd$edit <- list(row = names(upd$add), col = "from", val = "a")
      session$flushReact()
      expect_identical(upd$add, as_links(set_names(list(new_link("a")), lnk)))

      upd$edit <- list(row = names(upd$add), col = "to", val = "b")
      session$flushReact()
      expect_identical(
        upd$add,
        as_links(set_names(list(new_link("a", "b")), lnk))
      )

      upd$edit <- list(row = names(upd$add), col = "input", val = "data")
      session$flushReact()
      expect_identical(
        upd$add,
        as_links(set_names(list(new_link("a", "b", "data")), lnk))
      )

      session$setInputs(modify_links = 1)

      expect_identical(
        update()$links,
        list(
          add = as_links(set_names(list(new_link("a", "b", "data")), lnk)),
          rm = NULL
        )
      )

      expect_identical(upd$add, links())
      expect_identical(upd$rm, character())

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )

  testServer(
    manage_links_server,
    {
      expect_null(upd$edit)
      expect_length(upd$add, 0L)

      session$flushReact()

      expect_length(upd$obs, 1L)
      expect_named(upd$obs, "ac")
      expect_type(upd$obs, "list")

      expect_length(upd$obs[["ac"]], 3L)
      expect_named(upd$obs[["ac"]], c("from", "to", "input"))
      expect_type(upd$obs[["ac"]], "list")

      for (i in c("from", "to", "input")) {
        expect_s3_class(upd$obs[["ac"]][[i]], "Observer")
      }

      session$setInputs(ac_input = "data")
      expect_null(upd$edit)
      expect_length(upd$add, 0L)

      session$setInputs(ac_to = "d")
      expect_identical(upd$edit, list(row = "ac", col = "to", val = "d"))
      expect_length(upd$add, 1L)

      session$setInputs(ac_from = "b")
      expect_identical(upd$edit, list(row = "ac", col = "from", val = "b"))
      expect_length(upd$add, 1L)
    },
    args = list(
      board = list(
        board = new_board(
          blocks = c(
            a = new_dataset_block("iris"),
            b = new_dataset_block("mtcars"),
            c = new_subset_block(),
            d = new_subset_block()
          ),
          links = links(ac = new_link(from = "a", to = "c"))
        )
      ),
      update = reactiveVal()
    )
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(aa = new_link(from = "a", to = "b"))
  )

  testServer(
    manage_links_server,
    {
      expect_null(session$returned)
      expect_null(update())

      expect_s3_class(upd$curr, "links")
      expect_length(upd$curr, 1L)
      expect_named(upd$curr, "aa")

      expect_identical(upd$rm, character())

      session$setInputs(links_dt_rows_selected = 1, rm_link = 1)

      expect_identical(upd$rm, "aa")

      session$setInputs(modify_links = 1)

      expect_identical(
        update()$links,
        list(add = NULL, rm = "aa")
      )

      expect_identical(upd$add, links())
      expect_identical(upd$rm, character())
      expect_identical(upd$curr, links())

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("add/rm links return validation", {

  brd <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(ab = new_link(from = "a", to = "b"))
  )

  pay <- list(
    links = list(
      add = links(new_link(from = "a", to = "b", input = "data")),
      rm = "ab"
    )
  )

  expect_identical(
    validate_board_update(pay, brd),
    pay
  )

  expect_error(
    validate_board_update(list(links = "a"), new_board()),
    class = "board_update_component_type_invalid"
  )

  expect_error(
    validate_board_update(list(links = list(abc = NULL)), new_board()),
    class = "board_update_component_components_invalid"
  )

  expect_error(
    validate_board_update(list(links = list(add = "a")), new_board()),
    class = "board_update_add_component_invalid"
  )

  expect_error(
    validate_board_update(
      list(links = list(add = links(a = new_link()))),
      new_board(
        blocks = c(a = new_dataset_block(), b = new_subset_block()),
        links = links(a = new_link("a", "b"))
      )
    ),
    class = "board_update_links_add_invalid"
  )

  expect_error(
    validate_board_update(list(links = list(rm = 1)), new_board()),
    class = "board_update_rm_component_invalid"
  )

  expect_error(
    validate_board_update(list(links = list(rm = "a")), new_board()),
    class = "board_update_links_rm_invalid"
  )
})

test_that("merge_staged_links overlays staged edits on refreshed links", {

  applied <- links(
    l1 = new_link(from = "a", to = "c", input = "1"),
    l2 = new_link(from = "b", to = "c", input = "2")
  )

  expect_identical(
    merge_staged_links(applied, links(), character()),
    applied
  )

  added <- links(x = new_link(from = "", to = "", input = ""))
  expect_identical(
    merge_staged_links(applied, added, character()),
    c(applied, added)
  )

  edited <- links(l1 = new_link(from = "a", to = "c", input = "9"))
  in_place <- applied
  in_place["l1"] <- edited
  expect_identical(merge_staged_links(applied, edited, "l1"), in_place)

  expect_named(merge_staged_links(applied, links(), "l2"), "l1")

  expect_named(
    merge_staged_links(applied["l1"], added, character()),
    c("l1", "x")
  )
})

test_that("manage links keeps staged edits when the board re-emits", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block(),
      c = new_subset_block()
    )
  )

  testServer(
    manage_links_server,
    {
      session$flushReact()

      session$setInputs(add_link = 1)
      session$flushReact()

      row <- names(upd$add)
      expect_length(upd$curr, 1L)
      expect_identical(names(upd$curr), row)

      board$board <- new_board(
        blocks = c(
          a = new_dataset_block("iris"),
          b = new_subset_block(),
          c = new_subset_block()
        )
      )
      session$flushReact()

      expect_length(upd$curr, 1L)
      expect_identical(names(upd$curr), row)
      expect_identical(upd$add, upd$curr[row])

      upd$edit <- list(row = row, col = "from", val = "a")
      session$flushReact()
      upd$edit <- list(row = row, col = "to", val = "c")
      session$flushReact()
      upd$edit <- list(row = row, col = "input", val = "data")
      session$flushReact()

      session$setInputs(modify_links = 1)

      res <- update()$links

      expect_s3_class(res$add, "links")
      expect_length(res$add, 1L)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("manage links redraws only when the id set changes", {

  redraws <- 0L

  local_mocked_bindings(
    replaceData = function(...) redraws <<- redraws + 1L,
    .package = "DT"
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(aa = new_link(from = "a", to = "b"))
  )

  testServer(
    manage_links_server,
    {
      session$flushReact()

      expect_identical(names(upd$curr), "aa")
      expect_setequal(names(upd$obs), "aa")

      base <- redraws

      upd$curr <- links(aa = new_link(from = "a", to = ""))
      session$flushReact()

      expect_identical(names(upd$curr), "aa")
      expect_identical(redraws, base)
      expect_setequal(names(upd$obs), "aa")

      upd$curr <- links(
        aa = new_link(from = "a", to = ""),
        bb = new_link(from = "b", to = "")
      )
      session$flushReact()

      expect_gt(redraws, base)
      expect_setequal(names(upd$obs), c("aa", "bb"))
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("manage links ignores no-op board re-emits", {

  link_syncs <- 0L

  real_merge_links <- merge_staged_links

  local_mocked_bindings(
    merge_staged_links = function(...) {
      link_syncs <<- link_syncs + 1L
      real_merge_links(...)
    }
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(aa = new_link(from = "a", to = "b"))
  )

  testServer(
    manage_links_server,
    {
      session$flushReact()

      expect_gt(link_syncs, 0L)

      link_base <- link_syncs

      board$board <- new_board(
        blocks = c(
          a = new_dataset_block("iris"),
          b = new_subset_block()
        ),
        links = links(aa = new_link(from = "a", to = "b"))
      )
      session$flushReact()

      expect_identical(link_syncs, link_base)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("dummy ad/rm link ui test", {
  expect_s3_class(manage_links_ui("link", new_board()), "shiny.tag.list")
  expect_s3_class(links_modal(NS("links")), "shiny.tag")
})

test_that("variadic link inputs offer no positional-integer options", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(ac = new_link("a", "c"), bc = new_link("b", "c"))
  )

  dt <- dt_board_link(board_links(board), NS("x"), board)

  expect_false(grepl("<option[^>]*value=\"[0-9]", dt$Input[[1L]]))
  expect_false(grepl("<option[^>]*value=\"[0-9]", dt$Input[[2L]]))
})

test_that("a named variadic link renders its name in the editor", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(l1 = new_link("a", "c", "left"))
  )

  dt <- dt_board_link(board_links(board), NS("x"), board)

  expect_match(dt$Input[[1L]], "value=\"left\" selected", fixed = TRUE)
})

test_that("clearing a variadic input name registers as a staged change", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(l1 = new_link("a", "c", "left"))
  )

  testServer(
    manage_links_server,
    {
      session$flushReact()

      session$setInputs(l1_input = "")

      expect_identical(upd$edit, list(row = "l1", col = "input", val = ""))
      expect_length(upd$add, 1L)
      expect_identical(upd$add[["l1"]][["input"]], "")
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

#' Block argument specification
#'
#' Block constructor arguments can be documented with a structured
#' specification: each argument via `new_block_arg()` (a `description`, a single
#' worked `example`, and an optional machine-readable `type`), collected with
#' `new_block_args()`. A bare named character vector of descriptions, and the
#' empty `character()`, are also accepted and normalized into this form, so
#' existing registrations are unaffected. A single argument's fields are read
#' back with `block_arg_description()`, `block_arg_example()` and
#' `block_arg_type()`; block-level metadata (the whole argument set, worked
#' examples, guidance and keywords) is tabulated across blocks via
#' [block_metadata()].
#'
#' An argument's `type` is described with a small, dependency-free subset of
#' JSON Schema, built with the `arg_*()` constructors: `arg_string()`,
#' `arg_number()`, `arg_integer()` and `arg_boolean()` for scalars, `arg_enum()`
#' for a fixed set of string values, `arg_array()` for a homogeneous list and
#' `arg_object()` for a closed record of named fields (`additionalProperties:
#' false`). Each returns a plain nested list mirroring the schema it denotes,
#' consumed directly -- blockr.ai binds it via `ellmer::type_from_schema()`, an
#' MCP or raw tool schema reads the JSON as-is -- and worked examples registered
#' alongside an argument are validated against it (see [register_block()]).
#' Semantic intent (e.g. "a column in the upstream data", "an R expression") is
#' carried in `description`, not in the type vocabulary.
#'
#' The complete worked configuration of a block is the assembly of its
#' per-argument examples, keyed by argument name. When arguments interact, or
#' several few-shot examples are wanted, complete configurations are instead
#' supplied as a list via the `examples` argument of [register_block()] and
#' supersede that assembly; combining multiple per-argument examples is
#' intentionally not supported, as there is no safe way to form coherent
#' whole-block configurations from them.
#'
#' @param description Human- and model-facing description of an argument value
#' @param example A single worked value for an argument (or `NULL`)
#' @param type Optional machine-readable type for the argument, built with the
#'   `arg_*()` descriptor constructors. A plain JSON-Schema-subset list; worked
#'   examples are validated against it
#' @param x A `block_arg` object, or a bare string taken as its description
#' @param values Allowed string values, for `arg_enum()`
#' @param items Element descriptor, for `arg_array()`
#' @param required Whether the field is required, when nested in an
#'   `arg_object()`
#' @param ... For `new_block_args()`, the per-argument `block_arg` objects (or
#'   bare description strings); for `arg_object()`, the named field descriptors;
#'   ignored by the `block_arg_*()` getters
#'
#' @return
#' `new_block_arg()` returns a `block_arg` and `new_block_args()` a `block_args`
#' collection. The `arg_*()` constructors each return a plain JSON-Schema node
#' (a list). The `block_arg_*()` getters return the corresponding field of a
#' single argument (resolving a bare description string too).
#'
#' @examples
#' new_block_args(
#'   n = new_block_arg(
#'     "Number of rows to return",
#'     example = 5L,
#'     type = arg_integer()
#'   )
#' )
#'
#' arg_object(
#'   conditions = arg_array(
#'     arg_object(column = arg_string(), value = arg_string())
#'   ),
#'   operator = arg_enum(c("&", "|"))
#' )
#'
#' @name new_block_arg
#' @export
new_block_arg <- function(description = NULL, example = NULL, type = NULL) {

  if (not_null(description) && !is_string(description)) {
    blockr_abort(
      "A block argument `description` must be a string or `NULL`.",
      class = "block_arg_invalid"
    )
  }

  structure(
    list(description = description, example = example, type = type),
    class = "block_arg"
  )
}

#' @rdname new_block_arg
#' @export
new_block_args <- function(...) {
  as_block_args(list(...))
}

block_args_obj <- function(x) {
  structure(x, class = "block_args")
}

is_block_args <- function(x) inherits(x, "block_args")

as_block_arg <- function(x, ...) {
  UseMethod("as_block_arg")
}

#' @export
as_block_arg.block_arg <- function(x, ...) {
  x
}

#' @export
as_block_arg.character <- function(x, ...) {
  new_block_arg(description = x)
}

as_block_args <- function(x, ...) {
  UseMethod("as_block_args")
}

#' @export
as_block_args.block_args <- function(x, ...) {
  x
}

#' @export
as_block_args.list <- function(x, ...) {

  nms <- names(x)

  if (length(x) && (is.null(nms) || any(!nzchar(nms)))) {
    blockr_abort(
      "Every block argument must be named after a constructor formal.",
      class = "block_args_unnamed"
    )
  }

  block_args_obj(lapply(x, as_block_arg))
}

#' @export
as_block_args.character <- function(x, ...) {

  examples <- attr(x, "examples")

  nms <- names(x)

  if (is.null(nms)) {
    nms <- character()
  }

  spec <- set_names(
    lapply(
      nms,
      function(nm) {
        new_block_arg(description = x[[nm]], example = examples[[nm]])
      }
    ),
    nms
  )

  block_args_obj(spec)
}

normalize_arguments <- function(arguments, guidance) {

  prompt <- if (is.character(arguments)) attr(arguments, "prompt") else NULL

  if (is.character(arguments) &&
        (not_null(attr(arguments, "examples")) || not_null(prompt))) {
    deprecate_legacy_arg_attrs()
  }

  list(
    arguments = as_block_args(arguments),
    guidance = coal(guidance, prompt, fail_all = FALSE)
  )
}

assemble_example <- function(args) {

  ex <- lst_xtr(args, "example")
  ex <- ex[!lgl_ply(ex, is.null)]

  if (!length(ex)) {
    return(NULL)
  }

  ex
}

block_examples_list <- function(args, examples) {

  if (length(examples)) {
    return(examples)
  }

  assembled <- assemble_example(args)

  if (is.null(assembled)) {
    return(list())
  }

  list(assembled)
}

validate_block_spec <- function(args, examples, obj, ctor, ctor_ref, ctor_pkg,
                                check_names = TRUE) {

  inputs <- block_ctor_inputs(obj)

  if (check_names) {

    unknown <- setdiff(names(args), inputs)

    if (length(unknown)) {
      blockr_abort(
        "Block argument(s) {unknown} are not constructor formals ({inputs}).",
        class = "block_arg_unknown"
      )
    }
  }

  for (nm in names(args)) {

    type <- args[[nm]][["type"]]

    if (not_null(type) && !valid_descriptor(type)) {
      blockr_abort(
        "The `type` of block argument {nm} is not a valid descriptor.",
        class = "block_arg_type_invalid"
      )
    }
  }

  for (cfg in block_examples_list(args, examples)) {

    unknown <- setdiff(names(cfg), inputs)

    if (length(unknown)) {
      blockr_abort(
        "Example field(s) {unknown} are not constructor formals ({inputs}).",
        class = "block_example_unknown"
      )
    }

    for (nm in intersect(names(cfg), names(args))) {

      type <- args[[nm]][["type"]]

      if (not_null(type) && !conforms_to(cfg[[nm]], type)) {
        blockr_abort(
          "Example value for block argument {nm} does not conform to its ",
          "declared `type`.",
          class = "block_example_nonconforming"
        )
      }
    }

    res <- try(
      do.call(
        ctor,
        c(
          cfg,
          list(ctor = ctor_ref, ctor_pkg = ctor_pkg, block_metadata = FALSE)
        )
      ),
      silent = TRUE
    )

    if (inherits(res, "try-error")) {
      blockr_abort(
        "A registered block example does not construct: ",
        "{conditionMessage(attr(res, 'condition'))}",
        class = "block_example_invalid"
      )
    }
  }

  invisible()
}

as_legacy_arguments <- function(args, guidance, examples_list) {

  desc <- chr_ply(
    args,
    function(a) coal(a[["description"]], "", fail_all = FALSE)
  )

  example <- if (length(examples_list)) examples_list[[1L]] else NULL
  prompt <- if (length(guidance)) paste(guidance, collapse = "\n\n") else NULL

  structure(set_names(desc, names(args)), examples = example, prompt = prompt)
}

deprecate_legacy_arg_attrs <- function() {
  blockr_warn(
    "Passing block metadata as `examples`/`prompt` attributes on `arguments` ",
    "is deprecated; use `new_block_args()` / `new_block_arg()` and the ",
    "`guidance` argument of `register_block()` instead.",
    class = "deprecated_arg_attrs",
    frequency = "once",
    frequency_id = "blockr_deprecated_arg_attrs"
  )
}

#' @rdname new_block_arg
#' @export
block_arg_description <- function(x, ...) {
  UseMethod("block_arg_description")
}

#' @export
block_arg_description.block_arg <- function(x, ...) {
  x[["description"]]
}

#' @export
block_arg_description.default <- function(x, ...) {
  block_arg_description(as_block_arg(x), ...)
}

#' @rdname new_block_arg
#' @export
block_arg_example <- function(x, ...) {
  UseMethod("block_arg_example")
}

#' @export
block_arg_example.block_arg <- function(x, ...) {
  x[["example"]]
}

#' @export
block_arg_example.default <- function(x, ...) {
  block_arg_example(as_block_arg(x), ...)
}

#' @rdname new_block_arg
#' @export
block_arg_type <- function(x, ...) {
  UseMethod("block_arg_type")
}

#' @export
block_arg_type.block_arg <- function(x, ...) {
  x[["type"]]
}

#' @export
block_arg_type.default <- function(x, ...) {
  block_arg_type(as_block_arg(x), ...)
}

#' @rdname new_block_arg
#' @export
arg_string <- function(description = NULL, required = TRUE) {
  arg_node(list(type = "string", description = description), required)
}

#' @rdname new_block_arg
#' @export
arg_number <- function(description = NULL, required = TRUE) {
  arg_node(list(type = "number", description = description), required)
}

#' @rdname new_block_arg
#' @export
arg_integer <- function(description = NULL, required = TRUE) {
  arg_node(list(type = "integer", description = description), required)
}

#' @rdname new_block_arg
#' @export
arg_boolean <- function(description = NULL, required = TRUE) {
  arg_node(list(type = "boolean", description = description), required)
}

#' @rdname new_block_arg
#' @export
arg_enum <- function(values, description = NULL, required = TRUE) {

  if (!is.character(values) || !length(values)) {
    blockr_abort(
      "`arg_enum()` requires a non-empty character vector of values.",
      class = "arg_enum_invalid"
    )
  }

  arg_node(
    list(type = "string", enum = I(values), description = description),
    required
  )
}

#' @rdname new_block_arg
#' @export
arg_array <- function(items, description = NULL, required = TRUE) {
  arg_node(
    list(type = "array", items = items, description = description),
    required
  )
}

#' @rdname new_block_arg
#' @export
arg_object <- function(..., description = NULL, required = TRUE) {

  fields <- list(...)
  nms <- names(fields)

  if (length(fields) && (is.null(nms) || any(!nzchar(nms)))) {
    blockr_abort(
      "Every `arg_object()` field must be named.",
      class = "arg_object_unnamed"
    )
  }

  req <- nms[lgl_ply(fields, function(x) isTRUE(attr(x, "blockr_required")))]

  arg_node(
    list(
      type = "object",
      description = description,
      properties = fields,
      required = I(req),
      additionalProperties = FALSE
    ),
    required
  )
}

arg_node <- function(fields, required) {
  structure(
    fields[!lgl_ply(fields, is.null)],
    blockr_required = required
  )
}

valid_descriptor <- function(x) {

  if (!is.list(x) || !is_string(x[["type"]])) {
    return(FALSE)
  }

  type <- x[["type"]]
  kinds <- c("string", "number", "integer", "boolean", "array", "object")

  if (!type %in% kinds) {
    return(FALSE)
  }

  if (not_null(x[["enum"]]) &&
        (!identical(type, "string") || !is.character(x[["enum"]]) ||
           !length(x[["enum"]]))) {
    return(FALSE)
  }

  if (identical(type, "array")) {
    return(valid_descriptor(x[["items"]]))
  }

  if (identical(type, "object")) {
    return(valid_object_descriptor(x))
  }

  TRUE
}

valid_object_descriptor <- function(x) {

  props <- x[["properties"]]
  nms <- names(props)

  if (!is.list(props) ||
        (length(props) && (is.null(nms) || any(!nzchar(nms))))) {
    return(FALSE)
  }

  req <- x[["required"]]

  if (not_null(req) && (!is.character(req) || length(setdiff(req, nms)))) {
    return(FALSE)
  }

  all(lgl_ply(props, valid_descriptor))
}

conforms_to <- function(value, schema) {

  if (not_null(schema[["enum"]])) {
    return(is_string(value) && value %in% schema[["enum"]])
  }

  switch(
    schema[["type"]],
    string = is_string(value),
    number = is_number(value),
    integer = is_number(value) && value == round(value),
    boolean = is_bool(value),
    array = conforms_array(value, schema[["items"]]),
    object = conforms_object(value, schema),
    FALSE
  )
}

conforms_array <- function(value, items) {
  is.list(value) && all(lgl_ply(value, conforms_to, items))
}

conforms_object <- function(value, schema) {

  if (!is.list(value)) {
    return(FALSE)
  }

  nms <- names(value)

  if (length(value) && (is.null(nms) || any(!nzchar(nms)))) {
    return(FALSE)
  }

  if (is.null(nms)) {
    nms <- character()
  }

  props <- schema[["properties"]]

  if (length(setdiff(nms, names(props)))) {
    return(FALSE)
  }

  if (length(setdiff(schema[["required"]], nms))) {
    return(FALSE)
  }

  all(
    lgl_ply(
      seq_along(value),
      function(i) conforms_to(value[[i]], props[[nms[[i]]]])
    )
  )
}

test_that("the roclet reads block metadata off constructor tags", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo block
#' @blockDescr A demo
#' @blockCategory transform
#' @blockIcon funnel
#' @blockGuidance Keeps the first `n` rows;
#'   `direction` picks the end.
#' @blockKeywords head, tail, multi word
#' @blockArg n Number of rows to keep
#'   [example] 10L
#'   [type] arg_integer()
#' @blockArg direction Which end, head or tail
#'   [example] \"tail\"
#'   [type] arg_enum(c(\"head\", \"tail\"))
#' @export
new_demo_block <- function(n = 1, direction = \"head\") n
"

  res <- roxygen2::roc_proc_text(block_registration_roclet(), txt)[[1L]]

  expect_identical(res[["name"]], "demo block")
  expect_identical(res[["description"]], "A demo")
  expect_identical(res[["category"]], "transform")
  expect_identical(res[["icon"]], "funnel")
  expect_identical(
    res[["guidance"]],
    "Keeps the first `n` rows; `direction` picks the end."
  )
  expect_identical(res[["keywords"]], c("head", "tail", "multi word"))

  args <- res[["arguments"]]
  expect_identical(args[["n"]][["description"]], "Number of rows to keep")
  expect_identical(args[["n"]][["example"]], 10L)
  expect_identical(args[["n"]][["type"]][["type"]], "integer")
  expect_identical(
    args[["direction"]][["description"]],
    "Which end, head or tail"
  )
  expect_identical(args[["direction"]][["example"]], "tail")
  expect_identical(
    as.character(args[["direction"]][["type"]][["enum"]]),
    c("head", "tail")
  )
})

test_that("@blockArg allows an explicit description and type-only args", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo block
#' @blockDescr A demo
#' @blockCategory transform
#' @blockArg x
#'   [type] arg_string()
#'   [description] The x column
#' @export
new_demo_block <- function(x = \"\") x
"

  args <- roxygen2::roc_proc_text(
    block_registration_roclet(), txt
  )[[1L]][["arguments"]]

  expect_named(args, "x")
  expect_identical(args[["x"]][["description"]], "The x column")
  expect_identical(args[["x"]][["type"]][["type"]], "string")
  expect_null(args[["x"]][["example"]])
})

test_that("@blockExamples yields block-level worked configs", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo block
#' @blockDescr A demo
#' @blockCategory transform
#' @blockExamples list(
#'   list(n = 10L),
#'   list(n = 3L)
#' )
#' @export
new_demo_block <- function(n = 1) n
"

  res <- roxygen2::roc_proc_text(block_registration_roclet(), txt)[[1L]]

  expect_identical(res[["examples"]], list(list(n = 10L), list(n = 3L)))
})

test_that("details falls back to @section and @blockLink is captured", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo block
#' @blockDescr A demo
#' @blockCategory transform
#' @blockLink https://example.com/demo
#' @section Demo block: Prose about the block.
#' @export
new_demo_block <- function() NULL
"

  res <- roxygen2::roc_proc_text(block_registration_roclet(), txt)[[1L]]

  expect_identical(res[["details"]], "Prose about the block.")
  expect_identical(res[["link"]], "https://example.com/demo")
})

test_that("@blockDetails overrides the @section fallback", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo block
#' @blockDescr A demo
#' @blockCategory transform
#' @blockDetails An explicit detail.
#' @section Demo block: Ignored prose.
#' @export
new_demo_block <- function() NULL
"

  res <- roxygen2::roc_proc_text(block_registration_roclet(), txt)[[1L]]

  expect_identical(res[["details"]], "An explicit detail.")
})

test_that("block_link derives a pkgdown URL, gated on the URL existing", {

  skip_if_not_installed("roxygen2")

  block <- roxygen2::parse_text("
#' Title
#' @rdname some_topic
#' @export
new_x_block <- function() NULL
")[[1L]]

  expect_identical(
    derived_link(block, "https://example.com/pkg"),
    "https://example.com/pkg/reference/some_topic.html"
  )
  expect_null(derived_link(block, NULL))

  local_mocked_bindings(url_exists = function(url) TRUE)
  expect_identical(
    block_link(block, "https://example.com/pkg"),
    "https://example.com/pkg/reference/some_topic.html"
  )

  local_mocked_bindings(url_exists = function(url) FALSE)
  expect_null(block_link(block, "https://example.com/pkg"))
})

test_that("an explicit @blockLink is used without an HTTP check", {

  skip_if_not_installed("roxygen2")

  block <- roxygen2::parse_text("
#' Title
#' @blockLink https://example.com/custom
#' @export
new_x_block <- function() NULL
")[[1L]]

  local_mocked_bindings(url_exists = function(url) stop("must not be called"))

  expect_identical(
    block_link(block, "https://example.com/pkg"),
    "https://example.com/custom"
  )
})

test_that("only constructors carrying @block are picked up", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @export
untagged <- function() NULL
"

  expect_length(roxygen2::roc_proc_text(block_registration_roclet(), txt), 0L)
})

test_that("@blockCtor overrides the derived constructor name", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo
#' @blockDescr A demo
#' @blockCategory transform
#' @blockCtor make_demo
#' @export
new_demo_block <- function() NULL
"

  expect_named(
    roxygen2::roc_proc_text(block_registration_roclet(), txt),
    "make_demo"
  )
})

test_that("a @block missing a required companion tag aborts", {

  skip_if_not_installed("roxygen2")

  txt <- "
#' Title
#' @block demo
#' @export
new_demo_block <- function() NULL
"

  expect_error(
    roxygen2::roc_proc_text(block_registration_roclet(), txt),
    class = "block_roclet_invalid"
  )
})

test_that("YAML arguments become a block_args spec, examples included", {

  expect_identical(yaml_block_arguments(list()), character())
  expect_identical(yaml_block_arguments(list(arguments = list())), character())

  spec <- yaml_block_arguments(
    list(
      arguments = list(
        x = list(
          description = "d",
          example = "e",
          type = list(type = "integer")
        )
      )
    )
  )

  expect_s3_class(spec, "block_args")
  expect_identical(block_arg_description(spec[["x"]]), "d")
  expect_identical(block_arg_example(spec[["x"]]), "e")
  expect_identical(block_arg_type(spec[["x"]]), list(type = "integer"))

  desc_only <- yaml_block_arguments(
    list(arguments = list(x = list(description = "d")))
  )

  expect_identical(block_arg_description(desc_only[["x"]]), "d")
  expect_null(block_arg_example(desc_only[["x"]]))
  expect_null(block_arg_type(desc_only[["x"]]))
})

test_that("register_package_blocks is a no-op when no registry ships", {

  before <- list_blocks()

  expect_invisible(register_package_blocks(package = "utils"))
  expect_identical(list_blocks(), before)
})

test_that("register_package_blocks registers a subset and keeps examples", {

  curr <- list_blocks()

  withr::defer(
    {
      unregister_blocks()
      register_core_blocks(curr)
    }
  )

  unregister_blocks()
  register_package_blocks(package = "blockr.core", which = "new_dataset_block")

  expect_identical(list_blocks(), "dataset_block")

  args <- block_meta_arguments("dataset_block")

  expect_identical(
    block_arg_description(args[["dataset"]]),
    "Selects the dataset to use."
  )
  expect_identical(block_arg_example(args[["dataset"]]), "iris")

  expect_error(
    register_package_blocks(package = "blockr.core", which = "not_a_block")
  )
})

test_that("the shipped core registry is complete", {

  file <- system.file("registry/blocks.yml", package = "blockr.core")

  expect_true(nzchar(file))

  reg <- yaml::read_yaml(file)

  expect_length(reg, 10L)
  expect_identical(
    reg[["new_dataset_block"]][["arguments"]][["dataset"]][["example"]],
    "iris"
  )

  head <- reg[["new_head_block"]]
  expect_identical(head[["arguments"]][["n"]][["type"]][["type"]], "integer")
  expect_identical(
    head[["arguments"]][["direction"]][["type"]][["enum"]],
    c("head", "tail")
  )
  expect_true(is_string(head[["guidance"]]))
  expect_gt(length(head[["keywords"]]), 0L)

  expect_true(is_string(reg[["new_dataset_block"]][["details"]]))
  expect_length(reg[["new_merge_block"]][["examples"]], 2L)
  expect_match(
    reg[["new_dataset_block"]][["link"]],
    "/reference/new_data_block\\.html$"
  )
})

test_that("core blocks register with types, guidance and keywords", {

  args <- block_meta_arguments("head_block")

  expect_identical(block_arg_type(args[["n"]])[["type"]], "integer")

  direction <- block_arg_type(args[["direction"]])
  expect_identical(direction[["type"]], "string")
  expect_identical(as.character(direction[["enum"]]), c("head", "tail"))

  expect_true(is_string(block_meta_guidance("head_block")))
  expect_gt(length(block_meta_keywords("head_block")), 0L)

  merge_args <- block_meta_arguments("merge_block")
  expect_identical(block_arg_type(merge_args[["by"]])[["type"]], "array")
  expect_identical(block_arg_type(merge_args[["all_x"]])[["type"]], "boolean")

  expect_true(is_string(block_meta_details("dataset_block")))
  expect_match(
    block_meta_link("dataset_block"),
    "/reference/new_data_block\\.html$"
  )
  expect_identical(
    block_meta_examples("merge_block"),
    list(list(all_x = TRUE, all_y = FALSE), list(all_x = TRUE, all_y = TRUE))
  )
})

test_that("new_block_arg and new_block_args constructors", {

  a <- new_block_arg(description = "d", example = 1L, type = NULL)

  expect_s3_class(a, "block_arg")
  expect_identical(a[["description"]], "d")
  expect_identical(a[["example"]], 1L)
  expect_null(a[["type"]])

  spec <- new_block_args(n = "just a description")

  expect_s3_class(spec, "block_args")
  expect_s3_class(spec[["n"]], "block_arg")
  expect_identical(spec[["n"]][["description"]], "just a description")

  expect_length(new_block_args(), 0L)

  expect_error(new_block_arg(description = 1L), class = "block_arg_invalid")
  expect_error(new_block_args(new_block_arg("x")), class = "block_args_unnamed")
  expect_error(new_block_args(n = 1L))
})

test_that("block_arg field getters expose the spec without indexing", {

  a <- new_block_arg(description = "rows", example = 10L, type = NULL)

  expect_identical(block_arg_description(a), "rows")
  expect_identical(block_arg_example(a), 10L)
  expect_null(block_arg_type(a))

  # a bare description string resolves through as_block_arg()
  expect_identical(block_arg_description("just a description"),
                   "just a description")
  expect_null(block_arg_example("just a description"))
  expect_error(block_arg_description(1L))

  spec <- new_block_args(
    n = new_block_arg("rows", example = 10L),
    direction = new_block_arg("end", example = "tail")
  )

  expect_identical(
    vapply(spec, block_arg_description, character(1)),
    c(n = "rows", direction = "end")
  )
})

test_that("as_block_args dispatches by input type", {

  spec <- new_block_args(n = new_block_arg("rows"))

  expect_identical(as_block_args(spec), spec)
  expect_s3_class(as_block_args(list(n = new_block_arg("rows"))), "block_args")
  expect_s3_class(as_block_args(c(n = "rows")), "block_args")

  expect_error(as_block_args(1L))
  expect_error(as_block_arg(1L))
})

test_that("bare and empty arguments normalize without warning", {

  expect_silent(
    bare <- normalize_arguments(c(n = "rows", direction = "end"), NULL)
  )

  expect_s3_class(bare[["arguments"]], "block_args")
  expect_identical(bare[["arguments"]][["n"]][["description"]], "rows")
  expect_null(bare[["guidance"]])

  empty <- normalize_arguments(character(), NULL)

  expect_length(empty[["arguments"]], 0L)
})

test_that("legacy examples/prompt attributes are absorbed with a deprecation", {

  withr::local_options(rlib_warning_verbosity = "verbose")

  legacy <- structure(
    c(direction = "head or tail"),
    examples = list(direction = "tail"),
    prompt = "Pick head for the first rows."
  )

  expect_warning(
    norm <- normalize_arguments(legacy, NULL),
    class = "deprecated_arg_attrs"
  )

  expect_identical(norm[["guidance"]], "Pick head for the first rows.")
  expect_identical(norm[["arguments"]][["direction"]][["example"]], "tail")

  expect_identical(
    block_examples_list(norm[["arguments"]], list()),
    list(list(direction = "tail"))
  )
})

test_that("per-argument examples assemble into one whole-block configuration", {

  spec <- new_block_args(
    n = new_block_arg("count", example = 10L),
    direction = new_block_arg("end", example = "tail")
  )

  expect_identical(
    block_examples_list(spec, list()),
    list(list(n = 10L, direction = "tail"))
  )
})

test_that("block-level examples supersede the per-argument assembly", {

  spec <- new_block_args(direction = new_block_arg("end", example = "head"))

  authored <- list(
    list(direction = "head"),
    list(direction = "tail")
  )

  expect_identical(block_examples_list(spec, authored), authored)
})

test_that("registration hard-validates the structured spec form", {

  expect_error(
    register_block(
      new_head_block,
      name = "t",
      description = "t",
      uid = "ut_head_bad_arg",
      arguments = new_block_args(not_a_formal = new_block_arg("x"))
    ),
    class = "block_arg_unknown"
  )

  expect_error(
    register_block(
      new_head_block,
      name = "t",
      description = "t",
      uid = "ut_head_bad_example",
      arguments = new_block_args(
        direction = new_block_arg("end", example = "sideways")
      )
    ),
    class = "block_example_invalid"
  )

  expect_false("ut_head_bad_arg" %in% list_blocks())
  expect_false("ut_head_bad_example" %in% list_blocks())
})

test_that("registration validates worked examples against declared types", {

  new_stateful_block <- function(state = list(), ...) {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(expr = reactive(quote(identity(data))), state = list())
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "stateful_block",
      ...
    )
  }

  state_type <- arg_object(
    conditions = arg_array(arg_object(column = arg_string())),
    operator   = arg_enum(c("&", "|"))
  )

  withr::defer(
    if ("stateful_block" %in% list_blocks()) unregister_blocks("stateful_block")
  )

  expect_no_error(
    register_block(
      new_stateful_block,
      name = "Stateful",
      description = "d",
      uid = "stateful_block",
      arguments = new_block_args(
        state = new_block_arg(
          "the state",
          example = list(
            conditions = list(list(column = "Species")),
            operator = "&"
          ),
          type = state_type
        )
      )
    )
  )

  # malformed-but-constructible: conditions is a string, not an array
  expect_error(
    register_block(
      new_stateful_block,
      name = "Stateful",
      description = "d",
      uid = "stateful_block",
      overwrite = TRUE,
      arguments = new_block_args(
        state = new_block_arg(
          "the state",
          example = list(conditions = "not-a-list", operator = "&"),
          type = state_type
        )
      )
    ),
    class = "block_example_nonconforming"
  )

  # structural: a malformed type descriptor is rejected
  expect_error(
    register_block(
      new_stateful_block,
      name = "Stateful",
      description = "d",
      uid = "stateful_block",
      overwrite = TRUE,
      arguments = new_block_args(
        state = new_block_arg("the state", type = list(type = "frobnicate"))
      )
    ),
    class = "block_arg_type_invalid"
  )
})

test_that("block-level examples validate without a user-supplied spec", {

  withr::defer(unregister_blocks("ut_ex_only"))

  expect_no_error(
    register_block(
      new_dataset_block,
      name = "t",
      description = "t",
      uid = "ut_ex_only",
      examples = list(list(dataset = "iris"))
    )
  )

  expect_error(
    register_block(
      new_head_block,
      name = "t",
      description = "t",
      uid = "ut_ex_only_bad",
      examples = list(list(direction = "sideways"))
    ),
    class = "block_example_invalid"
  )
})

test_that("registry_metadata is deprecated but still projects legacy shape", {

  withr::local_options(rlib_warning_verbosity = "verbose")
  withr::defer(unregister_blocks("ut_head_legacy"))

  register_block(
    new_head_block,
    name = "Head",
    description = "First or last.",
    uid = "ut_head_legacy",
    arguments = new_block_args(
      direction = new_block_arg("head or tail", example = "tail")
    ),
    guidance = "Pick head for the first rows."
  )

  expect_warning(
    args <- registry_metadata("ut_head_legacy", "arguments"),
    class = "deprecated_registry_metadata"
  )

  args <- args[[1L]]

  expect_type(args, "character")
  expect_identical(as.character(args), "head or tail")
  expect_identical(attr(args, "examples"), list(direction = "tail"))
  expect_identical(attr(args, "prompt"), "Pick head for the first rows.")
})

test_that("arg type descriptor constructors build JSON-Schema nodes", {

  expect_identical(arg_string()[["type"]], "string")
  expect_identical(arg_number()[["type"]], "number")
  expect_identical(arg_integer()[["type"]], "integer")
  expect_identical(arg_boolean()[["type"]], "boolean")

  en <- arg_enum(c("a", "b"))
  expect_identical(en[["type"]], "string")
  expect_identical(as.character(en[["enum"]]), c("a", "b"))

  arr <- arg_array(arg_string())
  expect_identical(arr[["type"]], "array")
  expect_identical(arr[["items"]][["type"]], "string")

  obj <- arg_object(x = arg_string(), y = arg_integer(required = FALSE))
  expect_identical(obj[["type"]], "object")
  expect_identical(as.character(obj[["required"]]), "x")
  expect_false(obj[["additionalProperties"]])
  expect_named(obj[["properties"]], c("x", "y"))

  expect_error(arg_enum(character()), class = "arg_enum_invalid")
  expect_error(arg_object(arg_string()), class = "arg_object_unnamed")
})

test_that("valid_descriptor accepts the subset and rejects malformed types", {

  expect_true(valid_descriptor(arg_string()))
  expect_true(valid_descriptor(arg_object(x = arg_array(arg_enum(c("a"))))))

  expect_false(valid_descriptor("string"))
  expect_false(valid_descriptor(list(type = "frobnicate")))
  expect_false(valid_descriptor(list(type = "array")))
  expect_false(
    valid_descriptor(list(type = "object", properties = list(arg_string())))
  )
})

test_that("conforms_to checks R values against a descriptor", {

  expect_true(conforms_to("x", arg_string()))
  expect_false(conforms_to(5L, arg_string()))

  expect_true(conforms_to(5L, arg_integer()))
  expect_false(conforms_to(5.5, arg_integer()))

  expect_true(conforms_to(TRUE, arg_boolean()))
  expect_true(conforms_to("a", arg_enum(c("a", "b"))))
  expect_false(conforms_to("z", arg_enum(c("a", "b"))))

  arr <- arg_array(arg_string())
  expect_true(conforms_to(list("a", "b"), arr))
  expect_false(conforms_to("not-a-list", arr))
  expect_false(conforms_to(list("a", 5L), arr))

  obj <- arg_object(x = arg_string(), y = arg_integer(required = FALSE))
  expect_true(conforms_to(list(x = "a", y = 3L), obj))
  expect_true(conforms_to(list(x = "a"), obj))
  expect_false(conforms_to(list(y = 3L), obj))
  expect_false(conforms_to(list(x = "a", z = 1L), obj))
})

test_that("descriptors serialise to JSON Schema that ellmer reads back", {

  skip_if_not_installed("ellmer")

  d <- arg_object(
    conditions = arg_array(arg_object(column = arg_string())),
    operator   = arg_enum(c("&", "|"))
  )

  js <- as.character(jsonlite::toJSON(d, auto_unbox = TRUE))

  expect_match(js, '"required":["conditions","operator"]', fixed = TRUE)
  expect_match(js, '"enum":["&","|"]', fixed = TRUE)
  expect_no_error(ellmer::type_from_schema(text = js))
})

#' Parser block constructors
#'
#' Operating on results from blocks created via [new_file_block()], parser
#' blocks read (i.e. "parse") a file and make the contents available to
#' subsequent blocks for further analysis and visualization.
#'
#' If using the default validator for a parser block sub-class (i.e. not
#' overriding the `dat_valid` argument in the call to `new_parser_block()`),
#' the data argument corresponding to the input file name must be `file` in
#' order to match naming conventions in the validator function.
#'
#' @param ... Forwarded to `new_parser_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_parser_block()` inherit from
#' `parser_block`.
#'
#' @seealso The [blockr.io](https://bristolmyerssquibb.github.io/blockr.io/)
#' package provides real-world blocks for parsing external data formats (e.g.
#' `csv`, `xpt`).
#'
#' @export
new_parser_block <- function(server, ui, class, ctor = sys.parent(),
                             dat_valid = is_file, ...) {

  new_block(server, ui, c(class, "parser_block"), ctor, dat_valid = dat_valid,
            ...)
}

#' @export
block_output.parser_block <- function(x, result, session) {
  dt_result(result, x, session)
}

#' @export
block_ui.parser_block <- function(id, x, ...) {
  tagList(
    DT::dataTableOutput(NS(id, "result"))
  )
}

is_file <- function(file) {
  stopifnot(is_string(file), file.exists(file))
}

#' @export
#' @include utils-dt.R
block_render_trigger.parser_block <- dt_render_trigger

#' @export
board_options.parser_block <- function(x, ...) {
  combine_board_options(
    new_n_rows_option(...),
    new_page_size_option(...),
    new_filter_rows_option(...),
    NextMethod()
  )
}

#' Serve object
#'
#' Intended as entry point to start up a shiny app, the generic function
#' `serve()` can be dispatched either on a single block (mainly for previewing
#' purposes during block development) or an entire board
#'
#' @param x Object
#' @param ... Generic consistency
#'
#' @examplesShinylive
#' webr::install("blockr.core", repos = "https://cynkra.github.io/blockr.webR/")
#' library(blockr.core)
#' serve(
#'   new_merge_block("Time"),
#'   data = list(
#'     x = datasets::BOD,
#'     y = datasets::ChickWeight
#'   )
#' )
#'
#' @return The generic `serve()` is expected to return the result of a call to
#' [shiny::shinyApp()].
#'
#' @export
serve <- function(x, ...) {
  UseMethod("serve")
}

#' @param id Block ID
#' @param data Data inputs
#' @rdname serve
#' @export
serve.block <- function(x, id = "block", ..., data = list()) {

  init_data <- function(x, is_variadic) {

    if (!is_variadic) {
      return(reactiveVal(x))
    }

    const <- function(val) function() val

    do.call(reactives, lapply(x, const))
  }

  if (...length() && !length(data)) {
    data <- list(...)
  }

  nms <- names(data)

  if (is.null(nms)) {
    nms <- character(length(data))
  }

  dot_args <- !nms %in% block_inputs(x)

  if (!is.na(block_arity(x)) && any(dot_args)) {
    blockr_abort(
      "Unexpected arguments {nms[dot_args]}.",
      class = "unexpected_var_args"
    )
  }

  if (any(dot_args)) {
    data <- c(data[!dot_args], list(...args = data[dot_args]))
  }

  ui <- bslib::page_fluid(
    theme = bslib::bs_theme(version = 5),
    title = id,
    expr_ui(id, x),
    block_ui(id, x)
  )

  server <- function(input, output, session) {

    res <- block_server(id, x, Map(init_data, data, names(data) == "...args"))

    exportTestValues(
      result = export_safely(res$result())()
    )

    invisible()
  }

  shinyApp(ui, server)
}

#' @param id Board namespace ID
#' @param plugins Board plugins
#' @param options Board options
#' @param loader A [board_loader()] resolving the board to build for each
#' request; defaults to [local_loader()], the in-process save/restore handoff
#'
#' @rdname serve
#'
#' @examplesShinylive
#' webr::install("blockr.core", repos = "https://cynkra.github.io/blockr.webR/")
#' library(blockr.core)
#' serve(
#'   new_board(
#'     blocks = c(
#'       a = new_dataset_block("BOD"),
#'       b = new_dataset_block("ChickWeight"),
#'       c = new_merge_block("Time")
#'     ),
#'     links = c(
#'       ac = new_link("a", "c", "x"),
#'       bc = new_link("b", "c", "y")
#'     ),
#'     stacks = list(ac = c("a", "c"))
#'   )
#' )
#'
#' @export
serve.board <- function(x, id = rand_names(), plugins = blockr_app_plugins,
                        options = blockr_app_options, loader = local_loader(),
                        ...) {

  stopifnot(is_string(id), is.function(plugins), is.function(options))

  validate_board_loader(loader)

  shinyApp(
    serve_board_ui(id, x, loader, plugins, options),
    serve_board_srv(id, x, loader, plugins, options, ...)
  )
}

#' @rdname serve
#' @export
blockr_app_plugins <- function(x) {
  UseMethod("blockr_app_plugins")
}

#' @export
blockr_app_plugins.board <- function(x) {
  board_plugins(x)
}

#' @rdname serve
#' @export
custom_plugins <- function(x) {

  custom <- as_plugins(x)

  function(x) {

    default <- blockr_app_plugins(x)

    hit <- match(names(custom), names(default), nomatch = NA_integer_)

    if (all(is.na(hit))) {
      c(default, custom)
    } else {
      c(default[-hit[!is.na(hit)]], custom)
    }
  }
}

#' @rdname serve
#' @export
blockr_app_options <- function(x) {
  UseMethod("blockr_app_options")
}

#' @export
blockr_app_options.board <- function(x, ...) {
  combine_board_options(
    board_options(x),
    lapply(board_blocks(x), board_options),
    lapply(available_blocks(), board_options)
  )
}

#' @rdname serve
#' @export
custom_options <- function(x) {

  custom <- as_board_options(x)

  function(x) {

    default <- blockr_app_options(x)

    combine_board_options(
      custom,
      default
    )
  }
}

#' @rdname serve
#' @export
blockr_app_ui <- function(id, x, ...) {
  UseMethod("blockr_app_ui", x)
}

#' @export
blockr_app_ui.board <- function(id, x, ...) {
  bslib::page_fluid(
    theme = bslib::bs_theme(version = 5),
    board_ui(id, x, ...)
  )
}

#' @rdname serve
#' @export
blockr_app_server <- function(id, x, ...) {
  UseMethod("blockr_app_server", x)
}

#' @export
blockr_app_server.board <- function(id, x, ...) {
  board_server(id, x, ...)
}

serve_board_ui <- function(id, default, loader, plugins, options, ...) {

  args <- list(...)

  function(request) {

    x <- resolve_board(default, loader, request, NULL, options)
    id <- coal(attr(x, "id"), id)

    log_debug("building ui for board {id}")

    do.call(
      blockr_app_ui,
      c(list(id, x, plugins = plugins(x), options = options(x)), args)
    )
  }
}

serve_board_srv <- function(id, default, loader, plugins, options, ...) {

  args <- list(...)

  function(input, output, session) {

    onStop(revert(trace_observe()), session)

    x <- resolve_board(default, loader, session$request, session, options)
    id <- coal(attr(x, "id"), id)

    res <- do.call(
      blockr_app_server,
      c(list(id, x, plugins = plugins(x), options = options(x)), args)
    )

    board_refresh <- res[["board_refresh"]]

    if (not_null(board_refresh)) {
      observeEvent(
        board_refresh(),
        {
          if (not_null(loader$stage)) {
            loader$stage(board_refresh(), session)
          }

          session$reload()
        }
      )
    }

    blockr_test_exports(x, res)

    invisible()
  }
}

revert <- function(...) {
  funs <- Filter(is.function, list(...))
  function() {
    invisible(
      map(do.call, what = funs, MoreArgs = list(args = list()))
    )
  }
}

#' @export
serve.character <- function(x, ...) {

  stopifnot(is_string(x), file.exists(x))

  serve(
    blockr_deser(read_json(x)),
    ...
  )
}

#' @param rv Board [shiny::reactiveValues()]
#' @rdname serve
#' @export
blockr_test_exports <- function(x, rv, ...) {
  UseMethod("blockr_test_exports", x)
}

#' @export
blockr_test_exports.board <- function(x, rv, ...) {
  exportTestValues(
    result = {

      blocks <- rv[[1L]]$blocks
      ids <- intersect(board_block_ids(rv[[1L]]$board), names(blocks))

      lapply(
        lapply(
          lapply(lst_xtr(blocks[ids], "server", "result"), export_safely),
          reval
        ),
        reval
      )
    }
  )
}

#' User notification plugin module
#'
#' During the evaluation cycle of each block, conditions (errors, warnings and
#' messages) may be raised. The default `notify_user` plugin surfaces these to
#' the user as [shiny::showNotification()] toasts, displaying newly active
#' conditions and clearing ones that are no longer active via
#' [shiny::removeNotification()]. Each block's conditions (see [board_server()])
#' are tracked individually, so that a single block's change touches only its
#' own notifications rather than re-processing the whole board.
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `notify_user` is returned by
#' `notify_user()` and the UI component (e.g. `notify_user_ui()`) is expected
#' to return shiny UI (i.e. [shiny::tagList()]). The server component (i.e.
#' `notify_user_server()`) is called for the side effect of managing
#' notifications and returns `NULL`.
#'
#' @export
notify_user <- function(server = notify_user_server, ui = notify_user_ui) {
  new_plugin(server, ui, class = "notify_user")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname notify_user
#' @export
notify_user_server <- function(id, board, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      tracked <- new.env(parent = emptyenv())

      observe({

        ids <- names(board$blocks)

        isolate(
          sync_block_notifs(ids, tracked, board, session)
        )
      })

      invisible()
    }
  )
}

#' @rdname notify_user
#' @export
notify_user_ui <- function(id, board) {
  tagList(
    if (requireNamespace("cli", quietly = TRUE)) {
      tags$style(HTML(paste(format(cli::ansi_html_style()), collapse = "\n")))
    }
  )
}

sync_block_notifs <- function(ids, tracked, board, session) {

  for (id in setdiff(ids, ls(tracked))) {
    tracked[[id]] <- block_notif_observer(id, board, session)
  }

  for (id in setdiff(ls(tracked), ids)) {

    tracked[[id]]$destroy()
    rm(list = id, envir = tracked)
  }

  invisible()
}

block_notif_observer <- function(block_id, board, session) {

  conditions <- isolate(board$blocks[[block_id]]$server$conditions)

  shown <- character()

  obs <- observeEvent(
    conditions(),
    {
      frame <- notif_frame(conditions(), session)

      for (key in setdiff(shown, frame$key)) {
        notify_remove(key, session)
      }

      for (key in setdiff(frame$key, shown)) {
        notif_show(key, frame[frame$key == key, , drop = FALSE], session)
      }

      shown <<- frame$key
    },
    ignoreNULL = FALSE
  )

  list(
    destroy = function() {

      obs$destroy()

      for (key in shown) {
        notify_remove(key, session)
      }
    }
  )
}

notif_frame <- function(conditions, session = get_session()) {

  cnds <- coal(
    isolate(
      get_board_option_or_null("show_conditions", session)
    ),
    blockr_option("show_conditions", c("warning", "error"))
  )

  conditions <- conditions[conditions$severity %in% cnds, , drop = FALSE]

  conditions$key <- paste(
    conditions$block, conditions$severity, conditions$id,
    sep = "-"
  )

  conditions[!duplicated(conditions$key), , drop = FALSE]
}

notif_show <- function(key, row, session) {
  notify(
    "Block ", row$block, ": ", row$message,
    glue = FALSE,
    log = FALSE,
    duration = NULL,
    id = key,
    type = row$severity,
    session = session
  )
}

#' @section Head block:
#' Row-subsetting the first or last `n` rows of a `data.frame` (as provided by
#' [utils::head()] and [utils::tail()]) is implemented as `head_block`. This is
#' an example of a block that takes a single `data.frame` as input and produces
#' a single `data.frame` as output.
#'
#' @param n Number of rows
#' @param direction Either "head" or "tail"
#'
#' @block head/tail block
#' @blockDescr Data head/tail
#' @blockCategory transform
#' @blockIcon eye
#' @blockGuidance Keeps the first or last `n` rows of the input with
#'   `utils::head()` / `utils::tail()`. `direction` chooses the end: "head" for
#'   the first rows, "tail" for the last.
#' @blockKeywords head, tail, first, last, rows, limit
#' @blockArg n Number of rows to keep.
#'   [example] 10L
#'   [type] arg_integer()
#' @blockArg direction Row end to keep, either head (first) or tail (last).
#'   [example] "tail"
#'   [type] arg_enum(c("head", "tail"))
#'
#' @rdname new_transform_block
#' @export
new_head_block <- function(n = 6L, direction = c("head", "tail"), ...) {

  direction <- match.arg(direction)

  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {

          nrw <- reactiveVal(n)
          til <- reactiveVal(isTRUE(direction == "tail"))

          observeEvent(input$n, nrw(input$n))
          observeEvent(input$tail, til(input$tail))

          observeEvent(
            nrow(data()),
            updateNumericInput(
              inputId = "n",
              value = nrw(),
              min = 1L,
              max = nrow(data())
            )
          )

          list(
            expr = reactive(
              if (isTRUE(til())) {
                bbquote(utils::tail(.(data), n = .(n)), list(n = nrw()))
              } else {
                bbquote(utils::head(.(data), n = .(n)), list(n = nrw()))
              }
            ),
            state = list(
              n = nrw,
              direction = reactive(if (isTRUE(til())) "tail" else "head")
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        numericInput(
          inputId = NS(id, "n"),
          label = "Number of rows",
          value = n,
          min = 1L
        ),
        bslib::input_switch(
          id = NS(id, "tail"),
          label = "Tail",
          value = isTRUE(direction == "tail")
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data))
    },
    expr_type = "bquoted",
    class = "head_block",
    ...
  )
}

#' Plugin module for editing board blocks
#'
#' Logic and user experience for editing block attributes such as block titles
#' can be customized or enhanced by providing an alternate version of this
#' plugin. The default implementation only handles block titles, but if further
#' (editable) block attributes are to be introduced, corresponding UI and logic
#' can be included here. In addition to blocks titles, this default
#' implementation provides UI for removing, as well as inserting blocks before
#' or after the current one.
#'
#' @inheritParams new_plugin
#'
#' @return A plugin container inheriting from `edit_block` is returned by
#' `edit_block()`, while the UI component (e.g. `edit_block_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `edit_block_server()`) is expected to return `NULL`.
#'
#' @export
edit_block <- function(server = edit_block_server, ui = edit_block_ui,
                       validator = abort_not_null) {
  new_plugin(server, ui, validator, class = "edit_block")
}

#' @param id Namespace ID
#' @param block_id Block ID
#' @param board Reactive values object containing board information
#' @param update Reactive value object to initiate board updates
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname edit_block
#' @export
edit_block_server <- function(id, block_id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      initial_block <- isolate(
        board_blocks(board$board)[[block_id]]
      )

      cur_name <- reactive(
        {
          blk <- board_blocks(board$board)[[block_id]]
          if (is_block(blk)) block_name(blk) else NULL
        }
      )

      output$block_name_out <- renderUI(
        bslib::tooltip(cur_name(), paste("Block ID: ", block_id))
      )

      observeEvent(
        cur_name(),
        updateTextInput(
          session,
          "block_name_in",
          "Block name",
          cur_name()
        )
      )

      observeEvent(
        input$block_name_in,
        {
          req(input$block_name_in)
          if (!identical(cur_name(), input$block_name_in)) {
            update(
              list(
                blocks = list(
                  mod = set_names(
                    list(list(block_name = input$block_name_in)),
                    block_id
                  )
                )
              )
            )
          }
        }
      )

      output$block_summary <- renderText(
        block_summary(
          initial_block,
          reval_if(board$blocks[[block_id]]$server$result)
        )
      )

      observeEvent(
        input$rm_block,
        update(list(blocks = list(rm = block_id)))
      )

      position <- NULL

      if (block_has_inputs(initial_block)) {
        observeEvent(
          input$add_block_before,
          {
            inps <- setdiff(
              available_block_inputs(initial_block),
              occupied_block_inputs(block_id, board$board)
            )
            if (length(inps)) {
              position <<- "before"
              insert_block_modal(
                session,
                position,
                inputs = inps,
                stack_checkbox = is_block_in_stacks(block_id, board$board)
              )
            } else {
              notify(
                "All block inputs are used. Remove incoming link first.",
                type = "warning",
                session = session
              )
            }
          }
        )
      }

      observeEvent(
        input$add_block_after,
        {
          position <<- "after"
          insert_block_modal(
            session,
            position,
            stack_checkbox = is_block_in_stacks(block_id, board$board)
          )
        }
      )

      observeEvent(
        input$registry_select,
        {
          sel <- input$registry_select

          if (!is_string(sel) || !sel %in% list_blocks()) {

            notify(
              "Please choose a valid block type.",
              type = "warning",
              session = session
            )

            return()
          }

          if (identical(position, "after")) {
            updateSelectInput(
              session,
              "input_select",
              "Select input",
              available_block_inputs(sel)
            )
          }
        }
      )

      observeEvent(
        input$confirm_insert,
        {
          sel <- input$registry_select
          bid <- input$block_id

          if (!validate_block_addition(sel, bid, board$board, session)) {
            return()
          }

          blk <- create_block(sel)

          res <- list(
            blocks = list(
              add = as_blocks(set_names(list(blk), bid))
            )
          )

          if (identical(position, "before")) {

            if (!validate_link_addition(input$link_id, input$input_select,
                                        initial_block, board$board, session)) {
              return()
            }

            lnk <- set_names(
              list(new_link(bid, block_id, input$input_select)),
              input$link_id
            )

          } else if (identical(position, "after")) {

            if (!validate_link_addition(input$link_id, input$input_select,
                                        blk, board$board, session)) {
              return()
            }

            lnk <- set_names(
              list(new_link(block_id, bid, input$input_select)),
              input$link_id
            )

          } else {

            blockr_abort(
              "Unexpected postion {position}.",
              class = "unexpected_position_value"
            )
          }

          res <- c(res, list(links = list(add = as_links(lnk))))

          if (isTRUE(input$add_to_stack)) {

            stacks <- board_stacks(board$board)

            stk <- is_block_in_stack(block_id, stacks)
            stk <- names(stk)[stk]

            res <- c(
              res,
              list(
                stacks = list(
                  add = as_stacks(lapply(stacks[stk], union, bid)),
                  rm = stk
                )
              )
            )
          }

          update(res)

          removeModal()

          position <<- NULL
        }
      )

      observeEvent(
        input$cancel_insert,
        removeModal()
      )

      NULL
    }
  )
}

#' @param x Block
#' @rdname edit_block
#' @export
edit_block_ui <- function(x, id, ...) {
  tagList(
    bslib::card_header(
      class = "d-flex justify-content-between",
      uiOutput(NS(id, "block_name_out"), inline = TRUE),
      bslib::popover(
        bsicons::bs_icon("gear"),
        textInput(
          NS(id, "block_name_in"),
          "Block name"
        ),
        actionButton(
          NS(id, "rm_block"),
          "Remove block",
          icon = icon("circle-minus"),
          class = c("btn-danger", "mb-2")
        ),
        if (block_has_inputs(x)) {
          actionButton(
            NS(id, "add_block_before"),
            "Add block before",
            icon = icon("circle-plus"),
            class = c("btn-success", "mb-2")
          )
        },
        actionButton(
          NS(id, "add_block_after"),
          "Add block after",
          icon = icon("circle-plus"),
          class = "btn-success"
        ),
        title = "Block options",
        id = NS(id, "opts"),
        options = list(customClass = id)
      )
    ),
    ...,
    bslib::card_footer(
      class = "text-center",
      uiOutput(NS(id, "block_summary"), inline = TRUE)
    )
  )
}

#' @param data Result data
#' @rdname edit_block
#' @export
block_summary <- function(x, data) {
  UseMethod("block_summary")
}

#' @rdname edit_block
#' @export
block_summary.block <- function(x, data) {

  type <- try(type_desc(data), silent = TRUE)

  if (inherits(type, "try-error")) {
    type <- "??"
  }

  dims <- try(dim_desc(data), silent = TRUE)

  if (inherits(dims, "try-error")) {
    dims <- ""
  } else {
    dims <- paste0("[", dims, "]")
  }

  paste0("&lt;", type, dims, "&gt;")
}

type_desc <- function(x) {
  if (is.object(x)) {
    class(x)[[1]]
  } else if (rlang::is_vector(x)) {
    typeof(x)
  }
}

dim_desc <- function(x) {
  dim <- coal(dim(x), vec_size(x))
  format_dim <- big_mark(dim)
  format_dim[is.na(dim)] <- "??"
  paste0(format_dim, collapse = paste0(" &#10005; "))
}

big_mark <- function(x) {

  if (identical(getOption("OutDec"), ",")) {
    mark <- "."
  } else {
    mark <- ","
  }

  ret <- formatC(x, big.mark = mark, format = "d",
                 preserve.width = "individual")

  ret[is.na(x)] <- "??"

  ret
}

insert_block_modal <- function(session, pos, inputs = "",
                               stack_checkbox = FALSE) {

  ns <- session$ns

  ui <- modalDialog(
    title = switch(
      match.arg(pos, c("before", "after")),
      before = "Add block before",
      after = "Add block after"
    ),
    div(
      selectInput(
        ns("registry_select"),
        "Select block from registry",
        choices = list_suitable_blocks(pos)
      ),
      textInput(
        inputId = ns("block_id"),
        label = "Block ID",
        value = rand_names(),
        placeholder = "Enter a block ID."
      ),
      selectInput(
        ns("input_select"),
        "Select input",
        choices = inputs
      ),
      textInput(
        inputId = ns("link_id"),
        label = "Link ID",
        value = rand_names(),
        placeholder = "Enter a link ID."
      ),
      if (stack_checkbox) {
        bslib::input_switch(
          id = ns("add_to_stack"),
          label = "Add to stack",
          value = TRUE
        )
      }
    ),
    footer = tagList(
      actionButton(ns("cancel_insert"), "Cancel", class = "btn-danger"),
      actionButton(ns("confirm_insert"), "OK", class = "btn-success")
    )
  )

  showModal(ui, session)

  invisible()
}

list_suitable_blocks <- function(pos) {

  opts <- list_blocks()

  if (identical(pos, "before")) {
    return(opts)
  }

  stopifnot(identical(pos, "after"))

  arity <- int_ply(lapply(opts, create_block), block_arity)
  opts[is.na(arity) | arity >= 1L]
}

available_block_inputs <- function(block) {

  if (is.character(block)) {
    block <- create_block(block)
  }

  stopifnot(is_block(block))

  inp <- block_inputs(block)

  if (is.na(block_arity(block))) {
    inp <- c(inp, as.character(seq_len(length(inp) + 1L)))
  }

  inp
}

occupied_block_inputs <- function(block, links) {

  if (is_board(links)) {
    links <- board_links(links)
  }

  stopifnot(is_string(block), is_links(links))

  links[links$to == block]$input
}

validate_link_addition <- function(id, input, block, board, session) {

  if (nchar(id) == 0L || !is_string(id)) {

    notify(
      "Please choose a valid link ID.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  if (id %in% board_link_ids(board)) {

    notify(
      "Please choose a unique link ID.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  inps <- block_inputs(block)

  if (is.na(block_arity(block))) {
    inps <- c(inps, as.character(seq_len(length(inps) + 1L)))
  }

  if (!is_string(input) || !input %in% inps) {

    notify(
      "Please choose a valid block data input.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  TRUE
}

is_block_in_stack <- function(block, stacks) {

  is_el <- function(stk, el) is.element(el, stk)

  if (is_board(stacks)) {
    stacks <- board_stacks(stacks)
  }

  lgl_ply(stacks, is_el, block, use_names = TRUE)
}

is_block_in_stacks <- function(block, stacks) {
  any(is_block_in_stack(block, stacks))
}

block_has_inputs <- function(x) {
  is.na(block_arity(x)) || block_arity(x) >= 1L
}

#' Vector-valued links (i.e. `links` objects) manage IDs that are unique (with
#' empty strings being the only exception). These are either user-supplied or
#' generated automatically (via random strings) and operations like renaming,
#' concatenation and sub-assignments are guaranteed to not allow duplicated
#' IDs.
#'
#' In addition to unique IDs, `links` objects are guaranteed to be consistent
#' in that it is not possible to have multiple links pointing to the same
#' target (combination of `to` and `input` attributes). Furthermore, links
#' behave like edges in a directed acyclic graph (DAG) in that cycles are
#' detected and disallowed.
#'
#' @rdname new_link
#' @export
links <- function(...) {

  vals <- list_to_list_of_links(list(...))

  if (length(vals)) {

    if (is.null(names(vals))) {
      names(vals) <- rand_names(n = length(vals))
    }

    miss <- is.na(names(vals)) | names(vals) == ""

    if (any(miss)) {
      names(vals)[miss] <- rand_names(names(vals)[!miss], sum(miss))
    }

    ids <- names(vals)

  } else {

    vals <- list(lapply(new_link(), `[`, 0L))
    ids <- character()
  }

  res <- do.call(rbind, lapply(vals, as.data.frame))
  res <- cbind(id = ids, res)

  rownames(res) <- NULL

  validate_links(
    vctrs::new_rcrd(res, class = "links")
  )
}

harmonize_list_of_links <- function(x) {
  if (is_link(x)) {
    list(x)
  } else if (is_links(x)) {
    as.list(x)
  } else if (is.list(x) && all(lgl_ply(x, is_link))) {
    x
  } else {
    list(as_link(x))
  }
}

list_to_list_of_links <- function(x) {

  if (length(x) == 1L && is.character(x[[1L]])) {
    x <- list(as.list(x))
  }

  if (length(x) && all(lgl_ply(x, is.atomic))) {

    stopifnot(
      all(length(x[[1L]]) == lengths(x[-1L]))
    )

    vals <- lgl_ply(
      coal(names(x), rep("", length(x))),
      Negate(identical),
      "id"
    )

    res <- do.call(map, c(new_link, x[vals]))

    if ("id" %in% names(x)) {
      names(res) <- x[["id"]]
    } else {
      names(res) <- NULL
    }

    res

  } else {

    unlist(lapply(x, harmonize_list_of_links), recursive = FALSE)
  }
}

#' @rdname new_link
#' @export
is_links <- function(x) {
  inherits(x, "links")
}

#' @export
names.links <- function(x) {
  field(x, "id")
}

#' @export
`names<-.links` <- function(x, value) {

  if (is.null(value)) {
    value <- rep("", length(x))
  } else if (anyDuplicated(value) != 0L) {
    blockr_abort(
      "IDs are required to be unique.",
      class = "links_names_unique_invalid"
    )
  }

  field(x, "id") <- value

  x
}

#' @export
format.links <- function(x, ...) {
  res <- chr_ply(as.list(x), format, length = 2L, use_names = TRUE)
  res[2L, ]
}

#' @export
str_value.links <- function(x, ...) {
  str_value_collection(x, chr_ply(as.list(x), str_value))
}

#' @export
str.links <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
vec_restore.links <- function(x, to, ...) {
  validate_links(NextMethod())
}

#' @export
c.links <- function(...) {
  as_links(list_to_list_of_links(list(...)))
}

#' @export
`[.links` <- function(x, i, ...) {

  res <- links_slice(x, vec_as_location(i, length(x), names(x)))

  if (is.null(res)) {
    return(links())
  }

  res
}

#' @export
`[<-.links` <- function(x, i, ..., value) {

  i <- vec_as_location(i, length(x), names(x))

  if (is.null(value)) {
    return(links_slice(x, -i))
  }

  value <- list_to_list_of_links(value)

  trg_ids <- field(links_slice(x, i), "id")

  if (is.null(names(value))) {
    names(value) <- trg_ids
  }

  new_ids <- names(value)

  if (!setequal(new_ids, trg_ids)) {
    blockr_abort(
      "Replacing IDs {trg_ids} with {new_ids} is not allowed.",
      class = "links_assignment_name_invalid"
    )
  }

  links_assign(x, i, as_links(value[trg_ids]))
}

#' @export
`[[.links` <- function(x, i, ...) {

  res <- vec_proxy(
    links_slice(x, vec_as_location2(i, length(x), names(x)))
  )

  as_link(res[, colnames(res) != "id"])
}

#' @export
`[[<-.links` <- function(x, i, ..., value) {

  i <- vec_as_location2(i, length(x), names(x))
  val <- set_names(list(as_link(value)), names(x)[i])

  links_assign(x, i, as_links(val))
}

#' @export
`$.links` <- function(x, name) {
  field(x, name)
}

#' @export
`$<-.links` <- function(x, name, value) {
  `field<-`(x, name, value)
}

#' @rdname new_link
#' @export
as_links <- function(x, ...) {
  UseMethod("as_links")
}

#' @export
as_links.links <- function(x, ...) {
  validate_links(x)
}

#' @export
as_links.link <- function(x, ...) {
  as_links(as.list(x))
}

#' @export
as_links.list <- function(x, ...) {
  do.call(links, x)
}

#' @export
as_links.NULL <- function(x, ...) {
  links()
}

#' @method as_links data.frame
#' @export
as_links.data.frame <- function(x, ...) {
  as_links(as.list(x))
}

#' @export
as_links.board <- function(x, ...) {
  board_links(x)
}

#' @method as.data.frame links
#' @export
as.data.frame.links <- function(x, ...) {
  vec_proxy(x)
}

#' @method as.matrix links
#' @export
as.matrix.links <- function(x, ...) {
  todo <- x$from != "" & x$to != ""
  as_adjacency_matrix(x$from[todo], x$to[todo])
}

#' @rdname topo_sort
#' @export
is_acyclic.links <- function(x) {
  is_acyclic(as.matrix(x))
}

#' @method as.list links
#' @export
as.list.links <- function(x, ...) {

  df <- vec_proxy(x)

  lst <- split(df, seq_len(nrow(df)))

  res <- map(
    function(x, cols) as_link(lapply(x[, cols], na_to_empty)),
    lst,
    lapply(lapply(lst, colnames), setdiff, "id")
  )

  names(res) <- chr_xtr(lst, "id")

  res
}

links_slice <- function(...) {
  validate_links(vec_slice(...))
}

links_assign <- function(...) {
  validate_links(vec_assign(...))
}

#' @rdname new_link
#' @export
validate_links <- function(x) {

  any_dup <- function(x) {
    anyDuplicated(filter_empty(x))
  }

  if (is_board(x)) {
    x <- board_links(x)
  }

  if (!is_links(x)) {
    blockr_abort(
      "Expecting a board links objects to inherit from \"links\".",
      class = "links_class_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a board links objects behave list-like.",
      class = "links_list_like_invalid"
    )
  }

  fields <- c("id", "from", "to", "input")

  if (!all(fields %in% fields(x))) {
    blockr_abort(
      "Expecting the links to contain at least fields {fields}.",
      class = "links_fields_invalid"
    )
  }

  for (y in as.list(x)) {
    validate_link(y)
  }

  if (any_dup(field(x, "id")) != 0L) {
    blockr_abort(
      "Links IDs are required to be unique.",
      class = "links_names_unique_invalid"
    )
  }

  dup_inp <- lapply(split(field(x, "input"), field(x, "to")), any_dup)
  dup_inp <- lgl_ply(dup_inp, Negate(identical), 0L, use_names = TRUE)

  if (any(dup_inp)) {
    blockr_abort(
      "Block{?s} {names(dup_inp)[dup_inp]} {?has/have} multiple identical ",
      "inputs.",
      class = "links_block_inputs_invalid"
    )
  }

  if (!is_acyclic(x)) {
    blockr_abort("Links form a cycle.", class = "links_acyclic_invalid")
  }

  x
}

test_that("add/rm blocks", {

  testServer(
    manage_blocks_server,
    {
      expect_null(update())

      session$setInputs(
        registry_select = "dataset_block",
        block_id = "a",
        confirm_add = 1
      )

      session$flushReact()

      res <- update()$blocks

      expect_s3_class(res$add, "blocks")
      expect_length(res$add, 1L)
      expect_null(res$rm)

      expect_null(session$returned)
    },
    args = list(
      board = reactiveValues(board = new_board()),
      update = reactiveVal()
    )
  )

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(from = "a", to = "b")
  )

  testServer(
    manage_blocks_server,
    {
      expect_null(update())

      session$setInputs(block_select = "a", confirm_rm = 1)

      session$flushReact()

      res <- update()$blocks

      expect_type(res$rm, "character")
      expect_length(res$rm, 1L)
      expect_null(res$add)

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )

  testServer(
    manage_blocks_server,
    {
      expect_null(update())

      session$setInputs(confirm_add = 1)
      session$setInputs(block_id = "", confirm_add = 2)
      session$setInputs(block_id = "a", confirm_add = 3)
      session$setInputs(block_id = "c", registry_select = "abc",
                        confirm_add = 4)
      session$setInputs(cancel_add = 1)

      session$setInputs(confirm_rm = 1)
      session$setInputs(block_select = "x", confirm_rm = 2)
      session$setInputs(cancel_rm = 1)

      res <- update()$blocks

      expect_null(res$add)
      expect_null(res$rm)

      expect_null(session$returned)
    },
    args = list(board = reactiveValues(board = board), update = reactiveVal())
  )
})

test_that("add/rm blocks return validation", {

  expect_error(
    validate_board_update(list(blocks = "a"), new_board()),
    class = "board_update_component_type_invalid"
  )

  expect_error(
    validate_board_update(list(blocks = list(abc = NULL)), new_board()),
    class = "board_update_component_components_invalid"
  )

  expect_error(
    validate_board_update(list(blocks = list(add = "a")), new_board()),
    class = "board_update_add_component_invalid"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(add = blocks(a = new_dataset_block()))),
      new_board(blocks = blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_add_invalid"
  )

  expect_error(
    validate_board_update(list(blocks = list(rm = 1)), new_board()),
    class = "board_update_rm_component_invalid"
  )

  expect_error(
    validate_board_update(list(blocks = list(rm = "a")), new_board()),
    class = "board_update_blocks_rm_invalid"
  )
})

test_that("dummy add/rm block ui test", {
  expect_s3_class(manage_blocks_ui("add_rm", new_board()), "shiny.tag.list")
})

test_that("links class", {

  x <- new_link("a", "b")
  y <- new_link("b", "c")
  z <- new_link("c", "a")

  lnk <- links(x, y)

  expect_s3_class(lnk, "links")
  expect_true(is_links(lnk))
  expect_false(is_links("lnk"))
  expect_named(lnk)
  expect_named(links(x = x, y = y), c("x", "y"))

  expect_identical(lnk, as_links(lnk))
  expect_identical(lnk, as_links(as.list(lnk)))

  expect_error(links(x = x, x = y), class = "links_names_unique_invalid")
  expect_error(c(lnk, z), class = "links_acyclic_invalid")

  expect_snapshot(print(links(a = x, b = y)))

  z <- new_link("c", "d")

  lnks <- c(x, y)

  expect_s3_class(lnks, "links")

  expect_s3_class(c(lnks, z), "links")
  expect_s3_class(c(z, lnks), "links")

  lnks <- c(x = x, y = y)

  expect_s3_class(lnks, "links")
  expect_named(lnks, c("x", "y"))

  names(lnks) <- NULL

  expect_named(lnks, c("", ""))

  expect_error(
    names(lnks) <- c("x", "x"),
    class = "links_names_unique_invalid"
  )

  names(lnks) <- c("x", "y")

  expect_error(
    names(lnks)[1] <- "y",
    class = "links_names_unique_invalid"
  )

  expect_named(lnks, c("x", "y"))

  lnks[1] <- list(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  lnks[1] <- new_link(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  lnks[1] <- links(x = new_link(from = "a", to = "d"))
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  expect_error(
    lnks[1] <- links(y = new_link(from = "a", to = "d")),
    class = "links_assignment_name_invalid"
  )

  lnks[1] <- data.frame(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  lnks[[1]] <- list(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  lnks[[1]] <- new_link(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  expect_error(
    lnks[[1]] <- links(x = new_link(from = "a", to = "d"))
  )

  lnks[[1]] <- data.frame(from = "a", to = "d")
  expect_identical(lnks[["x"]], as_link(list(from = "a", to = "d")))

  expect_identical(as_links(lnks), lnks)
  expect_identical(as_links(as.list(lnks)), lnks)
  expect_identical(as_links(as.data.frame(lnks)), lnks)

  expect_error(
    validate_links("a"),
    class = "links_class_invalid"
  )

  expect_error(
    validate_links(structure(1, class = "links")),
    class = "links_list_like_invalid"
  )

  expect_error(
    validate_links(structure(list(input = "a"), class = "links")),
    class = "links_fields_invalid"
  )

  expect_error(
    validate_links(
      structure(
        list(
          id = c("x", "x"),
          from = c("a", "b"),
          to = c("d",  "c"),
          input = c("", "")
        ),
        class = c("links", "vctrs_rcrd", "vctrs_vctr")
      )
    ),
    class = "links_names_unique_invalid"
  )

  expect_error(
    validate_links(
      structure(
        list(
          id = c("x", "y"),
          from = c("a", "b"),
          to = c("c",  "c"),
          input = c("x", "x")
        ),
        class = c("links", "vctrs_rcrd", "vctrs_vctr")
      )
    ),
    class = "links_block_inputs_invalid"
  )

  expect_error(
    validate_links(
      structure(
        list(
          id = c("x", "y", "z"),
          from = c("a", "b", "c"),
          to = c("b",  "c", "a"),
          input = c("", "", "")
        ),
        class = c("links", "vctrs_rcrd", "vctrs_vctr")
      )
    ),
    class = "links_acyclic_invalid"
  )
})

test_that("a links container has a compact str_value()", {

  lks <- as_links(
    list(
      l1 = new_link(from = "a", to = "b"),
      l2 = new_link(from = "c", to = "d", input = "z")
    )
  )

  expect_identical(
    str_value(lks),
    "<links[2]>\n  l1: <link> a -> b\n  l2: <link> c -> d (z)"
  )

  expect_identical(str_value(links()), "<links[0]>")

  expect_identical(capture.output(str(lks))[1L], " <links[2]>")
})

#' Plugin module for editing board stacks
#'
#' Logic and user experience for editing stack attributes such as stack names
#' can be customized or enhanced by providing an alternate version of this
#' plugin. The default implementation only handles stack names, but if further
#' (editable) stack attributes are to be introduced, corresponding UI and logic
#' can be included here. In addition to stack names, this default
#' implementation provides UI for removing the current stack.
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `edit_stack` is returned by
#' `edit_stack()`, while the UI component (e.g. `edit_stack_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `edit_stack_server()`) is expected to return `NULL`.
#'
#' @export
edit_stack <- function(server = edit_stack_server, ui = edit_stack_ui) {
  new_plugin(server, ui, class = "edit_stack")
}

#' @param id Namespace ID
#' @param stack_id Stack ID
#' @param board Reactive values object containing board information
#' @param update Reactive value object to initiate board updates
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname edit_stack
#' @export
edit_stack_server <- function(id, stack_id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      cur_name <- reactive(
        stack_name(board_stacks(board$board)[[stack_id]])
      )

      output$stack_name_out <- renderUI(
        bslib::tooltip(cur_name(), paste("Stack ID: ", stack_id))
      )

      observeEvent(
        cur_name(),
        updateTextInput(
          session,
          "stack_name_in",
          "Stack name",
          cur_name()
        )
      )

      observeEvent(
        input$stack_name_in,
        {
          req(input$stack_name_in)
          if (!identical(cur_name(), input$stack_name_in)) {
            update(
              list(
                stacks = list(
                  mod = set_names(
                    list(list(name = input$stack_name_in)),
                    stack_id
                  )
                )
              )
            )
          }
        }
      )

      observeEvent(
        input$rm_stack,
        update(list(stacks = list(rm = stack_id)))
      )

      NULL
    }
  )
}

#' @param x Stack
#' @rdname edit_stack
#' @export
edit_stack_ui <- function(id, x, ...) {
  tagList(
    uiOutput(NS(id, "stack_name_out"), inline = TRUE),
    span(
      `data-bs-toggle` = "collapse",
      `data-bs-target` = "",
      bslib::popover(
        bsicons::bs_icon("gear"),
        textInput(
          NS(id, "stack_name_in"),
          "Stack name",
          stack_name(x)
        ),
        actionButton(
          NS(id, "rm_stack"),
          "Remove stack",
          icon = icon("circle-minus"),
          class = c("btn-danger", "mb-2")
        )
      )
    )
  )
}

#' Graph utils
#'
#' Block dependencies are represented by DAGs and graph utility functions
#' `topo_sort()` and `is_acyclic()` are used to create a topological ordering
#' (implemented as DFS) of blocks and to check for cycles. An adjacency matrix
#' corresponding to a board is available as `as.matrix()`.
#'
#' @param x Adjacency matrix/board object
#'
#' @examples
#' brd <- new_board(
#'   c(
#'      a = new_dataset_block(),
#'      b = new_dataset_block(),
#'      c = new_scatter_block(),
#'      d = new_subset_block()
#'   ),
#'   list(from = c("a", "d"), to = c("d", "c"))
#' )
#'
#' as.matrix(brd)
#' topo_sort(brd)
#' is_acyclic(brd)
#'
#' @return Topological ordering via `topo_sort()` returns a character vector
#' with sorted node IDs and the generic function `is_acyclic()` is expected to
#' return a scalar logical value.
#'
#' @export
topo_sort <- function(x) {

  if (is_board(x)) {
    x <- as.matrix(x)
  }

  stopifnot(
    is.matrix(x), nrow(x) == ncol(x),
    identical(rownames(x), colnames(x)),
    is.integer(x), all(x %in% c(0L, 1L))
  )

  num_nodes <- nrow(x)

  if (!num_nodes) {
    return(character())
  }

  dfs <- local({

    visited <- rep(FALSE, num_nodes)
    on_stack <- rep(FALSE, num_nodes)

    function(node, adj, stack) {

      if (on_stack[node]) {
        blockr_abort(
          "The graph contains a cycle and is not a DAG.",
          class = "graph_has_cycle"
        )
      }

      if (visited[node]) {
        return(stack)
      }

      visited[node] <<- TRUE
      on_stack[node] <<- TRUE

      for (neighbor in which(adj[node, ] == 1)) {
        stack <- Recall(neighbor, adj, stack)
      }

      on_stack[node] <<- FALSE

      c(node, stack)
    }
  })

  stack <- integer()

  for (node in seq_len(num_nodes)) {
    stack <- dfs(node, x, stack)
  }

  rownames(x)[stack]
}

as_adjacency_matrix <- function(from, to, nodes = union(from, to)) {

  stopifnot(
    all(from %in% nodes), all(to %in% nodes), anyDuplicated(nodes) == 0L
  )

  n <- length(nodes)

  res <- matrix(0L, n, n, dimnames = list(nodes, nodes))
  res[cbind(from, to)] <- 1L

  res
}

upstream_blocks <- function(ids, board) {

  links <- board_links(board)

  res <- ids
  new <- ids

  while (length(new)) {
    new <- setdiff(links$from[links$to %in% new], res)
    res <- c(res, new)
  }

  res
}

#' @param x Object
#' @rdname topo_sort
#' @export
is_acyclic <- function(x) {
  UseMethod("is_acyclic")
}

#' @rdname topo_sort
#' @export
is_acyclic.matrix <- function(x) {
  tryCatch(
    {
      topo_sort(x)
      TRUE
    },
    graph_has_cycle = function(e) FALSE
  )
}

test_that("topo sort", {

  # linear graph
  expect_identical(
    topo_sort(as_adjacency_matrix(letters[1:3], letters[2:4])),
    letters[1:4]
  )

  # diverging paths
  expect_identical(
    topo_sort(as_adjacency_matrix(from = c("a", "a", "b", "c"),
                                  to = c("b", "c", "d", "d"))),
    c("a", "c", "b", "d") # or c("a", "b", "c", "d")
  )

  # multiple roots
  expect_identical(
    topo_sort(as_adjacency_matrix(from = c("a", "c"), to = c("b", "d"))),
    c("c", "d", "a", "b") # or c("a", "b", "c", "d")
  )

  # cycle
  expect_error(
    topo_sort(as_adjacency_matrix(from = c("a", "b", "c"),
                                  to = c("b", "c", "a"))
    ),
    "The graph contains a cycle and is not a DAG."
  )

  # single node
  expect_identical(
    topo_sort(as_adjacency_matrix(character(), character(), "a")),
    "a"
  )

  # empty graph
  expect_identical(
    topo_sort(as_adjacency_matrix(from = character(), to = character())),
    character()
  )
})

test_that("upstream blocks", {

  board <- new_board(
    c(a = new_dataset_block("iris"), b = new_dataset_block("mtcars"),
      c = new_scatter_block(), d = new_subset_block()),
    list(from = c("a", "d"), to = c("d", "c"))
  )

  expect_setequal(upstream_blocks("c", board), c("a", "c", "d"))
  expect_setequal(upstream_blocks("d", board), c("a", "d"))
  expect_setequal(upstream_blocks("a", board), "a")
  expect_setequal(upstream_blocks("b", board), "b")
  expect_setequal(upstream_blocks(c("b", "c"), board), c("a", "b", "c", "d"))

  empty <- new_board(c(a = new_dataset_block("iris")))

  expect_setequal(upstream_blocks("a", empty), "a")
  expect_identical(upstream_blocks(character(), empty), character())
})

#' Serialization plugin module
#'
#' Board state can be preserved by serializing all contained objects and
#' restored via de-serialization. This mechanism can be used to power features
#' such as save/restore (via download, as implemented in the default
#' `preserve_board` plugin), but more refined user experience is conceivable
#' in terms of undo/redo functionality and (automatic) saving of board state.
#' Such enhancements can be implemented in a third-party `preserve_board`
#' module.
#'
#' The `session$reload()` that a restore triggers is handled by the app's
#' [board_loader()] (passed to [serve()]), not by the plugin -- the plugin
#' server simply returns the `board` to restore and never reloads.
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `preserve_board` is returned by
#' `preserve_board()`, while the UI component (e.g. `preserve_board_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component a [shiny::reactiveVal()] evaluating to `NULL` or the `board` to
#' restore.
#'
#' @export
preserve_board <- function(server = preserve_board_server,
                           ui = preserve_board_ui) {

  new_plugin(server, ui, validator = check_ser_deser_val,
             class = "preserve_board")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname preserve_board
#' @export
preserve_board_server <- function(id, board, ...) {

  dot_args <- list(...)

  moduleServer(
    id,
    function(input, output, session) {

      output$serialize <- downloadHandler(
        board_filename(board),
        do.call(
          write_board_to_disk,
          c(list(board), dot_args, list(session = session))
        )
      )

      res <- reactiveVal()

      observeEvent(
        input$restore,
        {
          board_ser <- read_json(input$restore$datapath)

          do.call(
            restore_board,
            c(
              list(board$board, board_ser, res),
              dot_args,
              list(session = session)
            )
          )
        }
      )

      res
    }
  )
}

read_json <- function(x) {
  jsonlite::fromJSON(
    x,
    simplifyDataFrame = FALSE,
    simplifyMatrix = FALSE
  )
}

#' @param x The current `board` object
#' @param new Serialized (list-based) representation of the new board
#' @param result A [shiny::reactiveVal()] to hold the new board object
#' @param session Shiny session
#'
#' @rdname preserve_board
#' @export
restore_board <- function(x, new, result, ..., session = get_session()) {
  UseMethod("restore_board")
}

#' @export
restore_board.board <- function(x, new, result, ..., session = get_session()) {
  result(blockr_deser(new))
}

#' @param board The initial `board` object
#' @rdname preserve_board
#' @export
preserve_board_ui <- function(id, board) {
  tagList(
    downloadButton(
      NS(id, "serialize"),
      "Save"
    ),
    htmltools::tagQuery(
      fileInput(
        NS(id, "restore"),
        "",
        buttonLabel = tagList(icon("upload"), "Restore")
      )
    )$addAttrs(
      style = "margin-bottom: 8px;"
    )$allTags()
  )
}

board_filename <- function(rv) {
  function() {
    paste0(
      rv$board_id,
      "_",
      format(Sys.time(), "%Y-%m-%d_%H-%M-%S"),
      ".json"
    )
  }
}

#' @param blocks Block state reactive values
#' @rdname preserve_board
#' @export
serialize_board <- function(x, blocks, id = NULL, ...,
                            session = get_session()) {

  UseMethod("serialize_board")
}

#' @export
serialize_board.board <- function(x, blocks, id = NULL, ...,
                                  session = get_session()) {

  states <- lst_xtr(blocks, "server", "state")
  states <- set_names(states[board_block_ids(x)], board_block_ids(x))

  blocks <- lapply(
    states,
    function(st) if (is.null(st)) NULL else lapply(st, reval_if)
  )

  opts <- lapply(
    set_names(nm = names(as_board_options(x))),
    get_board_option_or_null,
    session
  )

  blockr_ser(x, board_id = id, blocks = blocks, options = opts)
}

write_board_to_disk <- function(rv, ..., session = get_session()) {

  dot_args <- list(...)

  function(con) {

    json <- write_json(
      do.call(
        serialize_board,
        c(
          list(rv$board, rv$blocks, rv$board_id),
          dot_args,
          list(session = session)
        )
      )
    )

    writeLines(json, con)
  }
}

write_json <- function(x) {
  jsonlite::toJSON(x, null = "null")
}

check_ser_deser_val <- function(val) {
  observeEvent(
    TRUE,
    {
      if (!is.reactive(val)) {
        blockr_abort(
          "Expecting `preserve_board` to return a reactive value.",
          class = "preserve_board_return_invalid"
        )
      }
    },
    once = TRUE
  )

  observeEvent(
    val(),
    {
      board <- val()

      if (!is_board(board)) {
        blockr_abort(
          "Expecting the `preserve_board` return value to evaluate to a ",
          "`board` object.",
          class = "preserve_board_return_invalid"
        )
      }

      validate_board(board)
    },
    once = TRUE
  )

  val
}

#' @section Fixed block:
#' Mainly useful for testing and examples, this block applies a fixed
#' transformation to its data input. No UI elements are exposed and the
#' transformation consequently cannot be parametrized. The quoted expression
#' passed as `expr` is expected to refer to the input data as `data`.
#'
#' @param expr Quoted expression
#'
#' @rdname new_transform_block
#' @export
new_fixed_block <- function(expr, ...) {

  if (is_string(expr)) {
    expr <- str2lang(expr)
  }

  if (!is.language(expr)) {
    blockr_abort(
      "Expecting `expr` to be of type \"language\".",
      class = "invalid_block_input"
    )
  }

  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          list(
            expr = reactive(expr),
            state = list(expr = expr)
          )
        }
      )
    },
    class = "fixed_block",
    allow_empty_state = TRUE,
    external_ctrl = FALSE,
    block_metadata = list(),
    ...
  )
}

#' @export
blockr_ser.fixed_block <- function(x, ...) {
  res <- NextMethod()
  res$payload$expr <- paste(deparse(res$payload$expr), collapse = "\n")
  res
}

#' Board server
#'
#' A call to `board_server()`, dispatched on objects inheriting from `board`,
#' returns a [shiny::moduleServer()], containing all necessary logic to
#' manipulate board components via UI. Extensibility over currently available
#' functionality is provided in the form of S3, where a `board_server()`
#' implementation of `board` sub-classes may be provided, as well as via a
#' plugin architecture and callback functions which can be used to register
#' additional observers.
#'
#' @section Active conditions:
#' Conditions raised while blocks evaluate (errors, warnings and messages) are
#' exposed as a reactive data frame `board$conditions` on the read-only board
#' handed to plugins and callbacks, with one row per active condition and
#' columns `block`, `phase`, `severity`, `message` and `id`. It combines the
#' per-block `server$conditions` reactives (see [block_server()]), so a
#' consumer reads a single reactive — the whole board, or one block's frame
#' for fine-grained updates — rather than walking nested condition state. The
#' default [notify_user()] plugin renders its toasts from this source.
#'
#' @param x Board
#' @param id Parent namespace
#' @param ... Generic consistency
#'
#' @return A `board_server()` implementation (such as the default for the
#' `board` base class) is expected to return a [shiny::moduleServer()].
#'
#' @export
board_server <- function(id, x, ...) {
  UseMethod("board_server", x)
}

#' @param plugins Board plugins as modules
#' @param options Board options (`NULL` defaults to the union of board, block
#' and registry sourced options)
#' @param callbacks Single (or list of) callback function(s), called only
#' for their side-effects)
#' @param callback_location Location of callback invocation (before or after
#' plugins)
#' @rdname board_server
#' @export
board_server.board <- function(id, x, plugins = board_plugins(x),
                               options = board_options(x),
                               callbacks = list(),
                               callback_location = c("end", "start"),
                               ...) {

  plugins <- as_plugins(plugins)

  if (is.function(callbacks)) {
    callbacks <- list(callbacks)
  }

  validate_callbacks(callbacks)

  dot_args <- list(...) # nolint: object_usage_linter.

  callback_location <- match.arg(callback_location)

  stopifnot(is_board_options(options))

  # carry the resolved option set on rv$board so it serializes in full
  board_options(x) <- options

  moduleServer(
    id,
    function(input, output, session) {

      rv <- reactiveValues(
        blocks = list(),
        inputs = list(),
        sources = list(),
        board = x,
        board_id = id,
        stacks = list(),
        last_update = NULL,
        conditions = NULL,
        visible = TRUE
      )

      rv$eval <- reactiveValues()

      rv_ro <- list(board = make_read_only(rv))

      rv$conditions <- reactive(
        combine_block_conditions(
          lapply(rv$blocks, function(blk) blk$server$conditions())
        )
      )

      rv$needed <- reactiveVal(TRUE)

      observe(
        {
          cur <- if (isTRUE(rv$visible)) {
            TRUE
          } else {
            upstream_blocks(on_screen_blocks(rv$visible), rv$board)
          }

          old <- isolate(rv$needed())

          same <- if (isTRUE(cur) || isTRUE(old)) {
            identical(cur, old)
          } else {
            setequal(cur, old)
          }

          if (!same) {
            rv$needed(cur)
          }
        }
      )

      do.call(
        board_options_to_userdata,
        c(
          list(options),
          rv_ro,
          dot_args,
          list(session = session)
        )
      )

      board_update <- reactiveVal()
      board_visible <- reactiveVal()

      observe(
        {
          vis <- board_visible()

          if (is.null(vis) || !isTRUE(blockr_option("gate_visibility", TRUE))) {

            rv$visible <- TRUE

          } else {

            rv$visible <- vis

            log_debug(
              "visibility update: {length(on_screen_blocks(vis))}/",
              "{length(board_block_ids(rv$board))} on screen"
            )
          }
        }
      )

      cb_res <- set_names(
        vector("list", length(callbacks)),
        names(callbacks)
      )

      cb_args <- c(
        rv_ro,
        list(update = board_update, visible = board_visible),
        dot_args,
        list(session = session)
      )

      if (identical(callback_location, "start")) {

        for (i in seq_along(callbacks)) {
          cb_res[[i]] <- do.call(callbacks[[i]], cb_args)
        }

        if (length(cb_res) == 1L) {
          cb_res <- cb_res[[1L]]
        }

        dot_args <- c(dot_args, cb_res)
      }

      edit_block <- get_plugin("edit_block", plugins)
      ctrl_block <- get_plugin("ctrl_block", plugins)

      edit_stack <- get_plugin("edit_stack", plugins)

      edit_plugin_args <- c(
        rv_ro,
        list(update = board_update),
        dot_args
      )

      observeEvent(
        TRUE,
        setup_board(rv, edit_block, ctrl_block, edit_stack, edit_plugin_args,
                    session),
        once = TRUE
      )

      call_plugin_server(
        "manage_blocks",
        server_args = edit_plugin_args,
        plugins = plugins
      )

      call_plugin_server(
        "manage_links",
        server_args = edit_plugin_args,
        plugins = plugins
      )

      call_plugin_server(
        "manage_stacks",
        server_args = edit_plugin_args,
        plugins = plugins
      )

      update_seq <- 0L

      record_update_outcome <- function(ok, phase, message = NA_character_) {

        update_seq <<- update_seq + 1L

        rv$last_update <- list(
          seq = update_seq,
          ok = ok,
          phase = phase,
          message = message
        )
      }

      observeEvent(
        board_update(),
        {
          upd <- board_update()

          tryCatch(
            {
              if (is_board_locked(rv$board)) {

                log_debug("rejecting board update on locked board")
                board_update(NULL)

              } else {

                log_debug("starting board update")
                validate_board_update_structure(upd, rv$board)
                log_debug("board update validated")

                log_debug("preprocessing board update")
                if (!preprocess_board_update(board_update, rv$board)) {
                  log_debug("validating board links against board")
                  validate_board_update_xrefs(upd, rv$board)
                }
              }
            },
            error = function(e) {
              log_warn("board update rejected: {conditionMessage(e)}")
              notify(conditionMessage(e), type = "error", glue = FALSE,
                     session = session)
              record_update_outcome(FALSE, "validate", conditionMessage(e))
              board_update(NULL)
            }
          )
        },
        priority = Inf
      )

      observeEvent(
        board_update(),
        {
          upd <- board_update()

          tryCatch(
            {
              apply_core_board_update(
                rv, upd,
                session = session,
                edit_block = edit_block,
                ctrl_block = ctrl_block,
                edit_stack = edit_stack,
                edit_plugin_args = edit_plugin_args,
                dot_args = dot_args
              )

              new_board <- do.call(
                apply_board_update,
                c(
                  list(rv$board, upd),
                  dot_args,
                  list(session = session)
                )
              )

              stopifnot(is_board(new_board))
              rv$board <- new_board

              record_update_outcome(TRUE, "apply")
            },
            error = function(e) {
              log_warn("apply_board_update failed: {conditionMessage(e)}")
              notify(conditionMessage(e), type = "error", glue = FALSE,
                     session = session)
              record_update_outcome(FALSE, "apply", conditionMessage(e))
            }
          )

          board_update(NULL)

          log_debug("board update completed")
        },
        priority = -Inf
      )

      read_plugin_args <- c(rv_ro, dot_args)

      board_refresh <- call_plugin_server(
        "preserve_board",
        server_args = read_plugin_args,
        plugins = plugins
      )

      call_plugin_server(
        "notify_user",
        server_args = read_plugin_args,
        plugins = plugins
      )

      call_plugin_server(
        "generate_code",
        server_args = read_plugin_args,
        plugins = plugins
      )

      if (identical(callback_location, "end")) {

        for (i in seq_along(callbacks)) {
          cb_res[[i]] <- do.call(callbacks[[i]], cb_args)
        }

        dot_args <- c(dot_args, cb_res)
      }

      if (all(c("thematic", "dark_mode") %in% names(options))) {
        observeEvent(
          get_board_option_values(
            c("thematic", "dark_mode"),
            opts = options,
            if_not_found = "null"
          ),
          {
            if (isTRUE(get_board_option_or_null("thematic"))) {
              do.call(thematic::thematic_shiny, bs_theme_colors(session))
            } else if (isFALSE(get_board_option_or_null("thematic"))) {
              thematic::thematic_off()
            }
          }
        )
      }

      c(rv_ro, dot_args, list(board_refresh = board_refresh))
    }
  )
}

bs_theme_colors <- function(session) {

  theme <- bslib::bs_current_theme(session)

  if (!bslib::is_bs_theme(theme)) {
    return(
      list(bg = "auto", fg = "auto", accent = "auto")
    )
  }

  if ("3" %in% bslib::theme_version(theme)) {

    vars <- c("body-bg", "text-color", "link-color")

  } else {

    vars <- c("body-bg", "body-color", "link-color")

    if (identical(get_board_option_or_null("dark_mode"), "dark")) {
      vars <- paste0(vars, "-dark")
    }
  }

  set_names(
    as.list(bslib::bs_get_variables(theme, vars)),
    c("bg", "fg", "accent")
  )
}

deduped_board_reactive <- function(board, accessor) {

  val <- reactiveVal(isolate(accessor(board$board)))

  observe(
    {
      new <- accessor(board$board)

      if (!identical(new, isolate(val()))) {
        val(new)
      }
    }
  )

  val
}

setup_board <- function(rv, blk_ed, blk_ct, stk_mod, args, sess) {

  stopifnot(
    is.reactivevalues(rv),
    all(c("blocks", "inputs", "sources", "board", "stacks") %in% names(rv)),
    is_board(rv$board)
  )

  rv$blocks <- list()
  rv$inputs <- list()
  rv$sources <- list()
  rv$stacks <- list()

  observe(
    {
      need <- needed_block_ids(rv)
      construct_blocks(need, rv, blk_ed, blk_ct, args)
    }
  )

  construct_remaining_blocks(rv, blk_ed, blk_ct, args)

  setup_stacks(rv, stk_mod, args)
  add_blocks_to_stacks(rv, board_stacks(rv$board), sess)

  invisible()
}

combine_block_conditions <- function(frames) {

  res <- do.call(
    rbind,
    c(list(empty_conditions_frame()), unname(frames))
  )

  row.names(res) <- NULL

  res
}

construct_block <- function(id, rv, mod_ed, mod_ct, args) {

  if (id %in% names(rv$blocks)) {
    return(invisible())
  }

  blk <- board_blocks(rv$board)[[id]]

  rv$sources[[id]] <- reactiveValues()
  src_rv <- rv$sources[[id]]

  inpts <- set_names(
    lapply(block_inputs(blk), upstream_result, src_rv, rv, id),
    block_inputs(blk)
  )

  if (is.na(block_arity(blk))) {
    inpts <- c(inpts, list(`...args` = reactives()))
  }

  rv$inputs[[id]] <- inpts

  links <- board_links(rv$board)

  update_block_links(rv, links[links$to == id])

  inputs_ready <- reactive(block_inputs_ready(src_rv, blk, rv))

  srv <- do.call(
    block_server,
    c(
      list(paste0("block_", id), blk, rv$inputs[[id]], id, mod_ed, mod_ct),
      args,
      list(inputs_ready = inputs_ready)
    )
  )

  rv$blocks[[id]] <- list(block = blk, server = srv)

  rv$eval[[id]] <- reactive(block_eval_status(rv, id, inputs_ready, srv))

  invisible()
}

construct_blocks <- function(ids, rv, mod_ed, mod_ct, args) {

  ordered <- isolate(
    intersect(topo_sort(as.matrix(rv$board)), setdiff(ids, names(rv$blocks)))
  )

  if (!length(ordered)) {
    return(invisible())
  }

  log_debug("constructing block{?s} {ordered}")

  for (id in ordered) {
    isolate(construct_block(id, rv, mod_ed, mod_ct, args))
  }

  invisible()
}

construct_remaining_blocks <- function(rv, mod_ed, mod_ct, args) {

  if (background_construction_delay() <= 0) {

    construct_blocks(board_block_ids(rv$board), rv, mod_ed, mod_ct, args)

    return(invisible())
  }

  gate <- observe(
    {
      if (isTRUE(rv$visible) || visible_set_rendered(rv$visible)) {

        gate$destroy()

        construct_blocks_in_background(rv, mod_ed, mod_ct, args)
      }
    }
  )

  invisible()
}

construct_blocks_in_background <- function(rv, mod_ed, mod_ct, args) {

  started <- FALSE

  obs <- observe(
    {
      if (!started) {

        started <<- TRUE
        invalidateLater(background_construction_delay())

        return(invisible())
      }

      remaining <- isolate(
        setdiff(topo_sort(as.matrix(rv$board)), names(rv$blocks))
      )

      if (!length(remaining)) {

        obs$destroy()

        return(invisible())
      }

      isolate(construct_block(remaining[[1L]], rv, mod_ed, mod_ct, args))

      invalidateLater(background_construction_delay())
    }
  )

  invisible()
}

background_construction_delay <- function() {
  as.numeric(blockr_option("background_construction_delay", 50L))
}

on_screen_blocks <- function(status) {
  names(status)[status %in% c("required", "rendered")]
}

visible_set_rendered <- function(status) {
  all(status[on_screen_blocks(status)] == "rendered")
}

needed_block_ids <- function(rv) {

  need <- rv$needed()

  if (isTRUE(need)) board_block_ids(rv$board) else need
}

block_inputs_ready <- function(src_rv, blk, rv) {

  src <- reactiveValuesToList(src_rv)
  fixed <- block_inputs(blk)
  required <- setdiff(fixed, block_optional_inputs(blk))

  if (!all(lgl_ply(src[required], input_ready, rv))) {
    return(FALSE)
  }

  if (is.na(block_arity(blk))) {

    variadic <- src[setdiff(names(src), fixed)]

    if (sum(lgl_ply(variadic, input_ready, rv)) < block_min_args(blk)) {
      return(FALSE)
    }
  }

  TRUE
}

input_ready <- function(from, rv) {
  not_null(from) && identical(reval_if(rv$eval[[from]]), "ready")
}

block_eval_status <- function(rv, id, inputs_ready, srv) {

  if (!block_needed(rv, id)) {
    return("dormant")
  }

  if (!inputs_ready()) {
    return("waiting")
  }

  if (!isTRUE(srv$state_ready())) {
    return("unset")
  }

  if (isTRUE(srv$failed())) {
    return("failed")
  }

  "ready"
}

block_needed <- function(rv, id) {
  needed <- rv$needed()
  isTRUE(needed) || id %in% needed
}

apply_block_mod_delta <- function(blk_id, delta, rv) {

  state_rvs <- rv$blocks[[blk_id]]$server$state

  for (nm in names(delta)) {

    if (identical(nm, "block_name")) {

      blks <- board_blocks(rv$board)
      block_name(blks[[blk_id]]) <- delta[[nm]]
      board_blocks(rv$board) <- blks

    } else {

      cur <- reval_if(state_rvs[[nm]])

      if (!identical(cur, delta[[nm]])) {
        state_rvs[[nm]](delta[[nm]])
      }
    }
  }

  invisible()
}

destroy_rm_blocks <- function(ids, rv, sess, args) {

  links <- board_links(rv$board)

  update_block_links(
    rv,
    rm = links[links$from %in% ids | links$to %in% ids]
  )

  for (id in ids) {
    destroy_module(paste0("block_", id), session = sess)
    remove_block_from_stack(rv$board, id, rv$board_id, sess)
  }

  rv$inputs <- rv$inputs[!names(rv$inputs) %in% ids]
  rv$sources <- rv$sources[!names(rv$sources) %in% ids]
  rv$blocks <- rv$blocks[!names(rv$blocks) %in% ids]

  for (id in ids) {
    rv$eval[[id]] <- NULL
  }

  rv$board <- do.call(
    rm_blocks,
    c(list(rv$board, ids), args, list(session = sess))
  )

  invisible()
}

upstream_result <- function(key, src_rv, rv, to) {

  force(key)

  reactive(
    {
      needed <- rv$needed()
      req(isTRUE(needed) || to %in% needed)

      from <- src_rv[[key]]

      srv <- if (is.null(from)) NULL else isolate(rv$blocks[[from]])[["server"]]

      if (is.null(srv)) NULL else srv$result()
    }
  )
}

link_slot_key <- function(rv, to, id, input) {

  if (input %in% block_inputs(board_blocks(rv$board)[[to]])) input else id
}

setup_link <- function(rv, id, from, to, input) {

  rv$sources[[to]][[link_slot_key(rv, to, id, input)]] <- from

  invisible()
}

destroy_link <- function(rv, id, from, to, input) {

  src_rv <- rv$sources[[to]]

  if (input %in% block_inputs(board_blocks(rv$board)[[to]])) {
    src_rv[[input]] <- NULL
  } else {
    trim_rv(src_rv, id)
  }

  invisible()
}

variadic_links <- function(rv, to, add, rm) {

  lnks <- board_links(rv$board)

  present <- intersect(names(add), names(lnks))

  if (length(present)) {
    lnks[present] <- add[present]
  }

  fresh <- add[setdiff(names(add), present)]
  drop <- setdiff(names(rm), present)

  lnks <- c(lnks[!names(lnks) %in% drop], fresh)

  fixed <- block_inputs(board_blocks(rv$board)[[to]])

  lnks[lnks$to == to & !field(lnks, "input") %in% fixed]
}

sync_dot_args <- function(rv, to, lnks) {

  args <- rv$inputs[[to]][["...args"]]
  src_rv <- rv$sources[[to]]

  for (key in isolate(raw_keys(args))) {
    drop_reactive(args, key)
  }

  ids <- names(lnks)
  inputs <- field(lnks, "input")

  for (i in seq_along(ids)) {

    slot <- upstream_result(ids[[i]], src_rv, rv, to)

    if (nzchar(inputs[[i]])) {
      set_reactive(args, inputs[[i]], slot)
    } else {
      append_reactive(args, slot)
    }
  }

  invisible()
}

update_block_links <- function(rv, add = NULL, rm = NULL) {

  todo <- as.list(rm)

  for (i in names(todo)) {
    do.call(destroy_link, c(list(rv, i), todo[[i]]))
  }

  todo <- as.list(add)

  for (i in names(todo)) {
    do.call(setup_link, c(list(rv, i), todo[[i]]))
  }

  blks <- board_blocks(rv$board)

  touched <- unique(c(chr_xtr(as.list(rm), "to"), chr_xtr(as.list(add), "to")))
  touched <- touched[is.na(int_ply(blks[touched], block_arity))]

  for (to in touched) {
    sync_dot_args(rv, to, variadic_links(rv, to, add, rm))
  }

  invisible()
}

setup_stacks <- function(rv, mod, args, stacks = board_stacks(rv$board)) {

  serv <- get_plugin_server(mod)

  for (i in names(stacks)) {

    if (not_null(serv)) {
      serv(c(list(id = paste0("stack_", i), stack_id = i), args))
    }

    rv$stacks[[i]] <- character()
  }

  invisible()
}

destroy_stacks <- function(ids, rv, sess) {

  stopifnot(all(lengths(rv$stacks[ids]) == 0L))

  for (id in ids) {
    destroy_module(paste0("stack_", id), session = sess)
  }

  rv$stacks[ids] <- NULL

  invisible()
}

update_stack_blocks <- function(rv, upd, mod, args, session) {

  if (length(upd$rm)) {
    rm_blocks_from_stacks(rv, upd$rm, session)
    destroy_stacks(upd$rm, rv, session)
    remove_stack_ui(upd$rm, rv$board)
  }

  if (length(upd$add)) {
    setup_stacks(rv, mod, args, upd$add)
    insert_stack_ui(rv$board_id, upd$add, rv$board, mod)
    add_blocks_to_stacks(rv, upd$add, session)
  }

  if (length(upd$mod)) {
    update_blocks_in_stacks(rv, upd$mod, session)
  }

  invisible()
}

update_blocks_in_stacks <- function(rv, mod, sess) {

  for (i in names(mod)) {
    update_blocks_in_stack(i, rv, mod[[i]], sess)
  }

  invisible()
}

update_blocks_in_stack <- function(id, rv, val, sess) {

  stopifnot(is_string(id), id %in% names(rv$stacks))

  targ <- stack_blocks(val)

  for (i in setdiff(rv$stacks[[id]], targ)) {
    remove_block_from_stack(rv$board, i, rv$board_id, sess)
  }

  for (i in setdiff(targ, rv$stacks[[id]])) {
    add_block_to_stack(rv$board, i, id, sess)
  }

  rv$stacks[[id]] <- targ

  invisible()
}

rm_blocks_from_stacks <- function(rv, rm, session) {

  stopifnot(is.character(rm), all(rm %in% names(rv$stacks)))

  for (i in rm) {

    for (j in rv$stacks[[i]]) {
      remove_block_from_stack(rv$board, j, rv$board_id, session)
    }

    rv$stacks[[i]] <- character()
  }

  invisible()
}

add_blocks_to_stacks <- function(rv, add, session) {

  stopifnot(
    is_stacks(add),
    all(names(add) %in% names(rv$stacks)),
    all(lengths(rv$stacks[names(add)]) == 0L)
  )

  for (i in names(add)) {

    blks <- stack_blocks(add[[i]])

    for (j in stack_blocks(add[[i]])) {
      add_block_to_stack(rv$board, j, i, session)
    }

    rv$stacks[[i]] <- blks
  }

  invisible()
}

#' Board update
#'
#' Inside [board_server()] every state change flows through one
#' `board_update` reactive. Core registers two observers framing the
#' change: an initial one that validates the payload and runs
#' [augment_board_update()] for auto-fixups, and a final one that runs
#' [apply_board_update()] and resets the reactive. Plugins or
#' callbacks may register their own observers in between, provided they
#' use a *finite* priority — the highest and lowest reactive priorities
#' are reserved for core.
#'
#' All three functions dispatch on the `board` class. Subclasses
#' override to validate, augment, or react to their own payload slots,
#' typically composing with `NextMethod()`. [validate_board_update()]
#' is also a caller-facing entry point: it mirrors the initial
#' observer's checks against a caller-supplied payload, useful for
#' staging layers (e.g. accumulating LLM-proposed updates) that need
#' to fail loudly before publishing.
#'
#' @section Validation:
#' The default `.board` method runs a structural check on the payload
#' (block / link / stack per-slot rules) and a cross-reference check
#' that link endpoints and stack members resolve in the post-update
#' merged view. Unknown top-level keys are passed through, so subclass
#' payload slots reach subclass augment / apply methods.
#'
#' @section Augment:
#' The default `.board` method inserts implied link removals and stack
#' updates that follow from block removals, plus link-input
#' completion. Subclass methods may extend the payload with their own
#' fixups; an error thrown here aborts the update before apply runs.
#'
#' @section Apply:
#' The default `.board` method returns the supplied board unchanged —
#' the core apply path (block / link / stack mutation, block UI
#' insertion / removal) is not routed through this generic. Subclass
#' methods receive a plain `board` snapshot (no reactive surface) and
#' return a `board`, which the final observer assigns back to
#' `rv$board`. For piecemeal customization of the core apply path
#' itself, override the relevant sub-generic
#' ([modify_board_links()], [insert_block_ui()], etc.) instead.
#'
#' Errors thrown from either augment or apply are caught by the
#' observer, reported via [notify()], and the reactive is reset so the
#' app keeps running.
#'
#' @section Outcome:
#' Alongside the human-facing [notify()] toast, every update cycle
#' records a machine-readable result into `board$last_update` (the
#' read-only board handed to plugins and callbacks). It is a list with
#' a monotonically increasing `seq`, a logical `ok`, the `phase` it
#' ended in (`"validate"` or `"apply"`), and a `message`
#' (`conditionMessage()` on failure, `NA` on success); it is `NULL`
#' before the first update. The `seq` advances on every write so that
#' two consecutive identical outcomes still invalidate a downstream
#' observer. A programmatic caller can watch this field to learn
#' whether a dispatched update was rejected, failed to apply, or landed.
#'
#' @param payload,upd A board update payload — see Validation above
#' for the accepted shape.
#' @param board A `board` object.
#' @param ... Forwarded between methods. For [apply_board_update()],
#' the final observer also splices `board_server()`'s `...` in here.
#' @param session A shiny session, default [get_session()].
#'
#' @return [validate_board_update()] returns `invisible(payload)` (or
#' throws a `blockr_abort()` error). [augment_board_update()] returns
#' the (possibly extended) payload. [apply_board_update()] returns a
#' `board`.
#'
#' @seealso On a locked board (see [is_board_locked()]) the update is
#' dropped rather than applied.
#'
#' @examples
#' brd <- new_board(
#'   blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
#'   links = links(ab = new_link(from = "a", to = "b"))
#' )
#'
#' validate_board_update(
#'   list(links = list(rm = "ab")),
#'   brd
#' )
#'
#' try(
#'   validate_board_update(
#'     list(links = list(add = links(xy = new_link(from = "x", to = "y")))),
#'     brd
#'   )
#' )
#'
#' @name board_update
NULL

#' @rdname board_update
#' @export
validate_board_update <- function(payload, board, ...,
                                  session = get_session()) {
  UseMethod("validate_board_update", board)
}

#' @export
validate_board_update.board <- function(payload, board, ...,
                                        session = get_session()) {

  validate_board_update_structure(payload, board)
  validate_board_update_xrefs(payload, board)

  invisible(payload)
}

# nolint next: object_length_linter.
validate_board_update_structure <- function(payload, board) {

  exp_typ <- c("blocks", "links", "stacks")

  if (!is.list(payload)) {
    blockr_abort(
      "Expecting a board update to be specified as a list.",
      class = "board_update_type_invalid"
    )
  }

  exp_cmp <- c("rm", "add", "mod")

  for (typ in exp_typ) {

    if (!typ %in% names(payload)) {
      next
    }

    x <- payload[[typ]]

    if (!is.list(x)) {
      blockr_abort(
        "Expecting a board update component to be specified as a list.",
        class = "board_update_component_type_invalid"
      )
    }

    if (!length(names(x)) == length(x) || !all(names(x) %in% exp_cmp)) {
      blockr_abort(
        "Expecting a board update component to consist of components ",
        "{exp_cmp}. Please remove {setdiff(names(x), exp_cmp)}.",
        class = "board_update_component_components_invalid"
      )
    }

    if ("rm" %in% names(x) && !(is.null(x$rm) || is.character(x$rm))) {
      blockr_abort(
        "Expecting a board update `rm` component be specified as a character ",
        "vector (or NULL).",
        class = "board_update_rm_component_invalid"
      )
    }

    if ("add" %in% names(x) && !(is.null(x$add) || inherits(x$add, typ))) {
      blockr_abort(
        "Expecting a board update `add` component be specified as a {typ} ",
        "object (or NULL).",
        class = "board_update_add_component_invalid"
      )
    }

    if ("mod" %in% names(x) && !(is.null(x$mod) || is.list(x$mod))) {
      blockr_abort(
        "Expecting a board update `mod` component be specified as a named ",
        "list of per-entry argument deltas (or NULL).",
        class = "board_update_mod_component_invalid"
      )
    }
  }

  if ("blocks" %in% names(payload)) {
    validate_board_update_blocks(payload$blocks, board)
  }

  if ("links" %in% names(payload)) {
    validate_board_update_links(payload$links, board)
  }

  if ("stacks" %in% names(payload)) {
    validate_board_update_stacks(payload, board)
  }

  invisible()
}

has_comp <- function(comp, x) {
  comp %in% names(x) && length(x[[comp]])
}

has_comps <- function(comps, x, fun = `||`) {
  Reduce(fun, lgl_ply(comps, has_comp, x))
}

validate_board_update_blocks <- function(x, board) {

  all_ids <- board_block_ids(board)

  if (has_comp("rm", x)) {
    cur_ids <- setdiff(all_ids, x$rm)
  } else {
    cur_ids <- all_ids
  }

  if (has_comp("rm", x)) {

    if (!all(x$rm %in% all_ids)) {
      blockr_abort(
        "Expecting the removed block to be specified by a known ID.",
        class = "board_update_blocks_rm_invalid"
      )
    }
  }

  if (has_comp("add", x)) {

    if (any(names(x$add) %in% cur_ids)) {
      blockr_abort(
        "Expecting the newly added block to have a unique ID.",
        class = "board_update_blocks_add_invalid"
      )
    }

    validate_blocks(x$add)

    if (has_comp("mod", x) && length(intersect(names(x$add), names(x$mod)))) {
      blockr_abort(
        "Cannot add and modify the same IDs simulatneously.",
        class = "board_update_blocks_add_mod_clash"
      )
    }
  }

  if (has_comp("mod", x)) {

    validate_mod_deltas(x$mod, cur_ids, "blocks")

    blks <- board_blocks(board)

    for (blk_id in names(x$mod)) {

      delta <- x$mod[[blk_id]]
      allowed <- external_ctrl_vars(blks[[blk_id]])
      extra <- setdiff(names(delta), allowed)

      if (length(extra)) {
        blockr_abort(
          "Block `{blk_id}` mod delta contains argument{?s} {extra} which ",
          "{?is/are} not externally controllable. Use a `rm` + `add` payload ",
          "to replace the block.",
          class = "board_update_blocks_mod_not_ctrl"
        )
      }
    }
  }

  invisible()
}

validate_board_update_links <- function(x, board) {

  all_ids <- board_link_ids(board)

  if (has_comp("rm", x)) {
    cur_ids <- setdiff(all_ids, x$rm)
  } else {
    cur_ids <- all_ids
  }

  if (has_comp("rm", x)) {

    if (!all(x$rm %in% all_ids)) {
      blockr_abort(
        "Expecting all link IDs to be removed to be known.",
        class = "board_update_links_rm_invalid"
      )
    }
  }

  if (has_comp("add", x)) {

    if (any(names(x$add) %in% cur_ids)) {
      blockr_abort(
        "Expecting the newly added links to have a unique ID.",
        class = "board_update_links_add_invalid"
      )
    }

    validate_links(x$add)

    if (has_comp("mod", x) && length(intersect(names(x$add), names(x$mod)))) {
      blockr_abort(
        "Cannot add and modify the same IDs simulatneously.",
        class = "board_update_links_add_mod_clash"
      )
    }
  }

  if (has_comp("mod", x)) {

    validate_mod_deltas(x$mod, cur_ids, "links")
  }

  invisible()
}

validate_mod_deltas <- function(mod, cur_ids, typ) {

  err_class <- function(suffix) paste0("board_update_", typ, "_mod_", suffix)

  if (length(names(mod)) != length(mod) || any(nchar(names(mod)) == 0L)) {
    blockr_abort(
      "Expecting the {typ} `mod` component to be a named list keyed by ID.",
      class = err_class("unnamed")
    )
  }

  unknown <- setdiff(names(mod), cur_ids)

  if (length(unknown)) {
    blockr_abort(
      "Modified {typ} entries reference unknown ID{?s} {unknown}.",
      class = err_class("unknown_id")
    )
  }

  is_obj <- switch(typ, blocks = is_block, links = is_link, stacks = is_stack)

  for (id in names(mod)) {

    delta <- mod[[id]]

    if (is_obj(delta) || !is.list(delta) ||
          length(names(delta)) != length(delta) ||
          any(nchar(names(delta)) == 0L)) {
      blockr_abort(
        "Expecting each {typ} `mod` entry to be a named list of ",
        "constructor argument values.",
        class = err_class("entry_invalid")
      )
    }

    if (!length(delta)) {
      blockr_abort(
        "{typ} `mod` entry `{id}` is empty; omit it or supply at least one ",
        "argument.",
        class = err_class("entry_empty")
      )
    }
  }

  invisible()
}

combine_update_blocks <- function(upd, board) {

  all_blks <- board_blocks(board)

  if (!has_comp("blocks", upd)) {
    return(all_blks)
  }

  blk <- upd$blocks

  if (has_comp("rm", blk)) {
    all_blks <- all_blks[setdiff(names(all_blks), blk$rm)]
  }

  if (has_comp("add", blk)) {
    all_blks <- c(all_blks, blk$add)
  }

  all_blks
}

validate_board_update_xrefs <- function(payload, board) {

  payload <- augment_board_update(payload, board)

  if (has_comp("links", payload)) {
    lnk <- payload$links
  } else {
    return(invisible())
  }

  if (has_comps(c("add", "mod"), lnk)) {

    all_lnks <- board_links(board)

    if (has_comp("rm", lnk)) {
      all_lnks <- all_lnks[setdiff(names(all_lnks), lnk$rm)]
    }

    if (has_comp("add", lnk)) {
      all_lnks <- c(all_lnks, lnk$add)
    }

    if (has_comp("mod", lnk)) {
      merged <- Map(
        update_link,
        board_links(board)[names(lnk$mod)],
        lnk$mod
      )
      all_lnks <- c(
        all_lnks[setdiff(names(all_lnks), names(lnk$mod))],
        as_links(merged)
      )
    }

    validate_board_blocks_links(
      combine_update_blocks(payload, board),
      all_lnks
    )
  }

  invisible()
}

validate_board_update_stacks <- function(upd, board) {

  x <- upd$stacks

  all_stks <- board_stacks(board)

  if (has_comp("rm", x)) {

    if (!all(x$rm %in% names(all_stks))) {
      blockr_abort(
        "Expecting all stack IDs to be removed to be known.",
        class = "board_update_stacks_rm_invalid"
      )
    }

    all_stks <- all_stks[setdiff(names(all_stks), x$rm)]
  }

  if (has_comp("add", x)) {

    if (any(names(x$add) %in% names(all_stks))) {
      blockr_abort(
        "Expecting the newly added stacks to have a unique ID.",
        class = "board_update_stacks_add_invalid"
      )
    }

    if (has_comp("mod", x) && length(intersect(names(x$add), names(x$mod)))) {
      blockr_abort(
        "Cannot add and modify the same IDs simulatneously.",
        class = "board_update_stacks_add_mod_clash"
      )
    }

    validate_stacks(x$add)

    all_stks <- c(all_stks, x$add)
  }

  if (has_comp("mod", x)) {

    validate_mod_deltas(x$mod, names(all_stks), "stacks")

    merged <- Map(
      update_stack,
      board_stacks(board)[names(x$mod)],
      x$mod
    )

    all_stks <- c(
      all_stks[setdiff(names(all_stks), names(x$mod))],
      as_stacks(merged)
    )
  }

  if (has_comps(c("add", "mod"), x)) {
    validate_board_blocks_stacks(combine_update_blocks(upd, board), all_stks)
  }

  invisible()
}

#' @rdname board_update
#' @export
augment_board_update <- function(upd, board, ...,
                                 session = get_session()) {
  UseMethod("augment_board_update", board)
}

#' @export
augment_board_update.board <- function(upd, board, ...,
                                       session = get_session()) {

  if ("blocks" %in% names(upd) && "rm" %in% names(upd$blocks)) {

    rm <- upd$blocks$rm

    links <- board_links(board)

    mis_lnk <- setdiff(
      names(links[links$from %in% rm | links$to %in% rm]),
      upd$links$rm
    )

    merged_stks <- board_stacks(board)

    if (length(upd$stacks$mod)) {
      cur_for_mod <- board_stacks(board)[names(upd$stacks$mod)]
      mod_stks <- Map(update_stack, cur_for_mod, upd$stacks$mod)
      merged_stks[names(upd$stacks$mod)] <- as_stacks(mod_stks)
    }

    if (length(upd$stacks$rm)) {
      merged_stks <- merged_stks[setdiff(names(merged_stks), upd$stacks$rm)]
    }

    affected <- merged_stks[
      lengths(lapply(merged_stks, intersect, rm)) > 0L
    ]

    upd_stk <- lapply(
      affected,
      function(s) list(blocks = setdiff(stack_blocks(s), rm))
    )

  } else {

    mis_lnk <- NULL
    upd_stk <- NULL
  }

  add_lnk <- NULL

  if ("links" %in% names(upd) && "add" %in% names(upd$links)) {

    tmp <- complete_unary_inputs(upd$links$add, board_blocks(board))

    if (!identical(tmp$input, upd$links$add$input)) {
      add_lnk <- tmp
    }
  }

  if (length(mis_lnk)) {
    log_debug("adding link removal{?s} for {mis_lnk}")
    upd$links$rm <- c(mis_lnk, upd$links$rm)
  }

  if (length(upd_stk)) {
    log_debug("adding stack update{?s} for {names(upd_stk)}")
    upd$stacks$mod <- c(
      upd_stk,
      upd$stacks$mod[setdiff(names(upd$stacks$mod), names(upd_stk))]
    )
  }

  if (length(add_lnk)) {
    log_debug("adding link input update{?s} for {names(add_lnk)}")
    upd$links$add <- c(
      add_lnk,
      upd$links$add[setdiff(names(upd$links$add), names(add_lnk))]
    )
  }

  upd
}

#' @rdname board_update
#' @export
apply_board_update <- function(board, upd, ...,
                               session = get_session()) {
  UseMethod("apply_board_update", board)
}

#' @export
apply_board_update.board <- function(board, upd, ...,
                                     session = get_session()) {
  board
}

apply_core_board_update <- function(rv, upd, session,
                                    edit_block, ctrl_block, edit_stack,
                                    edit_plugin_args, dot_args = list()) {

  ns <- session$ns

  if (length(upd$blocks$add)) {

    log_debug("adding block{?s} {names(upd$blocks$add)}")

    do.call(
      insert_block_ui,
      c(
        list(ns(NULL), rv$board, upd$blocks$add),
        dot_args,
        list(
          edit_ui = edit_block,
          ctrl_ui = ctrl_block,
          session = session
        )
      )
    )

    board_blocks(rv$board) <- c(board_blocks(rv$board), upd$blocks$add)

    construct_blocks(names(upd$blocks$add), rv, edit_block, ctrl_block,
                     edit_plugin_args)
  }

  if (length(upd$blocks$mod)) {

    log_debug("modifying block{?s} {names(upd$blocks$mod)}")

    for (blk_id in names(upd$blocks$mod)) {

      delta <- upd$blocks$mod[[blk_id]]

      if (length(delta)) {
        construct_block(blk_id, rv, edit_block, ctrl_block, edit_plugin_args)
        apply_block_mod_delta(blk_id, delta, rv)
      }
    }
  }

  if (length(upd$links$mod)) {

    cur_lnks <- board_links(rv$board)[names(upd$links$mod)]
    merged <- as_links(Map(update_link, cur_lnks, upd$links$mod))

    upd$links$add <- c(upd$links$add, merged)
    upd$links$rm <- c(upd$links$rm, names(upd$links$mod))
  }

  if (length(upd$links$add) || length(upd$links$rm)) {

    if (length(upd$links$add)) {
      log_debug("adding link{?s} {names(upd$links$add)}")
    }

    if (length(upd$links$rm)) {
      log_debug("removing link{?s} {names(upd$links$rm)}")
    }

    rm <- board_links(rv$board)[upd$links$rm]
    update_block_links(rv, upd$links$add, rm)

    rv$board <- do.call(
      modify_board_links,
      c(
        list(rv$board, upd$links$add, upd$links$rm),
        dot_args,
        list(session = session)
      )
    )
  }

  if (length(upd$stacks$rm)) {
    log_debug("removing stack{?s} {names(upd$stacks$rm)}")
  }

  if (length(upd$stacks$add)) {
    log_debug("adding stack{?s} {names(upd$stacks$add)}")
  }

  if (length(upd$stacks$mod)) {

    log_debug("modifying stack{?s} {names(upd$stacks$mod)}")

    cur_stks <- board_stacks(rv$board)[names(upd$stacks$mod)]
    upd$stacks$mod <- as_stacks(
      Map(update_stack, cur_stks, upd$stacks$mod)
    )
  }

  update_stack_blocks(
    rv, upd$stacks, edit_stack, edit_plugin_args, session
  )

  rv$board <- do.call(
    modify_board_stacks,
    c(
      list(rv$board, upd$stacks$add, upd$stacks$rm, upd$stacks$mod),
      dot_args,
      list(session = session)
    )
  )

  if (length(upd$blocks$rm)) {

    log_debug("removing block{?s} {names(upd$blocks$rm)}")

    do.call(
      remove_block_ui,
      c(
        list(ns(NULL), rv$board, upd$blocks$rm),
        dot_args,
        list(
          edit_ui = edit_block,
          ctrl_ui = ctrl_block,
          session = session
        )
      )
    )

    destroy_rm_blocks(upd$blocks$rm, rv, session, dot_args)
  }

  invisible()
}

preprocess_board_update <- function(update, board) {

  upd <- update()
  augmented <- augment_board_update(upd, board)

  if (identical(augmented, upd)) {
    return(FALSE)
  }

  update(augmented)
  TRUE
}

probe_render <- new.env()
probe_render$ids <- character()

probe_eval <- new.env()
probe_eval$ids <- character()

probe_args <- new.env()
probe_args$entry_classes <- NULL

probe_construct <- new.env()
probe_construct$ids <- character()

registerS3method(
  "block_output", "probe_block",
  function(x, result, session) {
    probe_render$ids <- c(probe_render$ids, session$ns(NULL))
    NULL
  }
)

registerS3method(
  "expr_server", "probe_block",
  function(x, data, ...) {
    probe_construct$ids <- c(probe_construct$ids, attr(x, "probe_id"))
    NextMethod()
  }
)

registerS3method(
  "block_ui", "probe_block",
  function(id, x, ...) shiny::tagList()
)

registerS3method(
  "block_eval", "probe_block",
  function(x, expr, env, ...) {
    probe_eval$ids <- c(probe_eval$ids, attr(x, "probe_id"))
    NextMethod()
  }
)

probe_source <- function() {
  new_data_block(
    function(id) {
      moduleServer(
        id,
        function(input, output, session) {
          list(expr = reactive(quote(datasets::BOD)), state = list())
        }
      )
    },
    function(id) shiny::tagList(),
    class = "probe_block",
    block_metadata = FALSE
  )
}

probe_passthrough <- function() {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          list(expr = reactive(quote(identity(data))), state = list())
        }
      )
    },
    function(id) shiny::tagList(),
    class = "probe_block",
    block_metadata = FALSE
  )
}

probe_data_observer <- function() {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          observeEvent(data(), NULL)
          list(expr = reactive(quote(identity(data))), state = list())
        }
      )
    },
    function(id) shiny::tagList(),
    class = "probe_block",
    block_metadata = FALSE
  )
}

probe_variadic <- function() {
  new_transform_block(
    function(id, ...args) {
      moduleServer(
        id,
        function(input, output, session) {

          observe(
            {
              ks <- names(...args)
              req(length(ks) > 0)
              probe_args$entry_classes <- chr_ply(
                ks, function(k) class(...args[[k]])[1L]
              )
            }
          )

          list(expr = reactive(quote(datasets::BOD)), state = list())
        }
      )
    },
    function(id) shiny::tagList(),
    class = "probe_block",
    block_metadata = FALSE
  )
}

with_id <- function(blk, id) {
  attr(blk, "probe_id") <- id
  blk
}

reset_probes <- function() {
  probe_render$ids <- character()
  probe_eval$ids <- character()
  probe_construct$ids <- character()
}

rendered <- function(id) {
  any(endsWith(probe_render$ids, paste0("block_", id)))
}

evaluated <- function(id) {
  id %in% probe_eval$ids
}

constructed <- function(id) {
  id %in% probe_construct$ids
}

vis_map <- function(..., level = "rendered") {
  ids <- c(...)
  set_names(rep(level, length(ids)), ids)
}

test_that("with no producer every block is visible", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b")
    ),
    links = links(new_link(from = "a", to = "b"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(rv$visible)

      expect_true(evaluated("a"))
      expect_true(evaluated("b"))

      expect_true(rendered("a"))
      expect_true(rendered("b"))
    },
    args = list(x = board, plugins = list())
  )
})

test_that("a producer gates evaluation and rendering on visibility", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b"),
      c = with_id(probe_passthrough(), "c"),
      d = with_id(probe_passthrough(), "d")
    ),
    links = links(
      new_link(from = "a", to = "b"),
      new_link(from = "b", to = "c"),
      new_link(from = "a", to = "d")
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_setequal(on_screen_blocks(rv$visible), "b")

      expect_true(evaluated("b"))
      expect_true(rendered("b"))

      expect_true(evaluated("a"))
      expect_false(rendered("a"))

      expect_false(evaluated("c"))
      expect_false(rendered("c"))

      expect_false(evaluated("d"))
      expect_false(rendered("d"))

      board_visible(vis_map("b", "c", "d"))
      session$flushReact()

      expect_true(evaluated("c"))
      expect_true(rendered("c"))

      expect_true(evaluated("d"))
      expect_true(rendered("d"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b"))
        NULL
      }
    )
  )
})

test_that("the gate_visibility option disables gating", {

  reset_probes()

  withr::local_options(blockr.gate_visibility = FALSE)

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b"),
      c = with_id(probe_passthrough(), "c")
    ),
    links = links(
      new_link(from = "a", to = "b"),
      new_link(from = "b", to = "c")
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      for (id in c("a", "b", "c")) {
        expect_true(evaluated(id))
        expect_true(rendered(id))
      }
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b"))
        NULL
      }
    )
  )
})

test_that("a link change re-routes the pulled upstream", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      c = with_id(probe_source(), "c"),
      b = with_id(probe_passthrough(), "b")
    ),
    links = links(ab = new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(evaluated("a"))
      expect_true(evaluated("b"))
      expect_false(evaluated("c"))

      reset_probes()

      board_update(
        list(
          links = list(rm = "ab", add = links(cb = new_link("c", "b", "data")))
        )
      )
      session$flushReact()

      expect_true(evaluated("c"))
      expect_true(evaluated("b"))
      expect_false(evaluated("a"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b"))
        NULL
      }
    )
  )
})

test_that("an off-screen data-observing block does not pull its upstream", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_data_observer(), "b"),
      c = with_id(probe_source(), "c")
    ),
    links = links(new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(evaluated("c"))

      expect_false(evaluated("a"))
      expect_false(evaluated("b"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("c"))
        NULL
      }
    )
  )
})

test_that("an unrelated structural edit does not re-evaluate needed blocks", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b"),
      x = with_id(probe_source(), "x")
    ),
    links = links(new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(evaluated("a"))
      expect_true(evaluated("b"))
      expect_false(evaluated("x"))

      reset_probes()

      board_update(
        list(blocks = list(mod = list(x = list(block_name = "renamed"))))
      )
      session$flushReact()

      expect_false(evaluated("a"))
      expect_false(evaluated("b"))
      expect_false(evaluated("x"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b"))
        NULL
      }
    )
  )
})

test_that("adding a block does not re-evaluate existing needed blocks", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b")
    ),
    links = links(new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(evaluated("a"))
      expect_true(evaluated("b"))

      reset_probes()

      board_update(
        list(blocks = list(add = blocks(d = with_id(probe_source(), "d"))))
      )
      session$flushReact()

      expect_false(evaluated("a"))
      expect_false(evaluated("b"))
      expect_false(evaluated("d"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b"))
        NULL
      }
    )
  )
})

test_that("a variadic block receives its inputs as values, not reactives", {

  reset_probes()
  probe_args$entry_classes <- NULL

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_source(), "b"),
      c = with_id(probe_variadic(), "c")
    ),
    links = links(new_link("a", "c", "1"), new_link("b", "c", "2"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_length(probe_args$entry_classes, 2)
      expect_setequal(probe_args$entry_classes, "data.frame")
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("c"))
        NULL
      }
    )
  )
})

test_that("an off-screen variadic block does not pull its inputs", {

  reset_probes()

  board <- new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_source(), "b"),
      c = with_id(probe_variadic(), "c"),
      e = with_id(probe_source(), "e")
    ),
    links = links(new_link("a", "c", "1"), new_link("b", "c", "2"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_true(evaluated("e"))

      expect_false(evaluated("a"))
      expect_false(evaluated("b"))
      expect_false(evaluated("c"))
    },
    args = list(
      x = board,
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("e"))
        NULL
      }
    )
  )
})

ordered_board <- function() {
  new_board(
    blocks = c(
      a = with_id(probe_source(), "a"),
      b = with_id(probe_passthrough(), "b"),
      c = with_id(probe_passthrough(), "c"),
      d = with_id(probe_passthrough(), "d")
    ),
    links = links(
      new_link(from = "a", to = "b"),
      new_link(from = "b", to = "c"),
      new_link(from = "a", to = "d")
    )
  )
}

visible_b <- function(visible, ...) {
  visible(vis_map("b"))
  NULL
}

test_that("only the needed blocks are constructed before first paint", {

  reset_probes()

  testServer(
    get_s3_method("board_server", ordered_board()),
    {
      session$flushReact()

      expect_true(constructed("a"))
      expect_true(constructed("b"))

      expect_false(constructed("c"))
      expect_false(constructed("d"))

      session$elapse(5000)
      session$flushReact()

      expect_true(constructed("c"))
      expect_true(constructed("d"))
    },
    args = list(x = ordered_board(), plugins = list(), callbacks = visible_b)
  )
})

test_that("opening a view constructs its blocks without the background", {

  reset_probes()

  testServer(
    get_s3_method("board_server", ordered_board()),
    {
      session$flushReact()

      expect_false(constructed("c"))

      board_visible(vis_map("b", "c"))
      session$flushReact()

      expect_true(constructed("c"))
      expect_false(constructed("d"))
    },
    args = list(x = ordered_board(), plugins = list(), callbacks = visible_b)
  )
})

test_that("the background constructs every block exactly once", {

  reset_probes()

  testServer(
    get_s3_method("board_server", ordered_board()),
    {
      session$flushReact()
      session$elapse(5000)
      session$flushReact()

      built <- probe_construct$ids

      expect_setequal(built, c("a", "b", "c", "d"))
      expect_length(built, 4L)

      session$elapse(5000)
      session$flushReact()

      expect_identical(probe_construct$ids, built)
    },
    args = list(x = ordered_board(), plugins = list(), callbacks = visible_b)
  )
})

test_that("on_screen_blocks is the union of the required and rendered ids", {

  expect_setequal(
    on_screen_blocks(c(a = "rendered", b = "required", z = "hidden")),
    c("a", "b")
  )

  expect_length(on_screen_blocks(c(z = "hidden")), 0L)
  expect_length(on_screen_blocks(character()), 0L)
})

test_that("visible_set_rendered is TRUE only when all on-screen rendered", {

  expect_true(visible_set_rendered(c(a = "rendered", b = "rendered")))

  expect_false(visible_set_rendered(c(a = "rendered", b = "required")))
  expect_false(visible_set_rendered(c(b = "required")))

  expect_true(visible_set_rendered(c(a = "rendered", z = "hidden")))

  expect_true(visible_set_rendered(character()))
  expect_true(visible_set_rendered(c(z = "hidden")))
})

test_that("the background waits for the front-end's rendered report", {

  reset_probes()

  testServer(
    get_s3_method("board_server", ordered_board()),
    {
      session$flushReact()

      expect_true(constructed("a"))
      expect_true(constructed("b"))

      session$elapse(5000)
      session$flushReact()

      expect_false(constructed("c"))
      expect_false(constructed("d"))

      board_visible(vis_map("b"))
      session$flushReact()
      session$elapse(5000)
      session$flushReact()

      expect_true(constructed("c"))
      expect_true(constructed("d"))
    },
    args = list(
      x = ordered_board(),
      plugins = list(),
      callbacks = function(visible, ...) {
        visible(vis_map("b", level = "required"))
        NULL
      }
    )
  )
})

test_that("a zero background delay builds every block up front", {

  reset_probes()

  withr::local_options(blockr.background_construction_delay = 0)

  testServer(
    get_s3_method("board_server", ordered_board()),
    {
      session$flushReact()

      expect_true(constructed("a"))
      expect_true(constructed("b"))
      expect_true(constructed("c"))
      expect_true(constructed("d"))
    },
    args = list(x = ordered_board(), plugins = list(), callbacks = visible_b)
  )
})

test_that("board server", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(from = "a", to = "b")
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()
      expect_equal(rv$blocks$b$server$result(), iris)
    },
    args = list(x = board)
  )

  empty <- new_board()

  testServer(
    get_s3_method("board_server", empty),
    {
      expect_length(board_blocks(x), 0L)
      expect_length(board_blocks(rv$board), 0L)

      expect_null(board_update())

      board_update(
        list(
          blocks = list(add = as_blocks(new_dataset_block()))
        )
      )

      session$flushReact()

      expect_length(board_blocks(x), 0L)
      expect_length(board_blocks(rv$board), 1L)

      expect_null(board_update())

      board_update(
        list(
          blocks = list(add = as_blocks(new_subset_block()))
        )
      )

      session$flushReact()

      expect_length(board_blocks(x), 0L)
      expect_length(board_blocks(rv$board), 2L)

      expect_null(board_update())

      board_update(
        list(
          blocks = list(rm = board_block_ids(rv$board))
        )
      )

      session$flushReact()

      expect_length(board_blocks(x), 0L)
      expect_length(board_blocks(rv$board), 0L)

      expect_null(board_update())
    },
    args = list(
      x = empty,
      plugins = list(manage_blocks())
    )
  )

  testServer(
    get_s3_method("board_server", empty),
    {
      expect_length(board_blocks(x), 0L)
    },
    args = list(
      x = empty,
      callbacks = function(board, ...) {
        expect_length(board_blocks(board$board), 0L)
        NULL
      }
    )
  )

  test_xtra_args <- function(id, board, plugin_a, plugin_b) {
    moduleServer(
      id,
      function(input, output, session) {

        parent <- reactiveVal()

        board_server(
          "board",
          board,
          list(
            preserve_board(server = plugin_a, ui = NULL),
            manage_blocks(server = plugin_b, ui = NULL)
          ),
          parent = parent
        )
      }
    )
  }

  testServer(
    test_xtra_args,
    session$flushReact(),
    args = list(
      board = empty,
      plugin_a = function(id, board, update, parent) {
        moduleServer(
          id,
          function(input, output, session) {
            observeEvent(TRUE, parent(1L), once = TRUE)
            reactiveVal()
          }
        )
      },
      plugin_b = function(id, board, update, parent) {
        moduleServer(
          id,
          function(input, output, session) {
            observeEvent(
              parent(),
              expect_identical(parent(), 1L)
            )
            NULL
          }
        )
      }
    )
  )

  testServer(
    get_s3_method("board_server", empty),
    {
      session$flushReact()
      rv$abc <- 1L
    },
    args = list(
      x = empty,
      plugins = list(
        preserve_board(
          function(id, board, ...) {
            moduleServer(
              id,
              function(input, output, session) {
                observeEvent(
                  board$abc,
                  expect_identical(board$abc, 1L)
                )
                reactiveVal()
              }
            )
          },
          NULL
        )
      )
    )
  )

  testServer(
    function(id, board, plugin) {
      moduleServer(
        id,
        function(input, output, session) {
          board_server("board", board, preserve_board(plugin, NULL))
        }
      )
    },
    session$flushReact(),
    args = list(
      board = empty,
      plugin = function(id, board, ...) {
        moduleServer(
          id,
          function(input, output, session) {
            observeEvent(
              TRUE,
              {
                expect_error(
                  {
                    board$abc <- 1
                  }
                )
              },
              once = TRUE
            )
            reactiveVal()
          }
        )
      }
    )
  )
})

test_that("blocks$mod with ctrl-able delta writes reactiveVals in place", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    links = links(from = "a", to = "b")
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$blocks$b$server$result(), datasets::iris)

      pre_server <- rv$blocks$b$server

      board_update(
        list(
          blocks = list(
            mod = list(b = list(subset = "Species == \"setosa\""))
          )
        )
      )

      session$flushReact()

      expect_identical(rv$blocks$b$server, pre_server)
      expect_identical(
        rv$blocks$b$server$result(),
        subset(datasets::iris, Species == "setosa")
      )
    },
    args = list(x = board)
  )
})

test_that("blocks$mod routes block_name without disturbing the server", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_head_block())
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      pre_server <- rv$blocks$b$server
      pre_result <- rv$blocks$b$server$result()

      board_update(
        list(blocks = list(mod = list(b = list(block_name = "Renamed"))))
      )

      session$flushReact()

      expect_identical(rv$blocks$b$server, pre_server)
      expect_identical(rv$blocks$b$server$result(), pre_result)
      expect_identical(block_name(board_blocks(rv$board)$b), "Renamed")
    },
    args = list(x = board)
  )
})

test_that("blocks$mod with both ctrl-arg and block_name delta applies both", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    links = links(from = "a", to = "b")
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      pre_server <- rv$blocks$b$server

      board_update(
        list(
          blocks = list(
            mod = list(
              b = list(
                subset = "Species == \"setosa\"",
                block_name = "Setosa only"
              )
            )
          )
        )
      )

      session$flushReact()

      expect_identical(rv$blocks$b$server, pre_server)
      expect_identical(
        rv$blocks$b$server$result(),
        subset(datasets::iris, Species == "setosa")
      )
      expect_identical(block_name(board_blocks(rv$board)$b), "Setosa only")
    },
    args = list(x = board)
  )
})

test_that("blocks$mod rejects non-ctrl-able arguments", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_head_block(n = 6L)),
    links = links(from = "a", to = "b")
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(b = list(n = 3L)))),
      board
    ),
    class = "board_update_blocks_mod_not_ctrl"
  )
})

test_that("ctrl plugin sees block_name as a routable var", {

  blk <- new_dataset_block("iris")
  upd <- reactiveVal()

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      session$makeScope("ctrl_block")$setInputs(
        block_name = "NewName",
        submit = 1L
      )

      session$flushReact()

      expect_identical(
        upd(),
        list(
          blocks = list(
            mod = list(blk = list(block_name = "NewName"))
          )
        )
      )
    },
    args = list(
      x = blk,
      ctrl_block = ctrl_block(),
      board = reactiveValues(
        board = new_board(blocks = c(blk = new_dataset_block("iris")))
      ),
      block_id = "blk",
      update = upd
    )
  )
})

test_that("a delta-set block_name survives serialize + deserialize", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block())
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      board_update(
        list(blocks = list(mod = list(b = list(block_name = "Persisted"))))
      )

      session$flushReact()

      state <- lapply(
        blockr.core:::lst_xtr(rv$blocks, "server", "state"),
        lapply,
        reval_if
      )
      ser <- blockr_ser(rv$board, blocks = state)
      restored <- blockr_deser(ser)

      expect_identical(
        block_name(board_blocks(restored)$b),
        "Persisted"
      )
    },
    args = list(x = board)
  )
})

test_that("block_name attr and ctrl reactiveVal stay in sync without loops", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block())
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      n_updates <- 0L

      observe({
        upd <- board_update()
        if (!is.null(upd) && !is.null(upd$blocks$mod$b$block_name)) {
          n_updates <<- n_updates + 1L
        }
      })

      session$flushReact()

      board_update(
        list(blocks = list(mod = list(b = list(block_name = "FromBoard"))))
      )

      session$flushReact()

      expect_identical(block_name(board_blocks(rv$board)$b), "FromBoard")

      pre <- n_updates
      session$flushReact()
      expect_identical(n_updates, pre)
    },
    args = list(x = board)
  )
})

test_that("links$mod accepts partial-args deltas via update_link", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    links = links(ab = list(from = "a", to = "b", input = "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      board_update(
        list(links = list(mod = list(ab = list(input = "data"))))
      )

      session$flushReact()

      expect_identical(board_links(rv$board)$input, "data")
    },
    args = list(x = board)
  )
})

test_that("stacks$mod merges deltas onto current stack via update_stack", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    stacks = stacks(s1 = new_stack(c("a"), "Initial"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      board_update(
        list(stacks = list(mod = list(s1 = list(blocks = c("a", "b")))))
      )

      session$flushReact()

      stk <- board_stacks(rv$board)$s1
      expect_setequal(stack_blocks(stk), c("a", "b"))
      expect_identical(stack_name(stk), "Initial")

      board_update(
        list(stacks = list(mod = list(s1 = list(name = "Renamed"))))
      )

      session$flushReact()

      stk <- board_stacks(rv$board)$s1
      expect_setequal(stack_blocks(stk), c("a", "b"))
      expect_identical(stack_name(stk), "Renamed")
    },
    args = list(x = board)
  )
})

test_that("update_stack default preserves extra attrs through reconstruction", {

  stk <- new_stack(
    c("a", "b"),
    name = "S",
    color = "red",
    pkg = "blockr.core"
  )

  out <- update_stack(stk, list(blocks = "c"))
  expect_identical(stack_blocks(out), "c")
  expect_identical(stack_name(out), "S")
  expect_identical(attr(out, "color"), "red")

  out2 <- update_stack(stk, list(color = "blue"))
  expect_setequal(stack_blocks(out2), c("a", "b"))
  expect_identical(attr(out2, "color"), "blue")
})

test_that("update_link default preserves all fields through reconstruction", {

  lnk <- new_link("a", "b", "data")

  out <- update_link(lnk, list(input = "x"))
  expect_identical(out$from, "a")
  expect_identical(out$to, "b")
  expect_identical(out$input, "x")

  out2 <- update_link(lnk, list(to = "c", input = "y"))
  expect_identical(out2$from, "a")
  expect_identical(out2$to, "c")
  expect_identical(out2$input, "y")
})

test_that("removing a variadic link drops its ...args slot", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      v = new_rbind_block()
    ),
    links = links(av = new_link("a", "v"), bv = new_link("b", "v"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      args <- rv$inputs$v[["...args"]]
      expect_identical(isolate(length(args)), 2L)

      board_update(list(links = list(rm = "av")))
      session$flushReact()

      expect_identical(isolate(length(args)), 1L)
      expect_identical(rv$blocks$v$server$result(), datasets::BOD)
    },
    args = list(x = board)
  )
})

test_that("variadic ...args slots each bind to their own upstream", {

  board <- new_board(
    blocks = c(
      a = new_static_block(iris[1:3, ]),
      b = new_static_block(iris[4:6, ]),
      v = new_rbind_block()
    ),
    links = links(
      av = new_link("a", "v", "x"),
      bv = new_link("b", "v", "y")
    )
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(
        rv$blocks$v$server$result(),
        rbind(x = iris[1:3, ], y = iris[4:6, ])
      )
    },
    args = list(x = board)
  )
})

test_that("update validation", {

  expect_error(
    validate_board_update("a", new_board()),
    class = "board_update_type_invalid"
  )

  expect_silent(
    validate_board_update(list(test = list(add = "a")), new_board())
  )

  expect_error(
    validate_board_update(
      list(
        blocks = list(
          add = blocks(a = new_dataset_block()),
          mod = list(a = list(dataset = "iris"))
        )
      ),
      new_board()
    ),
    class = "board_update_blocks_add_mod_clash"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(a = "not a list"))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_entry_invalid"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(b = list(dataset = "iris")))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_unknown_id"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(a = list()))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_entry_empty"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(a = list(no_such_arg = 1)))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_not_ctrl"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(a = list(package = "datasets")))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_not_ctrl"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = "xyz")),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_mod_component_invalid"
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = blocks(a = new_dataset_block()))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_blocks_mod_entry_invalid"
  )

  expect_error(
    validate_board_update(
      list(
        links = list(
          add = links(ab = list(from = "a", to = "b", input = "data")),
          mod = links(ab = list(from = "a", to = "b", input = "data"))
        )
      ),
      new_board(blocks(a = new_dataset_block(), b = new_head_block()))
    ),
    class = "board_update_links_add_mod_clash"
  )

  expect_error(
    validate_board_update(
      list(links = list(mod = list(ab = "xyz"))),
      new_board(
        blocks(a = new_dataset_block(), b = new_head_block()),
        links(ab = list(from = "a", to = "b", input = "data"))
      )
    ),
    class = "board_update_links_mod_entry_invalid"
  )

  expect_error(
    validate_board_update(
      list(
        links = list(
          mod = links(ab = list(from = "a", to = "b", input = "data"))
        )
      ),
      new_board(
        blocks(a = new_dataset_block(), b = new_head_block()),
        links(ab = list(from = "a", to = "b", input = "data"))
      )
    ),
    class = "board_update_links_mod_entry_invalid"
  )

  expect_error(
    validate_board_update(
      list(
        links = list(
          add = links(xyz = list(from = "a", to = "c", input = "data"))
        )
      ),
      new_board(blocks(a = new_dataset_block(), b = new_head_block()))
    ),
    class = "board_block_link_name_mismatch"
  )

  expect_error(
    validate_board_update(
      list(links = list(mod = list(xyz = list(to = "c")))),
      new_board(
        blocks(a = new_dataset_block(), b = new_head_block()),
        links(xyz = list(from = "a", to = "b", input = "data"))
      )
    ),
    class = "board_block_link_name_mismatch"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(add = stacks(a = "a"), mod = stacks(a = "a"))),
      new_board(blocks(a = new_dataset_block()))
    ),
    class = "board_update_stacks_add_mod_clash"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(mod = list(a = "a"))),
      new_board(
        blocks(a = new_dataset_block()),
        stacks = stacks(a = "a")
      )
    ),
    class = "board_update_stacks_mod_entry_invalid"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(mod = list(b = list(blocks = "a")))),
      new_board(
        blocks(a = new_dataset_block()),
        stacks = stacks(a = "a")
      )
    ),
    class = "board_update_stacks_mod_unknown_id"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(add = stacks(b = "b"))),
      new_board(
        blocks(a = new_dataset_block()),
        stacks = stacks(a = "a")
      )
    ),
    class = "board_block_stack_name_mismatch"
  )

  expect_error(
    validate_board_update(
      list(stacks = list(mod = list(a = list(blocks = "b")))),
      new_board(
        blocks(a = new_dataset_block()),
        stacks = stacks(a = "a")
      )
    ),
    class = "board_block_stack_name_mismatch"
  )
})

test_that("public validate_board_update", {

  brd <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    links = links(ab = new_link(from = "a", to = "b"))
  )

  expect_identical(
    validate_board_update(list(), brd),
    list()
  )

  pay <- list(links = list(rm = "ab"))
  expect_identical(
    validate_board_update(pay, brd),
    pay
  )

  expect_error(
    validate_board_update("not-a-list", brd),
    class = "board_update_type_invalid"
  )

  expect_error(
    validate_board_update(
      list(
        links = list(
          add = links(xy = new_link(from = "x", to = "y", input = "data"))
        )
      ),
      brd
    ),
    class = "board_block_link_name_mismatch"
  )
})

test_that("validate_board_update handles block-rm orphan cleanup (#177)", {

  brd <- new_board(
    blocks = c(my_data = new_dataset_block("iris"), preview = new_head_block()),
    links = c(lnk1 = new_link("my_data", "preview", "data"))
  )

  payload <- list(
    blocks = list(add = c(preview_new = new_head_block()), rm = "preview"),
    links = list(add = c(lnk_new = new_link("my_data", "preview_new", "data")))
  )

  expect_identical(
    validate_board_update(payload, brd),
    payload
  )

  payload_mod <- list(
    blocks = list(rm = "preview"),
    links = list(mod = list(lnk1 = list(input = "data")))
  )

  expect_error(
    validate_board_update(payload_mod, brd),
    class = "board_block_link_name_mismatch"
  )
})

test_that("board server utils", {

  expect_identical(
    bs_theme_colors(MockShinySession$new()),
    list(bg = "auto", fg = "auto", accent = "auto")
  )

  expect_named(
    withr::with_options(
      list(bslib_theme = bslib::bs_theme()),
      bs_theme_colors(MockShinySession$new())
    ),
    c("bg", "fg", "accent")
  )
})
register_ext_method <- function(generic, method) {

  registerS3method(
    generic, "test_ext_board", method,
    envir = asNamespace("blockr.core")
  )
}

reset_ext_methods <- function() {

  register_ext_method(
    "validate_board_update",
    function(payload, board) NextMethod()
  )

  register_ext_method(
    "augment_board_update",
    function(upd, board, ...) NextMethod()
  )

  register_ext_method(
    "apply_board_update",
    function(board, upd, ...) NextMethod()
  )
}

test_that("augment_board_update dispatches and chains via NextMethod", {

  reset_ext_methods()

  register_ext_method(
    "augment_board_update",
    function(upd, board) {
      upd <- NextMethod()
      upd$ext$augmented <- TRUE
      upd
    }
  )

  ext_board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
    links = links(ab = new_link("a", "b")),
    class = "test_ext_board"
  )

  out <- augment_board_update(
    list(
      blocks = list(rm = "a"),
      ext = list(views = list(grid = "two-up"))
    ),
    ext_board
  )

  expect_identical(out$ext$augmented, TRUE)
  expect_identical(out$ext$views$grid, "two-up")
  expect_identical(out$links$rm, "ab")
})

test_that("subclass errors during augment_board_update propagate", {

  reset_ext_methods()

  register_ext_method(
    "augment_board_update",
    function(upd, board) stop("subclass augment failure")
  )

  ext_board <- new_board(class = "test_ext_board")

  expect_error(
    augment_board_update(list(blocks = list(rm = "x")), ext_board),
    "subclass augment failure"
  )
})

test_that("validate_board_update dispatches and chains via NextMethod", {

  reset_ext_methods()

  register_ext_method(
    "validate_board_update",
    function(payload, board) {
      NextMethod()
      if (length(payload$ext$views)) {
        stopifnot(is.character(payload$ext$views))
      }
      invisible(payload)
    }
  )

  ext_board <- new_board(class = "test_ext_board")

  expect_silent(
    validate_board_update(
      list(ext = list(views = "two-up")),
      ext_board
    )
  )

  expect_error(
    validate_board_update(
      list(ext = list(views = 42L)),
      ext_board
    )
  )

  expect_error(
    validate_board_update(
      list(blocks = list(mod = list(a = "not a list"))),
      ext_board
    ),
    class = "board_update_blocks_mod_unknown_id"
  )
})

test_that("board_update lifecycle runs augment before apply and resets", {

  reset_ext_methods()

  call_log <- character()

  register_ext_method(
    "augment_board_update",
    function(upd, board) {
      call_log <<- c(call_log, "augment")
      upd <- NextMethod()
      upd$ext$augmented <- TRUE
      upd
    }
  )

  register_ext_method(
    "apply_board_update",
    function(board, upd, ...) {
      call_log <<- c(call_log, "apply")
      board <- NextMethod()
      attr(board, "ext_seen") <- upd$ext
      board
    }
  )

  ext_board <- new_board(class = "test_ext_board")

  testServer(
    get_s3_method("board_server", ext_board),
    {
      session$flushReact()

      board_update(
        list(
          blocks = list(add = as_blocks(new_dataset_block())),
          ext = list(views = list(grid = "two-up"))
        )
      )

      session$flushReact()

      expect_true(any(call_log == "augment"))
      expect_true(any(call_log == "apply"))
      expect_lt(
        max(which(call_log == "augment")),
        min(which(call_log == "apply"))
      )

      ext_seen <- attr(rv$board, "ext_seen")
      expect_identical(ext_seen$augmented, TRUE)
      expect_identical(ext_seen$views$grid, "two-up")

      expect_length(board_blocks(rv$board), 1L)

      expect_null(board_update())
    },
    args = list(
      x = ext_board,
      plugins = list(manage_blocks())
    )
  )
})

test_that("apply_board_update splices board_server `...` as named args", {

  reset_ext_methods()

  seen <- NULL

  register_ext_method(
    "apply_board_update",
    function(board, upd, ...) {
      seen <<- list(...)
      NextMethod()
    }
  )

  ext_board <- new_board(class = "test_ext_board")

  testServer(
    get_s3_method("board_server", ext_board),
    {
      session$flushReact()
      board_update(list(blocks = list(add = as_blocks(new_dataset_block()))))
      session$flushReact()

      expect_identical(seen$passthrough, "carry me")
      expect_named(seen, c("passthrough", "session"))
    },
    args = list(
      x = ext_board,
      plugins = list(manage_blocks()),
      passthrough = "carry me"
    )
  )
})

test_that("apply_board_update.board returns the supplied board unchanged", {

  brd <- new_board()
  expect_identical(apply_board_update(brd, list()), brd)
})

test_that("apply_board_update runs after core has settled rv state", {

  reset_ext_methods()

  observed <- NULL

  register_ext_method(
    "apply_board_update",
    function(board, upd, ...) {
      observed <<- board_block_ids(board)
      NextMethod()
    }
  )

  ext_board <- new_board(class = "test_ext_board")

  testServer(
    get_s3_method("board_server", ext_board),
    {
      session$flushReact()
      board_update(
        list(blocks = list(add = as_blocks(c(a = new_dataset_block()))))
      )
      session$flushReact()

      expect_identical(observed, "a")
    },
    args = list(
      x = ext_board,
      plugins = list(manage_blocks())
    )
  )
})

test_that("successful board update records ok outcome with monotonic seq", {

  empty <- new_board()

  testServer(
    get_s3_method("board_server", empty),
    {
      session$flushReact()

      expect_null(rv$last_update)

      board_update(
        list(blocks = list(add = as_blocks(c(a = new_dataset_block()))))
      )

      session$flushReact()

      expect_length(board_blocks(rv$board), 1L)
      expect_true(rv$last_update$ok)
      expect_identical(rv$last_update$phase, "apply")
      expect_identical(rv$last_update$seq, 1L)
      expect_identical(rv$last_update$message, NA_character_)

      expect_identical(make_read_only(rv)$last_update, rv$last_update)

      board_update(
        list(blocks = list(add = as_blocks(c(b = new_subset_block()))))
      )

      session$flushReact()

      expect_length(board_blocks(rv$board), 2L)
      expect_true(rv$last_update$ok)
      expect_identical(rv$last_update$seq, 2L)
    },
    args = list(
      x = empty,
      plugins = list(manage_blocks())
    )
  )
})

test_that("rejected board update records validate-phase failure each time", {

  empty <- new_board()

  testServer(
    get_s3_method("board_server", empty),
    {
      session$flushReact()

      board_update(list(blocks = list(rm = "missing")))

      session$flushReact()

      expect_false(rv$last_update$ok)
      expect_identical(rv$last_update$phase, "validate")
      expect_true(nzchar(rv$last_update$message))
      expect_null(board_update())
      expect_length(board_blocks(rv$board), 0L)

      first_seq <- rv$last_update$seq

      board_update(list(blocks = list(rm = "missing")))

      session$flushReact()

      expect_false(rv$last_update$ok)
      expect_identical(rv$last_update$phase, "validate")
      expect_identical(rv$last_update$seq, first_seq + 1L)
    },
    args = list(
      x = empty,
      plugins = list(manage_blocks())
    )
  )
})

test_that("apply-phase failure records apply outcome", {

  local_mocked_bindings(
    apply_board_update = function(board, upd, ...) {
      stop("apply boom")
    }
  )

  empty <- new_board()

  testServer(
    get_s3_method("board_server", empty),
    {
      session$flushReact()

      board_update(list())

      session$flushReact()

      expect_false(rv$last_update$ok)
      expect_identical(rv$last_update$phase, "apply")
      expect_match(rv$last_update$message, "apply boom")
    },
    args = list(
      x = empty,
      plugins = list(manage_blocks())
    )
  )
})

boom_block <- function() {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          list(expr = reactive(quote(stop("boom"))), state = list())
        }
      )
    },
    function(id) tagList(),
    class = "boom_block",
    block_metadata = list()
  )
}

test_that("board combines per-block conditions and exposes them per block", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = boom_block()),
    links = links(ab = new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(make_read_only(rv)$conditions, rv$conditions)

      # block b evaluates against ready data but its expression raises, so it
      # is `failed` with an eval-phase error; a stays healthy
      expect_identical(nrow(rv$blocks$a$server$conditions()), 0L)
      expect_identical(rv$eval$b(), "failed")

      b_cond <- rv$blocks$b$server$conditions()

      expect_identical(nrow(b_cond), 1L)
      expect_identical(b_cond$block, "b")
      expect_identical(b_cond$phase, "eval")
      expect_identical(b_cond$severity, "error")

      combined <- rv$conditions()

      expect_named(combined, c("block", "phase", "severity", "message", "id"))
      expect_identical(nrow(combined), 1L)
      expect_identical(combined$block, "b")
    },
    args = list(x = board, plugins = list(manage_blocks()))
  )
})

test_that("board drops conditions of removed blocks", {

  board <- new_board(
    blocks = c(a = new_dataset_block("iris"), b = boom_block()),
    links = links(ab = new_link("a", "b", "data"))
  )

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      expect_identical(rv$eval$b(), "failed")
      expect_identical(nrow(rv$conditions()), 1L)

      board_update(list(blocks = list(rm = "b")))

      session$flushReact()

      expect_identical(nrow(rv$conditions()), 0L)
    },
    args = list(x = board, plugins = list(manage_blocks()))
  )
})

#' Block metadata
#'
#' Registry metadata for blocks is available both as a tabular overview and via
#' per-attribute accessors. `block_metadata()` returns a `data.frame` with one
#' row per block -- dispatching on a `block`, a `blocks` collection, a
#' `block_registry_entry` or a registry ID -- where scalar attributes are
#' atomic columns and the multi-valued ones (`arguments`, `examples`,
#' `keywords`) are list-columns. The `fields` argument selects a subset of
#' columns. Each attribute additionally has a dedicated getter
#' (`block_meta_name()`, `block_meta_guidance()`, ...) returning that attribute
#' for a single block. Missing fields are filled with display defaults in the
#' data.frame; the getters instead return the stored value (or `NA` / an empty
#' value).
#'
#' @param x A `block`, a `blocks` collection, a `block_registry_entry` or a
#'   registry ID
#' @param fields Metadata fields to include (defaults to `"all"`)
#' @param ... Generic consistency, passed on to methods
#'
#' @return `block_metadata()` returns a `data.frame`. The `block_meta_*()`
#'   getters return the named attribute: a string (or `NA`) for scalar fields, a
#'   character vector for `block_meta_keywords()`, a `block_args` object for
#'   `block_meta_arguments()`, and a list of worked configurations for
#'   `block_meta_examples()`.
#'
#' @rdname block_metadata
#' @export
block_metadata <- function(x, fields = "all", ...) {
  UseMethod("block_metadata")
}

#' @export
block_metadata.character <- function(x, fields = "all", ...) {
  build_block_catalog(
    lapply(x, block_metadata_record),
    fields,
    rep(NA_character_, length(x))
  )
}

#' @export
block_metadata.default <- function(x, fields = "all", ...) {

  blks <- as_blocks(x)

  build_block_catalog(
    lapply(blks, block_metadata_record),
    fields,
    chr_ply(blks, block_default_name),
    names(blks)
  )
}

block_default_name <- function(x) {
  gsub("_", " ", class(x)[1L])
}

block_metadata_fields <- function() {
  c(
    "id", "name", "description", "details", "link", "guidance", "keywords",
    "category", "icon", "arguments", "examples", "package"
  )
}

build_block_catalog <- function(records, fields, default_names,
                                row_names = NULL) {

  if (identical(fields, "all")) {
    fields <- block_metadata_fields()
  }

  fields <- match.arg(fields, block_metadata_fields(), several.ok = TRUE)

  cols <- lapply(
    set_names(nm = fields),
    catalog_column,
    records = records,
    default_names = default_names
  )

  res <- list2DF(cols)

  if (not_null(row_names) && length(row_names) == nrow(res)) {
    rownames(res) <- row_names
  }

  res
}

catalog_column <- function(field, records, default_names) {

  switch(
    field,
    id = scalar_column(records, "id", NA_character_),
    name = name_column(records, default_names),
    description = scalar_column(
      records, "description", "No description available."
    ),
    details = scalar_column(records, "details", NA_character_),
    link = scalar_column(records, "link", NA_character_),
    guidance = scalar_column(records, "guidance", NA_character_),
    category = scalar_column(records, "category", default_category()),
    icon = icon_column(records),
    package = scalar_column(records, "package", "local"),
    keywords = lapply(records, meta_default, "keywords", character()),
    arguments = lapply(records, meta_default, "arguments", new_block_args()),
    examples = lapply(records, record_examples)
  )
}

scalar_column <- function(records, field, default) {
  chr_ply(records, function(r) coal(r[[field]], default, fail_all = FALSE))
}

name_column <- function(records, default_names) {
  chr_mply(
    function(r, d) coal(r[["name"]], d, fail_all = FALSE),
    records,
    default_names
  )
}

icon_column <- function(records) {
  chr_ply(
    records,
    function(r) {
      category <- coal(r[["category"]], default_category())
      coal(r[["icon"]], default_icon(category), fail_all = FALSE)
    }
  )
}

meta_default <- function(record, field, default) {
  coal(record[[field]], default, fail_all = FALSE)
}

record_examples <- function(record) {
  block_examples_list(
    coal(record[["arguments"]], new_block_args(), fail_all = FALSE),
    coal(record[["examples"]], list(), fail_all = FALSE)
  )
}

block_record_fields <- function() {
  c(registry_metadata_fields, "examples")
}

block_metadata_record <- function(x, ...) {
  UseMethod("block_metadata_record")
}

#' @export
block_metadata_record.block <- function(x, ...) {

  record <- attr(x, "block_metadata")

  if (!is.list(record)) {
    blockr_abort(
      "Block {class(x)[1L]} carries no metadata.",
      class = "missing_block_metadata"
    )
  }

  record
}

#' @export
block_metadata_record.block_registry_entry <- function(x, ...) {
  c(
    list(id = registry_uid(attr(x, "classes"))),
    lapply(set_names(nm = block_record_fields()), get_attr, x)
  )
}

#' @export
block_metadata_record.character <- function(x, ...) {

  record <- block_metadata_record(get_registry_entry(x))
  record[["id"]] <- x

  record
}

scalar_meta <- function(x, field, ...) {
  coal(block_metadata_record(x, ...)[[field]], NA_character_, fail_all = FALSE)
}

#' @rdname block_metadata
#' @export
block_meta_id <- function(x, ...) {
  scalar_meta(x, "id", ...)
}

#' @rdname block_metadata
#' @export
block_meta_name <- function(x, ...) {
  scalar_meta(x, "name", ...)
}

#' @rdname block_metadata
#' @export
block_meta_description <- function(x, ...) {
  scalar_meta(x, "description", ...)
}

#' @rdname block_metadata
#' @export
block_meta_details <- function(x, ...) {
  scalar_meta(x, "details", ...)
}

#' @rdname block_metadata
#' @export
block_meta_link <- function(x, ...) {
  scalar_meta(x, "link", ...)
}

#' @rdname block_metadata
#' @export
block_meta_guidance <- function(x, ...) {
  scalar_meta(x, "guidance", ...)
}

#' @rdname block_metadata
#' @export
block_meta_category <- function(x, ...) {
  scalar_meta(x, "category", ...)
}

#' @rdname block_metadata
#' @export
block_meta_icon <- function(x, ...) {
  scalar_meta(x, "icon", ...)
}

#' @rdname block_metadata
#' @export
block_meta_package <- function(x, ...) {
  scalar_meta(x, "package", ...)
}

#' @rdname block_metadata
#' @export
block_meta_keywords <- function(x, ...) {
  coal(block_metadata_record(x, ...)[["keywords"]], character(),
       fail_all = FALSE)
}

#' @rdname block_metadata
#' @export
block_meta_arguments <- function(x, ...) {
  coal(block_metadata_record(x, ...)[["arguments"]], new_block_args(),
       fail_all = FALSE)
}

#' @rdname block_metadata
#' @export
block_meta_examples <- function(x, ...) {
  record_examples(block_metadata_record(x, ...))
}

test_that("block metadata", {

  meta1 <- block_metadata(new_dataset_block())

  expect_s3_class(meta1, "data.frame")
  expect_identical(nrow(meta1), 1L)
  expect_identical(ncol(meta1), 12L)
  expect_named(
    meta1,
    c("id", "name", "description", "details", "link", "guidance", "keywords",
      "category", "icon", "arguments", "examples", "package")
  )

  new_identity_block <- function() {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(
              expr = reactive(quote(identity(data))),
              state = list()
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      block_metadata = list(),
      class = "identity_block"
    )
  }

  meta2 <- block_metadata(new_identity_block())

  expect_s3_class(meta2, "data.frame")
  expect_identical(nrow(meta2), 1L)
  expect_identical(ncol(meta2), 12L)
  expect_named(
    meta2,
    c("id", "name", "description", "details", "link", "guidance", "keywords",
      "category", "icon", "arguments", "examples", "package")
  )

  meta3 <- block_metadata(
    blocks(a = new_dataset_block(), b = new_identity_block())
  )

  expect_s3_class(meta3, "data.frame")
  expect_identical(nrow(meta3), 2L)
  expect_identical(ncol(meta3), 12L)
  expect_named(
    meta3,
    c("id", "name", "description", "details", "link", "guidance", "keywords",
      "category", "icon", "arguments", "examples", "package")
  )
  expect_identical(rownames(meta3), c("a", "b"))

  meta4 <- block_metadata(blocks())

  expect_s3_class(meta4, "data.frame")
  expect_identical(nrow(meta4), 0L)
  expect_identical(ncol(meta4), 12L)
  expect_named(
    meta4,
    c("id", "name", "description", "details", "link", "guidance", "keywords",
      "category", "icon", "arguments", "examples", "package")
  )

  expect_type(meta1$keywords, "list")
  expect_type(meta1$arguments, "list")
  expect_s3_class(meta1$arguments[[1L]], "block_args")

  sel <- block_metadata(new_dataset_block(), fields = c("name", "icon"))

  expect_named(sel, c("name", "icon"))
})

test_that("construction-metadata accessors dispatch on id, entry and block", {

  withr::defer(unregister_blocks("ut_head_acc"))

  register_block(
    new_head_block,
    name = "Head",
    description = "First or last rows.",
    uid = "ut_head_acc",
    arguments = new_block_args(
      n = new_block_arg("How many rows", example = 10L),
      direction = new_block_arg("head or tail", example = "tail")
    ),
    guidance = "Use direction='tail' for the bottom rows.",
    keywords = c("first", "last", "top")
  )

  expect_s3_class(block_meta_arguments("ut_head_acc"), "block_args")
  expect_identical(
    block_meta_examples("ut_head_acc"),
    list(list(n = 10L, direction = "tail"))
  )
  expect_identical(
    block_meta_guidance("ut_head_acc"),
    "Use direction='tail' for the bottom rows."
  )
  expect_identical(
    block_meta_keywords("ut_head_acc"),
    c("first", "last", "top")
  )

  entry <- available_blocks()[["ut_head_acc"]]
  expect_identical(block_meta_keywords(entry), c("first", "last", "top"))

  blk <- new_head_block()
  expect_identical(
    block_meta_arguments(blk),
    block_meta_arguments("head_block")
  )
  expect_identical(block_meta_guidance(blk), block_meta_guidance("head_block"))

  expect_identical(
    vapply(block_meta_arguments("ut_head_acc"), block_arg_description,
           character(1)),
    c(n = "How many rows", direction = "head or tail")
  )
})

test_that("construction accessors read the block's attached metadata", {

  withr::defer({
    unregister_blocks()
    register_core_blocks()
  })

  register_block(
    new_head_block,
    name = "Head",
    description = "d",
    uid = "head_block",
    overwrite = TRUE,
    arguments = new_block_args(n = new_block_arg("rows", example = 3L)),
    guidance = "be careful"
  )

  blk <- new_head_block()

  unregister_blocks("head_block")

  args <- block_meta_arguments(blk)

  expect_identical(block_arg_description(args[["n"]]), "rows")
  expect_identical(block_meta_guidance(blk), "be careful")
  expect_identical(block_meta_examples(blk), list(list(n = 3L)))
})

test_that("block metadata accessors fall back when an attribute is absent", {

  withr::defer(unregister_blocks("ut_head_empty"))

  register_block(
    new_head_block,
    name = "Head",
    description = "First or last rows.",
    uid = "ut_head_empty"
  )

  expect_identical(block_meta_keywords("ut_head_empty"), character())
  expect_identical(block_meta_examples("ut_head_empty"), list())
  expect_identical(block_meta_guidance("ut_head_empty"), NA_character_)
})

test_that("block_metadata tabulates catalog fields with list-columns", {

  withr::defer(unregister_blocks("ut_head_cat"))

  register_block(
    new_head_block,
    name = "Head",
    description = "First or last.",
    uid = "ut_head_cat",
    category = "transform",
    details = "Defaults to six rows.",
    link = "https://example.com/head"
  )

  meta <- block_metadata("ut_head_cat")

  expect_s3_class(meta, "data.frame")
  expect_identical(meta$name, "Head")
  expect_identical(meta$details, "Defaults to six rows.")
  expect_identical(meta$link, "https://example.com/head")

  expect_type(meta$arguments, "list")
  expect_type(meta$keywords, "list")
  expect_s3_class(meta$arguments[[1L]], "block_args")

  blk <- new_head_block()
  expect_identical(block_metadata(blk)$id, "head_block")
})

reactives <- function(...) {

  args <- list(...)

  keys <- names(args)

  if (is.null(keys)) {
    keys <- character(length(args))
  }

  unnamed <- !nzchar(keys)
  keys[unnamed] <- positional_key(which(unnamed))

  if (anyDuplicated(keys)) {
    blockr_abort(
      "Each `reactives()` entry needs a distinct name.",
      class = "reactives_key_collision"
    )
  }

  res <- structure(
    list(
      store = new.env(parent = emptyenv()),
      keys = reactiveVal(character())
    ),
    class = "reactives"
  )

  for (i in seq_along(args)) {
    set_reactive(res, keys[[i]], args[[i]])
  }

  res
}

# An unnamed slot still needs a store key; it gets a hidden synthetic one,
# prefixed with a control character so `names()` can blank it while a real
# name (even "1") survives.
positional_marker <- function() {
  intToUtf8(1L)
}

positional_key <- function(i) {
  paste0(positional_marker(), i)
}

is_positional_key <- function(keys) {
  startsWith(keys, positional_marker())
}

is_reactives <- function(x) {
  inherits(x, "reactives")
}

raw_keys <- function(x) {
  .subset2(x, "keys")()
}

display_names <- function(keys) {

  pos <- is_positional_key(keys)

  if (all(pos)) {
    return(NULL)
  }

  replace(keys, pos, "")
}

set_reactive <- function(x, key, fun) {

  store <- .subset2(x, "store")

  if (exists(key, store, inherits = FALSE)) {
    rm(list = key, envir = store)
  }

  makeActiveBinding(key, fun, store)

  keys <- .subset2(x, "keys")

  if (!key %in% isolate(keys())) {
    keys(c(isolate(keys()), key))
  }

  invisible(x)
}

drop_reactive <- function(x, key) {

  rm(list = key, envir = .subset2(x, "store"))

  keys <- .subset2(x, "keys")
  keys(setdiff(isolate(keys()), key))

  invisible(x)
}

append_reactive <- function(x, fun) {

  keys <- isolate(raw_keys(x))
  pos <- keys[is_positional_key(keys)]

  idx <- if (length(pos)) max(as.integer(substring(pos, 2L))) + 1L else 1L

  set_reactive(x, positional_key(idx), fun)
}

reorder_reactives <- function(x, order) {

  keys <- .subset2(x, "keys")
  cur <- isolate(keys())

  if (is.numeric(order)) {
    order <- cur[vec_as_location(order, length(cur))]
  }

  if (!setequal(order, cur) || anyDuplicated(order)) {
    blockr_abort(
      "`order` must list each slot exactly once; unnamed slots by position.",
      class = "reactives_bad_reorder"
    )
  }

  keys(order)

  invisible(x)
}

get_reactive <- function(x, key) {

  store <- .subset2(x, "store")

  if (is_string(key) && exists(key, store, inherits = FALSE)) {
    get(key, envir = store)
  } else {
    NULL
  }
}

#' @export
`$.reactives` <- function(x, name) {
  get_reactive(x, name)
}

#' @export
`[[.reactives` <- function(x, i) {

  if (is.numeric(i)) {
    i <- raw_keys(x)[[vec_as_location2(i, length(x))]]
  }

  get_reactive(x, i)
}

#' @export
names.reactives <- function(x) {
  display_names(raw_keys(x))
}

#' @export
length.reactives <- function(x) {
  length(raw_keys(x))
}

#' @export
as.list.reactives <- function(x, ...) {

  keys <- raw_keys(x)

  set_names(lapply(keys, get_reactive, x = x), display_names(keys))
}

#' @export
print.reactives <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
format.reactives <- function(x, ...) {
  paste0("<reactives[", length(isolate(raw_keys(x))), "]>")
}

assign_value <- function(x, key, value) {

  if (!key %in% isolate(raw_keys(x))) {
    blockr_abort(
      "Cannot assign to unknown slot {key}.",
      class = "reactives_unknown_slot"
    )
  }

  store <- .subset2(x, "store")
  store[[key]] <- value

  invisible(x)
}

#' @export
`$<-.reactives` <- function(x, name, value) {
  assign_value(x, name, value)
}

#' @export
`[[<-.reactives` <- function(x, i, value) {
  assign_value(x, if (is.numeric(i)) raw_keys(x)[[i]] else i, value)
}

#' @export
`[<-.reactives` <- function(x, i, value) {

  keys <- if (is.numeric(i)) raw_keys(x)[vec_as_location(i, length(x))] else i

  value <- if (is.list(value)) value else list(value)

  for (j in seq_along(keys)) {
    set_reactive(x, keys[[j]], value[[j]])
  }

  invisible(x)
}

#' @export
`[.reactives` <- function(x, i) {

  keys <- if (is.numeric(i)) raw_keys(x)[vec_as_location(i, length(x))] else i

  store <- .subset2(x, "store")

  res <- reactives()

  for (key in keys) {
    set_reactive(res, key, activeBindingFunction(key, store))
  }

  res
}

#' @export
`names<-.reactives` <- function(x, value) {

  keys <- raw_keys(x)

  stopifnot(length(value) == length(keys))

  unnamed <- !nzchar(value)
  value[unnamed] <- positional_key(which(unnamed))

  if (anyDuplicated(value)) {
    blockr_abort(
      "Each `reactives()` entry needs a distinct name.",
      class = "reactives_key_collision"
    )
  }

  store <- .subset2(x, "store")
  funs <- lapply(keys, activeBindingFunction, env = store)

  for (key in keys) {
    rm(list = key, envir = store)
  }

  for (j in seq_along(value)) {
    makeActiveBinding(value[[j]], funs[[j]], store)
  }

  .subset2(x, "keys")(value)

  invisible(x)
}

test_that("reactives grows, shrinks and reorders honestly", {

  with_mock_session(
    {
      coll <- reactives()

      # an observer holding the bag sees membership changes through the
      # shared store / keys, the way a downstream server context would
      seen <- NULL
      observe(seen <<- as.list(coll))
      session$flushReact()

      expect_identical(seen, list())
      expect_identical(isolate(length(coll)), 0L)

      set_reactive(coll, "a", function() "one")
      set_reactive(coll, "b", function() "two")
      session$flushReact()

      expect_identical(seen, list(a = "one", b = "two"))
      expect_identical(isolate(length(coll)), 2L)

      drop_reactive(coll, "a")
      session$flushReact()

      expect_identical(seen, list(b = "two"))
      expect_false(exists("a", .subset2(coll, "store"), inherits = FALSE))

      # the slot is genuinely free, so the same key can be re-added (appended)
      set_reactive(coll, "a", function() "one-again")
      session$flushReact()

      expect_identical(seen, list(b = "two", a = "one-again"))

      # all slots named, so reorder by name is well-defined
      reorder_reactives(coll, c("a", "b"))
      session$flushReact()

      expect_identical(seen, list(a = "one-again", b = "two"))

      expect_error(
        reorder_reactives(coll, c("a", "b", "c")),
        class = "reactives_bad_reorder"
      )
    }
  )
})

test_that("reorder by position works on a mix; by name needs all named", {

  with_mock_session(
    {
      coll <- reactives(a = function() "x", function() "y", b = function() "z")

      seen <- NULL
      observe(seen <<- as.list(coll))
      session$flushReact()

      expect_identical(seen, list(a = "x", "y", b = "z"))

      # an unnamed slot has no public handle, so it reorders by position
      reorder_reactives(coll, c(3, 1, 2))
      session$flushReact()

      expect_identical(seen, list(b = "z", a = "x", "y"))

      # by name can't address the unnamed slot -- labelled error, not silence
      expect_error(
        reorder_reactives(coll, c("a", "b")),
        class = "reactives_bad_reorder"
      )
    }
  )
})

test_that("a reactives slot computes on read and tracks its source", {

  with_mock_session(
    {
      src <- reactiveVal(1)

      coll <- reactives(a = function() src() * 10)

      seen <- NULL
      observe(seen <<- as.list(coll))
      session$flushReact()

      expect_identical(seen, list(a = 10))

      src(5)
      session$flushReact()

      expect_identical(seen, list(a = 50))
    }
  )
})

test_that("a slot can hold a reactive() that is evaluated and memoized", {

  with_mock_session(
    {
      runs <- new.env(parent = emptyenv())
      runs$n <- 0L

      v <- reactiveVal(1)
      r <- reactive({
        runs$n <- runs$n + 1L
        v() * 10
      })

      coll <- reactives(a = r)

      seen <- NULL
      observe(seen <<- c(coll[["a"]], coll[["a"]]))
      session$flushReact()

      expect_identical(seen, c(10, 10))
      expect_identical(runs$n, 1L)

      v(2)
      session$flushReact()

      expect_identical(seen, c(20, 20))
      expect_identical(runs$n, 2L)
    }
  )
})

test_that("inspecting a reactives never fires its bindings", {

  with_mock_session(
    {
      fired <- new.env(parent = emptyenv())
      fired$n <- 0L

      counting <- function(val) {
        function() {
          fired$n <- fired$n + 1L
          val
        }
      }

      coll <- reactives(a = counting(42), b = counting(99))

      expect_identical(format(coll), "<reactives[2]>")
      expect_output(print(coll), "<reactives[2]>", fixed = TRUE)

      expect_identical(fired$n, 0L)
    }
  )
})

test_that("reactives read accessors mirror reactiveValues", {

  with_mock_session(
    isolate(
      {
        coll <- reactives(x = function() 10, y = function() 20)

        expect_identical(coll[["x"]], 10)
        expect_identical(coll$y, 20)
        expect_identical(coll[[1L]], 10)
        expect_identical(coll[[2L]], 20)
        expect_null(coll[["missing"]])
        expect_null(coll$nope)
        expect_error(coll[[3L]])
        expect_identical(names(coll), c("x", "y"))
        expect_identical(length(coll), 2L)
      }
    )
  )
})

test_that("as.list(reactives) is the list(...) it represents", {

  # named and unnamed args splice in declaration order: unnamed slots stay
  # anonymous, named slots keep their name
  k <- function(v) function() v

  with_mock_session(
    isolate(
      {
        expect_identical(
          as.list(reactives(k("x"), a = k("y"), k("z"))),
          list("x", a = "y", "z")
        )
        expect_identical(
          as.list(reactives(a = k("y"), k("x"), k("z"))),
          list(a = "y", "x", "z")
        )
        expect_identical(
          as.list(reactives(k("x"), k("z"))),
          list("x", "z")
        )
        expect_identical(
          as.list(reactives(a = k("y"), b = k("z"))),
          list(a = "y", b = "z")
        )

        # an explicit numeric name is a name, not a position
        expect_identical(
          as.list(reactives(`1` = k("x"), `3` = k("z"))),
          list(`1` = "x", `3` = "z")
        )
      }
    )
  )
})

test_that("reactives errors on a name collision", {

  k <- function(v) function() v

  expect_error(
    reactives(a = k("x"), a = k("y")),
    class = "reactives_key_collision"
  )
})

test_that("the public operators delegate to the bindings", {

  with_mock_session(
    {
      rv <- reactiveVal(1)
      coll <- reactives(a = rv, b = reactive(99))

      # $<- / [[<- push a value through the binding; a reactiveVal cell sets
      coll$a <- 5
      expect_identical(isolate(coll$a), 5)

      # a reactive cell has no setter, so it errors of its own accord
      expect_error(coll$b <- 7)

      # assigning to an unregistered slot errors
      expect_error(coll$nope <- 1, class = "reactives_unknown_slot")

      # [<- re-registers the expression
      coll["a"] <- function() 100
      expect_identical(isolate(coll$a), 100)

      # [ returns a sub-reactives that shares the cell
      sub <- isolate(coll["b"])
      expect_identical(isolate(names(sub)), "b")
      expect_identical(isolate(sub$b), 99)

      # names<- re-keys, keeping the cells
      coll2 <- reactives(function() "x", a = function() "y")
      names(coll2) <- c("p", "q")
      expect_identical(isolate(names(coll2)), c("p", "q"))
      expect_identical(isolate(coll2$p), "x")
    }
  )
})

test_that("set_reactive replaces an existing slot without duplicating it", {

  with_mock_session(
    isolate(
      {
        coll <- reactives(a = function() "first")
        expect_identical(coll[["a"]], "first")

        set_reactive(coll, "a", function() "second")

        expect_identical(coll[["a"]], "second")
        expect_identical(names(coll), "a")
        expect_identical(length(coll), 1L)
      }
    )
  )
})

test_that("append_reactive adds unnamed slots at the end", {

  with_mock_session(
    {
      coll <- reactives(a = function() 1)

      seen <- NULL
      observe(seen <<- as.list(coll))
      session$flushReact()

      expect_identical(seen, list(a = 1))

      append_reactive(coll, function() 2)
      append_reactive(coll, function() 3)
      session$flushReact()

      expect_identical(seen, list(a = 1, 2, 3))
      expect_identical(isolate(names(coll)), c("a", "", ""))
      expect_identical(isolate(length(coll)), 3L)

      drop_reactive(coll, isolate(raw_keys(coll))[[2L]])
      session$flushReact()

      expect_identical(seen, list(a = 1, 3))
    }
  )
})

#' @section Static block:
#' Mainly useful for testing and examples, this block simply returns the data
#' with which it was initialized. Serialization of static blocks is not allowed
#' and exported code will not be self-contained in the sense that it will not
#' be possible to reproduce results in a script that contains code from a
#' static block.
#'
#' @param data Data (used directly as block result)
#'
#' @rdname new_data_block
#' @export
new_static_block <- function(data, ...) {

  ctor_envir <- environment()

  new_data_block(
    function(id) {
      moduleServer(
        id,
        function(input, output, session) {
          list(
            expr = reactive(
              bquote(get("data", envir = .(env)), list(env = ctor_envir))
            ),
            state = list(data = data)
          )
        }
      )
    },
    class = "static_block",
    allow_empty_state = TRUE,
    external_ctrl = FALSE,
    block_metadata = list(),
    ...
  )
}

#' Text block constructors
#'
#' A text block produces (markdown styled) text, given some (optional) data
#' input.
#'
#' @param ... Forwarded to `new_text_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_text_block()` inherit from
#' `text_block`.
#'
#' @export
new_text_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "text_block"), ctor, ...)
}

#' @export
block_output.text_block <- function(x, result, session) {
  renderUI(markdown(result))
}

#' @export
block_ui.text_block <- function(id, x, ...) {
  tagList(
    uiOutput(NS(id, "result"))
  )
}

#' Board
#'
#' A set of blocks, optionally connected via links and grouped into stacks
#' are organized as a `board` object. Boards are constructed using `new_board()`
#' and inheritance can be tested with `is_board()`, while validation is
#' available as (generic function) `validate_board()`. This central data
#' structure can be extended by adding further attributes and sub-classes. S3
#' dispatch is used in many places to control how the UI looks and feels and
#' using this extension mechanism, UI aspects can be customized to user
#' requirements. Several utilities are available for retrieving and modifying
#' block attributes (see [board_blocks()]).
#'
#' @param blocks Set of blocks
#' @param links Set of links
#' @param stacks Set of stacks
#' @param options Board-level user settings
#' @param ... Further (metadata) attributes
#' @param ctor,pkg Constructor information (used for serialization)
#' @param class Board sub-class
#'
#' @examples
#' brd <- new_board(
#'   c(
#'      a = new_dataset_block(),
#'      b = new_subset_block()
#'   ),
#'   list(from = "a", to = "b")
#' )
#'
#' is_board(brd)
#'
#' @return The board constructor `new_board()` returns a `board` object, as does
#' the validator `validate_board()`, which typically is called for side effects
#' in the form of errors. Inheritance checking as `is_board()` returns a scalar
#' logical.
#'
#' @export
new_board <- function(blocks = list(), links = list(), stacks = list(),
                      options = default_board_options(), ...,
                      ctor = NULL, pkg = NULL, class = character()) {

  blocks <- as_blocks(blocks)
  links <- as_links(links)
  stacks <- as_stacks(stacks)
  options <- as_board_options(options)

  links <- complete_unary_inputs(links, blocks)

  validate_board(
    structure(
      list(
        blocks = blocks,
        links = links,
        stacks = stacks,
        options = options,
        ...
      ),
      ctor = resolve_ctor(forward_ctor(ctor), pkg),
      class = c(class, "board")
    )
  )
}

complete_unary_inputs <- function(x, blocks) {

  ids <- names(blocks)

  to_complete <- x$to %in% ids[int_ply(blocks, block_arity) == 1L] & (
    is.na(x$input) | nchar(x$input) == 0L
  )

  inputs <- lapply(blocks[unique(x$to[to_complete])], block_inputs)

  x$input[to_complete] <- chr_ply(inputs[x$to[to_complete]], identity)

  x
}

validate_board_blocks_links <- function(blocks, links) {

  ids <- names(blocks)

  if (!all(links$from %in% ids) || !all(links$to %in% ids)) {
    blockr_abort(
      "Expecting all links to refer to known block IDs.",
      class = "board_block_link_name_mismatch"
    )
  }

  arity <- int_ply(blocks, block_arity)

  for (i in Filter(Negate(is.na), unique(arity))) {

    fail <- table(factor(links$to, levels = ids[arity == i])) > i

    if (any(fail)) {
      blockr_abort(
        "{i}-ary blocks are expected to have at most {i} incoming edge{?s}, ",
        "which does not hold for {qty(sum(fail))}block{?s} ",
        "{names(fail)[fail]}.",
        class = "board_block_link_arity_mismatch"
      )
    }
  }

  inputs <- lapply(blocks[!is.na(arity)], block_inputs)

  for (i in intersect(links$to, names(inputs))) {

    unknown <- setdiff(links$input[links$to == i], inputs[[i]])

    if (length(unknown)) {
      blockr_abort(
        "Block {i} expects input{?s} {inputs[[i]]} but received {unknown} ",
        "instead.",
        class = "board_block_link_input_mismatch"
      )
    }
  }

  invisible()
}

validate_board_blocks_stacks <- function(blocks, stacks) {

  for (i in seq_along(stacks)) {

    stk <- stacks[[i]]
    chk <- is.element(stk, names(blocks))

    if (!all(chk)) {
      blockr_abort(
        "Unknown {qty(sum(!chk))}block{?s} {stk[!chk]} {?is/are} assigned to ",
        "stack {names(stacks)[i]}.",
        class = "board_block_stack_name_mismatch"
      )
    }
  }

  invisible()
}

#' @param x Board object
#' @rdname new_board
#' @export
validate_board <- function(x) {
  UseMethod("validate_board", x)
}

#' @export
validate_board.board <- function(x) {

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a board object to be list-like.",
      class = "board_list_like_invalid"
    )
  }

  cmps <- c("blocks", "links")

  if (!all(cmps %in% names(x))) {
    blockr_abort(
      "Expecting a board object to contain components {cmps}.",
      class = "board_list_components_invalid"
    )
  }

  blks <- validate_blocks(board_blocks(x))

  validate_board_blocks_links(blks, validate_links(board_links(x)))
  validate_board_blocks_stacks(blks, validate_stacks(board_stacks(x)))

  x
}

#' @export
validate_board.default <- function(x) {

  blockr_abort(
    "Expecting a board object to inherit from \"baord\".",
    class = "board_inheritance_invalid"
  )
}

#' @rdname new_board
#' @export
is_board <- function(x) {
  inherits(x, "board")
}

#' @export
sort.board <- function(x, decreasing = FALSE, ...) {

  res <- topo_sort(as.matrix(x))
  ids <- board_block_ids(x)

  ind <- match(res, ids)

  if (isTRUE(decreasing)) {
    ind <- rev(ind)
  }

  board_blocks(x) <- board_blocks(x)[ind]

  x
}

#' @export
as.matrix.board <- function(x, ...) {

  block_ids <- board_block_ids(x)
  links <- board_links(x)

  as_adjacency_matrix(links$from, links$to, block_ids)
}

#' @rdname topo_sort
#' @export
is_acyclic.board <- function(x) {
  is_acyclic(as.matrix(x))
}

#' Board utils
#'
#' A set of utility functions is available for querying and manipulating board
#' components (i.e. blocks, links and stacks). Functions for retrieving and
#' modifying board options are documented in [new_board_options()].
#'
#' @section Blocks:
#' Board blocks can be retrieved using `board_blocks()` and updated with the
#' corresponding replacement function `board_blocks<-()`. If just the current
#' board IDs are of interest, `board_block_ids()` is available as short for
#' `names(board_blocks(x))`. In order to remove block(s) from a board, the
#' (generic) convenience function `rm_blocks()` is exported, which takes care
#' (in the default implementation for `board`) of also updating links and
#' stacks accordingly. The more basic replacement function `board_blocks<-()`
#' might fail at validation of the updated board object if an inconsistent
#' state results from an update (e.g. a block referenced by a stack is no
#' longer available).
#'
#' @param x Board
#'
#' @examples
#' brd <- new_board(
#'   c(
#'      a = new_dataset_block(),
#'      b = new_subset_block()
#'   ),
#'   list(from = "a", to = "b")
#' )
#'
#' board_blocks(brd)
#' board_block_ids(brd)
#'
#' board_links(brd)
#' board_link_ids(brd)
#'
#' board_stacks(brd)
#' board_stack_ids(brd)
#'
#' board_options(brd)
#'
#' @return Functions for retrieving, as well as updating components
#' (`board_blocks()`/`board_links()`/`board_stacks()`/`board_options()` and
#' `board_blocks<-()`/`board_links<-()`/`board_stacks<-()`/`board_options<-()`)
#' return corresponding objects (i.e. `blocks`, `links`, `stacks` and
#' `board_options`), while ID getters (`board_block_ids()`, `board_link_ids()`,
#' `board_stack_ids()` and `board_option_ids()`) return character vectors, as
#' does `available_stack_blocks()`. Convenience functions `rm_blocks()`,
#' `modify_board_links()` and `modify_board_stacks()` return an updated `board`
#' object.
#'
#' @export
board_blocks <- function(x) {
  stopifnot(is_board(x))
  x[["blocks"]]
}

#' @param value Replacement value
#' @rdname board_blocks
#' @export
`board_blocks<-` <- function(x, value) {
  stopifnot(is_board(x))
  x[["blocks"]] <- value
  validate_board(x)
}

#' @rdname board_blocks
#' @export
board_block_ids <- function(x) {
  names(board_blocks(x))
}

#' @param rm Block/link/stack IDs to remove
#' @param ... Further arguments they may be passed from the board server context
#' @param session Shiny session object
#'
#' @rdname board_blocks
#' @export
rm_blocks <- function(x, rm, ..., session = get_session()) {
  UseMethod("rm_blocks", x)
}

#' @export
rm_blocks.board <- function(x, rm, ..., session = get_session()) {

  if (is_blocks(rm)) {
    rm <- names(rm)
  }

  blocks <- board_blocks(x)

  stopifnot(is.character(rm), all(rm %in% names(blocks)))

  links <- board_links(x)

  if (any(rm %in% c(links$from, links$to))) {

    # nolint start: object_usage_linter.
    blk <- intersect(rm, c(links$from, links$to))
    lnk <- names(links[links$from %in% rm | links$to %in% rm])
    # nolint end

    blockr_abort(
      "Cannot remove {qty(blk)} block{?s} {blk} which participate in ",
      "{qty(lnk)} link{?s} {lnk}.",
      class = "invalid_removal_of_used_block"
    )
  }

  stacks <- board_stacks(x)
  stck_blks <- lapply(stacks, stack_blocks)

  if (any(rm %in% unlst(stck_blks))) {

    # nolint start: object_usage_linter.
    blk <- intersect(rm, unlst(stck_blks))
    stk <- lengths(lapply(stck_blks, intersect, rm))
    # nolint end

    blockr_abort(
      "Cannot remove {qty(blk)} block{?s} {blk} which participate in ",
      "{qty(stk)} stack{?s} {stk}.",
      class = "invalid_removal_of_used_block"
    )
  }

  board_blocks(x) <- blocks[!names(blocks) %in% rm]

  x
}

#' @section Links:
#' Board links can be retrieved using `board_links()` and updated with the
#' corresponding replacement function `board_links<-()`. If only links IDs are
#' of interest, this is available as `board_link_ids()`, which is short for
#' `names(board_links(x))`. A (generic) convenience function for all kinds of
#' updates to board links in one is available as `modify_board_links()`. With
#' arguments `add`, `rm` and `mod`, links can be added, removed or modified in
#' one go.
#'
#' @rdname board_blocks
#' @export
board_links <- function(x) {
  stopifnot(is_board(x))
  x[["links"]]
}

#' @rdname board_blocks
#' @export
`board_links<-` <- function(x, value) {
  stopifnot(is_board(x))
  x[["links"]] <- value
  validate_board(x)
}

#' @rdname board_blocks
#' @export
board_link_ids <- function(x) {
  names(board_links(x))
}

#' @param add Links/stacks to add
#' @param mod Link/stacks to modify
#' @rdname board_blocks
#' @export
modify_board_links <- function(x, add = NULL, rm = NULL, mod = NULL, ...,
                               session = get_session()) {

  if (!length(add) && !length(rm) && !length(mod)) {
    return(x)
  }

  UseMethod("modify_board_links", x)
}

#' @export
modify_board_links.board <- function(x, add = NULL, rm = NULL, mod = NULL,
                                     ..., session = get_session()) {

  links <- board_links(x)

  if (is_links(rm)) {
    rm <- names(rm)
  }

  keep <- intersect(names(add), rm)

  if (length(keep)) {
    links[keep] <- add[keep]
    add <- add[setdiff(names(add), keep)]
    rm <- setdiff(rm, keep)
  }

  if (length(rm)) {
    stopifnot(is.character(rm), all(rm %in% names(links)))
    links <- links[!names(links) %in% rm]
  }

  if (length(mod)) {
    links[names(mod)] <- mod
  }

  board_links(x) <- c(links, add)

  x
}

#' @section Stacks:
#' Board stacks can be retrieved using `board_stacks()` and updated with the
#' corresponding replacement function `board_stacks<-()`. If only the stack IDs
#' are of interest, this is available as `board_stack_ids()`, which is short
#' for `names(board_stacks(x))`. A (generic) convenience function to update
#' stacks is available as `modify_board_stacks()`, which can add, remove and
#' modify stacks depending on arguments passed as `add`, `rm` and `mod`. If
#' block IDs that are not already associated with a stack (i.e. "free" blocks)
#' are of interest, this is available as `available_stack_blocks()`.
#'
#' @rdname board_blocks
#' @export
board_stacks <- function(x) {
  stopifnot(is_board(x))
  x[["stacks"]]
}

#' @rdname board_blocks
#' @export
`board_stacks<-` <- function(x, value) {
  stopifnot(is_board(x))
  x[["stacks"]] <- value
  validate_board(x)
}

#' @rdname board_blocks
#' @export
board_stack_ids <- function(x) {
  names(board_stacks(x))
}

#' @rdname board_blocks
#' @export
modify_board_stacks <- function(x, add = NULL, rm = NULL, mod = NULL, ...,
                                session = get_session()) {

  if (!length(add) && !length(rm) && !length(mod)) {
    return(x)
  }

  UseMethod("modify_board_stacks", x)
}

#' @export
modify_board_stacks.board <- function(x, add = NULL, rm = NULL, mod = NULL,
                                      ..., session = get_session()) {

  stacks <- board_stacks(x)

  if (is_stacks(rm)) {
    rm <- names(rm)
  }

  if (length(rm)) {
    stopifnot(is.character(rm), all(rm %in% names(stacks)))
    stacks <- stacks[!names(stacks) %in% rm]
  }

  if (length(mod)) {
    stacks[names(mod)] <- mod
  }

  board_stacks(x) <- c(stacks, add)

  x
}

#' @section Options:
#' Board options can be retrieved using `board_options()` and updated with the
#' corresponding replacement function `board_options<-()`.  If only the option
#' IDs are of interest, this is available as `board_option_ids()`, which calls
#' [board_option_id()] on each board option.
#'
#' @rdname board_blocks
#' @export
board_options <- function(x) {
  UseMethod("board_options")
}

#' @export
board_options.board <- function(x) {
  res <- x[["options"]]
  validate_board_options(res)
  res
}

#' @rdname board_blocks
#' @export
`board_options<-` <- function(x, value) {
  stopifnot(is_board(x))
  x[["options"]] <- validate_board_options(value)
  x
}

#' @rdname board_blocks
#' @export
board_option_ids <- function(x) {
  stopifnot(is_board(x))
  chr_ply(board_options(x), board_option_id)
}

#' @param blocks,stacks Sets of blocks/stacks
#' @rdname board_blocks
#' @export
available_stack_blocks <- function(x, stacks = board_stacks(x),
                                   blocks = board_stack_ids(x)) {

  Reduce(setdiff, lapply(stacks, as.character), blocks)
}

#' @export
block_inputs.board <- function(x) {
  lapply(set_names(board_blocks(x), board_block_ids(x)), block_inputs)
}

#' @rdname new_board_options
#' @export
board_ctor <- function(x) {
  stopifnot(is_board(x))
  attr(x, "ctor")
}

#' @export
format.board <- function(x, ...) {

  paste_name_append_empty <- function(x, nme) {
    c(paste0(nme, x[1L]), x[-1L], "")
  }

  out <- ""

  for (cl in rev(class(x))) {
    out <- paste0("<", cl, out, ">")
  }

  blk <- board_blocks(x)

  if (length(blk)) {

    blk <- lapply(blk, format, ...)
    blk <- map(paste_name_append_empty, blk, names(blk))

    out <- c(
      out,
      "",
      paste0("Blocks[", length(blk), "]:"),
      "",
      unlst(blk)
    )
  }

  lnk <- board_links(x)

  if (length(lnk)) {

    out <- c(
      out,
      paste0("Links[", length(lnk), "]:"),
      "",
      paste0(names(lnk), ": ", format(lnk))
    )
  }

  stk <- board_stacks(x)

  if (length(stk)) {

    stk <- lapply(stk, format, ...)
    stk <- map(paste_name_append_empty, stk, names(stk))

    out <- c(
      out,
      if (length(lnk)) "",
      paste0("Stacks[", length(stk), "]:"),
      "",
      unlst(stk)
    )
  }

  if ((length(blk) && !length(lnk)) || length(stk)) {
    out <- out[-length(out)]
  }

  c(
    out,
    "",
    "Options:",
    format(as_board_options(x))
  )
}

#' @export
print.board <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.board <- function(x, ...) {
  paste(
    paste0("<", class(x)[1L], ">"),
    str_value(board_blocks(x)),
    str_value(board_links(x)),
    str_value(board_stacks(x)),
    str_value(as_board_options(x)),
    sep = "\n"
  )
}

#' @export
str.board <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @rdname board_blocks
#' @export
clear_board <- function(x) {

  stopifnot(is_board(x))

  res <- modify_board_stacks(x, rm = board_stacks(x))
  res <- modify_board_links(res, rm = board_links(res))
  res <- rm_blocks(res, board_blocks(res))

  res
}

test_that("board constructor", {

  expect_s3_class(new_board(new_dataset_block()), "board")

  board <- new_board(
    list(
      d = new_merge_block(),
      a = new_dataset_block(),
      c = new_subset_block(),
      e = new_subset_block(),
      b = new_dataset_block()
    ),
    data.frame(
      id = c("ad", "cd", "bc", "de"),
      from = c("a", "c", "b", "d"),
      to = c("d", "d", "c", "e"),
      input = c("x", "y", "", "")
    ),
    list(bc = c("b", "c"))
  )

  expect_s3_class(board, "board")
  expect_snapshot(print(board))

  srt_inc <- sort(board)

  expect_true(
    match("a", board_block_ids(srt_inc)) < match("d", board_block_ids(srt_inc))
  )

  expect_true(
    match("b", board_block_ids(srt_inc)) < match("d", board_block_ids(srt_inc))
  )

  expect_true(
    match("b", board_block_ids(srt_inc)) < match("c", board_block_ids(srt_inc))
  )

  expect_true(
    match("c", board_block_ids(srt_inc)) < match("d", board_block_ids(srt_inc))
  )

  srt_dec <- sort(board, decreasing = TRUE)

  expect_true(
    match("a", board_block_ids(srt_dec)) > match("d", board_block_ids(srt_dec))
  )

  expect_true(
    match("b", board_block_ids(srt_dec)) > match("d", board_block_ids(srt_dec))
  )

  expect_true(
    match("b", board_block_ids(srt_dec)) > match("c", board_block_ids(srt_dec))
  )

  expect_true(
    match("c", board_block_ids(srt_dec)) > match("d", board_block_ids(srt_dec))
  )

  expect_true(is_acyclic(board))

  expect_error(
    new_board(
      list(
        a = new_dataset_block(),
        b = new_subset_block()
      ),
      new_link("a", "b", "foo")
    ),
    class = "board_block_link_input_mismatch"
  )

  expect_error(
    new_board(
      list(
        a = new_dataset_block(),
        b = new_subset_block()
      ),
      data.frame(from = "a", to = "b", input = "foo")
    ),
    class = "board_block_link_input_mismatch"
  )

  expect_error(
    new_board(
      list(
        a = new_dataset_block(),
        b = new_dataset_block()
      ),
      data.frame(from = "a", to = "b")
    ),
    class = "board_block_link_arity_mismatch"
  )

  expect_error(
    new_board(
      list(
        a = new_dataset_block(),
        b = new_subset_block()
      ),
      stacks = "ab"
    ),
    class = "board_block_stack_name_mismatch"
  )

  expect_error(
    rm_blocks(board, "e"),
    class = "invalid_removal_of_used_block"
  )

  lnks <- board_links(board)
  board_links(board) <- lnks[setdiff(names(lnks), "de")]

  expect_snapshot(print(rm_blocks(board, "e")))

  upd <- reactiveVal(list(blocks = list(rm = "b")))

  isolate(preprocess_board_update(upd, board))

  upd <- isolate(upd())

  expect_type(upd, "list")

  expect_named(upd, c("blocks", "links", "stacks"), ignore.order = TRUE)

  expect_length(upd$links, 1L)
  expect_named(upd$links, "rm")
  expect_identical(upd$links$rm, "bc")

  expect_length(upd$stacks, 1L)
  expect_named(upd$stacks, "mod")
  expect_identical(stack_blocks(upd$stacks$mod[[1L]]), "c")

  expect_error(
    validate_board(structure("123", class = "board")),
    class = "board_list_like_invalid"
  )

  expect_error(
    validate_board(structure(list(), class = "board")),
    class = "board_list_components_invalid"
  )

  expect_error(
    validate_board(NULL),
    class = "board_inheritance_invalid"
  )

  expect_error(
    rm_blocks(
      new_board(
        blocks(a = new_dataset_block()),
        stacks = stacks(a = "a")
      ),
      blocks(a = new_dataset_block())
    ),
    class = "invalid_removal_of_used_block"
  )

  inps <- block_inputs(board)

  expect_type(inps, "list")
  expect_true(all(lgl_ply(inps, is.character)))

  expect_type(board_option_ids(board), "character")

  opts <- board_options(board)
  board_options(board) <- opts[-1L]

  expect_length(board_options(board), length(opts) - 1L)

  empty <- clear_board(board)

  expect_length(board_blocks(empty), 0L)
  expect_length(board_links(empty), 0L)
  expect_length(board_stacks(empty), 0L)

  expect_identical(
    board_options(board),
    board_options(empty)
  )
})

test_that("validate_board checks each collection, not just cross-relations", {

  board <- new_board(c(a = new_dataset_block(), b = new_subset_block()))

  # Board accessors are pure reads, so validate_board must itself run each
  # collection's validator. A bare list (e.g. from a botched deserialization)
  # slips past the cross-relationship checks but not the per-collection ones.

  blocks_corrupt <- board
  blocks_corrupt[["blocks"]] <- list()

  expect_error(
    validate_board(blocks_corrupt),
    class = "blocks_class_invalid"
  )

  links_corrupt <- board
  links_corrupt[["links"]] <- list()

  expect_error(
    validate_board(links_corrupt),
    class = "links_class_invalid"
  )

  stacks_corrupt <- board
  stacks_corrupt[["stacks"]] <- list()

  expect_error(
    validate_board(stacks_corrupt),
    class = "stacks_class_invalid"
  )
})

test_that("a board has a compact str_value()", {

  board <- new_board(
    blocks = c(a = new_dataset_block(), b = new_subset_block()),
    stacks = list(s1 = new_stack(c("a", "b"), name = "my stack"))
  )

  res <- str_value(board)
  lines <- strsplit(res, "\n")[[1L]]

  expect_length(res, 1L)
  expect_identical(lines[1L], "<board>")
  expect_true("<blocks[2]>" %in% lines)
  expect_true("  a: <dataset_block> dataset*, package" %in% lines)
  expect_true("<links[0]>" %in% lines)
  expect_true("<stacks[1]>" %in% lines)
  expect_true("  s1: <stack> \"my stack\": a, b" %in% lines)
  expect_true(any(startsWith(lines, "<board_options[")))

  expect_identical(capture.output(str(board))[1L], " <board>")
})

test_that("empty variadic link inputs stay unnamed", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(
      ac = new_link("a", "c"),
      bc = new_link("b", "c")
    )
  )

  expect_identical(board_links(board)$input, c("", ""))
})

test_that("editing a link via add + rm overlap preserves its position", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(
      ac = new_link("a", "c", "left"),
      bc = new_link("b", "c")
    )
  )

  edited <- modify_board_links(
    board,
    add = links(ac = new_link("a", "c", "x")),
    rm = "ac"
  )

  expect_identical(names(board_links(edited)), c("ac", "bc"))
  expect_identical(board_links(edited)[["ac"]][["input"]], "x")
})

test_that("blocks utils", {

  a <- new_dataset_block()
  b <- new_subset_block()

  ab <- c(a, b)

  expect_s3_class(ab, "blocks")
  expect_length(ab, 2L)
  expect_named(ab)

  names(ab) <- c("a", "b")

  expect_named(ab, c("a", "b"))

  names(ab)[2] <- "c"

  expect_named(ab, c("a", "c"))

  expect_error(
    names(ab)[1] <- "c",
    class = "blocks_names_unique_invalid"
  )

  expect_error(
    names(ab) <- c("a", "a"),
    class = "blocks_names_unique_invalid"
  )

  names(ab) <- NULL

  expect_named(ab, c("", ""))

  names(ab) <- c("a", "b")

  expect_snapshot(print(ab))

  c <- new_dataset_block()

  abc <- c(ab, c = c)

  expect_s3_class(abc, "blocks")
  expect_length(abc, 3L)
  expect_named(abc, c("a", "b", "c"))

  abc["c"] <- NULL

  expect_identical(abc, ab)

  expect_error(
    abc["c"] <- NULL
  )

  abc <- c(ab, c = c)

  expect_error(
    abc["c"] <- c,
    class = "blocks_assignment_name_invalid"
  )

  abc["c"] <- blocks(c = c)

  expect_identical(abc, c(ab, c = c))

  abc["c"] <- list(c = c)

  expect_identical(abc, c(ab, c = c))

  expect_error(
    abc["c"] <- list(d = c),
    class = "blocks_assignment_name_invalid"
  )

  abc[["c"]] <- c

  expect_identical(abc, c(ab, c = c))

  expect_s3_class(abc["c"], "blocks")
  expect_length(abc["c"], 1L)
  expect_named(abc["c"], "c")

  expect_s3_class(abc[["c"]], "block")

  expect_error(
    validate_blocks("a"),
    class = "blocks_class_invalid"
  )

  expect_error(
    validate_blocks(
      structure(
        1,
        class = c("blocks", "vctrs_vctr", "list")
      )
    ),
    class = "blocks_contains_invalid"
  )

  expect_error(
    validate_blocks(
      structure(
        list(a),
        class = c("blocks", "vctrs_vctr", "list")
      )
    ),
    class = "blocks_names_invalid"
  )

  expect_error(
    validate_blocks(
      structure(
        list(a = a, a = b),
        class = c("blocks", "vctrs_vctr", "list")
      )
    ),
    class = "blocks_names_invalid"
  )
})

test_that("a blocks container has a compact str_value()", {

  brd <- new_board(blocks = c(a = new_dataset_block(), b = new_subset_block()))

  expect_identical(
    str_value(board_blocks(brd)),
    paste(
      "<blocks[2]>",
      "  a: <dataset_block> dataset*, package",
      "  b: <subset_block> subset*, select*",
      sep = "\n"
    )
  )

  expect_identical(str_value(board_blocks(new_board())), "<blocks[0]>")

  out <- capture.output(res <- withVisible(str(board_blocks(brd))))

  expect_identical(
    out,
    c(
      " <blocks[2]>",
      "  a: <dataset_block> dataset*, package",
      "  b: <subset_block> subset*, select*"
    )
  )
  expect_false(res$visible)
  expect_identical(res$value, board_blocks(brd))
})

test_that("str_value() of a blocks container is extensible via NextMethod()", {

  registerS3method(
    "str_value", "test_sub_blocks",
    function(x, ...) paste0(NextMethod(), "\n[extra]"),
    envir = asNamespace("blockr.core")
  )

  blk <- board_blocks(new_board(blocks = c(a = new_dataset_block())))
  class(blk) <- c("test_sub_blocks", class(blk))

  expect_identical(
    str_value(blk),
    "<test_sub_blocks[1]>\n  a: <dataset_block> dataset*, package\n[extra]"
  )
})

test_that("the default server produces the deserialized board", {

  test_board <- new_board(
    blocks = c(
      a = new_dataset_block("iris"),
      b = new_subset_block()
    ),
    links = links(from = "a", to = "b")
  )

  temp <- withr::local_tempfile(fileext = ".json")

  testServer(
    preserve_board_server,
    file.copy(output$serialize, temp),
    args = list(board = reactiveValues(board = test_board))
  )

  testServer(
    preserve_board_server,
    {
      session$setInputs(restore = list(datapath = temp))

      restored <- res()

      expect_s3_class(restored, class(test_board))
      expect_length(board_blocks(restored), length(board_blocks(test_board)))
      expect_setequal(board_block_ids(restored), board_block_ids(test_board))
      expect_setequal(board_link_ids(restored), board_link_ids(test_board))
    },
    args = list(board = reactiveValues(board = new_board()))
  )
})

test_that("board_server exposes the restored board through board_refresh", {

  test_board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_merge_block(by = "Time")
    ),
    links = links(
      from = c("a", "b"),
      to = c("c", "c"),
      input = c("x", "y")
    )
  )

  temp <- withr::local_tempfile(fileext = ".json")

  testServer(
    get_s3_method("board_server", test_board),
    {
      ser_deser <- session$makeScope("preserve_board")
      file.copy(ser_deser$output$serialize, temp)
    },
    args = list(x = test_board, plugins = preserve_board())
  )

  testServer(
    get_s3_method("board_server", test_board),
    {
      ser_deser <- session$makeScope("preserve_board")
      ser_deser$setInputs(restore = list(datapath = temp))

      restored <- board_refresh()

      expect_length(board_blocks(restored), length(board_blocks(test_board)))
      expect_setequal(board_block_ids(restored), board_block_ids(test_board))
      expect_setequal(board_link_ids(restored), board_link_ids(test_board))
    },
    args = list(x = new_board(), plugins = preserve_board())
  )
})

test_that("serve_board_srv stages and reloads when board_refresh fires", {

  refresh <- reactiveVal()
  staged <- NULL

  spy <- board_loader(
    resolve = function(request, session, default) NULL,
    stage = function(board, session) staged <<- board
  )

  local_mocked_bindings(
    board_server = function(id, x, ...) list(board_refresh = refresh),
    blockr_test_exports = function(x, rv, ...) invisible()
  )

  board <- new_board(blocks = c(a = new_dataset_block("iris")))

  srv <- serve_board_srv(
    "app", board, spy, blockr_app_plugins, blockr_app_options
  )

  testServer(
    srv,
    {
      refresh(board)
      session$flushReact()

      expect_identical(staged, board)
    }
  )
})

test_that("preserve_board is the save/restore plugin (no loader on it)", {

  plug <- preserve_board()

  expect_true(is_plugin(plug))
  expect_s3_class(plug, "preserve_board")
  expect_setequal(names(plug), c("server", "ui"))
  expect_null(attr(plug, "loader"))
})

test_that("board_loader bundles a resolve and an optional validated stage", {

  ld <- board_loader(
    resolve = function(request, session, default) NULL,
    stage = function(board, session) NULL
  )

  expect_true(is_board_loader(ld))
  expect_true(is.function(ld$resolve))
  expect_true(is.function(ld$stage))

  stageless <- board_loader(function(request, session, default) NULL)
  expect_true(is_board_loader(stageless))
  expect_null(stageless$stage)

  expect_error(
    board_loader(function(x) NULL),
    class = "board_loader_resolve_invalid"
  )
  expect_error(
    board_loader(
      function(request, session, default) NULL,
      function(b, s) NULL
    ),
    class = "board_loader_stage_invalid"
  )

  # formals must match exactly -- missing/extra args (or `...`) are rejected
  expect_error(
    board_loader(function(request, session) NULL),
    class = "board_loader_resolve_invalid"
  )
  expect_error(
    board_loader(function(...) NULL),
    class = "board_loader_resolve_invalid"
  )
  expect_error(
    board_loader(
      function(request, session, default) NULL,
      function(board, session, x) NULL
    ),
    class = "board_loader_stage_invalid"
  )
})

test_that("the default loader stages a board and resolves it back", {

  ld <- local_loader()
  board <- new_board(blocks = c(a = new_dataset_block("iris")))

  written <- NULL
  session <- list(
    clientData = list(url_search = ""),
    updateQueryString = function(url, mode) written <<- url
  )

  ld$stage(board, session)
  expect_true(grepl(reload_param, written, fixed = TRUE))

  # the staged board resolves back from the URL the stage wrote
  get_req <- list(QUERY_STRING = sub("^[?]", "", written))
  expect_identical(ld$resolve(get_req, NULL, board), board)
  expect_null(ld$resolve(list(QUERY_STRING = ""), NULL, board))

  # a separate loader instance has its own store
  expect_null(local_loader()$resolve(get_req, NULL, board))
})

test_that("the default loader consumes the token at the WS read", {

  ld <- local_loader()
  board <- new_board(blocks = c(a = new_dataset_block("iris")))

  written <- NULL
  staging <- list(
    clientData = list(url_search = ""),
    updateQueryString = function(url, mode) written <<- url
  )
  ld$stage(board, staging)

  # WS read off clientData consumes the token from the store
  ws <- list(
    clientData = list(url_search = written),
    updateQueryString = function(url, mode) NULL
  )
  expect_identical(ld$resolve(NULL, ws, board), board)

  # consumed: a GET read by the same query no longer finds it
  get_req <- list(QUERY_STRING = sub("^[?]", "", written))
  expect_null(ld$resolve(get_req, NULL, board))
})

test_that("resolve_board prefers the loader board, validates, falls back", {

  default <- new_board()
  staged <- new_board(blocks = c(a = new_dataset_block("iris")))

  # resolve_board augments the returned board's options (covered below), so the
  # picked board is no longer identical to the input -- compare by blocks.
  pref <- board_loader(function(request, session, default) staged)
  expect_identical(
    board_blocks(resolve_board(default, pref, list(), NULL)),
    board_blocks(staged)
  )

  none <- board_loader(function(request, session, default) NULL)
  expect_identical(
    board_blocks(resolve_board(default, none, list(), NULL)),
    board_blocks(default)
  )

  bad <- board_loader(function(request, session, default) "nope")
  expect_error(
    resolve_board(default, bad, list(), NULL),
    class = "invalid_board_loader"
  )
})

test_that("resolve_board augments the board with the managed option set", {

  # A board declares only its own options, but the settings UI manages the
  # wider blockr_app_options() set -- options contributed by registered blocks
  # (e.g. the preview-row count `n_rows`). resolve_board() bakes that set onto
  # the board it returns so the server and serialization see the full set.
  brd <- new_board()

  expect_false("n_rows" %in% names(board_options(brd)))
  expect_true("n_rows" %in% names(blockr_app_options(brd)))

  null_loader <- board_loader(function(request, session, default) NULL)
  via_default <- resolve_board(
    brd, null_loader, request = list(), session = NULL
  )
  expect_true("n_rows" %in% names(board_options(via_default)))

  load_loader <- board_loader(function(request, session, default) new_board())
  via_loaded <- resolve_board(
    brd, load_loader, request = list(), session = NULL
  )
  expect_true("n_rows" %in% names(board_options(via_loaded)))
})

test_that("a user-changed block-backed board option survives save/restore", {

  # End to end of the augmentation: `n_rows` is contributed by blocks, not the
  # board itself, so before resolve_board() widened the board it was dropped on
  # save and reverted to its registry default on reload.
  x <- resolve_board(
    new_board(),
    board_loader(function(request, session, default) NULL),
    request = list(),
    session = NULL
  )

  ser <- NULL

  testServer(
    get_s3_method("board_server", x),
    {
      session$flushReact()

      session$setInputs(n_rows = 100L)
      session$flushReact()

      ser <<- serialize_board(rv$board, rv$blocks, id = "b", session = session)
    },
    args = list(
      x = x,
      plugins = list(manage_blocks()),
      options = blockr_app_options(x)
    )
  )

  loaded <- blockr_deser(ser)

  expect_true("n_rows" %in% names(board_options(loaded)))
  expect_identical(
    board_option_value(blockr_app_options(loaded)[["n_rows"]]),
    100L
  )
})

test_that("preserve_board return validation", {

  with_mock_session(
    {
      check_ser_deser_val(list(a = 1))
      sink_msg(
        expect_warning(
          session$flushReact(),
          "Expecting `preserve_board` to return a reactive value"
        )
      )
    }
  )

  with_mock_session(
    {
      check_ser_deser_val(reactiveVal(1))
      sink_msg(
        expect_warning(
          session$flushReact(),
          "Expecting the `preserve_board` return value to evaluate to a"
        )
      )
    }
  )
})

test_that("restore_board returns the deserialized board", {

  test_board <- new_board(
    blocks = c(a = new_dataset_block("iris"))
  )

  ser <- blockr_ser(test_board)
  result <- reactiveVal()

  with_mock_session(
    {
      restore_board(test_board, ser, result, session = session)

      val <- result()
      expect_true(is_board(val))
      expect_length(board_blocks(val), length(board_blocks(test_board)))
    }
  )
})

test_that("block-contributed board options survive save/restore", {

  brd <- new_board(blocks = c(a = new_dataset_block("iris")))

  temp <- withr::local_tempfile(fileext = ".json")

  testServer(
    get_s3_method("board_server", brd),
    {
      session$flushReact()
      session$setInputs(page_size = "25")
      session$flushReact()

      ser_deser <- session$makeScope("preserve_board")
      file.copy(ser_deser$output$serialize, temp)
    },
    args = list(
      x = brd,
      plugins = preserve_board(),
      options = blockr_app_options(brd)
    )
  )

  restored <- blockr_deser(read_json(temp))

  expect_true("page_size" %in% board_option_ids(restored))
  expect_identical(
    board_option_value(blockr_app_options(restored)[["page_size"]]),
    25L
  )
})

test_that("serialize_board uses constructor state for a not-yet-built block", {

  board <- new_board(blocks = c(a = new_dataset_block("iris")))

  # `blocks` empty: block `a` has not been constructed (still building in the
  # background), so it serializes from its constructor defaults.
  ser <- serialize_board(board, list(), id = "board")

  expect_identical(
    ser$payload$blocks$payload$a$payload$dataset,
    "iris"
  )
})

test_that("dummy ser/deser ui test", {
  expect_s3_class(preserve_board_ui("ser_deser", new_board()), "shiny.tag.list")
})

#' Plugin module for external control of block inputs
#'
#' This plugin enables setting block reactive state values from outside the
#' block expression server context. Blocks opt in to external control via the
#' `external_ctrl` argument to [new_block()], which can be set to `TRUE` (all
#' constructor inputs) or a character vector of specific input names. The
#' default UI renders a [shiny::textInput()] for each externally controllable
#' input along with a submit [shiny::actionButton()]. Both the server and UI
#' can be replaced with custom implementations by passing alternate functions
#' to `ctrl_block()`.
#'
#' The default server validates submitted values by evaluating the block
#' expression (via the `eval` reactive) after updating state. On success,
#' a reactive gate is returned as `TRUE`, allowing downstream evaluation to
#' proceed. On failure, state values are reverted to their previous values,
#' the user is notified, and the gate is set to `FALSE`, which blocks
#' downstream evaluation until a subsequent successful submit.
#'
#' @inheritParams new_plugin
#'
#' @return A plugin container inheriting from `ctrl_block` is returned by
#' `ctrl_block()`, while the UI component (i.e. `ctrl_block_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `ctrl_block_server()`) is expected to return a value
#' that passes validation (i.e. `TRUE` or a reactive gate).
#'
#' @export
ctrl_block <- function(server = ctrl_block_server, ui = ctrl_block_ui) {
  new_plugin(server, ui, validate_ctrl, class = "ctrl_block")
}

#' @param id Namespace ID
#' @param x Block object
#' @param vars Reactive state values (list of `reactiveVal` objects keyed by
#' input name). `block_name` is included by the default `block_server.block`
#' even though it is not part of the block's `expr_server` state.
#' @param data Input data paseed as list of reactive values
#' @param eval Reactive that evaluates the block expression against input
#' data. May be used to validate that the new values produce a successful
#' evaluation.
#'
#' @rdname ctrl_block
#' @export
ctrl_block_server <- function(id, x, vars, data, eval) {
  moduleServer(
    id,
    function(input, output, session) {

      inps <- names(vars)

      gate <- reactiveVal(TRUE)

      observeEvent(
        input$submit,
        {
          old <- lapply(vars[inps], reval)

          for (inp in inps) {
            val <- session$input[[inp]]
            if (!is.null(val) && !identical(vars[[inp]](), val)) {
              vars[[inp]](val)
            }
          }

          result <- try(eval(), silent = TRUE)

          if (inherits(result, "try-error")) {

            for (inp in inps) {
              if (!identical(vars[[inp]](), old[[inp]])) {
                vars[[inp]](old[[inp]])
              }
            }

            err <- attr(result, "condition")

            if (!inherits(err, "shiny.silent.error")) {
              notify(conditionMessage(err), type = "error", glue = FALSE)
            }

            gate(FALSE)

          } else {
            gate(TRUE)
          }
        }
      )

      gate
    }
  )
}

#' @param x Block
#' @rdname ctrl_block
#' @export
ctrl_block_ui <- function(id, x) {

  inps <- external_ctrl_vars(x)

  if (!length(inps)) {
    return(do.call(tagList, list()))
  }

  fields <- map(
    textInput,
    inputId = chr_ply(inps, NS(id)),
    label = paste0(toupper(substr(inps, 1L, 1L)), substring(inps, 2L))
  )

  do.call(
    tagList,
    c(
      fields,
      list(actionButton(NS(id, "submit"), "Submit"))
    )
  )
}

validate_ctrl <- function(x) {

  if (is.reactive(x) || isTRUE(x)) {
    return(invisible(x))
  }

  blockr_abort(
    "Expected `TRUE` or a reactive value, but got {class(x)} instead.",
    class = "expect_true_or_rv"
  )
}

test_that("ctrl_block constructor", {

  plg <- ctrl_block()

  expect_s3_class(plg, "ctrl_block")
  expect_s3_class(plg, "plugin")

  expect_identical(plg$server, ctrl_block_server)
  expect_identical(plg$ui, ctrl_block_ui)
})

test_that("ctrl_block_ui generates text inputs and submit button", {

  blk <- new_dataset_block("mtcars")
  ui <- ctrl_block_ui("ctrl", blk)

  expect_s3_class(ui, "shiny.tag.list")
  expect_length(ui, 3L)

  html <- paste(vapply(ui, as.character, character(1L)), collapse = "")
  expect_match(html, "ctrl-dataset")
  expect_match(html, "Dataset")
  expect_match(html, "ctrl-block_name")
  expect_match(html, "Block_name")
  expect_match(html, "ctrl-submit")
  expect_match(html, "Submit")

  blk2 <- new_subset_block()
  ui2 <- ctrl_block_ui("ctrl", blk2)

  expect_s3_class(ui2, "shiny.tag.list")
  expect_length(ui2, 4L)

  html2 <- paste(vapply(ui2, as.character, character(1L)), collapse = "")
  expect_match(html2, "ctrl-subset")
  expect_match(html2, "ctrl-select")
  expect_match(html2, "ctrl-block_name")
  expect_match(html2, "ctrl-submit")
})

test_that("ctrl_block_ui renders block_name even without other ctrl vars", {

  blk <- new_dataset_block("mtcars")
  attr(blk, "external_ctrl") <- FALSE

  ui <- ctrl_block_ui("ctrl", blk)
  expect_s3_class(ui, "shiny.tag.list")
  expect_length(ui, 2L)

  html <- paste(vapply(ui, as.character, character(1L)), collapse = "")
  expect_match(html, "ctrl-block_name")
  expect_match(html, "ctrl-submit")
})

test_that("ctrl_block_server syncs input to reactive state on submit", {

  blk <- new_dataset_block("mtcars")
  dataset_rv <- reactiveVal("mtcars")

  testServer(
    ctrl_block_server,
    {
      session$flushReact()

      expect_equal(vars$dataset(), "mtcars")

      session$setInputs(dataset = "iris")
      expect_equal(vars$dataset(), "mtcars")

      session$setInputs(submit = 1L)
      expect_equal(vars$dataset(), "iris")
    },
    args = list(
      x = blk,
      vars = list(dataset = dataset_rv),
      eval = reactive(TRUE)
    )
  )
})

test_that("ctrl_block_server syncs multiple inputs on submit", {

  blk <- new_subset_block(subset = "cyl > 4", select = "mpg")
  subset_rv <- reactiveVal("cyl > 4")
  select_rv <- reactiveVal("mpg")

  testServer(
    ctrl_block_server,
    {
      session$flushReact()

      expect_equal(vars$subset(), "cyl > 4")
      expect_equal(vars$select(), "mpg")

      session$setInputs(subset = "cyl > 6", select = "disp")
      expect_equal(vars$subset(), "cyl > 4")
      expect_equal(vars$select(), "mpg")

      session$setInputs(submit = 1L)
      expect_equal(vars$subset(), "cyl > 6")
      expect_equal(vars$select(), "disp")
    },
    args = list(
      x = blk,
      vars = list(subset = subset_rv, select = select_rv),
      eval = reactive(TRUE)
    )
  )
})

test_that("ctrl_block_server returns a reactive gate", {

  blk <- new_dataset_block("mtcars")

  testServer(
    ctrl_block_server,
    {
      session$flushReact()
      expect_true(is.reactive(session$returned))
      expect_true(session$returned())
    },
    args = list(
      x = blk,
      vars = list(dataset = reactiveVal("mtcars")),
      eval = reactive(TRUE)
    )
  )
})

test_that("ctrl_block_server does not update when input matches state", {

  blk <- new_dataset_block("mtcars")

  testServer(
    ctrl_block_server,
    {
      session$flushReact()

      session$setInputs(dataset = "mtcars")
      session$setInputs(submit = 1L)
      expect_equal(vars$dataset(), "mtcars")
    },
    args = list(
      x = blk,
      vars = list(dataset = reactiveVal("mtcars")),
      eval = reactive(TRUE)
    )
  )
})

test_that("ctrl_block_server reverts vars and blocks gate on eval error", {

  blk <- new_dataset_block("mtcars")
  dataset_rv <- reactiveVal("mtcars")
  fail <- reactiveVal(FALSE)

  testServer(
    ctrl_block_server,
    {
      session$flushReact()

      expect_true(session$returned())
      expect_equal(vars$dataset(), "mtcars")

      fail(TRUE)
      session$setInputs(dataset = "bad_value")
      session$setInputs(submit = 1L)

      expect_equal(vars$dataset(), "mtcars")
      expect_false(session$returned())
    },
    args = list(
      x = blk,
      vars = list(dataset = dataset_rv),
      eval = reactive(if (fail()) stop("eval failed") else TRUE)
    )
  )
})

test_that("ctrl_block_server recovers gate after successful submit", {

  blk <- new_dataset_block("mtcars")
  dataset_rv <- reactiveVal("mtcars")
  fail <- reactiveVal(FALSE)

  testServer(
    ctrl_block_server,
    {
      session$flushReact()

      fail(TRUE)
      session$setInputs(dataset = "bad")
      session$setInputs(submit = 1L)

      expect_false(session$returned())
      expect_equal(vars$dataset(), "mtcars")

      fail(FALSE)
      session$setInputs(dataset = "iris")
      session$setInputs(submit = 2L)

      expect_true(session$returned())
      expect_equal(vars$dataset(), "iris")
    },
    args = list(
      x = blk,
      vars = list(dataset = dataset_rv),
      eval = reactive(if (fail()) stop("eval failed") else TRUE)
    )
  )
})

test_that("validate_ctrl accepts TRUE and reactives", {

  expect_invisible(validate_ctrl(TRUE))
  expect_invisible(validate_ctrl(reactive(NULL)))
  expect_invisible(validate_ctrl(reactiveVal(TRUE)))

  expect_error(validate_ctrl(FALSE), class = "expect_true_or_rv")
  expect_error(validate_ctrl("abc"), class = "expect_true_or_rv")
  expect_error(validate_ctrl(42), class = "expect_true_or_rv")
})

test_that("ctrl_block integrates with block_server", {

  blk <- new_dataset_block("mtcars")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()

      expect_identical(session$returned$result(), datasets::mtcars)

      session$makeScope("ctrl_block")$setInputs(dataset = "iris")
      session$makeScope("ctrl_block")$setInputs(submit = 1L)

      expect_identical(session$returned$result(), datasets::iris)
    },
    args = list(x = blk, ctrl_block = ctrl_block())
  )
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----setup--------------------------------------------------------------------
library(blockr.core)


## flowchart LR
##   subgraph blockr_ggplot2[blockr.ggplot2]
##     new_block1[New block]
##     new_block2[New block]
##   end
##   subgraph blockr_echarts4r[blockr.echarts4r]
##     new_block3[New block]
##     new_block4[New block]
##   end
##   blockr_ggplot2 --> |register| registry
##   blockr_echarts4r --> |register| registry
##   subgraph registry[Registry]
##     subgraph select_reg[Select block]
##       reg_name[Name: select block]
##       reg_descr[Description: select columns in a table]
##       reg_classes[Classes: select_block, tranform_block]
##       reg_category[Category: transform]
##       reg_ctor[Construcor: new_select_block]
##       reg_package[Package: blockr.dplyr]
##     end
##     subgraph filter_reg[Filter block]
##     end
##     filter_reg --x |unregister| trash['fa:fa-trash']
##   end

## ----eval=FALSE---------------------------------------------------------------
# available_blocks()[["dataset_block"]]


## ----eval=TRUE----------------------------------------------------------------
names(available_blocks())


## ----eval=TRUE----------------------------------------------------------------
# ./R/dummy-block.R
new_dummy_block <- function(text = "Hello World", ...) {
  new_data_block(
    function(id) {
      moduleServer(id, function(input, output, session) {
        list(
          expr = reactive(quote(text)),
          state = list(text = text)
        )
      })
    },
    function(id) {
      shiny::tagList()
    },
    class = "dummy_block",
    ...
  )
}

register_dummy_blocks <- function() {
  register_blocks(
    c(new_dummy_block),
    name = c("dummy block"),
    description = c("A block that does nothing"),
    overwrite = TRUE
  )
}

register_dummy_blocks()


## -----------------------------------------------------------------------------
available_blocks()[["dummy_block"]]


## -----------------------------------------------------------------------------
unregister_blocks(uid = "dummy_block")

# Check it out
names(available_blocks())

#' Compact rendering
#'
#' `str_value()` returns a compact string describing an object. It is the
#' value-returning half of the compact rendering tier: where [utils::str()]
#' only displays (it `cat()`s and returns `NULL`), `str_value()` returns the
#' string, mirroring how [format()] returns what [print()] displays in the
#' full, multi-line tier. The [utils::str()] methods are thin wrappers that
#' display `str_value()`.
#'
#' A scalar object (a `block`, `stack`, `link`, `board_option` or `plugin`)
#' renders as a single line; a container (`blocks`, `stacks`, `links`,
#' `board_options`, `plugins`) and a whole `board` render as one element per
#' line below a `<class[n]>` header. The `block` method lists a block's
#' constructor inputs, marking the externally controllable ones (those
#' reported by [external_ctrl_vars()]) with a trailing `*`; the `stack`
#' method shows the stack name and its member block ids.
#'
#' This is the blockr extension point for token-dense renderings such as a
#' board summary. A home package surfaces a subclass's state by defining a
#' `str_value()` method, typically extending the parent's via [NextMethod()]
#' (the way `format.dock_stack()` appends a stack colour). [print()] and
#' [format()] are unaffected and remain the full, multi-line tier.
#'
#' @param x Object to render.
#' @param ... Generic consistency.
#'
#' @return `str_value()` returns a length-one character vector (multi-line for
#' containers). The [utils::str()] methods are called for their side effect
#' (display) and return their `object` invisibly.
#'
#' @examples
#' str_value(new_dataset_block())
#'
#' str_value(new_stack(c("plot", "data"), name = "My stack"))
#'
#' board <- new_board(c(a = new_dataset_block()))
#' str(board)
#'
#' @export
str_value <- function(x, ...) {
  UseMethod("str_value")
}

#' @rdname str_value
#' @export
str_value.default <- function(x, ...) {
  paste0("<", class(x)[1L], ">")
}

cat_str_value <- function(object, ...) {
  cat(" ", str_value(object), "\n", sep = "")
  invisible(object)
}

str_value_collection <- function(x, vals, ids = names(x)) {

  items <- if (length(x)) {
    if (is.null(ids)) paste0("  ", vals) else paste0("  ", ids, ": ", vals)
  }

  paste(
    c(paste0("<", class(x)[1L], "[", length(x), "]>"), items),
    collapse = "\n"
  )
}

#' Blocks
#'
#' Steps in a data analysis pipeline are represented by blocks. Each block
#' combines data input with user inputs to produce an output. In order to
#' create a block, which is implemented as a shiny module, we require a server
#' function, a function that produces some UI and a class vector.
#'
#' A block constructor may have arguments, which taken together define the
#' block state. It is good practice to expose all user-selectable arguments of
#' a block (i.e. everything excluding the "data" input) as block arguments such
#' that block can be fully initialized via the constructor. Some default values
#' are required such that blocks can be constructed via constructor calls
#' without arguments. Where it is sensible to do so, specific default values
#' are acceptable, but if in any way data dependent, defaults should map to
#' an "empty" input. For example, a block that provides [utils::head()]
#' functionality, one such argument could be `n` and a reasonable default value
#' could be `6L` (in line with corresponding default S3 method implementation).
#' On the other hand, a block that performs a [base::merge()] operation might
#' expose a `by` argument, but a general purpose default value (that does not
#' depend on the data) is not possible. Therefore, [new_merge_block()] has
#' `by = character()`.
#'
#' The return value of a block constructor should be the result of a call to
#' `new_block()` and `...` should be contained in the constructor signature
#' such that general block arguments (e.g. `name`) are available from the
#' constructor.
#'
#' @section Server:
#' The server function (passed as `server`) is expected to be a function that
#' returns a [shiny::moduleServer()]. This function is expected to have at
#' least an argument `id` (string-valued), which will be used as the module ID.
#' Further arguments may be used in the function signature, one for each "data"
#' input. A block implementing [utils::head()] for example could have a single
#' extra argument `data`, while a block that performs [base::merge()] requires
#' two extra arguments, e.g. `x` and `y`. Finally, a variadic block, e.g.
#' a block implementing something like [base::rbind()], needs to accommodate for
#' an arbitrary number of inputs. This is achieved by passing a container
#' object as `...args` and thus such a variadic block needs `...args` as part
#' of the server function signature. The container mirrors
#' [shiny::reactiveValues()] for reads (`names()`, `[[`, `$` and `as.list()`),
#' and accessing a slot yields the value of its (lazily evaluated) input. The
#' fixed per-data input arguments are instead passed as [shiny::reactive()] or
#' [shiny::reactiveVal()] objects.
#'
#' The server function may implement arbitrary shiny logic and is expected to
#' return a list with components `expr` and `state`. The expression corresponds
#' to the R code necessary to perform the block task and is expected to be
#' a reactive quoted expression. It should contain user-chosen values for all
#' user inputs and placeholders for all data inputs (using the same names for
#' data inputs as in the server function signature). Such an expression for a
#' [base::merge()] block could be created using [base::bquote()] as
#'
#' ```r
#' bquote(
#'   merge(x, y, by = .(cols)),
#'   list(cols = current_val())
#' }
#' ```
#'
#' where `current_val()` is a reactive that evaluates to the current user
#' selection of the `by` columns. This should then be wrapped in a
#' [shiny::reactive()] call such that `current_val()` can be evaluated whenever
#' the current expression is required.
#'
#' The `state` component is expected to be a named list with either reactive or
#' "static" values. In most cases, components of `state` will be reactives,
#' but it might make sense in some scenarios to have constructor arguments that
#' are not exposed via UI components but are fixed at construction time. An
#' example for this could be the `dataset_block` implementation where we have
#' constructor arguments `dataset` and `package`, but only expose `dataset`
#' as UI element. This means that `package` is fixed at construction time.
#' Nevertheless, `package` is required as state component, as this is used for
#' re-creating blocks from saved state.
#'
#' State component names are required to match block constructor arguments and
#' re-creating saved objects basically calls the block constructor with values
#' obtained from block state.
#'
#' @section UI:
#' Block UI is generated using the function passed as `ui` to the `new_block`
#' constructor. This function is required to take a single argument `id` and
#' shiny UI components have to be namespaced such that they are nested within
#' this ID (i.e. by creating IDs as `shiny::NS(id, "some_value")`). Some care
#' has to be taken to properly initialize inputs with constructor values. As a
#' rule of thumb, input elements exposed to the UI should have corresponding
#' block constructor arguments such that blocks can be created with a given
#' initial state.
#'
#' Block UI should be limited to displaying and arranging user inputs to set
#' block arguments. For outputs, use generics [block_output()] and
#' [block_ui()].
#'
#' @section Sub-classing:
#' In addition to the specific class of a block, the core package uses virtual
#' classes to group together blocks with similar behavior (e.g.
#' `transform_block`) and makes use of this inheritance structure in S3
#' dispatch for methods like [block_output()] and [block_ui()]. This pattern is
#' not required but encouraged.
#'
#' @section Initialization/evaluation:
#' Some control over when a block is considered "ready for evaluation" is
#' available via arguments `dat_valid` and `allow_empty_state`. Data input
#' validation can optionally be performed by passing a predicate function with
#' the same arguments as in the server function (not including `id`) and the
#' block expression will not be evaluated as long as this function throws an
#' error.
#'
#' Ahead of validation, a block must have its inputs available: every required
#' *data* input connected to a ready upstream (variadic blocks need at least the
#' `...args` minimum declared via `allow_empty_state`) and every required *user*
#' input (`state`) provided. Until then -- including while an upstream is itself
#' still pending -- neither the validator nor the block expression run and no
#' output is produced. See [block_server()] for the resulting eval status
#' (`dormant` / `waiting` / `unset` / `failed` / `ready`).
#'
#' Other conditions (messages and warnings) may be thrown as will be caught
#' and displayed to the user but they will not interrupt evaluation. Errors
#' are safe in that they will be caught as well but the will interrupt
#' evaluation as long as block data input does not satisfy validation.
#'
#' @param server A function returning [shiny::moduleServer()]
#' @param ui A function with a single argument (`ns`) returning a `shiny.tag`
#' @param class Block subclass
#' @param ctor String-valued constructor name or function/frame number (mostly
#'   for internal use or when defining constructors for virtual classes)
#' @param ctor_pkg String-valued package name when passing a string-valued
#'   constructor name or `NULL`
#' @param dat_valid (Optional) input data validator
#' @param block_name Block name
#' @param allow_empty_state Relaxes the block's readiness requirements. By
#' default every data input must be connected and a variadic block needs at
#' least one `...args` input. As `TRUE`, `FALSE` or a character vector of
#' `state` names it relaxes *user* inputs (which `state` values may be empty).
#' A `list(input = ..., data = ...)` also relaxes *data* inputs: `input` takes
#' the same `TRUE`/`FALSE`/character form, while `data` names non-variadic data
#' inputs that may stay unconnected and, via a `...args` entry, overrides the
#' required number of variadic inputs, e.g.
#' `list(input = "n", data = list("y", ...args = 2))`
#' @param expr_type Expression type (experimental)
#' @param external_ctrl Set up external control (experimental)
#' @param block_metadata Block metadata
#' @param ... Further (metadata) attributes
#'
#' @examples
#' new_identity_block <- function() {
#'   new_transform_block(
#'     function(id, data) {
#'       moduleServer(
#'         id,
#'         function(input, output, session) {
#'           list(
#'             expr = reactive(quote(identity(data))),
#'             state = list()
#'           )
#'         }
#'       )
#'     },
#'     function(id) {
#'       htmltools::tagList()
#'     },
#'     class = "identity_block"
#'   )
#' }
#'
#' blk <- new_identity_block()
#' is_block(blk)
#'
#' @return Both `new_block()` and `as_block()` return an object inheriting from
#' `block`, while `is_block()` returns a boolean indicating whether an object
#' inherits from `block` or not. Block vectors, created using `blocks()`,
#' `as_blocks()`, or by combining multiple blocks using [base::c()] all inherit
#' from `blocks` and `is_blocks()` returns a boolean indicating whether an
#' object
#' inherits from `blocks` or not.
#'
#' @export
new_block <- function(server, ui, class, ctor = sys.parent(), ctor_pkg = NULL,
                      dat_valid = NULL, allow_empty_state = FALSE,
                      block_name = default_block_name,
                      expr_type = c("quoted", "bquoted"),
                      external_ctrl = FALSE, block_metadata = NULL, ...) {

  stopifnot(
    is.character(class), length(class) > 0L,
    is_bool(external_ctrl) || is.character(external_ctrl)
  )

  expr_type <- match.arg(expr_type)

  if (missing(ui)) {
    ui <- function(id) {
      tagList()
    }
  }

  if (is.null(block_name)) {
    block_name <- default_block_name
  }

  class <- c(class, "block")

  if (is.function(block_name)) {
    block_name <- block_name(class)
  }

  stopifnot(is_string(block_name))

  if (is.null(block_metadata) || isFALSE(block_metadata)) {

    uid <- registry_id_from_block(class)

    if (length(uid)) {

      block_metadata <- lapply(
        set_names(nm = block_record_fields()),
        get_attr,
        get_registry_entry(uid)
      )

      block_metadata <- c(list(id = uid), block_metadata)

    } else if (!isFALSE(block_metadata)) {
      blockr_warn(
        "No block metadata available for block {class[1L]}.",
        class = "missing_block_metadata"
      )
    }
  }

  validate_block(
    new_vctr(
      list(
        expr_server = server,
        expr_ui = ui,
        dat_valid = dat_valid
      ),
      ...,
      ctor = resolve_ctor(ctor, ctor_pkg),
      block_name = block_name,
      allow_empty_state = allow_empty_state,
      expr_type = expr_type,
      external_ctrl = external_ctrl,
      block_metadata = block_metadata,
      class = class
    ),
    ui_eval = TRUE
  )
}

static_block_arguments <- function() {
  c(
    "server",
    "ui",
    "class",
    "dat_valid",
    "allow_empty_state",
    "expr_type",
    "external_ctrl",
    "block_metadata"
  )
}

internal_block_attributes <- function() {
  c(
    "ctor",
    "class",
    "allow_empty_state",
    "expr_type",
    "external_ctrl",
    "block_metadata"
  )
}

#' @rdname new_block
#' @export
default_block_name <- function(class) {
  res <- gsub("_", " ", sub("_block$", "", class[1L]))
  paste0(toupper(substr(res, 1L, 1L)), substring(res, 2L))
}

validate_block_server <- function(server) {

  if (!is.function(server)) {
    blockr_abort(
      "A block server component is expected to be a function.",
      class = "block_server_function_invalid"
    )
  }

  args <- names(formals(server))

  if (!identical(args[1L], "id")) {
    blockr_abort(
      "A block server function is expected to have an argument `id` in ",
      "first postion.",
      class = "block_server_arg_id_invalid"
    )
  }

  if ("..." %in% args) {
    blockr_abort(
      "Variadic blocks are supported not by passing arguments as `...` ",
      "but using a special argument `...args`.",
      class = "block_server_dots_invalid"
    )
  }

  invisible(server)
}

validate_block_ui <- function(ui) {

  if (!is.function(ui)) {
    blockr_abort(
      "A block UI component is expected to be a function.",
      class = "block_ui_function_invalid"
    )
  }

  if (!identical(names(formals(ui)), "id")) {
    blockr_abort(
      "A block UI function is expected to have a single argument `id`.",
      class = "block_ui_arg_id_invalid"
    )
  }

  invisible(ui)
}

validate_data_validator <- function(validator, server) {

  server_args <- setdiff(names(formals(server)), "id")

  if (!length(server_args) && not_null(validator)) {
    blockr_abort(
      "A nullary server function cannot accompany a data input validator.",
      class = "block_validator_nullary_invalid"
    )
  }

  if (not_null(validator)) {

    if (!is.function(validator)) {
      blockr_abort(
        "Data input validator is expected to be a function.",
        class = "block_validator_function_invalid"
      )
    }

    val_args <- names(formals(validator))

    if (!identical(server_args, val_args)) {
      blockr_abort(
        "Server args {server_args} do not match validator args {val_args}.",
        class = "block_validator_args_invalid"
      )
    }
  }

  invisible(validator)
}

validate_allow_empty_state <- function(x) {

  spec <- attr(x, "allow_empty_state")

  if (!is.list(spec)) {
    return(invisible(x))
  }

  nms <- names(spec)

  if (length(spec) &&
        (is.null(nms) || any(!nzchar(nms)) ||
           length(setdiff(nms, c("input", "data"))))) {
    blockr_abort(
      "A list-valued `allow_empty_state` may only contain `input` and ",
      "`data` entries.",
      class = "allow_empty_state_invalid"
    )
  }

  data <- spec[["data"]]

  if (is.null(data)) {
    return(invisible(x))
  }

  dnms <- coal(names(data), character(length(data)))
  is_args <- dnms == "...args"

  if (sum(is_args) > 1L) {
    blockr_abort(
      "`allow_empty_state$data` accepts at most one `...args` entry.",
      class = "allow_empty_state_invalid"
    )
  }

  if (any(is_args)) {

    if (!is.na(block_arity(x))) {
      blockr_abort(
        "A `...args` entry in `allow_empty_state$data` is only valid for a ",
        "variadic block (one with a `...args` server argument).",
        class = "allow_empty_state_invalid"
      )
    }

    if (!is_count(data[[which(is_args)]])) {
      blockr_abort(
        "The `...args` entry in `allow_empty_state$data` is expected to be a ",
        "single non-negative number.",
        class = "allow_empty_state_invalid"
      )
    }
  }

  unknown <- setdiff(unlst(data[!is_args]), block_inputs(x))

  if (length(unknown)) {
    blockr_abort(
      "`allow_empty_state$data` names unknown block inputs {unknown}.",
      class = "allow_empty_state_invalid"
    )
  }

  invisible(x)
}

validate_block <- function(x, ui_eval = FALSE) {

  if (!is_block(x)) {
    blockr_abort(
      "Expecting blocks to inherit from \"block\".",
      class = "block_class_invalid"
    )
  }

  if (class(x)[1L] == "block") {
    blockr_abort(
      "Expecting blocks to have at least one sub-class.",
      class = "block_class_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting blocks to behave list-like.",
      class = "block_list_like_invalid"
    )
  }

  ctor <- block_ctor(x)

  if (!is_blockr_ctor(ctor)) {
    blockr_abort(
      "Expecting a block constructor to inherit from `blockr_ctor`.",
      class = "block_ctor_invalid"
    )
  }

  if (!is_string(block_name(x))) {
    blockr_abort(
      "Expecting a string-valued block name.",
      class = "block_name_invalid"
    )
  }

  srv <- block_expr_server(x)

  validate_block_server(srv)
  validate_block_ui(block_expr_ui(x))
  validate_data_validator(block_data_validator(x), srv)
  validate_allow_empty_state(x)

  if (isTRUE(ui_eval)) {

    ui <- expr_ui("block", x)

    if (!inherits(ui, "shiny.tag", "shiny.tag.list", agg = any)) {
      blockr_abort(
        "A block UI function is expected to return a shiny UI object, i.e. ",
        "either a `shiny.tag` or a `shiny.tag.list`.",
        class = "block_ui_eval_invalid"
      )
    }
  }

  x
}

#' @param x An object inheriting from `"block"`
#' @rdname new_block
#' @export
is_block <- function(x) {
  inherits(x, "block")
}

#' @rdname new_block
#' @export
as_block <- function(x, ...) {
  UseMethod("as_block")
}

#' @export
as_block.block <- function(x, ...) {
  validate_block(x)
}

#' @export
c.block <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_blocks),
    recursive = FALSE
  )

  as_blocks(res)
}

#' Block utilities
#'
#' Several utilities for working (and manipulating) `block` objects are exported
#' and developers are encouraged to use these instead of relying on object
#' implementation to extract or modify attributes. If functionality for working
#' with blocks in lacking, please consider opening an
#' [issue](https://github.com/BristolMyersSquibb/blockr.core/issues/new).
#'
#' @section Block name:
#' Each block can have a name (by default constructed from the class vector)
#' intended for users to easily identify different blocks. This name can freely
#' be changed during the lifetime of a block and no uniqueness restrictions are
#' in place. The current block name can be retrieved with `block_name()` and
#' set as `block_name(x) <- "some name"`.
#'
#' @section Input validation:
#' Data input validation is available via `validate_data_inputs()` which uses
#' the (optional) validator function passed to [new_block()] at construction
#' time. This mechanism can be used to prevent premature evaluation of the
#' block expression as this might lead to unexpected errors.
#'
#' @section Block arity/inputs:
#' The set of explicit (named) data inputs for a block is available as
#' `block_inputs()`, while the block arity can be queried with `block_arity()`.
#' In case of variadic blocks (i.e. blocks that take a variable number of
#' inputs like for example a block providing [base::rbind()]-like
#' functionality), `block_arity()` returns `NA` and the special block server
#' function argument `...args`, signalling variadic behavior is stripped from
#' `block_inputs()`.
#'
#' @section External control:
#' Blocks can expose constructor inputs for programmatic control from outside
#' the block server (see the `external_ctrl` argument to [new_block()] and the
#' [ctrl_block()] plugin). `external_ctrl_vars()` is a generic that resolves
#' this declaration into the concrete set of controllable variable names:
#' `TRUE` expands to all constructor inputs, `FALSE` to none and a character
#' vector is taken as a (validated) subset. For blocks, `"block_name"` is
#' always included as every block can be renamed. The predicate
#' `has_external_ctrl()` reports whether this set is non-empty (always `TRUE`
#' for blocks).
#'
#' @param x An object inheriting from `"block"`
#'
#' @examples
#' blk <- new_dataset_block()
#' block_name(blk)
#' block_name(blk) <- "My dataset block"
#' block_name(blk)
#'
#' block_inputs(new_dataset_block())
#' block_arity(new_dataset_block())
#'
#' block_inputs(new_merge_block())
#' block_arity(new_merge_block())
#'
#' block_inputs(new_rbind_block())
#' block_arity(new_rbind_block())
#'
#' external_ctrl_vars(new_dataset_block())
#' has_external_ctrl(new_dataset_block())
#'
#' @return Return types vary among the set of exported utilities:
#' * `block_name()`: string valued block name,
#' * `block_name<-()`: `x` (invisibly),
#' * `validate_data_inputs()`: `NULL` if no validator is set and the result of
#'   the validator function otherwise,
#' * `block_inputs()`: a (possibly empty) character vector of data input names,
#' * `block_arity()`: a scalar integer with `NA` in case of variadic behavior,
#' * `external_ctrl_vars()`: a character vector of externally controllable
#'   variable names (always including `"block_name"` for blocks),
#' * `has_external_ctrl()`: a scalar logical.
#'
#' @export
block_name <- function(x) {
  stopifnot(is_block(x))
  attr(x, "block_name")
}

#' @param value New value
#' @rdname block_name
#' @export
`block_name<-` <- function(x, value) {
  stopifnot(is_block(x), is_string(value))
  attr(x, "block_name") <- value
  invisible(x)
}

block_ctor <- function(x) {
  stopifnot(is_block(x))
  attr(x, "ctor")
}

block_expr_server <- function(x) {
  x[["expr_server"]]
}

block_expr_ui <- function(x) {
  x[["expr_ui"]]
}

block_data_validator <- function(x) {
  x[["dat_valid"]]
}

block_has_data_validator <- function(x) {
  not_null(block_data_validator(x))
}

parse_allow_empty_state <- function(x) {

  empty_state <- FALSE
  optional_inputs <- character()
  min_args <- 1L

  if (is.list(x)) {

    if (not_null(x[["input"]])) {
      empty_state <- x[["input"]]
    }

    data <- x[["data"]]
    nms <- coal(names(data), character(length(data)))
    is_args <- nms == "...args"

    optional_inputs <- unlst(data[!is_args])

    if (any(is_args)) {
      min_args <- as.integer(data[[which(is_args)[1L]]])
    }

  } else if (not_null(x)) {
    empty_state <- x
  }

  if (!length(empty_state)) {
    empty_state <- FALSE
  }

  if (!length(optional_inputs)) {
    optional_inputs <- character()
  }

  list(
    empty_state = empty_state,
    optional_inputs = optional_inputs,
    min_args = min_args
  )
}

block_allow_empty_state <- function(x) {
  parse_allow_empty_state(attr(x, "allow_empty_state"))[["empty_state"]]
}

block_optional_inputs <- function(x) {
  parse_allow_empty_state(attr(x, "allow_empty_state"))[["optional_inputs"]]
}

block_min_args <- function(x) {
  parse_allow_empty_state(attr(x, "allow_empty_state"))[["min_args"]]
}

#' @param data Data input values
#' @rdname block_name
#' @export
validate_data_inputs <- function(x, data) {

  if (block_has_data_validator(x)) {
    return(do.call(block_data_validator(x), data))
  }

  NULL
}

#' @rdname block_name
#' @export
block_inputs <- function(x) {
  UseMethod("block_inputs", x)
}

#' @export
block_inputs.block <- function(x) {
  setdiff(block_expr_inputs(x), "...args")
}

block_expr_inputs <- function(x) {
  setdiff(names(formals(block_expr_server(x))), "id")
}

block_ctor_inputs <- function(x) {
  setdiff(names(formals(block_ctor(x))), "...")
}

initial_block_state <- function(x) {
  lapply(
    set_names(nm = block_ctor_inputs(x)),
    get,
    envir = environment(block_expr_server(x))
  )
}

block_base_attrs <- function() {
  setdiff(
    names(formals(new_block)),
    c("server", "ui", "class", "ctor", "ctor_pkg", "dat_valid",
      "allow_empty_state")
  )
}

#' @rdname block_name
#' @export
block_arity <- function(x) {

  if (is_board(x)) {
    return(
      set_names(int_ply(board_blocks(x), block_arity), board_block_ids(x))
    )
  }

  args <- block_expr_inputs(x)

  if ("...args" %in% args) {
    return(NA_integer_)
  }

  length(args)
}

#' @export
format.block <- function(x, ...) {

  out <- ""

  for (cl in rev(setdiff(class(x), c("list", "vctrs_vctr")))) {
    out <- paste0("<", cl, out, ">")
  }

  out <- c(
    out,
    paste0("Name: \"", attr(x, "block_name"), "\"")
  )

  arity <- block_arity(x)

  if (is.na(arity)) {
    dat <- "Indefinite arity"
  } else if (arity > 0L) {
    dat <- paste0("Data inputs: ", paste_enum(block_inputs(x), quotes = "\""))
  } else {
    dat <- "No data inputs"
  }

  out <- c(out, dat)

  if (length(block_ctor_inputs(x))) {

    args <- initial_block_state(x)
    args <- trimws(utils::capture.output(utils::str(args))[-1L], "right")

    out <- c(out, "Initial block state:", args)

  } else {

    out <- c(out, "Stateless block")
  }

  ctor <- block_ctor(x)

  if (is.null(ctor_name(ctor))) {
    ctor <- "<local function>"
  } else {
    ctor <- paste0(ctor_pkg(ctor), "::", ctor_name(ctor), "()")
  }

  c(out, paste("Constructor:", ctor))
}

#' @export
print.block <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.block <- function(x, ...) {

  out <- paste0("<", class(x)[1L], ">")

  args <- block_ctor_inputs(x)

  if (length(args)) {
    ctrl <- external_ctrl_vars(x)
    args <- ifelse(args %in% ctrl, paste0(args, "*"), args)
    out <- paste0(out, " ", paste0(args, collapse = ", "))
  }

  out
}

#' @importFrom utils str
#' @export
str.block <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
board_options.block <- function(x, ...) {
  new_board_options()
}

block_expr_type <- function(x) {
  stopifnot(is_block(x))
  attr(x, "expr_type")
}

#' @rdname block_name
#' @export
external_ctrl_vars <- function(x) {
  UseMethod("external_ctrl_vars")
}

#' @export
external_ctrl_vars.block <- function(x) {

  res <- attr(x, "external_ctrl")

  user <- if (isTRUE(res)) {
    block_ctor_inputs(x)
  } else if (isFALSE(res)) {
    character()
  } else {
    stopifnot(is.character(res), all(res %in% block_ctor_inputs(x)))
    res
  }

  union(user, "block_name")
}

#' @rdname block_name
#' @export
has_external_ctrl <- function(x) {
  length(external_ctrl_vars(x)) > 0
}

test_that("block constructor", {

  new_identity_block <- function() {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(
              expr = reactive(quote(identity(data))),
              state = list()
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "identity_block",
      block_metadata = list()
    )
  }

  expect_s3_class(
    new_identity_block(),
    c("identity_block", "transform_block", "block")
  )

  expect_error(
    validate_block_server("abc"),
    class = "block_server_function_invalid"
  )

  expect_error(
    validate_block_server(function() {}),
    class = "block_server_arg_id_invalid"
  )

  expect_error(
    validate_block_server(function(id, ...) {}),
    class = "block_server_dots_invalid"
  )

  expect_identical(
    validate_block_server(function(id) {}),
    function(id) {}
  )

  expect_error(
    validate_block_ui("abc"),
    class = "block_ui_function_invalid"
  )

  expect_error(
    validate_block_ui(function() {}),
    class = "block_ui_arg_id_invalid"
  )

  expect_error(
    validate_block_ui(function(abc) {}),
    class = "block_ui_arg_id_invalid"
  )

  expect_error(
    validate_data_validator(function(x) {}, function() {}),
    class = "block_validator_nullary_invalid"
  )

  expect_error(
    validate_data_validator("abc", function(id, x) {}),
    class = "block_validator_function_invalid"
  )

  expect_silent(
    validate_data_validator(NULL, function(id, x) {})
  )

  expect_error(
    validate_data_validator(function(y) {}, function(id, x) {}),
    class = "block_validator_args_invalid"
  )

  expect_error(
    validate_block("abc"),
    class = "block_class_invalid"
  )

  expect_error(
    validate_block(structure("abc", class = "block")),
    class = "block_class_invalid"
  )

  expect_error(
    validate_block(structure("abc", class = c("some_block", "block"))),
    class = "block_list_like_invalid"
  )

  expect_error(
    validate_block(
      structure(list(), ctor = "abc", class = c("some_block", "block"))
    ),
    class = "block_ctor_invalid"
  )

  expect_error(
    validate_block(
      structure(list(), ctor = new_blockr_ctor(function() {}), name = 1,
                class = c("some_block", "block"))
    ),
    class = "block_name_invalid"
  )

  expect_error(
    validate_block(
      structure(
        list(
          expr_server = function(id) {},
          expr_ui = function(id) {},
          dat_valid = NULL
        ),
        ctor = new_blockr_ctor(function() {}),
        block_name = "1",
        class = c("some_block", "block")
      ),
      ui_eval = TRUE
    ),
    class = "block_ui_eval_invalid"
  )
})

test_that("block class", {

  x <- new_dataset_block()

  expect_s3_class(x, "block")
  expect_true(is_block(x))
  expect_false(is_block("x"))

  expect_identical(x, as_block(x))

  expect_snapshot(print(x))
})

test_that("blocks report controllability in str_value()", {

  expect_identical(
    str_value(new_scatter_block(x = "wt", y = "mpg")),
    "<scatter_block> x, y"
  )
  expect_identical(
    str_value(new_subset_block()),
    "<subset_block> subset*, select*"
  )
  expect_identical(
    str_value(new_dataset_block()),
    "<dataset_block> dataset*, package"
  )
  expect_identical(str_value(new_rbind_block()), "<rbind_block>")
  expect_identical(str_value(structure(list(), class = "made_up")), "<made_up>")

  blk <- new_dataset_block()
  out <- capture.output(res <- withVisible(str(blk)))

  expect_identical(out, " <dataset_block> dataset*, package")
  expect_false(res$visible)
  expect_identical(res$value, blk)
})

test_that("block utils", {

  blk <- new_dataset_block()
  lst <- list(blk)

  expect_s3_class(c(blk, blk), "blocks")
  expect_s3_class(c(blk, lst), "blocks")
})

test_that("Without package blocks can print", {
  blk <- new_dummy_block <- function(text = "Hello World", ...) {
    new_data_block(
      function(id) {
        moduleServer(id, function(input, output, session) {
          list(
            expr = reactive(quote(text)),
            state = list(text = text)
          )
        })
      },
      function(id) {
        tagList()
      },
      class = "dummy_block",
      block_metadata = list(),
      ...
    )
  }
  expect_snapshot(blk())
})

test_that("external_ctrl_vars resolves the external_ctrl declaration", {

  expect_identical(
    external_ctrl_vars(new_static_block(mtcars)),
    "block_name"
  )

  expect_setequal(
    external_ctrl_vars(new_subset_block()),
    c("subset", "select", "block_name")
  )

  expect_setequal(
    external_ctrl_vars(new_dataset_block("mtcars")),
    c("dataset", "block_name")
  )

  blk <- new_dataset_block("mtcars")
  attr(blk, "external_ctrl") <- "not_an_input"

  expect_error(external_ctrl_vars(blk))
  expect_error(external_ctrl_vars("not a block"))
})

test_that("has_external_ctrl is TRUE whenever any var is controllable", {

  expect_true(has_external_ctrl(new_static_block(mtcars)))
  expect_true(has_external_ctrl(new_subset_block()))
  expect_true(has_external_ctrl(new_dataset_block("mtcars")))
})

test_that("allow_empty_state parses user- and data-input specs", {

  expect_identical(
    parse_allow_empty_state(FALSE),
    list(empty_state = FALSE, optional_inputs = character(), min_args = 1L)
  )

  expect_identical(
    parse_allow_empty_state(c("a", "b")),
    list(
      empty_state = c("a", "b"),
      optional_inputs = character(),
      min_args = 1L
    )
  )

  expect_identical(
    parse_allow_empty_state(list(input = "n", data = list("y", ...args = 2))),
    list(empty_state = "n", optional_inputs = "y", min_args = 2L)
  )

  expect_identical(
    parse_allow_empty_state(list(data = list(...args = 1))),
    list(empty_state = FALSE, optional_inputs = character(), min_args = 1L)
  )

  expect_identical(
    parse_allow_empty_state(list(input = TRUE)),
    list(empty_state = TRUE, optional_inputs = character(), min_args = 1L)
  )
})

test_that("block accessors expose user- and data-input specs", {

  # rbind declares nothing: variadic minimum defaults to 1
  rb <- new_rbind_block()
  expect_identical(block_min_args(rb), 1L)
  expect_identical(block_allow_empty_state(rb), FALSE)
  expect_identical(block_optional_inputs(rb), character())

  # glue opts out of the default, working with zero data inputs
  expect_identical(block_min_args(new_glue_block()), 0L)
})

test_that("allow_empty_state validates its structured form", {

  expect_error(
    new_head_block(allow_empty_state = list("data")),
    class = "allow_empty_state_invalid"
  )

  expect_error(
    new_head_block(allow_empty_state = list(data = list(...args = 2))),
    class = "allow_empty_state_invalid"
  )

  expect_error(
    new_head_block(allow_empty_state = list(data = list("nope"))),
    class = "allow_empty_state_invalid"
  )

  var_block <- function(aes) {
    new_transform_block(
      function(id, ...args) {
        moduleServer(
          id,
          function(input, output, session) {
            list(expr = reactive(quote(rbind(...args))), state = list())
          }
        )
      },
      function(id) tagList(),
      class = "test_variadic_block",
      allow_empty_state = aes,
      block_metadata = list()
    )
  }

  expect_error(
    var_block(list(data = list(...args = -1))),
    class = "allow_empty_state_invalid"
  )

  expect_s3_class(
    var_block(list(data = list(...args = 2))),
    "test_variadic_block"
  )
})

#' Block server
#'
#' A block is represented by several (nested) shiny modules and the top level
#' module is created using the `block_server()` generic. S3 dispatch is offered
#' as a way to add flexibility, but in most cases the default method for the
#' `block` class should suffice at top level. Further entry points for
#' customization are offered by the generics `expr_server()` and `block_eval()`,
#' which are responsible for initializing the block "expression" module (i.e.
#' the block server function passed in [new_block()]) and block evaluation
#' (evaluating the interpolated expression in the context of input data),
#' respectively.
#'
#' The module returned from `block_server()`, at least in the default
#' implementation, provides much of the essential but block-type agnostic
#' functionality, including data input validation (if available), instantiation
#' of the block expression server (handling the block-specific functionality,
#' i.e. block user inputs and expression), and instantiation of the
#' `edit_block` module (if passed from the parent scope).
#'
#' Each block carries an *eval status* -- one of `dormant`, `waiting`, `unset`,
#' `failed` or `ready` -- which, together with the orthogonal `visible` flag,
#' determines its behaviour. The status separates the two input kinds (data
#' inputs from links, user inputs from `state`) and a genuine failure:
#' * `dormant` -- not *needed* (neither on screen nor feeding, transitively over
#'   [board_links()], an on-screen block); inputs stay unfulfilled
#'   ([shiny::req()] out) and nothing evaluates.
#' * `waiting` -- needed, but a required *data* input is missing: unconnected,
#'   below the required number of variadic `...args` inputs (one by default),
#'   or fed by an upstream block that is not itself `ready` (see
#'   `allow_empty_state`).
#' * `unset` -- data inputs are ready, but a required *user* input (`state`
#'   value) has not been provided (unless permitted by `allow_empty_state`).
#' * `failed` -- all inputs are present, but the block cannot produce a result:
#'   the data validator ([validate_data_inputs()]) or the block expression
#'   raised. The offending condition is surfaced through the block conditions.
#' * `ready` -- evaluation succeeded and a result (possibly a legitimate `NULL`)
#'   is available for downstream blocks to consume.
#'
#' A block reaches `ready` only once its upstreams have, so an unconnected or
#' pending block holds its whole downstream chain `waiting` without any of them
#' evaluating against missing data. Output rendering follows the status: the
#' block output is shown only while `ready` and cleared otherwise, so a block
#' leaving `ready` never displays a stale result. While not `ready` the block
#' surfaces a condition explaining why -- a `status`-phase note for `waiting`
#' and `unset`, or the raised error for `failed`. Conditions raised during
#' validation and evaluation are caught and returned to be surfaced to the app
#' user.
#'
#' Block-level user inputs (provided by the expression module) are separated
#' from output, the behavior of which can be customized via the
#' [block_output()] generic. The [block_ui()] generic can then be used to
#' control rendering of outputs.
#'
#' When a front-end (such as blockr.dock) drives the `visible` write-channel
#' that [board_server()] hands to the board callback, reporting each block's
#' visibility status (off screen, on screen, or rendered into its view),
#' evaluation and rendering are gated on visibility.
#' Rendering is gated on plain visibility: the render observer is suspended
#' while a block is off screen and resumed once it is on screen, starting
#' suspended so nothing renders before the front-end first reports. Evaluation
#' is gated on the *needed* set, the on-screen blocks together with their
#' upstream closure over [board_links()] (derived from `visible`, recomputed
#' only when it or the links change). A block's input data reactives stay
#' unfulfilled (they [shiny::req()] out) unless the block is needed, so a block
#' that is neither visible nor feeding a visible block pulls no input and stays
#' fully quiescent: its result reactive, and any observer its expression server
#' registers on the incoming data, all short-circuit and do nothing. A needed
#' but off-screen block (one feeding a visible block) evaluates but does not
#' render. Block-server *construction* is prioritized the same way: the needed
#' set is instantiated first so that first paint waits only for the on-screen
#' blocks and their upstreams, and the remaining block servers are built
#' progressively in the background. That background pass holds until the
#' front-end reports every on-screen block as rendered (arranged into its
#' view), so it never competes with first paint. Until a block is built it is
#' absent from
#' the `board$blocks` handed to plugins and callbacks, which simply see it
#' appear once constructed. The background cadence is set by the
#' `background_construction_delay` [blockr_option()] (milliseconds between
#' successive blocks, default 50); a value of 0 disables the staggering and
#' builds every block up front. With nothing driving
#' `visible` every block is needed and behaviour is unchanged; the
#' `gate_visibility` [blockr_option()] (default `TRUE`) turns gating off
#' entirely.
#'
#' @param id Namespace ID
#' @param x Object for which to generate a [shiny::moduleServer()]
#' @param data Input data (list of reactives)
#' @param ... Generic consistency
#'
#' @return Both `block_server()` and `expr_server()` return shiny server module
#' (i.e. a call to [shiny::moduleServer()]), while `block_eval()` evaluates
#' an interpolated (w.r.t. block "user" inputs) block expression in the context
#' of block data inputs.
#'
#' @export
block_server <- function(id, x, data = list(), ...) {
  UseMethod("block_server", x)
}

#' @param block_id Block ID
#' @param edit_block,ctrl_block Block plugins
#' @param board Reactive values object containing board information
#' @param update Reactive value object to initiate board updates
#' @param inputs_ready Reactive flag signaling whether the block's required
#' inputs are all connected to ready upstream blocks (supplied by
#' [board_server()]; defaults to always-ready when a block server is run
#' standalone)
#' @rdname block_server
#' @export
block_server.block <- function(id, x, data = list(), block_id = id,
                               edit_block = NULL, ctrl_block = NULL,
                               board = reactiveValues(),
                               update = reactiveVal(),
                               inputs_ready = reactive(TRUE), ...) {

  dot_args <- list(...)

  moduleServer(
    id,
    function(input, output, session) {

      cond <- reactiveValues(
        data = NULL,
        state = NULL,
        eval = NULL,
        render = NULL,
        block = NULL,
        status = NULL
      )

      exp <- check_expr_val(
        expr_server(x, data),
        x
      )

      lang <- reactive(
        exprs_to_lang(exp$expr())
      )

      state <- exp$state

      dat <- reactive(
        {
          res <- lapply(data[names(data) != "...args"], reval)

          if ("...args" %in% names(data)) {
            res <- c(res, list(`...args` = dot_arg_values(data[["...args"]])))
          }

          res
        }
      )

      data_valid <- validate_block_reactive(block_id, x, dat, cond, session,
                                            inputs_ready)

      state_ready <- state_ready_reactive(block_id, x, state, session)

      dat_eval <- reactive(
        {
          req(isTRUE(data_valid()), isTRUE(state_ready()))
          lapply(state, reval_if)
          try(lang(), silent = TRUE)

          res <- dat()

          if ("...args" %in% names(res)) {
            res <- c(res[names(res) != "...args"], res[["...args"]])
          }

          block_eval_trigger(x, session)

          res
        },
        domain = session
      )

      cur_name <- reactiveVal(block_name(x))

      if (is_board(isolate(board$board))) {

        reg_name <- reactive(
          {
            blk <- board_blocks(board$board)[[block_id]]
            if (is_block(blk)) block_name(blk) else NULL
          }
        )

        observeEvent(
          cur_name(),
          {
            new_name <- cur_name()
            if (!identical(reg_name(), new_name)) {
              update(
                list(
                  blocks = list(
                    mod = set_names(
                      list(list(block_name = new_name)),
                      block_id
                    )
                  )
                )
              )
            }
          },
          ignoreInit = TRUE
        )

        observeEvent(
          reg_name(),
          {
            if (!identical(cur_name(), reg_name())) {
              cur_name(reg_name())
            }
          },
          ignoreInit = TRUE
        )
      }

      ctrl_vars <- c(
        state[setdiff(external_ctrl_vars(x), "block_name")],
        list(block_name = cur_name)
      )

      if (!all(lgl_ply(ctrl_vars, inherits, "reactiveVal"))) {
        blockr_abort(
          "All externally controllable variables for {class(x)[1L]} are ",
          "expected to inherit from `reactiveVal`.",
          class = "unsupported_external_ctrl_variable"
        )
      }

      cb_res <- coal(
        call_plugin_server(
          ctrl_block,
          list(
            x = x,
            vars = ctrl_vars,
            data = dat_eval,
            eval = reactive(eval_impl(x, lang(), dat_eval()))
          )
        ),
        TRUE
      )

      gate <- cb_res

      res <- reactive(
        {
          if (!isTRUE(reval_if(gate)) || !isTRUE(data_valid()) ||
                !isTRUE(state_ready())) {
            return(NULL)
          }

          eval_data <- dat_eval()

          log_debug("evaluating block ", block_id)

          isolate(
            capture_conditions(
              eval_impl(x, lang(), eval_data),
              cond,
              "eval",
              session = session
            )
          )
        },
        domain = session
      )

      failed <- reactive(
        {
          if (!inputs_ready() || !isTRUE(state_ready())) {
            return(FALSE)
          }

          if (!isTRUE(data_valid())) {
            return(TRUE)
          }

          res()

          length(cond$eval$error) > 0L
        },
        domain = session
      )

      block_ready <- reactive(
        isTRUE(reval_if(gate)) && inputs_ready() && isTRUE(state_ready()) &&
          !failed()
      )

      gated <- is_board(isolate(board$board)) &&
        isTRUE(blockr_option("gate_visibility", TRUE))

      render_obs <- output_render_observer(x, block_ready, inputs_ready,
                                           state_ready, res, cond, session,
                                           suspended = gated)

      if (gated) {
        render_gate_observer(block_id, board, render_obs, session)
      }

      eb_res <- call_plugin_server(
        edit_block,
        server_args = c(
          list(block_id = block_id, board = board, update = update),
          dot_args
        )
      )

      if ("cond" %in% names(exp)) {

        blk_cnd <- reactive(
          {
            include <- coal(
              get_board_option_or_null("show_conditions", session),
              match.arg(
                blockr_option("show_conditions", c("warning", "error")),
                c("message", "warning", "error"),
                several.ok = TRUE
              )
            )
            res <- reactiveValuesToList(exp[["cond"]])[include]
            set_names(lapply(res, coal, list()), include)
          }
        )

        block_cond_observer(blk_cnd, cond, session)
      }

      conditions <- reactive(
        blk_cnds(reactiveValuesToList(cond), block_id)
      )

      c(
        list(
          result = res,
          state_ready = state_ready,
          failed = failed,
          expr = lang,
          state = state,
          conditions = conditions
        ),
        eb_res
      )
    }
  )
}

#' @rdname block_server
#' @export
expr_server <- function(x, data, ...) {
  UseMethod("expr_server")
}

#' @export
expr_server.block <- function(x, data, ...) {
  do.call(block_expr_server(x), c(list(id = "expr"), data))
}

validate_block_reactive <- function(id, x, dat, cond, sess, inputs_ready) {

  has_validator <- block_has_data_validator(x)

  reactive(
    {
      if (!inputs_ready()) {

        isolate(clear_data_conditions(cond))

        return(FALSE)
      }

      if (!has_validator) {

        isolate(clear_data_conditions(cond))

        return(TRUE)
      }

      log_debug("performing input validation for block ", id)

      isolate(
        capture_conditions(
          {
            validate_data_inputs(x, dat())
            TRUE
          },
          cond,
          "data",
          session = sess
        )
      )
    },
    domain = sess
  )
}

clear_data_conditions <- function(cond) {

  if (any(lengths(cond$data))) {
    cond$data <- empty_block_condition()
  }
}

state_ready_reactive <- function(id, x, state, sess) {

  reactive(
    {
      log_debug("checking returned state values of block ", id)

      allow_empty <- block_allow_empty_state(x)

      if (isTRUE(allow_empty) || !length(state)) {
        return(TRUE)
      }

      check <- if (isFALSE(allow_empty)) {
        TRUE
      } else {
        setdiff(names(state), allow_empty)
      }

      all(lgl_ply(lapply(state[check], reval_if), Negate(is_empty)))
    },
    domain = sess
  )
}

eval_impl <- function(x, expr, dat) {

  if (identical(block_expr_type(x), "bquoted")) {
    expr <- do.call(
      bquote,
      list(
        expr,
        lapply(set_names(nm = names(dat)), as.name),
        splice = is.na(block_arity(x))
      )
    )
  }

  block_eval(x, expr, eval_env(dat))
}

#' @rdname block_server
#' @export
block_render_trigger <- function(x, session = get_session()) {
  UseMethod("block_render_trigger", x)
}

#' @export
block_render_trigger.block <- function(x, session = get_session()) {
  NULL
}

output_render_observer <- function(x, ready, inputs_ready, state_ready, res,
                                   cond, sess, suspended = FALSE) {

  observe(
    {
      block_render_trigger(x, sess)

      if (isTRUE(ready())) {

        sess$output$result <- capture_conditions(
          block_output(x, res(), sess),
          cond,
          "render",
          session = sess
        )

        isolate(explain_block_status(cond, NULL))

      } else {

        sess$output$result <- NULL

        reason <- if (!inputs_ready()) {
          "This block is waiting for its data input to be connected."
        } else if (!isTRUE(state_ready())) {
          "This block is waiting for its inputs to be set."
        } else {
          NULL
        }

        isolate(explain_block_status(cond, reason))
      }
    },
    domain = sess,
    suspended = suspended
  )
}

explain_block_status <- function(cond, reason) {

  new <- empty_block_condition()

  if (not_null(reason)) {
    new[["warning"]] <- list(new_blk_cnd(reason))
  }

  if (!identical(cond$status, new)) {
    cond$status <- new
  }
}

render_gate_observer <- function(id, board, render_obs, sess) {

  prev_render <- NA

  observe(
    {
      do_render <- isTRUE(board$visible) ||
        isTRUE(board$visible[id] %in% c("required", "rendered"))

      if (do_render) render_obs$resume() else render_obs$suspend()

      if (!identical(do_render, prev_render)) {
        prev_render <<- do_render
        log_debug(
          "block {id} rendering ",
          "{if (do_render) 'resumed' else 'suspended'}"
        )
      }
    },
    domain = sess
  )
}

block_cond_observer <- function(blk, cond, sess) {

  observeEvent(
    blk(),
    {
      new_cnds <- blk()
      cur_cnds <- set_names(
        coal(cond$block, empty_block_condition())[names(new_cnds)],
        names(new_cnds)
      )

      if (any(lengths(new_cnds))) {

        new_cnds <- lapply(new_cnds, lapply, as_blk_cnd)

        chk <- lgl_mply(
          Negate(setequal),
          lapply(new_cnds, chr_ply, attr, "id"),
          lapply(cur_cnds, chr_ply, attr, "id")
        )

        if (any(chk)) {
          cond$block <- new_cnds
        }

      } else if (any(lengths(cur_cnds) > 0L)) {
        cond$block <- empty_block_condition()[names(new_cnds)]
      }
    }
  )
}

check_expr_val <- function(val, x) {

  if (!is.list(val)) {
    blockr_abort(
      "The block server for {class(x)[1L]} is expected to return a list.",
      class = "expr_server_return_type_invalid"
    )
  }

  required <- c("expr", "state")

  if (!all(required %in% names(val))) {
    blockr_abort(
      "The block server for {class(x)[1L]} is expected to return values ",
      "{setdiff(required, names(val))}.",
      class = "expr_server_return_required_component_missing"
    )
  }

  if (!is.reactive(val[["expr"]])) {
    blockr_abort(
      "The `expr` component of the return value for {class(x)[1L]} is ",
      "expected to be a reactive.",
      class = "expr_server_return_expr_invalid"
    )
  }

  if (!is.list(val[["state"]])) {
    blockr_abort(
      "The `state` component of the return value for {class(x)[1L]} is ",
      "expected to be a list.",
      class = "expr_server_return_state_type_invalid"
    )
  }

  expected <- block_ctor_inputs(x)
  current <- names(val[["state"]])
  missing <- setdiff(expected, current)

  if (length(missing)) {
    blockr_abort(
      "The `state` component of the return value for {class(x)[1L]} is ",
      "expected to additionally return {missing}.",
      class = "expr_server_return_state_missing_component"
    )
  }

  disallowed <- intersect(current, static_block_arguments())

  if (length(disallowed)) {
    blockr_abort(
      "The `state` component of the return value for {class(x)[1L]} is ",
      "is not allowed to return components {disallowed}.",
      class = "expr_server_return_state_invalid_component"
    )
  }

  if ("cond" %in% names(val)) {

    if (!is.reactivevalues(val[["cond"]])) {
      blockr_abort(
        "The `cond` component of the return value for {class(x)[1L]} is ",
        "expected to be a `reactiveValues` object.",
        class = "expr_server_return_cond_invalid"
      )
    }

    conds <- c("message", "warning", "error")

    if (!all(names(val[["cond"]]) %in% conds)) {
      blockr_abort(
        "The `cond` component of the return value for {class(x)[1L]} ",
        "is expected to contain any of components {conds}.",
        class = "expr_server_return_cond_invalid"
      )
    }

    lapply(
      conds,
      function(cnd, cnds) {
        observeEvent(
          req(length(cnds[[cnd]]) > 0L),
          {
            for (y in cnds[[cnd]]) {
              if (!(is.character(y) || is_list_of_blk_cnds(y))) {
                blockr_abort(
                  "The `cond` component of the return value for ",
                  "{class(x)[1L]} is expected to contain a nested list of ",
                  "character vectors or list of objects inheriting from ",
                  "`blk_cnd`.",
                  class = "expr_server_return_cond_invalid"
                )
              }
            }
          },
          once = TRUE
        )
      },
      val[["cond"]]
    )
  }

  val
}

#' @section File browser block:
#' In order to make user data available to blockr, this block provides file-
#' upload functionality via [shiny::fileInput()]. Given that data provided in
#' this way are only available for the life-time of the shiny session, exported
#' code is not self-contained and a script containing code from an upload block
#' is cannot be run in a new session. Also, serialization of upload blocks is
#' currently not allowed as the full data would have to be included during
#' serialization.
#'
#' @param file_path File path
#' @param volumes Parent namespace
#'
#' @block file browser block
#' @blockDescr Browse local files
#' @blockCategory input
#' @blockIcon folder2-open
#' @blockGuidance Selects a file from the server's file system for a downstream
#'   parser such as a CSV block. `file_path` is the chosen path; `volumes`
#'   bounds the root directories the browser may reach.
#' @blockKeywords file, browser, path, filesystem, select
#' @blockArg file_path Path of the selected file.
#'   [type] arg_string()
#'
#' @rdname new_file_block
#' @export
new_filebrowser_block <- function(file_path = character(),
                                  volumes = filebrowser_volumes(),
                                  ...) {
  new_file_block(
    function(id) {
      moduleServer(
        id,
        function(input, output, session) {

          path <- reactiveVal(file_path)
          file <- reactive(
            shinyFiles::parseFilePaths(volumes, input$file)$datapath
          )

          shinyFiles::shinyFileChoose(input, "file", roots = volumes)

          observeEvent(
            file(),
            {
              req(file())
              path(unname(file()))
            }
          )

          list(
            expr = reactive(
              bquote(.(file), list(file = path()))
            ),
            state = list(
              file_path = path,
              volumes = volumes
            )
          )
        }
      )
    },
    function(id) {
      shinyFiles::shinyFilesButton(
        NS(id, "file"),
        label = "File select",
        title = "Please select a file",
        multiple = FALSE
      )
    },
    class = "filebrowser_block",
    ...
  )
}

#' @param default Default volumes specification (use the blockr option
#' "volumes" to override)
#' @rdname new_file_block
#' @export
filebrowser_volumes <- function(default = c(home = path.expand("~"))) {

  res <- blockr_option("volumes", default)

  if (is_string(res) && grepl(":", res, fixed = TRUE)) {
    res <- strsplit(res, ":", fixed = TRUE)[[1L]]
  }

  if (is.null(names(res))) {
    if (length(res) == 1L) {
      names(res) <- "volume"
    } else if (length(res) > 1L) {
      names(res) <- paste0("volume", seq_along(res))
    }
  }

  stopifnot(
    is.character(res),
    length(res) >= 1L,
    length(unique(names(res))) == length(res)
  )

  res
}

skip_on_os("windows")

test_that("filebrowser block constructor", {

  file <- normalizePath(withr::local_tempfile(lines = "abc"))

  blk <- new_filebrowser_block(volumes = c(tmp = dirname(file)))

  expect_s3_class(blk, "filebrowser_block")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$makeScope("expr")$setInputs(
        file = list(files = basename(file), root = "tmp")
      )
      session$flushReact()
      expect_identical(
        session$returned$result(),
        file
      )
    },
    args = list(x = blk)
  )

  vols <- filebrowser_volumes()

  expect_type(vols, "character")
  expect_named(vols, "home")

  vols <- filebrowser_volumes(file.path("foo", "bar"))

  expect_type(vols, "character")
  expect_named(vols, "volume")

  vols <- filebrowser_volumes("foo/bar:foo/baz")

  expect_type(vols, "character")
  expect_named(vols, c("volume1", "volume2"))
})

#' Assertions
#'
#' Utility functions, mainly intended for asserting common preconditions are
#' exported for convenience in dependent packages.
#'
#' @param x Object to check
#'
#' @return Scalar logical value.
#'
#' @rdname assertions
#' @export
is_scalar <- function(x) length(x) == 1L

#' @rdname assertions
#' @export
is_string <- function(x) {
  is.character(x) && is_scalar(x)
}

#' @rdname assertions
#' @export
is_bool <- function(x) {
  is_scalar(x) && (isTRUE(x) || isFALSE(x))
}

#' @rdname assertions
#' @export
is_intish <- function(x) {
  is.integer(x) || (is.numeric(x) && all(x == trunc(x)) && !is.na(x))
}

#' @param allow_zero Determines whether the value 0 is considered a valid count
#' @rdname assertions
#' @export
is_count <- function(x, allow_zero = TRUE) {

  if (!is_scalar(x)) {
    return(FALSE)
  }

  if (!is_intish(x)) {
    return(FALSE)
  }

  if (isTRUE(allow_zero)) {
    x >= 0 && !is.na(x)
  } else {
    x > 0 && !is.na(x)
  }
}

#' @rdname assertions
#' @export
is_number <- function(x) {
  is.numeric(x) && is_scalar(x) && !is.na(x) && !is.nan(x) && is.finite(x)
}

#' @param ... Silently ignored
#' @rdname assertions
#' @export
not_null <- Negate(is.null)

#' @rdname assertions
#' @export
has_length <- function(x) length(x) > 0

is_zero_len <- function(x) {
  length(x) == 0L
}

all_zero_len <- function(x) {
  if (is.list(x)) {
    all(lgl_ply(x, all_zero_len))
  } else {
    is_zero_len(x)
  }
}

is_empty <- function(x) {
  is_zero_len(x) || all(is.na(x) | !nchar(x))
}

filter_all_zero_len <- function(x) {
  if (all_zero_len(x)) {
    NULL
  } else if (is.list(x)) {
    Filter(Negate(is.null), lapply(x, filter_all_zero_len))
  } else {
    x
  }
}

filter_empty <- function(x) Filter(Negate(is_empty), x)

test_that("assertions", {

  expect_true(is_bool(TRUE))
  expect_true(is_bool(FALSE))

  expect_false(is_bool(0L))
  expect_false(is_bool(1))
  expect_false(is_bool(NA))
  expect_false(is_bool(list()))
  expect_false(is_bool(NULL))
  expect_false(is_bool(c(TRUE, FALSE)))

  expect_true(is_count(1))
  expect_true(is_count(1L))

  expect_false(is_count(1.5))
  expect_false(is_count(-1L))
  expect_false(is_count(1L:3L))
  expect_false(is_count(0L, allow_zero = FALSE))

  expect_true(is_number(1))
  expect_true(is_number(1L))
  expect_true(is_number(1.5))
  expect_true(is_number(-1L))

  expect_false(is_number(1:3))
  expect_false(is_number(NaN))
  expect_false(is_number(Inf))
  expect_false(is_number(NA_integer_))
})

test_that("zero length", {

  x <- list()

  expect_true(all_zero_len(x))
  expect_null(filter_all_zero_len(x))

  x <- list(1)

  expect_false(all_zero_len(x))
  expect_identical(x, filter_all_zero_len(x))

  x <- list(
    list(
      a = list(),
      b = NULL,
      c = character()
    )
  )

  expect_true(all_zero_len(x))
  expect_null(filter_all_zero_len(x))

  x <- list(
    list(
      a = list("foo"),
      b = NULL,
      c = character()
    )
  )

  expect_false(all_zero_len(x))
  expect_identical(
    filter_all_zero_len(x),
    list(list(a = list("foo")))
  )

  x <- list(
    list(
      a = list(
        a = NULL,
        b = list()
      ),
      b = NULL,
      c = character()
    )
  )

  expect_true(all_zero_len(x))
  expect_null(filter_all_zero_len(x))

  x <- list(
    list(
      a = list(
        a = NULL,
        b = list()
      ),
      b = "NULL",
      c = character()
    )
  )

  expect_false(all_zero_len(x))
  expect_identical(
    filter_all_zero_len(x),
    list(list(b = "NULL"))
  )

  x <- list(
    list(
      a = list(
        a = NULL,
        b = list(
          a = character(),
          b = list()
        )
      ),
      b = NULL,
      c = character()
    )
  )

  expect_true(all_zero_len(x))
  expect_null(filter_all_zero_len(x))
})

test_that("*_xtr", {

  x <- list(
    list(
      a = "1",
      b = "2"
    ),
    list(
      a = "3",
      b = "4"
    )
  )

  expect_identical(lst_xtr(x, "a"), list("1", "3"))

  x <- list(
    list(
      a = list(a = "1", b = "2"),
      b = list(a = "3", b = "4")
    ),
    list(
      a = list(a = "5", b = "6"),
      b = list(a = "7", b = "8")
    )
  )

  expect_identical(lst_xtr(x, "a", "a"), list("1", "5"))
})

#' Stack UI
#'
#' Several generics are exported in order to integrate stack UI into board UI.
#' We have `stack_ui()` which is dispatched on the `board` (and in the default
#' implementation) on individual `stack` objects. This renders stacks as
#' bootstrap accordion items (using [bslib::accordion()]). If a different way
#' of displaying stacks and integrating them with a board is desired, this can
#' be implemented by introducing a board subclass and providing a `stack_ui()`
#' method for that subclass. Inserting stacks into (and removing stacks from)
#' a board is available as `insert_stack_ui()`/`remove_stack_ui()` and blocks
#' into/from stacks via `add_block_to_stack()`/`remove_block_from_stack()`.
#' All are S3 generics with implementations for `board` and alternative
#' implementation may be provided for board sub-classes.
#'
#' @param id Parent namespace
#' @param x Object
#' @param ... Generic consistency
#'
#' @return UI set up via `stack_ui()` is expected to return [shiny::tag()] or
#' [shiny::tagList()] objects while stack/block insertion/removal functions
#' (into/from board/stack objects) are called for their side-effects. Both
#' `insert_stack_ui()`/`remove_stack_ui` and
#' `add_block_to_stack()`/`remove_block_from_stack()` return `NULL` invisibly
#' and where the former call [shiny::insertUI()]/[shiny::removeUI()] and the
#' latter modify the DOM via [shiny::session] custom messages.
#'
#' @export
stack_ui <- function(id, x, ...) {
  UseMethod("stack_ui", x)
}

#' @param stacks (Additional) stacks (or IDs) for which to generate the UI
#' @param edit_ui Stack edit plugin
#' @rdname stack_ui
#' @export
stack_ui.board <- function(id, x, stacks = NULL, edit_ui = NULL, ...) {

  stopifnot(is_string(id))

  if (is.null(stacks)) {
    stacks <- board_stacks(x)
  } else if (is.character(stacks)) {
    stacks <- board_stacks(x)[stacks]
  }

  stopifnot(is_stacks(stacks))

  cont_id <- paste0(id, "_stacks")

  log_debug("setting up stack ui container: ", cont_id)

  ns <- NS(id)

  tagList(
    do.call(
      bslib::accordion,
      c(
        list(id = cont_id),
        map(
          stack_ui,
          chr_ply(paste0("stack_", names(stacks)), ns),
          as.list(stacks),
          MoreArgs = list(edit_ui = edit_ui)
        )
      )
    ),
    htmltools::htmlDependency(
      "move-block-ui",
      pkg_version(),
      src = pkg_file("assets", "js"),
      script = "moveBlockUi.js"
    )
  )
}

#' @rdname stack_ui
#' @export
stack_ui.stack <- function(id, x, edit_ui = NULL, ...) {

  accordion_id <- paste0("stack-accordion-panel-", id)

  log_debug("setting up stack item: ", accordion_id)

  if (is.null(edit_ui)) {
    title <- stack_name(x)
  } else {
    title <- get_plugin_ui(edit_ui)(id, x)
  }

  btn <- tags$button(
    class = "accordion-button collapsed",
    type = "button",
    `data-bs-toggle` = "collapse",
    `data-bs-target` = paste0("#", accordion_id),
    `aria-expanded` = "false",
    `aria-controls` = accordion_id,
    do.call(div, c(list(class = "accordion-title"), title))
  )

  div(
    id = paste0("stack-accordion-item-", id),
    class = "accordion-item",
    `data-value` = id,
    div(class = "accordion-header", btn),
    div(
      id = accordion_id,
      class = "accordion-collapse collapse",
      div(class = "accordion-body")
    )
  )
}

#' @param board Board object
#' @param session Shiny session
#' @rdname stack_ui
#' @export
insert_stack_ui <- function(id, x, board, edit_ui = NULL,
                            session = get_session(), ...) {

  UseMethod("insert_stack_ui", board)
}

#' @rdname stack_ui
#' @export
insert_stack_ui.board <- function(id, x, board, edit_ui = NULL,
                                  session = get_session(), ...) {

  x <- as_stacks(x)

  insertUI(
    paste0("#", id, "_stacks"),
    "beforeEnd",
    map(
      stack_ui,
      chr_ply(paste0("stack_", names(x)), session$ns),
      x,
      MoreArgs = list(edit_ui = edit_ui)
    ),
    immediate = TRUE,
    session = session
  )

  invisible()
}

#' @rdname stack_ui
#' @export
remove_stack_ui <- function(id, board, session = get_session(), ...) {
  UseMethod("remove_stack_ui", board)
}

#' @rdname stack_ui
#' @export
remove_stack_ui.board <- function(id, board, session = get_session(), ...) {

  for (x in id) {

    log_debug("removing stack item: ", x)

    removeUI(
      paste0(
        "#stack-accordion-item-stack_",
        session$ns(paste0("stack_", x))
      ),
      immediate = TRUE,
      session = session
    )
  }

  invisible()
}

#' @param block_id,stack_id,board_id Block/stack/board IDs
#' @rdname stack_ui
#' @export
add_block_to_stack <- function(board, block_id, stack_id,
                               session = get_session(), ...) {

  UseMethod("add_block_to_stack", board)
}

#' @rdname stack_ui
#' @export
add_block_to_stack.board <- function(board, block_id, stack_id,
                                     session = get_session(), ...) {

  log_debug("adding block ", block_id, " to stack ", stack_id)

  session$sendCustomMessage(
    "move-block-ui",
    list(
      sel = paste0("#", block_id, "_block"),
      dest = paste0(
        "#stack-accordion-panel-",
        session$ns(paste0("stack_", stack_id)),
        " > div.accordion-body"
      )
    )
  )

  invisible()
}

#' @rdname stack_ui
#' @export
remove_block_from_stack <- function(board, block_id, board_id,
                                    session = get_session(), ...) {

  UseMethod("remove_block_from_stack", board)
}

#' @rdname stack_ui
#' @export
remove_block_from_stack.board <- function(board, block_id, board_id,
                                          session = get_session(), ...) {

  log_debug("removing block ", block_id, " from stacks")

  session$sendCustomMessage(
    "move-block-ui",
    list(
      sel = paste0("#", block_id, "_block"),
      dest = paste0("#", board_id, "_blocks")
    )
  )

  invisible()
}

test_that("dummy stack ui tests", {

  x <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("ChickWeight"),
      c = new_merge_block("Time")
    ),
    links = c(
      ac = new_link("a", "c", "x"),
      bc = new_link("b", "c", "y")
    ),
    stacks = list(ac = c("a", "c"))
  )

  expect_s3_class(
    stack_ui("stacks", x),
    "shiny.tag.list"
  )

  plugins <- board_plugins(new_board())

  expect_s3_class(
    stack_ui("stacks", x, "ac", plugins[["edit_stack"]]),
    "shiny.tag.list"
  )

  expect_null(
    insert_stack_ui(
      "board",
      board_stacks(x),
      x,
      session = MockShinySession$new()
    )
  )

  expect_null(
    insert_stack_ui(
      "board",
      board_stacks(x),
      x,
      plugins[["edit_stack"]],
      MockShinySession$new()
    )
  )

  expect_null(
    remove_stack_ui("board", x, MockShinySession$new())
  )

  expect_null(
    add_block_to_stack(x, "a", "a", MockShinySession$new())
  )

  expect_null(
    remove_block_from_stack(x, "a", "board", MockShinySession$new())
  )
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)


## ----packages, message=FALSE, warning=FALSE-----------------------------------
library(shiny)
library(bslib)
library(scoutbaR)
library(blockr.core)


## ----custom-plugin-setup, eval = FALSE, echo=FALSE----------------------------
# chr_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
#   vapply(x, fun, character(length), ..., USE.NAMES = use_names)
# }
# 
# #' @keywords internal
# lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
#   vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
# }
# 
# dropNulls <- function(x) {
#   x[!lgl_ply(x, is.null)]
# }
# 
# blk_icon <- function(category) {
#   switch(
#     category,
#     "data" = "table",
#     "file" = "file-import",
#     "parse" = "cogs",
#     "plot" = "chart-line",
#     "transform" = "wand-magic-sparkles",
#     "table" = "table"
#   )
# }
# 
# blk_choices <- function() {
#   blk_cats <- sort(
#     unique(chr_ply(available_blocks(), \(b) attr(b, "category")))
#   )
# 
#   lapply(blk_cats, \(cat) {
#     scout_section(
#       label = cat,
#       .list = dropNulls(
#         unname(
#           lapply(available_blocks(), \(choice) {
#             if (attr(choice, "category") == cat) {
#               scout_action(
#                 id = attr(choice, "classes")[1],
#                 label = attr(choice, "name"),
#                 description = attr(choice, "description"),
#                 icon = blk_icon(cat)
#               )
#             }
#           })
#         )
#       )
#     )
#   })
# }


## flowchart TD
##   subgraph board[board]
##     subgraph plugins[plugins]
##       subgraph manage_blocks[Manage blocks]
##       end
##       subgraph manage_links[Manage links]
##       end
##       subgraph manage_stacks[Manage stacks]
##       end
##       subgraph preserve_board[Preserve board]
##       end
##       subgraph generate_code[Generate code]
##       end
##       subgraph notify_user[Notify user]
##       end
##       subgraph edit_block[Edit block]
##       end
##       subgraph edit_stack[Edit stack]
##       end
##     end
##   end

## ----eval=FALSE---------------------------------------------------------------
# main_ui <- function(id, board) {
#   ns <- NS(id)
#   board_ui(
#     ns("board"),
#     board,
#     plugins = board_plugins(
#       board,
#       c(
#         "preserve_board",
#         "manage_blocks",
#         "manage_links",
#         "manage_stacks",
#         "generate_code",
#         "notify_user"
#       )
#     )
#   )
# }


## ----eval=FALSE---------------------------------------------------------------
# main_server <- function(id, board) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       ns <- session$n
# 
#       app_state <- reactiveValues(
#         # App state for module communication
#       )
# 
#       # Board module
#       board_server(
#         "board",
#         board,
#         plugins = board_plugins(
#           board,
#           c(
#             "preserve_board",
#             "manage_blocks",
#             "manage_links",
#             "manage_stacks",
#             "generate_code",
#             "notify_user"
#           )
#         ),
#         callbacks = list(),
#         parent = app_state
#       )
#     }
#   )
# }


## ----eval=FALSE---------------------------------------------------------------
# board_plugins.board <- function(x, which = NULL) {
# 
#   plugins <- plugins(
#     preserve_board(),
#     manage_blocks(),
#     manage_links(),
#     manage_stacks(),
#     notify_user(),
#     generate_code(),
#     edit_block(),
#     edit_stack()
#   )
# 
#   if (is.null(which)) {
#     return(plugins)
#   }
# 
#   plugins[which]
# }


## ----eval=FALSE---------------------------------------------------------------
# manage_blocks <- function(server, ui) {
#   new_plugin(server, ui, validator = expect_null, class = "manage_blocks")
# }


## ----custom-plugin-ui, eval=FALSE---------------------------------------------
# add_rm_block_ui <- function(id, board) {
#   tagList(
#     scoutbar(
#       NS(id, "scoutbar"),
#       placeholder = "Search for a block",
#       actions = blk_choices(),
#       theme = "dark",
#       showRecentSearch = TRUE
#     ),
#     actionButton(
#       NS(id, "add_block"),
#       "New block",
#       icon = icon("circle-plus"),
#     )
#   )
# }


## ----eval=FALSE---------------------------------------------------------------
# #' Add/remove block module
# #'
# #' Customizable logic for adding/removing blocks to the board.
# #'
# #' @param id Namespace ID
# #' @param board Reactive values object
# #' @param update Reactive value object to initiate board updates
# #' @param ... Extra arguments passed from parent scope
# #'
# #' @return A [shiny::reactiveValues()] object with components `add` and `rm`,
# #' where `add` may be `NULL` or a `block` object and `rm` be `NULL` or a string
# #' (block ID).
# #'
# #' @rdname add_rm_block
# #' @export
# add_rm_block_server <- function(id, board, update, ...) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       # SERVER LOGIC
# 
#       NULL
#     }
#   )
# }


## ----eval=FALSE---------------------------------------------------------------
# add_rm_block_server <- function(id, board, update, ...) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       # Trigger add block
#       observeEvent(
#         input$add_block,
#         {
#           update_scoutbar(
#             session,
#             "scoutbar",
#             revealScoutbar = TRUE
#           )
#         }
#       )
# 
#       NULL
#     }
#   )
# }


## ----custom-plugin-server, eval=FALSE-----------------------------------------
# add_rm_block_server <- function(id, board, update, ...) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       # Trigger add block
#       observeEvent(
#         input$add_block,
#         {
#           update_scoutbar(
#             session,
#             "scoutbar",
#             revealScoutbar = TRUE
#           )
#         }
#       )
# 
#       observeEvent(input$scoutbar, {
#         new_blk <- as_blocks(create_block(input$scoutbar))
#         update(
#           list(blocks = list(add = new_blk))
#         )
#       })
# 
#       NULL
#     }
#   )
# }


## ----custom-plugin-helpers, eval=FALSE----------------------------------------
# custom_board_plugins <- function(which = NULL) {
#   plugins <- plugins(
#     manage_blocks(server = add_rm_block_server, ui = add_rm_block_ui)
#   )
# 
#   if (is.null(which)) {
#     return(plugins)
#   }
# 
#   plugins[which]
# }


## ----custom-plugin-app, eval=FALSE--------------------------------------------
#| code-fold: true
# main_ui <- function(id, board) {
#   ns <- NS(id)
#   board_ui(
#     ns("board"),
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     )
#   )
# }
# 
# main_server <- function(id, board) {
#   moduleServer(
#     id,
#     function(input, output, session) {
#       ns <- session$n
# 
#       # Board module
#       board_server(
#         "board",
#         board,
#         plugins = custom_board_plugins(
#           c(
#             "manage_blocks"
#           )
#         ),
#         callbacks = list()
#       )
#     }
#   )
# }
# 
# board <- new_board()
# 
# ui <- page_fluid(
#   main_ui("app", board)
# )
# 
# server <- function(input, output, session) {
#   main_server("app", board)
# }
# 
# shinyApp(ui, server)


## ----shinylive_url, echo = FALSE, results = 'asis'----------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    knitr::knit_code$get("packages"),
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-plugin-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive_iframe, echo = FALSE, eval = TRUE------------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)


## ----custom-board-ui, eval=FALSE----------------------------------------------
# board_ui.custom_board <- function(id, x, plugins = list(), ...) {
#   plugins <- as_plugins(plugins)
#   div(
#     id = paste0(id, "_board"),
#     board_ui(id, plugins[["manage_blocks"]], x),
#     div(
#       id =  paste0(id, "_blocks"),
#       block_ui(id, x)
#     )
#   )
# }


## ----custom-block-ui, eval=FALSE----------------------------------------------
# get_block_registry <- function(x) {
#   stopifnot(is_block(x))
#   available_blocks()[[strsplit(attr(x, "ctor"), "new_")[[1]][2]]]
# }
# 
# block_ui.custom_board <- function(id, x, blocks = NULL, ...) {
#   block_card <- function(x, id, ns) {
#     id <- paste0("block_", id)
# 
#     blk_info <- get_block_registry(x)
# 
#     div(
#       class = "m-2",
#       id = ns(id),
#       shinyNextUI::card(
#         variant = "bordered",
#         shinyNextUI::card_header(
#           className = "d-flex justify-content-between",
#           icon(blk_icon(attr(blk_info, "category"))),
#           sprintf(
#             "Block: %s (id: %s)",
#             attr(blk_info, "name"),
#             gsub("block_", "", id)
#           ),
#           shinyNextUI::tooltip(
#             icon("info-circle"),
#             content = tagList(
#               p(
#               icon("lightbulb"),
#               "How to use this block?",
#               ),
#               p(attr(blk_info, "description"), ".")
#             )
#           )
#         ),
#         shinyNextUI::divider(),
#         shinyNextUI::card_body(
#           expr_ui(ns(id), x),
#           block_ui(ns(id), x)
#         ),
#         shinyNextUI::divider(),
#         shinyNextUI::card_footer(
#           sprintf(
#             "Type: %s; Package: %s",
#             attr(blk_info, "category"),
#             attr(blk_info, "package")
#           )
#         )
#       )
#     )
#   }
# 
#   stopifnot(is.character(id) && length(id) == 1L)
# 
#   if (is.null(blocks)) {
#     blocks <- board_blocks(x)
#   } else if (is.character(blocks)) {
#     blocks <- board_blocks(x)[blocks]
#   }
# 
#   stopifnot(is_blocks(blocks))
# 
#   tagList(
#     Map(
#       block_card,
#       blocks,
#       names(blocks),
#       MoreArgs = list(ns = NS(id)),
#       USE.NAMES = FALSE
#     )
#   )
# }


## ----custom-plugin-ui-nextui, eval=FALSE--------------------------------------
# add_rm_block_ui <- function(id, board) {
#   tagList(
#     scoutbar(
#       NS(id, "scoutbar"),
#       placeholder = "Search for a block",
#       actions = blk_choices(),
#       theme = "dark",
#       showRecentSearch = TRUE
#     ),
#     shinyNextUI::actionButton(
#       NS(id, "add_block"),
#       "New block",
#       icon = icon("circle-plus"),
#     )
#   )
# }


## ----custom-ui-app, eval=FALSE------------------------------------------------
#| code-fold: true
# board <- new_board(class = "custom_board")
# 
# ui <- nextui_page(
#   board_ui(
#     "board",
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     )
#   )
# )
# 
# server <- function(input, output, session) {
#   board_server(
#     "board",
#     board,
#     plugins = custom_board_plugins(
#       c(
#         "manage_blocks"
#       )
#     ),
#     callbacks = list()
#   )
# }
# 
# shinyApp(ui, server)


## ----shinylive2_url, echo = FALSE, results = 'asis'---------------------------
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(shiny)",
    "library(scoutbaR)",
    "library(blockr.core)",
    "library(shinyNextUI)",
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui-nextui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-block-ui"),
    knitr::knit_code$get("custom-board-ui"),
    knitr::knit_code$get("custom-ui-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)


## ----shinylive2_iframe, echo = FALSE, eval = TRUE-----------------------------
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)

#' Board options
#'
#' User settings at the board level are managed by a `board_options` object.
#' This can be constructed via `new_board_options()` and in case the set of
#' user options is to be extended, the constructor is designed with sub-classing
#' in mind. Consequently, the associated validator `validate_board_options()`
#' is available as S3 generic. Inheritance checking is available as
#' `is_board_options()` and coercion as `as_board_options()`.
#'
#' @param ... Options passed as individual arguments
#'
#' @examples
#' opt <- new_board_options(
#'    new_board_name_option(),
#'    new_page_size_option()
#'  )
#'
#' is_board_options(opt)
#' names(opt)
#'
#' opt[["page_size"]]
#'
#' @return All of `new_board_options()` and `as_board_options()` return a
#' `board_options` object, as does the validator `validate_board_options()`,
#' which is typically called for side effects of throwing errors is validation
#' does not pass. Inheritance checking as `is_board_options()` returns a scalar
#' logical, while `board_option_values()` returns a named list of option values.
#'
#' @export
new_board_options <- function(...) {

  res <- new_vctr(
    list_to_list_of_opts(list(...)),
    class = "board_options"
  )

  validate_board_options(res)

  res
}

#' @rdname new_board_options
#' @export
default_board_options <- function(...) {
  new_board_options(
    new_board_name_option(...),
    new_show_conditions_option(...),
    new_thematic_option(...),
    new_dark_mode_option(...)
  )
}

#' @param x Board options object
#' @rdname new_board_options
#' @export
is_board_options <- function(x) {
  inherits(x, "board_options")
}

#' @rdname new_board_options
#' @export
as_board_options <- function(x) {
  UseMethod("as_board_options", x)
}

#' @rdname new_board_options
#' @export
as_board_options.board_options <- function(x) {
  validate_board_options(x)
}

#' @rdname new_board_options
#' @export
as_board_options.board_option <- function(x) {
  new_board_options(x)
}

#' @rdname new_board_options
#' @export
as_board_options.list <- function(x) {
  do.call(new_board_options, x)
}

#' @rdname new_board_options
#' @export
as_board_options.board <- function(x) {
  board_options(x)
}

#' @rdname new_board_options
#' @export
validate_board_options <- function(x) {

  if (!is_board_options(x)) {
    blockr_abort(
      "Expecting board options to inherit from `board_options`.",
      class = "board_options_inheritance_invalid"
    )
  }

  for (opt in x) {
    validate_board_option(opt)
  }

  ids <- names(x)

  if (length(unique(ids)) != length(x)) {
    blockr_abort(
      "Non-unique board options IDs are being used.",
      class = "board_options_duplicated_ids"
    )
  }

  invisible(x)
}

#' @export
as.list.board_options <- function(x, ...) {

  res <- unclass(x)

  names(res) <- chr_ply(res, board_option_id)

  res
}

#' @export
names.board_options <- function(x) {
  chr_ply(x, board_option_id)
}

#' @rdname new_board_options
#' @export
board_option_values <- function(x) {

  x <- as_board_options(x)

  set_names(
    lapply(x, board_option_value),
    names(x)
  )
}

#' @rdname board_ui
#' @export
board_ui.board_options <- function(id, x, ...) {
  do.call(tagList, lapply(x, board_option_ui, id))
}

board_option_to_userdata <- function(x, ..., session = get_session()) {

  stopifnot(is_board_option(x))

  id <- board_option_id(x)
  rv <- reactiveVal(board_option_value(x))

  res <- board_option_server(x, ..., session = session)
  trg <- board_option_trigger(x)

  if (inherits(res, "Observer")) {
    res <- list(res)
  } else if (is.null(res)) {
    res <- list()
  }

  res <- Filter(not_null, res)

  if (!is.list(res) || any(lgl_ply(res, Negate(inherits), "Observer"))) {
    blockr_abort(
      "Expecting a `board_option` server function to return either a ",
      "single or list of observers.",
      class = "invalid_board_option_server_return_value"
    )
  }

  if (not_null(trg)) {

    board_rv <- list(...)$board

    exp <- as.call(
      c(
        quote(`{`),
        lapply(trg, function(v) substitute(session$input[[v]], list(v = v)))
      )
    )

    obs <- list(
      observeEvent(
        exp,
        {
          if (is_scalar(trg)) {
            new <- board_option_value(x, session$input[[trg]])
          } else {
            new <- board_option_value(
              x,
              lapply(set_names(nm = trg), function(v) session$input[[v]])
            )
          }

          cur <- rv()

          if (!identical(new, cur)) {
            set_board_option_value(id, new, board_rv$board, session)
          }
        },
        event.quoted = TRUE,
        ignoreInit = TRUE
      )
    )
  } else {
    obs <- list()
  }

  attr(rv, "observers") <- c(obs, res)

  env <- session$userData

  if (!exists("board_options", envir = env, inherits = FALSE)) {
    assign("board_options", list(), envir = env, inherits = FALSE)
  }

  env$board_options[[id]] <- rv

  invisible()
}

board_options_to_userdata <- function(options, ...) {

  stopifnot(is_board_options(options))

  for (opt in options) {
    board_option_to_userdata(opt, ...)
  }

  invisible()
}

#' @param opt Option name
#' @param session Shiny session
#' @rdname new_board_options
#' @export
get_board_option_value <- function(opt, session = get_session()) {
  rv <- get_board_opt_rv_from_userdata(opt, session)
  rv()
}

get_board_opts_from_userdata <- function(session = get_session()) {

  env <- session$userData

  stopifnot(is.environment(env))

  if (!exists("board_options", envir = env, inherits = FALSE)) {
    assign("board_options", list(), envir = env, inherits = FALSE)
  }

  env$board_options
}

get_board_opt_rv_from_userdata <- function(opt, session = get_session()) {

  stopifnot(is_string(opt))

  opts <- get_board_opts_from_userdata(session)

  if (!opt %in% names(opts)) {
    blockr_abort(
      "Could not find option {opt}.",
      class = "board_option_not_found"
    )
  }

  opts[[opt]]
}

#' @param val New value
#' @param board Board the option belongs to, used to resolve the lock
#' state (see [is_board_locked()]).
#' @rdname new_board_options
#' @export
set_board_option_value <- function(opt, val, board, session = get_session()) {

  if (is_board_locked(board)) {
    log_debug("ignoring write to option {opt} on locked board")
    return(invisible(val))
  }

  log_debug("setting option {opt}")

  rv <- get_board_opt_rv_from_userdata(opt, session)
  rv(val)

  invisible(val)
}

#' @param opts Board options
#' @rdname new_board_options
#' @export
get_board_option_or_default <- function(opt, opts = default_board_options(),
                                        session = get_session()) {
  tryCatch(
    get_board_option_value(opt, session),
    board_option_not_found = function(e) {
      board_option_value(opts[[opt]])
    }
  )
}

#' @rdname new_board_options
#' @export
get_board_option_or_null <- function(opt, session = get_session()) {

  if (is.null(session)) {
    return(NULL)
  }

  tryCatch(
    get_board_option_value(opt, session),
    board_option_not_found = function(e) NULL
  )
}

#' @param if_not_found Behavior in case an option is not found
#' @rdname new_board_options
#' @export
get_board_option_values <- function(..., opts = default_board_options(),
                                    if_not_found = c("error", "default",
                                                     "null"),
                                    session = get_session()) {

  fun <- switch(
    match.arg(if_not_found),
    error = get_board_option_value,
    default = function(opt, session) {
      get_board_option_or_default(opt, opts, session)
    },
    null = get_board_option_or_null
  )

  lapply(set_names(nm = c(...)), fun, session = session)
}

#' @export
format.board_options <- function(x, ...) {
  c(
    paste0("<", class(x)[1L], "[", length(x), "]>"),
    if (length(x)) paste0("  ", chr_ply(x, format))
  )
}

#' @export
print.board_options <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.board_options <- function(x, ...) {
  str_value_collection(x, chr_ply(x, str_value), ids = NULL)
}

#' @export
str.board_options <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
c.board_options <- function(...) {
  as_board_options(list_to_list_of_opts(list(...)))
}

#' @export
`[.board_options` <- function(x, i, ...) {

  res <- opts_slice(x, vec_as_location(i, length(x), names(x)))

  if (!length(res)) {
    return(new_board_options())
  }

  res
}

#' @export
`[<-.board_options` <- function(x, i, ..., value) {
  validate_board_options(NextMethod())
}

#' @export
`[[.board_options` <- function(x, i, ...) {
  validate_board_option(
    .subset2(x, vec_as_location2(i, length(x), names(x)))
  )
}

#' @export
`[[<-.board_options` <- function(x, i, ..., value) {
  validate_board_options(NextMethod())
}

opts_slice <- function(...) {
  validate_board_options(vec_slice(...))
}

harmonize_list_of_opts <- function(x) {
  if (is_board_option(x)) {
    list(x)
  } else if (is_board_options(x)) {
    as.list(x)
  } else if (is.list(x) && all(lgl_ply(x, is_board_option))) {
    x
  } else if (is.list(x) && any(lgl_ply(x, is_board_options))) {
    c(
      x[lgl_ply(x, Negate(is_board_options))],
      unlst(lapply(x[lgl_ply(x, is_board_options)], as.list))
    )
  } else {
    list(as_board_option(x))
  }
}

list_to_list_of_opts <- function(x) {

  x <- Filter(not_null, x)

  if (length(x)) {
    unlst(lapply(x, harmonize_list_of_opts))
  } else {
    list()
  }
}

#' @rdname new_board_options
#' @export
combine_board_options <- function(...) {

  opts <- list_to_list_of_opts(list(...))

  as_board_options(
    opts[!duplicated(chr_ply(opts, board_option_id))]
  )
}

test_that("is_board_locked reflects the blockr.locked option", {

  withr::local_options(blockr.locked = NULL)
  expect_false(is_board_locked(new_board()))

  withr::local_options(blockr.locked = TRUE)
  expect_true(is_board_locked(new_board()))
})

test_that("is_board_locked dispatches on board subclass", {

  withr::local_options(blockr.locked = TRUE)

  registerS3method(
    "is_board_locked", "test_unlockable_board",
    function(board, ...) FALSE,
    envir = asNamespace("blockr.core")
  )

  expect_false(is_board_locked(new_board(class = "test_unlockable_board")))
})

test_that("a locked board drops structural board updates", {

  withr::local_options(blockr.locked = TRUE)

  board <- new_board()

  testServer(
    get_s3_method("board_server", board),
    {
      board_update(
        list(blocks = list(add = as_blocks(new_dataset_block())))
      )

      session$flushReact()

      expect_length(board_blocks(rv$board), 0L)
      expect_null(board_update())
    },
    args = list(x = board, plugins = list(manage_blocks()))
  )
})

test_that("an unlocked board applies structural board updates", {

  board <- new_board()

  testServer(
    get_s3_method("board_server", board),
    {
      board_update(
        list(blocks = list(add = as_blocks(new_dataset_block())))
      )

      session$flushReact()

      expect_length(board_blocks(rv$board), 1L)
    },
    args = list(x = board, plugins = list(manage_blocks()))
  )
})

test_that("the gate consults is_board_locked, not the raw option", {

  withr::local_options(blockr.locked = TRUE)

  registerS3method(
    "is_board_locked", "test_unlockable_board",
    function(board, ...) FALSE,
    envir = asNamespace("blockr.core")
  )

  board <- new_board(class = "test_unlockable_board")

  testServer(
    get_s3_method("board_server", board),
    {
      board_update(
        list(blocks = list(add = as_blocks(new_dataset_block())))
      )

      session$flushReact()

      expect_length(board_blocks(rv$board), 1L)
    },
    args = list(x = board, plugins = list(manage_blocks()))
  )
})

test_that("set_board_option_value is ignored on a locked board", {

  withr::local_options(blockr.locked = TRUE)

  board <- new_board()

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      set_board_option_value("n_rows", 7L, board, session)
      expect_equal(get_board_option_value("n_rows", session), 50L)
    },
    args = list(
      x = board,
      options = as_board_options(new_n_rows_option())
    )
  )
})

test_that("a forged option input is ignored on a locked board", {

  withr::local_options(blockr.locked = TRUE)

  board <- new_board()

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      session$setInputs(n_rows = 7L)
      session$flushReact()

      expect_equal(get_board_option_value("n_rows", session), 50L)
    },
    args = list(
      x = board,
      options = as_board_options(new_n_rows_option())
    )
  )
})

test_that("an option input updates an unlocked board", {

  board <- new_board()

  testServer(
    get_s3_method("board_server", board),
    {
      session$flushReact()

      session$setInputs(n_rows = 8L)
      session$flushReact()

      expect_equal(get_board_option_value("n_rows", session), 8L)
    },
    args = list(
      x = board,
      options = as_board_options(new_n_rows_option())
    )
  )
})

#' Logging
#'
#' Internally used infrastructure for emitting log messages is exported, hoping
#' that other packages which depend on this, use it and thereby logging is
#' carried out consistently both in terms of presentation and output device.
#' All log messages are associated with an (ordered) level ("fatal", "error",
#' "warn", "info", "debug" or "trace") which is compared against the currently
#' set value (available as `get_log_level()`) and output is only generated if
#' the message level is greater or equal to the currently set value.
#'
#' @param ... Concatenated as `paste0(..., "\n")`
#' @param level Logging level (possible values are "fatal", "error", "warn",
#' "info", "debug" and "trace"
#' @param envir Environment where the logging call originated from
#' @param asis Flag to disable re-wrapping of text to terminal width
#' @param use_glue Flag to disable use of glue
#' @param pkg Package name
#'
#' @return Logging function `write_log()`, wrappers `log_*()` and loggers
#' provided as `cnd_logger()`/cat_logger() all return `NULL` invisibly and are
#' called for their side effect of emitting a message. Helpers `as_log_level()`
#' and `get_log_level()` return a scalar-valued ordered factor.
#'
#' @export
write_log <- function(..., level = "info", envir = parent.frame(),
                      asis = FALSE, use_glue = TRUE, pkg = pkg_name(envir)) {

  stopifnot(is.environment(envir), is.null(pkg) || is_string(pkg))

  lvl <- as_log_level(level)

  if (lvl > get_log_level()) {
    return(invisible(NULL))
  }

  logger <- get_logger()

  pfx <- paste0(
    "[", toupper(level), "]",
    if (isTRUE(blockr_option("log_time", TRUE))) get_timmestamp("[", "]"),
    if (isTRUE(blockr_option("log_mem", FALSE))) get_mem_use("[", "]"),
    if (not_null(pkg)) paste0("[", pkg, "]"),
    " "
  )

  if (isTRUE(use_glue)) {
    msg <- glue_plur(..., envir = envir, .trim = !asis)
  } else {
    msg <- paste0(...)
  }

  stopifnot(is_string(msg))

  msg <- strsplit(msg, "\n", fixed = TRUE)

  if (!isTRUE(asis)) {

    msg <- strwrap(
      msg[[1L]],
      width = getOption("width") - nchar(pfx),
      exdent = 2L
    )

  } else {
    msg <- msg[[1L]]
  }

  logger(paste0(pfx, msg, collapse = "\n"), level = lvl)
}

get_mem_use <- function(prefix = "", suffix = "") {

  if (!has_memuse()) {
    return("")
  }

  mem <- memuse::Sys.procmem()

  if (is.null(mem$peak)) {
    peak <- ""
  } else {
    peak <- paste0("/", mem$peak)
  }

  paste0(prefix, mem$size, peak, suffix)
}

has_memuse <- function() {
  requireNamespace("memuse", quietly = TRUE)
}

get_timmestamp <- function(prefix = "", suffix = "") {
  paste0(prefix, format(Sys.time()), suffix)
}

#' @rdname write_log
#' @export
log_fatal <- function(..., envir = parent.frame()) {
  write_log(..., level = "fatal", envir = envir)
}

#' @rdname write_log
#' @export
log_error <- function(..., envir = parent.frame()) {
  write_log(..., level = "error", envir = envir)
}

#' @rdname write_log
#' @export
log_warn <- function(..., envir = parent.frame()) {
  write_log(..., level = "warn", envir = envir)
}

#' @rdname write_log
#' @export
log_info <- function(..., envir = parent.frame()) {
  write_log(..., level = "info", envir = envir)
}

#' @rdname write_log
#' @export
log_debug <- function(..., envir = parent.frame()) {
  write_log(..., level = "debug", envir = envir)
}

#' @rdname write_log
#' @export
log_trace <- function(..., envir = parent.frame()) {
  write_log(..., level = "trace", envir = envir)
}

#' @rdname write_log
#' @export
as_log_level <- function(level) {
  log_levels <- c("fatal", "error", "warn", "info", "debug", "trace")
  ordered(match.arg(level, log_levels), log_levels)
}

#' @rdname write_log
#' @export
fatal_log_level <- as_log_level("fatal")

#' @rdname write_log
#' @export
error_log_level <- as_log_level("error")

#' @rdname write_log
#' @export
warn_log_level <- as_log_level("warn")

#' @rdname write_log
#' @export
info_log_level <- as_log_level("info")

#' @rdname write_log
#' @export
debug_log_level <- as_log_level("debug")

#' @rdname write_log
#' @export
trace_log_level <- as_log_level("trace")

#' @rdname write_log
#' @export
get_log_level <- function() {
  as_log_level(
    blockr_option("log_level", "info")
  )
}

get_logger <- function() {

  fun <- blockr_option("logger", "cat_logger")

  if (is.function(fun)) {
    return(fun)
  }

  get(fun, mode = "function")
}

#' @param msg Message (string)
#' @rdname write_log
#' @export
cnd_logger <- function(msg, level) {

  if (level == fatal_log_level) {
    stop(msg, call. = FALSE)
  }

  if (level <= warn_log_level) {
    warning(msg, call. = FALSE)
  } else {
    message(msg)
  }

  invisible()
}

#' @rdname write_log
#' @export
cat_logger <- function(msg, level) {

  if (level == fatal_log_level) {
    stop(msg, call. = FALSE)
  }

  if (level == error_log_level) {
    out <- stderr()
  } else {
    out <- stdout()
  }

  cat(paste0(msg, "\n"), file = out)

  invisible()
}

test_that("logging", {

  withr::local_envvar(
    BLOCKR_LOG_LEVEL = "",
    BLOCKR_LOG_MEM = "",
    BLOCKR_LOG_TIME = "",
    BLOCKR_LOGGER = ""
  )

  withr::with_options(
    list(blockr.log_level = "info"),
    {
      expect_identical(get_log_level(), info_log_level)
      expect_silent(log_debug("abc"))
      expect_output(log_warn("abc"), " abc$")
      expect_output(log_warn("abc"), "^\\[WARN\\]")
    }
  )

  withr::with_options(
    list(
      blockr.log_level = "info",
      blockr.log_mem = TRUE,
      blockr.log_time = TRUE
    ),
    {
      expect_true(blockr_option("log_time", FALSE))
      expect_true(blockr_option("log_mem", FALSE))

      with_mocked_bindings(
        {
          expect_output(log_info("abc"), "\\]\\[42B\\]\\[blockr\\.core\\] ")
          expect_output(log_info("abc"), "\\]\\[.+\\]\\[")
        },
        get_mem_use = function(prefix = "", suffix = "") {
          paste0(prefix, "42B", suffix)
        }
      )
    }
  )

  withr::with_options(
    list(
      blockr.log_level = "info",
      blockr.logger = cnd_logger
    ),
    {
      expect_message(log_info("abc"))
      expect_warning(log_warn("abc"))
      expect_warning(log_error("abc"))
      expect_error(log_fatal("abc"))
    }
  )
})

test_that("memory marker is omitted when memuse is unavailable", {

  expect_identical(
    with_mocked_bindings(
      get_mem_use("[", "]"),
      has_memuse = function() FALSE
    ),
    ""
  )
})

#' @section Row-bind block:
#' Row-wise concatenation of an arbitrary number of `data.frame`s, as performed
#' by [base::rbind()] is available as an `rbind_block`. This mainly serves as
#' an example for a variadic block via the "special" `...args` block data
#' argument.
#'
#' @block rbind block
#' @blockDescr Row-binding of datasets
#' @blockCategory transform
#' @blockIcon chevron-bar-expand
#' @blockGuidance Row-binds two or more inputs with `base::rbind()`; the inputs
#'   must share the same columns. Takes a variable number of data inputs and has
#'   no configurable arguments.
#' @blockKeywords rbind, bind, concatenate, rows, append, union
#'
#' @rdname new_transform_block
#' @export
new_rbind_block <- function(...) {
  new_transform_block(
    function(id, ...args) {
      moduleServer(
        id,
        function(input, output, session) {

          arg_names <- reactive(
            dot_arg_refs(...args)
          )

          list(
            expr = reactive(
              bquote(
                rbind(..(dat)),
                list(dat = lapply(arg_names(), as_dot_call)),
                splice = TRUE
              )
            ),
            state = list()
          )
        }
      )
    },
    expr_type = "bquoted",
    class = "rbind_block",
    ...
  )
}

#' @param expr Quoted expression to evaluate in the context of `data`
#' @param env Environment in which to evaluate `expr`
#' @rdname block_server
#' @export
block_eval <- function(x, expr, env, ...) {
  UseMethod("block_eval")
}

#' @export
block_eval.block <- function(x, expr, env, ...) {
  eval(expr, env)
}

#' @rdname block_server
#' @export
eval_env <- function(data) {
  if (isTRUE(blockr_option("attach_default_packages", FALSE))) {
    parent <- default_eval_parent()
  } else {
    parent <- baseenv()
  }
  list2env(data, parent = parent)
}

eval_env_default_pkgs <- c(
  "stats", "graphics", "grDevices", "utils", "datasets", "methods"
)

default_eval_parent <- local({
  cached <- NULL
  function() {
    if (!is.null(cached)) return(cached)
    parent <- baseenv()
    for (pkg in rev(eval_env_default_pkgs)) {
      parent <- pkg_export_env(pkg, parent)
    }
    cached <<- parent
    cached
  }
})

pkg_export_env <- function(pkg, parent) {
  ns <- loadNamespace(pkg)
  e <- new.env(parent = parent)
  exports <- getNamespaceExports(ns)
  importIntoEnv(e, exports, ns, exports)
  lazydata <- getNamespaceInfo(ns, "lazydata")
  for (nm in ls(lazydata, all.names = TRUE)) {
    do.call(delayedAssign,
            list(nm, as.symbol(nm), eval.env = lazydata, assign.env = e))
  }
  lockEnvironment(e, bindings = TRUE)
  e
}

#' @param session Shiny session object
#' @rdname block_server
#' @export
block_eval_trigger <- function(x, session = get_session()) {
  UseMethod("block_eval_trigger", x)
}

#' @export
block_eval_trigger.block <- function(x, session = get_session()) {
  NULL
}

test_that("eval_env defaults to baseenv (utils not visible)", {
  e <- eval_env(list())
  expect_identical(parent.env(e), baseenv())
  expect_error(eval(quote(head(1:10)), e), "could not find function")
})

test_that("attach_default_packages exposes vanilla R defaults", {
  withr::with_options(list(blockr.attach_default_packages = TRUE), {
    e <- eval_env(list())
    expect_equal(eval(quote(head(1:10, 3)), e), 1:3)
    expect_s3_class(
      eval(quote(lm(y ~ x, data.frame(x = 1:3, y = c(2, 4, 6)))), e),
      "lm"
    )
    expect_s3_class(eval(quote(iris), e), "data.frame")
    expect_equal(dim(eval(quote(mtcars), e)), c(32L, 11L))
    expect_true(is.function(eval(quote(setMethod), e)))
    expect_true(is.function(eval(quote(png), e)))
  })
})

test_that("non-default base packages remain namespace-only", {
  withr::with_options(list(blockr.attach_default_packages = TRUE), {
    e <- eval_env(list())
    expect_error(
      eval(quote(mclapply(1:2, identity)), e),
      "could not find function"
    )
    expect_error(
      eval(quote(viewport()), e),
      "could not find function"
    )
  })
})

test_that("evals do not leak state across calls", {
  withr::with_options(list(blockr.attach_default_packages = TRUE), {
    eval(quote(x <- 1), eval_env(list()))
    expect_false(exists("x", envir = eval_env(list()), inherits = FALSE))
  })
})

test_that("default eval parent chain is locked end-to-end", {
  withr::with_options(list(blockr.attach_default_packages = TRUE), {
    parent <- parent.env(eval_env(list()))
    while (!identical(parent, baseenv())) {
      expect_true(environmentIsLocked(parent))
      expect_error(
        assign("head", 1L, envir = parent),
        "locked"
      )
      parent <- parent.env(parent)
    }
  })
})

test_that("default eval parent is cached across calls", {
  withr::with_options(list(blockr.attach_default_packages = TRUE), {
    expect_identical(default_eval_parent(), default_eval_parent())
    expect_identical(parent.env(eval_env(list())), default_eval_parent())
  })
})

test_that("data inputs are bound in the eval env", {
  e <- eval_env(list(x = 1:5, y = letters[1:3]))
  expect_equal(get("x", envir = e), 1:5)
  expect_equal(get("y", envir = e), letters[1:3])
})

#' @section Glue block:
#' Using [glue::glue()], this block allows evaluation of a text string in the
#' context of datasets to produce (markdown formatted) text as block result.
#'
#' @param text String evaluated using [glue::glue()]
#'
#' @block glue string block
#' @blockDescr String interpolation using glue
#' @blockCategory utility
#' @blockIcon braces
#' @blockGuidance Interpolates a template against the input data with
#'   `glue::glue()` to produce (markdown) text. Reference columns of the data in
#'   `text` with braces, e.g. "{nrow(data)} rows".
#' @blockKeywords glue, template, text, string, interpolate, markdown
#' @blockArg text Template string evaluated with glue::glue().
#'   [example] "{nrow(data)} rows"
#'   [type] arg_string()
#'
#' @rdname new_text_block
#' @export
new_glue_block <- function(text = character(), ...) {

  new_text_block(
    function(id, ...args) {
      moduleServer(
        id,
        function(input, output, session) {

          arg_names <- reactive(
            dot_arg_refs(...args)
          )

          list(
            expr = reactive(
              bquote(
                glue::glue(.(txt), .envir = .(env)),
                list(
                  txt = input$text,
                  env = bquote(
                    list2env(list(..(data)), parent = baseenv()),
                    list(data = lapply(arg_names(), as_dot_call)),
                    splice = TRUE
                  )
                )
              )
            ),
            state = list(
              text = reactive(input$text)
            )
          )
        }
      )
    },
    function(id) {
      textAreaInput(
        inputId = NS(id, "text"),
        label = NULL,
        value = text,
        placeholder = "You may use markdown syntax to style the text."
      )
    },
    allow_empty_state = list(data = list(...args = 0)),
    expr_type = "bquoted",
    class = "glue_block",
    ...
  )
}

#' File block constructors
#'
#' Similarly to [new_data_block()], blocks created via `new_file_block()` serve
#' as starting points in analysis pipelines by providing data to down-stream
#' blocks. They typically will not have data inputs and represent root nodes in
#' analysis graphs.
#'
#' @param ... Forwarded to `new_file_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_file_block()` inherit from
#' `file_block`.
#'
#' @seealso The [blockr.io](https://bristolmyerssquibb.github.io/blockr.io/)
#' package provides real-world blocks for sourcing external files.
#'
#' @export
new_file_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "file_block"), ctor, ...)
}

#' @param id Board option ID
#' @param default Default value
#' @param ui Option UI
#' @param server (Optional) option server
#' @param update_trigger Shiny `input` entry/entries that trigger an update
#' @param transform (Optional) transform function
#' @param category (Optional) string-valued category
#' @param ctor,pkg Constructor information (used for serialization)

#' @rdname new_board_options
#' @export
new_board_option <- function(id, default, ui,
                             server = function(board, ..., session) {},
                             update_trigger = id,
                             transform = identity,
                             category = NULL, ctor = sys.parent(),
                             pkg = NULL) {

  res <- structure(
    list(ui = ui, server = server, transform = transform),
    default = default,
    id = id,
    trigger = update_trigger,
    category = category,
    ctor = resolve_ctor(ctor, pkg),
    class = c(paste0(id, "_option"), "board_option")
  )

  validate_board_option(res)

  res
}

#' @rdname new_board_options
#' @export
is_board_option <- function(x) {
  inherits(x, "board_option")
}

#' @rdname new_board_options
#' @export
validate_board_option <- function(x) {
  UseMethod("validate_board_option", x)
}

#' @rdname new_board_options
#' @export
as_board_option <- function(x, ...) {
  UseMethod("as_board_option", x)
}

#' @rdname new_board_options
#' @export
as_board_option.board_option <- function(x, ...) {
  x
}

#' @rdname new_board_options
#' @export
board_option_id <- function(x) {
  stopifnot(is_board_option(x))
  attr(x, "id")
}

#' @rdname new_board_options
#' @export
board_option_trigger <- function(x) {
  stopifnot(is_board_option(x))
  attr(x, "trigger")
}

#' @rdname new_board_options
#' @export
board_option_default <- function(x) {
  stopifnot(is_board_option(x))
  attr(x, "default")
}

#' @rdname new_board_options
#' @export
board_option_category <- function(x) {
  stopifnot(is_board_option(x))
  attr(x, "category")
}

#' @rdname new_board_options
#' @export
board_option_ui <- function(x, id = NULL) {

  stopifnot(is_board_option(x))

  fun <- x[["ui"]]

  if (is.null(id)) {
    return(fun)
  }

  fun(id)
}

#' @rdname new_board_options
#' @export
board_option_server <- function(x, ...) {

  stopifnot(is_board_option(x))

  fun <- x[["server"]]

  if (...length()) {
    return(fun(...))
  }

  fun
}

#' @rdname new_board_options
#' @export
board_option_transform <- function(x) {
  stopifnot(is_board_option(x))
  x[["transform"]]
}

#' @param value Option value
#' @rdname new_board_options
#' @export
board_option_value <- function(x, value = board_option_default(x)) {
  trans <- board_option_transform(x)
  trans(value)
}

#' @rdname new_board_options
#' @export
board_option_ctor <- function(x) {
  stopifnot(is_board_option(x))
  attr(x, "ctor")
}

#' @rdname new_board_options
#' @export
validate_board_option.default <- function(x) {

  if (!is_board_option(x)) {
    blockr_abort(
      "Expecting a board option to inherit from `board_option`.",
      class = "board_option_inheritance_invalid"
    )
  }

  if (!is_string(board_option_id(x))) {
    blockr_abort(
      "Expecting a board option ID to be string-valued.",
      class = "board_option_component_id_invalid"
    )
  }

  trigger <- board_option_trigger(x)

  if (!is.null(trigger) && (!is.character(trigger) || !length(trigger))) {
    blockr_abort(
      "Expecting a board option trigger to be a non-zero length character",
      "vector or `NULL`.",
      class = "board_option_component_trigger_invalid"
    )
  }

  ui <- board_option_ui(x)

  if (!is.function(ui) || !identical(names(formals(ui)), "id")) {
    blockr_abort(
      "Expecting a board option UI function to have a single argument `id`.",
      class = "board_option_component_ui_invalid"
    )
  }

  srv <- board_option_server(x)

  if (!is.function(srv)) {
    blockr_abort(
      "Expecting a board option server function to be a function.",
      class = "board_option_component_server_invalid"
    )
  }

  arg <- names(formals(srv))
  len <- length(arg)
  exp <- c("...", "session")

  if (len < 2L || !identical(arg[seq.int(len - 1L, len)], exp)) {
    blockr_abort(
      "Expecting a board option server function signature to end with {exp}.",
      class = "board_option_component_server_invalid"
    )
  }

  tfm <- board_option_transform(x)

  if (!is.function(tfm) || !identical(names(formals(tfm)), "x")) {
    blockr_abort(
      "Expecting a board option transform to have a single argument `x`.",
      class = "board_option_component_transform_invalid"
    )
  }

  invisible(x)
}

#' @export
format.board_option <- function(x, session = get_session(), ...) {

  id <- board_option_id(x)
  vl <- coal(
    get_board_option_or_null(id, session),
    board_option_value(x),
    fail_all = FALSE
  )

  paste0(id, ": ", format_head(vl, 5L))
}

#' @export
print.board_option <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.board_option <- function(x, ...) {
  format(x, ...)
}

#' @export
str.board_option <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
c.board_option <- function(...) {
  as_board_options(list_to_list_of_opts(list(...)))
}

#' @rdname new_board_options
#' @export
new_board_name_option <- function(value = NULL, category = "Board options",
                                  ...) {

  new_board_option(
    id = "board_name",
    default = value,
    ui = function(id) {
      tagList(
        htmltools::htmlDependency(
          "change-board-title",
          pkg_version(),
          src = pkg_file("assets", "js"),
          script = "changeBoardTitle.js"
        ),
        textInput(
          NS(id, "board_name"),
          "Board name",
          value
        )
      )
    },
    server = function(board, ..., session) {
      list(
        if (is.null(value)) {
          observeEvent(
            TRUE,
            {
              val <- id_to_sentence_case(board$board_id)
              set_board_option_value("board_name", val, board$board, session)
              updateTextInput(session, "board_name", value = val)
              session$sendCustomMessage("change-board-title", val)
            },
            once = TRUE
          )
        },
        observeEvent(
          get_board_option_or_null("board_name", session),
          {
            val <- get_board_option_value("board_name", session)
            updateTextInput(session, "board_name", value = val)
            session$sendCustomMessage("change-board-title", val)
          }
        )
      )
    },
    category = category,
    ...
  )
}

#' @export
validate_board_option.board_name_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!(is_string(val) || is.null(val))) {
    blockr_abort(
      "Expecting `board_name` to be `NULL` or string-valued.",
      class = "board_name_option_invalid"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_n_rows_option <- function(value = blockr_option("n_rows", 50L),
                              category = "Table options", ...) {

  new_board_option(
    id = "n_rows",
    default = value,
    ui = function(id) {
      numericInput(
        NS(id, "n_rows"),
        "Preview rows",
        value,
        min = 1L,
        step = 1L
      )
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("n_rows", session),
        {
          updateNumericInput(
            session,
            "n_rows",
            value = get_board_option_value("n_rows", session)
          )
        }
      )
    },
    transform = function(x) as.integer(x),
    category = category,
    ...
  )
}

#' @export
validate_board_option.n_rows_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!is_count(val)) {
    blockr_abort(
      "Expecting `n_rows` to represent a count.",
      class = "board_options_n_rows_invalid"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_page_size_option <- function(value = blockr_option("page_size", 5L),
                                 category = "Table options", ...) {

  new_board_option(
    id = "page_size",
    default = value,
    ui = function(id) {
      selectInput(
        NS(id, "page_size"),
        "Preview page size",
        c(5, 10, 25, 50, 100),
        value
      )
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("page_size", session),
        {
          updateSelectInput(
            session,
            "page_size",
            selected = get_board_option_value("page_size", session)
          )
        }
      )
    },
    transform = function(x) as.integer(x),
    category = category,
    ...
  )
}

#' @export
validate_board_option.page_size_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!is_count(val)) {
    blockr_abort(
      "Expecting `page_size` to represent a count.",
      class = "board_options_page_size_invalid"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_filter_rows_option <- function(value = blockr_option("filter_rows",
                                                         FALSE),
                                   category = "Table options", ...) {

  new_board_option(
    id = "filter_rows",
    default = value,
    ui = function(id) {
      bslib::input_switch(
        NS(id, "filter_rows"),
        "Enable preview search",
        value
      )
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("filter_rows", session),
        {
          bslib::toggle_switch(
            "filter_rows",
            value = get_board_option_value("filter_rows", session),
            session = session
          )
        }
      )
    },
    category = category,
    ...
  )
}

#' @export
validate_board_option.filter_rows_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!is_bool(val)) {
    blockr_abort(
      "Expecting `filter_rows` to represent a boolean.",
      class = "board_options_filter_rows_invalid"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_thematic_option <- function(value = blockr_option("thematic", NULL),
                                category = "Theme options", ...) {

  new_board_option(
    id = "thematic",
    default = value,
    ui = function(id) {
      if (pkg_avail("thematic")) {
        bslib::input_switch(
          NS(id, "thematic"),
          "Enable thematic",
          coal(value, FALSE)
        )
      }
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("thematic", session),
        {
          bslib::toggle_switch(
            "thematic",
            value = get_board_option_value("thematic", session),
            session = session
          )
        }
      )
    },
    category = category,
    ...
  )
}

#' @export
validate_board_option.thematic_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!(is.null(val) || is_bool(val))) {
    blockr_abort(
      "Expecting `thematic` to be `NULL` or a boolean.",
      class = "board_options_thematic_invalid"
    )
  }

  if (isTRUE(val) && !pkg_avail("thematic")) {
    blockr_abort(
      "Please install `thematic` to enable auto theming of plots.",
      class = "thematic_not_installed"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_dark_mode_option <- function(value = blockr_option("dark_mode", NULL),
                                 category = "Theme options", ...) {

  if (isTRUE(value)) {
    value <- "dark"
  }

  if (isFALSE(value)) {
    value <- "light"
  }

  if (length(value) && is.na(value)) {
    value <- NULL
  }

  new_board_option(
    id = "dark_mode",
    default = value,
    ui = function(id) {
      span(
        bslib::input_dark_mode(
          id = NS(id, "dark_mode"),
          mode = value
        ),
        tags$label(
          "Light/dark mode",
          style = "vertical-align: top; margin-top: 3px;"
        )
      )
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("dark_mode", session),
        {
          bslib::toggle_dark_mode(
            mode = get_board_option_value("dark_mode", session),
            session = session
          )
        }
      )
    },
    category = category,
    ...
  )
}

#' @export
validate_board_option.dark_mode_option <- function(x) {

  val <- board_option_value(NextMethod())

  if (!(is.null(val) || (is_string(val) && val %in% c("light", "dark")))) {
    blockr_abort(
      "Expecting `dark_mode` to be either `NULL`, \"light\" or \"dark\".",
      class = "board_options_dark_mode_invalid"
    )
  }

  invisible(x)
}

#' @rdname new_board_options
#' @export
new_show_conditions_option <- function(value = blockr_option("show_conditions",
                                                             c("warning",
                                                               "error")),
                                       category = "Board options", ...) {

  new_board_option(
    id = "show_conditions",
    default = value,
    ui = function(id) {
      selectInput(
        NS(id, "show_conditions"),
        "Message types to display",
        c("message", "warning", "error"),
        value,
        multiple = TRUE
      )
    },
    server = function(..., session) {
      observeEvent(
        get_board_option_or_null("show_conditions", session),
        {
          updateSelectInput(
            session,
            "show_conditions",
            selected = get_board_option_value("show_conditions", session)
          )
        }
      )
    },
    category = category,
    ...
  )
}

#' @export
validate_board_option.show_conditions_option <- function(x) {

  val <- board_option_value(NextMethod())

  opt <- c("message", "warning", "error")

  if (!(is.character(val) || all(val %in% opt))) {
    blockr_abort(
      "Expecting `show_conditions` to be any of {opt}.",
      class = "board_options_show_conditions_invalid"
    )
  }

  invisible(x)
}

default_chat <- function(system_prompt = NULL, params = NULL) {
  ellmer::chat_openai(system_prompt = system_prompt, params = params)
}

#' @rdname new_board_options
#' @export
new_llm_model_option <- function(value = NULL, category = "Board options",
                                 ...) {

  if (!(is.null(value) || is_string(value))) {
    blockr_abort(
      "Expecting an LLM model option to be specified by a string (or be ",
      "`NULL`).",
      class = "invalid_llm_model_option_value"
    )
  }

  options <- get_chat_fun_opts()

  if (!is.function(options) && length(options) == 1L) {

    if (is_string(value) && identical(value, names(options))) {
      value <- NULL
    }

    options <- options[[1L]]
  }

  if (is.null(value) && !is.function(options)) {
    value <- names(options)[1L]
  }

  if (is.function(options) && !is.null(value)) {

    blockr_warn(
      "Ignoring value {value} for option `llm_model_option`.",
      class = "single_llm_model"
    )

    value <- NULL
  }

  if (is.list(options) && is_string(value) && !value %in% names(options)) {
    blockr_abort(
      "Cannot choose {value} among option{?s} {names(options)}.",
      class = "invalid_llm_model_option_value"
    )
  }

  new_board_option(
    id = "llm_model",
    default = value,
    ui = function(id) {
      if (is.function(options)) {
        return(NULL)
      }
      selectInput(
        NS(id, "llm_model"),
        "LLM Model",
        names(options),
        value
      )
    },
    server = function(..., session) {
      if (!is.function(options)) {
        observeEvent(
          get_board_option_or_null("llm_model", session),
          {
            opt <- get_board_option_value("llm_model", session)
            nme <- attr(opt, "chat_name")

            stopifnot(is_string(nme), nme %in% names(options))

            updateSelectInput(session, "llm_model", selected = nme)
          }
        )
      }
    },
    transform = function(x) {

      if (is.function(options)) {
        return(options)
      }

      structure(options[[x]], chat_name = x)
    },
    update_trigger = if (is.function(options)) NULL else "llm_model",
    category = category,
    ...
  )
}

get_chat_fun_opts <- function() {

  opt <- blockr_option("chat_function", default_chat)

  is_loo <- is.list(opt) &&
    all(lgl_ply(opt, is.function)) &&
    length(unique(names(opt))) == length(opt)

  if (!(is.function(opt) || is_loo)) {
    blockr_abort(
      "Expecting the blockr option `chat_function` to be either a function ",
      "or a list of functions with unique names.",
      class = "board_options_llm_model_invalid"
    )
  }

  opt
}

#' @export
format.llm_model_option <- function(x, session = get_session(), ...) {

  opts <- get_chat_fun_opts()

  if (is.function(opts)) {
    opts <- ""
  } else {
    opts <- paste0(" (", format_head(names(opts), 3L), ")")
  }

  id <- board_option_id(x)
  vl <- coal(get_board_option_or_null(id, session), board_option_value(x))
  nm <- coal(attr(vl, "chat_name"), "Default chat function")

  paste0(id, ": ", nm, opts)
}

#' @export
str_value.llm_model_option <- function(x, ...) {
  format(x, ...)
}

#' @export
validate_board_option.llm_model_option <- function(x) {

  val <- board_option_value(NextMethod())
  arg <- list(system_prompt = NULL, params = NULL)

  if (!is.function(val) || !identical(arg, as.list(formals(val)))) {
    blockr_abort(
      "Expecting `llm_model` to be a function with arguments ",
      "{paste0(names(arg), \" = \", arg)}.",
      class = "board_options_llm_model_invalid"
    )
  }

  nme <- attr(val, "chat_name")
  opt <- get_chat_fun_opts()

  is_fun <- is.function(opt) || (is.list(opt) && length(opt) == 1L)

  if (is_fun && !is.null(nme)) {
    blockr_abort(
      "Expecting `llm_model` name contain no attribute `chat_name`.",
      class = "board_options_llm_model_invalid"
    )
  }

  nms <- names(opt)

  if (is.list(opt) && length(opt) > 1L && !(is_string(nme) && nme %in% nms)) {
    blockr_abort(
      "Expecting `llm_model` name contain a string-valued attribute ",
      "`chat_name` among options {nms}.",
      class = "board_options_llm_model_invalid"
    )
  }

  invisible(x)
}

test_that("board opts", {

  opts1 <- new_board_options(
    new_board_name_option(),
    new_n_rows_option(),
    new_page_size_option()
  )

  expect_s3_class(opts1, "board_options")
  expect_length(opts1, 3L)
  expect_named(opts1, c("board_name", "n_rows", "page_size"))

  expect_s3_class(opts1[[1]], "board_option")
  expect_s3_class(opts1[["board_name"]], "board_option")

  expect_error(
    opts1[[1]] <- opts1[[2]],
    class = "board_options_duplicated_ids"
  )

  expect_error(
    opts1[["board_name"]] <- opts1[["n_rows"]],
    class = "board_options_duplicated_ids"
  )

  opts2 <- c(new_filter_rows_option(), opts1)

  expect_s3_class(opts2, "board_options")
  expect_length(opts2, 4L)
  expect_named(opts2, c("filter_rows", "board_name", "n_rows", "page_size"))

  opts3 <- c(opts1, new_filter_rows_option())

  expect_s3_class(opts3, "board_options")
  expect_length(opts3, 4L)
  expect_named(opts3, c("board_name", "n_rows", "page_size", "filter_rows"))

  chat_fun <- function(system_prompt = NULL, params = NULL) {}

  expect_warning(
    withr::with_options(
      list(blockr.chat_function = chat_fun),
      new_llm_model_option("foo")
    ),
    class = "single_llm_model"
  )

  expect_error(
    withr::with_options(
      list(blockr.chat_function = list(foo = chat_fun, bar = chat_fun)),
      new_llm_model_option("baz")
    ),
    class = "invalid_llm_model_option_value"
  )

  expect_error(
    withr::with_options(
      list(blockr.chat_function = list(foo = chat_fun, bar = chat_fun)),
      new_llm_model_option("baz")
    ),
    class = "invalid_llm_model_option_value"
  )

  expect_error(
    withr::with_options(
      list(blockr.chat_function = chat_fun),
      new_llm_model_option(chat_fun)
    ),
    class = "invalid_llm_model_option_value"
  )
})

test_that("opt ser/deser", {

  for (opt in default_board_options()) {
    expect_identical(
      opt,
      blockr_deser(blockr_ser(opt)),
      ignore_function_env = TRUE
    )
  }

  reg_opts <- combine_board_options(lapply(available_blocks(), board_options))

  for (opt in reg_opts) {
    expect_identical(
      opt,
      blockr_deser(blockr_ser(opt)),
      ignore_function_env = TRUE
    )
  }

  opts <- list(
    new_board_name_option("foo"),
    new_thematic_option(TRUE),
    new_dark_mode_option("dark"),
    new_show_conditions_option(character()),
    new_show_conditions_option("error"),
    new_show_conditions_option(c("warning", "error", "message"))
  )

  for (opt in opts) {
    expect_identical(
      opt,
      blockr_deser(blockr_ser(opt)),
      ignore_function_env = TRUE
    )
  }

  chat_fun <- function(system_prompt = NULL, params = NULL) {}

  withr::with_options(
    list(blockr.chat_function = chat_fun),
    {
      expect_identical(
        new_llm_model_option(),
        blockr_deser(blockr_ser(new_llm_model_option())),
        ignore_function_env = TRUE
      )
    }
  )

  withr::with_options(
    list(blockr.chat_function = list(foo = chat_fun, bar = chat_fun)),
    {
      expect_identical(
        new_llm_model_option(),
        blockr_deser(blockr_ser(new_llm_model_option())),
        ignore_function_env = TRUE
      )
    }
  )

  withr::with_options(
    list(blockr.chat_function = list(foo = chat_fun, bar = chat_fun)),
    {
      expect_identical(
        new_llm_model_option("bar"),
        blockr_deser(blockr_ser(new_llm_model_option("bar"))),
        ignore_function_env = TRUE
      )
    }
  )
})

test_that("a board options container has a compact str_value()", {

  opts <- as_board_options(new_board())

  res <- str_value(opts)
  lines <- strsplit(res, "\n")[[1L]]

  expect_length(res, 1L)
  expect_identical(lines[1L], "<board_options[4]>")
  expect_length(lines, 5L)
  expect_true(any(startsWith(lines[-1L], "  board_name: ")))

  expect_identical(str_value(new_board_options()), "<board_options[0]>")

  expect_identical(capture.output(str(opts))[1L], " <board_options[4]>")
})

test_that("opt utils", {

  opt <- new_board_option(
    "test",
    "test",
    ui = function(id) textInput(NS(id, "test"), "Test", "test"),
    ctor = "new_board_option",
    pkg = "blockr.core"
  )

  expect_true(is_board_option(opt))
  expect_identical(validate_board_option(opt), opt)
  expect_identical(as_board_option(opt), opt)

  expect_error(
    validate_board_option(list()),
    class = "board_option_inheritance_invalid"
  )

  expect_error(
    validate_board_option(structure(list(), id = 1L, class = "board_option")),
    class = "board_option_component_id_invalid"
  )

  expect_error(
    validate_board_option(
      structure(list(), id = "test", trigger = 1L, class = "board_option")
    ),
    class = "board_option_component_trigger_invalid"
  )

  expect_error(
    validate_board_option(
      structure(list(ui = "test"), id = "test", class = "board_option")
    ),
    class = "board_option_component_ui_invalid"
  )

  expect_error(
    validate_board_option(
      structure(
        list(ui = function(id) id, server = "test"),
        id = "test",
        class = "board_option"
      )
    ),
    class = "board_option_component_server_invalid"
  )

  expect_error(
    validate_board_option(
      structure(
        list(ui = function(id) id, server = function(id) id),
        id = "test",
        class = "board_option"
      )
    ),
    class = "board_option_component_server_invalid"
  )

  expect_error(
    validate_board_option(
      structure(
        list(
          ui = function(id) id,
          server = function(board, ..., session) { },
          transform = "test"
        ),
        id = "test",
        class = "board_option"
      )
    ),
    class = "board_option_component_transform_invalid"
  )

  expect_snapshot(opt)
})

test_that("board name opt", {

  expect_true(is_board_option(new_board_name_option()))

  expect_error(
    new_board_name_option(1L),
    class = "board_name_option_invalid"
  )
})

test_that("board n_rows opt", {

  expect_true(is_board_option(new_n_rows_option()))

  expect_error(
    new_n_rows_option(-1L),
    class = "board_options_n_rows_invalid"
  )
})

test_that("board page_size opt", {

  expect_true(is_board_option(new_page_size_option()))

  expect_error(
    new_page_size_option(-1L),
    class = "board_options_page_size_invalid"
  )
})

test_that("board filter_rows opt", {

  expect_true(is_board_option(new_filter_rows_option()))

  expect_error(
    new_filter_rows_option(-1L),
    class = "board_options_filter_rows_invalid"
  )
})

test_that("board thematic opt", {

  expect_true(is_board_option(new_thematic_option()))

  expect_error(
    new_thematic_option(-1L),
    class = "board_options_thematic_invalid"
  )

  expect_error(
    with_mocked_bindings(
      new_thematic_option(TRUE),
      pkg_avail = function(...) FALSE
    ),
    class = "thematic_not_installed"
  )
})

test_that("board dark_mode opt", {

  expect_true(is_board_option(new_dark_mode_option()))
  expect_true(is_board_option(new_dark_mode_option(TRUE)))
  expect_true(is_board_option(new_dark_mode_option(FALSE)))

  expect_error(
    new_dark_mode_option(-1L),
    class = "board_options_dark_mode_invalid"
  )
})

test_that("board show_conditions opt", {

  expect_true(is_board_option(new_show_conditions_option()))

  expect_error(
    new_show_conditions_option(-1L),
    class = "board_options_show_conditions_invalid"
  )
})

test_that("llm model opt", {

  opt0 <- withr::with_options(
    list(blockr.chat_function = default_chat),
    new_llm_model_option()
  )

  expect_null(board_option_ui(opt0, "chat"))

  chat_fun <- function(system_prompt = NULL, params = NULL) {
    list(system_prompt = system_prompt, params = params)
  }

  opt1 <- withr::with_options(
    list(blockr.chat_function = chat_fun),
    new_llm_model_option()
  )

  expect_null(board_option_ui(opt1, "chat"))
  expect_null(board_option_server(opt1, board = "foo"))
  expect_null(board_option_trigger(opt1))

  expect_identical(board_option_value(opt1), chat_fun)

  withr::with_options(
    list(blockr.chat_function = chat_fun),
    expect_snapshot(opt1)
  )

  opt2 <- withr::with_options(
    list(blockr.chat_function = list(chat1 = chat_fun, chat2 = chat_fun)),
    new_llm_model_option()
  )

  expect_s3_class(board_option_ui(opt2, "chat"), "shiny.tag")

  withr::with_options(
    list(blockr.chat_function = list(chat1 = chat_fun, chat2 = chat_fun)),
    expect_snapshot(opt2)
  )

  sess <- MockShinySession$new()
  sess$setInputs(llm_model = "chat1")

  board_option_to_userdata(opt2, session = sess)

  expect_s3_class(board_option_server(opt2, session = sess), "Observer")

  res <- board_option_value(opt2)
  fun <- attr(res, "chat_name")

  expect_identical(fun, "chat1")

  opt3 <- withr::with_options(
    list(blockr.chat_function = list(chat1 = chat_fun, chat2 = chat_fun)),
    new_llm_model_option("chat2")
  )

  expect_true(is_board_option(opt3))

  opt4 <- withr::with_options(
    list(blockr.chat_function = list(chat1 = chat_fun)),
    new_llm_model_option("chat1")
  )

  expect_true(is_board_option(opt4))

  expect_error(
    withr::with_options(
      list(blockr.chat_function = identity),
      new_llm_model_option()
    ),
    class = "board_options_llm_model_invalid"
  )

  expect_error(
    withr::with_options(
      list(blockr.chat_function = "test"),
      new_llm_model_option()
    ),
    class = "board_options_llm_model_invalid"
  )
})

test_that("a board option has a compact str_value()", {

  opt <- new_board_name_option("My board")

  expect_identical(str_value(opt), "board_name: My board")
  expect_identical(str_value(opt), format(opt))

  expect_identical(capture.output(str(opt)), " board_name: My board")

  chat_fun <- function(system_prompt = NULL, params = NULL) NULL

  llm <- withr::with_options(
    list(blockr.chat_function = chat_fun),
    new_llm_model_option()
  )

  expect_identical(str_value(llm), "llm_model: Default chat function")
  expect_identical(str_value(llm), format(llm))
})

#' Plugin module for managing board blocks
#'
#' Logic and user experience for adding/removing blocks to the board can be
#' customized or enhanced by providing an alternate version of this plugin. The
#' default implementation provides a modal-based UI with simple shiny inputs
#' such as drop-downs and text fields.
#'
#' Updates are mediated via the [shiny::reactiveVal()] object passed as
#' `update`, where block updates are communicated as list entry `blocks` with
#' components `add` and `rm`, where
#' * `add` may be `NULL` or a `block` object (block IDs may not already exist),
#' * `rm` may be `NULL` or a string (of existing block IDs).
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `manage_blocks` is returned by
#' `manage_blocks()`, while the UI component (e.g. `manage_blocks_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `manage_blocks_server()`) is expected to return `NULL`.
#'
#' @export
manage_blocks <- function(server = manage_blocks_server,
                          ui = manage_blocks_ui) {

  new_plugin(server, ui, class = "manage_blocks")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param update Reactive value object to initiate board updates
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname manage_blocks
#' @export
manage_blocks_server <- function(id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      observeEvent(
        input$add_block,
        showModal(add_block_modal(session$ns))
      )

      observeEvent(
        input$confirm_add,
        {
          sel <- input$registry_select
          bid <- input$block_id

          if (!validate_block_addition(sel, bid, board$board, session)) {
            return()
          }

          blk <- create_block(sel, name = id_to_sentence_case(bid))
          add <- as_blocks(set_names(list(blk), bid))

          update(
            list(blocks = list(add = add))
          )

          removeModal()
        }
      )

      observeEvent(
        input$cancel_add,
        removeModal()
      )

      observeEvent(
        input$rm_block,
        rm_block_modal(session$ns, board$board)
      )

      observeEvent(
        input$confirm_rm,
        {
          sel <- input$block_select

          if (!length(sel)) {

            notify(
              "Please choose at least one block.",
              type = "warning",
              session = session
            )

            return()
          }

          if (!all(sel %in% board_block_ids(board$board))) {

            notify(
              "Please choose valid block IDs.",
              type = "warning",
              session = session
            )

            return()
          }

          update(
            list(blocks = list(rm = sel))
          )

          removeModal()
        }
      )

      observeEvent(
        input$cancel_rm,
        removeModal()
      )

      NULL
    }
  )
}

#' @param board The initial `board` object
#' @rdname manage_blocks
#' @export
manage_blocks_ui <- function(id, board) {
  tagList(
    actionButton(
      NS(id, "add_block"),
      "Add block",
      icon = icon("circle-plus"),
      class = "btn-success"
    ),
    actionButton(
      NS(id, "rm_block"),
      "Remove block",
      icon = icon("circle-minus"),
      class = "btn-danger"
    )
  )
}

add_block_modal <- function(ns) {
  modalDialog(
    title = "Add block",
    div(
      selectInput(
        ns("registry_select"),
        "Select block from registry",
        choices = list_blocks()
      ),
      textInput(
        inputId = ns("block_id"),
        label = "Block ID",
        value = rand_names(),
        placeholder = "Enter a block ID."
      )
    ),
    footer = tagList(
      actionButton(ns("cancel_add"), "Cancel", class = "btn-danger"),
      actionButton(ns("confirm_add"), "OK", class = "btn-success")
    )
  )
}

rm_block_modal <- function(ns, board) {
  showModal(
    modalDialog(
      title = "Remove blocks",
      selectInput(
        ns("block_select"),
        "Select block(s) from board",
        choices = board_block_ids(board),
        multiple = TRUE
      ),
      footer = tagList(
        actionButton(ns("cancel_rm"), "Cancel", class = "btn-danger"),
        actionButton(ns("confirm_rm"), "OK", class = "btn-success")
      )
    )
  )
}

validate_block_addition <- function(block, id, board, session) {

  if (nchar(id) == 0L || !is_string(id)) {

    notify(
      "Please choose a valid block ID.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  if (id %in% board_block_ids(board)) {

    notify(
      "Please choose a unique block ID.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  if (!is_string(block) || !block %in% list_blocks()) {

    notify(
      "Please choose a valid block type.",
      type = "warning",
      session = session
    )

    return(FALSE)
  }

  TRUE
}

#' @section Upload block:
#' In order to make user data available to blockr, this block provides file-
#' upload functionality via [shiny::fileInput()]. Given that data provided in
#' this way are only available for the life-time of the shiny session, exported
#' code is not self-contained and a script containing code from an upload block
#' is cannot be run in a new session. Also, serialization of upload blocks is
#' currently not allowed as the full data would have to be included during
#' serialization.
#'
#' @block data upload block
#' @blockDescr Upload data
#' @blockCategory input
#' @blockIcon upload
#' @blockGuidance Provides a data.frame from a file uploaded in the running app.
#'   Uploaded data lives only for the session, so serialization and reproducible
#'   export are unavailable. Has no configurable arguments.
#' @blockKeywords upload, file, import, data, session
#'
#' @rdname new_file_block
#' @export
new_upload_block <- function(...) {
  new_file_block(
    function(id) {
      moduleServer(
        id,
        function(input, output, session) {
          list(
            expr = reactive(
              bquote(.(file), list(file = input$upload$datapath))
            ),
            state = list()
          )
        }
      )
    },
    function(id) {
      fileInput(
        NS(id, "upload"),
        "Upload data"
      )
    },
    class = "upload_block",
    ...
  )
}

#' Locked boards
#'
#' A board can be deployed read-only by setting the `blockr.locked`
#' option (see [blockr_option()]). Locking is enforced server-side, not
#' by UI hiding: a forged `Shiny.setInputValue` bypasses any
#' `display: none` and fires the underlying observer, so the refusal has
#' to happen on the server. While locked, the two channels every
#' mutation funnels through -- the `board_update` lifecycle (see
#' [board_update()]) and [set_board_option_value()] -- refuse to apply
#' changes.
#'
#' `is_board_locked()` reports whether a board is locked and dispatches
#' on the board, so a subclass can source the state however it likes
#' (for example gating an unlock behind an authenticated request). The
#' default `board` method reads the `blockr.locked` option, and
#' `blockr.dock`'s `is_dock_locked()` delegates here.
#'
#' @section Scope: Locking guards against forged *client input* -- a
#' browser that sets Shiny inputs but cannot run server-side R. It is
#' **not** a defense against code executing in the session: blockr
#' evaluates arbitrary R, which can flip the option, rebind this generic
#' or mutate board state directly, and no in-process flag can prevent
#' that. A subclass method may raise the bar (e.g. an authenticated
#' unlock) but cannot close the gap on its own -- a hard boundary has to
#' live where the user's code does not run (a sandboxed evaluator, or an
#' ephemeral read-only deployment). Treat locking as a guard for
#' non-malicious users.
#'
#' @param board A `board` object to dispatch on.
#' @param ... Generic consistency.
#'
#' @return A logical flag.
#'
#' @examples
#' is_board_locked(new_board())
#'
#' @name locked-board
#' @export
is_board_locked <- function(board, ...) {
  UseMethod("is_board_locked", board)
}

#' @rdname locked-board
#' @export
is_board_locked.board <- function(board, ...) {
  isTRUE(blockr_option("locked", FALSE))
}

#' @section Block vectors:
#' Multiple blocks can be combined into a `blocks` object, a container for
#' an (ordered) set of blocks. Block IDs are handled at the `blocks` level
#' which will ensure uniqueness.
#'
#' @examples
#' blks <- c(a = new_dataset_block(), b = new_subset_block())
#'
#' is_block(blks)
#' is_blocks(blks)
#'
#' names(blks)
#'
#' tryCatch(
#'   names(blks["a"]) <- "b",
#'   error = function(e) conditionMessage(e)
#' )
#'
#' @rdname new_block
#' @export
blocks <- function(...) {

  args <- list(...)

  args <- lapply(args, as_block)

  if (is.null(names(args))) {
    names(args) <- rand_names(n = length(args))
  }

  miss <- is.na(names(args)) | names(args) == ""

  if (any(miss)) {
    names(args)[miss] <- rand_names(names(args)[!miss], sum(miss))
  }

  validate_blocks(
    new_vctr(args, class = "blocks")
  )
}

#' @rdname new_block
#' @export
is_blocks <- function(x) {
  inherits(x, "blocks")
}

#' @rdname new_block
#' @export
as_blocks <- function(x, ...) {
  UseMethod("as_blocks")
}

#' @export
as_blocks.blocks <- function(x, ...) {
  validate_blocks(x)
}

#' @export
as_blocks.list <- function(x, ...) {
  do.call(blocks, x)
}

#' @export
as_blocks.block <- function(x, ...) {
  as_blocks(list(x))
}

#' @export
as_blocks.board <- function(x, ...) {
  board_blocks(x)
}

#' @export
as.list.blocks <- function(x, ...) {
  vec_data(x)
}

#' @export
`names<-.blocks` <- function(x, value) {

  if (is.null(value)) {
    value <- rep("", length(x))
  } else if (anyDuplicated(value) != 0L) {
    blockr_abort(
      "IDs are required to be unique.",
      class = "blocks_names_unique_invalid"
    )
  }

  attr(x, "names") <- value

  x
}

#' @export
format.blocks <- function(x, ...) {

  out <- lapply(x, format)
  out <- Map(c, names(x), lapply(out, function(x) paste0(" ", x)))
  out <- lapply(out, c, "")
  out <- unlst(out)

  c(
    paste0("<", class(x)[1L], "[", length(x), "]>"),
    if (length(x)) "",
    out[-length(out)]
  )
}

#' @export
print.blocks <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.blocks <- function(x, ...) {
  str_value_collection(x, chr_ply(x, str_value))
}

#' @export
str.blocks <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
vec_restore.blocks <- function(x, to, ...) {
  validate_blocks(NextMethod())
}

harmonize_list_of_blocks <- function(x) {
  if (is_block(x)) {
    list(x)
  } else if (is_blocks(x)) {
    as.list(x)
  } else if (is.list(x) && all(lgl_ply(x, is_block))) {
    x
  } else {
    list(as_block(x))
  }
}

#' @export
c.blocks <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_blocks),
    recursive = FALSE
  )

  as_blocks(res)
}

#' @export
`[<-.blocks` <- function(x, i, ..., value) {

  i <- vec_as_location(i, length(x), names(x))

  if (is.null(value)) {
    return(blocks_slice(x, -i))
  }

  value <- as_blocks(value)

  trg_ids <- names(x)[i]

  if (is.null(names(value))) {
    names(value) <- trg_ids
  }

  new_ids <- names(value)

  if (!setequal(new_ids, trg_ids)) {
    blockr_abort(
      paste0(
        "Replacing IDs ", paste_enum(trg_ids), " with ", paste_enum(new_ids),
        " is not allowed."
      ),
      class = "blocks_assignment_name_invalid"
    )
  }

  blocks_assign(x, i, value[trg_ids])
}

#' @export
`[[<-.blocks` <- function(x, i, ..., value) {

  i <- vec_as_location2(i, length(x), names(x))
  val <- set_names(list(as_block(value)), names(x)[i])

  blocks_assign(x, i, as_blocks(val))
}

blocks_slice <- function(...) {
  validate_blocks(vec_slice(...))
}

blocks_assign <- function(...) {
  validate_blocks(vec_assign(...))
}

validate_blocks <- function(x) {

  if (!is_blocks(x)) {
    blockr_abort(
      "Expecting blocks to inherit from \"blocks\".",
      class = "blocks_class_invalid"
    )
  }

  if (!is.list(x) || !all(lgl_ply(x, is_block))) {
    blockr_abort(
      "Expecting the board to contain a set of blocks.",
      class = "blocks_contains_invalid"
    )
  }

  ids <- names(x)

  if (length(ids) != length(x) || any(is.na(ids) | !nchar(ids))) {
    blockr_abort(
      "Block IDs are required to be nonempty strings.",
      class = "blocks_names_invalid"
    )
  }

  if (anyDuplicated(ids) != 0) {
    blockr_abort(
      "Block IDs are required to be unique.",
      class = "blocks_names_invalid"
    )
  }

  x
}

#' Serialization utilities
#'
#' Blocks are serialized by writing out information on the constructor used to
#' create the object, combining this with block state information, which
#' constitutes values such that when passed to the constructor the original
#' object can be re-created.
#'
#' Helper functions `blockr_ser()` and `blockr_deser()` are implemented as
#' generics and perform most of the heavy lifting for (de-)serialization:
#' representing objects as easy-to-serialize (nested) lists containing mostly
#' strings and no objects which are hard/impossible to truthfully re-create in
#' new sessions (such as environments).
#'
#' During deserialization, `blockr_deser()` forwards `...` to the dispatched
#' per-class method. This lets callers (and outer methods deserializing nested
#' objects) thread additional context down to inner deserializers.
#'
#' @param x Object to (de)serialize
#' @param ... Generic consistency
#'
#' @examples
#' blk <- new_dataset_block("iris")
#'
#' blockr_ser(blk)
#'
#' all.equal(blk, blockr_deser(blockr_ser(blk)), check.environment = FALSE)
#'
#' @return Serialization helper function `blockr_ser()` returns lists, which
#' for most objects contain slots `object` and `payload`, where `object`
#' contains a class vector which is used by `blockr_deser()` to instantiate an
#' empty object of that class and use S3 dispatch to identify the correct method
#' that, given the content in `payload`, can re-create the original object.
#'
#' @export
blockr_ser <- function(x, ...) {
  UseMethod("blockr_ser")
}

#' @param state Object state (as returned from an `expr_server`)
#' @rdname blockr_ser
#' @export
blockr_ser.block <- function(x, state = NULL, ...) {

  if (is.null(state)) {
    state <- initial_block_state(x)
  }

  attrs <- attributes(x)

  state <- c(
    state,
    attrs[setdiff(names(attrs), c("names", internal_block_attributes()))]
  )

  ctor <- block_ctor(x)

  list(
    object = class(x),
    payload = state,
    constructor = blockr_ser(ctor)
  )
}

#' @param blocks Block states (`NULL`, or a per-block `NULL` entry, defaults to
#' values from the constructor scope)
#' @rdname blockr_ser
#' @export
blockr_ser.blocks <- function(x, blocks = NULL, ...) {

  if (!length(blocks)) {

    res <- lapply(x, blockr_ser)

  } else {

    stopifnot(
      is.list(blocks),
      all(lgl_ply(blocks, function(b) is.null(b) || is.list(b))),
      length(blocks) == length(x), setequal(names(blocks), names(x))
    )

    res <- Map(blockr_ser, x, blocks[names(x)])
  }

  list(
    object = class(x),
    payload = res
  )
}

#' @param options Board option values (`NULL` uses values provided by `x`)
#' @rdname blockr_ser
#' @export
blockr_ser.board_options <- function(x, options = NULL, ...) {

  if (is.null(options)) {
    options <- board_option_values(x)
  }

  expected <- names(x)

  stopifnot(is.list(options), setequal(expected, names(options)))

  list(
    object = class(x),
    payload = map(blockr_ser, x, options[expected])
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.blockr_ctor <- function(x, ...) {

  fun <- attr(x, "fun")
  pkg <- attr(x, "pkg")

  if (is.null(fun)) {
    fun <- serialize(x, NULL)
    ver <- NULL
  } else {
    ver <- as.character(pkg_version(pkg))
  }

  list(
    object = class(x),
    constructor = fun,
    package = pkg,
    version = ver
  )
}

#' @param option Board option value (`NULL` uses values provided by `x`)
#' @rdname blockr_ser
#' @export
blockr_ser.board_option <- function(x, option = NULL, ...) {

  val <- coal(option, board_option_value(x), list())

  if (is.atomic(val)) {
    val <- list(val)
  }

  ctor <- board_option_ctor(x)
  name <- setdiff(names(formals(ctor)), c("...", "category"))

  if (is.null(names(val))) {

    if (length(val) && length(name) != length(val)) {
      blockr_abort(
        "Cannot match option values to constructor arguments for option ",
        "{board_option_id(x)}.",
        class = "option_value_arg_number_mismatch"
      )
    }

    if (length(val) && length(name) > 1L) {
      log_warn(
        "Uncertainty matching option values to constructor arguments for ",
        "option {board_option_id(x)}. Consider providing a specific ",
        "`blockr_ser()` method."
      )
    }

    if (length(val)) {
      names(val) <- name
    }
  }

  unknown <- setdiff(names(val), name)

  if (length(unknown)) {
    blockr_abort(
      "Unrecognized option {qty(unknown)} argument{?s} {unknown} for option ",
      "{board_option_id(x)}.",
      class = "option_value_arg_name_mismatch"
    )
  }

  list(
    object = class(x),
    payload = c(val, list(category = board_option_category(x))),
    constructor = blockr_ser(ctor)
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.llm_model_option <- function(x, option = NULL, ...) {
  NextMethod(
    option = coal(
      attr(coal(option, board_option_value(x)), "chat_name"),
      list()
    )
  )
}

#' @param board_id Board ID
#' @rdname blockr_ser
#' @export
blockr_ser.board <- function(x, board_id = NULL, ...) {

  stopifnot(is.null(board_id) || is_string(board_id))

  list(
    object = class(x),
    payload = Map(
      blockr_ser, x, list(...)[names(x)]
    ),
    constructor = blockr_ser(board_ctor(x)),
    version = as.character(pkg_version()),
    id = board_id
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.link <- function(x, ...) {
  list(
    object = class(x),
    payload = as.list(x),
    constructor = blockr_ser(attr(x, "ctor"))
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.links <- function(x, ...) {
  list(
    object = class(x),
    payload = lapply(x, blockr_ser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.stack <- function(x, ...) {
  list(
    object = class(x),
    payload = as.list(x),
    constructor = blockr_ser(attr(x, "ctor"))
  )
}

#' @rdname blockr_ser
#' @export
blockr_ser.stacks <- function(x, ...) {
  list(
    object = class(x),
    payload = lapply(x, blockr_ser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser <- function(x, ...) {
  UseMethod("blockr_deser")
}

#' @rdname blockr_ser
#' @export
blockr_deser.list <- function(x, ...) {

  stopifnot("object" %in% names(x))

  cls <- x[["object"]]

  res <- blockr_deser(
    structure(list(), class = cls),
    data = x,
    ...
  )

  if (!identical(class(res), cls)) {
    blockr_abort(
      "Could not deserialize object: expected {qty(cls)} class{?es} {cls}, ",
      "but received {qty(res)} {class(res)}.",
      class = "block_deser_class_error"
    )
  }

  res
}

#' @param data List valued data (converted from JSON)
#' @rdname blockr_ser
#' @export
blockr_deser.block <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "payload") %in% names(data))
  )

  ctor <- blockr_deser(data[["constructor"]])

  args <- c(
    data[["payload"]],
    list(
      ctor = coal(ctor_name(ctor), ctor_fun(ctor)),
      ctor_pkg = ctor_pkg(ctor)
    )
  )

  do.call(ctor_fun(ctor), args)
}

#' @rdname blockr_ser
#' @export
blockr_deser.blocks <- function(x, data, ...) {
  as_blocks(
    lapply(data[["payload"]], blockr_deser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.board <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "payload") %in% names(data))
  )

  ctor <- blockr_deser(data[["constructor"]])

  args <- c(
    lapply(data[["payload"]], blockr_deser),
    list(
      ctor = coal(ctor_name(ctor), ctor_fun(ctor)),
      pkg = ctor_pkg(ctor)
    )
  )

  res <- do.call(ctor_fun(ctor), args)

  attr(res, "id") <- data[["id"]]

  res
}

#' @rdname blockr_ser
#' @export
blockr_deser.link <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "payload") %in% names(data))
  )

  do.call(
    blockr_deser(data[["constructor"]]),
    data[["payload"]]
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.links <- function(x, data, ...) {
  as_links(
    lapply(data[["payload"]], blockr_deser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.stack <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "payload") %in% names(data))
  )

  do.call(
    blockr_deser(data[["constructor"]]),
    data[["payload"]]
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.stacks <- function(x, data, ...) {
  as_stacks(
    lapply(data[["payload"]], blockr_deser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.board_options <- function(x, data, ...) {
  as_board_options(
    lapply(data[["payload"]], blockr_deser)
  )
}

#' @rdname blockr_ser
#' @export
blockr_deser.blockr_ctor <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "package") %in% names(data))
  )

  pkg <- data[["package"]]
  ctr <- data[["constructor"]]

  if (is.null(pkg)) {
    new_blockr_ctor(unserialize(ctr))
  } else if (pkg_avail(pkg)) {
    new_blockr_ctor(NULL, ctr, pkg)
  } else {
    blockr_abort(
      "Cannot deserialize object that depends on not installed ",
      "package {pkg}",
      class = "blockr_deser_missing_pkg"
    )
  }
}

#' @rdname blockr_ser
#' @export
blockr_deser.board_option <- function(x, data, ...) {

  stopifnot(
    all(c("constructor", "payload") %in% names(data))
  )

  ctor <- blockr_deser(data[["constructor"]])

  args <- c(
    data[["payload"]],
    list(
      ctor = coal(ctor_name(ctor), ctor_fun(ctor)),
      pkg = ctor_pkg(ctor)
    )
  )

  do.call(ctor_fun(ctor), args)
}

test_that("serialization", {

  blk_1 <- new_dataset_block("iris", "datasets")

  expect_identical(
    blk_1,
    blockr_deser(blockr_ser(blk_1)),
    ignore_function_env = TRUE
  )

  ser_1 <- blockr_ser(
    new_dataset_block(),
    list(dataset = "iris", package = "datasets")
  )

  expect_identical(
    new_dataset_block("iris", "datasets"),
    blockr_deser(ser_1),
    ignore_function_env = TRUE
  )

  blks_1 <- c(
    a = new_dataset_block(),
    b = new_subset_block()
  )

  expect_identical(
    blks_1,
    blockr_deser(blockr_ser(blks_1)),
    ignore_function_env = TRUE
  )

  blks_2 <- blocks(
    a = new_dataset_block("iris", "datasets")
  )

  json <- blockr_ser(
    blocks(a = new_dataset_block()),
    list(
      a = list(dataset = "iris", package = "datasets")
    )
  )

  expect_identical(blks_2, blockr_deser(json), ignore_function_env = TRUE)

  lnks <- links(from = "a", to = "b")

  expect_identical(lnks, blockr_deser(blockr_ser(lnks)))

  stks <- stacks(ab = c("a", "b"))

  expect_identical(stks, blockr_deser(blockr_ser(stks)))

  board <- new_board(
    blocks = c(
      a = new_dataset_block(),
      b = new_subset_block()
    ),
    links = links(from = "a", to = "b")
  )

  expect_identical(
    board,
    blockr_deser(blockr_ser(board)),
    ignore_function_env = TRUE
  )

  new_identity_block <- function(...) {
    new_transform_block(
      function(id, data) {
        moduleServer(
          id,
          function(input, output, session) {
            list(
              expr = reactive(quote(identity(data))),
              state = list()
            )
          }
        )
      },
      function(id) {
        tagList()
      },
      class = "identity_block",
      block_metadata = list(),
      ...
    )
  }

  blk_2 <- new_identity_block()

  expect_identical(
    blk_2,
    blockr_deser(blockr_ser(blk_2)),
    ignore_function_env = TRUE
  )

  blk_3 <- with_mocked_bindings(
    new_identity_block(
      ctor = "new_identity_block",
      ctor_pkg = "some_imaginary_package"
    ),
    get_ctor_from_pkg = function(...) new_identity_block
  )

  expect_error(
    blockr_deser(
      with_mocked_bindings(blockr_ser(blk_3), pkg_version = function(x) "1.0.0")
    ),
    class = "blockr_deser_missing_pkg"
  )

  ser_2 <- blockr_ser(blk_1)

  ser_2$object <- c("foo_block", ser_2$object)

  expect_error(
    blockr_deser(ser_2),
    class = "block_deser_class_error"
  )

  new_test_opt <- function(val1 = 1, val2 = "2", ...) {
    new_board_option(
      "test_opt",
      list(val1 = val1, val2 = val2),
      ui = function(id) tagList()
    )
  }

  opt <- new_test_opt()

  expect_error(
    blockr_ser(opt, option = 3),
    class = "option_value_arg_number_mismatch"
  )

  expect_silent(blockr_ser(opt, option = list(val1 = 3, val2 = "4")))
  expect_output(blockr_ser(opt, option = list(3, "4")))

  expect_error(
    blockr_ser(opt, option = list(val1 = 3, val2 = "4", val3 = "boom")),
    class = "option_value_arg_name_mismatch"
  )
})

test_that("blockr_deser.list forwards `...` to per-class methods", {

  captured <- NULL

  registerS3method(
    "blockr_deser", "deser_ctx_probe",
    function(x, data, version = NULL, ...) {
      captured <<- version
      structure(list(), class = "deser_ctx_probe")
    },
    envir = asNamespace("blockr.core")
  )

  with_ctx <- blockr_deser(
    list(object = "deser_ctx_probe"),
    version = "1.2.3"
  )

  expect_s3_class(with_ctx, "deser_ctx_probe")
  expect_identical(captured, "1.2.3")

  no_ctx <- blockr_deser(list(object = "deser_ctx_probe"))

  expect_s3_class(no_ctx, "deser_ctx_probe")
  expect_null(captured)
})

test_that("a board with variadic links round-trips and still evaluates", {

  board <- new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("BOD"),
      c = new_rbind_block()
    ),
    links = links(
      ac = new_link("a", "c"),
      bc = new_link("b", "c")
    )
  )

  expect_identical(
    board,
    blockr_deser(blockr_ser(board)),
    ignore_function_env = TRUE
  )

  restored <- blockr_deser(blockr_ser(board))

  testServer(
    get_s3_method("board_server", restored),
    {
      session$flushReact()

      expect_identical(
        rv$blocks$c$server$result(),
        rbind(datasets::BOD, datasets::BOD)
      )
    },
    args = list(x = restored)
  )
})

test_that("fixed block constructor", {

  blk <- new_fixed_block(quote(utils::head(data)))

  expect_s3_class(blk, "fixed_block")

  expect_error(new_fixed_block())
  expect_error(new_fixed_block(123), class = "invalid_block_input")

  testServer(
    get_s3_method("block_server", blk),
    {
      session$flushReact()
      expect_identical(
        session$returned$result(),
        head(datasets::mtcars)
      )
    },
    args = list(
      x = blk,
      data = list(data = function() datasets::mtcars)
    )
  )

  expect_identical(
    blk,
    blockr_deser(blockr_ser(blk)),
    ignore_function_env = TRUE
  )
})

#' Plugin module for managing board stacks
#'
#' Logic and user experience for adding new, removing and modifying existing
#' stacks to/from the board can be customized or enhanced by providing an
#' alternate version of this plugin. The default implementation provides a
#' table-based UI, presented in a modal.
#'
#' Updates are mediated via the [shiny::reactiveVal()] object passed as
#' `update`, where stack updates are communicated as list entry `stacks` with
#' components `add`, `rm` or `mod`, where
#' * `add` is either `NULL` or a `stacks` object (stack IDs may not already
#'   exists),
#' * `rm` is either `NULL` or a character vector of (existing) stack IDs,
#' * `mod` is either `NULL` or a `stacks` object (where stack IDs must already
#'   exist).
#'
#' @param server,ui Server/UI for the plugin module
#'
#' @return A plugin container inheriting from `manage_stacks` is returned by
#' `manage_stacks()`, while the UI component (e.g. `manage_stacks_ui()`) is
#' expected to return shiny UI (i.e. [shiny::tagList()]) and the server
#' component (i.e. `manage_stacks_server()`) is expected to return `NULL`.
#'
#' @export
manage_stacks <- function(server = manage_stacks_server,
                          ui = manage_stacks_ui) {

  new_plugin(server, ui, class = "manage_stacks")
}

#' @param id Namespace ID
#' @param board Reactive values object
#' @param update Reactive value object to initiate board updates
#' @param ... Extra arguments passed from parent scope
#'
#' @rdname manage_stacks
#' @export
manage_stacks_server <- function(id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {

      observeEvent(
        input$stacks_mod,
        showModal(stacks_modal(session$ns))
      )

      upd <- reactiveValues(
        add = stacks(),
        rm = character(),
        mod = stacks(),
        curr = isolate(board_stacks(board$board)),
        obs = list(),
        edit = NULL
      )

      applied_stacks <- deduped_board_reactive(board, board_stacks)

      observeEvent(
        applied_stacks(),
        {
          upd$curr <- merge_staged_stacks(
            applied_stacks(), upd$add, upd$rm, upd$mod
          )
        }
      )

      output$stacks_dt <- DT::renderDataTable(
        stack_dt(isolate(upd$curr), session$ns, isolate(board$board)),
        server = TRUE
      )

      stacks_proxy <- DT::dataTableProxy("stacks_dt", session)

      create_stack_obs_observer(input, board, upd, session, stacks_proxy)

      edit_stack_observer(upd, board)

      add_stack_observer(input, board, upd, session)

      rm_stack_observer(input, board, upd, session)

      cancel_stack_observer(input, board, upd, session)

      modify_stack_observer(input, board, upd, session, stacks_proxy, update)

      NULL
    }
  )
}

#' @param board The initial `board` object
#' @rdname manage_stacks
#' @export
manage_stacks_ui <- function(id, board) {
  tagList(
    actionButton(
      NS(id, "stacks_mod"),
      "Edit stacks",
      icon = icon("stack-overflow")
    )
  )
}

stacks_modal <- function(ns) {

  new_id_ns <- ns("new_stack_id")

  modalDialog(
    title = "Board stacks",
    DT::dataTableOutput(ns("stacks_dt")),
    footer = tagList(
      htmltools::tagQuery(
        textInput(new_id_ns, NULL, placeholder = "Next ID")
      )$addAttrs(
        style = "width: 180px; margin: 0 8px;"
      )$allTags(),
      tags$style(
        type = "text/css",
        paste0(
          "#", new_id_ns, " {padding: 0.75em 2em; margin: 4px;}"
        )
      ),
      tags$style(
        type = "text/css",
        paste0(
          "#", new_id_ns, "-label {text-align: left;}"
        )
      ),
      actionButton(ns("add_stack"), "Add", icon = icon("plus")),
      actionButton(ns("rm_stack"), "Remove", icon = icon("minus")),
      actionButton(ns("cancel_stacks"), "Cancel", class = "btn-danger"),
      actionButton(ns("modify_stacks"), "OK", class = "btn-success")
    ),
    size = "l"
  )
}

stack_dt <- function(dat, ns, board) {
  res <- DT::datatable(
    dt_board_stack(dat, ns, board),
    options = list(
      pageLength = 5,
      preDrawCallback = DT::JS(
        "function() { Shiny.unbindAll(this.api().table().node()); }"
      ),
      drawCallback = DT::JS(
        "function() { Shiny.bindAll(this.api().table().node()); }"
      ),
      dom = "tp",
      ordering = FALSE,
      columnDefs = list(
        list(targets = 0, width = "125px")
      )
    ),
    rownames = FALSE,
    escape = FALSE
  )

  DT::formatStyle(res, 1L, `vertical-align` = "middle")
}

dt_board_stack <- function(stk, ns, board) {

  ids <- names(stk)

  data.frame(
    ID = ids,
    Name = chr_mply(
      dt_text,
      lapply(paste0(ids, "_name"), ns),
      chr_ply(stk, attr, "name")
    ),
    Blocks = chr_mply(
      dt_select,
      lapply(paste0(ids, "_blocks"), ns),
      lapply(stk, as.character),
      MoreArgs = list(
        rem = available_stack_blocks(
          stacks = stk,
          blocks = board_block_ids(board)
        )
      )
    )
  )
}

dt_select <- function(id, val, rem) {

  res <- selectInput(id, label = NULL, choices = c("", val, rem),
                     selected = val, multiple = TRUE)

  res <- htmltools::tagQuery(
    res
  )$addAttrs(
    style = "width: 400px; margin-bottom: 0;"
  )$allTags()

  as.character(res)
}

dt_text <- function(id, val) {

  res <- textInput(id, label = NULL, value = val)

  res <- htmltools::tagQuery(
    res
  )$addAttrs(
    style = "width: 200px; margin-bottom: 0;"
  )$allTags()

  as.character(res)
}

create_dt_stack_obs <- function(ids, upd, ...) {

  create_obs <- function(col, row, upd, input, blks, sess) {

    inp <- paste0(row, "_", col)

    log_debug("creating stack DT observer ", inp)

    observeEvent(
      input[[inp]],
      {
        new <- input[[inp]]

        if (col == "blocks" && setequal(stack_blocks(upd$curr[[row]]), new)) {
          return()
        }

        if (col == "name" && identical(stack_name(upd$curr[[row]]), new)) {
          return()
        }

        if (col == "blocks") {

          if (is.null(new)) {
            new <- character()
          }

          rem <- setdiff(names(upd$curr), row)
          ava <- available_stack_blocks(
            stacks = c(upd$curr[rem], new),
            blocks = names(blks)
          )

          for (i in rem) {

            cur <- stack_blocks(upd$curr[[i]])

            updateSelectInput(
              sess,
              paste0(i, "_blocks"),
              label = NULL,
              choices = c(cur, ava),
              selected = cur
            )
          }

        } else if (col != "name") {

          blockr_abort(
            "Unexpected input: column {col}.",
            class = "unexpected_stack_col_input"
          )
        }

        upd$edit <- list(row = row, col = col, val = new)
      },
      ignoreInit = TRUE,
      ignoreNULL = FALSE
    )
  }

  create_obs_for_id <- function(id, ...) {
    lapply(
      set_names(nm = c("name", "blocks")),
      create_obs,
      id,
      ...
    )
  }

  upd$obs[ids] <- lapply(ids, create_obs_for_id, upd, ...)

  upd
}

destroy_dt_stack_obs <- function(ids, update) {

  for (row in ids) {
    for (col in c("name", "blocks")) {
      log_debug("destroying stack DT observer ", row, " ", col)
      update$obs[[row]][[col]]$destroy()
    }
    update$obs[[row]] <- NULL
  }

  update
}

create_stack_obs_observer <- function(input, rv, upd, sess, proxy) {

  observeEvent(
    names(upd$curr),
    {
      ids <- names(upd$curr)

      if (setequal(ids, names(upd$obs))) {
        return()
      }

      DT::replaceData(
        proxy,
        dt_board_stack(upd$curr, sess$ns, rv$board),
        rownames = FALSE
      )

      create_dt_stack_obs(setdiff(ids, names(upd$obs)), upd, input,
                          board_blocks(rv$board), sess)
      destroy_dt_stack_obs(setdiff(names(upd$obs), ids), upd)
    }
  )
}

edit_stack_observer <- function(upd, rv) {

  observeEvent(
    upd$edit,
    {
      row <- upd$edit$row
      col <- upd$edit$col
      val <- upd$edit$val

      if (col == "name") {
        stack_name(upd$curr[[row]]) <- val
      } else if (col == "blocks") {
        stack_blocks(upd$curr[[row]]) <- val
      }

      if (row %in% board_stack_ids(rv$board)) {

        if (row %in% names(upd$mod)) {
          upd$mod[row] <- upd$curr[row]
        } else {
          upd$mod <- c(upd$mod, upd$curr[row])
        }

      } else {

        if (row %in% names(upd$add)) {
          upd$add[row] <- upd$curr[row]
        } else {
          upd$add <- c(upd$add, upd$curr[row])
        }
      }
    }
  )
}

merge_staged_stacks <- function(applied, add, rm, mod) {

  keep <- setdiff(names(applied), rm)
  out <- applied[keep]

  modded <- intersect(keep, names(mod))

  if (length(modded)) {
    out[modded] <- mod[modded]
  }

  c(out, add[setdiff(names(add), names(applied))])
}

add_stack_observer <- function(input, rv, upd, sess) {

  observeEvent(
    input$add_stack,
    {
      avail <- available_stack_blocks(
        stacks = upd$curr,
        blocks = board_block_ids(rv$board)
      )

      if (length(avail)) {

        if (length(input$new_stack_id) && nchar(input$new_stack_id)) {

          updateTextInput(
            session = sess,
            inputId = "new_stack_id",
            label = NULL,
            value = "",
            placeholder = "Next ID"
          )

          if (input$new_stack_id %in% names(upd$curr)) {
            notify(
              "Please choose a unique stack ID.",
              type = "warning",
              session = sess
            )
            return()
          }

          new_id <- input$new_stack_id

        } else {

          new_id <- rand_names(names(upd$curr))
        }

        new <- set_names(
          list(new_stack(name = id_to_sentence_case(new_id))),
          new_id
        )

        upd$curr <- c(upd$curr, new)
        upd$add <- c(upd$add, upd$curr[length(upd$curr)])

      } else {

        notify(
          "No new stacks can be added. Remove a row first.",
          type = "warning",
          session = sess
        )
      }
    }
  )
}

rm_stack_observer <- function(input, rv, upd, sess) {

  observeEvent(
    input$rm_stack,
    {
      sel <- input$stacks_dt_rows_selected

      if (length(sel)) {

        ids <- names(upd$curr[sel])

        upd$rm <- c(upd$rm, ids[ids %in% board_stack_ids(rv$board)])

        upd$add <- upd$add[setdiff(names(upd$add), ids)]
        upd$mod <- upd$mod[setdiff(names(upd$mod), ids)]
        upd$curr <- upd$curr[setdiff(names(upd$curr), ids)]

      } else {

        notify("No row selected", type = "warning", session = sess)
      }
    }
  )
}

cancel_stack_observer <- function(input, rv, upd, session) {

  observeEvent(
    input$cancel_stacks,
    {
      removeModal(session)

      upd$add <- stacks()
      upd$rm <- character()
      upd$mod <- stacks()
      upd$curr <- board_stacks(rv$board)
    }
  )
}

modify_stack_observer <- function(input, rv, upd, sess, proxy, res) {

  observeEvent(
    input$modify_stacks,
    {
      if (!length(upd$add) && !length(upd$rm) && !length(upd$mod)) {

        notify(
          "No changes specified.",
          type = "warning",
          session = sess
        )

        return()
      }

      new <- try_catch_continue(
        modify_board_stacks(rv$board, upd$add, upd$rm, upd$mod),
        message = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "message",
                 glue = FALSE, session = sess)
        },
        warning = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "warning",
                 glue = FALSE, session = sess)
        },
        error = function(e) {
          notify(conditionMessage(e), duration = NULL, type = "error",
                 glue = FALSE, session = sess)
          NULL
        }
      )

      if (is.null(new)) {
        return()
      }

      mod_payload <- if (length(upd$mod)) {
        lapply(
          upd$mod,
          function(s) {
            list(name = stack_name(s), blocks = stack_blocks(s))
          }
        )
      }

      res(
        list(
          stacks = list(
            add = if (length(upd$add)) upd$add,
            rm = if (length(upd$rm)) upd$rm,
            mod = mod_payload
          )
        )
      )

      upd$add <- stacks()
      upd$rm <- character()
      upd$mod <- stacks()
      upd$curr <- board_stacks(new)

      DT::replaceData(
        proxy,
        dt_board_stack(upd$curr, sess$ns, rv$board),
        rownames = FALSE
      )

      removeModal(sess)
    }
  )
}

#' @section CSV block:
#' Files in CSV format provided for example by a block created via
#' [new_file_block()] may be parsed into `data.frame` by CSV blocks.
#'
#' @param sep,quote Forwarded to [utils::read.table()]
#'
#' @block csv parser block
#' @blockDescr Read CSV file
#' @blockCategory utility
#' @blockIcon filetype-csv
#' @blockGuidance Parses a delimited text file (typically from a file or upload
#'   block) into a data.frame with `utils::read.table()`. `sep` is the field
#'   separator and `quote` the quoting character.
#' @blockKeywords csv, parse, read, delimited, import
#' @blockArg sep Field separator character.
#'   [example] ","
#'   [type] arg_string()
#' @blockArg quote Quoting character.
#'   [type] arg_string()
#'
#' @rdname new_parser_block
#' @export
new_csv_block <- function(sep = ",", quote = "\"", ...) {
  new_parser_block(
    function(id, file) {
      moduleServer(
        id,
        function(input, output, session) {

          sp <- reactiveVal(sep)
          qo <- reactiveVal(quote)

          observeEvent(input$sep, sp(input$sep))
          observeEvent(input$quote, qo(input$quote))

          list(
            expr = reactive(
              bbquote(
                utils::read.table(.(file), header = TRUE, sep = .(sp),
                                  quote = .(qo), dec = ".", fill = TRUE,
                                  comment.char = ""),
                list(sp = sp(), qo = qo())
              )
            ),
            state = list(sep = sp, quote = qo)
          )
        }
      )
    },
    function(id) {
      tagList(
        textInput(
          inputId = NS(id, "sep"),
          label = "Field separator",
          value = sep
        ),
        textInput(
          inputId = NS(id, "quote"),
          label = "Quoting characters",
          value = quote
        )
      )
    },
    expr_type = "bquoted",
    class = "csv_block",
    ...
  )
}

#' Plot block constructors
#'
#' Blocks for data visualization using base R graphics can be created via
#' `new_plot_block()`.
#'
#' Due to the current block evaluation procedure, where block evaluation is
#' separated from block "rendering" (via  [shiny::renderPlot()]) integration of
#' base R graphics requires some mechanism to achieve this decoupling. This
#' is implemented by adding a `plot` attribute to the result of
#' [block_eval()], generated with [grDevices::recordPlot()] and containing the
#' required information to re-create the plot at a later time. As part of
#' [block_output()], the attribute is retrieved and passed to
#' [grDevices::replayPlot()]. Consequently, any block that inherits from
#' `plot_block` is required to support this type of decoupling.
#'
#' @param ... Forwarded to `new_plot_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_plot_block()` inherit from
#' `plot_block`.
#'
#' @seealso Real-world plot blocks built on `new_plot_block()` are provided by
#' the [blockr.ggplot](https://bristolmyerssquibb.github.io/blockr.ggplot/)
#' package.
#'
#' @export
new_plot_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "plot_block"), ctor, ...)
}

#' @export
block_output.plot_block <- function(x, result, session) {
  renderPlot(evaluate::replay(result))
}

#' @export
block_ui.plot_block <- function(id, x, ...) {
  tagList(
    plotOutput(NS(id, "result"))
  )
}

#' @export
block_eval.plot_block <- function(x, expr, env, ...) {

  inherits_rec_plt <- function(x) {
    inherits(x, "recordedplot")
  }

  res <- evaluate::evaluate(expr, env, stop_on_error = 1L)

  show_condition(res)

  Filter(inherits_rec_plt, res)
}

#' @noRd
#' @export
block_eval_trigger.plot_block <- function(x, session = get_session()) {

  invisible(
    get_board_option_values(
      c("thematic", "dark_mode"),
      if_not_found = "null",
      session = session
    )
  )
}

#' Board links
#'
#' Two blocks can be connected via a (directed) link. This means the result from
#' one block is passed as (data) input to the next. Source and destination are
#' identified by `from` and `to` attributes and in case of polyadic receiving
#' blocks, the `input` attribute identified which of the data inputs is the
#' intended destination. In principle, the `link` object may be extended via
#' sub-classing and passing further attributes, but this has not been properly
#' tested so far.
#'
#' A links is created via the `new_link()` constructor and for a vector of
#' links, the container object `links` is provided and a corresponding
#' constructor `links()` exported from the package. Testing whether an object
#' inherits from `link` (or `links`) is available via `is_link()` (or
#' `is_links()`, respectively). Coercion to `link` (and `links`) objects is
#' implemented as `as_link()` (and `as_links()`, respectively). Finally, links
#' can be validated by calling `validate_links()`.
#'
#' @param from,to Block ID(s)
#' @param input Block argument
#' @param ... Extensibility
#' @param ctor,pkg Constructor information (used for serialization)
#' @param class (Optional) link sub-class
#'
#' @examples
#' lnks <- links(from = c("a", "b"), to = c("b", "c"), input = c("x", "y"))
#' is_links(lnks)
#' names(lnks)
#'
#' tryCatch(
#'   c(lnks, new_link("a", "b", "x")),
#'   error = function(e) conditionMessage(e)
#' )
#' tryCatch(
#'   c(lnks, new_link("b", "a")),
#'   error = function(e) conditionMessage(e)
#' )
#'
#' lnks <- links(a = new_link("a", "b"), b = new_link("b", "c"))
#' names(lnks)
#'
#' tryCatch(
#'   c(lnks, a = new_link("a", "b")),
#'   error = function(e) conditionMessage(e)
#' )
#'
#' @return Both `new_link()`/`as_link()`, and `links()`/`as_links()` return
#' `link` and `links` objects, respectively. Testing for inheritance is
#' available as `is_link()`/`is_links()` and validation (for `links`) is
#' performed with `validate_links()`, which returns its input (`x`) or throws
#' an error.
#'
#' @export
new_link <- function(from = "", to = "", input = "", ...,
                     ctor = "new_link", pkg = pkg_name(), class = character()) {

  stopifnot(length(from) == length(to))

  if (length(from) && is_empty(input)) {
    input <- ""
  }

  validate_link(
    new_vctr(
      list(from = from, to = to, input = input, ...),
      ctor = new_blockr_ctor(NULL, ctor, pkg),
      class = c(class, "link")
    )
  )
}

#' @section Partial-arg updates:
#' `update_link()` is an S3 generic that produces a modified link by
#' merging a named-list `delta` of constructor argument values onto an
#' existing link. The default `.link` method passes the current
#' `as.list(x)` (the link's `from`, `to`, `input` plus any extra
#' *list-element* fields a sub-class adds) merged with the delta to the
#' stored constructor. Sub-class extensions stored as attributes (rather
#' than list elements) are **not** preserved through the default
#' reconstruction — sub-class owners must register a class-specific
#' method when their constructor uses attributes or a non-standard
#' signature.
#'
#' @param delta A named list of constructor argument values to apply on
#' top of `x`'s current fields.
#'
#' @rdname new_link
#' @export
update_link <- function(x, delta) {
  UseMethod("update_link")
}

#' @export
update_link.link <- function(x, delta) {

  ctor <- attr(x, "ctor")

  args <- as.list(x)
  args[names(delta)] <- delta

  do.call(
    ctor_fun(ctor),
    c(
      args,
      list(ctor = coal(ctor_name(ctor), ctor_fun(ctor)), pkg = ctor_pkg(ctor))
    )
  )
}

validate_link <- function(x) {

  if (!is_link(x)) {
    blockr_abort(
      "Expecting a link to inherit from \"link\".",
      class = "link_class_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a link to be represented by a list.",
      class = "link_list_like_invalid"
    )
  }

  if (all_zero_len(x)) {
    return(x)
  }

  fields <- c("from", "to", "input")

  if (!all(fields %in% names(x))) {
    blockr_abort(
      "Expecting a link to contain attributes {fields}.",
      class = "link_components_missing"
    )
  }

  if (!all(lgl_ply(x[fields], is_string))) {
    blockr_abort(
      "Expecting link attributes {fields} to be strings.",
      class = "link_components_invalid"
    )
  }

  if (anyNA(x[fields])) {
    blockr_abort(
      "Missing values for {fields} are not allowed.",
      class = "link_components_invalid"
    )
  }

  if (x[["from"]] == x[["to"]] && x[["to"]] != "") {
    blockr_abort(
      "Self-referencing links are not allowed.",
      class = "link_self_referencing"
    )
  }

  x
}

#' @rdname new_link
#' @export
is_link <- function(x) {
  inherits(x, "link")
}

#' @param x Links object
#' @rdname new_link
#' @export
as_link <- function(x) {
  UseMethod("as_link")
}

#' @export
as_link.link <- function(x) {
  x
}

#' @export
as_link.character <- function(x) {
  do.call(new_link, as.list(x))
}

#' @export
as_link.list <- function(x) {
  do.call(new_link, x)
}

#' @method as_link data.frame
#' @export
as_link.data.frame <- function(x) {
  do.call(new_link, x)
}

#' @export
as.character.link <- function(x, ...) {
  do.call(c, as.list(x))
}

#' @export
as.list.link <- function(x, ...) {
  vec_data(x)
}

#' @method as.data.frame link
#' @export
as.data.frame.link <- function(x, ...) {
  as.data.frame(as.list(x))
}

#' @export
`[.link` <- function(x, i, ...) {
  x <- vec_data(x)
  x[vec_as_location(i, length(x), names(x))]
}

#' @export
`[[.link` <- function(x, i, ...) {
  x <- vec_data(x)
  x[[vec_as_location2(i, length(x), names(x))]]
}

#' @export
format.link <- function(x, ...) {

  field_miss <- function(x) {
    if (is.na(x) || !nzchar(x)) "?" else x
  }

  out <- ""

  for (cl in rev(setdiff(class(x), c("list", "vctrs_vctr")))) {
    out <- paste0("<", cl, out, ">")
  }

  if (all_zero_len(x)) {
    return(out)
  }

  inp <- x["input"]

  if (is.na(inp) || !nzchar(inp)) {
    inp <- ""
  } else {
    inp <- paste0(" (", inp, ")")
  }

  c(out, paste0(field_miss(x["from"]), " -> ", field_miss(x["to"]), inp))
}

#' @export
print.link <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.link <- function(x, ...) {
  paste(format(x, ...), collapse = " ")
}

#' @export
str.link <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
c.link <- function(...) {
  as_links(list_to_list_of_links(list(...)))
}

test_that("link class", {

  x <- new_link("a", "b")

  expect_s3_class(x, "link")
  expect_snapshot(print(x))

  expect_error(
    validate_link("abc"),
    class = "link_class_invalid"
  )

  expect_error(
    validate_link(new_vctr("abc", class = "link")),
    class = "link_list_like_invalid"
  )

  expect_error(
    validate_link(new_vctr(list("abc"), class = "link")),
    class = "link_components_missing"
  )

  expect_error(
    validate_link(
      new_vctr(list(from = "abc", to = 1, input = ""), class = "link")
    ),
    class = "link_components_invalid"
  )

  expect_error(
    validate_link(
      new_vctr(
        list(from = "abc", to = "abc", input = ""),
        class = "link"
      )
    ),
    class = "link_self_referencing"
  )

  lnk <- new_link("a", "b")

  dtf <- as.data.frame(lnk)
  lst <- as.list(lnk)
  chr <- as.character(lnk)

  expect_s3_class(c(lnk, dtf), "links")
  expect_s3_class(c(lnk, lst), "links")
  expect_s3_class(c(lnk, chr), "links")
})

test_that("a link has a compact str_value()", {

  expect_identical(
    str_value(new_link(from = "a", to = "b", input = "x")),
    "<link> a -> b (x)"
  )
  expect_identical(str_value(new_link(from = "a", to = "b")), "<link> a -> b")
  expect_identical(str_value(new_link()), "<link> ? -> ?")

  expect_identical(
    capture.output(str(new_link(from = "a", to = "b"))),
    " <link> a -> b"
  )
})

#' @section Subset block:
#' This block allows to perform row and column subsetting on `data.frame`
#' objects via [base::subset()]. Using non-standard evaluation, strings passed
#' as `subset`/`select` arguments or entered via shiny UI are turned into
#' `language` objects by [base::parse()].
#'
#' @param subset,select Expressions (passed as strings)
#'
#' @block subset block
#' @blockDescr Row and column subsetting
#' @blockCategory transform
#' @blockIcon funnel
#' @blockGuidance Row- and column-subsets the input with `base::subset()`.
#'   `subset` is a logical expression over the columns keeping rows where it is
#'   TRUE; `select` picks columns. Both are R expressions written as strings and
#'   evaluated with the data columns in scope.
#' @blockKeywords filter, subset, select, rows, columns, where
#' @blockArg subset Logical row-filter expression, as a string.
#'   [example] "Sepal.Width > 3"
#'   [type] arg_string()
#' @blockArg select Column-selection expression, as a string.
#'   [example] "Species"
#'   [type] arg_string()
#'
#' @rdname new_transform_block
#' @export
new_subset_block <- function(subset = "", select = "", ...) {

  pasrse_first <- function(x) {
    parse(text = x)[[1L]]
  }

  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {

          sub <- reactiveVal(subset)
          sel <- reactiveVal(select)

          observeEvent(
            input$eval,
            {
              sub(input$subset)
              sel(input$select)
            }
          )

          observeEvent(
            req(sub()),
            {
              if (!identical(sub(), input$subset)) {
                updateTextInput(session, "subset", value = sub())
              }
            }
          )

          observeEvent(
            req(sel()),
            {
              if (!identical(sel(), input$select)) {
                updateTextInput(session, "select", value = sel())
              }
            }
          )

          list(
            expr = reactive(
              {
                su <- sub()
                se <- sel()

                if (nzchar(su) && nzchar(se)) {
                  bbquote(
                    subset(.(data), .(su), .(se)),
                    list(su = pasrse_first(su), se = pasrse_first(se))
                  )
                } else if (nzchar(se)) {
                  bbquote(
                    subset(.(data), select = .(se)),
                    list(se = pasrse_first(se))
                  )
                } else if (nzchar(su)) {
                  bbquote(
                    subset(.(data), .(su)),
                    list(su = pasrse_first(su))
                  )
                } else {
                  quote(subset(.(data)))
                }
              }
            ),
            state = list(subset = sub, select = sel)
          )
        }
      )
    },
    function(id) {
      tagList(
        textInput(
          inputId = NS(id, "subset"),
          label = "Subset",
          value = subset,
          placeholder = "Enter a row subsetting expression."
        ),
        textInput(
          inputId = NS(id, "select"),
          label = "Select",
          value = select,
          placeholder = "Enter a column selection expression."
        ),
        actionButton(
          inputId = NS(id, "eval"),
          label = "Evaluate",
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data))
    },
    allow_empty_state = TRUE,
    expr_type = "bquoted",
    external_ctrl = TRUE,
    class = "subset_block",
    ...
  )
}

#' Board plugin
#'
#' A core mechanism for extending or customizing UX aspects of the board module
#' is a "plugin" architecture. All plugins inherit from `plugin` and a sub-class
#' is assigned to each specific plugin. The "manage blocks" plugin for example
#' has a class vector `c("manage_blocks", "plugin")`. Sets of plugins are
#' handled via a wrapper class `plugins`. Each plugin needs a server component,
#' in most cases accompanied by a UI component and is optionally bundled with a
#' validator function.
#'
#' @param server,ui Server/UI for the plugin module
#' @param validator Validator function that validates server return values
#' @param class Plugin subclass
#'
#' @examples
#' plg <- board_plugins(new_board())
#'
#' is_plugins(plg)
#' names(plg)
#'
#' plg[1:3]
#'
#' is_plugin(plg[["preserve_board"]])
#'
#' @return Constructors `new_plugin()`/`plugins()` return `plugin` and
#' `plugins` objects, respectively, as do `as_plugin()`/`as_plugins()` and
#' validators `validate_plugin()`/`validate_plugins()`, which are typically
#' called for their side effects of throwing errors in case of validation
#' failure. Inheritance checkers `is_plugin()`/`is_plugins()` return scalar
#' logicals and finally, the convenience function `board_plugins()` returns a
#' `plugins` object with all known plugins (or a selected subset thereof).
#'
#' @export
new_plugin <- function(server, ui = NULL, validator = abort_not_null,
                       class = character()) {

  res <- structure(
    list(server = server, ui = ui),
    validator = validator,
    class = c(class, "plugin")
  )

  validate_plugin(res)
}

#' @param x Plugin object
#' @rdname new_plugin
#' @export
is_plugin <- function(x) {
  inherits(x, "plugin") &&
    sum(inherits(x, known_plugins(), which = TRUE) != 0L) == 1L
}

#' @rdname new_plugin
#' @export
abort_not_null <- function(x) {

  if (is.null(x)) {
    return(invisible(x))
  }

  blockr_abort(
    "Expected `NULL`, but got {class(x)} instead.",
    class = "expect_not_null"
  )
}

known_plugins <- function() {
  c("preserve_board", "manage_blocks", "manage_links", "manage_stacks",
    "notify_user", "generate_code", "edit_block", "edit_stack", "ctrl_block")
}

#' @rdname new_plugin
#' @export
as_plugin <- function(x) {
  UseMethod("as_plugin")
}

#' @export
as_plugin.plugin <- function(x) x

#' @export
as_plugin.list <- function(x) do.call(new_plugin, x)

#' @export
as_plugin.plugins <- function(x) {

  if (length(x) != 1L) {
    blockr_abort(
      "Cannot cast length != 1 \"plugins\" to \"plugin\".",
      class = "plugins_as_plugin_invalid"
    )
  }

  .subset2(x, 1L)
}

#' @export
as.list.plugin <- function(x, ...) {
  list(
    server = plugin_server(x),
    ui = plugin_ui(x),
    validator = plugin_validator(x),
    class = setdiff(class(x), "plugin")
  )
}

#' @rdname new_plugin
#' @export
plugin_server <- function(x) x[["server"]]

#' @rdname new_plugin
#' @export
plugin_ui <- function(x) x[["ui"]]

#' @rdname new_plugin
#' @export
plugin_validator <- function(x) attr(x, "validator")

#' @rdname new_plugin
#' @export
plugin_id <- function(x) class(x)[1L]

validate_plugin <- function(x) {

  if (!is_plugin(x)) {
    blockr_abort(
      "Expecting a plugin to inherit from `plugin` and any one of ",
      "{known_plugins()}.",
      class = "plugin_inheritance_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a plugin to behave list-like.",
      class = "plugin_type_invalid"
    )
  }

  if (!setequal(names(x), c("server", "ui"))) {
    blockr_abort(
      "Expecting a plugin to contain components \"server\" and \"ui\".",
      class = "plugin_components_invalid"
    )
  }

  srv <- plugin_server(x)

  if (!is.null(srv) && !is.function(srv)) {
    blockr_abort(
      "Expecting a plugin server to be a function.",
      class = "plugin_server_invalid"
    )
  }

  ui <- plugin_ui(x)

  if (!is.null(ui) && !is.function(ui)) {
    blockr_abort(
      "Expecting a plugin ui to be `NULL` or a function.",
      class = "plugin_ui_invalid"
    )
  }

  if (!is.function(plugin_validator(x))) {
    blockr_abort(
      "Expecting a plugin validator to be a function.",
      class = "plugin_validator_invalid"
    )
  }

  x
}

#' @export
format.plugin <- function(x, ...) {

  out <- ""

  for (cl in rev(class(x))) {
    out <- paste0("<", cl, out, ">")
  }

  out
}

#' @export
print.plugin <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.plugin <- function(x, ...) {
  format(x, ...)
}

#' @export
str.plugin <- function(object, ...) {
  cat_str_value(object, ...)
}

harmonize_list_of_plugins <- function(x) {
  if (is_plugin(x)) {
    list(x)
  } else if (is_plugins(x)) {
    as.list(x)
  } else if (is.list(x) && all(lgl_ply(x, is_plugin))) {
    x
  } else {
    list(as_plugin(x))
  }
}

#' @export
c.plugin <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_plugins),
    recursive = FALSE
  )

  as_plugins(res)
}

get_plugin <- function(plugin, plugins = NULL, component = NULL) {

  if (is.null(plugins)) {

    if (is.null(plugin)) {
      return(NULL)
    }

    res <- validate_plugin(plugin)

  } else {

    stopifnot(is_string(plugin), is_plugins(plugins))

    if (!plugin %in% chr_ply(plugins, plugin_id)) {
      return(NULL)
    }

    res <- plugins[[plugin]]
  }

  if (is.null(component)) {
    return(res)
  }

  switch(
    match.arg(component, c("server", "ui")),
    ui = plugin_ui(res),
    server = function(server_args, validator_args = NULL) {
      do.call(
        plugin_validator(res),
        c(
          list(do.call(plugin_server(res), server_args)),
          validator_args
        )
      )
    }
  )
}

get_plugin_ui <- function(plugin, plugins = NULL) {
  get_plugin(plugin, plugins, component = "ui")
}

get_plugin_server <- function(plugin, plugins = NULL) {
  get_plugin(plugin, plugins, component = "server")
}

call_plugin_server <- function(plugin, server_args, plugins = NULL,
                               validator_args = NULL) {

  server <- get_plugin_server(plugin, plugins)

  if (is.null(server)) {
    return(NULL)
  }

  if (is_plugin(plugin)) {
    id <- plugin_id(plugin)
  } else {
    id <- plugin
  }

  server(c(list(id), server_args), validator_args)
}

call_plugin_ui <- function(plugin, ns, ..., plugins = NULL) {

  ui <- get_plugin_ui(plugin, plugins)

  if (is.null(ui)) {
    return(tagList())
  }

  if (is_plugin(plugin)) {
    id <- plugin_id(plugin)
  } else {
    id <- plugin
  }

  ui(ns(id), ...)
}

#' @param ... Generic consistency
#' @rdname new_plugin
#' @export
board_plugins <- function(x, ...) {
  UseMethod("board_plugins")
}

#' @export
board_plugins.board <- function(x, which = NULL, ...) {

  plugins <- plugins(
    preserve_board(),
    manage_blocks(),
    manage_links(),
    manage_stacks(),
    notify_user(),
    generate_code(),
    edit_block(),
    edit_stack()
  )

  if (is.null(which)) {
    return(plugins)
  }

  plugins[which]
}

validate_callbacks <- function(x) {

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a list of callbacks.",
      class = "invalid_callbacks"
    )
  }

  for (f in x) {
    if (!is.function(f)) {
      blockr_abort(
        "Expecting callbacks to be passed as functions.",
        class = "invalid_callbacks"
      )
    }
  }

  if (!is.null(names(x)) && length(unique(names(x))) != length(x)) {
    blockr_abort(
      "Callbacks must either be unnamed or named with unique names.",
      class = "invalid_callbacks"
    )
  }

  invisible(x)
}

#' @param ... Plugin objects
#' @rdname new_plugin
#' @export
plugins <- function(...) {
  validate_plugins(
    new_vctr(unname(list(...)), class = c("plugins", "list"))
  )
}

#' @rdname new_plugin
#' @export
is_plugins <- function(x) {
  inherits(x, "plugins")
}

#' @rdname new_plugin
#' @export
as_plugins <- function(x) {
  UseMethod("as_plugins")
}

#' @export
as_plugins.plugins <- function(x) {
  validate_plugins(x)
}

#' @export
as_plugins.list <- function(x) {
  do.call(plugins, x)
}

#' @export
as_plugins.plugin <- function(x) {
  as_plugins(list(x))
}

#' @export
as.list.plugins <- function(x, ...) {
  vec_data(x)
}

#' @export
anyDuplicated.plugins <- function(x, incomparables = FALSE, ...) {
  anyDuplicated(chr_ply(x, plugin_id), incomparables = incomparables, ...)
}

#' @rdname new_plugin
#' @export
validate_plugins <- function(x) {

  if (!is_plugins(x)) {
    blockr_abort(
      "Expecting a board plugins object to inherit from \"plugins\".",
      class = "plugins_inheritance_invalid"
    )
  }

  if (!is.list(x)) {
    blockr_abort(
      "Expecting a board plugins object behave list-like.",
      class = "plugins_type_invalid"
    )
  }

  if (!all(lgl_ply(x, is_plugin))) {
    blockr_abort(
      "Expecting a board plugins to contain a set of plugins.",
      class = "plugins_contains_invalid"
    )
  }

  if (anyDuplicated(x) != 0L) {
    blockr_abort(
      "Duplicate board plugins are not allowed.",
      class = "plugins_duplicate_invalid"
    )
  }

  x
}

#' @export
format.plugins <- function(x, ...) {
  c(
    paste0("<", class(x)[1L], "[", length(x), "]>"),
    chr_mply(paste0, names(x), ": ", chr_ply(x, format, ...))
  )
}

#' @export
print.plugins <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export
str_value.plugins <- function(x, ...) {
  str_value_collection(x, chr_ply(x, str_value))
}

#' @export
str.plugins <- function(object, ...) {
  cat_str_value(object, ...)
}

#' @export
names.plugins <- function(x) {
  chr_ply(x, plugin_id)
}

#' @export
`names<-.plugins` <- function(x, value) {

  reset <- is.null(value) ||
    identical(value, chr_ply(x, plugin_id)) ||
    all(is.na(value) | value == "")

  if (isTRUE(reset)) {
    attr(x, "names") <- NULL
    return(x)
  }

  blockr_abort(
    "Cannot modify plugin names.",
    class = "plugin_names_assign_invalid"
  )
}

#' @export
board_ui.plugin <- function(id, x, ...) {
  call_plugin_ui(x, NS(id), ...)
}

#' @export
board_ui.plugins <- function(id, x, ...) {
  do.call(
    tagList,
    lapply(x, function(y, id, ...) board_ui(id, y, ...), id, ...)
  )
}

#' @export
vec_restore.plugins <- function(x, to, ...) {
  validate_plugins(NextMethod())
}

#' @export
c.plugins <- function(...) {

  res <- unlist(
    lapply(list(...), harmonize_list_of_plugins),
    recursive = FALSE
  )

  as_plugins(res)
}

#' @export
`[.plugins` <- function(x, i, ...) {

  loc <- vec_as_location(
    i,
    length(x),
    chr_ply(x, plugin_id),
    missing = "remove"
  )

  vec_slice(x, loc)
}

#' @export
`[<-.plugins` <- function(x, i, ..., value) {
  blockr_abort(
    "Cannot assign into `plugins`.",
    class = "plugins_assignment_invalid"
  )
}

#' @export
`[[.plugins` <- function(x, i, ...) {

  loc <- vec_as_location2(
    i,
    length(x),
    chr_ply(x, plugin_id),
    missing = "propagate"
  )

  as_plugin(vec_slice(x, loc))
}

#' @export
`[[<-.plugins` <- function(x, i, ..., value) {
  blockr_abort(
    "Cannot assign into `plugins`.",
    class = "plugins_assignment_invalid"
  )
}

test_that("plugins", {

  a <- preserve_board()
  b <- manage_blocks()

  expect_s3_class(a, "plugin")
  expect_s3_class(a, "preserve_board")
  expect_true(is_plugin(a))

  expect_snapshot(print(a))

  expect_error(
    validate_plugin("a"),
    class = "plugin_inheritance_invalid"
  )

  expect_error(
    validate_plugin(
      structure("a", class = c("preserve_board", "plugin"))
    ),
    class = "plugin_type_invalid"
  )

  expect_error(
    validate_plugin(
      structure(list("a"), class = c("preserve_board", "plugin"))
    ),
    class = "plugin_components_invalid"
  )

  expect_error(
    validate_plugin(
      structure(
        list(server = "a", ui = NULL),
        class = c("preserve_board", "plugin")
      )
    ),
    class = "plugin_server_invalid"
  )

  expect_error(
    validate_plugin(
      structure(
        list(server = identity, ui = "a"),
        class = c("preserve_board", "plugin")
      )
    ),
    class = "plugin_ui_invalid"
  )

  expect_error(
    validate_plugin(
      structure(
        list(server = identity, ui = identity),
        validator = "a",
        class = c("preserve_board", "plugin")
      )
    ),
    class = "plugin_validator_invalid"
  )

  expect_identical(as_plugin(as.list(a)), a)
  expect_identical(as_plugin(as_plugins(a)), a)

  c <- manage_links()
  d <- manage_stacks()

  cd <- plugins(c, d)

  expect_s3_class(cd, "plugins")
  expect_true(is_plugins(cd))
  expect_length(cd, 2L)

  expect_snapshot(print(cd))

  expect_error(
    validate_plugins("a"),
    class = "plugins_inheritance_invalid"
  )

  expect_error(
    validate_plugins(
      structure(
        "a",
        class = c("plugins", "list", "vctrs_vctr", "list")
      )
    ),
    class = "plugins_type_invalid"
  )

  expect_error(
    validate_plugins(
      structure(
        list(a, a),
        class = c("plugins", "list", "vctrs_vctr", "list")
      )
    ),
    class = "plugins_duplicate_invalid"
  )

  expect_identical(as_plugins(cd), cd)
  expect_identical(as_plugins(as.list(cd)), cd)

  expect_identical(c(a), as_plugins(a))
  expect_identical(c(cd), cd)

  expect_identical(
    names(cd),
    c("manage_links", "manage_stacks")
  )

  names(cd) <- NULL

  expect_identical(
    names(cd),
    c("manage_links", "manage_stacks")
  )

  expect_error(
    names(cd) <- c("c", "d"),
    class = "plugin_names_assign_invalid"
  )

  res <- c(a, cd)

  expect_s3_class(res, "plugins")
  expect_length(res, 3L)

  res <- c(a, b)

  expect_s3_class(res, "plugins")
  expect_length(res, 2L)

  res <- c(a, b, c)

  expect_s3_class(res, "plugins")
  expect_length(res, 3L)

  res <- c(a, list(c, d))

  expect_s3_class(res, "plugins")
  expect_length(res, 3L)

  res <- c(cd, a)

  expect_s3_class(res, "plugins")
  expect_length(res, 3L)

  res <- c(cd, list(a))

  expect_s3_class(res, "plugins")
  expect_length(res, 3L)

  expect_equal(cd[1L], as_plugins(c), ignore_function_env = TRUE)
  expect_equal(cd[c(1L, 2L)], cd, ignore_function_env = TRUE)

  expect_equal(cd["manage_links"], as_plugins(c), ignore_function_env = TRUE)
  expect_equal(cd[c("manage_links", "manage_stacks")], cd,
               ignore_function_env = TRUE)

  expect_error(
    cd[1] <- c,
    class = "plugins_assignment_invalid"
  )

  expect_equal(cd[[1L]], c, ignore_function_env = TRUE)
  expect_equal(cd[["manage_links"]], c, ignore_function_env = TRUE)

  expect_error(
    cd[[1]] <- c,
    class = "plugins_assignment_invalid"
  )
})

test_that("check default validator", {

  expect_null(abort_not_null(NULL))
  expect_invisible(abort_not_null(NULL))

  expect_error(
    abort_not_null(1),
    class = "expect_not_null"
  )
})

test_that("plugins have a compact str_value()", {

  expect_identical(str_value(preserve_board()), "<preserve_board<plugin>>")
  expect_identical(str_value(preserve_board()), format(preserve_board()))

  pls <- plugins(preserve_board(), manage_blocks())

  expect_identical(
    str_value(pls),
    paste(
      "<plugins[2]>",
      "  preserve_board: <preserve_board<plugin>>",
      "  manage_blocks: <manage_blocks<plugin>>",
      sep = "\n"
    )
  )

  expect_identical(str_value(plugins()), "<plugins[0]>")

  expect_identical(capture.output(str(pls))[1L], " <plugins[2]>")
})

block_registry <- new.env()

#' Block registry
#'
#' Listing of blocks is available via a block registry, which associates a block
#' constructor with metadata in order to provide a browsable block directory.
#' Every constructor is identified by a unique ID (uid), which by default is
#' generated from the class vector (first element). If the class vector is not
#' provided during registration, an object is instantiated (by calling the
#' constructor with arguments `ctor` and `ctor_pkg` only) to derive this
#' information. Block constructors therefore should be callable without block-
#' specific arguments.
#'
#' Due to current requirements for serialization/deserialization, we keep track
#' the constructor that was used for block instantiation. This works most
#' reliable whenever a block constructor is an exported function from a package
#' as this function is guaranteed to be available in a new session (give the
#' package is installed in an appropriate version). While it is possible to
#' register a block passing a "local" function as `ctor`, this may introduce
#' failure modes that are less obvious (for example when such a constructor
#' calls another function that is only defined within the scope of the session).
#' It is therefore encouraged to only rely on exported function constructors.
#' These can also be passed as strings and together with the value of `package`,
#' the corresponding function can easily be retrieved in any session.
#'
#' Blocks can be registered (i.e. added to the registry) via `register_block()`
#' with scalar-valued arguments and `register_blocks()`, where arguments may be
#' vector-valued, while de-registration (or removal) is handled via
#' `unregister_blocks()`. A listing of all available blocks can be created as
#' `list_blocks()`, which will return registry IDs and `available_blocks()`,
#' which provides a set of (named) `block_registry_entry` objects. Finally,
#' block construction via a registry ID is available as `create_block()`.
#'
#' @param ctor Block constructor
#' @param name,description Metadata describing the block
#' @param classes Block classes
#' @param uid Unique ID for a registry entry
#' @param category Useful to sort blocks by topics. If not specified,
#'   blocks are uncategorized.
#' @param icon Icon
#' @param arguments Block argument specification, either a [new_block_args()]
#'   object or a (possibly empty) named character vector of argument
#'   descriptions
#' @param details Optional longer human-facing description (e.g. for a help
#'   popover), complementing the short `description`
#' @param link Optional URL to a block's help or documentation page
#' @param guidance Optional model-facing construction guidance (do/don't rules,
#'   enums, pitfalls), distinct from the human-facing `details`. Retrieved with
#'   [block_metadata()]
#' @param examples Optional list of complete worked configurations (each a
#'   named list keyed by argument name); supersedes the per-argument example
#'   assembly. Retrieved with [block_metadata()]
#' @param keywords Optional character vector of free-text search terms for block
#'   discovery, retrieved with [block_metadata()]
#' @param package Package where constructor is defined (or `NULL`)
#' @param overwrite Overwrite existing entry
#'
#' @examples
#' blks <- list_blocks()
#' register_block("new_dataset_block", "Test", "Registry test",
#'                uid = "test_block", package = "blockr.core")
#'
#' new <- setdiff(list_blocks(), blks)
#' unregister_blocks(new)
#' setequal(list_blocks(), blks)
#'
#' @return `register_block()` and `register_blocks()` are invoked for their side
#' effects and return `block_registry_entry` object(s) invisibly, while
#' `unregister_blocks()` returns `NULL` (invisibly). `register_package_blocks()`
#' registers the blocks a package declares in its YAML registry (written by the
#' [block_registration_roclet()]), likewise returning their entries invisibly.
#' Listing via `list_blocks()`
#' returns a character vector and a list of `block_registry_entry` object(s) for
#' `available_blocks()`. `create_block()` returns a newly instantiated `block`
#' object. A block's registered metadata is read with [block_metadata()] and the
#' `block_meta_*()` accessors, and its argument specification is built and read
#' with [new_block_arg()] and friends.
#'
#' @export
register_block <- function(ctor, name, description, classes = NULL, uid = NULL,
                           category = NULL, icon = NULL, arguments = NULL,
                           details = NULL, link = NULL, guidance = NULL,
                           examples = list(), keywords = NULL,
                           package = NULL, overwrite = FALSE) {

  if (is.null(category)) {
    category <- default_category()
  }

  stopifnot(is_string(category))

  if (!category %in% names(suggested_categories())) {
    blockr_warn(
      "Block category {category} is not among suggested categories ",
      "{names(suggested_categories())}. Consider choosing a different one.",
      class = "block_category_discouraged",
      frequency = "once",
      frequency_id = paste0("block_category_", category, "_discouraged")
    )
  }

  if (is.null(icon)) {

    icon <- default_icon(category)

  } else {

    stopifnot(is_string(icon))

    if (grepl("-fill$", icon)) {
      blockr_warn(
        "Using block icons with 'fill' style, such as {icon} is discouraged.",
        class = "block_icon_fill_discouraged",
        frequency = "once",
        frequency_id = paste0("block_icon_", icon, "_discouraged")
      )
    }

    if (!icon %in% bsicon_icons()) {
      blockr_abort("Unknown icon {icon}.", class = "block_icon_invalid")
    }

    icon <- as.character(bsicons::bs_icon(icon))
  }

  ctor_name <- NULL

  if (is_string(ctor)) {
    stopifnot(is_string(package))
    ctor_name <- ctor
    ctor <- get(ctor, asNamespace(package), mode = "function")
  } else {
    stopifnot(is.function(ctor), is.null(package))
    package <- pkg_name(environment(ctor))
  }

  ctor_ref <- if (is.null(ctor_name)) ctor else ctor_name
  ctor_pkg <- if (is.null(ctor_name)) NULL else package

  user_spec <- is_block_args(arguments)
  validate <- user_spec || length(examples) > 0L

  if (is.null(classes) || is.null(arguments) || validate) {

    obj <- ctor(ctor = ctor_ref, ctor_pkg = ctor_pkg, block_metadata = FALSE)

    if (is.null(classes)) {
      classes <- class(obj)
    }

    if (is.null(arguments)) {
      arguments <- external_ctrl_vars(obj)
    }
  }

  norm <- normalize_arguments(arguments, guidance)
  arguments <- norm[["arguments"]]
  guidance <- norm[["guidance"]]

  if (validate) {
    validate_block_spec(arguments, examples, obj, ctor, ctor_ref, ctor_pkg,
                        check_names = user_spec)
  }

  if (is.null(uid)) {
    uid <- registry_uid(classes)
  }

  if (uid %in% list_blocks() && !isTRUE(overwrite)) {
    blockr_abort(
      "Block {uid} already exists. Set `overwrite = TRUE` to replace it, ",
      "or remove it first with `unregister_blocks(\"{uid}\")`.",
      class = "block_already_in_registry"
    )
  }

  entry <- new_registry_entry(
    ctor,
    name = name,
    description = description,
    classes = classes,
    category = category,
    icon = icon,
    arguments = arguments,
    details = details,
    link = link,
    guidance = guidance,
    examples = examples,
    keywords = keywords,
    ctor_name = ctor_name,
    package = package
  )

  assign(uid, entry, envir = block_registry)

  invisible(entry)
}

registry_uid <- function(x) {

  if (is_block(x)) {
    x <- class(x)
  }

  stopifnot(is.character(x), length(x) >= 1L)

  x[1L]
}

#' @rdname register_block
#' @export
default_icon <- function(category = default_category()) {

  stopifnot(is_string(category))

  res <- switch(
    category,
    input = "upload",
    transform = "magic",
    structured = "asterisk",
    plot = "file-bar-graph",
    table = "table",
    model = "gear",
    output = "save",
    utility = "tools",
    "question-square"
  )

  as.character(bsicons::bs_icon(res))
}

#' @rdname register_block
#' @export
default_category <- function() "uncategorized"

bsicon_icons <- function() {
  get("icon_info", envir = asNamespace("bsicons"), mode = "list")$name
}

#' @rdname register_block
#' @export
suggested_categories <- function() {
  c(
    input = paste(
      "Data loading and import blocks, e.g. dataset selection, file uploads,",
      "API connections, database queries."
    ),
    transform = paste(
      "Data manipulation and transformation (incl. joins, binds), e.g. filter,",
      "select, mutate, arrange, summarize, join, bind_rows."
    ),
    structured = paste(
      "Processing of specialized structured data formats such as time series",
      "or spatial data, hierarchical structures and graph/network data"
    ),
    plot = paste(
      "Visualization and plotting blocks, e.g. ggplotz visualizations, scatter",
      "plots, bar charts, histograms."
    ),
    table = paste(
      "Tabular output and display blocks, e.g. data tables, formatted tables,",
      "interactive tables, table viewers."
    ),
    model = paste(
      "Statistical modeling and Al/ML blocks, e.g. linear models, GLMs,",
      "neural networks, LLMs, predictions, classifiers."
    ),
    output = paste(
      "Data export and output blocks, e.g. save to file, export tables,",
      "generate reports."
    ),
    utility = paste(
      "Utility and tool blocks, e.g. parse, format, text manipulation,",
      "helpers."
    ),
    uncategorized = "Blocks without assigned category."
  )
}

new_registry_entry <- function(ctor, ...) {
  structure(ctor, ..., class = "block_registry_entry")
}

is_registry_entry <- function(x) inherits(x, "block_registry_entry")

#' @export
board_options.block_registry_entry <- function(x, ...) {
  board_options(structure(list(), class = attr(x, "classes")), ...)
}

#' @rdname register_block
#' @export
list_blocks <- function() {
  ls(envir = block_registry)
}

#' @param block Block object
#' @rdname register_block
#' @export
registry_id_from_block <- function(block) {

  uid <- registry_uid(block)
  ids <- list_blocks()

  hit <- sum(uid == ids)

  if (!hit) {
    return(character())
  }

  stopifnot(identical(hit, 1L))

  uid
}

#' @rdname register_block
#' @export
unregister_blocks <- function(uid = list_blocks()) {
  rm(list = uid, envir = block_registry, inherits = FALSE)
  invisible()
}

#' @param ... For `register_blocks()`, arguments forwarded to
#'   `register_block()`; for `new_block_args()`, named `new_block_arg()` objects
#'   (or strings), one per constructor formal
#' @rdname register_block
#' @export
register_blocks <- function(...) {
  arg_processor <- function(ctor, ...) {
    wrap_list <- function(x) {
      if (length(x) > 1L) list(x) else x
    }

    if (length(ctor) > 1L) {
      return(c(list(ctor), list(...)))
    }

    c(list(ctor), lapply(list(...), wrap_list))
  }

  invisible(
    do.call(Map, c(register_block, arg_processor(...)))
  )
}

get_registry_entry <- function(uid) {
  res <- get(uid, envir = block_registry, inherits = FALSE)
  stopifnot(is_registry_entry(res))
  res
}

#' @rdname register_block
#' @export
available_blocks <- function() {
  lapply(set_names(nm = list_blocks()), get_registry_entry)
}

registry_metadata_fields <- c(
  "name",
  "description",
  "details",
  "link",
  "guidance",
  "keywords",
  "category",
  "icon",
  "arguments",
  "package"
)

registry_scalar_field <- function(entry, field) {
  coal(attr(entry, field), NA_character_, fail_all = FALSE)
}

registry_keywords_field <- function(entry) {
  coal(attr(entry, "keywords"), character(), fail_all = FALSE)
}

registry_arguments_field <- function(entry) {

  args <- attr(entry, "arguments")

  as_legacy_arguments(
    args,
    attr(entry, "guidance"),
    block_examples_list(args, attr(entry, "examples"))
  )
}

#' @param blocks Character vector of registry IDs
#' @param fields Metadata fields
#'
#' @rdname register_block
#' @export
registry_metadata <- function(blocks = list_blocks(), fields = "all") {

  blockr_warn(
    "`registry_metadata()` is deprecated; use `block_metadata()` or the ",
    "`block_meta_*()` accessors instead.",
    class = "deprecated_registry_metadata",
    frequency = "once",
    frequency_id = "blockr_deprecated_registry_metadata"
  )

  stopifnot(is.character(blocks), is.character(fields))

  if (identical(fields, "all")) {
    fields <- registry_metadata_fields
  }

  fields <- match.arg(fields, registry_metadata_fields, several.ok = TRUE)

  reg <- lapply(blocks, get_registry_entry)
  scalar <- setdiff(fields, c("keywords", "arguments", "package"))

  res <- lapply(
    set_names(nm = scalar),
    function(f) chr_ply(reg, registry_scalar_field, f)
  )

  if ("keywords" %in% fields) {
    res <- c(res, list(keywords = lapply(reg, registry_keywords_field)))
  }

  if ("arguments" %in% fields) {
    res <- c(res, list(arguments = lapply(reg, registry_arguments_field)))
  }

  if ("package" %in% fields) {
    res <- c(
      res,
      list(package = chr_ply(lapply(reg, attr, "package"), coal, "local"))
    )
  }

  res <- res[intersect(fields, names(res))]

  if (length(fields) == 1L && length(blocks) == 1L) {
    return(res[[1L]])
  } else if (length(fields) == 1L) {
    return(set_names(res[[1L]], blocks))
  }

  list2DF(c(list(id = blocks), res))
}

#' @param id Block ID as reported by `list_blocks()`
#' @rdname register_block
#' @export
create_block <- function(id, ...) {
  ctor <- get_registry_entry(id)
  ctor(..., ctor = attr(ctor, "ctor_name"), ctor_pkg = attr(ctor, "package"))
}

#' @param which Character vector of block constructor names to register (or
#'   `"all"`), keying into the YAML registry generated by the
#'   [block_registration_roclet()]
#'
#' @rdname register_block
#' @export
register_package_blocks <- function(which = "all", package = pkg_name(),
                                    overwrite = TRUE) {

  file <- system.file("registry/blocks.yml", package = package)

  if (!nzchar(file)) {
    return(invisible())
  }

  entries <- yaml::read_yaml(file)

  if (identical(which, "all")) {
    which <- names(entries)
  } else {
    stopifnot(all(which %in% names(entries)), anyDuplicated(which) == 0L)
  }

  entries <- entries[which]

  register_blocks(
    names(entries),
    name = chr_xtr(entries, "name"),
    description = chr_xtr(entries, "description"),
    details = lapply(entries, `[[`, "details"),
    link = lapply(entries, `[[`, "link"),
    category = chr_xtr(entries, "category"),
    icon = lapply(entries, `[[`, "icon"),
    guidance = lapply(entries, `[[`, "guidance"),
    keywords = lapply(entries, `[[`, "keywords"),
    arguments = lapply(entries, yaml_block_arguments),
    examples = lapply(entries, yaml_block_examples),
    package = package,
    overwrite = overwrite
  )
}

yaml_block_examples <- function(entry) {
  coal(entry[["examples"]], list(), fail_all = FALSE)
}

yaml_block_arguments <- function(entry) {

  args <- entry[["arguments"]]

  if (!length(args)) {
    return(character())
  }

  do.call(
    new_block_args,
    lapply(
      args,
      function(a) new_block_arg(a[["description"]], a[["example"]], a[["type"]])
    )
  )
}

register_core_blocks <- function(which = blockr_option("core_blocks", "all")) {

  if (!identical(which, "all")) {
    which <- paste0("new_", which)
  }

  register_package_blocks(which = which)
}

test_that("registry", {

  curr <- list_blocks()

  withr::defer(
    {
      unregister_blocks()
      register_core_blocks(curr)
    }
  )

  unregister_blocks()

  expect_length(list_blocks(), 0L)

  candidates <- ls(
    envir = asNamespace("blockr.core"),
    pattern = "^new_.+_block$"
  )

  is_cand <- lgl_ply(
    candidates,
    function(x) {

      res <- try(do.call(x, list(block_metadata = list())), silent = TRUE)

      if (inherits(res, "try-error")) {
        return(FALSE)
      }

      length(grep("_block$", class(res), value = TRUE)) > 1L
    }
  )

  candidates <- candidates[is_cand]

  cand_1 <- do.call(candidates[1L], list(block_metadata = list()))

  expect_identical(
    registry_id_from_block(cand_1),
    character()
  )

  register_core_blocks("all")

  expect_setequal(
    paste0("new_", list_blocks()),
    candidates
  )

  expect_identical(
    registry_id_from_block(cand_1),
    sub("^new_", "", candidates[1L])
  )

  expect_error(register_block(candidates[1L], "test", "test"))

  expect_error(
    register_block(get(candidates[1L]), "test", "test"),
    class = "block_already_in_registry",
    regexp = "overwrite = TRUE"
  )

  expect_error(
    register_block(candidates[1L], "test", "test", package = "blockr.core"),
    class = "block_already_in_registry"
  )

  rlang::reset_warning_verbosity("block_category_test_discouraged")

  expect_warning(
    register_block(get(candidates[1L]), "test", "test", category = "test",
                   overwrite = TRUE),
    class = "block_category_discouraged"
  )

  rlang::reset_warning_verbosity("block_icon_cloud-arrow-up-fill_discouraged")

  expect_warning(
    register_block(get(candidates[1L]), "test", "test",
                   icon = "cloud-arrow-up-fill", overwrite = TRUE),
    class = "block_icon_fill_discouraged"
  )

  expect_error(
    register_block(get(candidates[1L]), "test", "test", icon = "test"),
    class = "block_icon_invalid"
  )
})

test_that("registry metadata", {

  withr::local_options(rlib_warning_verbosity = "quiet")

  meta <- registry_metadata()

  expect_s3_class(meta, "data.frame")
  expect_identical(nrow(meta), length(list_blocks()))
  expect_identical(ncol(meta), 11L)
  expect_named(
    meta,
    c("id", "name", "description", "details", "link", "guidance", "keywords",
      "category", "icon", "arguments", "package")
  )

  meta <- registry_metadata(list_blocks()[1L], "name")

  expect_type(meta, "character")
  expect_length(meta, 1L)

  meta <- registry_metadata(fields = "name")

  expect_type(meta, "character")
  expect_length(meta, length(list_blocks()))
  expect_named(meta, list_blocks())

  icon <- default_icon(default_category())

  expect_type(icon, "character")
  expect_length(icon, 1L)
})

#' Utilities for code export
#'
#' To facilitate other means of code export than implemented by the default
#' [generate_code()] plugin, this utility performs much of the heavy lifting
#' to properly arrange and scope block-level expressions.
#'
#' @param expressions Block expressions
#' @param board Board object
#'
#' @return String containing properly arranged block expressions.
#'
#' @export
export_code <- function(expressions, board) {

  lnks <- board_links(board)

  arg_map <- lapply(
    split(as.list(lnks), lnks$to),
    function(x) {
      set_names(
        lapply(lst_xtr(x, "from"), as.name),
        arg_refs(chr_xtr(x, "input"))
      )
    }
  )

  expr_type <- set_names(
    chr_ply(board_blocks(board), block_expr_type),
    board_block_ids(board)
  )

  ordering <- topo_sort(as.matrix(board))

  list(
    exprs = expressions[ordering],
    args = set_names(arg_map[ordering], ordering),
    types = expr_type[ordering]
  )
}

export_wrapped_code <- function(expressions, board) {

  exprs <- do.call(
    Map,
    c(
      list(wrap_expr),
      export_code(expressions, board)
    )
  )

  exprs <- map(assignment, names(exprs), exprs)
  exprs <- lapply(exprs, deparse)
  exprs <- chr_ply(exprs, paste0, collapse = "\n")

  paste0(exprs, collapse = "\n\n")
}

wrap_expr <- function(exprs, args, types) {

  if (identical(types, "bquoted")) {
    exprs <- do.call(bquote, list(exprs, args))
  }

  if (length(args) && identical(types, "quoted")) {
    call("with", args, exprs)
  } else {
    call("local", exprs)
  }
}

assignment <- function(name, value) {
  bquote(.(nme) <- .(val), list(nme = as.name(name), val = value))
}

test_that("code generation", {

  board <- generate_plugin_args(
    new_board(
      blocks = c(
        a = new_dataset_block("BOD"),
        b = new_dataset_block("ChickWeight"),
        c = new_merge_block(by = "Time")
      ),
      links = links(
        from = c("a", "b"),
        to = c("c", "c"),
        input = c("x", "y")
      )
    )
  )$board

  code_exp <- export_code(
    isolate(lst_xtr_reval(board$blocks, "server", "expr")),
    isolate(board$board)
  )

  expect_type(code_exp, "list")
  expect_length(code_exp, 3L)
  expect_named(code_exp, c("exprs", "args", "types"))

  expect_named(code_exp[["exprs"]], c("a", "b", "c"), ignore.order = TRUE)
  expect_named(code_exp[["args"]], names(code_exp[["exprs"]]))
  expect_named(code_exp[["types"]], names(code_exp[["exprs"]]))

  code_str <- export_wrapped_code(
    isolate(lst_xtr_reval(board$blocks, "server", "expr")),
    isolate(board$board)
  )

  expect_type(code_str, "character")
  expect_length(code_str, 1L)
})

#' @section Dataset block:
#' This data block allows to select a dataset from a package, such as
#' the datasets package available in most R installations as one of the
#' packages with "recommended" priority. The source package can be chosen at
#' time of block instantiation and can be set to any R package, for which then
#' a set of candidate datasets is computed. This includes exported objects that
#' inherit from `data.frame`.
#'
#' @param dataset Selected dataset
#' @param package Name of an R package containing datasets
#'
#' @block dataset block
#' @blockDescr Choose a dataset from a package
#' @blockCategory input
#' @blockIcon database
#' @blockGuidance Loads a data.frame that ships with an installed package. Set
#'   `package` to the source package (default "datasets"), then `dataset` to the
#'   name of a data.frame object it exports.
#' @blockKeywords dataset, data, import, package, source
#' @blockArg dataset Selects the dataset to use.
#'   [example] "iris"
#'   [type] arg_string()
#' @blockArg package R package to source the dataset from.
#'   [example] "datasets"
#'   [type] arg_string()
#'
#' @rdname new_data_block
#' @export
new_dataset_block <- function(dataset = character(), package = "datasets",
                              ...) {

  is_dataset_eligible <- function(x, pkg) {
    inherits(do.call("::", list(pkg, x)), "data.frame")
  }

  list_datasets <- function(package) {
    datasets <- utils::data(package = package)
    datasets <- datasets$results[, "Item"]

    options <- gsub("\\s+\\(.+\\)$", "", datasets)

    options[lgl_ply(options, is_dataset_eligible, package)]
  }

  new_data_block(
    function(id) {
      moduleServer(
        id,
        function(input, output, session) {

          dat <- reactiveVal(dataset)

          observeEvent(
            req(input$dataset),
            dat(input$dataset)
          )

          observeEvent(
            req(dat()),
            {
              if (!identical(dat(), input$dataset)) {
                updateSelectInput(
                  session,
                  "dataset",
                  choices = list_datasets(package),
                  selected = dat()
                )
              }
            }
          )

          list(
            expr = reactive(
              eval(
                bquote(
                  as.call(c(as.symbol("::"), quote(.(pkg)), quote(.(dat)))),
                  list(pkg = as.name(package), dat = as.name(dat()))
                )
              )
            ),
            state = list(
              dataset = dat,
              package = package
            )
          )
        }
      )
    },
    function(id) {
      selectInput(
        inputId = NS(id, "dataset"),
        label = "Dataset",
        choices = list_datasets(package),
        selected = dataset
      )
    },
    class = "dataset_block",
    external_ctrl = "dataset",
    ...
  )
}

#' Data block constructors
#'
#' Data blocks typically do not have data inputs and represent root nodes in
#' analysis graphs. Intended as initial steps in a pipeline, such blocks are
#' responsible for providing down-stream blocks with data.
#'
#' @param ... Forwarded to `new_data_block()` and [new_block()]
#' @inheritParams new_block
#'
#' @return All blocks constructed via `new_data_block()` inherit from
#' `data_block`.
#'
#' @seealso Real-world data blocks built on `new_data_block()` are provided by
#' the [blockr.io](https://bristolmyerssquibb.github.io/blockr.io/) package,
#' which sources data from a range of external formats.
#'
#' @export
new_data_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "data_block"), ctor, ...)
}

#' @export
block_output.data_block <- function(x, result, session) {
  dt_result(result, x, session)
}

#' @export
block_ui.data_block <- function(id, x, ...) {
  tagList(
    DT::dataTableOutput(NS(id, "result"))
  )
}

#' @export
#' @include utils-dt.R
block_render_trigger.data_block <- dt_render_trigger

#' @export
board_options.data_block <- function(x, ...) {
  combine_board_options(
    new_n_rows_option(...),
    new_page_size_option(...),
    new_filter_rows_option(...),
    NextMethod()
  )
}

#' @section Scatter block:
#' Mainly for demonstration purposes, this block draws a scattter plot using
#' [base::plot()]. In its current simplistic implementation, apart from axis
#' labels (fixed to the corresponding column names), no further plotting
#' options are available and for any "production" application, a more
#' sophisticated (set of) block(s) for data visualization will most likely be
#' required.
#'
#' @param x,y Columns to place on respective axes
#'
#' @block scatter plot block
#' @blockDescr Scatter plotting
#' @blockCategory plot
#' @blockIcon dice-5
#' @blockGuidance Draws a scatter plot of `x` against `y` with `base::plot()`.
#'   `x` and `y` are column names from the input data; axis labels follow the
#'   column names.
#' @blockKeywords scatter, scatter plot, chart, visualization, points
#' @blockArg x Column for the horizontal axis.
#'   [example] "Sepal.Length"
#'   [type] arg_string()
#' @blockArg y Column for the vertical axis.
#'   [example] "Sepal.Width"
#'   [type] arg_string()
#'
#' @rdname new_plot_block
#' @export
new_scatter_block <- function(x = character(), y = character(), ...) {
  new_plot_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {

          cols <- reactive(colnames(data()))

          x_col <- reactiveVal(x)
          y_col <- reactiveVal(y)

          observeEvent(input$xcol, x_col(input$xcol))
          observeEvent(input$ycol, y_col(input$ycol))

          observeEvent(
            cols(),
            {
              updateSelectInput(
                session,
                inputId = "xcol",
                choices = cols(),
                selected = x_col()
              )
              updateSelectInput(
                session,
                inputId = "ycol",
                choices = cols(),
                selected = y_col()
              )
            }
          )

          list(
            expr = reactive(
              bbquote(
                plot(.(data)[[.(x)]], data[[.(y)]], xlab = .(x_col),
                     ylab = .(y_col)),
                list(x = x_col(), y = y_col(), x_col = x_col(),
                     y_col = y_col())
              )
            ),
            state = list(x = x_col, y = y_col)
          )
        }
      )
    },
    function(id) {
      tagList(
        selectInput(
          inputId = NS(id, "xcol"),
          label = "X-axis",
          choices = x,
          selected = x
        ),
        selectInput(
          inputId = NS(id, "ycol"),
          label = "Y-axis",
          choices = y,
          selected = y
        )
      )
    },
    dat_valid = function(data) {
      stopifnot(is.data.frame(data) || is.matrix(data))
    },
    expr_type = "bquoted",
    class = "scatter_block",
    ...
  )
}

#' Block registration roclet
#'
#' A custom roxygen2 roclet that turns block registration metadata, declared as
#' tags on block constructors, into a YAML registry (`inst/registry/blocks.yml`)
#' that [register_package_blocks()] reads at load time. It runs alongside the
#' standard roclets during [roxygen2::roxygenise()] (or `devtools::document()`)
#' for any package that lists `blockr.core::block_registration_roclet` in the
#' `Roxygen` field of its `DESCRIPTION`.
#'
#' @section Tags:
#' Placed in the roxygen block above a block constructor:
#' \describe{
#'   \item{`@block <name>`}{Human-readable block name. Required: its presence is
#'     what marks a constructor for registration.}
#'   \item{`@blockDescr <text>`}{Short block description. Required.}
#'   \item{`@blockCategory <category>`}{One of the [suggested_categories()].
#'     Required.}
#'   \item{`@blockIcon <icon>`}{Bootstrap icon name. Optional, defaulting to the
#'     category icon (see [default_icon()]).}
#'   \item{`@blockGuidance <text>`}{Model-facing construction guidance -- do and
#'     don't rules, enumerations, pitfalls. Optional.}
#'   \item{`@blockKeywords <terms>`}{Comma-separated search terms for block
#'     discovery; a term may contain spaces. Optional.}
#'   \item{`@blockDetails <text>`}{Longer human-facing description (e.g. for a
#'     help popover). Optional; when omitted it falls back to the constructor's
#'     `@details` or, failing that, its `@section` prose.}
#'   \item{`@blockLink <url>`}{URL of the block's help or documentation page.
#'     Optional; when omitted it is derived from the package's pkgdown `url`
#'     (`_pkgdown.yml` or `DESCRIPTION`) and the documented topic, as
#'     `<url>/reference/<topic>.html`.}
#'   \item{`@blockArg <name> <description>`}{One constructor argument's
#'     specification. The text after the name is a free-text description;
#'     `[example] <expr>` and `[type] <expr>` markers (each on its own line)
#'     supply an R expression -- evaluated when documentation is generated --
#'     for a worked example value and an `arg_*()` type descriptor (see
#'     [new_block_arg()]). An explicit `[description] <text>` marker may replace
#'     the inline description. Optional and repeatable.}
#'   \item{`@blockExamples <expr>`}{Block-level worked example configurations,
#'     as an R expression -- evaluated when documentation is generated -- that
#'     yields a list of complete configurations (each a named list keyed by
#'     argument). Optional; supersedes the per-argument `[example]` assembly.}
#'   \item{`@blockCtor <name>`}{Constructor name override. Optional: the
#'     documented object otherwise supplies the name.}
#' }
#' Multi-argument worked `examples` (whole-block configurations, as opposed to a
#' per-argument `[example]`) remain out of the tags' scope; a block needing
#' those registers via [register_block()] directly.
#'
#' @return `block_registration_roclet()` returns a roclet object.
#'
#' @name block_roclet
#' @export
block_registration_roclet <- function() {
  roxygen2::roclet("block_registration")
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_block <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockDescr <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockCategory <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockIcon <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockGuidance <- function(x) {
  roxygen2::tag_value(x, multiline = TRUE)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockKeywords <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockDetails <- function(x) {
  roxygen2::tag_value(x, multiline = TRUE)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockLink <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockExamples <- function(x) {
  roxygen2::tag_value(x, multiline = TRUE)
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockArg <- function(x) {

  raw <- trimws(x[["raw"]])
  split <- regexpr("[[:space:]]", raw)

  if (split < 0L) {
    return(roxygen2::roxy_tag_warning(x, "requires an argument name"))
  }

  x[["val"]] <- c(
    list(name = substr(raw, 1L, split - 1L)),
    block_arg_fields(substring(raw, split + 1L))
  )

  x
}

#' @exportS3Method roxygen2::roxy_tag_parse
roxy_tag_parse.roxy_tag_blockCtor <- function(x) {
  roxygen2::tag_value(x)
}

#' @exportS3Method roxygen2::roclet_process
roclet_process.roclet_block_registration <- function(x, blocks, env,
                                                     base_path) {
  base_url <- pkgdown_site_url(base_path)
  entries <- list()

  for (block in blocks) {

    if (is.null(roxygen2::block_get_tag(block, "block"))) {
      next
    }

    entries[[block_roclet_ctor(block)]] <- block_roclet_entry(block, base_url)
  }

  entries
}

#' @exportS3Method roxygen2::roclet_output
roclet_output.roclet_block_registration <- function(x, results, base_path,
                                                    ...) {
  if (!length(results)) {
    return(invisible(character()))
  }

  dir <- file.path(base_path, "inst", "registry")
  dir.create(dir, recursive = TRUE, showWarnings = FALSE)

  path <- file.path(dir, "blocks.yml")

  writeLines(
    c(block_roclet_header(), "", yaml::as.yaml(results)),
    path
  )

  invisible(path)
}

#' @exportS3Method roxygen2::roclet_clean
roclet_clean.roclet_block_registration <- function(x, base_path) {

  path <- file.path(base_path, "inst", "registry", "blocks.yml")

  generated <- file.exists(path) &&
    identical(readLines(path, n = 1L, warn = FALSE), block_roclet_header())

  if (generated) {
    unlink(path)
  }

  invisible()
}

block_roclet_header <- function() {
  "# Generated by blockr.core block_registration roclet: do not edit by hand"
}

block_roclet_ctor <- function(block) {

  ctor <- roxygen2::block_get_tag(block, "blockCtor")

  if (not_null(ctor)) {
    return(ctor$val)
  }

  name <- coal(block$object$alias, block$object$topic, fail_all = FALSE)

  if (is.null(name) && not_null(block$call)) {
    name <- as.character(block$call[[2L]])
  }

  if (!is_string(name)) {
    block_roclet_abort(
      roxygen2::block_get_tag(block, "block"),
      "Unable to determine the block constructor name."
    )
  }

  name
}

block_roclet_entry <- function(block, base_url) {

  require_block_tag(block, "blockDescr")
  require_block_tag(block, "blockCategory")

  entry <- list(
    name = block_tag_value(block, "block"),
    description = block_tag_value(block, "blockDescr"),
    category = block_tag_value(block, "blockCategory")
  )

  details <- block_details(block)

  if (not_null(details)) {
    entry[["details"]] <- details
  }

  link <- block_link(block, base_url)

  if (not_null(link)) {
    entry[["link"]] <- link
  }

  icon <- block_tag_value(block, "blockIcon")

  if (not_null(icon)) {
    entry[["icon"]] <- icon
  }

  guidance <- block_guidance(block)

  if (not_null(guidance)) {
    entry[["guidance"]] <- guidance
  }

  keywords <- block_keywords(block)

  if (length(keywords)) {
    entry[["keywords"]] <- keywords
  }

  args <- block_roclet_args(block)

  if (length(args)) {
    entry[["arguments"]] <- args
  }

  examples <- block_examples(block)

  if (length(examples)) {
    entry[["examples"]] <- examples
  }

  entry
}

block_roclet_args <- function(block) {

  vals <- lapply(roxygen2::block_get_tags(block, "blockArg"), `[[`, "val")

  set_names(lapply(vals, block_arg_spec), chr_xtr(vals, "name"))
}

block_arg_fields <- function(body) {

  lines <- strsplit(body, "\n", fixed = TRUE)[[1L]]

  fields <- list()
  key <- "description"

  for (line in lines) {

    marker <- regmatches(
      line,
      regexec("^[[:space:]]*\\[(description|example|type)\\][[:space:]]*(.*)$",
              line)
    )[[1L]]

    if (length(marker)) {
      key <- marker[[2L]]
      value <- marker[[3L]]
    } else {
      value <- trimws(line)
    }

    fields[[key]] <- trimws(paste(c(fields[[key]], value), collapse = " "))
  }

  Filter(nzchar, fields)
}

block_arg_spec <- function(val) {
  compact_arg(
    val[["description"]],
    eval_arg_expr(val[["example"]]),
    eval_arg_expr(val[["type"]])
  )
}

eval_arg_expr <- function(expr) {

  if (is.null(expr)) {
    return(NULL)
  }

  eval(str2lang(expr), new.env(parent = asNamespace("blockr.core")))
}

compact_arg <- function(description, example, type) {

  arg <- list(description = description, example = example, type = type)

  arg[!lgl_ply(arg, is.null)]
}

require_block_tag <- function(block, tag) {

  if (is.null(roxygen2::block_get_tag(block, tag))) {
    block_roclet_abort(
      roxygen2::block_get_tag(block, "block"),
      paste0("A block declared with @block must also provide @", tag, ".")
    )
  }

  invisible()
}

block_tag_value <- function(block, tag) {

  hit <- roxygen2::block_get_tag(block, tag)

  if (is.null(hit)) NULL else hit$val
}

block_tag_raw <- function(block, tag) {

  hit <- roxygen2::block_get_tag(block, tag)

  if (is.null(hit)) NULL else trimws(hit[["raw"]])
}

block_keywords <- function(block) {

  value <- block_tag_value(block, "blockKeywords")

  if (is.null(value)) {
    return(character())
  }

  parts <- trimws(strsplit(value, ",", fixed = TRUE)[[1L]])

  parts[nzchar(parts)]
}

block_guidance <- function(block) {
  reflow_text(block_tag_value(block, "blockGuidance"))
}

block_details <- function(block) {

  explicit <- coal(
    block_tag_value(block, "blockDetails"),
    block_tag_raw(block, "details"),
    fail_all = FALSE
  )

  if (not_null(explicit)) {
    return(reflow_text(explicit))
  }

  section <- block_tag_raw(block, "section")

  if (is.null(section)) {
    return(NULL)
  }

  reflow_text(sub("^[^:]+:[[:space:]]*", "", section))
}

block_link <- function(block, base_url) {

  explicit <- block_tag_value(block, "blockLink")

  if (not_null(explicit)) {
    return(explicit)
  }

  url <- derived_link(block, base_url)

  if (not_null(url) && url_exists(url)) {
    return(url)
  }

  NULL
}

derived_link <- function(block, base_url) {

  if (is.null(base_url)) {
    return(NULL)
  }

  topic <- coal(
    block_tag_value(block, "rdname"),
    block_tag_value(block, "name"),
    block_roclet_ctor(block),
    fail_all = FALSE
  )

  paste0(sub("/?$", "/", base_url), "reference/", topic, ".html")
}

url_exists <- function(url) {

  status <- tryCatch(
    attr(curlGetHeaders(url, timeout = 10L), "status"),
    error = function(e) NA_integer_
  )

  isTRUE(status >= 200L && status < 300L)
}

pkgdown_site_url <- function(base_path) {

  configs <- file.path(
    base_path,
    c("_pkgdown.yml", "pkgdown/_pkgdown.yml", "inst/_pkgdown.yml")
  )

  for (path in configs[file.exists(configs)]) {

    url <- yaml::read_yaml(path)[["url"]]

    if (is_string(url)) {
      return(url)
    }
  }

  desc <- file.path(base_path, "DESCRIPTION")

  if (!file.exists(desc)) {
    return(NULL)
  }

  url <- read.dcf(desc, "URL")[1L]

  if (is.na(url)) {
    return(NULL)
  }

  trimws(strsplit(url, "[,[:space:]]+")[[1L]][1L])
}

block_examples <- function(block) {
  eval_arg_expr(block_tag_value(block, "blockExamples"))
}

reflow_text <- function(text) {

  if (is.null(text)) {
    return(NULL)
  }

  gsub("\\s+", " ", trimws(text))
}

block_roclet_abort <- function(tag, message) {
  cli::cli_abort(
    paste0("{tag$file}:{tag$line}: ", message),
    class = "block_roclet_invalid"
  )
}

#' Board loader
#'
#' Which `board` to build for an incoming request -- and how a board is carried
#' across the `session$reload()` that a [preserve_board] restore triggers -- is
#' the job of an app-level **board loader**, passed to [serve()] as its
#' `loader` argument. A `board_loader()` pairs a `resolve(request, session,
#' default)` -- returning the `board` to build for an incoming request, or
#' `NULL` for the [serve()] default -- with an optional `stage(board,
#' session)`, which persists a board and returns the URL query parameters that
#' reference it (a resolve-only loader, e.g. one not backing a restore, leaves
#' `stage` `NULL`). [serve()] uses that one loader for both the request-phase
#' resolution (at the GET, where `session` is `NULL`, and at the WS connect)
#' and the in-session staging when a restore fires; **core** writes those
#' parameters into the URL and drives the reload, so the reload stays a
#' guaranteed core mechanism that no loader can opt out of.
#'
#' `resolve` receives the raw HTTP `request` at both phases (the UI request at
#' the GET, `session$request` at the WS), the `session` (`NULL` at the GET),
#' and `default` -- the `board` passed to [serve()] (the [serve()] default,
#' built when `resolve` returns `NULL`), which a loader can also derive its own
#' result from (e.g. `clear_board(default)`). A loader that keys off URL query
#' parameters reads them itself, minding the phase split: the query is on
#' `request$QUERY_STRING` at the GET but `session$clientData$url_search` at the
#' WS (the websocket request carries neither).
#'
#' The default [local_loader()] keeps its handoff in a per-loader store (no
#' process global) and is therefore single-process; multi-user deployments pass
#' a loader resolving from a shared backend (as blockr.session does).
#'
#' @param resolve,stage Paired functions backing a `board_loader`: `resolve` is
#' `function(request, session, default)` returning the `board` to build or
#' `NULL`; `stage` is `function(board, session)` (or `NULL` for a loader
#' that does not stage, e.g. resolve-only) which persists `board` and returns
#' the URL query parameters referencing it (core writes them and reloads). They
#' share private state, so a `board_loader` is built as a unit, not supplied as
#' two loose functions.
#' @param x Object to test for `board_loader`-ness
#'
#' @return `board_loader()` and `local_loader()` return a `board_loader` object
#' and `is_board_loader()` a scalar logical.
#'
#' @export
board_loader <- function(resolve, stage = NULL) {
  validate_board_loader(
    structure(
      list(resolve = resolve, stage = stage),
      class = "board_loader"
    )
  )
}

validate_board_loader <- function(x) {

  if (!is_board_loader(x)) {
    blockr_abort(
      "Expecting a `board_loader` object.",
      class = "board_loader_invalid"
    )
  }

  resolve_ok <- is.function(x$resolve) &&
    formals_match(x$resolve, c("request", "session", "default"))

  if (!resolve_ok) {
    blockr_abort(
      "A `board_loader` `resolve` must be a function of `request`, `session` ",
      "and `default`.",
      class = "board_loader_resolve_invalid"
    )
  }

  stage_ok <- is.null(x$stage) ||
    (is.function(x$stage) && formals_match(x$stage, c("board", "session")))

  if (!stage_ok) {
    blockr_abort(
      "A `board_loader` `stage` must be `NULL` or a function of `board` and ",
      "`session`.",
      class = "board_loader_stage_invalid"
    )
  }

  x
}

formals_match <- function(f, args) {
  identical(names(formals(f)), args)
}

#' @rdname board_loader
#' @export
is_board_loader <- function(x) {
  inherits(x, "board_loader")
}

reload_param <- "__blockr_reload__"

#' @rdname board_loader
#' @export
local_loader <- function() {

  store <- new.env(parent = emptyenv())

  read_query <- function(request, session) {

    search <- if (is.null(session)) {
      request[["QUERY_STRING"]]
    } else {
      isolate(session$clientData$url_search)
    }

    parseQueryString(coal(search, ""))
  }

  query_string <- function(query) {

    if (!length(query)) {
      return("?")
    }

    nms <- chr_ply(names(query), utils::URLencode, reserved = TRUE)
    vals <- chr_ply(unlst(query), utils::URLencode, reserved = TRUE)

    paste0("?", paste(nms, vals, sep = "=", collapse = "&"))
  }

  write_token <- function(session, token) {

    query <- read_query(NULL, session)
    query[[reload_param]] <- token

    updateQueryString(query_string(query), mode = "replace", session = session)
  }

  resolve <- function(request, session, default) {

    token <- read_query(request, session)[[reload_param]]

    if (is.null(token)) {
      return(NULL)
    }

    board <- get0(token, envir = store, inherits = FALSE)

    if (not_null(session)) {

      if (not_null(board)) {
        rm(list = token, envir = store)
      }

      write_token(session, NULL)
    }

    board
  }

  stage <- function(board, session) {

    token <- rand_names(ls(envir = store))
    assign(token, board, envir = store)

    log_debug("staging board for reload handoff {token}")

    write_token(session, token)
  }

  board_loader(resolve, stage)
}

resolve_board <- function(default, loader, request, session,
                          options = blockr_app_options) {

  board <- loader$resolve(request, session, default)

  if (is.null(board)) {
    board <- default
  } else if (!is_board(board)) {
    blockr_abort(
      "A board loader's `resolve` must return `NULL` or a `board`.",
      class = "invalid_board_loader"
    )
  } else {
    board <- validate_board(board)
  }

  # A board declares only its own options, but the settings UI manages the
  # wider `options(board)` set (board options plus those contributed by blocks
  # and the registry). Bake it onto the resolved board so the UI, the server
  # and serialization all see the same set -- otherwise a user-changed,
  # block-contributed option is lost on save.
  board_options(board) <- options(board)

  board
}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)


## ----eval = FALSE-------------------------------------------------------------
# test_that("block constructor", {
#     expect_s3_class(new_test_block(), "test_block")
# })


## ----eval = FALSE-------------------------------------------------------------
# test_that("test block server input widgets update reactive values", {
#     testServer(
#         app = new_test_block()$expr_server, # Capture the expression server
#         expr = {
#             session$setInputs(color = "green") # Set a color
#             expect_equal(color(), "green")
# 
#             session$setInputs(color = "red") # Change color
#             expect_equal(color(), "red")
#         }
#     )
# })


## ----eval = FALSE-------------------------------------------------------------
# test_that("state is correctly returned", {
#     testServer(
#         app = new_test_block()$expr_server,
#         expr = {
#             # First check default values
#             expect_equal(session$returned$state$number(), numeric())
#             expect_equal(session$returned$state$color(), character())
# 
#             # Then toggle the inputs and recheck state is updated correctly
#             session$setInputs(number = 1)
#             expect_equal(session$returned$state$number(), 1)
#             session$setInputs(color = "pink")
#             expect_equal(session$returned$state$color(), "pink")
#         }
#     )
# })


## ----eval = FALSE-------------------------------------------------------------
# test_that("expr evaluates correctly", {
#     testServer(
#         app = new_test_block()$expr_server,
#         expr = {
#             session$setInputs(number = 1)
#             session$setInputs(color = "cyan")
# 
#             # Call `base::eval()` on our expression
#             evaluated_expr <- eval(session$returned$expr())
#             expect_s3_class(evaluated_expr, "colored_number")
#         }
#     )
# })


## ----eval = FALSE-------------------------------------------------------------
# test_that("incorrect colours throw an error", {
#     testServer(
#         app = new_test_block()$expr_server,
#         expr = {
#             session$setInputs(number = 2)
# 
#             # Set invalid colour
#             session$setInputs(color = "Ooops")
# 
#             # Check that an error is thrown
#             expect_error(eval(session$returned$expr()))
#         }
#     )
# })


## ----eval = FALSE-------------------------------------------------------------
# test_that("test block server input widgets update reactive values", {
#     testServer(
#         app = new_test_block()$expr_server,
#         # Set block arguments with the "args" argument
#         args = list(df = reactive(gt::gt(mtcars))),
#         expr = {
#             session$setInputs(color = "green")
#             expect_equal(color(), "green")
#         }
#     )
# })

---
title: "5. Testing blocks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{5. Testing blocks}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

As a block developer, you may want to write unit tests for any blocks you create.
Testing blocks is more complex than writing unit tests for standard R code.
For blocks, we need methods to both interact with the reactive shiny values in the blocks, and to capture the state and expressions returned by our blocks.

In both instances, we can take advantage of `shiny::testServer()`.
This function allows us to simulate a blockr application, without having to spin up a full app (which would be time consuming to create and result in slow tests). This vignette will show how to apply `shiny::testServer()` to a suite of common test cases. For more details on how `shiny::testServer()` works, we recommend you see [this](https://shiny.posit.co/r/articles/improve/server-function-testing/) article in the official Shiny docs.

## Common test cases

#### Test class of object

Before interacting with a block, it is often a good idea to test that the block constructor has worked as intended and created the correct class of object.
For this example, let's create a pretend "test_block" class:

```{r, eval = FALSE}
test_that("block constructor", {
    expect_s3_class(new_test_block(), "test_block")
})
```

#### Test input widgets work

Next, you may want to test that the input widgets to your block work as intended.
That is, that changes in inputs are reflected by changes in reactive values in the block.
For this example, let's imagine out test block has an input widget that allows the user to select a color:

```{r, eval = FALSE}
test_that("test block server input widgets update reactive values", {
    testServer(
        app = new_test_block()$expr_server, # Capture the expression server
        expr = {
            session$setInputs(color = "green") # Set a color
            expect_equal(color(), "green")

            session$setInputs(color = "red") # Change color
            expect_equal(color(), "red")
        }
    )
})
```

#### Test that state is correctly returned

After confirming our input widgets work as expected, we may then want to check that the "state" of our block is returned correctly.
Let's expand our previous example to add a second input widget, number, which allows the user to select a numeric value.
Here we access the state returned by our block by indexing into the `session$returned` object returned by `shiny::testServer()`:

```{r, eval = FALSE}
test_that("state is correctly returned", {
    testServer(
        app = new_test_block()$expr_server,
        expr = {
            # First check default values
            expect_equal(session$returned$state$number(), numeric())
            expect_equal(session$returned$state$color(), character())

            # Then toggle the inputs and recheck state is updated correctly
            session$setInputs(number = 1)
            expect_equal(session$returned$state$number(), 1)
            session$setInputs(color = "pink")
            expect_equal(session$returned$state$color(), "pink")
        }
    )
})
```

#### Test that expressions are correctly returned

In addition to checking the "state" returned by our block, we can also check that the "expr" is correctly evaluated.
To evaluate our expressions, we can call `base::eval()` on our returned expression, and then check that this evaluated expressions matches our expression.
In this example, let's assume that our input widgets combine in an expression to build a new "colored_number" class which just prints a colored number to the screen:

```{r, eval = FALSE}
test_that("expr evaluates correctly", {
    testServer(
        app = new_test_block()$expr_server,
        expr = {
            session$setInputs(number = 1)
            session$setInputs(color = "cyan")

            # Call `base::eval()` on our expression
            evaluated_expr <- eval(session$returned$expr())
            expect_s3_class(evaluated_expr, "colored_number")
        }
    )
})
```

#### Test that expressions correctly throw errors
We may also want to test that errors are correctly thrown for incorrect expressions:

```{r, eval = FALSE}
test_that("incorrect colours throw an error", {
    testServer(
        app = new_test_block()$expr_server,
        expr = {
            session$setInputs(number = 2)

            # Set invalid colour
            session$setInputs(color = "Ooops")

            # Check that an error is thrown
            expect_error(eval(session$returned$expr()))
        }
    )
})
```

#### Passing additional args

It is likely that some of your blocks require data or other blocks to be passed in as arguments to the block.
To do this in our test suite, we can pass any such objects in a list to the `args` object of `testServer()`.
For instance, let's assume our test block requires a data.frame to be passed to the argument "df":

```{r, eval = FALSE}
test_that("test block server input widgets update reactive values", {
    testServer(
        app = new_test_block()$expr_server,
        # Set block arguments with the "args" argument
        args = list(df = reactive(gt::gt(mtcars))),
        expr = {
            session$setInputs(color = "green")
            expect_equal(color(), "green")
        }
    )
})
```

---
title: "3. Browse for blocks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{3. Browse for blocks}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(blockr.core)
```

## Introduction

The __registry__ is a __environment__ which provides access to multiple __blocks__ as well as some
metadata:

- The block __description__.
- The block __category__.
- The block __package__.
- ...

In other words, the __registry__ is a __"supermarket"__ for data analysis. As shown below, if you develop your own blocks package and registers blocks on load, these blocks become available to the end user. Therefore this makes it powerful for __collaboration__ between data science teams.

```{r, echo=FALSE}
include_mermaid("registry")
```

## Previewing available blocks

Upon loading, `{blockr}` __registers__ its internal __blocks__ with `register_core_blocks()`.
You won't have to call this function as it is not exported anyway. This makes the __registry__ environment
ready to be queried by `available_blocks()`. A truncated output example below:

```{r, eval=FALSE}
available_blocks()[["dataset_block"]]
```

```r
function(dataset = character(), package = "datasets", ...) {
  ...
}
<environment: namespace:blockr.core>
attr(,"name")
[1] "dataset block"
attr(,"description")
[1] "Choose a dataset from a package"
attr(,"classes")
[1] "dataset_block" "data_block"    "block"         "vctrs_vctr"    "list"         
attr(,"category")
[1] "input"
attr(,"ctor_name")
[1] "new_dataset_block"
attr(,"package")
[1] "blockr.core"
attr(,"class")
[1] "block_registry_entry"
```

```{r, eval=TRUE}
names(available_blocks())
```

## Register a block
To register your own blocks, user facing functions are:

- `register_block()` to register a block in the __registry__. If the __block__ is already registered,
it __overwrites__ the existing one.
- `register_blocks()` to register multiple blocks.

Note that, if you develop your block outside a package context, you must call `register_block` (or `register_blocks`)
passing the constructor as a function and not as a string (see below).

Let's say you want to create a new `new_dummy_block` which does nothing specific in the `./R/dummy-block.R` script:

```{r, eval=TRUE}
# ./R/dummy-block.R
new_dummy_block <- function(text = "Hello World", ...) {
  new_data_block(
    function(id) {
      moduleServer(id, function(input, output, session) {
        list(
          expr = reactive(quote(text)),
          state = list(text = text)
        )
      })
    },
    function(id) {
      shiny::tagList()
    },
    class = "dummy_block",
    ...
  )
}

register_dummy_blocks <- function() {
  register_blocks(
    c(new_dummy_block),
    name = c("dummy block"),
    description = c("A block that does nothing"),
    overwrite = TRUE
  )
}

register_dummy_blocks()
```

Finally, we create a `.R/zzz.R` script where you run this code to register the block(s) whenever the package loads:

```r
# ./R/zzz.R
.onLoad <- function(libname, pkgname) {
  register_dummy_blocks()
  invisible(NULL)
}
```

If we now query the registry, the new block is available:

```{r}
available_blocks()[["dummy_block"]]
```

## Unregister a block

The counterpart of `register_block()` is `unregister_blocks()`. We can remove
our new `dummy_block` from the registry:

```{r}
unregister_blocks(uid = "dummy_block")

# Check it out
names(available_blocks())
```

where __uid__ is/are the block(s) class(s) passed in the constructor.

---
title: "1. Getting started"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{1. Getting started}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(blockr.core)
```

## {blockr} for non coders

In that case, you likely just want to get started to create your first data pipeline. After all,
isn't that a promise of blockr framework?

### The user interface

Leveraging the default `blockr.core` UI displayed below, let's see how easily you can set your first data pipeline.

```{r empty-board, eval = FALSE}
serve(
  new_board()
)
```

```{r empty-board-shinylive_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "webr::install(\"blockr.dplyr\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(blockr.dplyr)",
    "library(bsicons)",
    knitr::knit_code$get("empty-board")
  ),
  collapse = "\n"
)

empty_board_url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", empty_board_url))
```

```{r empty-board-demo, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()}
knitr::include_url(empty_board_url, height = "800px")
```

1. In the top navigation bar, click on `Add block`. Select `dataset_block` and give it `data` as custom id. Click `ok`.

Each block is represented as a __card__ containing the block name, the __output__ data and a parameter icon on the top right side. The latter may be used to remove the block, change its name or append (add after) a new block. Some block can contain __inputs__. In our case, we can change the __output__ data.

2. Click on `Add block` and add a new `Select block`, giving it `transform` as custom id.

Notice the error message in red. You can also see that none of the select block field can be changed. This basically means no data is fed into this block. Let's change this.

3. In the top navbar, click on `Edit links`. This opens a modal window. Click on `Add` and then select `data` in the `From` dropdown, `transform` in `To` and `data` as `Input`. This creates a link from the data block to the select block. Click on ok.
4. On the select block, you can now select the `Chick` column. A new table is displayed when you do so.

Notice that we could have done that in a much faster way.

5. Click on the select block top right option icon and remove the block.
6. From the dataset block, click on `Append block` and select the `Select block`. A __link__, that is a connection, is automatically created for you. This now means that any change in the dataset block propagates down to the select block. 
7. Select the `Chick` column. you can try to change the data back to `BOD` and see an error message `Block transform: Can't select columns that don't exist. ✖ Column 'Chick' doesn't exist.`.

Did you enjoy this first round? Let's meet for the next steps.

### Group nodes

As your analysis grows, it may be convenient to group blocks by functions. This can be achieved through __stacks__. From the previous analysis

8. Click on `Edit stacks` (top navbar), click `Add` and select the blocks to add and click ok.

Blocks are now part of a collapsible unit (accordion) for convenience.

### Code export

Want to share or reproduce your analysis? Click on `Show code` and copy and paste the produce code.

### Save and restore your work

Wouldn't it be nice to save this work? At anytime you may click on `Save`. Then restart the app, click on `Restore` and select a previously saved file (snapshot). Note that this won't work from the documentation website due to limitations with shinylive.

### Other options

You may have notice in the top right side of the navbar. An option icon is also available at the entire application level. At the moment, you can change some table options like the number of pages, the number of previewed data, toggle search.

### More sophisticated UIs

`blockr.ui` exposes a fully customized [user experience](https://shinylive.io/r/app/#h=0&code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAdzgCMAnRRASwgGdSoAbbgCgA6YOtyIEA1gwwBXFkNwACBnFRF2CgLwKhAC1KlU7RAHpjBAJ4RJUDAHMWpHdLoYWRYyLGSMtOlmNCAJQCEL5MrBxcvILCohJSACao3OYM8koqapraYHoGRqYWVgw29o7Oru6e8T70-kEhYcxsnDz8QtXe7AmkMOnKqupauvqGJmaW1nYOTi5uHnHevvVgwaH04S1R7bFeUm79mUM5eWOFkyXT5XNVi1LLAash3CyMUAzmfJ1Ssmsvbx8vncMEkUgw-q8SoDvhhur0IQDPjC3GsGNIIAB9BJwGBEDFoVB8QJgAC+AF0gA) built on top of `blockr.core`.

If none of the solution meet your needs, keep in mind that `blockr.core` is entirely customizable, which means you can talk to your developer team and get a new app.

## {blockr} for developers

As a developer you may have to serve blockr applications or develop custom blocks. We provide a introduction below.

### Introduction

`blockr.core` decomposes entire data pipeline workflows into smaller units called __block__, responsible for performing a single __task__. At the end of the day, you obtain a __DAG__ to create powerful data __workflows__.

```{r, echo=FALSE}
include_mermaid("board-dag")
```

At the top level, a blockr app is made of a __board__ object. That board may contains __blocks__, __stacks__ or group of blocks, and connections between blocks, also known as __links__. You may find interesting to read more about block metadata in this [vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/blocks-registry.html).

### Serve blocks

This is mostly available for testing, you can serve a block as an entire application.
Below is all what you need to spin up an app with a single block:

```{r eval = FALSE}
serve(new_dataset_block("iris"))
```

For a block that requires input, additionally pass static data:

```{r, eval = FALSE}
serve(
  new_merge_block(by = "name"),
  data = list(x = datasets::BOD, y = datasets::ChickWeight)
)
```

Notice that in the above example, data is a named list with `x` and `y`. How to name this list is given by the signature
of the corresponding block server function (see truncated example below with `new_merge_block()`):

```r
new_merge_block <- function(by = character(), all_x = FALSE, all_y = FALSE, ...) {

  new_transform_block(
    server = function(id, x, y) {
      moduleServer(
        id,
        function(input, output, session) {
          # ... server logic ...
        })
    })
}
```

### Serve a board

In most of the case, you want to serve an entire board.

```{r eval = FALSE}
serve(new_board())
```

This single line of code spins up an ready-to-use blockr board. In some cases, it is 
possible to initialize the board with few blocks, stacks and links:

```{r simple-board, eval = FALSE}
serve(
  new_board(
    blocks = c(
      a = new_dataset_block("BOD"),
      b = new_dataset_block("ChickWeight"),
      c = new_merge_block("Time")
    ),
    links = c(
      ac = new_link("a", "c", "x"),
      bc = new_link("b", "c", "y")
    ),
    stacks = list(ac = c("a", "c"))
  )
)
```

```{r shinylive_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")", #nolint
    "library(blockr.core)",
    "library(bsicons)",
    knitr::knit_code$get("simple-board")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
cat(sprintf("[Open in Shinylive](%s)\n\n", url))
```

```{r shinylive_iframe, echo = FALSE, out.width = '150%', out.extra = 'style = "position: relative; z-index:1"', eval = knitr::is_html_output()}
knitr::include_url(url, height = "800px")
```

You can then deploy this application to any hosting solution like 
Posit Connect, Shiny server or even shinylive like in this example.

### Use the board as a module

`serve()` assumes the board to be at the top level of the shiny app module structure. Under the hood, `serve()` returns
a `shinyapp` object. You may need to embed the board into a larger app, which is also possible. There are two entry points and you
may look at the `serve.board` method for reference:

- `board_ui()`: the board UI module.
- `board_server()`: the board server module.

```r
ui <- bslib::page_fluid(
  board_ui(id, x, plugins)
)
server <- function(input, output, session) {
  res <- board_server(id, x, plugins)
}
shinyApp(ui, server)
```

### What's next?

#### Develop new blocks

If you had to develop custom blocks, you can read this [vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/create-block.html).

#### Extend blockr.core

If you wish to customize the look and feel of block, look no further than this [article](https://bristolmyerssquibb.github.io/blockr.core/articles/extend-blockr.html).

---
title: "2. Create a block"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. Create a block}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(blockr.core)
```

## Creating blocks

**Note:** If you are not already familiar with [Shiny modules](https://shiny.posit.co/r/articles/improve/modules/), it is recommended you familiarize yourself with them before continuing with this section.

`blockr.core` is built on top of Shiny. At it's heart, a __block__ is nothing more than a specialized Shiny __module__ that returns two additional values:

- **An expression** (called "expr"), a reactive expression defining a blocks computation. This allows R code to be exported and  data analyses to be recreated, outside of blockr and a reactive context.
- **A state object** (called "state"), a list of reactive values tracking user inputs. This allows blocks to separate user-defined state from inputs specified by blocks.

```{mermaid}
flowchart LR
  subgraph block[Block]
    subgraph ctor[constructor]
      block_ui[UI]
      subgraph block_server[server]
        blk_state[State]
        blk_expr[Expression]
      end
    end
  end
```

While blocks *return* the "expr" and "state" values, blocks *consist* of three elements:

1. A **UI function** to define the user interface. 
2. A **Server function** to handle reactive logic that returns an expression and state object.
3. A **Constructor function** wraps the UI and server to initialize the block state.

Now, let's go through each element, building up a block template as we go.

### UI function

A UI function in blockr is the same as a UI function in a Shiny module.
This means that:

- The UI function signature is expected to contain a single `id` argument, which can be used with `shiny::NS()` to construct namespaced IDs.
- A call to appropriate shiny UI functions is expected that return `shiny.tag` or `shiny.tag.list` objects, typically via the use of `shiny::tagList()` in the UI to list separate UI elements.

Let's start to build our block template:

```{r, eval = FALSE}
ui <- function(id) {
  tagList(
    # Wrap widgets in `tagList()`
    textInput(
      NS(id, "my_input") # Use `shiny::NS()` to construct namespaces
    )
  )
}
```

### Server function

As inputs, a server function takes an id and any additional inputs from other blocks (e.g., data). For example, for the blocks shipped with blockr.core, this would result in zero inputs for a __data__ block, one for a __transform__ block such as a select block, two for a __join__ block and the special argument `...args` for __variadic__ blocks, such as an `rbind` block.

```{mermaid}
flowchart TB
  data_blk[data block 1]
  data_blk_2[data block 2]
  data_blk_3[data block 3]
  data_blk_4[data block 4]
  select_blk[select block]
  join_blk[join block]
  rbind_blk[rbind block]
  data_blk --> |data| select_blk
  data_blk_2 -->|data1| join_blk
  data_blk_3 --> |data2| join_blk
  
  data_blk --> |1| rbind_blk
  select_blk --> |2| rbind_blk
  data_blk_4 --> |3| rbind_blk
  join_blk --> |4| rbind_blk
```

A server function should return as output a `moduleServer()` call, defining:

- **expr**: A quoted reactive expression representing the block’s computation. Data names should match between expression and the top-level server function arguments.
- **state**: A list of reactive values tracking user selections. The set of returned values from the __state__ should match (both in count and names) that of the constructor signature.

```{r eval = FALSE}
server <- function(id, data) {
  moduleServer(id, function(input, output, session) {
    # Reactive logic goes here

    # Return a list with "expr" and "state"
    list(
      expr = reactive(quote(identity(data))),
      state = list(
        input_one = reactive(input_one()),
        input_two = reactive(input_two())
      )
    )
  })
}
```

Many options exist to build expressions, such as using tools offered by the `rlang` package or by passing a string to `base::parse()`.
As a block developer you are free to choose the tools that your prefer, as long as the evaluated expression returns non-reactive R code which can be executed outside of blockr.

The state is needed for __serialization__ and __deserialization__ to respectively save and restore the state of a blockr application.

### Constructor function

The __constructor__ function wraps the UI and server functions and initializes the block. Note that both `server` and `ui` are [closures](https://adv-r.hadley.nz/environments.html#function-environments) and therefore may refer to names bound in the constructor scope (e.g. `dataset` and `choices`).

It should expose as arguments anything the user might set via the UI to control the block __state__.
In other words, any inputs in the UI function, should have a corresponding argument in the constructor function.
On the other hand, inputs from other blocks should not be exposed as arguments, as these are dynamically provided in the server function.

The return value should be a call to `new_block()` (or if applicable a call to the more specific *virtual* constructors `new_data_block()`, `new_transform_block()`, etc.).

We can finish our block template:

```{r eval = FALSE}
example_constructor <- function(ui_state = character(), ...) {
  ui <- function(id) {
    tagList(textInput(NS(id, "ui_state")))
  }

  server <- function(id, data) {
    moduleServer(id, function(input, output, session) {
      # Reactive logic goes here

      # Return a list with "expr" and "state"
      list(
        expr = reactive(quote(identity(data))),
        state = list(
          # name must match what is defined in the constructor signature
          ui_state = <STATE_VALUE>
        )
      )
    })
  }

  # Return call to `new_block()`
  new_block(
    server = server,
    ui = ui,
    class = "my_block",
    ...
  )
}
```

The `...` is forwarded to `new_block` or any `new_*_block`.

### Example

Putting this all together, we can create a `utils::head()` block, such as the one offered as `new_head_block()`:

```{r}
new_head_block <- function(n = 6L, ...) {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          n_rows <- reactiveVal(n)

          observeEvent(input$n, n_rows(input$n))

          observeEvent(
            nrow(data()),
            updateNumericInput(
              inputId = "n",
              value = n_rows(),
              min = 1L,
              max = nrow(data())
            )
          )

          list(
            expr = reactive(
              bquote(utils::head(data, n = .(n)), list(n = n_rows()))
            ),
            state = list(
              n = n_rows
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        numericInput(
          inputId = NS(id, "n"),
          label = "Number of rows",
          value = n,
          min = 1L
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data) || is.matrix(data))
    },
    class = "head_block",
    ...
  )
}
```

Here, the state return value is a list of length 1, containing the current value for the only constructor argument. The current value for `n` is represented by a `shiny::reactiveVal()`, `n_rows()`, which is initialized with the values `n` in the constructor scope and updated on every change to `input$n`. Furthermore the `max` value for the `shiny::numericInput()` field is updated on every change to the number of data rows.

A final `block` constructor argument of note might be `dat_val`, which is an optional function that can be passed which signals to the framework, if data passed to a block, can actually be processed by the block.

An app containing such a head block can be spun up as

```{r, eval = FALSE}
serve(new_head_block(n = 10L), list(data = mtcars))
```

For an example with multiple data inputs, refer to examples such as `new_merge_block()`. Such a binary block (with arguments `x` and `y`) can be explored in a standalone app (with nonsensical inputs) as

```{r, eval = FALSE}
serve(
  new_merge_block(by = "Time"),
  data = list(x = datasets::BOD, y = datasets::ChickWeight)
)
```

The `data` argument to `serve()` expects a list with names components that match the server function signature, i.e. `data` for `new_head_block()` and `x`, `y` for `new_merge_block()`. Such names can be chosen freely by the block implementer (with the exception of integer-valued names, such as `` `1` ``, `` `2` ``, etc. which are reserved as positional arguments in `...args`).


### More complex examples

The above example showed how to embed a typical R function into a block. However,
`blockr.core` can do much more. What about creating an entire block around an existing
complex Shiny module?

In the following, we expose how you could wrap a block around `esquisse`, a [package](https://dreamrs.github.io/esquisse/index.html) to build
ggplot without coding. `esquisse` exports standalone [modules](https://dreamrs.github.io/esquisse/articles/shiny-usage.html#use-esquisse-as-a-shiny-module) you can embed in any existing shiny app. Specifically, we consider `esquisse_server()` and `esquisse_ui()`. `esquisse_server()` needs some rectangular data as input and returns a list of 3 elements: modified data, code plot and filters code:

```r
results <- esquisse::esquisse_server(
  id = "esquisse",
  data_rv = data
)

results$data
results$code_filters
results$code_plot
```

When filters are applied with `esquisse`, the returned data are filtered, otherwise the module returns the unchanged input data. Based on what we already know about constructing block, we can write the following code for the constructor server function. This block will have 1 input materialized as the `data` parameter (data coming from the upstream block):

```{r esquisse-server, eval=FALSE}
esquisse_block_server <- function(id, data) {
  moduleServer(
    id,
    function(input, output, session) {
      results <- esquisse::esquisse_server(
        id = "esquisse",
        data_rv = data
      )
      list(
        expr = reactive({
          bquote(
            list(
              dat = as.data.frame(.(dat)),
              filters = .(filters)
            ),
            list(
              filters = results$code_plot,
              dat = results$data
            )
          )
        }),
        state = list()
      )
    }
  )
}
```

In the server module, we first call `esquisse_server()`. Then, the most technical part is to construct the returned value. We decide that the block is __stateless__, as in any case `esquisse` doesn't handle starting the module in a specific state, so that would be useless to save something. The __expression__ is composed of 2 elements, the data and plot code. By default, if we assume our `esquisse` block to be a `transform_block,` `blockr.core` uses `block_output.transform_block` to display the block output:

```r
block_output.transform_block <- function(x, result, session) {
  session$output$result <- dt_result(result$dat, session)
}
```

This S3 [method](https://adv-r.hadley.nz/s3.html#s3-methods) can only handle rectangular data and not a list as we specified in `esquisse_block_server()`. We therefore have to create a new class and constructor for our `esquisse` block, we call it `complex_block` and `new_complex_block()`, respectively:

```{r esquisse-complex-block, eval=FALSE}
new_complex_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "complex_block"), ctor, ...)
}
```

As you may notice, `new_complex_block()` is very similar to `new_transform_block`, the only difference being the class name. We subsequently define a new S3 method for `block_output()`, which is able to process our `list` result (don't forget the `@export` roxygen tag):

```{r esquisse-block-output, eval=FALSE}
#' @export
block_output.complex_block <- function(x, result, session) {
  session$output$filters <- renderPrint(result$filters)
  # result must come at the end of the output list if you have multiple outputs
  session$output$result <- dt_result(result$dat, session)
}
```

We also provide the UI counter part. Here you could go for more fancy layout but for sake of simplicity, we design a minimalistic UI:

```{r esquisse-block-ui, eval=FALSE}
#' @export
block_ui.complex_block <- function(id, x, ...) {
  tagList(
    h1("Transformed data from {esquisse}"),
    DT::dataTableOutput(NS(id, "result")),
    verbatimTextOutput(NS(id, "filters"))
  )
}
```
The UI function of our constructor is simple, even though you would be totally free to customize it further:

```{r esquisse-ui, eval=FALSE}
esquisse_block_ui <- function(id) {
  tagList(
    esquisse::esquisse_ui(
      id = NS(id, "esquisse"), 
      header = FALSE # dont display gadget title
    )
  )
}
```

Finally our new `esquisse` block constructor given by:

```{r esquisse-ctor, eval=FALSE}
new_esquisse_block <- function(...) {
  new_complex_block(
    server = esquisse_block_server,
    ui = esquisse_block_ui,
    class = "esquisse_block",
    dat_valid = NULL,
    allow_empty_state = TRUE,
    ...
  )
}
```

To test our new block we call:

```{r esquisse-app, eval=FALSE}
serve(
  new_board(
    blocks = list(
      a = new_dataset_block(iris),
      b = new_esquisse_block()
    ),
    links = list(
      new_link("a", "b", "data")
    )
  )
)
```

As closing remark, since the `esquisse` block is stateless, save and restore won't work for this block.

::: {.callout-note}
The demo below runs with shinylive. Not all feature may work as expected due to compatibility issues with webR.
:::

```{r shinylive_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(esquisse)",
    "library(plotly)",
    knitr::knit_code$get("esquisse-server"),
    knitr::knit_code$get("esquisse-ui"),
    knitr::knit_code$get("esquisse-complex-block"),
    knitr::knit_code$get("esquisse-ctor"),
    knitr::knit_code$get("esquisse-block-ui"),
    knitr::knit_code$get("esquisse-block-output"),
    knitr::knit_code$get("esquisse-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
```

```{r shinylive_iframe, echo = FALSE, eval = TRUE}
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)
```

---
title: "3. Browse for blocks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{3. Browse for blocks}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(blockr.core)
```

## Introduction

The __registry__ is a __environment__ which provides access to multiple __blocks__ as well as some
metadata:

- The block __description__.
- The block __category__.
- The block __package__.
- ...

In other words, the __registry__ is a __"supermarket"__ for data analysis. As shown below, if you develop your own blocks package and registers blocks on load, these blocks become available to the end user. Therefore this makes it powerful for __collaboration__ between data science teams.

```{mermaid}
flowchart LR
  subgraph blockr_ggplot2[blockr.ggplot2]
    new_block1[New block]
    new_block2[New block]
  end
  subgraph blockr_echarts4r[blockr.echarts4r]
    new_block3[New block]
    new_block4[New block]
  end
  blockr_ggplot2 --> |register| registry
  blockr_echarts4r --> |register| registry
  subgraph registry[Registry]
    subgraph select_reg[Select block]
      reg_name[Name: select block]
      reg_descr[Description: select columns in a table]
      reg_classes[Classes: select_block, tranform_block]
      reg_category[Category: transform]
      reg_ctor[Construcor: new_select_block]
      reg_package[Package: blockr.dplyr]
    end
    subgraph filter_reg[Filter block]
    end
    filter_reg --x |unregister| trash['fa:fa-trash']
  end
```

## Previewing available blocks

Upon loading, `{blockr}` __registers__ its internal __blocks__ with `register_blockr_blocks()`.
You won't have to call this function as it is not exported anyway. This makes the __registry__ environment
ready to be queried by `available_blocks()`. A truncated output example below:

```{r, eval=FALSE}
available_blocks()[["dataset_block"]]
```

```r
function(dataset = character(), package = "datasets", ...) {
  ...
}
<environment: namespace:blockr.core>
attr(,"name")
[1] "dataset block"
attr(,"description")
[1] "Choose a dataset from a package"
attr(,"classes")
[1] "dataset_block" "data_block"    "block"         "vctrs_vctr"    "list"         
attr(,"category")
[1] "data"
attr(,"ctor_name")
[1] "new_dataset_block"
attr(,"package")
[1] "blockr.core"
attr(,"class")
[1] "block_registry_entry"
```

```{r, eval=TRUE}
names(available_blocks())
```

## Register a block
To register your own blocks, user facing functions are:

- `register_block()` to register a block in the __registry__. If the __block__ is already registered,
it __overwrites__ the existing one.
- `register_blocks()` to register multiple blocks.

Note that, if you develop your block outside a package context, you must call `register_block` (or `register_blocks`)
passing the constructor as a function and not as a string (see below).

Let's say you want to create a new `new_dummy_block` which does nothing specific in the `./R/dummy-block.R` script:

```{r, eval=TRUE}
# ./R/dummy-block.R
new_dummy_block <- function(text = "Hello World", ...) {
  new_data_block(
    function(id) {
      moduleServer(id, function(input, output, session) {
        list(
          expr = reactive(quote(text)),
          state = list(text = text)
        )
      })
    },
    function(id) {
      shiny::tagList()
    },
    class = "dummy_block",
    ...
  )
}

register_dummy_blocks <- function() {
  register_blocks(
    c(new_dummy_block),
    name = c("dummy block"),
    description = c("A block that does nothing"),
    overwrite = TRUE
  )
}

register_dummy_blocks()
```

Finally, we create a `.R/zzz.R` script where you run this code to register the block(s) whenever the package loads:

```r
# ./R/zzz.R
.onLoad <- function(libname, pkgname) {
  register_dummy_blocks()
  invisible(NULL)
}
```

If we now query the registry, the new block is available:

```{r}
available_blocks()[["dummy_block"]]
```

## Unregister a block

The counterpart of `register_block()` is `unregister_blocks()`. We can remove
our new `dummy_block` from the registry:

```{r}
unregister_blocks(uid = "dummy_block")

# Check it out
names(available_blocks())
```

where __uid__ is/are the block(s) class(s) passed in the constructor.

---
title: "4. Extend blockr"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{4. Extend blockr}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r packages, message=FALSE, warning=FALSE}
library(shiny)
library(bslib)
library(scoutbaR)
library(blockr.core)
```

```{r custom-plugin-setup, eval = FALSE, echo=FALSE}
chr_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, character(length), ..., USE.NAMES = use_names)
}

#' @keywords internal
lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
}

dropNulls <- function(x) {
  x[!lgl_ply(x, is.null)]
}

blk_icon <- function(category) {
  switch(
    category,
    "data" = "table",
    "file" = "file-import",
    "parse" = "cogs",
    "plot" = "chart-line",
    "transform" = "wand-magic-sparkles",
    "table" = "table"
  )
}

blk_choices <- function() {
  blk_cats <- sort(
    unique(chr_ply(available_blocks(), \(b) attr(b, "category")))
  )

  lapply(blk_cats, \(cat) {
    scout_section(
      label = cat,
      .list = dropNulls(
        unname(
          lapply(available_blocks(), \(choice) {
            if (attr(choice, "category") == cat) {
              scout_action(
                id = attr(choice, "classes")[1],
                label = attr(choice, "name"),
                description = attr(choice, "description"),
                icon = blk_icon(cat)
              )
            }
          })
        )
      )
    )
  })
}
```

## Introduction

`blockr` can be extended in 3 main ways. The first way is to create new blocks, as described in the `Create block` [vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/create-block.html). The second way is to change the behavior at __board__ level through the __plugin__ system. Finally we can customize the UI library, as we will in the last section.

## Plugins

Most major features of a __board__ are implemented through what we call __plugins__. A board of a given class, such as `"board"` (the basic legacy board) or `"dock_board"` (the board used by `blockr::run_app()`) supports a given set of __plugins__, implemented using shiny modules. Users can override the ui and server functions of these modules from the outside if they wish to tweak the display and logic of these features, as we'll show further below.

### Available plugins

The legacy board supports the following plugins :

- `manage_blocks()`: Implements the way __blocks__ are created, removed, appened...
- `manage_links()`: Implements how __blocks__ are connected. Linking __block__ A to __block__ B means that __block__ A passes its output data to __block__ B.
- `manage_stacks()`: Implements how __blocks__ can be grouped using __stacks__.
- `preserve_board()`: Implements how __boards__ can be saved and restored.
- `generate_code()`: Implements the way the data manipulation code executed by the __blocks__ is displayed to the user.
- `notify_user()`: Implements how notifications and warnings are relayed to the user.
- `edit_block()`: Implements how a block attributes such as titles are processed and rendered.
- `edit_stack()`: Implements how a stack attributes such as names are processed and rendered.

### Override a plugin's server and ui functions

We will work from an example, we want to modify the `manage_blocks()` plugin to use the `scoutbaR` package to search for blocks to add.

By default `manage_blocks()` uses `manage_blocks_ui()` and `manage_blocks_server()`, a good workflow is to start from those and implement our modifications.

```r
# manage_blocks_ui
function(id, board) {
  tagList(
    actionButton(
      NS(id, "add_block"),
      "Add block",
      icon = icon("circle-plus"),
      class = "btn-success"
    ),
    actionButton(
      NS(id, "rm_block"),
      "Remove block",
      icon = icon("circle-minus"),
      class = "btn-danger"
    )
  )
}
```

For simplicity let's assume we also want to drop the "Remove block" button, our new ui becomes: 

```r
new_manage_blocks_ui <- function(id, board) {
  tagList(
    scoutbaR::scoutbar(
      NS(id, "scoutbar"),
      placeholder = "Search for a block",
      actions = blk_choices(),
      theme = "dark",
      showRecentSearch = TRUE
    ),
    actionButton(
      NS(id, "add_block"),
      "New block",
      icon = icon("circle-plus"),
    )
  )
}
```

`blk_choices()` is a list of scoutbar actions, the details are not relevant here but the definition is provided below for information.

<details>

```r
blk_choices <- function() {
  blk_cats <- sort(
    unique(chr_ply(available_blocks(), \(b) attr(b, "category")))
  )

  lapply(blk_cats, \(cat) {
    scout_section(
      label = cat,
      .list = dropNulls(
        unname(
          lapply(available_blocks(), \(choice) {
            if (attr(choice, "category") == cat) {
              scout_action(
                id = attr(choice, "classes")[1],
                label = attr(choice, "name"),
                description = attr(choice, "description"),
                icon = blk_icon(cat)
              )
            }
          })
        )
      )
    )
  })
}

blk_icon <- function(category) {
  switch(
    category,
    "data" = "table",
    "file" = "file-import",
    "parse" = "cogs",
    "plot" = "chart-line",
    "transform" = "wand-magic-sparkles",
    "table" = "table"
  )
}

chr_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, character(length), ..., USE.NAMES = use_names)
}

lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
}

dropNulls <- function(x) {
  x[!lgl_ply(x, is.null)]
}
```
</details>

In the same fashion we override the server function to handle the new logic.

We end up with the following server function, see additional explanations below.

```{r custom-plugin-server, eval=FALSE}
new_manage_blocks_server <- function(id, board, update, ...) {
  moduleServer(
    id,
    function(input, output, session) {
      # Trigger add block
      observeEvent(
        input$add_block,
        {
          update_scoutbar(
            session,
            "scoutbar",
            revealScoutbar = TRUE
          )
        }
      )

      observeEvent(input$scoutbar, {
        new_blk <- as_blocks(create_block(input$scoutbar))
        update(
          list(blocks = list(add = new_blk))
        )
      })

      NULL
    }
  )
}
```

The server function's __signature__ must start with the module id, `board` refers to internal reactive values (read-only), `update` is a reactive value
to send updates to the board module and `...` is used to recover parameters passed from the top level like `parent`. The plugin always returns `NULL`.

In the first observer we open the `scoutbaR` widget whenever the users clicks on the `Add block` button. We can achieve that by calling `update_scoutbar` passing `revealScoutbar = TRUE`.

In the second observer we listen to `input$scoutbar` which holds the name of the selected block, and use to create a "blocks" object with `create_block()` and `as_blocks()`. Finally, we signal this change to the board by refreshing the `update` reactive value, saying we want to add a new block `list(blocks = list(add = new_blk))`.

### Putting everything together

Once we have updated ui and server functions we can use the updated plugin by defining new main ui and server functions around them.

```{r custom-plugin-app, eval=FALSE}
#| code-fold: true
main_ui <- function(id, board) {
  ns <- NS(id)
  board_ui(
    ns("board"),
    board,
    plugins = custom_board_plugins(
      c(
        "manage_blocks"
      )
    )
  )
}

main_server <- function(id, board) {
  moduleServer(
    id,
    function(input, output, session) {
      ns <- session$ns

      # Board module
      board_server(
        "board",
        board,
        plugins = plugins(
          manage_blocks(server = new_manage_blocks_server, ui = new_manage_blocks_ui)
        ),
        callbacks = list()
      )
    }
  )
}

board <- new_board()

ui <- page_fluid(
  main_ui("app", board)
)

server <- function(input, output, session) {
  main_server("app", board)
}

shinyApp(ui, server)
```

::: {.callout-note}
The demo below runs with shinylive. Not all feature may work as expected due to compatibility issues with webR.
:::

```{r shinylive_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    knitr::knit_code$get("packages"),
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-plugin-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
```

```{r shinylive_iframe, echo = FALSE, eval = TRUE}
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)
```

## Custom UI components

If you'd like to use the board with another UI kit than `bslib` you can create a new method for `board_ui()`. For that,
you'll need a little bit of S3 [knowledge](https://adv-r.hadley.nz/s3.html).

The function signature should contain `id` (module namespace), `x` (board object), and `plugins` to use `blockr.core` plugins.
In the following, we leverage the brand new `shinyNextUI` to power the custom board UI:

```{r custom-board-ui, eval=FALSE}
board_ui.custom_board <- function(id, x, plugins = list(), ...) {
  plugins <- as_plugins(plugins)
  div(
    id = paste0(id, "_board"),
    board_ui(id, plugins[["manage_blocks"]], x),
    div(
      id =  paste0(id, "_blocks"),
      block_ui(id, x)
    )
  )
}
```

We have to customize the `block_ui` too. Overall, we leverage the `shinyNextUI::card` component to create the block layout:

```{r custom-block-ui, eval=FALSE}
get_block_registry <- function(x) {
  stopifnot(is_block(x))
  available_blocks()[[strsplit(attr(x, "ctor"), "new_")[[1]][2]]]
}

block_ui.custom_board <- function(id, x, blocks = NULL, ...) {
  block_card <- function(x, id, ns) {
    id <- paste0("block_", id)

    blk_info <- get_block_registry(x)

    div(
      class = "m-2",
      id = ns(id),
      shinyNextUI::card(
        variant = "bordered",
        shinyNextUI::card_header(
          className = "d-flex justify-content-between",
          icon(blk_icon(attr(blk_info, "category"))),
          sprintf(
            "Block: %s (id: %s)",
            attr(blk_info, "name"),
            gsub("block_", "", id)
          ),
          shinyNextUI::tooltip(
            icon("info-circle"),
            content = tagList(
              p(
              icon("lightbulb"),
              "How to use this block?",
              ),
              p(attr(blk_info, "description"), ".")
            )
          )
        ),
        shinyNextUI::divider(),
        shinyNextUI::card_body(
          expr_ui(ns(id), x),
          block_ui(ns(id), x)
        ),
        shinyNextUI::divider(),
        shinyNextUI::card_footer(
          sprintf(
            "Type: %s; Package: %s",
            attr(blk_info, "category"),
            attr(blk_info, "package")
          )
        )
      )
    )
  }

  stopifnot(is.character(id) && length(id) == 1L)

  if (is.null(blocks)) {
    blocks <- board_blocks(x)
  } else if (is.character(blocks)) {
    blocks <- board_blocks(x)[blocks]
  }

  stopifnot(is_blocks(blocks))

  tagList(
    Map(
      block_card,
      blocks,
      names(blocks),
      MoreArgs = list(ns = NS(id)),
      USE.NAMES = FALSE
    )
  )
}
```

Notice the use of few `blockr.core` helpers along the way:

- `board_blocks()` to extract and validate the blocks of a board.
- `is_blocks()` check whether an object correspond to a list of blocks.
- `get_block_registry()` to get the current block metadata from the [registry](https://bristolmyerssquibb.github.io/blockr.core/articles/blocks-registry.html).

`add_rm_block_ui()` now leverages `shinyNextUI::actionButton`:

```{r custom-plugin-ui-nextui, eval=FALSE}
add_rm_block_ui <- function(id, board) {
  tagList(
    scoutbar(
      NS(id, "scoutbar"),
      placeholder = "Search for a block",
      actions = blk_choices(),
      theme = "dark",
      showRecentSearch = TRUE
    ),
    shinyNextUI::actionButton(
      NS(id, "add_block"),
      "New block",
      icon = icon("circle-plus"),
    )
  )
}
```

Since `blockr.core` blocks utilizes `shiny`/`bslib` UI, you'd also have to rewrite the UI and/or server part whenever necessary. This [vignette](https://bristolmyerssquibb.github.io/blockr.core/articles/create-block.html) provides a starting point to authoring blocks.

As a final step, when you call `new_board()` don't forget to add it the `custom_board` class so that the custom S3 methods are invoked.

```{r custom-ui-app, eval=FALSE}
#| code-fold: true
board <- new_board(class = "custom_board")

ui <- nextui_page(
  board_ui(
    "board",
    board,
    plugins = custom_board_plugins(
      c(
        "manage_blocks"
      )
    )
  ) 
)

server <- function(input, output, session) {
  board_server(
    "board",
    board,
    plugins = custom_board_plugins(
      c(
        "manage_blocks"
      )
    ),
    callbacks = list()
  )
}

shinyApp(ui, server)
```

::: {.callout-note}
The demo below runs with shinylive. Not all feature may work as expected due to compatibility issues with webR.
:::

```{r shinylive2_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(shiny)",
    "library(scoutbaR)",
    "library(blockr.core)",
    "library(shinyNextUI)",
    knitr::knit_code$get("custom-plugin-setup"),
    knitr::knit_code$get("custom-plugin-ui-nextui"),
    knitr::knit_code$get("custom-plugin-server"),
    knitr::knit_code$get("custom-plugin-helpers"),
    knitr::knit_code$get("custom-block-ui"),
    knitr::knit_code$get("custom-board-ui"),
    knitr::knit_code$get("custom-ui-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
```

```{r shinylive2_iframe, echo = FALSE, eval = TRUE}
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)
```

---
title: "2. Create a block"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{2. Create a block}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(blockr.core)
```

## Creating blocks

**Note:** If you are not already familiar with [Shiny modules](https://mastering-shiny.org/scaling-modules.html), it is recommended you familiarize yourself with them before continuing with this section.

`blockr.core` is built on top of Shiny. At its heart, a __block__ is nothing more than a specialized Shiny __module__ that returns two additional values:

- **An expression** (called "expr"), a reactive expression defining a blocks computation. This allows R code to be exported and  data analyses to be recreated, outside of blockr and a reactive context.
- **A state object** (called "state"), a list of reactive values tracking user inputs. This allows blocks to separate user-defined state from inputs specified by blocks.

```{r, echo=FALSE}
include_mermaid("block-anatomy")
```

While blocks *return* the "expr" and "state" values, blocks *consist* of three elements:

1. A **UI function** to define the user interface. 
2. A **Server function** to handle reactive logic that returns an expression and state object.
3. A **Constructor function** wraps the UI and server to initialize the block state.

Now, let's go through each element, building up a block template as we go.

### UI function

A UI function in blockr is the same as a UI function in a Shiny module.
This means that:

- The UI function signature is expected to contain a single `id` argument, which can be used with `shiny::NS()` to construct namespaced IDs.
- A call to appropriate shiny UI functions is expected that return `shiny.tag` or `shiny.tag.list` objects, typically via the use of `shiny::tagList()` in the UI to list separate UI elements.

Let's start to build our block template:

```{r, eval = FALSE}
ui <- function(id) {
  tagList(
    # Wrap widgets in `tagList()`
    textInput(
      NS(id, "my_input") # Use `shiny::NS()` to construct namespaces
    )
  )
}
```

### Server function

As inputs, a server function takes an id and any additional inputs from other blocks (e.g., data). For example, for the blocks shipped with blockr.core, this would result in zero inputs for a __data__ block, one for a __transform__ block such as a select block, two for a __join__ block and the special argument `...args` for __variadic__ blocks, such as an `rbind` block.

```{r, echo=FALSE}
include_mermaid("block-inputs")
```

A server function should return as output a `moduleServer()` call, defining:

- **expr**: A quoted reactive expression representing the block’s computation. Data names should match between expression and the top-level server function arguments.
- **state**: A list of reactive values tracking user selections. The set of returned values from the __state__ should match (both in count and names) that of the constructor signature.

```{r eval = FALSE}
server <- function(id, data) {
  moduleServer(id, function(input, output, session) {
    # Reactive logic goes here

    # Return a list with "expr" and "state"
    list(
      expr = reactive(quote(identity(data))),
      state = list(
        input_one = reactive(input_one()),
        input_two = reactive(input_two())
      )
    )
  })
}
```

Many options exist to build expressions, such as using tools offered by the `rlang` package or by passing a string to `base::parse()`.
As a block developer you are free to choose the tools that your prefer, as long as the evaluated expression returns non-reactive R code which can be executed outside of blockr.

The state is needed for __serialization__ and __deserialization__ to respectively save and restore the state of a blockr application.

### Constructor function

The __constructor__ function wraps the UI and server functions and initializes the block. Note that both `server` and `ui` are [closures](https://adv-r.hadley.nz/environments.html#function-environments) and therefore may refer to names bound in the constructor scope (e.g. `dataset` and `choices`).

It should expose as arguments anything the user might set via the UI to control the block __state__.
In other words, any inputs in the UI function, should have a corresponding argument in the constructor function.
On the other hand, inputs from other blocks should not be exposed as arguments, as these are dynamically provided in the server function.

The return value should be a call to `new_block()` (or if applicable a call to the more specific *virtual* constructors `new_data_block()`, `new_transform_block()`, etc.).

We can finish our block template:

```{r eval = FALSE}
example_constructor <- function(ui_state = character(), ...) {
  ui <- function(id) {
    tagList(textInput(NS(id, "ui_state")))
  }

  server <- function(id, data) {
    moduleServer(id, function(input, output, session) {
      # Reactive logic goes here

      # Return a list with "expr" and "state"
      list(
        expr = reactive(quote(identity(data))),
        state = list(
          # name must match what is defined in the constructor signature
          ui_state = <STATE_VALUE>
        )
      )
    })
  }

  # Return call to `new_block()`
  new_block(
    server = server,
    ui = ui,
    class = "my_block",
    ...
  )
}
```

The `...` is forwarded to `new_block` or any `new_*_block`.

### Example

Putting this all together, we can create a `utils::head()` block, such as the one offered as `new_head_block()`:

```{r}
new_head_block <- function(n = 6L, ...) {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          n_rows <- reactiveVal(n)

          observeEvent(input$n, n_rows(input$n))

          observeEvent(
            nrow(data()),
            updateNumericInput(
              inputId = "n",
              value = n_rows(),
              min = 1L,
              max = nrow(data())
            )
          )

          list(
            expr = reactive(
              bquote(utils::head(data, n = .(n)), list(n = n_rows()))
            ),
            state = list(
              n = n_rows
            )
          )
        }
      )
    },
    function(id) {
      tagList(
        numericInput(
          inputId = NS(id, "n"),
          label = "Number of rows",
          value = n,
          min = 1L
        )
      )
    },
    dat_val = function(data) {
      stopifnot(is.data.frame(data) || is.matrix(data))
    },
    class = "head_block",
    ...
  )
}
```

Here, the state return value is a list of length 1, containing the current value for the only constructor argument. The current value for `n` is represented by a `shiny::reactiveVal()`, `n_rows()`, which is initialized with the values `n` in the constructor scope and updated on every change to `input$n`. Furthermore the `max` value for the `shiny::numericInput()` field is updated on every change to the number of data rows.

A final `block` constructor argument of note might be `dat_val`, which is an optional function that can be passed which signals to the framework, if data passed to a block, can actually be processed by the block.

An app containing such a head block can be spun up as

```{r, eval = FALSE}
serve(new_head_block(n = 10L), list(data = mtcars))
```

For an example with multiple data inputs, refer to examples such as `new_merge_block()`. Such a binary block (with arguments `x` and `y`) can be explored in a standalone app (with nonsensical inputs) as

```{r, eval = FALSE}
serve(
  new_merge_block(by = "Time"),
  data = list(x = datasets::BOD, y = datasets::ChickWeight)
)
```

The `data` argument to `serve()` expects a list with names components that match the server function signature, i.e. `data` for `new_head_block()` and `x`, `y` for `new_merge_block()`. Such names can be chosen freely by the block implementer (with the exception of integer-valued names, such as `` `1` ``, `` `2` ``, etc. which are reserved as positional arguments in `...args`).


### More complex examples

The above example showed how to embed a typical R function into a block. However,
`blockr.core` can do much more. What about creating an entire block around an existing
complex Shiny module?

In the following, we expose how you could wrap a block around `esquisse`, a [package](https://dreamrs.github.io/esquisse/index.html) to build
ggplot without coding. `esquisse` exports standalone [modules](https://dreamrs.github.io/esquisse/articles/shiny-usage.html#use-esquisse-as-a-shiny-module) you can embed in any existing shiny app. Specifically, we consider `esquisse_server()` and `esquisse_ui()`. `esquisse_server()` needs some rectangular data as input and returns a list of 3 elements: modified data, code plot and filters code:

```r
results <- esquisse::esquisse_server(
  id = "esquisse",
  data_rv = data
)

results$data
results$code_filters
results$code_plot
```

When filters are applied with `esquisse`, the returned data are filtered, otherwise the module returns the unchanged input data. Based on what we already know about constructing block, we can write the following code for the constructor server function. This block will have 1 input materialized as the `data` parameter (data coming from the upstream block):

```{r esquisse-server, eval=FALSE}
esquisse_block_server <- function(id, data) {
  moduleServer(
    id,
    function(input, output, session) {
      results <- esquisse::esquisse_server(
        id = "esquisse",
        data_rv = data
      )
      list(
        expr = reactive({
          bquote(
            list(
              dat = as.data.frame(.(dat)),
              filters = .(filters)
            ),
            list(
              filters = results$code_plot,
              dat = results$data
            )
          )
        }),
        state = list()
      )
    }
  )
}
```

In the server module, we first call `esquisse_server()`. Then, the most technical part is to construct the returned value. We decide that the block is __stateless__, as in any case `esquisse` doesn't handle starting the module in a specific state, so that would be useless to save something. The __expression__ is composed of 2 elements, the data and plot code. By default, if we assume our `esquisse` block to be a `transform_block,` `blockr.core` uses `block_output.transform_block` to display the block output:

```r
block_output.transform_block <- function(x, result, session) {
  dt_result(result, x, session)
}
```

This S3 [method](https://adv-r.hadley.nz/s3.html#s3-methods) can only handle rectangular data and not a list as we specified in `esquisse_block_server()`. We therefore have to create a new class and constructor for our `esquisse` block, we call it `complex_block` and `new_complex_block()`, respectively:

```{r esquisse-complex-block, eval=FALSE}
new_complex_block <- function(server, ui, class, ctor = sys.parent(), ...) {
  new_block(server, ui, c(class, "complex_block"), ctor, ...)
}
```

As you may notice, `new_complex_block()` is very similar to `new_transform_block`, the only difference being the class name. We subsequently define a new S3 method for `block_output()`, which is able to process our `list` result (don't forget the `@export` roxygen tag):

```{r esquisse-block-output, eval=FALSE}
#' @export
block_output.complex_block <- function(x, result, session) {
  session$output$filters <- renderPrint(result$filters)
  dt_result(result$dat, x, session)
}
```

We also provide the UI counter part. Here you could go for more fancy layout but for sake of simplicity, we design a minimalistic UI:

```{r esquisse-block-ui, eval=FALSE}
#' @export
block_ui.complex_block <- function(id, x, ...) {
  tagList(
    h1("Transformed data from {esquisse}"),
    DT::dataTableOutput(NS(id, "result")),
    verbatimTextOutput(NS(id, "filters"))
  )
}
```
The UI function of our constructor is simple, even though you would be totally free to customize it further:

```{r esquisse-ui, eval=FALSE}
esquisse_block_ui <- function(id) {
  tagList(
    esquisse::esquisse_ui(
      id = NS(id, "esquisse"), 
      header = FALSE # dont display gadget title
    )
  )
}
```

Finally our new `esquisse` block constructor given by:

```{r esquisse-ctor, eval=FALSE}
new_esquisse_block <- function(...) {
  new_complex_block(
    server = esquisse_block_server,
    ui = esquisse_block_ui,
    class = "esquisse_block",
    dat_valid = NULL,
    allow_empty_state = TRUE,
    ...
  )
}
```

To test our new block we call:

```{r esquisse-app, eval=FALSE}
serve(
  new_board(
    blocks = list(
      a = new_dataset_block(iris),
      b = new_esquisse_block()
    ),
    links = list(
      new_link("a", "b", "data")
    )
  )
)
```

As closing remark, since the `esquisse` block is stateless, save and restore won't work for this block.

::: {.callout-note}
The demo below runs with shinylive. Not all feature may work as expected due to compatibility issues with webR.
:::

```{r shinylive_url, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"blockr.core\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    "library(blockr.core)",
    "library(esquisse)",
    "library(plotly)",
    knitr::knit_code$get("esquisse-server"),
    knitr::knit_code$get("esquisse-ui"),
    knitr::knit_code$get("esquisse-complex-block"),
    knitr::knit_code$get("esquisse-ctor"),
    knitr::knit_code$get("esquisse-block-ui"),
    knitr::knit_code$get("esquisse-block-output"),
    knitr::knit_code$get("esquisse-app")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
```

```{r shinylive_iframe, echo = FALSE, eval = TRUE}
shiny::tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)
```

# opt utils

    Code
      opt
    Output
      test: test

# llm model opt

    Code
      opt1
    Output
      llm_model: Default chat function

---

    Code
      opt2
    Output
      llm_model: chat1 (chat1, chat2)

# board constructor

    Code
      print(board)
    Output
      <board>
      
      Blocks[5]:
      
      d<merge_block<transform_block<block>>>
      Name: "Merge"
      Data inputs: "x" and "y"
      Initial block state:
       $ by   : chr(0)
       $ all_x: logi FALSE
       $ all_y: logi FALSE
      Constructor: blockr.core::new_merge_block()
      
      a<dataset_block<data_block<block>>>
      Name: "Dataset"
      No data inputs
      Initial block state:
       $ dataset: chr(0)
       $ package: chr "datasets"
      Constructor: blockr.core::new_dataset_block()
      
      c<subset_block<transform_block<block>>>
      Name: "Subset"
      Data inputs: "data"
      Initial block state:
       $ subset: chr ""
       $ select: chr ""
      Constructor: blockr.core::new_subset_block()
      
      e<subset_block<transform_block<block>>>
      Name: "Subset"
      Data inputs: "data"
      Initial block state:
       $ subset: chr ""
       $ select: chr ""
      Constructor: blockr.core::new_subset_block()
      
      b<dataset_block<data_block<block>>>
      Name: "Dataset"
      No data inputs
      Initial block state:
       $ dataset: chr(0)
       $ package: chr "datasets"
      Constructor: blockr.core::new_dataset_block()
      
      Links[4]:
      
      ad: a -> d (x)
      cd: c -> d (y)
      bc: b -> c (data)
      de: d -> e (data)
      
      Stacks[1]:
      
      bc<stack[2]>
      Name: "Stack"
      Blocks: "b" and "c"
      
      Options:
      <board_options[4]>
        board_name: NULL
        show_conditions: warning, error
        thematic: NULL
        dark_mode: NULL

---

    Code
      print(rm_blocks(board, "e"))
    Output
      <board>
      
      Blocks[4]:
      
      d<merge_block<transform_block<block>>>
      Name: "Merge"
      Data inputs: "x" and "y"
      Initial block state:
       $ by   : chr(0)
       $ all_x: logi FALSE
       $ all_y: logi FALSE
      Constructor: blockr.core::new_merge_block()
      
      a<dataset_block<data_block<block>>>
      Name: "Dataset"
      No data inputs
      Initial block state:
       $ dataset: chr(0)
       $ package: chr "datasets"
      Constructor: blockr.core::new_dataset_block()
      
      c<subset_block<transform_block<block>>>
      Name: "Subset"
      Data inputs: "data"
      Initial block state:
       $ subset: chr ""
       $ select: chr ""
      Constructor: blockr.core::new_subset_block()
      
      b<dataset_block<data_block<block>>>
      Name: "Dataset"
      No data inputs
      Initial block state:
       $ dataset: chr(0)
       $ package: chr "datasets"
      Constructor: blockr.core::new_dataset_block()
      
      Links[3]:
      
      ad: a -> d (x)
      cd: c -> d (y)
      bc: b -> c (data)
      
      Stacks[1]:
      
      bc<stack[2]>
      Name: "Stack"
      Blocks: "b" and "c"
      
      Options:
      <board_options[4]>
        board_name: NULL
        show_conditions: warning, error
        thematic: NULL
        dark_mode: NULL

# stacks class

    Code
      print(stk)
    Output
      <stacks[2]>
      
      a
       <stack[5]>
       Name: "stack a"
       Blocks: "a", "b", "c", "d" and "e"
      
      b
       <stack[3]>
       Name: "stack b"
       Blocks: "f", "g" and "h"

# blocks utils

    Code
      print(ab)
    Output
      <blocks[2]>
      
      a
       <dataset_block<data_block<block>>>
       Name: "Dataset"
       No data inputs
       Initial block state:
        $ dataset: chr(0)
        $ package: chr "datasets"
       Constructor: blockr.core::new_dataset_block()
      
      b
       <subset_block<transform_block<block>>>
       Name: "Subset"
       Data inputs: "data"
       Initial block state:
        $ subset: chr ""
        $ select: chr ""
       Constructor: blockr.core::new_subset_block()

# plugins

    Code
      print(a)
    Output
      <preserve_board<plugin>>

---

    Code
      print(cd)
    Output
      <plugins[2]>
      manage_links: <manage_links<plugin>>
      manage_stacks: <manage_stacks<plugin>>

# block class

    Code
      print(x)
    Output
      <dataset_block<data_block<block>>>
      Name: "Dataset"
      No data inputs
      Initial block state:
       $ dataset: chr(0)
       $ package: chr "datasets"
      Constructor: blockr.core::new_dataset_block()

# Without package blocks can print

    Code
      blk()
    Output
      <dummy_block<data_block<block>>>
      Name: "Dummy"
      No data inputs
      Initial block state:
       $ text: chr "Hello World"
      Constructor: <local function>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-server.R
\name{board_server}
\alias{board_server}
\alias{board_server.board}
\title{Board server}
\usage{
board_server(id, x, ...)

\method{board_server}{board}(
  id,
  x,
  plugins = board_plugins(x),
  options = board_options(x),
  callbacks = list(),
  callback_location = c("end", "start"),
  ...
)
}
\arguments{
\item{id}{Parent namespace}

\item{x}{Board}

\item{...}{Generic consistency}

\item{plugins}{Board plugins as modules}

\item{options}{Board options (\code{NULL} defaults to the union of board, block
and registry sourced options)}

\item{callbacks}{Single (or list of) callback function(s), called only
for their side-effects)}

\item{callback_location}{Location of callback invocation (before or after
plugins)}
}
\value{
A \code{board_server()} implementation (such as the default for the
\code{board} base class) is expected to return a \code{\link[shiny:moduleServer]{shiny::moduleServer()}}.
}
\description{
A call to \code{board_server()}, dispatched on objects inheriting from \code{board},
returns a \code{\link[shiny:moduleServer]{shiny::moduleServer()}}, containing all necessary logic to
manipulate board components via UI. Extensibility over currently available
functionality is provided in the form of S3, where a \code{board_server()}
implementation of \code{board} sub-classes may be provided, as well as via a
plugin architecture and callback functions which can be used to register
additional observers.
}
\section{Active conditions}{

Conditions raised while blocks evaluate (errors, warnings and messages) are
exposed as a reactive data frame \code{board$conditions} on the read-only board
handed to plugins and callbacks, with one row per active condition and
columns \code{block}, \code{phase}, \code{severity}, \code{message} and \code{id}. It combines the
per-block \code{server$conditions} reactives (see \code{\link[=block_server]{block_server()}}), so a
consumer reads a single reactive — the whole board, or one block's frame
for fine-grained updates — rather than walking nested condition state. The
default \code{\link[=notify_user]{notify_user()}} plugin renders its toasts from this source.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-stacks.R
\name{manage_stacks}
\alias{manage_stacks}
\alias{manage_stacks_server}
\alias{manage_stacks_ui}
\title{Plugin module for managing board stacks}
\usage{
manage_stacks(server = manage_stacks_server, ui = manage_stacks_ui)

manage_stacks_server(id, board, update, ...)

manage_stacks_ui(id, board)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{The initial \code{board} object}

\item{update}{Reactive value object to initiate board updates}

\item{...}{Extra arguments passed from parent scope}
}
\value{
A plugin container inheriting from \code{manage_stacks} is returned by
\code{manage_stacks()}, while the UI component (e.g. \code{manage_stacks_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{manage_stacks_server()}) is expected to return \code{NULL}.
}
\description{
Logic and user experience for adding new, removing and modifying existing
stacks to/from the board can be customized or enhanced by providing an
alternate version of this plugin. The default implementation provides a
table-based UI, presented in a modal.
}
\details{
Updates are mediated via the \code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} object passed as
\code{update}, where stack updates are communicated as list entry \code{stacks} with
components \code{add}, \code{rm} or \code{mod}, where
\itemize{
\item \code{add} is either \code{NULL} or a \code{stacks} object (stack IDs may not already
exists),
\item \code{rm} is either \code{NULL} or a character vector of (existing) stack IDs,
\item \code{mod} is either \code{NULL} or a \code{stacks} object (where stack IDs must already
exist).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-eval.R, R/block-server.R
\name{block_eval}
\alias{block_eval}
\alias{eval_env}
\alias{block_eval_trigger}
\alias{block_server}
\alias{block_server.block}
\alias{expr_server}
\alias{block_render_trigger}
\title{Block server}
\usage{
block_eval(x, expr, env, ...)

eval_env(data)

block_eval_trigger(x, session = get_session())

block_server(id, x, data = list(), ...)

\method{block_server}{block}(
  id,
  x,
  data = list(),
  block_id = id,
  edit_block = NULL,
  ctrl_block = NULL,
  board = reactiveValues(),
  update = reactiveVal(),
  inputs_ready = reactive(TRUE),
  ...
)

expr_server(x, data, ...)

block_render_trigger(x, session = get_session())
}
\arguments{
\item{x}{Object for which to generate a \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{expr}{Quoted expression to evaluate in the context of \code{data}}

\item{env}{Environment in which to evaluate \code{expr}}

\item{...}{Generic consistency}

\item{data}{Input data (list of reactives)}

\item{session}{Shiny session object}

\item{id}{Namespace ID}

\item{block_id}{Block ID}

\item{edit_block, ctrl_block}{Block plugins}

\item{board}{Reactive values object containing board information}

\item{update}{Reactive value object to initiate board updates}

\item{inputs_ready}{Reactive flag signaling whether the block's required
inputs are all connected to ready upstream blocks (supplied by
\code{\link[=board_server]{board_server()}}; defaults to always-ready when a block server is run
standalone)}
}
\value{
Both \code{block_server()} and \code{expr_server()} return shiny server module
(i.e. a call to \code{\link[shiny:moduleServer]{shiny::moduleServer()}}), while \code{block_eval()} evaluates
an interpolated (w.r.t. block "user" inputs) block expression in the context
of block data inputs.
}
\description{
A block is represented by several (nested) shiny modules and the top level
module is created using the \code{block_server()} generic. S3 dispatch is offered
as a way to add flexibility, but in most cases the default method for the
\code{block} class should suffice at top level. Further entry points for
customization are offered by the generics \code{expr_server()} and \code{block_eval()},
which are responsible for initializing the block "expression" module (i.e.
the block server function passed in \code{\link[=new_block]{new_block()}}) and block evaluation
(evaluating the interpolated expression in the context of input data),
respectively.
}
\details{
The module returned from \code{block_server()}, at least in the default
implementation, provides much of the essential but block-type agnostic
functionality, including data input validation (if available), instantiation
of the block expression server (handling the block-specific functionality,
i.e. block user inputs and expression), and instantiation of the
\code{edit_block} module (if passed from the parent scope).

Each block carries an \emph{eval status} -- one of \code{dormant}, \code{waiting}, \code{unset},
\code{failed} or \code{ready} -- which, together with the orthogonal \code{visible} flag,
determines its behaviour. The status separates the two input kinds (data
inputs from links, user inputs from \code{state}) and a genuine failure:
\itemize{
\item \code{dormant} -- not \emph{needed} (neither on screen nor feeding, transitively over
\code{\link[=board_links]{board_links()}}, an on-screen block); inputs stay unfulfilled
(\code{\link[shiny:req]{shiny::req()}} out) and nothing evaluates.
\item \code{waiting} -- needed, but a required \emph{data} input is missing: unconnected,
below the required number of variadic \code{...args} inputs (one by default),
or fed by an upstream block that is not itself \code{ready} (see
\code{allow_empty_state}).
\item \code{unset} -- data inputs are ready, but a required \emph{user} input (\code{state}
value) has not been provided (unless permitted by \code{allow_empty_state}).
\item \code{failed} -- all inputs are present, but the block cannot produce a result:
the data validator (\code{\link[=validate_data_inputs]{validate_data_inputs()}}) or the block expression
raised. The offending condition is surfaced through the block conditions.
\item \code{ready} -- evaluation succeeded and a result (possibly a legitimate \code{NULL})
is available for downstream blocks to consume.
}

A block reaches \code{ready} only once its upstreams have, so an unconnected or
pending block holds its whole downstream chain \code{waiting} without any of them
evaluating against missing data. Output rendering follows the status: the
block output is shown only while \code{ready} and cleared otherwise, so a block
leaving \code{ready} never displays a stale result. While not \code{ready} the block
surfaces a condition explaining why -- a \code{status}-phase note for \code{waiting}
and \code{unset}, or the raised error for \code{failed}. Conditions raised during
validation and evaluation are caught and returned to be surfaced to the app
user.

Block-level user inputs (provided by the expression module) are separated
from output, the behavior of which can be customized via the
\code{\link[=block_output]{block_output()}} generic. The \code{\link[=block_ui]{block_ui()}} generic can then be used to
control rendering of outputs.

When a front-end (such as blockr.dock) drives the \code{visible} write-channel
that \code{\link[=board_server]{board_server()}} hands to the board callback, reporting each block's
visibility status (off screen, on screen, or rendered into its view),
evaluation and rendering are gated on visibility.
Rendering is gated on plain visibility: the render observer is suspended
while a block is off screen and resumed once it is on screen, starting
suspended so nothing renders before the front-end first reports. Evaluation
is gated on the \emph{needed} set, the on-screen blocks together with their
upstream closure over \code{\link[=board_links]{board_links()}} (derived from \code{visible}, recomputed
only when it or the links change). A block's input data reactives stay
unfulfilled (they \code{\link[shiny:req]{shiny::req()}} out) unless the block is needed, so a block
that is neither visible nor feeding a visible block pulls no input and stays
fully quiescent: its result reactive, and any observer its expression server
registers on the incoming data, all short-circuit and do nothing. A needed
but off-screen block (one feeding a visible block) evaluates but does not
render. Block-server \emph{construction} is prioritized the same way: the needed
set is instantiated first so that first paint waits only for the on-screen
blocks and their upstreams, and the remaining block servers are built
progressively in the background. That background pass holds until the
front-end reports every on-screen block as rendered (arranged into its
view), so it never competes with first paint. Until a block is built it is
absent from
the \code{board$blocks} handed to plugins and callbacks, which simply see it
appear once constructed. The background cadence is set by the
\code{background_construction_delay} \code{\link[=blockr_option]{blockr_option()}} (milliseconds between
successive blocks, default 50); a value of 0 disables the staggering and
builds every block up front. With nothing driving
\code{visible} every block is needed and behaviour is unchanged; the
\code{gate_visibility} \code{\link[=blockr_option]{blockr_option()}} (default \code{TRUE}) turns gating off
entirely.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-code.R
\name{export_code}
\alias{export_code}
\title{Utilities for code export}
\usage{
export_code(expressions, board)
}
\arguments{
\item{expressions}{Block expressions}

\item{board}{Board object}
}
\value{
String containing properly arranged block expressions.
}
\description{
To facilitate other means of code export than implemented by the default
\code{\link[=generate_code]{generate_code()}} plugin, this utility performs much of the heavy lifting
to properly arrange and scope block-level expressions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/link-class.R, R/links-class.R
\name{new_link}
\alias{new_link}
\alias{update_link}
\alias{is_link}
\alias{as_link}
\alias{links}
\alias{is_links}
\alias{as_links}
\alias{validate_links}
\title{Board links}
\usage{
new_link(
  from = "",
  to = "",
  input = "",
  ...,
  ctor = "new_link",
  pkg = pkg_name(),
  class = character()
)

update_link(x, delta)

is_link(x)

as_link(x)

links(...)

is_links(x)

as_links(x, ...)

validate_links(x)
}
\arguments{
\item{from, to}{Block ID(s)}

\item{input}{Block argument}

\item{...}{Extensibility}

\item{ctor, pkg}{Constructor information (used for serialization)}

\item{class}{(Optional) link sub-class}

\item{x}{Links object}

\item{delta}{A named list of constructor argument values to apply on
top of \code{x}'s current fields.}
}
\value{
Both \code{new_link()}/\code{as_link()}, and \code{links()}/\code{as_links()} return
\code{link} and \code{links} objects, respectively. Testing for inheritance is
available as \code{is_link()}/\code{is_links()} and validation (for \code{links}) is
performed with \code{validate_links()}, which returns its input (\code{x}) or throws
an error.
}
\description{
Two blocks can be connected via a (directed) link. This means the result from
one block is passed as (data) input to the next. Source and destination are
identified by \code{from} and \code{to} attributes and in case of polyadic receiving
blocks, the \code{input} attribute identified which of the data inputs is the
intended destination. In principle, the \code{link} object may be extended via
sub-classing and passing further attributes, but this has not been properly
tested so far.

In addition to unique IDs, \code{links} objects are guaranteed to be consistent
in that it is not possible to have multiple links pointing to the same
target (combination of \code{to} and \code{input} attributes). Furthermore, links
behave like edges in a directed acyclic graph (DAG) in that cycles are
detected and disallowed.
}
\details{
A links is created via the \code{new_link()} constructor and for a vector of
links, the container object \code{links} is provided and a corresponding
constructor \code{links()} exported from the package. Testing whether an object
inherits from \code{link} (or \code{links}) is available via \code{is_link()} (or
\code{is_links()}, respectively). Coercion to \code{link} (and \code{links}) objects is
implemented as \code{as_link()} (and \code{as_links()}, respectively). Finally, links
can be validated by calling \code{validate_links()}.
}
\section{Partial-arg updates}{

\code{update_link()} is an S3 generic that produces a modified link by
merging a named-list \code{delta} of constructor argument values onto an
existing link. The default \code{.link} method passes the current
\code{as.list(x)} (the link's \code{from}, \code{to}, \code{input} plus any extra
\emph{list-element} fields a sub-class adds) merged with the delta to the
stored constructor. Sub-class extensions stored as attributes (rather
than list elements) are \strong{not} preserved through the default
reconstruction — sub-class owners must register a class-specific
method when their constructor uses attributes or a non-standard
signature.
}

\examples{
lnks <- links(from = c("a", "b"), to = c("b", "c"), input = c("x", "y"))
is_links(lnks)
names(lnks)

tryCatch(
  c(lnks, new_link("a", "b", "x")),
  error = function(e) conditionMessage(e)
)
tryCatch(
  c(lnks, new_link("b", "a")),
  error = function(e) conditionMessage(e)
)

lnks <- links(a = new_link("a", "b"), b = new_link("b", "c"))
names(lnks)

tryCatch(
  c(lnks, a = new_link("a", "b")),
  error = function(e) conditionMessage(e)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-class.R
\name{board_blocks}
\alias{board_blocks}
\alias{board_blocks<-}
\alias{board_block_ids}
\alias{rm_blocks}
\alias{board_links}
\alias{board_links<-}
\alias{board_link_ids}
\alias{modify_board_links}
\alias{board_stacks}
\alias{board_stacks<-}
\alias{board_stack_ids}
\alias{modify_board_stacks}
\alias{board_options}
\alias{board_options<-}
\alias{board_option_ids}
\alias{available_stack_blocks}
\alias{clear_board}
\title{Board utils}
\usage{
board_blocks(x)

board_blocks(x) <- value

board_block_ids(x)

rm_blocks(x, rm, ..., session = get_session())

board_links(x)

board_links(x) <- value

board_link_ids(x)

modify_board_links(
  x,
  add = NULL,
  rm = NULL,
  mod = NULL,
  ...,
  session = get_session()
)

board_stacks(x)

board_stacks(x) <- value

board_stack_ids(x)

modify_board_stacks(
  x,
  add = NULL,
  rm = NULL,
  mod = NULL,
  ...,
  session = get_session()
)

board_options(x)

board_options(x) <- value

board_option_ids(x)

available_stack_blocks(
  x,
  stacks = board_stacks(x),
  blocks = board_stack_ids(x)
)

clear_board(x)
}
\arguments{
\item{x}{Board}

\item{value}{Replacement value}

\item{rm}{Block/link/stack IDs to remove}

\item{...}{Further arguments they may be passed from the board server context}

\item{session}{Shiny session object}

\item{add}{Links/stacks to add}

\item{mod}{Link/stacks to modify}

\item{blocks, stacks}{Sets of blocks/stacks}
}
\value{
Functions for retrieving, as well as updating components
(\code{board_blocks()}/\code{board_links()}/\code{board_stacks()}/\code{board_options()} and
\verb{board_blocks<-()}/\verb{board_links<-()}/\verb{board_stacks<-()}/\verb{board_options<-()})
return corresponding objects (i.e. \code{blocks}, \code{links}, \code{stacks} and
\code{board_options}), while ID getters (\code{board_block_ids()}, \code{board_link_ids()},
\code{board_stack_ids()} and \code{board_option_ids()}) return character vectors, as
does \code{available_stack_blocks()}. Convenience functions \code{rm_blocks()},
\code{modify_board_links()} and \code{modify_board_stacks()} return an updated \code{board}
object.
}
\description{
A set of utility functions is available for querying and manipulating board
components (i.e. blocks, links and stacks). Functions for retrieving and
modifying board options are documented in \code{\link[=new_board_options]{new_board_options()}}.
}
\section{Blocks}{

Board blocks can be retrieved using \code{board_blocks()} and updated with the
corresponding replacement function \verb{board_blocks<-()}. If just the current
board IDs are of interest, \code{board_block_ids()} is available as short for
\code{names(board_blocks(x))}. In order to remove block(s) from a board, the
(generic) convenience function \code{rm_blocks()} is exported, which takes care
(in the default implementation for \code{board}) of also updating links and
stacks accordingly. The more basic replacement function \verb{board_blocks<-()}
might fail at validation of the updated board object if an inconsistent
state results from an update (e.g. a block referenced by a stack is no
longer available).
}

\section{Links}{

Board links can be retrieved using \code{board_links()} and updated with the
corresponding replacement function \verb{board_links<-()}. If only links IDs are
of interest, this is available as \code{board_link_ids()}, which is short for
\code{names(board_links(x))}. A (generic) convenience function for all kinds of
updates to board links in one is available as \code{modify_board_links()}. With
arguments \code{add}, \code{rm} and \code{mod}, links can be added, removed or modified in
one go.
}

\section{Stacks}{

Board stacks can be retrieved using \code{board_stacks()} and updated with the
corresponding replacement function \verb{board_stacks<-()}. If only the stack IDs
are of interest, this is available as \code{board_stack_ids()}, which is short
for \code{names(board_stacks(x))}. A (generic) convenience function to update
stacks is available as \code{modify_board_stacks()}, which can add, remove and
modify stacks depending on arguments passed as \code{add}, \code{rm} and \code{mod}. If
block IDs that are not already associated with a stack (i.e. "free" blocks)
are of interest, this is available as \code{available_stack_blocks()}.
}

\section{Options}{

Board options can be retrieved using \code{board_options()} and updated with the
corresponding replacement function \verb{board_options<-()}.  If only the option
IDs are of interest, this is available as \code{board_option_ids()}, which calls
\code{\link[=board_option_id]{board_option_id()}} on each board option.
}

\examples{
brd <- new_board(
  c(
     a = new_dataset_block(),
     b = new_subset_block()
  ),
  list(from = "a", to = "b")
)

board_blocks(brd)
board_block_ids(brd)

board_links(brd)
board_link_ids(brd)

board_stacks(brd)
board_stack_ids(brd)

board_options(brd)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-blocks.R
\name{manage_blocks}
\alias{manage_blocks}
\alias{manage_blocks_server}
\alias{manage_blocks_ui}
\title{Plugin module for managing board blocks}
\usage{
manage_blocks(server = manage_blocks_server, ui = manage_blocks_ui)

manage_blocks_server(id, board, update, ...)

manage_blocks_ui(id, board)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{The initial \code{board} object}

\item{update}{Reactive value object to initiate board updates}

\item{...}{Extra arguments passed from parent scope}
}
\value{
A plugin container inheriting from \code{manage_blocks} is returned by
\code{manage_blocks()}, while the UI component (e.g. \code{manage_blocks_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{manage_blocks_server()}) is expected to return \code{NULL}.
}
\description{
Logic and user experience for adding/removing blocks to the board can be
customized or enhanced by providing an alternate version of this plugin. The
default implementation provides a modal-based UI with simple shiny inputs
such as drop-downs and text fields.
}
\details{
Updates are mediated via the \code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} object passed as
\code{update}, where block updates are communicated as list entry \code{blocks} with
components \code{add} and \code{rm}, where
\itemize{
\item \code{add} may be \code{NULL} or a \code{block} object (block IDs may not already exist),
\item \code{rm} may be \code{NULL} or a string (of existing block IDs).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot-block.R, R/plot-scatter.R
\name{new_plot_block}
\alias{new_plot_block}
\alias{new_scatter_block}
\title{Plot block constructors}
\usage{
new_plot_block(server, ui, class, ctor = sys.parent(), ...)

new_scatter_block(x = character(), y = character(), ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{...}{Forwarded to \code{new_plot_block()} and \code{\link[=new_block]{new_block()}}}

\item{x, y}{Columns to place on respective axes}
}
\value{
All blocks constructed via \code{new_plot_block()} inherit from
\code{plot_block}.
}
\description{
Blocks for data visualization using base R graphics can be created via
\code{new_plot_block()}.
}
\details{
Due to the current block evaluation procedure, where block evaluation is
separated from block "rendering" (via  \code{\link[shiny:renderPlot]{shiny::renderPlot()}}) integration of
base R graphics requires some mechanism to achieve this decoupling. This
is implemented by adding a \code{plot} attribute to the result of
\code{\link[=block_eval]{block_eval()}}, generated with \code{\link[grDevices:recordPlot]{grDevices::recordPlot()}} and containing the
required information to re-create the plot at a later time. As part of
\code{\link[=block_output]{block_output()}}, the attribute is retrieved and passed to
\code{\link[grDevices:replayPlot]{grDevices::replayPlot()}}. Consequently, any block that inherits from
\code{plot_block} is required to support this type of decoupling.
}
\section{Scatter block}{

Mainly for demonstration purposes, this block draws a scattter plot using
\code{\link[base:plot]{base::plot()}}. In its current simplistic implementation, apart from axis
labels (fixed to the corresponding column names), no further plotting
options are available and for any "production" application, a more
sophisticated (set of) block(s) for data visualization will most likely be
required.
}

\seealso{
Real-world plot blocks built on \code{new_plot_block()} are provided by
the \href{https://bristolmyerssquibb.github.io/blockr.ggplot/}{blockr.ggplot}
package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-shiny.R
\name{get_session}
\alias{get_session}
\alias{notify}
\title{Shiny utilities}
\usage{
get_session()

notify(
  ...,
  envir = parent.frame(),
  action = NULL,
  duration = 5,
  close_button = TRUE,
  id = NULL,
  type = c("message", "warning", "error"),
  glue = TRUE,
  log = TRUE,
  session = get_session()
)
}
\arguments{
\item{...}{Concatenated as \code{paste0(..., "\\n")}}

\item{envir}{Environment where the logging call originated from}

\item{action}{Message content that represents an action. For example, this
could be a link that the user can click on. This is separate from \code{ui}
so customized layouts can handle the main notification content separately
from action content.}

\item{duration}{Number of seconds to display the message before it
disappears. Use \code{NULL} to make the message not automatically
disappear.}

\item{close_button}{Passed as \code{closeButton} to \code{\link[shiny:showNotification]{shiny::showNotification()}}}

\item{id}{A unique identifier for the notification.

\code{id} is optional for \code{showNotification()}: Shiny will automatically create
one if needed. If you do supply it, Shiny will update an existing
notification if it exists, otherwise it will create a new one.

\code{id} is required for \code{removeNotification()}.}

\item{type}{A string which controls the color of the notification. One of
"default" (gray), "message" (blue), "warning" (yellow), or "error" (red).}

\item{glue, log}{Whether to \code{\link[glue:glue]{glue::glue()}}-interpolate \code{...} and whether to
emit a log message. Set both to \code{FALSE} to surface pre-formatted,
already-logged text (e.g. a captured condition message, which may contain
braces that would otherwise fail interpolation).}

\item{session}{Session object to send notification to.}
}
\value{
Either \code{NULL} or a shiny session object for \code{get_session()}, a list
of arguments for plugin server functions in the case of
\code{generate_plugin_args()} and \code{notify()} is called for the side-effect of
displaying a browser notification (and returns \code{NULL} invisibly).
}
\description{
Utility functions for shiny:
\itemize{
\item \code{get_session}: See \code{\link[shiny:getDefaultReactiveDomain]{shiny::getDefaultReactiveDomain()}}.
\item \code{generate_plugin_args}: Meant for unit testing plugins.
\item \code{notify}: Glue-capable wrapper for \code{\link[shiny:showNotification]{shiny::showNotification()}}.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-roclet.R
\name{block_roclet}
\alias{block_roclet}
\alias{block_registration_roclet}
\title{Block registration roclet}
\usage{
block_registration_roclet()
}
\value{
\code{block_registration_roclet()} returns a roclet object.
}
\description{
A custom roxygen2 roclet that turns block registration metadata, declared as
tags on block constructors, into a YAML registry (\code{inst/registry/blocks.yml})
that \code{\link[=register_package_blocks]{register_package_blocks()}} reads at load time. It runs alongside the
standard roclets during \code{\link[roxygen2:roxygenise]{roxygen2::roxygenise()}} (or \code{devtools::document()})
for any package that lists \code{blockr.core::block_registration_roclet} in the
\code{Roxygen} field of its \code{DESCRIPTION}.
}
\section{Tags}{

Placed in the roxygen block above a block constructor:
\describe{
\item{\verb{@block <name>}}{Human-readable block name. Required: its presence is
what marks a constructor for registration.}
\item{\verb{@blockDescr <text>}}{Short block description. Required.}
\item{\verb{@blockCategory <category>}}{One of the \code{\link[=suggested_categories]{suggested_categories()}}.
Required.}
\item{\verb{@blockIcon <icon>}}{Bootstrap icon name. Optional, defaulting to the
category icon (see \code{\link[=default_icon]{default_icon()}}).}
\item{\verb{@blockGuidance <text>}}{Model-facing construction guidance -- do and
don't rules, enumerations, pitfalls. Optional.}
\item{\verb{@blockKeywords <terms>}}{Comma-separated search terms for block
discovery; a term may contain spaces. Optional.}
\item{\verb{@blockDetails <text>}}{Longer human-facing description (e.g. for a
help popover). Optional; when omitted it falls back to the constructor's
\verb{@details} or, failing that, its \verb{@section} prose.}
\item{\verb{@blockLink <url>}}{URL of the block's help or documentation page.
Optional; when omitted it is derived from the package's pkgdown \code{url}
(\verb{_pkgdown.yml} or \code{DESCRIPTION}) and the documented topic, as
\verb{<url>/reference/<topic>.html}.}
\item{\verb{@blockArg <name> <description>}}{One constructor argument's
specification. The text after the name is a free-text description;
\verb{[example] <expr>} and \verb{[type] <expr>} markers (each on its own line)
supply an R expression -- evaluated when documentation is generated --
for a worked example value and an \verb{arg_*()} type descriptor (see
\code{\link[=new_block_arg]{new_block_arg()}}). An explicit \verb{[description] <text>} marker may replace
the inline description. Optional and repeatable.}
\item{\verb{@blockExamples <expr>}}{Block-level worked example configurations,
as an R expression -- evaluated when documentation is generated -- that
yields a list of complete configurations (each a named list keyed by
argument). Optional; supersedes the per-argument \verb{[example]} assembly.}
\item{\verb{@blockCtor <name>}}{Constructor name override. Optional: the
documented object otherwise supplies the name.}
}
Multi-argument worked \code{examples} (whole-block configurations, as opposed to a
per-argument \verb{[example]}) remain out of the tags' scope; a block needing
those registers via \code{\link[=register_block]{register_block()}} directly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-class.R, R/links-class.R,
%   R/utils-graph.R
\name{is_acyclic.board}
\alias{is_acyclic.board}
\alias{is_acyclic.links}
\alias{topo_sort}
\alias{is_acyclic}
\alias{is_acyclic.matrix}
\title{Graph utils}
\usage{
\method{is_acyclic}{board}(x)

\method{is_acyclic}{links}(x)

topo_sort(x)

is_acyclic(x)

\method{is_acyclic}{matrix}(x)
}
\arguments{
\item{x}{Object}
}
\value{
Topological ordering via \code{topo_sort()} returns a character vector
with sorted node IDs and the generic function \code{is_acyclic()} is expected to
return a scalar logical value.
}
\description{
Block dependencies are represented by DAGs and graph utility functions
\code{topo_sort()} and \code{is_acyclic()} are used to create a topological ordering
(implemented as DFS) of blocks and to check for cycles. An adjacency matrix
corresponding to a board is available as \code{as.matrix()}.
}
\examples{
brd <- new_board(
  c(
     a = new_dataset_block(),
     b = new_dataset_block(),
     c = new_scatter_block(),
     d = new_subset_block()
  ),
  list(from = c("a", "d"), to = c("d", "c"))
)

as.matrix(brd)
topo_sort(brd)
is_acyclic(brd)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-server.R
\name{board_update}
\alias{board_update}
\alias{validate_board_update}
\alias{augment_board_update}
\alias{apply_board_update}
\title{Board update}
\usage{
validate_board_update(payload, board, ..., session = get_session())

augment_board_update(upd, board, ..., session = get_session())

apply_board_update(board, upd, ..., session = get_session())
}
\arguments{
\item{payload, upd}{A board update payload — see Validation above
for the accepted shape.}

\item{board}{A \code{board} object.}

\item{...}{Forwarded between methods. For \code{\link[=apply_board_update]{apply_board_update()}},
the final observer also splices \code{board_server()}'s \code{...} in here.}

\item{session}{A shiny session, default \code{\link[=get_session]{get_session()}}.}
}
\value{
\code{\link[=validate_board_update]{validate_board_update()}} returns \code{invisible(payload)} (or
throws a \code{blockr_abort()} error). \code{\link[=augment_board_update]{augment_board_update()}} returns
the (possibly extended) payload. \code{\link[=apply_board_update]{apply_board_update()}} returns a
\code{board}.
}
\description{
Inside \code{\link[=board_server]{board_server()}} every state change flows through one
\code{board_update} reactive. Core registers two observers framing the
change: an initial one that validates the payload and runs
\code{\link[=augment_board_update]{augment_board_update()}} for auto-fixups, and a final one that runs
\code{\link[=apply_board_update]{apply_board_update()}} and resets the reactive. Plugins or
callbacks may register their own observers in between, provided they
use a \emph{finite} priority — the highest and lowest reactive priorities
are reserved for core.
}
\details{
All three functions dispatch on the \code{board} class. Subclasses
override to validate, augment, or react to their own payload slots,
typically composing with \code{NextMethod()}. \code{\link[=validate_board_update]{validate_board_update()}}
is also a caller-facing entry point: it mirrors the initial
observer's checks against a caller-supplied payload, useful for
staging layers (e.g. accumulating LLM-proposed updates) that need
to fail loudly before publishing.
}
\section{Validation}{

The default \code{.board} method runs a structural check on the payload
(block / link / stack per-slot rules) and a cross-reference check
that link endpoints and stack members resolve in the post-update
merged view. Unknown top-level keys are passed through, so subclass
payload slots reach subclass augment / apply methods.
}

\section{Augment}{

The default \code{.board} method inserts implied link removals and stack
updates that follow from block removals, plus link-input
completion. Subclass methods may extend the payload with their own
fixups; an error thrown here aborts the update before apply runs.
}

\section{Apply}{

The default \code{.board} method returns the supplied board unchanged —
the core apply path (block / link / stack mutation, block UI
insertion / removal) is not routed through this generic. Subclass
methods receive a plain \code{board} snapshot (no reactive surface) and
return a \code{board}, which the final observer assigns back to
\code{rv$board}. For piecemeal customization of the core apply path
itself, override the relevant sub-generic
(\code{\link[=modify_board_links]{modify_board_links()}}, \code{\link[=insert_block_ui]{insert_block_ui()}}, etc.) instead.

Errors thrown from either augment or apply are caught by the
observer, reported via \code{\link[=notify]{notify()}}, and the reactive is reset so the
app keeps running.
}

\section{Outcome}{

Alongside the human-facing \code{\link[=notify]{notify()}} toast, every update cycle
records a machine-readable result into \code{board$last_update} (the
read-only board handed to plugins and callbacks). It is a list with
a monotonically increasing \code{seq}, a logical \code{ok}, the \code{phase} it
ended in (\code{"validate"} or \code{"apply"}), and a \code{message}
(\code{conditionMessage()} on failure, \code{NA} on success); it is \code{NULL}
before the first update. The \code{seq} advances on every write so that
two consecutive identical outcomes still invalidate a downstream
observer. A programmatic caller can watch this field to learn
whether a dispatched update was rejected, failed to apply, or landed.
}

\examples{
brd <- new_board(
  blocks = c(a = new_dataset_block("iris"), b = new_subset_block()),
  links = links(ab = new_link(from = "a", to = "b"))
)

validate_board_update(
  list(links = list(rm = "ab")),
  brd
)

try(
  validate_board_update(
    list(links = list(add = links(xy = new_link(from = "x", to = "y")))),
    brd
  )
)

}
\seealso{
On a locked board (see \code{\link[=is_board_locked]{is_board_locked()}}) the update is
dropped rather than applied.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-assertions.R
\name{is_scalar}
\alias{is_scalar}
\alias{is_string}
\alias{is_bool}
\alias{is_intish}
\alias{is_count}
\alias{is_number}
\alias{not_null}
\alias{has_length}
\title{Assertions}
\usage{
is_scalar(x)

is_string(x)

is_bool(x)

is_intish(x)

is_count(x, allow_zero = TRUE)

is_number(x)

not_null(...)

has_length(x)
}
\arguments{
\item{x}{Object to check}

\item{allow_zero}{Determines whether the value 0 is considered a valid count}

\item{...}{Silently ignored}
}
\value{
Scalar logical value.
}
\description{
Utility functions, mainly intended for asserting common preconditions are
exported for convenience in dependent packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-misc.R
\name{blockr_option}
\alias{blockr_option}
\alias{set_blockr_options}
\title{Blockr Options}
\usage{
blockr_option(name, default)

set_blockr_options(...)
}
\arguments{
\item{name}{Option name}

\item{default}{Default value}

\item{...}{Option key value pairs as named arguments}
}
\value{
The value set as option \code{name} or \code{default} if not set. In case of
the option being available only as environment variable, the value will be
a string and if available as \code{\link[base:options]{base::options()}} entry it may be of any R type.
}
\description{
Retrieves options via \code{\link[base:getOption]{base::getOption()}} or \code{\link[base:Sys.getenv]{base::Sys.getenv()}}, in that
order, and prefixes the option name passed as \code{name} with \code{blockr.} or
\code{blockr_} respectively. Additionally, the name is converted to lower case
for \code{getOption()} and upper case for environment variables. In case no value
is available for a given \code{name}, \code{default} is returned.
}
\examples{
blockr_option("test-example", "default")

options(`blockr.test-example` = "non-default")
blockr_option("test-example", "default")

Sys.setenv(`BLOCKR_TEST-EXAMPLE` = "another value")
tryCatch(
  blockr_option("test-example", "default"),
  error = function(e) conditionMessage(e)
)
options(`blockr.test-example` = NULL)
blockr_option("test-example", "default")

Sys.unsetenv("BLOCKR_TEST-EXAMPLE")
blockr_option("test-example", "default")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/text-block.R, R/text-glue.R
\name{new_text_block}
\alias{new_text_block}
\alias{new_glue_block}
\title{Text block constructors}
\usage{
new_text_block(server, ui, class, ctor = sys.parent(), ...)

new_glue_block(text = character(), ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{...}{Forwarded to \code{new_text_block()} and \code{\link[=new_block]{new_block()}}}

\item{text}{String evaluated using \code{\link[glue:glue]{glue::glue()}}}
}
\value{
All blocks constructed via \code{new_text_block()} inherit from
\code{text_block}.
}
\description{
A text block produces (markdown styled) text, given some (optional) data
input.
}
\section{Glue block}{

Using \code{\link[glue:glue]{glue::glue()}}, this block allows evaluation of a text string in the
context of datasets to produce (markdown formatted) text as block result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-meta.R
\name{block_metadata}
\alias{block_metadata}
\alias{block_meta_id}
\alias{block_meta_name}
\alias{block_meta_description}
\alias{block_meta_details}
\alias{block_meta_link}
\alias{block_meta_guidance}
\alias{block_meta_category}
\alias{block_meta_icon}
\alias{block_meta_package}
\alias{block_meta_keywords}
\alias{block_meta_arguments}
\alias{block_meta_examples}
\title{Block metadata}
\usage{
block_metadata(x, fields = "all", ...)

block_meta_id(x, ...)

block_meta_name(x, ...)

block_meta_description(x, ...)

block_meta_details(x, ...)

block_meta_link(x, ...)

block_meta_guidance(x, ...)

block_meta_category(x, ...)

block_meta_icon(x, ...)

block_meta_package(x, ...)

block_meta_keywords(x, ...)

block_meta_arguments(x, ...)

block_meta_examples(x, ...)
}
\arguments{
\item{x}{A \code{block}, a \code{blocks} collection, a \code{block_registry_entry} or a
registry ID}

\item{fields}{Metadata fields to include (defaults to \code{"all"})}

\item{...}{Generic consistency, passed on to methods}
}
\value{
\code{block_metadata()} returns a \code{data.frame}. The \verb{block_meta_*()}
getters return the named attribute: a string (or \code{NA}) for scalar fields, a
character vector for \code{block_meta_keywords()}, a \code{block_args} object for
\code{block_meta_arguments()}, and a list of worked configurations for
\code{block_meta_examples()}.
}
\description{
Registry metadata for blocks is available both as a tabular overview and via
per-attribute accessors. \code{block_metadata()} returns a \code{data.frame} with one
row per block -- dispatching on a \code{block}, a \code{blocks} collection, a
\code{block_registry_entry} or a registry ID -- where scalar attributes are
atomic columns and the multi-valued ones (\code{arguments}, \code{examples},
\code{keywords}) are list-columns. The \code{fields} argument selects a subset of
columns. Each attribute additionally has a dedicated getter
(\code{block_meta_name()}, \code{block_meta_guidance()}, ...) returning that attribute
for a single block. Missing fields are filled with display defaults in the
data.frame; the getters instead return the stored value (or \code{NA} / an empty
value).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-class.R, R/board-option.R,
%   R/board-options.R
\name{board_ctor}
\alias{board_ctor}
\alias{new_board_option}
\alias{is_board_option}
\alias{validate_board_option}
\alias{as_board_option}
\alias{as_board_option.board_option}
\alias{board_option_id}
\alias{board_option_trigger}
\alias{board_option_default}
\alias{board_option_category}
\alias{board_option_ui}
\alias{board_option_server}
\alias{board_option_transform}
\alias{board_option_value}
\alias{board_option_ctor}
\alias{validate_board_option.default}
\alias{new_board_name_option}
\alias{new_n_rows_option}
\alias{new_page_size_option}
\alias{new_filter_rows_option}
\alias{new_thematic_option}
\alias{new_dark_mode_option}
\alias{new_show_conditions_option}
\alias{new_llm_model_option}
\alias{new_board_options}
\alias{default_board_options}
\alias{is_board_options}
\alias{as_board_options}
\alias{as_board_options.board_options}
\alias{as_board_options.board_option}
\alias{as_board_options.list}
\alias{as_board_options.board}
\alias{validate_board_options}
\alias{board_option_values}
\alias{get_board_option_value}
\alias{set_board_option_value}
\alias{get_board_option_or_default}
\alias{get_board_option_or_null}
\alias{get_board_option_values}
\alias{combine_board_options}
\title{Board options}
\usage{
board_ctor(x)

new_board_option(
  id,
  default,
  ui,
  server = function(board, ..., session) {
 },
  update_trigger = id,
  transform = identity,
  category = NULL,
  ctor = sys.parent(),
  pkg = NULL
)

is_board_option(x)

validate_board_option(x)

as_board_option(x, ...)

\method{as_board_option}{board_option}(x, ...)

board_option_id(x)

board_option_trigger(x)

board_option_default(x)

board_option_category(x)

board_option_ui(x, id = NULL)

board_option_server(x, ...)

board_option_transform(x)

board_option_value(x, value = board_option_default(x))

board_option_ctor(x)

\method{validate_board_option}{default}(x)

new_board_name_option(value = NULL, category = "Board options", ...)

new_n_rows_option(
  value = blockr_option("n_rows", 50L),
  category = "Table options",
  ...
)

new_page_size_option(
  value = blockr_option("page_size", 5L),
  category = "Table options",
  ...
)

new_filter_rows_option(
  value = blockr_option("filter_rows", FALSE),
  category = "Table options",
  ...
)

new_thematic_option(
  value = blockr_option("thematic", NULL),
  category = "Theme options",
  ...
)

new_dark_mode_option(
  value = blockr_option("dark_mode", NULL),
  category = "Theme options",
  ...
)

new_show_conditions_option(
  value = blockr_option("show_conditions", c("warning", "error")),
  category = "Board options",
  ...
)

new_llm_model_option(value = NULL, category = "Board options", ...)

new_board_options(...)

default_board_options(...)

is_board_options(x)

as_board_options(x)

\method{as_board_options}{board_options}(x)

\method{as_board_options}{board_option}(x)

\method{as_board_options}{list}(x)

\method{as_board_options}{board}(x)

validate_board_options(x)

board_option_values(x)

get_board_option_value(opt, session = get_session())

set_board_option_value(opt, val, board, session = get_session())

get_board_option_or_default(
  opt,
  opts = default_board_options(),
  session = get_session()
)

get_board_option_or_null(opt, session = get_session())

get_board_option_values(
  ...,
  opts = default_board_options(),
  if_not_found = c("error", "default", "null"),
  session = get_session()
)

combine_board_options(...)
}
\arguments{
\item{x}{Board options object}

\item{id}{Board option ID}

\item{default}{Default value}

\item{ui}{Option UI}

\item{server}{(Optional) option server}

\item{update_trigger}{Shiny \code{input} entry/entries that trigger an update}

\item{transform}{(Optional) transform function}

\item{category}{(Optional) string-valued category}

\item{ctor, pkg}{Constructor information (used for serialization)}

\item{...}{Options passed as individual arguments}

\item{value}{Option value}

\item{opt}{Option name}

\item{session}{Shiny session}

\item{val}{New value}

\item{board}{Board the option belongs to, used to resolve the lock
state (see \code{\link[=is_board_locked]{is_board_locked()}}).}

\item{opts}{Board options}

\item{if_not_found}{Behavior in case an option is not found}
}
\value{
All of \code{new_board_options()} and \code{as_board_options()} return a
\code{board_options} object, as does the validator \code{validate_board_options()},
which is typically called for side effects of throwing errors is validation
does not pass. Inheritance checking as \code{is_board_options()} returns a scalar
logical, while \code{board_option_values()} returns a named list of option values.
}
\description{
User settings at the board level are managed by a \code{board_options} object.
This can be constructed via \code{new_board_options()} and in case the set of
user options is to be extended, the constructor is designed with sub-classing
in mind. Consequently, the associated validator \code{validate_board_options()}
is available as S3 generic. Inheritance checking is available as
\code{is_board_options()} and coercion as \code{as_board_options()}.
}
\examples{
opt <- new_board_options(
   new_board_name_option(),
   new_page_size_option()
 )

is_board_options(opt)
names(opt)

opt[["page_size"]]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-links.R
\name{manage_links}
\alias{manage_links}
\alias{manage_links_server}
\alias{manage_links_ui}
\title{Plugin module for managing board links}
\usage{
manage_links(server = manage_links_server, ui = manage_links_ui)

manage_links_server(id, board, update, ...)

manage_links_ui(id, board)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{The initial \code{board} object}

\item{update}{Reactive value object to initiate board updates}

\item{...}{Extra arguments passed from parent scope}
}
\value{
A plugin container inheriting from \code{manage_links} is returned by
\code{manage_links()}, while the UI component (e.g. \code{manage_links_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:reexports]{shiny::tagList()}}) and the server
component (i.e. \code{manage_links_server()}) is expected to return \code{NULL}.
}
\description{
Logic and user experience for adding new, removing and modifying existing
links to/from the board can be customized or enhanced by providing an
alternate version of this plugin. The default implementation provides a
table-based UI, presented in a modal.
}
\details{
Updates are mediated via the \code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} object passed as
\code{update}, where link updates are communicated as list entry \code{stacks} with
components \code{add}, \code{rm} or \code{mod}, where
\itemize{
\item \code{add} is either \code{NULL} or a \code{links} object (link IDs may not already
exists),
\item \code{rm} is either \code{NULL} or a character vector of (existing) link IDs,
\item \code{mod} is either \code{NULL} or a \code{links} object (where link IDs must already
exist).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-block.R, R/data-dataset.R,
%   R/data-static.R
\name{new_data_block}
\alias{new_data_block}
\alias{new_dataset_block}
\alias{new_static_block}
\title{Data block constructors}
\usage{
new_data_block(server, ui, class, ctor = sys.parent(), ...)

new_dataset_block(dataset = character(), package = "datasets", ...)

new_static_block(data, ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{...}{Forwarded to \code{new_data_block()} and \code{\link[=new_block]{new_block()}}}

\item{dataset}{Selected dataset}

\item{package}{Name of an R package containing datasets}

\item{data}{Data (used directly as block result)}
}
\value{
All blocks constructed via \code{new_data_block()} inherit from
\code{data_block}.
}
\description{
Data blocks typically do not have data inputs and represent root nodes in
analysis graphs. Intended as initial steps in a pipeline, such blocks are
responsible for providing down-stream blocks with data.
}
\section{Dataset block}{

This data block allows to select a dataset from a package, such as
the datasets package available in most R installations as one of the
packages with "recommended" priority. The source package can be chosen at
time of block instantiation and can be set to any R package, for which then
a set of candidate datasets is computed. This includes exported objects that
inherit from \code{data.frame}.
}

\section{Static block}{

Mainly useful for testing and examples, this block simply returns the data
with which it was initialized. Serialization of static blocks is not allowed
and exported code will not be self-contained in the sense that it will not
be possible to reproduce results in a script that contains code from a
static block.
}

\seealso{
Real-world data blocks built on \code{new_data_block()} are provided by
the \href{https://bristolmyerssquibb.github.io/blockr.io/}{blockr.io} package,
which sources data from a range of external formats.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-lock.R
\name{locked-board}
\alias{locked-board}
\alias{is_board_locked}
\alias{is_board_locked.board}
\title{Locked boards}
\usage{
is_board_locked(board, ...)

\method{is_board_locked}{board}(board, ...)
}
\arguments{
\item{board}{A \code{board} object to dispatch on.}

\item{...}{Generic consistency.}
}
\value{
A logical flag.
}
\description{
A board can be deployed read-only by setting the \code{blockr.locked}
option (see \code{\link[=blockr_option]{blockr_option()}}). Locking is enforced server-side, not
by UI hiding: a forged \code{Shiny.setInputValue} bypasses any
\code{display: none} and fires the underlying observer, so the refusal has
to happen on the server. While locked, the two channels every
mutation funnels through -- the \code{board_update} lifecycle (see
\code{\link[=board_update]{board_update()}}) and \code{\link[=set_board_option_value]{set_board_option_value()}} -- refuse to apply
changes.
}
\details{
\code{is_board_locked()} reports whether a board is locked and dispatches
on the board, so a subclass can source the state however it likes
(for example gating an unlock behind an authenticated request). The
default \code{board} method reads the \code{blockr.locked} option, and
\code{blockr.dock}'s \code{is_dock_locked()} delegates here.
}
\section{Scope}{
 Locking guards against forged \emph{client input} -- a
browser that sets Shiny inputs but cannot run server-side R. It is
\strong{not} a defense against code executing in the session: blockr
evaluates arbitrary R, which can flip the option, rebind this generic
or mutate board state directly, and no in-process flag can prevent
that. A subclass method may raise the bar (e.g. an authenticated
unlock) but cannot close the gap on its own -- a hard boundary has to
live where the user's code does not run (a sandboxed evaluator, or an
ephemeral read-only deployment). Treat locking as a guard for
non-malicious users.
}

\examples{
is_board_locked(new_board())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-misc.R, R/utils-pkg.R
\name{set_names}
\alias{set_names}
\alias{paste_enum}
\alias{coal}
\alias{reval}
\alias{reval_if}
\alias{unlst}
\alias{pkg_name}
\alias{pkg_version}
\alias{pkg_file}
\alias{pkg_avail}
\title{Miscellaneous utilities}
\usage{
set_names(object = nm, nm)

paste_enum(x, sep = ", ", conj = " and ", quotes = "`")

coal(..., fail_all = TRUE, test_fun = is.null)

reval(x)

reval_if(x)

unlst(x, recursive = FALSE, use_names = FALSE)

pkg_name(env = parent.frame())

pkg_version(pkg = parent.frame())

pkg_file(..., pkg = parent.frame())

pkg_avail(...)
}
\arguments{
\item{object, nm}{See \code{\link[stats:setNames]{stats::setNames()}}}

\item{x}{Character vector to \code{\link[base:paste]{base::paste()}}}

\item{sep, conj}{Separation strings for all but last and last positions}

\item{quotes}{Quotes to wrap each entry in \code{x} with}

\item{...}{Set of objects to iterate over}

\item{fail_all}{Error if no non-null objects are present}

\item{test_fun}{Function to test each element with}

\item{recursive, use_names}{See \code{\link[base:unlist]{base::unlist()}}}

\item{env}{An environment that is resolved to a package name}

\item{pkg}{A string.valued package name or an environment passed to
\code{\link[=pkg_name]{pkg_name()}}}
}
\value{
Function \code{set_names()} returns a names object, \code{paste_enum()} a
string, \code{coal()} the first non-null object and \code{unlst()} performs the same
action as \code{\link[base:unlist]{base::unlist()}} but with differing defaults.
}
\description{
Several internal utility functions are exported for convenience in case
dependent packages can make use of this functionality.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-serve.R
\name{serve}
\alias{serve}
\alias{serve.block}
\alias{serve.board}
\alias{blockr_app_plugins}
\alias{custom_plugins}
\alias{blockr_app_options}
\alias{custom_options}
\alias{blockr_app_ui}
\alias{blockr_app_server}
\alias{blockr_test_exports}
\title{Serve object}
\usage{
serve(x, ...)

\method{serve}{block}(x, id = "block", ..., data = list())

\method{serve}{board}(
  x,
  id = rand_names(),
  plugins = blockr_app_plugins,
  options = blockr_app_options,
  loader = local_loader(),
  ...
)

blockr_app_plugins(x)

custom_plugins(x)

blockr_app_options(x)

custom_options(x)

blockr_app_ui(id, x, ...)

blockr_app_server(id, x, ...)

blockr_test_exports(x, rv, ...)
}
\arguments{
\item{x}{Object}

\item{...}{Generic consistency}

\item{id}{Board namespace ID}

\item{data}{Data inputs}

\item{plugins}{Board plugins}

\item{options}{Board options}

\item{loader}{A \code{\link[=board_loader]{board_loader()}} resolving the board to build for each
request; defaults to \code{\link[=local_loader]{local_loader()}}, the in-process save/restore handoff}

\item{rv}{Board \code{\link[shiny:reactiveValues]{shiny::reactiveValues()}}}
}
\value{
The generic \code{serve()} is expected to return the result of a call to
\code{\link[shiny:shinyApp]{shiny::shinyApp()}}.
}
\description{
Intended as entry point to start up a shiny app, the generic function
\code{serve()} can be dispatched either on a single block (mainly for previewing
purposes during block development) or an entire board
}
\section{Examples in Shinylive}{
\describe{
  \item{example-1}{
    \href{https://shinylive.io/r/app/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAdzgCMAnRRASwgGdSoAbbgCgA6YOtyIEA1gwzEGcIbgAEs1EXYKAvAqEALUqVTtEAeiMEAnhElQMAcxaltAVzoYWRIyLGSMtOliNCAJQCENwsjFAMZnyeElIycMEccAwAbnCCEAoKEHDUAPrwDDZw+bHigmAAKizwQbgh2QAmUFwaCmGcmdnZAB7tLVzscKSGiABCAPIAIg1ZPWYDrVDDo8wAwtosEgDqcCw2uo0KSYFgAL4AukA}{Open in Shinylive}
    \if{html}{\out{<iframe class="iframe_shinylive" src="https://shinylive.io/r/app/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAdzgCMAnRRASwgGdSoAbbgCgA6YOtyIEA1gwzEGcIbgAEs1EXYKAvAqEALUqVTtEAeiMEAnhElQMAcxaltAVzoYWRIyLGSMtOliNCAJQCENwsjFAMZnyeElIycMEccAwAbnCCEAoKEHDUAPrwDDZw+bHigmAAKizwQbgh2QAmUFwaCmGcmdnZAB7tLVzscKSGiABCAPIAIg1ZPWYDrVDDo8wAwtosEgDqcCw2uo0KSYFgAL4AukA" style="height: 800px; width: 100vw; max-width: 1400px; border: 1px solid rgba(0,0,0,0.175); border-radius: .375rem; position: absolute; left: 50\%; margin-top: 30px; transform: translateX(-50\%); z-index: 1"></iframe>}}
    \if{html}{\out{<a style='height: 800px; display: block;'></a>}}
  }
  \item{example-2}{
    \href{https://shinylive.io/r/app/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAdzgCMAnRRASwgGdSoAbbgCgA6YOtyIEA1gwzEGcIbgAEs1EXYKAvAqEALUqVTtEAeiMEAnhElQMAcxaltAVzoYWRIyLGSMtOliNCAJQCENwsjFAMZnyeElIycMEccAwAbnCCEAoKEHDUAPp0RJEAJpnZ2bHiapoE5RXZUBo5efklUFzscKSFohKCYABCAPIAIkG4IQ2VzbkF7Z3dvV4DAMLaLBIA6nAsNroTU9MEs63wDDZwy-1CACos8EFH2YGTWRVhljUKdc8VUCdNHN8p9xAMoPItIRIUIAB6Hd4NOiAloFUEDOgw6F4KFmJ6IhSvP6cAHVZphTh8AHNX4oLEEIJJF4hQJgAC+AF0gA}{Open in Shinylive}
    \if{html}{\out{<iframe class="iframe_shinylive" src="https://shinylive.io/r/app/#code=NobwRAdghgtgpmAXGKAHVA6ASmANGAYwHsIAXOMpMAdzgCMAnRRASwgGdSoAbbgCgA6YOtyIEA1gwzEGcIbgAEs1EXYKAvAqEALUqVTtEAeiMEAnhElQMAcxaltAVzoYWRIyLGSMtOliNCAJQCENwsjFAMZnyeElIycMEccAwAbnCCEAoKEHDUAPp0RJEAJpnZ2bHiapoE5RXZUBo5efklUFzscKSFohKCYABCAPIAIkG4IQ2VzbkF7Z3dvV4DAMLaLBIA6nAsNroTU9MEs63wDDZwy-1CACos8EFH2YGTWRVhljUKdc8VUCdNHN8p9xAMoPItIRIUIAB6Hd4NOiAloFUEDOgw6F4KFmJ6IhSvP6cAHVZphTh8AHNX4oLEEIJJF4hQJgAC+AF0gA" style="height: 800px; width: 100vw; max-width: 1400px; border: 1px solid rgba(0,0,0,0.175); border-radius: .375rem; position: absolute; left: 50\%; margin-top: 30px; transform: translateX(-50\%); z-index: 1"></iframe>}}
    \if{html}{\out{<a style='height: 800px; display: block;'></a>}}
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-code.R
\name{generate_code}
\alias{generate_code}
\alias{generate_code_server}
\alias{generate_code_ui}
\title{Code generation plugin module}
\usage{
generate_code(server = generate_code_server, ui = generate_code_ui)

generate_code_server(id, board, ...)

generate_code_ui(id, board)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{Reactive values object}

\item{...}{Extra arguments passed from parent scope}
}
\value{
A plugin container inheriting from \code{generate_code} is returned by
\code{generate_code()}, while the UI component (e.g. \code{generate_code_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{generate_code_server()}) is expected to return \code{NULL}.
}
\description{
All code necessary for reproducing a data analysis as set up in blockr can
be made available to the user. Several ways of providing such a script or
code snippet are conceivable and currently implemented, we have a modal
with copy-to-clipboard functionality. This is readily extensible, for example
by offering a download button, by providing this functionality as a
\code{generate_code} module.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-class.R
\name{block_name}
\alias{block_name}
\alias{block_name<-}
\alias{validate_data_inputs}
\alias{block_inputs}
\alias{block_arity}
\alias{external_ctrl_vars}
\alias{has_external_ctrl}
\title{Block utilities}
\usage{
block_name(x)

block_name(x) <- value

validate_data_inputs(x, data)

block_inputs(x)

block_arity(x)

external_ctrl_vars(x)

has_external_ctrl(x)
}
\arguments{
\item{x}{An object inheriting from \code{"block"}}

\item{value}{New value}

\item{data}{Data input values}
}
\value{
Return types vary among the set of exported utilities:
\itemize{
\item \code{block_name()}: string valued block name,
\item \verb{block_name<-()}: \code{x} (invisibly),
\item \code{validate_data_inputs()}: \code{NULL} if no validator is set and the result of
the validator function otherwise,
\item \code{block_inputs()}: a (possibly empty) character vector of data input names,
\item \code{block_arity()}: a scalar integer with \code{NA} in case of variadic behavior,
\item \code{external_ctrl_vars()}: a character vector of externally controllable
variable names (always including \code{"block_name"} for blocks),
\item \code{has_external_ctrl()}: a scalar logical.
}
}
\description{
Several utilities for working (and manipulating) \code{block} objects are exported
and developers are encouraged to use these instead of relying on object
implementation to extract or modify attributes. If functionality for working
with blocks in lacking, please consider opening an
\href{https://github.com/BristolMyersSquibb/blockr.core/issues/new}{issue}.
}
\section{Block name}{

Each block can have a name (by default constructed from the class vector)
intended for users to easily identify different blocks. This name can freely
be changed during the lifetime of a block and no uniqueness restrictions are
in place. The current block name can be retrieved with \code{block_name()} and
set as \code{block_name(x) <- "some name"}.
}

\section{Input validation}{

Data input validation is available via \code{validate_data_inputs()} which uses
the (optional) validator function passed to \code{\link[=new_block]{new_block()}} at construction
time. This mechanism can be used to prevent premature evaluation of the
block expression as this might lead to unexpected errors.
}

\section{Block arity/inputs}{

The set of explicit (named) data inputs for a block is available as
\code{block_inputs()}, while the block arity can be queried with \code{block_arity()}.
In case of variadic blocks (i.e. blocks that take a variable number of
inputs like for example a block providing \code{\link[base:rbind]{base::rbind()}}-like
functionality), \code{block_arity()} returns \code{NA} and the special block server
function argument \code{...args}, signalling variadic behavior is stripped from
\code{block_inputs()}.
}

\section{External control}{

Blocks can expose constructor inputs for programmatic control from outside
the block server (see the \code{external_ctrl} argument to \code{\link[=new_block]{new_block()}} and the
\code{\link[=ctrl_block]{ctrl_block()}} plugin). \code{external_ctrl_vars()} is a generic that resolves
this declaration into the concrete set of controllable variable names:
\code{TRUE} expands to all constructor inputs, \code{FALSE} to none and a character
vector is taken as a (validated) subset. For blocks, \code{"block_name"} is
always included as every block can be renamed. The predicate
\code{has_external_ctrl()} reports whether this set is non-empty (always \code{TRUE}
for blocks).
}

\examples{
blk <- new_dataset_block()
block_name(blk)
block_name(blk) <- "My dataset block"
block_name(blk)

block_inputs(new_dataset_block())
block_arity(new_dataset_block())

block_inputs(new_merge_block())
block_arity(new_merge_block())

block_inputs(new_rbind_block())
block_arity(new_rbind_block())

external_ctrl_vars(new_dataset_block())
has_external_ctrl(new_dataset_block())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-loader.R
\name{board_loader}
\alias{board_loader}
\alias{is_board_loader}
\alias{local_loader}
\title{Board loader}
\usage{
board_loader(resolve, stage = NULL)

is_board_loader(x)

local_loader()
}
\arguments{
\item{resolve, stage}{Paired functions backing a \code{board_loader}: \code{resolve} is
\verb{function(request, session, default)} returning the \code{board} to build or
\code{NULL}; \code{stage} is \verb{function(board, session)} (or \code{NULL} for a loader
that does not stage, e.g. resolve-only) which persists \code{board} and returns
the URL query parameters referencing it (core writes them and reloads). They
share private state, so a \code{board_loader} is built as a unit, not supplied as
two loose functions.}

\item{x}{Object to test for \code{board_loader}-ness}
}
\value{
\code{board_loader()} and \code{local_loader()} return a \code{board_loader} object
and \code{is_board_loader()} a scalar logical.
}
\description{
Which \code{board} to build for an incoming request -- and how a board is carried
across the \code{session$reload()} that a \link{preserve_board} restore triggers -- is
the job of an app-level \strong{board loader}, passed to \code{\link[=serve]{serve()}} as its
\code{loader} argument. A \code{board_loader()} pairs a \code{resolve(request, session, default)} -- returning the \code{board} to build for an incoming request, or
\code{NULL} for the \code{\link[=serve]{serve()}} default -- with an optional \code{stage(board, session)}, which persists a board and returns the URL query parameters that
reference it (a resolve-only loader, e.g. one not backing a restore, leaves
\code{stage} \code{NULL}). \code{\link[=serve]{serve()}} uses that one loader for both the request-phase
resolution (at the GET, where \code{session} is \code{NULL}, and at the WS connect)
and the in-session staging when a restore fires; \strong{core} writes those
parameters into the URL and drives the reload, so the reload stays a
guaranteed core mechanism that no loader can opt out of.
}
\details{
\code{resolve} receives the raw HTTP \code{request} at both phases (the UI request at
the GET, \code{session$request} at the WS), the \code{session} (\code{NULL} at the GET),
and \code{default} -- the \code{board} passed to \code{\link[=serve]{serve()}} (the \code{\link[=serve]{serve()}} default,
built when \code{resolve} returns \code{NULL}), which a loader can also derive its own
result from (e.g. \code{clear_board(default)}). A loader that keys off URL query
parameters reads them itself, minding the phase split: the query is on
\code{request$QUERY_STRING} at the GET but \code{session$clientData$url_search} at the
WS (the websocket request carries neither).

The default \code{\link[=local_loader]{local_loader()}} keeps its handoff in a per-loader store (no
process global) and is therefore single-process; multi-user deployments pass
a loader resolving from a shared backend (as blockr.session does).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-options.R, R/board-ui.R
\name{board_ui.board_options}
\alias{board_ui.board_options}
\alias{board_ui}
\alias{board_ui.board}
\alias{board_ui.NULL}
\alias{insert_block_ui}
\alias{insert_block_ui.board}
\alias{remove_block_ui}
\alias{remove_block_ui.board}
\alias{toolbar_ui}
\alias{toolbar_ui.board}
\title{Board UI}
\usage{
\method{board_ui}{board_options}(id, x, ...)

board_ui(id, x, ...)

\method{board_ui}{board}(id, x, plugins = board_plugins(x), options = NULL, ...)

\method{board_ui}{`NULL`}(id, x, ...)

insert_block_ui(id, x, blocks = NULL, ..., session = get_session())

\method{insert_block_ui}{board}(id, x, blocks = NULL, ..., session = get_session())

remove_block_ui(id, x, blocks = NULL, ..., session = get_session())

\method{remove_block_ui}{board}(id, x, blocks = NULL, ..., session = get_session())

toolbar_ui(id, x, plugins = list(), ...)

\method{toolbar_ui}{board}(id, x, plugins = list(), options = NULL, ...)
}
\arguments{
\item{id}{Namespace ID}

\item{x}{Board}

\item{...}{Generic consistency}

\item{plugins}{UI for board plugins}

\item{options}{Board options (\code{NULL} defaults to the union of board, block
and registry sourced options)}

\item{blocks}{(Additional) blocks (or IDs) for which to generate the UI}

\item{session}{Shiny session}
}
\value{
A \code{board_ui()} implementation is expected to return \link[shiny:tag]{shiny::tag} or
\code{\link[shiny:tagList]{shiny::tagList()}} objects, as does \code{toolbar_ui()}, while updater functions
(\code{insert_block_ui()} and \code{remove_block_ui()}) are called for their side
effects (which includes UI updates such as \code{\link[shiny:insertUI]{shiny::insertUI()}},
\code{\link[shiny:removeUI]{shiny::removeUI()}}) and return the board object passed as \code{x} invisibly.
}
\description{
As counterpart to \code{\link[=board_server]{board_server()}}, \code{board_ui()} is responsible for rendering
UI for a board module. This top-level entry point for customizing board
appearance and functionality can be overridden by sub-classing the boar
object and providing an implementation for this sub-class. Such an
implementation is expected to handle UI for plugins and all board
components, including blocks, links and stacks, but may rely on
functionality that generates UI for these components, such as \code{\link[=block_ui]{block_ui()}}
or \code{\link[=stack_ui]{stack_ui()}}, as well as already available UI provided by plugins
themselves. Additionally, \code{toolbar_ui()} is responsible for creating a
toolbar UI component from several plugin UI components.
}
\details{
Dynamic UI updates are handled by functions \code{insert_block_ui()} and
\code{remove_block_ui()} for adding and removing block-level UI elements to and
from \code{board} UI, whenever blocks are added or removed. These update functions
are provided as S3 generics with implementations for \code{board} and can be
extended if so desired.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-logging.R
\name{write_log}
\alias{write_log}
\alias{log_fatal}
\alias{log_error}
\alias{log_warn}
\alias{log_info}
\alias{log_debug}
\alias{log_trace}
\alias{as_log_level}
\alias{fatal_log_level}
\alias{error_log_level}
\alias{warn_log_level}
\alias{info_log_level}
\alias{debug_log_level}
\alias{trace_log_level}
\alias{get_log_level}
\alias{cnd_logger}
\alias{cat_logger}
\title{Logging}
\usage{
write_log(
  ...,
  level = "info",
  envir = parent.frame(),
  asis = FALSE,
  use_glue = TRUE,
  pkg = pkg_name(envir)
)

log_fatal(..., envir = parent.frame())

log_error(..., envir = parent.frame())

log_warn(..., envir = parent.frame())

log_info(..., envir = parent.frame())

log_debug(..., envir = parent.frame())

log_trace(..., envir = parent.frame())

as_log_level(level)

fatal_log_level

error_log_level

warn_log_level

info_log_level

debug_log_level

trace_log_level

get_log_level()

cnd_logger(msg, level)

cat_logger(msg, level)
}
\arguments{
\item{...}{Concatenated as \code{paste0(..., "\\n")}}

\item{level}{Logging level (possible values are "fatal", "error", "warn",
"info", "debug" and "trace"}

\item{envir}{Environment where the logging call originated from}

\item{asis}{Flag to disable re-wrapping of text to terminal width}

\item{use_glue}{Flag to disable use of glue}

\item{pkg}{Package name}

\item{msg}{Message (string)}
}
\value{
Logging function \code{write_log()}, wrappers \verb{log_*()} and loggers
provided as \code{cnd_logger()}/cat_logger() all return \code{NULL} invisibly and are
called for their side effect of emitting a message. Helpers \code{as_log_level()}
and \code{get_log_level()} return a scalar-valued ordered factor.
}
\description{
Internally used infrastructure for emitting log messages is exported, hoping
that other packages which depend on this, use it and thereby logging is
carried out consistently both in terms of presentation and output device.
All log messages are associated with an (ordered) level ("fatal", "error",
"warn", "info", "debug" or "trace") which is compared against the currently
set value (available as \code{get_log_level()}) and output is only generated if
the message level is greater or equal to the currently set value.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-serdes.R
\name{blockr_ser}
\alias{blockr_ser}
\alias{blockr_ser.block}
\alias{blockr_ser.blocks}
\alias{blockr_ser.board_options}
\alias{blockr_ser.blockr_ctor}
\alias{blockr_ser.board_option}
\alias{blockr_ser.llm_model_option}
\alias{blockr_ser.board}
\alias{blockr_ser.link}
\alias{blockr_ser.links}
\alias{blockr_ser.stack}
\alias{blockr_ser.stacks}
\alias{blockr_deser}
\alias{blockr_deser.list}
\alias{blockr_deser.block}
\alias{blockr_deser.blocks}
\alias{blockr_deser.board}
\alias{blockr_deser.link}
\alias{blockr_deser.links}
\alias{blockr_deser.stack}
\alias{blockr_deser.stacks}
\alias{blockr_deser.board_options}
\alias{blockr_deser.blockr_ctor}
\alias{blockr_deser.board_option}
\title{Serialization utilities}
\usage{
blockr_ser(x, ...)

\method{blockr_ser}{block}(x, state = NULL, ...)

\method{blockr_ser}{blocks}(x, blocks = NULL, ...)

\method{blockr_ser}{board_options}(x, options = NULL, ...)

\method{blockr_ser}{blockr_ctor}(x, ...)

\method{blockr_ser}{board_option}(x, option = NULL, ...)

\method{blockr_ser}{llm_model_option}(x, option = NULL, ...)

\method{blockr_ser}{board}(x, board_id = NULL, ...)

\method{blockr_ser}{link}(x, ...)

\method{blockr_ser}{links}(x, ...)

\method{blockr_ser}{stack}(x, ...)

\method{blockr_ser}{stacks}(x, ...)

blockr_deser(x, ...)

\method{blockr_deser}{list}(x, ...)

\method{blockr_deser}{block}(x, data, ...)

\method{blockr_deser}{blocks}(x, data, ...)

\method{blockr_deser}{board}(x, data, ...)

\method{blockr_deser}{link}(x, data, ...)

\method{blockr_deser}{links}(x, data, ...)

\method{blockr_deser}{stack}(x, data, ...)

\method{blockr_deser}{stacks}(x, data, ...)

\method{blockr_deser}{board_options}(x, data, ...)

\method{blockr_deser}{blockr_ctor}(x, data, ...)

\method{blockr_deser}{board_option}(x, data, ...)
}
\arguments{
\item{x}{Object to (de)serialize}

\item{...}{Generic consistency}

\item{state}{Object state (as returned from an \code{expr_server})}

\item{blocks}{Block states (\code{NULL}, or a per-block \code{NULL} entry, defaults to
values from the constructor scope)}

\item{options}{Board option values (\code{NULL} uses values provided by \code{x})}

\item{option}{Board option value (\code{NULL} uses values provided by \code{x})}

\item{board_id}{Board ID}

\item{data}{List valued data (converted from JSON)}
}
\value{
Serialization helper function \code{blockr_ser()} returns lists, which
for most objects contain slots \code{object} and \code{payload}, where \code{object}
contains a class vector which is used by \code{blockr_deser()} to instantiate an
empty object of that class and use S3 dispatch to identify the correct method
that, given the content in \code{payload}, can re-create the original object.
}
\description{
Blocks are serialized by writing out information on the constructor used to
create the object, combining this with block state information, which
constitutes values such that when passed to the constructor the original
object can be re-created.
}
\details{
Helper functions \code{blockr_ser()} and \code{blockr_deser()} are implemented as
generics and perform most of the heavy lifting for (de-)serialization:
representing objects as easy-to-serialize (nested) lists containing mostly
strings and no objects which are hard/impossible to truthfully re-create in
new sessions (such as environments).

During deserialization, \code{blockr_deser()} forwards \code{...} to the dispatched
per-class method. This lets callers (and outer methods deserializing nested
objects) thread additional context down to inner deserializers.
}
\examples{
blk <- new_dataset_block("iris")

blockr_ser(blk)

all.equal(blk, blockr_deser(blockr_ser(blk)), check.environment = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-class.R
\name{new_board}
\alias{new_board}
\alias{validate_board}
\alias{is_board}
\title{Board}
\usage{
new_board(
  blocks = list(),
  links = list(),
  stacks = list(),
  options = default_board_options(),
  ...,
  ctor = NULL,
  pkg = NULL,
  class = character()
)

validate_board(x)

is_board(x)
}
\arguments{
\item{blocks}{Set of blocks}

\item{links}{Set of links}

\item{stacks}{Set of stacks}

\item{options}{Board-level user settings}

\item{...}{Further (metadata) attributes}

\item{ctor, pkg}{Constructor information (used for serialization)}

\item{class}{Board sub-class}

\item{x}{Board object}
}
\value{
The board constructor \code{new_board()} returns a \code{board} object, as does
the validator \code{validate_board()}, which typically is called for side effects
in the form of errors. Inheritance checking as \code{is_board()} returns a scalar
logical.
}
\description{
A set of blocks, optionally connected via links and grouped into stacks
are organized as a \code{board} object. Boards are constructed using \code{new_board()}
and inheritance can be tested with \code{is_board()}, while validation is
available as (generic function) \code{validate_board()}. This central data
structure can be extended by adding further attributes and sub-classes. S3
dispatch is used in many places to control how the UI looks and feels and
using this extension mechanism, UI aspects can be customized to user
requirements. Several utilities are available for retrieving and modifying
block attributes (see \code{\link[=board_blocks]{board_blocks()}}).
}
\examples{
brd <- new_board(
  c(
     a = new_dataset_block(),
     b = new_subset_block()
  ),
  list(from = "a", to = "b")
)

is_board(brd)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/str-value.R
\name{str_value}
\alias{str_value}
\alias{str_value.default}
\title{Compact rendering}
\usage{
str_value(x, ...)

\method{str_value}{default}(x, ...)
}
\arguments{
\item{x}{Object to render.}

\item{...}{Generic consistency.}
}
\value{
\code{str_value()} returns a length-one character vector (multi-line for
containers). The \code{\link[utils:str]{utils::str()}} methods are called for their side effect
(display) and return their \code{object} invisibly.
}
\description{
\code{str_value()} returns a compact string describing an object. It is the
value-returning half of the compact rendering tier: where \code{\link[utils:str]{utils::str()}}
only displays (it \code{cat()}s and returns \code{NULL}), \code{str_value()} returns the
string, mirroring how \code{\link[=format]{format()}} returns what \code{\link[=print]{print()}} displays in the
full, multi-line tier. The \code{\link[utils:str]{utils::str()}} methods are thin wrappers that
display \code{str_value()}.
}
\details{
A scalar object (a \code{block}, \code{stack}, \code{link}, \code{board_option} or \code{plugin})
renders as a single line; a container (\code{blocks}, \code{stacks}, \code{links},
\code{board_options}, \code{plugins}) and a whole \code{board} render as one element per
line below a \verb{<class[n]>} header. The \code{block} method lists a block's
constructor inputs, marking the externally controllable ones (those
reported by \code{\link[=external_ctrl_vars]{external_ctrl_vars()}}) with a trailing \code{*}; the \code{stack}
method shows the stack name and its member block ids.

This is the blockr extension point for token-dense renderings such as a
board summary. A home package surfaces a subclass's state by defining a
\code{str_value()} method, typically extending the parent's via \code{\link[=NextMethod]{NextMethod()}}
(the way \code{format.dock_stack()} appends a stack colour). \code{\link[=print]{print()}} and
\code{\link[=format]{format()}} are unaffected and remain the full, multi-line tier.
}
\examples{
str_value(new_dataset_block())

str_value(new_stack(c("plot", "data"), name = "My stack"))

board <- new_board(c(a = new_dataset_block()))
str(board)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack-ui.R
\name{stack_ui}
\alias{stack_ui}
\alias{stack_ui.board}
\alias{stack_ui.stack}
\alias{insert_stack_ui}
\alias{insert_stack_ui.board}
\alias{remove_stack_ui}
\alias{remove_stack_ui.board}
\alias{add_block_to_stack}
\alias{add_block_to_stack.board}
\alias{remove_block_from_stack}
\alias{remove_block_from_stack.board}
\title{Stack UI}
\usage{
stack_ui(id, x, ...)

\method{stack_ui}{board}(id, x, stacks = NULL, edit_ui = NULL, ...)

\method{stack_ui}{stack}(id, x, edit_ui = NULL, ...)

insert_stack_ui(id, x, board, edit_ui = NULL, session = get_session(), ...)

\method{insert_stack_ui}{board}(id, x, board, edit_ui = NULL, session = get_session(), ...)

remove_stack_ui(id, board, session = get_session(), ...)

\method{remove_stack_ui}{board}(id, board, session = get_session(), ...)

add_block_to_stack(board, block_id, stack_id, session = get_session(), ...)

\method{add_block_to_stack}{board}(board, block_id, stack_id, session = get_session(), ...)

remove_block_from_stack(
  board,
  block_id,
  board_id,
  session = get_session(),
  ...
)

\method{remove_block_from_stack}{board}(
  board,
  block_id,
  board_id,
  session = get_session(),
  ...
)
}
\arguments{
\item{id}{Parent namespace}

\item{x}{Object}

\item{...}{Generic consistency}

\item{stacks}{(Additional) stacks (or IDs) for which to generate the UI}

\item{edit_ui}{Stack edit plugin}

\item{board}{Board object}

\item{session}{Shiny session}

\item{block_id, stack_id, board_id}{Block/stack/board IDs}
}
\value{
UI set up via \code{stack_ui()} is expected to return \code{\link[shiny:tag]{shiny::tag()}} or
\code{\link[shiny:tagList]{shiny::tagList()}} objects while stack/block insertion/removal functions
(into/from board/stack objects) are called for their side-effects. Both
\code{insert_stack_ui()}/\code{remove_stack_ui} and
\code{add_block_to_stack()}/\code{remove_block_from_stack()} return \code{NULL} invisibly
and where the former call \code{\link[shiny:insertUI]{shiny::insertUI()}}/\code{\link[shiny:removeUI]{shiny::removeUI()}} and the
latter modify the DOM via \link[shiny:session]{shiny::session} custom messages.
}
\description{
Several generics are exported in order to integrate stack UI into board UI.
We have \code{stack_ui()} which is dispatched on the \code{board} (and in the default
implementation) on individual \code{stack} objects. This renders stacks as
bootstrap accordion items (using \code{\link[bslib:accordion]{bslib::accordion()}}). If a different way
of displaying stacks and integrating them with a board is desired, this can
be implemented by introducing a board subclass and providing a \code{stack_ui()}
method for that subclass. Inserting stacks into (and removing stacks from)
a board is available as \code{insert_stack_ui()}/\code{remove_stack_ui()} and blocks
into/from stacks via \code{add_block_to_stack()}/\code{remove_block_from_stack()}.
All are S3 generics with implementations for \code{board} and alternative
implementation may be provided for board sub-classes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-ui.R, R/board-ui.R
\name{block_ui}
\alias{block_ui}
\alias{expr_ui}
\alias{block_output}
\alias{block_ui.board}
\title{Block UI}
\usage{
block_ui(id, x, ...)

expr_ui(id, x, ...)

block_output(x, result, session)

\method{block_ui}{board}(id, x, blocks = NULL, edit_ui = NULL, ctrl_ui = NULL, ...)
}
\arguments{
\item{id}{Namespace ID}

\item{x}{Object for which to generate a UI container}

\item{...}{Generic consistency}

\item{result}{Block result}

\item{session}{Shiny session object}

\item{blocks}{(Additional) blocks (or IDs) for which to generate the UI}

\item{edit_ui, ctrl_ui}{Block plugin UI}
}
\value{
Both \code{expr_ui()} and \code{block_ui()} are expected to return shiny UI
(e.g. objects wrapped in a \code{\link[shiny:tagList]{shiny::tagList()}}). For rendering the UI,
\code{block_output()} is required to return the result of a shiny render function.
For example, a transform block might show the resulting \code{data.frame} as an
HTML table using the DT package. The corresponding \code{block_ui()} function
would then contain UI created by \code{\link[DT:dataTableOutput]{DT::dataTableOutput()}} and rendering in
\code{block_output()} would then be handled by \code{\link[DT:renderDT]{DT::renderDT()}}.
}
\description{
The UI associated with a block is created via the generics \code{expr_ui()} and
\code{block_ui()}. The former is mainly responsible for user inputs that are
specific to every block type (i.e. a \code{subset_block} requires different user
inputs compared to a \code{head_block}, see \code{\link[=new_transform_block]{new_transform_block()}}) and
essentially calls the UI function passed as \code{ui} to \code{\link[=new_block]{new_block()}}. UI that
represents block outputs typically is shared among similar block types (i.e.
blocks with shared inheritance structure, such as \code{subset_block} and
\code{head_block}, which both inherit from \code{transform_block}). This type of UI us
created by \code{block_ui()} and block inheritance is used to deduplicate shared
functionality (i.e. by implementing a method for the \code{transform_block} class
only instead of separate methods for \code{subset_block} and \code{head_block}.
Working in tandem with \code{block_ui()}, the generic \code{block_output()} generates
the output to be displayed by the UI portion defined via \code{block_ui()}.
}
\details{
The result of \code{block_output()}, which is evaluated in the \code{\link[=block_server]{block_server()}}
context is assigned to \code{output$result}. Consequently, when referencing
the block result in \code{block_ui()}, this naming convention has to be followed
by referring to this as something like \code{shiny::NS(id, "result")}.
}
\section{Board-level block UI}{

While the contents of block-level UI are created by dispatching \code{block_ui()}
on blocks another dispatch on \code{board} (see \code{\link[=new_board]{new_board()}}) occurs as well.
This can be used to control how blocks are integrated into the board UI. For
the default board, this uses \code{\link[bslib:card]{bslib::card()}} to represent blocks. For boards
that extend the default \code{board} class, control is available for how blocks
are displayed by providing a board-specific \code{block_ui()} method.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-misc.R
\name{include_mermaid}
\alias{include_mermaid}
\title{Include a pre-rendered mermaid diagram}
\usage{
include_mermaid(name, chromium_args = c("--no-sandbox"))
}
\arguments{
\item{name}{Diagram name (without extension), resolved relative to a
\verb{mermaid/} directory alongside the calling document.}

\item{chromium_args}{Character vector of extra Chromium flags passed to
\code{mmdc} via a temporary puppeteer config file.}
}
\description{
Renders mermaid \code{.mmd} source to \code{.svg} when the source is newer than the
existing SVG and \code{mmdc} is available, then includes the SVG via
\code{\link[knitr:include_graphics]{knitr::include_graphics()}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stack-class.R, R/stacks-class.R
\name{new_stack}
\alias{new_stack}
\alias{default_stack_name}
\alias{is_stack}
\alias{stack_blocks}
\alias{stack_blocks<-}
\alias{stack_name}
\alias{stack_name<-}
\alias{update_stack}
\alias{validate_stack}
\alias{as_stack}
\alias{stacks}
\alias{is_stacks}
\alias{as_stacks}
\title{Stacks}
\usage{
new_stack(
  blocks = character(),
  name = default_stack_name,
  ...,
  ctor = "new_stack",
  pkg = pkg_name(),
  class = character()
)

default_stack_name()

is_stack(x)

stack_blocks(x)

stack_blocks(x) <- value

stack_name(x, name)

stack_name(x) <- value

update_stack(x, delta)

validate_stack(x)

as_stack(x)

stacks(...)

is_stacks(x)

as_stacks(x, ...)
}
\arguments{
\item{blocks}{Set of blocks}

\item{name}{Stack name}

\item{...}{Extensibility}

\item{ctor, pkg}{Constructor information (used for serialization)}

\item{class}{(Optional) stack sub-class}

\item{x}{Stack object}

\item{value}{Replacement value}

\item{delta}{A named list of constructor argument values to apply on
top of \code{x}'s current fields.}
}
\value{
Construction and coercion via \code{new_stack()}/\code{as_stack()} and
\code{stacks()}/\code{as_stacks()} results in \code{stack} and \code{stacks} objects,
respectively, while inheritance testing via \code{is_stack()} and \code{is_stacks()}
returns scalar logicals. Attribute getters \code{stack_name()} and
\code{stack_blocks()} return scalar and vector-valued character vectors while
setters \verb{stack_name()<-} and \verb{stack_blocks()<-} return modified stack
objects.
}
\description{
Multiple (related) blocks can be grouped together into stacks. Such a
grouping has no functional implications, rather it is an organizational tool
to help users manage more complex pipelines. Stack objects constitute a set
of attributes, the most important of which is \code{blocks} (a character vector
of block IDs). Each \code{stack} may have an arbitrary \code{name} and the class can
be extended by adding further attributes, maybe something like \code{color},
coupled with sub-classing.

Stack container objects (\code{stacks} objects) can be created with \code{stacks()}
or \code{as_stacks()} and inheritance can be tested via \code{is_stacks()}. Further
basic operations such as concatenation, subsetting and sub-assignments is
available by means of base R generics.
}
\details{
Individual stacks can be created using \code{new_stack()} or \code{as_stack()} and
inheritance can be tested with \code{is_stack()}. Attributes can be retrieved
(and modified) with \code{stack_blocks()}/\verb{stack_blocks<-()} and
\code{stack_name()}/\verb{stack_name<-()}, while validation is available as
(generic) \code{validate_stack()}.
}
\section{Partial-arg updates}{

\code{update_stack()} is an S3 generic that produces a modified stack by
merging a named-list \code{delta} of constructor argument values onto an
existing stack. The default \code{.stack} method reconstructs the object
via its stored constructor, treating the underlying character vector
(block IDs) as \code{blocks} and every other attribute as a named
constructor argument. Sub-class owners (e.g. \code{dock_stack} adding
\code{color}) only need to register a method when their constructor
deviates from this convention. The reserved delta key \code{blocks}
replaces the member block IDs; every other key updates the named
attribute.
}

\examples{
stk <- new_stack(letters[1:5], "Alphabet 1")

stack_blocks(stk)
stack_name(stk)
stack_name(stk) <- "Alphabet start"

stks <- c(start = stk, cont = new_stack(letters[6:10], "Alphabet cont."))
names(stks)

tryCatch(
  stack_blocks(stks[[2]]) <- letters[4:8],
  error = function(e) conditionMessage(e)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-shiny.R
\name{destroy_module}
\alias{destroy_module}
\title{Destroy a Shiny module}
\usage{
destroy_module(
  id,
  what = c("inputs", "outputs", "observers"),
  session = get_session()
)
}
\arguments{
\item{id}{Module namespace id.}

\item{what}{Character vector indicating which components to destroy.
Defaults to all of \code{"inputs"}, \code{"outputs"} and \code{"observers"}.}

\item{session}{Shiny session object.}
}
\value{
The namespaced id (invisibly).
}
\description{
Cleans up inputs, outputs and/or observers associated with a Shiny module.
}
\note{
This function relies on non-documented, internal Shiny APIs that are
not officially supported or recommended to use. These internals may break
if Shiny's internal source code changes. Use with caution and only
whenever necessary.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-ply.R
\name{chr_ply}
\alias{chr_ply}
\alias{lgl_ply}
\alias{int_ply}
\alias{dbl_ply}
\alias{chr_mply}
\alias{lgl_mply}
\alias{int_mply}
\alias{dbl_mply}
\alias{chr_xtr}
\alias{lgl_xtr}
\alias{int_xtr}
\alias{dbl_xtr}
\alias{lst_xtr}
\alias{map}
\title{Functional programming utilities}
\usage{
chr_ply(x, fun, ..., length = 1L, use_names = FALSE)

lgl_ply(x, fun, ..., length = 1L, use_names = FALSE)

int_ply(x, fun, ..., length = 1L, use_names = FALSE)

dbl_ply(x, fun, ..., length = 1L, use_names = FALSE)

chr_mply(..., length = 1L)

lgl_mply(..., length = 1L)

int_mply(..., length = 1L)

dbl_mply(..., length = 1L)

chr_xtr(x, i, ...)

lgl_xtr(x, i, ...)

int_xtr(x, i, ...)

dbl_xtr(x, i, ...)

lst_xtr(x, ...)

map(fun, ..., use_names = FALSE)
}
\arguments{
\item{x}{Object to iterate over}

\item{fun}{Function to apply to each component}

\item{...}{Forwarded to \code{fun}}

\item{length}{Expected result length}

\item{use_names}{Name the result using \code{names(x)}}

\item{i}{Index to extract}
}
\value{
The result of a call to \code{\link[base:lapply]{base::lapply()}}, \code{\link[base:vapply]{base::vapply()}} or
\code{\link[base:Map]{base::Map()}}.
}
\description{
A set of wrappers for \code{\link[base:vapply]{base::vapply()}} with some convenient defaults.
Intended mainly for internal use, but available for dependent packages
wherever this is deemed convenient.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-block.R
\name{edit_block}
\alias{edit_block}
\alias{edit_block_server}
\alias{edit_block_ui}
\alias{block_summary}
\alias{block_summary.block}
\title{Plugin module for editing board blocks}
\usage{
edit_block(
  server = edit_block_server,
  ui = edit_block_ui,
  validator = abort_not_null
)

edit_block_server(id, block_id, board, update, ...)

edit_block_ui(x, id, ...)

block_summary(x, data)

\method{block_summary}{block}(x, data)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{validator}{Validator function that validates server return values}

\item{id}{Namespace ID}

\item{block_id}{Block ID}

\item{board}{Reactive values object containing board information}

\item{update}{Reactive value object to initiate board updates}

\item{...}{Extra arguments passed from parent scope}

\item{x}{Block}

\item{data}{Result data}
}
\value{
A plugin container inheriting from \code{edit_block} is returned by
\code{edit_block()}, while the UI component (e.g. \code{edit_block_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{edit_block_server()}) is expected to return \code{NULL}.
}
\description{
Logic and user experience for editing block attributes such as block titles
can be customized or enhanced by providing an alternate version of this
plugin. The default implementation only handles block titles, but if further
(editable) block attributes are to be introduced, corresponding UI and logic
can be included here. In addition to blocks titles, this default
implementation provides UI for removing, as well as inserting blocks before
or after the current one.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-control.R
\name{ctrl_block}
\alias{ctrl_block}
\alias{ctrl_block_server}
\alias{ctrl_block_ui}
\title{Plugin module for external control of block inputs}
\usage{
ctrl_block(server = ctrl_block_server, ui = ctrl_block_ui)

ctrl_block_server(id, x, vars, data, eval)

ctrl_block_ui(id, x)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{x}{Block}

\item{vars}{Reactive state values (list of \code{reactiveVal} objects keyed by
input name). \code{block_name} is included by the default \code{block_server.block}
even though it is not part of the block's \code{expr_server} state.}

\item{data}{Input data paseed as list of reactive values}

\item{eval}{Reactive that evaluates the block expression against input
data. May be used to validate that the new values produce a successful
evaluation.}
}
\value{
A plugin container inheriting from \code{ctrl_block} is returned by
\code{ctrl_block()}, while the UI component (i.e. \code{ctrl_block_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{ctrl_block_server()}) is expected to return a value
that passes validation (i.e. \code{TRUE} or a reactive gate).
}
\description{
This plugin enables setting block reactive state values from outside the
block expression server context. Blocks opt in to external control via the
\code{external_ctrl} argument to \code{\link[=new_block]{new_block()}}, which can be set to \code{TRUE} (all
constructor inputs) or a character vector of specific input names. The
default UI renders a \code{\link[shiny:textInput]{shiny::textInput()}} for each externally controllable
input along with a submit \code{\link[shiny:actionButton]{shiny::actionButton()}}. Both the server and UI
can be replaced with custom implementations by passing alternate functions
to \code{ctrl_block()}.
}
\details{
The default server validates submitted values by evaluating the block
expression (via the \code{eval} reactive) after updating state. On success,
a reactive gate is returned as \code{TRUE}, allowing downstream evaluation to
proceed. On failure, state values are reverted to their previous values,
the user is notified, and the gate is set to \code{FALSE}, which blocks
downstream evaluation until a subsequent successful submit.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-args.R
\name{new_block_arg}
\alias{new_block_arg}
\alias{new_block_args}
\alias{block_arg_description}
\alias{block_arg_example}
\alias{block_arg_type}
\alias{arg_string}
\alias{arg_number}
\alias{arg_integer}
\alias{arg_boolean}
\alias{arg_enum}
\alias{arg_array}
\alias{arg_object}
\title{Block argument specification}
\usage{
new_block_arg(description = NULL, example = NULL, type = NULL)

new_block_args(...)

block_arg_description(x, ...)

block_arg_example(x, ...)

block_arg_type(x, ...)

arg_string(description = NULL, required = TRUE)

arg_number(description = NULL, required = TRUE)

arg_integer(description = NULL, required = TRUE)

arg_boolean(description = NULL, required = TRUE)

arg_enum(values, description = NULL, required = TRUE)

arg_array(items, description = NULL, required = TRUE)

arg_object(..., description = NULL, required = TRUE)
}
\arguments{
\item{description}{Human- and model-facing description of an argument value}

\item{example}{A single worked value for an argument (or \code{NULL})}

\item{type}{Optional machine-readable type for the argument, built with the
\verb{arg_*()} descriptor constructors. A plain JSON-Schema-subset list; worked
examples are validated against it}

\item{...}{For \code{new_block_args()}, the per-argument \code{block_arg} objects (or
bare description strings); for \code{arg_object()}, the named field descriptors;
ignored by the \verb{block_arg_*()} getters}

\item{x}{A \code{block_arg} object, or a bare string taken as its description}

\item{required}{Whether the field is required, when nested in an
\code{arg_object()}}

\item{values}{Allowed string values, for \code{arg_enum()}}

\item{items}{Element descriptor, for \code{arg_array()}}
}
\value{
\code{new_block_arg()} returns a \code{block_arg} and \code{new_block_args()} a \code{block_args}
collection. The \verb{arg_*()} constructors each return a plain JSON-Schema node
(a list). The \verb{block_arg_*()} getters return the corresponding field of a
single argument (resolving a bare description string too).
}
\description{
Block constructor arguments can be documented with a structured
specification: each argument via \code{new_block_arg()} (a \code{description}, a single
worked \code{example}, and an optional machine-readable \code{type}), collected with
\code{new_block_args()}. A bare named character vector of descriptions, and the
empty \code{character()}, are also accepted and normalized into this form, so
existing registrations are unaffected. A single argument's fields are read
back with \code{block_arg_description()}, \code{block_arg_example()} and
\code{block_arg_type()}; block-level metadata (the whole argument set, worked
examples, guidance and keywords) is tabulated across blocks via
\code{\link[=block_metadata]{block_metadata()}}.
}
\details{
An argument's \code{type} is described with a small, dependency-free subset of
JSON Schema, built with the \verb{arg_*()} constructors: \code{arg_string()},
\code{arg_number()}, \code{arg_integer()} and \code{arg_boolean()} for scalars, \code{arg_enum()}
for a fixed set of string values, \code{arg_array()} for a homogeneous list and
\code{arg_object()} for a closed record of named fields (\code{additionalProperties: false}). Each returns a plain nested list mirroring the schema it denotes,
consumed directly -- blockr.ai binds it via \code{ellmer::type_from_schema()}, an
MCP or raw tool schema reads the JSON as-is -- and worked examples registered
alongside an argument are validated against it (see \code{\link[=register_block]{register_block()}}).
Semantic intent (e.g. "a column in the upstream data", "an R expression") is
carried in \code{description}, not in the type vocabulary.

The complete worked configuration of a block is the assembly of its
per-argument examples, keyed by argument name. When arguments interact, or
several few-shot examples are wanted, complete configurations are instead
supplied as a list via the \code{examples} argument of \code{\link[=register_block]{register_block()}} and
supersede that assembly; combining multiple per-argument examples is
intentionally not supported, as there is no safe way to form coherent
whole-block configurations from them.
}
\examples{
new_block_args(
  n = new_block_arg(
    "Number of rows to return",
    example = 5L,
    type = arg_integer()
  )
)

arg_object(
  conditions = arg_array(
    arg_object(column = arg_string(), value = arg_string())
  ),
  operator = arg_enum(c("&", "|"))
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-tests.R
\name{generate_plugin_args}
\alias{generate_plugin_args}
\alias{sink_msg}
\alias{new_mock_session}
\alias{with_mock_session}
\alias{with_mock_context}
\alias{get_s3_method}
\alias{export_safely}
\title{Testing utilities}
\usage{
generate_plugin_args(board, ..., mode = c("edit", "read"))

sink_msg(...)

new_mock_session()

with_mock_session(expr, session = new_mock_session())

with_mock_context(session, expr)

get_s3_method(generic, object)

export_safely(x)
}
\arguments{
\item{board}{A board object}

\item{...}{Forwarded to \code{\link[utils:capture.output]{utils::capture.output()}}}

\item{mode}{Edit plugins, such as \code{manage_blocks} get an additional argument
\code{update} over read plugins such as \code{preserve_board}.}

\item{expr}{Test code containing expectations. The objects from inside the
server function environment will be made available in the environment of
the test expression (this is done using a data mask with
\code{\link[rlang:eval_tidy]{rlang::eval_tidy()}}). This includes the parameters of the server function
(e.g. \code{input}, \code{output}, and \code{session}), along with any other values
created inside of the server function.}

\item{session}{The \code{\link[shiny]{MockShinySession}} object to use as the \link[shiny:domains]{reactive domain}. The same session object is used as the domain both
during invocation of the server or module under test and during evaluation
of \code{expr}.}

\item{generic}{Generic function name (passed as string)}

\item{object}{S3 Object}

\item{x}{Reactive object to use in \code{\link[shiny:exportTestValues]{shiny::exportTestValues()}}}
}
\value{
For testing plugins, \code{generate_plugin_args()} returns objects that
mimic how plugins are called in the board server (with reactive components
re-created on a \code{NULL} reactive domain so they outlive the
\code{\link[shiny:testServer]{shiny::testServer()}} session used to set them up), \code{sink_msg()} is called
mainly for the side-effect of muting shiny messages (and returns them
invisibly), \code{with_mock_session()} returns \code{NULL} (invisibly) and
\code{with_mock_context()} returns the result of a call to
\code{\link[shiny:withReactiveDomain]{shiny::withReactiveDomain()}}. Finally, \code{get_s3_method()} returns a
class-specific implementation of the specified generic (and throws an error
if none is found).
}
\description{
Several utilities for unit testing, mainly with \code{\link[shiny:testServer]{shiny::testServer()}} that
have proven themselves useful for testing this package are exported for
re-use in other packages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-shiny.R
\name{trim_rv}
\alias{trim_rv}
\title{Remove entries from a \code{reactiveValues} object}
\usage{
trim_rv(x, rm)
}
\arguments{
\item{x}{A \code{reactiveValues} object.}

\item{rm}{Character vector of keys to remove; all must be present in \code{x}.}
}
\value{
\code{x}, invisibly.
}
\description{
shiny offers no public way to delete a key from a \code{\link[shiny:reactiveValues]{shiny::reactiveValues()}}
object -- assigning \code{NULL} stores a \code{NULL} value but leaves the key in
\code{names()}. \code{trim_rv()} removes the named entries outright and invalidates
the affected reactive dependencies, so a key can be truly dropped (and
later re-added) -- for instance when a variadic block argument is unlinked.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-class.R, R/blocks-class.R
\name{new_block}
\alias{new_block}
\alias{default_block_name}
\alias{is_block}
\alias{as_block}
\alias{blocks}
\alias{is_blocks}
\alias{as_blocks}
\title{Blocks}
\usage{
new_block(
  server,
  ui,
  class,
  ctor = sys.parent(),
  ctor_pkg = NULL,
  dat_valid = NULL,
  allow_empty_state = FALSE,
  block_name = default_block_name,
  expr_type = c("quoted", "bquoted"),
  external_ctrl = FALSE,
  block_metadata = NULL,
  ...
)

default_block_name(class)

is_block(x)

as_block(x, ...)

blocks(...)

is_blocks(x)

as_blocks(x, ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{ctor_pkg}{String-valued package name when passing a string-valued
constructor name or \code{NULL}}

\item{dat_valid}{(Optional) input data validator}

\item{allow_empty_state}{Relaxes the block's readiness requirements. By
default every data input must be connected and a variadic block needs at
least one \code{...args} input. As \code{TRUE}, \code{FALSE} or a character vector of
\code{state} names it relaxes \emph{user} inputs (which \code{state} values may be empty).
A \code{list(input = ..., data = ...)} also relaxes \emph{data} inputs: \code{input} takes
the same \code{TRUE}/\code{FALSE}/character form, while \code{data} names non-variadic data
inputs that may stay unconnected and, via a \code{...args} entry, overrides the
required number of variadic inputs, e.g.
\code{list(input = "n", data = list("y", ...args = 2))}}

\item{block_name}{Block name}

\item{expr_type}{Expression type (experimental)}

\item{external_ctrl}{Set up external control (experimental)}

\item{block_metadata}{Block metadata}

\item{...}{Further (metadata) attributes}

\item{x}{An object inheriting from \code{"block"}}
}
\value{
Both \code{new_block()} and \code{as_block()} return an object inheriting from
\code{block}, while \code{is_block()} returns a boolean indicating whether an object
inherits from \code{block} or not. Block vectors, created using \code{blocks()},
\code{as_blocks()}, or by combining multiple blocks using \code{\link[base:c]{base::c()}} all inherit
from \code{blocks} and \code{is_blocks()} returns a boolean indicating whether an
object
inherits from \code{blocks} or not.
}
\description{
Steps in a data analysis pipeline are represented by blocks. Each block
combines data input with user inputs to produce an output. In order to
create a block, which is implemented as a shiny module, we require a server
function, a function that produces some UI and a class vector.
}
\details{
A block constructor may have arguments, which taken together define the
block state. It is good practice to expose all user-selectable arguments of
a block (i.e. everything excluding the "data" input) as block arguments such
that block can be fully initialized via the constructor. Some default values
are required such that blocks can be constructed via constructor calls
without arguments. Where it is sensible to do so, specific default values
are acceptable, but if in any way data dependent, defaults should map to
an "empty" input. For example, a block that provides \code{\link[utils:head]{utils::head()}}
functionality, one such argument could be \code{n} and a reasonable default value
could be \code{6L} (in line with corresponding default S3 method implementation).
On the other hand, a block that performs a \code{\link[base:merge]{base::merge()}} operation might
expose a \code{by} argument, but a general purpose default value (that does not
depend on the data) is not possible. Therefore, \code{\link[=new_merge_block]{new_merge_block()}} has
\code{by = character()}.

The return value of a block constructor should be the result of a call to
\code{new_block()} and \code{...} should be contained in the constructor signature
such that general block arguments (e.g. \code{name}) are available from the
constructor.
}
\section{Server}{

The server function (passed as \code{server}) is expected to be a function that
returns a \code{\link[shiny:moduleServer]{shiny::moduleServer()}}. This function is expected to have at
least an argument \code{id} (string-valued), which will be used as the module ID.
Further arguments may be used in the function signature, one for each "data"
input. A block implementing \code{\link[utils:head]{utils::head()}} for example could have a single
extra argument \code{data}, while a block that performs \code{\link[base:merge]{base::merge()}} requires
two extra arguments, e.g. \code{x} and \code{y}. Finally, a variadic block, e.g.
a block implementing something like \code{\link[base:rbind]{base::rbind()}}, needs to accommodate for
an arbitrary number of inputs. This is achieved by passing a container
object as \code{...args} and thus such a variadic block needs \code{...args} as part
of the server function signature. The container mirrors
\code{\link[shiny:reactiveValues]{shiny::reactiveValues()}} for reads (\code{names()}, \code{[[}, \code{$} and \code{as.list()}),
and accessing a slot yields the value of its (lazily evaluated) input. The
fixed per-data input arguments are instead passed as \code{\link[shiny:reactive]{shiny::reactive()}} or
\code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} objects.

The server function may implement arbitrary shiny logic and is expected to
return a list with components \code{expr} and \code{state}. The expression corresponds
to the R code necessary to perform the block task and is expected to be
a reactive quoted expression. It should contain user-chosen values for all
user inputs and placeholders for all data inputs (using the same names for
data inputs as in the server function signature). Such an expression for a
\code{\link[base:merge]{base::merge()}} block could be created using \code{\link[base:bquote]{base::bquote()}} as

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bquote(
  merge(x, y, by = .(cols)),
  list(cols = current_val())
\}
}\if{html}{\out{</div>}}

where \code{current_val()} is a reactive that evaluates to the current user
selection of the \code{by} columns. This should then be wrapped in a
\code{\link[shiny:reactive]{shiny::reactive()}} call such that \code{current_val()} can be evaluated whenever
the current expression is required.

The \code{state} component is expected to be a named list with either reactive or
"static" values. In most cases, components of \code{state} will be reactives,
but it might make sense in some scenarios to have constructor arguments that
are not exposed via UI components but are fixed at construction time. An
example for this could be the \code{dataset_block} implementation where we have
constructor arguments \code{dataset} and \code{package}, but only expose \code{dataset}
as UI element. This means that \code{package} is fixed at construction time.
Nevertheless, \code{package} is required as state component, as this is used for
re-creating blocks from saved state.

State component names are required to match block constructor arguments and
re-creating saved objects basically calls the block constructor with values
obtained from block state.
}

\section{UI}{

Block UI is generated using the function passed as \code{ui} to the \code{new_block}
constructor. This function is required to take a single argument \code{id} and
shiny UI components have to be namespaced such that they are nested within
this ID (i.e. by creating IDs as \code{shiny::NS(id, "some_value")}). Some care
has to be taken to properly initialize inputs with constructor values. As a
rule of thumb, input elements exposed to the UI should have corresponding
block constructor arguments such that blocks can be created with a given
initial state.

Block UI should be limited to displaying and arranging user inputs to set
block arguments. For outputs, use generics \code{\link[=block_output]{block_output()}} and
\code{\link[=block_ui]{block_ui()}}.
}

\section{Sub-classing}{

In addition to the specific class of a block, the core package uses virtual
classes to group together blocks with similar behavior (e.g.
\code{transform_block}) and makes use of this inheritance structure in S3
dispatch for methods like \code{\link[=block_output]{block_output()}} and \code{\link[=block_ui]{block_ui()}}. This pattern is
not required but encouraged.
}

\section{Initialization/evaluation}{

Some control over when a block is considered "ready for evaluation" is
available via arguments \code{dat_valid} and \code{allow_empty_state}. Data input
validation can optionally be performed by passing a predicate function with
the same arguments as in the server function (not including \code{id}) and the
block expression will not be evaluated as long as this function throws an
error.

Ahead of validation, a block must have its inputs available: every required
\emph{data} input connected to a ready upstream (variadic blocks need at least the
\code{...args} minimum declared via \code{allow_empty_state}) and every required \emph{user}
input (\code{state}) provided. Until then -- including while an upstream is itself
still pending -- neither the validator nor the block expression run and no
output is produced. See \code{\link[=block_server]{block_server()}} for the resulting eval status
(\code{dormant} / \code{waiting} / \code{unset} / \code{failed} / \code{ready}).

Other conditions (messages and warnings) may be thrown as will be caught
and displayed to the user but they will not interrupt evaluation. Errors
are safe in that they will be caught as well but the will interrupt
evaluation as long as block data input does not satisfy validation.
}

\section{Block vectors}{

Multiple blocks can be combined into a \code{blocks} object, a container for
an (ordered) set of blocks. Block IDs are handled at the \code{blocks} level
which will ensure uniqueness.
}

\examples{
new_identity_block <- function() {
  new_transform_block(
    function(id, data) {
      moduleServer(
        id,
        function(input, output, session) {
          list(
            expr = reactive(quote(identity(data))),
            state = list()
          )
        }
      )
    },
    function(id) {
      htmltools::tagList()
    },
    class = "identity_block"
  )
}

blk <- new_identity_block()
is_block(blk)

blks <- c(a = new_dataset_block(), b = new_subset_block())

is_block(blks)
is_blocks(blks)

names(blks)

tryCatch(
  names(blks["a"]) <- "b",
  error = function(e) conditionMessage(e)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file-block.R, R/file-browser.R,
%   R/file-upload.R
\name{new_file_block}
\alias{new_file_block}
\alias{new_filebrowser_block}
\alias{filebrowser_volumes}
\alias{new_upload_block}
\title{File block constructors}
\usage{
new_file_block(server, ui, class, ctor = sys.parent(), ...)

new_filebrowser_block(
  file_path = character(),
  volumes = filebrowser_volumes(),
  ...
)

filebrowser_volumes(default = c(home = path.expand("~")))

new_upload_block(...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{...}{Forwarded to \code{new_file_block()} and \code{\link[=new_block]{new_block()}}}

\item{file_path}{File path}

\item{volumes}{Parent namespace}

\item{default}{Default volumes specification (use the blockr option
"volumes" to override)}
}
\value{
All blocks constructed via \code{new_file_block()} inherit from
\code{file_block}.
}
\description{
Similarly to \code{\link[=new_data_block]{new_data_block()}}, blocks created via \code{new_file_block()} serve
as starting points in analysis pipelines by providing data to down-stream
blocks. They typically will not have data inputs and represent root nodes in
analysis graphs.
}
\section{File browser block}{

In order to make user data available to blockr, this block provides file-
upload functionality via \code{\link[shiny:fileInput]{shiny::fileInput()}}. Given that data provided in
this way are only available for the life-time of the shiny session, exported
code is not self-contained and a script containing code from an upload block
is cannot be run in a new session. Also, serialization of upload blocks is
currently not allowed as the full data would have to be included during
serialization.
}

\section{Upload block}{

In order to make user data available to blockr, this block provides file-
upload functionality via \code{\link[shiny:fileInput]{shiny::fileInput()}}. Given that data provided in
this way are only available for the life-time of the shiny session, exported
code is not self-contained and a script containing code from an upload block
is cannot be run in a new session. Also, serialization of upload blocks is
currently not allowed as the full data would have to be included during
serialization.
}

\seealso{
The \href{https://bristolmyerssquibb.github.io/blockr.io/}{blockr.io}
package provides real-world blocks for sourcing external files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-notification.R
\name{notify_user}
\alias{notify_user}
\alias{notify_user_server}
\alias{notify_user_ui}
\title{User notification plugin module}
\usage{
notify_user(server = notify_user_server, ui = notify_user_ui)

notify_user_server(id, board, ...)

notify_user_ui(id, board)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{Reactive values object}

\item{...}{Extra arguments passed from parent scope}
}
\value{
A plugin container inheriting from \code{notify_user} is returned by
\code{notify_user()} and the UI component (e.g. \code{notify_user_ui()}) is expected
to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}). The server component (i.e.
\code{notify_user_server()}) is called for the side effect of managing
notifications and returns \code{NULL}.
}
\description{
During the evaluation cycle of each block, conditions (errors, warnings and
messages) may be raised. The default \code{notify_user} plugin surfaces these to
the user as \code{\link[shiny:showNotification]{shiny::showNotification()}} toasts, displaying newly active
conditions and clearing ones that are no longer active via
\code{\link[shiny:removeNotification]{shiny::removeNotification()}}. Each block's conditions (see \code{\link[=board_server]{board_server()}})
are tracked individually, so that a single block's change touches only its
own notifications rather than re-processing the whole board.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/block-registry.R
\name{register_block}
\alias{register_block}
\alias{default_icon}
\alias{default_category}
\alias{suggested_categories}
\alias{list_blocks}
\alias{registry_id_from_block}
\alias{unregister_blocks}
\alias{register_blocks}
\alias{available_blocks}
\alias{registry_metadata}
\alias{create_block}
\alias{register_package_blocks}
\title{Block registry}
\usage{
register_block(
  ctor,
  name,
  description,
  classes = NULL,
  uid = NULL,
  category = NULL,
  icon = NULL,
  arguments = NULL,
  details = NULL,
  link = NULL,
  guidance = NULL,
  examples = list(),
  keywords = NULL,
  package = NULL,
  overwrite = FALSE
)

default_icon(category = default_category())

default_category()

suggested_categories()

list_blocks()

registry_id_from_block(block)

unregister_blocks(uid = list_blocks())

register_blocks(...)

available_blocks()

registry_metadata(blocks = list_blocks(), fields = "all")

create_block(id, ...)

register_package_blocks(which = "all", package = pkg_name(), overwrite = TRUE)
}
\arguments{
\item{ctor}{Block constructor}

\item{name, description}{Metadata describing the block}

\item{classes}{Block classes}

\item{uid}{Unique ID for a registry entry}

\item{category}{Useful to sort blocks by topics. If not specified,
blocks are uncategorized.}

\item{icon}{Icon}

\item{arguments}{Block argument specification, either a \code{\link[=new_block_args]{new_block_args()}}
object or a (possibly empty) named character vector of argument
descriptions}

\item{details}{Optional longer human-facing description (e.g. for a help
popover), complementing the short \code{description}}

\item{link}{Optional URL to a block's help or documentation page}

\item{guidance}{Optional model-facing construction guidance (do/don't rules,
enums, pitfalls), distinct from the human-facing \code{details}. Retrieved with
\code{\link[=block_metadata]{block_metadata()}}}

\item{examples}{Optional list of complete worked configurations (each a
named list keyed by argument name); supersedes the per-argument example
assembly. Retrieved with \code{\link[=block_metadata]{block_metadata()}}}

\item{keywords}{Optional character vector of free-text search terms for block
discovery, retrieved with \code{\link[=block_metadata]{block_metadata()}}}

\item{package}{Package where constructor is defined (or \code{NULL})}

\item{overwrite}{Overwrite existing entry}

\item{block}{Block object}

\item{...}{For \code{register_blocks()}, arguments forwarded to
\code{register_block()}; for \code{new_block_args()}, named \code{new_block_arg()} objects
(or strings), one per constructor formal}

\item{blocks}{Character vector of registry IDs}

\item{fields}{Metadata fields}

\item{id}{Block ID as reported by \code{list_blocks()}}

\item{which}{Character vector of block constructor names to register (or
\code{"all"}), keying into the YAML registry generated by the
\code{\link[=block_registration_roclet]{block_registration_roclet()}}}
}
\value{
\code{register_block()} and \code{register_blocks()} are invoked for their side
effects and return \code{block_registry_entry} object(s) invisibly, while
\code{unregister_blocks()} returns \code{NULL} (invisibly). \code{register_package_blocks()}
registers the blocks a package declares in its YAML registry (written by the
\code{\link[=block_registration_roclet]{block_registration_roclet()}}), likewise returning their entries invisibly.
Listing via \code{list_blocks()}
returns a character vector and a list of \code{block_registry_entry} object(s) for
\code{available_blocks()}. \code{create_block()} returns a newly instantiated \code{block}
object. A block's registered metadata is read with \code{\link[=block_metadata]{block_metadata()}} and the
\verb{block_meta_*()} accessors, and its argument specification is built and read
with \code{\link[=new_block_arg]{new_block_arg()}} and friends.
}
\description{
Listing of blocks is available via a block registry, which associates a block
constructor with metadata in order to provide a browsable block directory.
Every constructor is identified by a unique ID (uid), which by default is
generated from the class vector (first element). If the class vector is not
provided during registration, an object is instantiated (by calling the
constructor with arguments \code{ctor} and \code{ctor_pkg} only) to derive this
information. Block constructors therefore should be callable without block-
specific arguments.
}
\details{
Due to current requirements for serialization/deserialization, we keep track
the constructor that was used for block instantiation. This works most
reliable whenever a block constructor is an exported function from a package
as this function is guaranteed to be available in a new session (give the
package is installed in an appropriate version). While it is possible to
register a block passing a "local" function as \code{ctor}, this may introduce
failure modes that are less obvious (for example when such a constructor
calls another function that is only defined within the scope of the session).
It is therefore encouraged to only rely on exported function constructors.
These can also be passed as strings and together with the value of \code{package},
the corresponding function can easily be retrieved in any session.

Blocks can be registered (i.e. added to the registry) via \code{register_block()}
with scalar-valued arguments and \code{register_blocks()}, where arguments may be
vector-valued, while de-registration (or removal) is handled via
\code{unregister_blocks()}. A listing of all available blocks can be created as
\code{list_blocks()}, which will return registry IDs and \code{available_blocks()},
which provides a set of (named) \code{block_registry_entry} objects. Finally,
block construction via a registry ID is available as \code{create_block()}.
}
\examples{
blks <- list_blocks()
register_block("new_dataset_block", "Test", "Registry test",
               uid = "test_block", package = "blockr.core")

new <- setdiff(list_blocks(), blks)
unregister_blocks(new)
setequal(list_blocks(), blks)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/board-plugins.R
\name{new_plugin}
\alias{new_plugin}
\alias{is_plugin}
\alias{abort_not_null}
\alias{as_plugin}
\alias{plugin_server}
\alias{plugin_ui}
\alias{plugin_validator}
\alias{plugin_id}
\alias{board_plugins}
\alias{plugins}
\alias{is_plugins}
\alias{as_plugins}
\alias{validate_plugins}
\title{Board plugin}
\usage{
new_plugin(server, ui = NULL, validator = abort_not_null, class = character())

is_plugin(x)

abort_not_null(x)

as_plugin(x)

plugin_server(x)

plugin_ui(x)

plugin_validator(x)

plugin_id(x)

board_plugins(x, ...)

plugins(...)

is_plugins(x)

as_plugins(x)

validate_plugins(x)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{validator}{Validator function that validates server return values}

\item{class}{Plugin subclass}

\item{x}{Plugin object}

\item{...}{Plugin objects}
}
\value{
Constructors \code{new_plugin()}/\code{plugins()} return \code{plugin} and
\code{plugins} objects, respectively, as do \code{as_plugin()}/\code{as_plugins()} and
validators \code{validate_plugin()}/\code{validate_plugins()}, which are typically
called for their side effects of throwing errors in case of validation
failure. Inheritance checkers \code{is_plugin()}/\code{is_plugins()} return scalar
logicals and finally, the convenience function \code{board_plugins()} returns a
\code{plugins} object with all known plugins (or a selected subset thereof).
}
\description{
A core mechanism for extending or customizing UX aspects of the board module
is a "plugin" architecture. All plugins inherit from \code{plugin} and a sub-class
is assigned to each specific plugin. The "manage blocks" plugin for example
has a class vector \code{c("manage_blocks", "plugin")}. Sets of plugins are
handled via a wrapper class \code{plugins}. Each plugin needs a server component,
in most cases accompanied by a UI component and is optionally bundled with a
validator function.
}
\examples{
plg <- board_plugins(new_board())

is_plugins(plg)
names(plg)

plg[1:3]

is_plugin(plg[["preserve_board"]])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-expr.R
\name{bbquote}
\alias{bbquote}
\alias{.}
\alias{..}
\title{Quoting utilities}
\usage{
bbquote(expr, where = parent.frame(), splice = FALSE)

.(x)

..(x)
}
\arguments{
\item{expr}{A \link[base]{language object}.}

\item{where}{An environment.}

\item{splice}{Logical; if \code{TRUE} splicing is enabled.}

\item{x}{Object}
}
\value{
A language object in the same way as returned by \code{\link[base:bquote]{base::bquote()}}.
Functions \code{.()} and \code{..()} throw errors when invoked an only exist to mask
check notes "no visible global function definition" for their use.
}
\description{
Block expressions in blockr are evaluated in a 2-step manner: first in the
context of arguments supplied by the user via UI elements and in a second
step in the context of input data. The function \code{\link[base:bquote]{base::bquote()}} does not
allow for terms wrapped in \code{.()} or \code{..()} to be missing and this makes it
incompatible with this 2-step approach. A drop-in replacement, provided as
\code{bbquote()} addresses this shortcoming.
}
\details{
A block like \code{new_head_block()} is expected to return an expression of the
form \code{utils::head(data, n = 10L)}, which will then be evaluated in an
environment where he have a name \code{data} bound to some dataset. In order to
perform some manipulations of such block expressions it is required to
somehow mark the terms that correspond to input data and for that we can
use the syntax introduced by \code{\link[base:bquote]{base::bquote()}}. What we would prefer to have
as block expression therefore is not the above, but something like
\code{utils::head(.(data), n = 10L)}, as this affords us more possibilities for
performing substitutions (and therefore generates cleaner code).

In order to interpolate certain arguments in a first step, we unfortunately
cannot use \code{\link[base:bquote]{base::bquote()}}, but we can use \code{bbquote()} instead to generate
the desired expression.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{bquote(utils::head(.(data), n = .(n)), list(n = 10L))
#> Error in eval(e[[2L]], where) : object 'data' not found
bbquote(utils::head(.(data), n = .(n)), list(n = 10L))
#> utils::head(.(data), n = 10L)
}\if{html}{\out{</div>}}

This also works with \code{..()} and splicing.
}
\examples{
bbquote(utils::head(.(data), n = .(n)), list(n = 10L))
bbquote(c(.(a), ..(bc)), list(a = "a"))
bbquote(c(.(a), ..(bc)), list(a = "a", bc = c("b", "c")), splice = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parser-block.R, R/parser-csv.R
\name{new_parser_block}
\alias{new_parser_block}
\alias{new_csv_block}
\title{Parser block constructors}
\usage{
new_parser_block(
  server,
  ui,
  class,
  ctor = sys.parent(),
  dat_valid = is_file,
  ...
)

new_csv_block(sep = ",", quote = "\\"", ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{dat_valid}{(Optional) input data validator}

\item{...}{Forwarded to \code{new_parser_block()} and \code{\link[=new_block]{new_block()}}}

\item{sep, quote}{Forwarded to \code{\link[utils:read.table]{utils::read.table()}}}
}
\value{
All blocks constructed via \code{new_parser_block()} inherit from
\code{parser_block}.
}
\description{
Operating on results from blocks created via \code{\link[=new_file_block]{new_file_block()}}, parser
blocks read (i.e. "parse") a file and make the contents available to
subsequent blocks for further analysis and visualization.
}
\details{
If using the default validator for a parser block sub-class (i.e. not
overriding the \code{dat_valid} argument in the call to \code{new_parser_block()}),
the data argument corresponding to the input file name must be \code{file} in
order to match naming conventions in the validator function.
}
\section{CSV block}{

Files in CSV format provided for example by a block created via
\code{\link[=new_file_block]{new_file_block()}} may be parsed into \code{data.frame} by CSV blocks.
}

\seealso{
The \href{https://bristolmyerssquibb.github.io/blockr.io/}{blockr.io}
package provides real-world blocks for parsing external data formats (e.g.
\code{csv}, \code{xpt}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform-block.R, R/transform-fixed.R,
%   R/transform-head.R, R/transform-merge.R, R/transform-rbind.R,
%   R/transform-subset.R
\name{new_transform_block}
\alias{new_transform_block}
\alias{new_fixed_block}
\alias{new_head_block}
\alias{new_merge_block}
\alias{new_rbind_block}
\alias{new_subset_block}
\title{Transform block constructors}
\usage{
new_transform_block(server, ui, class, ctor = sys.parent(), ...)

new_fixed_block(expr, ...)

new_head_block(n = 6L, direction = c("head", "tail"), ...)

new_merge_block(by = character(), all_x = FALSE, all_y = FALSE, ...)

new_rbind_block(...)

new_subset_block(subset = "", select = "", ...)
}
\arguments{
\item{server}{A function returning \code{\link[shiny:moduleServer]{shiny::moduleServer()}}}

\item{ui}{A function with a single argument (\code{ns}) returning a \code{shiny.tag}}

\item{class}{Block subclass}

\item{ctor}{String-valued constructor name or function/frame number (mostly
for internal use or when defining constructors for virtual classes)}

\item{...}{Forwarded to \code{new_transform_block()} and \code{\link[=new_block]{new_block()}}}

\item{expr}{Quoted expression}

\item{n}{Number of rows}

\item{direction}{Either "head" or "tail"}

\item{by}{Column(s) tp join on}

\item{all_x, all_y}{Join type, see \code{\link[base:merge]{base::merge()}}}

\item{subset, select}{Expressions (passed as strings)}
}
\value{
All blocks constructed via \code{new_transform_block()} inherit from
\code{transform_block}.
}
\description{
Many data transformations are be provided by blocks constructed via
\code{new_transform_block()}, including examples where a single \code{data.frame} is
transformed into another (e.g. \code{subset_block}), and two or more \code{data.frame}s
are combined (e.g. \code{merge_block} or \code{rbind_block}).
}
\section{Fixed block}{

Mainly useful for testing and examples, this block applies a fixed
transformation to its data input. No UI elements are exposed and the
transformation consequently cannot be parametrized. The quoted expression
passed as \code{expr} is expected to refer to the input data as \code{data}.
}

\section{Head block}{

Row-subsetting the first or last \code{n} rows of a \code{data.frame} (as provided by
\code{\link[utils:head]{utils::head()}} and \code{\link[utils:tail]{utils::tail()}}) is implemented as \code{head_block}. This is
an example of a block that takes a single \code{data.frame} as input and produces
a single \code{data.frame} as output.
}

\section{Merge block}{

Joining together two \code{data.frame}s, based on a set of index columns, using
\code{\link[base:merge]{base::merge()}} is available as \code{merge_block}. Depending on values passed
as \code{all_x}/\code{all_y} the result will correspond to an "inner", "outer", "left"
or "right" join. See \code{\link[base:merge]{base::merge()}} for details. This block class serves
as an example for a transform block that takes exactly two data inputs \code{x}
and \code{y} to produce a single \code{data.frame} as output.
}

\section{Row-bind block}{

Row-wise concatenation of an arbitrary number of \code{data.frame}s, as performed
by \code{\link[base:rbind]{base::rbind()}} is available as an \code{rbind_block}. This mainly serves as
an example for a variadic block via the "special" \code{...args} block data
argument.
}

\section{Subset block}{

This block allows to perform row and column subsetting on \code{data.frame}
objects via \code{\link[base:subset]{base::subset()}}. Using non-standard evaluation, strings passed
as \code{subset}/\code{select} arguments or entered via shiny UI are turned into
\code{language} objects by \code{\link[base:parse]{base::parse()}}.
}

\seealso{
Real-world transform blocks (e.g. \code{select}, \code{mutate}, \code{filter} and
\code{join}) built on \code{new_transform_block()} are provided by the
\href{https://bristolmyerssquibb.github.io/blockr.dplyr/}{blockr.dplyr} package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-stack.R
\name{edit_stack}
\alias{edit_stack}
\alias{edit_stack_server}
\alias{edit_stack_ui}
\title{Plugin module for editing board stacks}
\usage{
edit_stack(server = edit_stack_server, ui = edit_stack_ui)

edit_stack_server(id, stack_id, board, update, ...)

edit_stack_ui(id, x, ...)
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{stack_id}{Stack ID}

\item{board}{Reactive values object containing board information}

\item{update}{Reactive value object to initiate board updates}

\item{...}{Extra arguments passed from parent scope}

\item{x}{Stack}
}
\value{
A plugin container inheriting from \code{edit_stack} is returned by
\code{edit_stack()}, while the UI component (e.g. \code{edit_stack_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component (i.e. \code{edit_stack_server()}) is expected to return \code{NULL}.
}
\description{
Logic and user experience for editing stack attributes such as stack names
can be customized or enhanced by providing an alternate version of this
plugin. The default implementation only handles stack names, but if further
(editable) stack attributes are to be introduced, corresponding UI and logic
can be included here. In addition to stack names, this default
implementation provides UI for removing the current stack.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-cnd.R
\name{blockr_abort}
\alias{blockr_abort}
\alias{blockr_warn}
\alias{blockr_inform}
\title{Blockr conditions}
\usage{
blockr_abort(..., class = character(), envir = parent.frame())

blockr_warn(
  ...,
  class = character(),
  envir = parent.frame(),
  frequency = "always",
  frequency_id = NULL
)

blockr_inform(
  ...,
  class = character(),
  envir = parent.frame(),
  frequency = "always",
  frequency_id = NULL
)
}
\arguments{
\item{...}{Forwarded to \code{\link[cli:pluralize]{cli::pluralize()}}}

\item{class}{Condition class}

\item{envir}{Forwarded to \code{\link[cli:pluralize]{cli::pluralize()}}}

\item{frequency, frequency_id}{Forwarded to \code{\link[rlang:warn]{rlang::warn()}}}
}
\value{
Called for side-effect of signaling conditions.
}
\description{
Wrappers for \code{\link[rlang:abort]{rlang::abort()}}, \code{\link[rlang:warn]{rlang::warn()}} and \code{\link[rlang:inform]{rlang::inform()}}. In
addition to \code{class}, conditions inherit from "blockr_error".
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-misc.R
\name{rand_names}
\alias{rand_names}
\alias{adjective_animal}
\alias{sample_letters}
\alias{resolve_ctor}
\alias{forward_ctor}
\alias{is_blockr_ctor}
\alias{ctor_name}
\alias{ctor_pkg}
\alias{ctor_fun}
\alias{to_sentence_case}
\alias{id_to_sentence_case}
\title{Random IDs}
\usage{
rand_names(
  old_names = character(0L),
  n = 1L,
  max_tries = 100L,
  id_fun = blockr_option("rand_id", NULL)
)

adjective_animal(n)

sample_letters(n)

resolve_ctor(ctor, ctor_pkg = NULL)

forward_ctor(x)

is_blockr_ctor(x)

ctor_name(x)

ctor_pkg(x)

ctor_fun(x)

to_sentence_case(x, replace = character(), with = character())

id_to_sentence_case(x)
}
\arguments{
\item{old_names}{Disallowed IDs}

\item{n}{Number of IDs to generate}

\item{max_tries}{Max number of attempts to create IDs that do not intersect
with \code{old_names}}

\item{id_fun}{A function with a single argument \code{n} that generates random
IDs. A value of \code{NULL} defaults to \code{\link[ids:adjective_animal]{ids::adjective_animal()}} if available and
\code{sample_letters} otherwise.}

\item{ctor}{Function (either a string, a function or number used to index
the call stack}

\item{ctor_pkg}{The package where \code{ctor} is defined (either a string or
\code{NULL} which will use the function environment)}

\item{x}{Character vector to transform}

\item{replace, with}{Mapped to \code{\link[base:gsub]{base::gsub()}}}
}
\value{
A character vector of length \code{n} where each entry contains \code{length}
characters (all among \code{chars} and start/end with \code{prefix}/\code{suffix}), is
guaranteed to be unique and not present among values passed as \code{old_names}.
}
\description{
Randomly generated unique IDs are used throughout the package, created by
\code{rand_names()}. If random strings are required that may not clash with a set
of existing values, this can be guaranteed by passing them as \code{old_names}.
A \code{\link[=blockr_option]{blockr_option()}} \code{rand_id} can be set to swap out the function responsible
for ID generation.
}
\examples{
rand_names()
rand_names(n = 5L)
rand_names(id_fun = sample_letters)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugin-serdes.R
\name{preserve_board}
\alias{preserve_board}
\alias{preserve_board_server}
\alias{restore_board}
\alias{preserve_board_ui}
\alias{serialize_board}
\title{Serialization plugin module}
\usage{
preserve_board(server = preserve_board_server, ui = preserve_board_ui)

preserve_board_server(id, board, ...)

restore_board(x, new, result, ..., session = get_session())

preserve_board_ui(id, board)

serialize_board(x, blocks, id = NULL, ..., session = get_session())
}
\arguments{
\item{server, ui}{Server/UI for the plugin module}

\item{id}{Namespace ID}

\item{board}{The initial \code{board} object}

\item{...}{Extra arguments passed from parent scope}

\item{x}{The current \code{board} object}

\item{new}{Serialized (list-based) representation of the new board}

\item{result}{A \code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} to hold the new board object}

\item{session}{Shiny session}

\item{blocks}{Block state reactive values}
}
\value{
A plugin container inheriting from \code{preserve_board} is returned by
\code{preserve_board()}, while the UI component (e.g. \code{preserve_board_ui()}) is
expected to return shiny UI (i.e. \code{\link[shiny:tagList]{shiny::tagList()}}) and the server
component a \code{\link[shiny:reactiveVal]{shiny::reactiveVal()}} evaluating to \code{NULL} or the \code{board} to
restore.
}
\description{
Board state can be preserved by serializing all contained objects and
restored via de-serialization. This mechanism can be used to power features
such as save/restore (via download, as implemented in the default
\code{preserve_board} plugin), but more refined user experience is conceivable
in terms of undo/redo functionality and (automatic) saving of board state.
Such enhancements can be implemented in a third-party \code{preserve_board}
module.
}
\details{
The \code{session$reload()} that a restore triggers is handled by the app's
\code{\link[=board_loader]{board_loader()}} (passed to \code{\link[=serve]{serve()}}), not by the plugin -- the plugin
server simply returns the \code{board} to restore and never reloads.
}
